// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
/*
   TODO:
   - remove dependency on sphere.hh
 */

#ifndef DUNE_TRIANGULATION_HH
#define DUNE_TRIANGULATION_HH

#include <list>
#include <vector>
#include <memory>
#include <map>

#include "vtriangulation.hh"
#include "2dtriangulation/sphere.hh"
#include <dune/common/version.hh>
#if !DUNE_VERSION_NEWER(DUNE_COMMON,2,4)
#include <dune/common/static_assert.hh>
#endif
#include <dune/common/shared_ptr.hh>
#include <dune/common/deprecated.hh>
#include <dune/udg/common/cuboid.hh>
#include <dune/udg/common/subentity.hh>
#include <dune/udg/common/subintersection.hh>

template <class GridViewT>
class SubTriangulation2D : public VirtualSubTriangulation<GridViewT> {
#if DUNE_VERSION_NEWER(DUNE_COMMON,2,4)
  static_assert( (GridViewT::Grid::dimension==2), "Only 2d Grid supported" );
#else
  dune_static_assert( (GridViewT::Grid::dimension==2), "Only 2d Grid supported" );
#endif
#if DUNE_VERSION_NEWER(DUNE_COMMON,2,4)
  static_assert( (GridViewT::Grid::dimensionworld==2), "Only 2d Grid supported" );
#else
  dune_static_assert( (GridViewT::Grid::dimensionworld==2), "Only 2d Grid supported" );
#endif
public:
  enum { dim=2 };
  enum { dimworld=2 };

  typedef GridViewT GridView;
  typedef typename GridView::Grid Grid;
  typedef typename Grid::ctype ct;
  typedef typename Grid::LevelIntersectionIterator IntersectionIterator;
  typedef typename Grid::template Codim<0>::EntityPointer EntityPointer;
  typedef typename Grid::template Codim<0>::Entity Entity;

  typedef Dune::UDG::EntityPart< Grid > EntityPart;
  typedef Dune::UDG::IntersectionPart< Grid > IntersectionPart;
  typedef Dune::GridCuboid< Grid > BoundingBox;

  typedef Dune::UDG::Sphere<dim,ct> Sphere;
  typedef Dune::UDG::Parable<dim,ct> Parable;
  typedef Dune::UDG::Line<dim,ct> Line;
  typedef Dune::UDG::InnerDomain<dim,ct> InnerDomain;

private:
  // store a list of InnerDomains for each codim<0>::entity
  typedef Dune::GridCuboid< Grid > GridCuboid;
  typedef Dune::shared_ptr<const InnerDomain> InnerDomainPointer;
  typedef std::vector<InnerDomainPointer> InnerDomainList;
  mutable std::map<GridCuboid, InnerDomainList > domainmapping;

  typedef Dune::FieldVector<double,2> Point;
  // List of InnerDomains
  InnerDomainList domains;
  // get the sphere intersecting with this codim<0>::entity
  const InnerDomainList &
  getInnerDomains(const GridCuboid & rect) const;
  // create subtriangulation for a rectangle
  void createSubGeometries(GridCuboid& rect,
                           GridCuboid& bbox,
                           std::list<EntityPart> & geometries) const;
  void createRectangles(const EntityPointer & e,
                        std::list<GridCuboid> & rectangles) const;
  void createSubRectangles(GridCuboid& rect,
                           std::list<GridCuboid> & rectangles);
  void createSubRectanglesTop(GridCuboid& rect,
                              std::list<GridCuboid> & rectangles);
  void createSubRectanglesBottom(GridCuboid& rect,
                                 std::list<GridCuboid> & rectangles);
  void createSubRectanglesLeft(GridCuboid& rect,
                               std::list<GridCuboid> & rectangles);
  void createSubRectanglesRight(GridCuboid& rect,
                                std::list<GridCuboid> & rectangles);
  void createSubEdges(IntersectionPart & e,
                      std::list<IntersectionPart> & edges);
  void createSubBoundaries(IntersectionPart & b,
                           std::list<IntersectionPart> & boundaries);
  void createCrossSubBoundaries(IntersectionPart& b, int vertexId,
                                const InnerDomain& omega,
                                const InnerDomainList & domains,
                                std::list<IntersectionPart> & boundaries);
  ///////////////////////
  void createGeometries(const EntityPointer & e,
                        std::list<EntityPart> & geometries) const;
  void createEdges(const EntityPointer & e,
                   std::list<IntersectionPart> & edges) const;
  void createBoundaries(const EntityPointer & e,
                        std::list<IntersectionPart> & boundaries) const;
  ///////////////////////
  bool bisect(const GridCuboid & r, const Point & p,
              std::list<GridCuboid> & rlist) const;
  inline void getPoints(const GridCuboid& rect,
                        const InnerDomain& omega,
                        std::vector<Point> & points) const;
  inline void getCrossPoints(const GridCuboid& rect,
                             const InnerDomain& omega,
                             std::vector<Point> & points) const;
  inline int chooseCase(const GridCuboid& rect,
                        const InnerDomain& omega,
                        std::vector<bool> & missing) const;
  inline bool entartet(Point & A, Point & B, Point & C) const;
  inline int push_rect(const GridCuboid & r, std::list<GridCuboid> & rlist) const;
  inline Point coord(const GridCuboid & r, int i) const
  {
    static const int corners[] = { 0, 2, 3, 1 };
    return r.coord(corners[i%4]);
  }
  SubTriangulation2D(const SubTriangulation2D &);
public:
  SubTriangulation2D(const GridView & _g, int i=0) : g(_g), rect_sub(true) { domains.clear(); domainmapping.clear(); };
  SubTriangulation2D() : rect_sub(true) { domains.clear(); domainmapping.clear(); };
  const GridView & gridView() const { return g; }
  // register boundaries
  void loadInnerDomains(const char * fname);
  void push_sphere(const Sphere & s) DUNE_DEPRECATED {
    addInnerDomain(s);
  };
  void addInnerDomain(const InnerDomain & d) {
    domains.push_back(Dune::stackobject_to_shared_ptr(d));
    domainmapping.clear();
  }
  void addInnerDomain(InnerDomainPointer p) {
    domains.push_back(p);
    domainmapping.clear();
  };
  const InnerDomainList & get_innerDomains() const {
    return domains;
  };
  InnerDomainList & get_innerDomains() {
    domainmapping.clear();   // data might change
    return domains;
  };
  // enable / disable cube refinement
  bool & forceRefinement() { return rect_sub; };
  // create Entities
  void create_entity_parts(const EntityPointer & e,
                           std::list<EntityPart> & geometries) const
  {
    createGeometries(e,geometries);
  }
  void create_intersection_parts(const EntityPointer & e,
                                 std::list<IntersectionPart> & intersections) const
  {
    createEdges(e, intersections);
    createBoundaries(e, intersections);
  }

  bool isInDomain(const Entity & e, const int & domain_index) const
  {
    if (domain_index != 0)
      return false;
    std::list<EntityPart> g2;
    createGeometries(e,g2);
    return g2.size() > 0;
  }
  ct getEdgeNorm(const IntersectionPart & intersection) const
  {
#ifndef NDEBUG
    #warning ERROR!
#endif
    return 1.0;
  }

  // get the bounding box for this codim<0>::entity
  const BoundingBox BBox(const unsigned int domain_index, const Entity & e) const;
private:
  const GridView & g;
  bool rect_sub;
};

#include "2dtriangulation/2dtriangulation.cc"

#endif
