// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#include <config.h>

#include <cmath>
#include <stdio.h>

#include <assert.h>

#include <string>
#include <iterator>

#include "../2dtriangulation.hh"
#include <dune/common/exceptions.hh>
#include <dune/common/fvector.hh>

enum whatWeCouldDoWithEmptyElements
{
  vierteln, halbieren, nichts
};

static const enum whatWeCouldDoWithEmptyElements
  whatToDoWithEmptyElements = nichts;
static const enum whatWeCouldDoWithEmptyElements
  whatToDoWithEmptyQuadrilaterals = nichts;

static const Dune::FieldVector<double, 1> zeroPoint(0.0);

static double reference_length;
static double eq_epsilon=1e-6;

namespace Dune {
  template<class K, int n>
  inline bool coord_eq(const FieldVector<K,n> & c1,
                       const FieldVector<K,n> & c2)
  {
    double ref = reference_length * eq_epsilon;
    return (c1-c2).two_norm() < ref;
  }

  template<class K, int n>
  inline bool coord_lt(const FieldVector<K,n> & c1,
                       const FieldVector<K,n> & c2)
  {
    for (int i=0; i<n; i++)
    {
      if (c1[i] != c2[i])
        return (c1[i] < c2[i]);
    }
    return false;
  }
}

template <class GridView>
void
SubTriangulation2D<GridView>::getPoints(const GridCuboid& rect,
                                        const InnerDomain& omega,
                                        std::vector<Point> & points) const
{
  points.clear();
  for (int i=0; i<4; i++)
  {
    // i-th corner in rect
    if (! omega.isInside(coord(rect,i)))
      points.push_back(coord(rect,i));
    // i-th edge of rect
    Line l (coord(rect,i), coord(rect,i+1));
    omega.getIntersections(l, points);
  }
}

template <class GridView>
void
SubTriangulation2D<GridView>::getCrossPoints(const GridCuboid& rect,
                                             const InnerDomain& omega,
                                             std::vector<Point> & points) const
{
  points.clear();
  for (int i=0; i<4; i++)
  {
    // i-th edge of rect
    Line l (coord(rect,i), coord(rect,i+1));
    omega.getIntersections(l, points);
  }
}

template <class GridView>
int
SubTriangulation2D<GridView>::chooseCase(const GridCuboid& rect,
                                         const InnerDomain& omega,
                                         std::vector<bool> & missing) const
{
  int missingVertices = 0;
  for (int a=0; a< 4; a++) {
    missing[a] = omega.isInside(coord(rect,a));
    if (missing[a]) missingVertices++;
  }
  return missingVertices;
}

template <class GridView>
bool
SubTriangulation2D<GridView>::entartet(Point & A, Point & B, Point & C) const
{
  bool e = false;
  A[0] += 1.0; A[0] -= 1.0;
  B[0] += 1.0; B[0] -= 1.0;
  C[0] += 1.0; C[0] -= 1.0;
  A[1] += 1.0; A[1] -= 1.0;
  B[1] += 1.0; B[1] -= 1.0;
  C[1] += 1.0; C[1] -= 1.0;
  ct epsilon = 1e-8 * (std::abs(A[0] - B[0]) + std::abs(A[0] - C[0])
                       + std::abs(A[1] - B[1]) + std::abs(A[1] - C[1]));
  if (std::abs(A[0]-B[0]) < epsilon &&
      std::abs(A[1]-B[1]) < epsilon)
    e = true;
  if (std::abs(C[0]-B[0]) < epsilon &&
      std::abs(C[1]-B[1]) < epsilon)
    e = true;
  if (std::abs(A[0]-C[0]) < epsilon &&
      std::abs(A[1]-C[1]) < epsilon)
    e = true;
  if (e)
  {
    std::cout << "Warning! Entartet" << std::endl
              << A << B << C << std::endl;
  }
  return false;
  return e;
}

template <class GridView>
const typename SubTriangulation2D<GridView>::InnerDomainList &
SubTriangulation2D<GridView>::getInnerDomains(const GridCuboid & rect) const
{
  typename std::map<GridCuboid,InnerDomainList>::const_iterator
  rectDomains = domainmapping.find(rect);
  if (rectDomains != domainmapping.end())
  {
    return rectDomains->second;
  }
  // is there an inner domain intersecting with this rectangle?
  typename InnerDomainList::const_iterator intersectingDomain =
    domains.begin();
  typename InnerDomainList::const_iterator domEnd =
    domains.end();
  // add it to the local domain list
  InnerDomainList localdomains;
  for (; intersectingDomain != domEnd; ++intersectingDomain) {
    InnerDomainPointer omega = *intersectingDomain;
    // sphere inside the rectangle
    if (rect.isInside(omega->begin())) {
      localdomains.push_back(omega);
      continue;
    };
    // rectangle inside the sphere
    if (omega->isInside(rect.coord(0))) {
      localdomains.push_back(omega);
      continue;
    };
    // intersection
    std::vector<Point> points;
    getCrossPoints(rect, **intersectingDomain, points);
    if (points.size()) {
      localdomains.push_back(omega);
      continue;
    };
  }
  domainmapping[rect] = localdomains;
  return domainmapping[rect];
}

template <class GridView>
const typename SubTriangulation2D<GridView>::BoundingBox
SubTriangulation2D<GridView>::BBox(const unsigned int domain_index, const Entity & e) const
{
  EntityPointer E(e);
  GridCuboid cube(E);
  const InnerDomainList & intersectionSpheres = getInnerDomains(cube);
  if (intersectionSpheres.size() == 0)
    return cube;

  // get bounding box
  Point lu,ro;
  lu[0] = 1e100;
  lu[1] = 1e100;
  ro[0] = -1e100;
  ro[1] = -1e100;
  // loop over all spheres intersecting with this element
  for (typename InnerDomainList::const_iterator
       si = intersectionSpheres.begin();
       si != intersectionSpheres.end(); ++si)
  {
    // intersect with this sphere
    const InnerDomain & sp = *(*si);
    std::vector<bool> missing(4, false);
    chooseCase(cube, sp, missing);
    std::vector<Point> p; p.clear();
    getPoints(cube, sp, p);

    for (std::vector<Point>::iterator i=p.begin(); i!=p.end(); ++i)
    {
      if((*i)[0] < lu[0]) lu[0]=(*i)[0];
      if((*i)[1] < lu[1]) lu[1]=(*i)[1];
      if((*i)[0] > ro[0]) ro[0]=(*i)[0];
      if((*i)[1] > ro[1]) ro[1]=(*i)[1];
    }
  }
  return GridCuboid(E, lu, ro, Dune::CuboidMode::corners);
}

template <class GridView>
void SubTriangulation2D<GridView>::createSubGeometries(GridCuboid& rect,
                                                       GridCuboid& bbox,
                                                       std::list<EntityPart> & geometries) const
{
  // is there a sphere intersecting with this rectangle?
  const InnerDomainList & intersectionSpheres = getInnerDomains(rect);
  if (intersectionSpheres.size() == 0)
  {
    switch(whatToDoWithEmptyElements)
    {
    case vierteln :
    {
      Point m = rect.center();
      geometries.push_back(
        EntityPart(bbox, coord(rect,1), coord(rect,0), m));
      geometries.push_back(
        EntityPart(bbox, coord(rect,2), coord(rect,1), m));
      geometries.push_back(
        EntityPart(bbox, coord(rect,3), coord(rect,2), m));
      geometries.push_back(
        EntityPart(bbox, coord(rect,0), coord(rect,3), m));
      break;
    }
    case halbieren :
    {
      geometries.push_back(
        EntityPart(bbox, coord(rect,1), coord(rect,0), coord(rect,2)));
      geometries.push_back(
        EntityPart(bbox, coord(rect,2), coord(rect,0), coord(rect,3)));
      break;
    }
    case nichts :
    {
      geometries.push_back(EntityPart(rect));
      break;
    }
    }
    return;
  }

  // now we must be so small that we only have one intersecting sphere
#ifndef NDEBUG
  if (intersectionSpheres.size() > 1)
    DUNE_THROW(Dune::MathError,"Wrong triangulation: " << rect << " has " << intersectionSpheres.size() << " intersecting domains: " << *intersectionSpheres[0] << *intersectionSpheres[1]);
#endif

  // intersect with this sphere
  const InnerDomain & sp = *(intersectionSpheres.front());

  std::vector<bool> missing(4, false);
  chooseCase(rect, sp, missing);
  std::vector<Point> points; points.clear();
  getPoints(rect, sp, points);
  // erase duplicate points
  reference_length = (rect.coord(3) - rect.coord(0)).one_norm();
  std::vector<Point>::iterator new_end =
    std::unique(points.begin(), points.end(), Dune::coord_eq<double,2>);
  points.erase(new_end, points.end());
  if (points.size())
    if (Dune::coord_eq(points[0],points[points.size()-1])) points.pop_back();
  // select intersection type
  switch (points.size()) {
  case 4 :
    switch(whatToDoWithEmptyQuadrilaterals)
    {
    case vierteln :
    {
      Point m = rect.center();
      geometries.push_back(
        EntityPart(bbox, coord(rect,1), coord(rect,0), m));
      geometries.push_back(
        EntityPart(bbox, coord(rect,2), coord(rect,1), m));
      geometries.push_back(
        EntityPart(bbox, coord(rect,3), coord(rect,2), m));
      geometries.push_back(
        EntityPart(bbox, coord(rect,0), coord(rect,3), m));
      break;
    }
    case halbieren :
    {
      geometries.push_back(
        EntityPart(bbox, coord(rect,1), coord(rect,0), coord(rect,2)));
      geometries.push_back(
        EntityPart(bbox, coord(rect,2), coord(rect,0), coord(rect,3)));
      break;
    }
    case nichts :
    {
      geometries.push_back(
        //         EntityPart(bbox,
        //                   coord(rect,0), coord(rect,3), coord(rect,1), coord(rect,2)));
        EntityPart(bbox,
                   coord(rect,0), coord(rect,2)));
      break;
    }
    }
    return;
  case 3 :
  {
    int id=0;
    for (int a=0; a<3; a++) {
      if (sp.isInsideInc(points[a])) {
        id += 1<<a;
      }
    }
    switch (id) {
    case 3 : // sphere intersects in 0,1
      assert (! entartet(points[2], points[0], points[1]));
      geometries.push_back(
        EntityPart(bbox, points[2], points[1], points[0], sp));
      break;
    case 5 : // sphere intersects in 0,2
      assert (! entartet(points[1], points[2], points[0]));
      geometries.push_back(
        EntityPart(bbox, points[1], points[0], points[2], sp));
      break;
    case 6 : // sphere intersects in 1,2
      assert (! entartet(points[0], points[1], points[2]));
      geometries.push_back(
        EntityPart(bbox, points[0], points[2], points[1], sp));
      break;
    case 1 : // just points of contact
    case 2 :
    case 4 :
      assert (! entartet(points[0], points[1], points[2]));
      geometries.push_back(
        EntityPart(bbox, points[1], points[0], points[2]));
      break;
    default :
      std::cout << "Id = " << id << std::endl;
      std::cout << "InnerDomain = " << sp << std::endl;
      for (int i=0; i<3; i++) {
        std::cout << "points[" << i << "] = ("
                  << points[i][0] << ", " << points[i][1] << ") ... ";
        if (id & (1<<i))
          std::cout << "crosspoint\n";
        else
          std::cout << "no crosspoint\n";
      }
      geometries.push_back(
        EntityPart(bbox, coord(rect,0), coord(rect,3)));
      DUNE_THROW(Dune::Exception, "imposible triangulation ARGH!");
    }
  }
  break;
  case 2 :
  case 1 :
  case 0 :
    // do nothing :-)
    break;
  default :
    std::cout << "Don't know how to handle a " << points.size() << "-gon : "
              << std::endl << "reference_length = " << reference_length
              << std::endl << "\\path ";
    std::copy(points.begin(), points.end(),
              std::ostream_iterator<Point>(std::cout, " "));
    std::cout << std::endl;
    break;
    DUNE_THROW(Dune::NotImplemented,
               "Don't know how to handle a " << points.size() << "-gon : "
                                             << std::endl << "reference_length = " << reference_length
                                             << std::endl << "\\path ";
               std::copy(points.begin(), points.end(),
                         std::ostream_iterator<Point>(th__out," ")));
    break;
  }
}

template <class GridView>
int
SubTriangulation2D<GridView>::push_rect(const GridCuboid & r,
                                        std::list<GridCuboid> & rlist) const
{
  ct eps = 1e-10;
  Point h = r.coord(3) - r.coord(0);
  if (h[0] > eps && h[1] > eps)
  {
    rlist.push_back(r);
    return 1;
  }
  return 0;
}

template <class GridView>
bool
SubTriangulation2D<GridView>::bisect(const GridCuboid & r, const Point & p,
                                     std::list<GridCuboid> & rlist) const
{
#ifdef DEBUG_2DST
  std::cout << "BISECT " << r << " at " << p << std::endl;
#endif
  if (! r.isInside(p)) return false;

#ifdef DEBUG_2DST
  std::cout << "\tis inside\n";
#endif

  reference_length = (r.coord(3) - r.coord(0)).one_norm();
  for (int i=0; i<4; i++)
  {
    if (Dune::coord_eq(coord(r,i), p))
    {
#ifdef DEBUG_2DST
      std::cout << "\tp == corner " << i << " ... " << coord(r,i) << "\n";
#endif
      return false;
    }
  }

#ifdef DEBUG_2DST
  std::cout << "\tno corner\n";
#endif

  /*
     _________
   |   |     | r1: ( r.x | r.y ) [ p.x-r.x | p.y-r.y ]
   | r2|  r3 | r2: ( r.x | p.y ) [ p.x-r.x | r.y+r.hy-p.y ]
   |   |     | r3: ( p.x | p.y ) [ r.x+r.hx-p.x | r.y+r.hy-p.y ]
   ||---+-----| r4: ( p.x | r.y ) [ r.x+r.hx-p.x | p.y-r.y ]
   | r1|  r4 |
   ||___|_____|

   */

  int count=0;
  for (int i=0; i<4; i++)
    count += push_rect(
      GridCuboid(r.entity(),p,r.coord(i), Dune::CuboidMode::corners),
      rlist);

#ifdef DEBUG_2DST
  std::cout << "\t" << count << " sub rectangles\n";
#endif

  assert(count > 1);
  return true;
}

template <class GridView>
void SubTriangulation2D<GridView>::createRectangles(const EntityPointer & e,
                                                    std::list<GridCuboid> & rectangles) const
{
  std::list<GridCuboid> todo;
  todo.push_back(GridCuboid(e));

  while (todo.size() > 0)
  {
    bool was_bisected = false;
    // get current working target
    GridCuboid current = todo.back();
    todo.pop_back();
#ifdef DEBUG_2DST
    std::cout << "RECTANGLE" << current << "\n";
#endif
    // get all intersecting spheres
    const InnerDomainList & innerDomains = getInnerDomains(current);
    if (innerDomains.size() == 0)
    {
      rectangles.push_back(current);
      continue;
    }
    // find the first ,,important'' sphere
    typename InnerDomainList::const_iterator dit = innerDomains.begin();
    typename InnerDomainList::const_iterator dend = innerDomains.end();
    for (; dit!=dend; ++dit)
    {
      const InnerDomain & omega = **dit;
      std::vector<Point> points;

      // bisect at interesting points
#ifdef DEBUG_2DST
      std::cout << "CHECK SPECIAL POINTS\n";
#endif
      omega.specialPoints(points);
      std::vector<Point>::iterator pit = points.begin();
      std::vector<Point>::iterator pend = points.begin();
      for (; pit != points.end(); ++pit)
      {
        was_bisected = bisect(current, *pit, todo);
        if (was_bisected) break;
      }
      if (was_bisected) break;

      // bisect at cross points
      // getCrossPoints(current, omega, points);
      getPoints(current, omega, points);
      pit = points.begin();
      pend = points.begin();
#ifdef DEBUG_2DST
      std::cout << "CHECK CROSSPOINTS\n";
#endif
      for (; pit != points.end(); ++pit)
      {
        was_bisected = bisect(current, *pit, todo);
        if (was_bisected) break;
      }
      if (was_bisected) break;
    }
    if (!was_bisected)
    {
      if (innerDomains.size() == 1)
      {
        std::vector<Point> points;
        getPoints(current, *innerDomains[0], points);
        // erase duplicate points
        reference_length = (current.coord(3) - current.coord(0)).one_norm();
        std::vector<Point>::iterator new_end =
          std::unique(points.begin(), points.end(), Dune::coord_eq<double,2>);
        points.erase(new_end, points.end());
        if (points.size())
          if (Dune::coord_eq(points[0],points[points.size()-1])) points.pop_back();
        std::cout << std::flush;
        if(points.size() > 4)
        {
          std::cout << "WARNING!!! "
                    << "Don't know how to handle a " << points.size() << "-gon : "
                    << std::endl << "reference_length = " << reference_length
                    << std::endl << "\\path ";
          std::copy(points.begin(), points.end(),
                    std::ostream_iterator<Point>(std::cout," "));
          std::cout << std::endl;
        }
      }
      if (innerDomains.size() > 1)
      {
        // educated guess: split in the middle
        was_bisected = bisect(current, current.center(), todo);
        assert (was_bisected);
        continue;
      }
      rectangles.push_back(current);
    }
  }
}

template <class GridView>
void SubTriangulation2D<GridView>::createGeometries(const EntityPointer & e,
                                                    std::list<EntityPart> & geometries) const
{
  geometries.clear();
  std::list<GridCuboid> rectangles;
  GridCuboid bbox = BBox(0,*e);
  createRectangles(e, rectangles);
  for (typename std::list<GridCuboid>::iterator r=rectangles.begin();
       r!=rectangles.end(); ++r) {
    createSubGeometries(*r, bbox, geometries);
  }
}

template <class GridView>
void SubTriangulation2D<GridView>::createEdges(const EntityPointer & e,
                                               std::list<IntersectionPart> & edges) const
{
  edges.clear();

  GridCuboid rect = e;
  GridCuboid bbox = BBox(0,*e);

  reference_length = (bbox.coord(3) - bbox.coord(0)).one_norm();

  const InnerDomainList & innerDomains = getInnerDomains(rect);

  Point A, B;
  IntersectionIterator it = g.ibegin(*e);
  IntersectionIterator endit = g.iend(*e);
  //  assert(e->index() != -1);
  for(; it != endit; ++it)
  {
    if (it->neighbor())
    {
      A = it->geometry().corner(0);
      B = it->geometry().corner(1);
      if (A[0] == B[0] && A[1] == B[1]) {
        DUNE_THROW(Dune::Exception,
                   "A = " << A << " B = " << B << " A==B!!!\n");
      }
      else {
        Line l(A,B);
        std::vector<Point> points;

        // add all intersections and begin/end to points
        typename InnerDomainList::const_iterator dit = innerDomains.begin();
        typename InnerDomainList::const_iterator dend = innerDomains.end();
        bool Ainside = false;
        bool Binside = false;
        for (; dit != dend; ++dit)
        {
          const InnerDomain & omega = (**dit);
          if (omega.isInside(A)) Ainside = true;
          if (omega.isInside(B)) Binside = true;
          omega.getIntersections(l, points);
        }
        if (! Ainside) points.push_back(A);
        if (! Binside) points.push_back(B);

        // sort points
        std::sort(points.begin(), points.end(), Dune::coord_lt<double,2>);
        while  (points.size() > 1 && Dune::coord_eq(points[0],points[1]))
          points.erase(points.begin());
        while (points.size() > 1 &&
               coord_eq(points[points.size()-1],points[points.size()-2]))
          points.pop_back();

        // creat edges from pairs of points
        if (points.size() > 1)
        {
          if (points.size()%2 != 0)
          {
            DUNE_THROW(Dune::Exception, "points.size()%2 != 0 ("
                       << points.size() << ")\n";
                       std::copy(points.begin(), points.end(),
                                 std::ostream_iterator<Point>(th__out)));
          }
          GridCuboid bbox_n = BBox(0,*it->outside());
          for (size_t i=0; i<points.size(); i+=2)
          {
            Point a = points[i];
            Point b = points[i+1];
            IntersectionPart
            edge(bbox, bbox_n, a, b, it->unitOuterNormal(zeroPoint));
            edges.push_back(edge);
          }
        }
      }
    }
  }
}

template <class GridView>
void SubTriangulation2D<GridView>::createBoundaries(const EntityPointer & e,
                                                    std::list<IntersectionPart> & boundaries) const
{
  boundaries.clear();

  GridCuboid rect = e;
  GridCuboid bbox = BBox(0,*e);

  reference_length = (bbox.coord(3) - bbox.coord(0)).one_norm();

  const InnerDomainList & innerDomains = getInnerDomains(rect);

  Point A, B;
  IntersectionIterator it = g.ibegin(*e);
  IntersectionIterator endit = g.iend(*e);
  for(; it != endit; ++it)
  {
    if (it->boundary())
    {
      assert(it->geometry().corners() == 2);
      A = it->geometry().corner(0);
      B = it->geometry().corner(1);

      Line l(A,B);
      std::vector<Point> points;

      // add all intersections and begin/end to points
      typename InnerDomainList::const_iterator dit = innerDomains.begin();
      typename InnerDomainList::const_iterator dend = innerDomains.end();
      bool Ainside = false;
      bool Binside = false;
      for (; dit != dend; ++dit)
      {
        const InnerDomain & omega = (**dit);
        if (omega.isInside(A)) Ainside = true;
        if (omega.isInside(B)) Binside = true;
        omega.getIntersections(l, points);
      }
      if (! Ainside) points.push_back(A);
      if (! Binside) points.push_back(B);

      // sort points
      std::sort(points.begin(), points.end(), Dune::coord_lt<double,2>);
      while  (points.size() > 1 && coord_eq(points[0],points[1]))
        points.erase(points.begin());
      while (points.size() > 1 &&
             coord_eq(points[points.size()-1],points[points.size()-2]))
        points.pop_back();

      // creat boundaries from pairs of points
      if (points.size() > 1)
      {
        if (points.size()%2 != 0)
        {
          DUNE_THROW(Dune::Exception, "points.size()%2 != 0 ("
                     << points.size() << ")\n";
                     std::copy(points.begin(), points.end(),
                               std::ostream_iterator<Point>(th__out)));
        }
        for (size_t i=0; i<points.size(); i+=2)
        {
          Point a = points[i];
          Point b = points[i+1];
          IntersectionPart boundary(bbox, a, b, it->boundarySegmentIndex(), it->unitOuterNormal(zeroPoint));
          boundaries.push_back(boundary);
        }
      }
    }
  }

  std::list<GridCuboid> rectangles;
  createRectangles(e, rectangles);
  for (typename std::list<GridCuboid>::iterator r=rectangles.begin();
       r!=rectangles.end(); ++r) {
    const InnerDomainList & intersectionSpheres = getInnerDomains(*r);
    if (intersectionSpheres.size() == 0) continue;
    InnerDomainPointer sit = intersectionSpheres.front();
    std::vector<Point> points;
    getCrossPoints(*r, *sit, points);
    reference_length = (r->coord(3) - r->coord(0)).one_norm();
    std::sort(points.begin(), points.end(), Dune::coord_lt<double,2>);
    std::vector<Point>::iterator new_end =
      std::unique(points.begin(), points.end(), Dune::coord_eq<double,2>);
    points.erase(new_end, points.end());
    if (points.size())
      if (coord_eq(points[0],points[points.size()-1])) points.pop_back();
    switch (points.size())
    {
    case 0 : // no intersection
    case 1 : // point of contact
      break;
    case 2 : // real intersection
    {
      const InnerDomainList& domains = get_innerDomains();
      int domain_index = 0;
      for (domain_index = 0; domains[domain_index] != sit; domain_index++) ;
      boundaries.push_back(IntersectionPart(bbox, points[0], points[1],
                                            // /*bid*/99,
                                            99 + domain_index,
                                            *sit));
    }
    break;
    default : // error
      DUNE_THROW(Dune::Exception,
                 "Impossible IntersectionPart " << points.size() << *r << " -> ";
                 std::copy(points.begin(), points.end(),
                           std::ostream_iterator<Point>(th__out, " ")));
    }
  }
}

template <class GridView>
void SubTriangulation2D<GridView>::loadInnerDomains(const char * fname)
{
  FILE * fp;
  const size_t len = 2048;
  char * buffer = (char*) malloc(len);
  buffer[0] = 0;
  fp = fopen(fname, "r");
  if (fp == 0)
    DUNE_THROW(Dune::IOError, "Could not read " << fname);
  for(; !feof(fp); fgets(buffer, len, fp)) {
    char * line = buffer;
    // strip
    while(line[0] == ' ' || line[0] == '\t') line++;
    // skip epmty lines
    if (strlen(line) == 0) continue;
    // comments
    if (line[0] == '#') continue;
    // sphere
    {
      Point x;
      double r;
      if (3 == sscanf(line, " Sphere(%lf, %lf, %lf)", &x[0], &x[1], &r))
      {
        InnerDomainPointer p(new Sphere(x, r));
        domains.push_back(p);
        continue;
      }
    }
    // parable
    {
      Point x;
      if (2 == sscanf(line, " Parable (%lf, %lf)", &x[0], &x[1]))
      {
        InnerDomainPointer p(new Parable(x));
        domains.push_back(p);
        continue;
      }
    }
    // throw
    DUNE_THROW(Dune::Exception, "Error parsing line \"" << line << "\"");
  }
  if (buffer)
    free(buffer);
  fclose(fp);
  domainmapping.clear();
}
