// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef INNERDOMAIN_HH
#define INNERDOMAIN_HH

#include <dune/common/fvector.hh>

#include <vector>
#include <iostream>
#include <limits>

namespace Dune {
  namespace UDG  {

    template <int dim, class ct>
    class Line;

    template <class ct>
    class Line<2,ct>
    {
    private:
      enum { dim=2 };
    public:
      Dune::FieldVector<ct, dim> c;
      Dune::FieldVector<ct, dim> d;
      Line (ct _x, ct _y, ct _hx, ct _hy)
      { c[0] = _x; c[1] = _y; d[0] = _hx; d[1] = _hy; }
      Line (Dune::FieldVector<ct, dim> a, Dune::FieldVector<ct, dim> b) :
        c(a), d(b-a) {}
      int dir() const {
        if (d[1] < 1e-14 && d[0] > 0) return 0;
        if (d[0] < 1e-14 && d[1] > 0) return 1;
        //     DUNE_THROW(Dune::Exception,
        //                "lines must be parallel to the coordinate axis");
        return 1;
      }
      bool isInside(Dune::FieldVector<ct, dim> p) const {
        double eps = std::numeric_limits<ct>::epsilon();
        switch(dir())
        {
        case 0 :
        {
          // x-axis
          assert(p[1] == c[1]);
          return (p[0] - c[0] > eps &&
                  p[0] - (c[0]+d[0]) < -eps);
        }
        case 1 :
        {
          // y-axis
          assert(p[0] == c[0]);
          return (p[1] - c[1] > eps &&
                  p[1] - (c[1]+d[1]) < -eps);
        }
        }
      }
      bool isInsideInc(Dune::FieldVector<ct, dim> p) const {
        double eps = std::numeric_limits<ct>::epsilon();
        switch(dir())
        {
        case 0 :
        {
          // x-axis
          assert(p[1] == c[1]);
          return (p[0] - c[0] >= -eps &&
                  p[0] - (c[0]+d[0]) <= eps);
        }
        case 1 :
        {
          // y-axis
          assert(p[0] == c[0]);
          return (p[1] - c[1] >= -eps &&
                  p[1] - (c[1]+d[1]) <= eps);
        }
        }
        throw (-1);
      }
    };

    template<int dim, class ct>
    class InnerDomain {
    public:
      enum { dimworld=dim };
      virtual ~InnerDomain() {};
      virtual void polynom(const Dune::FieldVector<ct,dim> & A,
                           const Dune::FieldVector<ct,dim> & B, int & side,
                           Dune::FieldVector<ct,3> & a,
                           Dune::FieldVector<ct,3> & b) const = 0;
      virtual void specialPoints(std::vector<Dune::FieldVector<ct,dim> > & points) const = 0;
      virtual const Dune::FieldVector< ct, dimworld >
      normal (const Dune::FieldVector< ct, dimworld > & p) const = 0;
      virtual const Dune::FieldVector<ct,dim> & begin() const = 0;
      virtual bool isInside(const Dune::FieldVector<ct,dim> & p) const = 0;
      virtual bool isInsideInc(const Dune::FieldVector<ct,dim> & p) const = 0;
      virtual void getIntersections(const Line<dim,ct>& l, std::vector<Dune::FieldVector<ct,dim> >& points) const = 0;
      virtual void print(std::ostream& s) const = 0;
    };

    template<int dim, class ct>
    inline std::ostream& operator<< (std::ostream& s, const InnerDomain<dim, ct>& a)
    {
      a.print(s);
      return s;
    }

    template<int dim, class ct>
    class Parable : public InnerDomain<dim,ct> {
      Dune::FieldVector<ct,dim> origin;
      inline ct sqr(ct x) const { return x*x; }
    public:
      enum { dimworld=dim };
      Parable() : origin(0) {}
      Parable(const Dune::FieldVector<ct,dim> o) : origin(o) {}
      virtual void polynom(const Dune::FieldVector<ct,dim> & _A,
                           const Dune::FieldVector<ct,dim> & _B, int & side,
                           Dune::FieldVector<ct,3> & a,
                           Dune::FieldVector<ct,3> & b) const
      {
        Dune::FieldVector<ct,dim> A = _A;
        Dune::FieldVector<ct,dim> B = _B;
        A -= origin;
        B -= origin;
        // x
        a[0] = A[0]+origin[0];
        a[1] = B[0]-A[0];
        a[2] = 0;
        // y
        b[0] = A[1]+origin[1];
        b[1] = 2*A[0]*B[0] - 2*A[1];
        b[2] = B[1] - 2*A[0]*B[0] + A[1];
        // it's bended downwards
        side = -1;
      };
      virtual ~Parable() {};
      virtual void specialPoints(std::vector<Dune::FieldVector<ct,dim> > & points) const {
        points.push_back(origin);
      };
      virtual const Dune::FieldVector< ct, dimworld >
      normal (const Dune::FieldVector< ct, dimworld > & p) const
      {
        Dune::FieldVector< ct, dimworld > n;
        n[0] = -2/5/(p[0]-origin[0]);
        n[1] = 1/5/(p[0]-origin[0]);
        return n;
      };
      virtual const Dune::FieldVector<ct,dim> & begin() const
      {
        return origin;
      };
      virtual bool isInside(const Dune::FieldVector<ct,dim> & p) const
      {
        Dune::FieldVector<ct,dim> a = p;
        a -= origin;
        return a[1] > sqr(a[0]);
      };
      virtual bool isInsideInc(const Dune::FieldVector<ct,dim> & p) const
      {
        Dune::FieldVector<ct,dim> a = p;
        a -= origin;
        return a[1] >= sqr(a[0]) - 2*a[0]*std::numeric_limits<ct>::epsilon();
      };
      virtual void getIntersections(const Line<dim,ct>& l, std::vector<Dune::FieldVector<ct,dim> >& points) const
      {
        Dune::FieldVector<ct,dim> p;
        switch(l.dir())
        {
        case 0 :
          // x-axis
          p[0] = sqrt(l.c[1] - origin[1]) + origin[0];
          p[1] = l.c[1];
          if (l.isInsideInc(p))
          {
            points.push_back(p);
          }
          return;
        case 1 :
          // y-axis
          p[0] = l.c[0];
          p[1] = sqr(l.c[0] - origin[0]) + origin[1];
          if (l.isInsideInc(p))
          {
            points.push_back(p);
          }
          return;
        };
      };
      virtual void print(std::ostream& s) const {};
    };

  } // end namespace UDG
} // end namespace DUNE

#endif // INNERDOMAIN_HH
