// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef ENTITY_PART_HH
#define ENTITY_PART_HH

#include <cstdarg>
#include <list>
#include <dune/udg/common/cuboid.hh>
#include <dune/udg/common/geometryinterface.hh>

namespace Dune {
  namespace UDG {

    /** \brief Traits class for entity parts */
    template< typename GV >
    struct EntityPartTraits
    {

      //! Grid view containing this part
      typedef GV GridView;
      //! Field type of domain coordinates
      typedef typename GridView::ctype DomainField;
      //! Local coordinate
      typedef Dune::FieldVector<DomainField,GV::dimension>
      LocalCoordinate;
      //! Global coordinate
      typedef Dune::FieldVector<DomainField,GV::dimension>
      GlobalCoordinate;

      //! Dimension of computational domain
      enum { dimension=GV::dimension };

      //! Geometry for entity parts
      typedef GeometryInterface<DomainField,dimension,dimension> Geometry;

      //! Bounding box for shape functions
      typedef Dune::GridCuboid<typename GridView::Grid> BoundingBox;

      typedef typename GV::template Codim<0>::Entity Entity;

      //! Domain tag
      typedef std::vector<int> DomainTag;
    };

    /** \brief Interface for the geometry engine of an entity part. */
    template <typename GV>
    class EntityPartGeometryInterface
    {
    public:
      //! Traits class
      typedef EntityPartTraits<GV> Traits;

      /** \brief Access the geometry mapping from the local domain of
         the entity part to the local domain of the bounding box. */
      virtual const typename Traits::Geometry& geometryInCell () const = 0;

      /** \brief Access the geometry mapping from the local domain of
         the entity part to the local domain of the bounding box. */
      virtual const Dune::shared_ptr<typename Traits::Geometry>
      geometryInCellPointer () const = 0;

      /** \brief Access the geometry mapping from the local domain of
         the entity part to the local domain of the entity. */
      virtual const typename Traits::Geometry& geometryInEntity () const = 0;

      /** \brief Access the geometry mapping from the local domain of
         the entity part to the local domain of the entity. */
      virtual const Dune::shared_ptr<typename Traits::Geometry>
      geometryInEntityPointer () const = 0;

      //! Access the global geometry
      virtual const typename Traits::Geometry& geometry () const = 0;

      //! Access the entity part's home entity in level set mesh or in
      //! fundamental mesh if the entity part was constructed from a whole
      //! fundamental mesh cell
      virtual const typename Traits::Entity & entity() const = 0;

      //! Access the bounding box
      virtual const typename Traits::BoundingBox & boundingBox() const = 0;
    };

    /** \brief Interface for an entity part. */
    template <typename GV>
    class EntityPartInterface : public virtual EntityPartGeometryInterface<GV>
    {
    public:
      //! Traits class
      typedef EntityPartTraits<GV> Traits;

      //! Shared pointer on entity part
      typedef Dune::shared_ptr<EntityPartInterface> EntityPartPointer;
      //! List for entity parts
      typedef std::list< EntityPartPointer > EntityPartList;

      virtual ~EntityPartInterface() {}

      /** \brief Legacy function to be removed */
      virtual int domainTag () const = 0;

      /** \brief Domain index of this entity part */
      virtual int domainIndex () const = 0;

      /** \brief Index of this part within the entity */
      virtual int partIndex() const = 0;

      //! Set to the intersection part in the current layer
      virtual void setLayer(const size_t layer) const {}

      //! Return the current layer index
      virtual size_t currentLayer() const { return 0;}

      //! Set to the intersection part in the current layer
      virtual size_t resetLayer() const { return 0; }

      /** \brief Sign vector indicating the placement of this
          entity part relative to the interfaces separating
          the sub triangulation's domains.

          For each interface of the sub triangulation, this sign
          vector has one entry holding an element of {-1,0,1},
          where 0 indicates that this entity part is intersected
          by the corresponding interface.
       */
      virtual const typename Traits::DomainTag & newDomainTag () const = 0;

      /** \brief Volume of the domain cell */
      virtual typename Traits::DomainField volume() const = 0;

    };


    /** \brief Base class for implementations of
        IntersectionPartGeometryInterface .

        This call represents intersections with the generic DUNE
        geometries.

        \tparam GV The grid view for which this part was constructed
     */
    template <class GV>
    class DefaultEntityPartGeometry :
      virtual public EntityPartGeometryInterface<GV>
    {
    public:

      //! Traits class
      typedef EntityPartTraits<GV> Traits;

      //! Geometry type
      typedef DefaultGeometry<typename Traits::DomainField,
          Traits::dimension,
          Traits::dimension>
      GeometryImp;

      //! Constructor
      DefaultEntityPartGeometry
        (const GeometryImp & geometry,
        const typename Traits::Entity & entity,
        const typename Traits::BoundingBox & bbox) :
        geometryInCell_(computeGeometryInBoundingBox(geometry,bbox)),
        geometryInEntity_(computeGeometryInEntity(geometry,entity)),
        geometry_(geometry),
        entity_(entity),
        bbox_(bbox)
      {}

      //! Setup face geometry
      static GeometryImp
      createGeometry(const int size ...)
      {
        std::vector<typename Traits::GlobalCoordinate> coords(size);
        Dune::GeometryType type;
        va_list ap;
        va_start(ap,size);
        for(size_t c=0; c<coords.size(); ++c)
          coords[c] = *(va_arg(ap,typename Traits::GlobalCoordinate*));
#if DUNE_VERSION_NEWER(DUNE_GEOMETRY, 2, 6)
        return FaceGeometryImp(Dune::geometryTypeFromVertexCount(Traits::dimension, size), coords);
#else
        return FaceGeometryImp(type,coords);
#endif
      }

      /** \brief Access the geometry mapping from the local domain of
         the entity part to the local domain of the bounding box. */
      const typename Traits::Geometry& geometryInCell () const
      { return *geometryInCell_; }

      /** \brief Access the geometry mapping from the local domain of
         the entity part to the local domain of the bounding box. */
      const Dune::shared_ptr<typename Traits::Geometry>
      geometryInCellPointer () const
      { return geometryInCell_; }

      /** \brief Access the geometry mapping from the local domain of
         the entity part to the local domain of the entity. */
      const typename Traits::Geometry& geometryInEntity() const
      { return *geometryInEntity_; }

      /** \brief Access the geometry mapping from the local domain of
         the entity part to the local domain of the entity. */
      const Dune::shared_ptr<typename Traits::Geometry>
      geometryInEntityPointer () const
      { return geometryInEntity_; }

      //! Access the global geometry
      const typename Traits::Geometry& geometry () const
      { return geometry_; }

      //! Access the entity part's home entity in level set mesh or in
      //! fundamental mesh if the entity part was constructed from a whole
      //! fundamental mesh cell
      const typename Traits::Entity & entity() const
      { return entity_; }

      //! Access the bounding box
      const typename Traits::BoundingBox & boundingBox() const
      { return bbox_; }

    private:

      // Return a mapping that maps the local domain of the global
      // face geometry into the local domain of the bounding box.
      static GeometryImp*  computeGeometryInBoundingBox
        (const GeometryImp & geometry,
        const typename Traits::BoundingBox & bbox)
      {
        std::vector<typename Traits::GlobalCoordinate>
        cornersInBBox(geometry.corners());
        for(int c=0; c<geometry.corners(); ++c)
          cornersInBBox[c] = bbox.local(geometry.corner(c));
        return new GeometryImp(geometry.type(),cornersInBBox);
      }

      // Return a mapping that maps the local domain of the global
      // face geometry into the local domain of the father entity
      static GeometryImp*  computeGeometryInEntity
        (const GeometryImp & geometry,
        const typename Traits::Entity& entity)
      {
        std::vector<typename Traits::GlobalCoordinate>
        cornersInEntity(geometry.corners());
        for(int c=0; c<geometry.corners(); ++c)
          cornersInEntity[c] = entity.geometry().local(geometry.corner(c));
        return new GeometryImp(geometry.type(),cornersInEntity);
      }

      const Dune::shared_ptr<GeometryImp> geometryInCell_;
      const Dune::shared_ptr<GeometryImp> geometryInEntity_;
      const GeometryImp geometry_;
      const typename Traits::Entity entity_;
      const typename Traits::BoundingBox bbox_;
    };

    /** \brief Default implementation for entity part.

        \tparam EPG Implementation of EntityPartGeometryInterface.
     */
    template <class EPG>
    class DefaultEntityPart :
      public EPG,
      public EntityPartInterface<typename EPG::Traits::GridView>
    {
    public:
      //! Geometry engine
      typedef EPG EntityPartGeometry;
      //! Traits class
      typedef typename EPG::Traits Traits;

      //! Constructor
      DefaultEntityPart
        (const int partIndex,
        const EPG & geometryEngine,
        const int oldDomainTag,
        const typename Traits::DomainTag & domainTag,
        const int domainIndex,
        const typename Traits::DomainField volume) :
        EPG(geometryEngine),
        domainTag_(domainTag),
        oldDomainTag_(oldDomainTag),
        domainIndex_(domainIndex),
        partIndex_(partIndex),
        volume_(volume)
      {}

      /** \brief Volume of the domain cell */
      typename Traits::DomainField volume() const {return volume_;}

      /** \brief Sign vector indicating the placement of this
          entity part relative to the interfaces separating
          the sub triangulation's domains.

          For each interface of the sub triangulation, this sign
          vector has one entry holding an element of {-1,0,1},
          where 0 indicates that this entity part is intersected
          by the corresponding interface.
       */
      const typename Traits::DomainTag & newDomainTag () const {
        return domainTag_;
      }

      /** \brief Legacy function */
      int domainTag () const {
        return oldDomainTag_;
      }

      /** \brief Domain index of this entity part*/
      int domainIndex () const {
        return domainIndex_;
      }

      /** \brief Index of this part within the entity */
      int partIndex() const {
        return partIndex_;
      }

    private:
      typename Traits::DomainTag domainTag_;
      const int oldDomainTag_;
      int domainIndex_;
      int partIndex_;
      typename Traits::DomainField volume_;
    };


  }
}

#endif
