// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef GEOMETRY_HH
#define GEOMETRY_HH

#include <dune/common/fvector.hh>
#include <dune/common/fmatrix.hh>
#include <dune/geometry/type.hh>
#include <dune/common/version.hh>

#if DUNE_VERSION_NEWER(DUNE_COMMON,2,4)
#include <dune/localfunctions/monomial.hh>
#else
#include <dune/localfunctions/monom.hh>
#endif

template<int mydim>
struct MaxGeometryOrder
{
  enum { order = (mydim>2 ? mydim : 2) };
};

inline int getGeometryOrder(const Dune::GeometryType & t)
{
  if (t.isLine()) return 1;
  if (t.isSimplex()) return 1;
  if (t.isCube()) return t.dim();
  if (t.isPyramid()) return 2;
  if (t.isPrism()) return 2;
  DUNE_THROW(Dune::Exception,"Unknown GeometryType " << t);
}

template<int mydim>
struct GeometryParamSize
{
  enum { linear = mydim+1 };
  //   enum { quadratic = (mydim+1)*(mydim+2)/2 };
  //   enum { multilinear = 1<<mydim-1 /* 2^mydim - 1*/ };
  enum {
#if DUNE_VERSION_LTE(DUNE_LOCALFUNCTIONS, 2, 6)
    higherorder = Dune::MonomImp::Size<mydim, MaxGeometryOrder<mydim>::order>::val
#else
    higherorder =
      Dune::MonomialLocalBasis<double,double,mydim,
        MaxGeometryOrder<mydim>::order>::size()
#endif
  };
};

template<int mydim, int cdim, class ct>
class SubTriangulationGeometry {
  inline ct sqr(ct x) const { return x*x; }
public:
  //! @brief export geometry dimension
  enum { mydimension=mydim /*!< geometry dimension */ };
  //! @brief export coordinate dimension
  enum { coorddimension=cdim /*!< dimension of embedding coordsystem */ };
  //! @brief export geometry dimension
  enum { dimension=cdim /*!< geometry dimension */ };
  //! @brief export coordinate dimension
  enum { dimensionworld=cdim /*!< dimension of embedding coordsystem */ };

  //! @brief export polynomial order of geometric mapping
  enum { order = MaxGeometryOrder<mydim>::order };

  //! @brief define type used for coordinates in grid module
  typedef ct ctype;
  typedef Dune::FieldMatrix<ct, GeometryParamSize<mydim>::linear, cdim>
  LinearParamMatrix;
  typedef Dune::FieldMatrix<ct, GeometryParamSize<mydim>::higherorder, cdim>
  ParamMatrix;
private:
  typedef Dune::FieldVector<ct, cdim> GlobalVector;
  typedef Dune::FieldVector<ct, mydim> LocalVector;
  typedef Dune::FieldMatrix<ct, mydim, mydim> LocalMatrix;
  //  typedef Dune::FieldMatrix<ct, cdim, mydim> JacobianMatrix;
  typedef Dune::FieldMatrix<ct, cdim, cdim> JacobianMatrix;

  ParamMatrix c;
  Dune::GeometryType _type;
  bool islinear;
  mutable JacobianMatrix JinvT;
  ct detj;

#if DUNE_VERSION_NEWER(DUNE_COMMON,2,4)
  typedef Dune::MonomialLocalFiniteElement<ct,ct,mydim,order> LocalFiniteElement;
#else
  typedef Dune::MonomLocalFiniteElement<ct,ct,mydim,order> LocalFiniteElement;
#endif
  typedef typename LocalFiniteElement::Traits::LocalBasisType LocalBasis;
  mutable std::vector<typename LocalBasis::Traits::RangeType> evalBuffer;
  static inline const LocalBasis & localBasis()
  {
    static const LocalFiniteElement finiteElement(Dune::GeometryType(Dune::GeometryType::none, mydim));
    return finiteElement.localBasis();
  }
  void setupJacobiMatrix(const LocalVector& local) const;
  void setupInverseTransposed() const;
  ct setupDetJ() const;
public:
  typedef GlobalVector GlobalCoordinate;
  typedef LocalVector LocalCoordinate;

  SubTriangulationGeometry() {};
  SubTriangulationGeometry(const Dune::GeometryType & t,
                           const LinearParamMatrix & _c);
  SubTriangulationGeometry(const Dune::GeometryType & t,
                           const ParamMatrix & _c);
  Dune::GeometryType type () const { return _type; };
  const JacobianMatrix &
  jacobianInverseTransposed (const LocalCoordinate& local) const;
  ct integrationElement(const LocalCoordinate& local) const;
  const GlobalCoordinate global (const LocalCoordinate& local) const;
  bool isLinear() const { return islinear; };
  bool operator == (const SubTriangulationGeometry & g) const
  {
    if (detj != g.detj) return false;
    if (islinear != g.islinear) return false;
    if (_type != g._type) return false;
    for (int i=0; i<GeometryParamSize<mydim>::higherorder; i++)
      for (int j=0; j<cdim; j++)
        if (c[i][j] != g.c[i][j]) return false;
    return true;
  }
};

#include "geometry_impl.cc"

#endif // GEOMETRY_HH
