// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef GEOMETRYINTERFACE_HH
#define GEOMETRYINTERFACE_HH

#include <dune/common/version.hh>
#include <dune/common/shared_ptr.hh>

#if DUNE_VERSION_NEWER(DUNE_GEOMETRY,2,3)
#include <dune/geometry/multilineargeometry.hh>
#else
#include <dune/geometry/genericgeometry/geometry.hh>
#endif

namespace Dune {
  namespace UDG {

    /** \brief Interface class for geometry implementations as
        provided by the entity and intersection part.

        \tparam T Specialization of Dune::GenericGeometry::DefaultGeometryTraits
     */
    template < typename DF, int localDim, int globalDim >
    class GeometryInterface
    {

#if DUNE_VERSION_NEWER(DUNE_GEOMETRY,2,3)
      // We don't need the caching, but the uncached MultiLinearGeometry is not defined at all vertices
      typedef Dune::CachedMultiLinearGeometry<DF, localDim, globalDim> G;
#else
      typedef Dune::GenericGeometry::
      DefaultGeometryTraits<DF,localDim,globalDim> T;
      typedef Dune::GenericGeometry::
      BasicGeometry<localDim, T> G;
#endif

    public:

      /** For compatibility with grids which have dimw==dim, both
          dimensions are equal to the world dimension. */
      enum { dimension = globalDim,
             dimensionworld = globalDim };

      /** For compatibility with grids which have dimw==dim, both
          dimensions are equal to the world dimension. */
      enum { mydimension = localDim,
             coorddimension = globalDim };

      //! \brief Interface types @{
      typedef typename G::ctype ctype;
      typedef typename G::LocalCoordinate LocalCoordinate;
      typedef typename G::GlobalCoordinate GlobalCoordinate;
      typedef typename G::JacobianTransposed JacobianTransposed;
      // typedef typename G::Jacobian Jacobian;
      typedef typename G::JacobianInverseTransposed JacobianInverseTransposed;
      //! @}

      //! \brief Interface functions @{
      virtual Dune::GeometryType type() const = 0;
      virtual int corners() const = 0;
      virtual GlobalCoordinate corner(const int i) const = 0;
      virtual GlobalCoordinate global(const LocalCoordinate & local) const = 0;
      virtual LocalCoordinate   local(const GlobalCoordinate & global) const = 0;
      virtual GlobalCoordinate center() const = 0;
      virtual bool affine() const = 0;
      virtual ctype integrationElement(const LocalCoordinate & local) const = 0;
      virtual ctype volume() const = 0;
#if DUNE_VERSION_NEWER(DUNE_GEOMETRY,2,3)
      virtual JacobianTransposed jacobianTransposed
#else
      virtual const JacobianTransposed & jacobianTransposed
#endif
        (const LocalCoordinate & local) const = 0;
#if DUNE_VERSION_NEWER(DUNE_GEOMETRY,2,3)
      virtual JacobianInverseTransposed jacobianInverseTransposed
#else
      virtual const JacobianInverseTransposed & jacobianInverseTransposed
#endif
        (const LocalCoordinate & local) const = 0;
      //! @}

      virtual ~GeometryInterface() {}
    };


    /** \brief Combine the consecutive mapping of two geometries into
        a combined geometry.
     */
    template< typename Geometry, typename HostGeometry >
    void combineGeometries(const Geometry & guest, const HostGeometry & host,
                           const bool forward,
                           std::vector<typename Geometry::GlobalCoordinate> & coords)
    {
      typedef typename Geometry::GlobalCoordinate Coordinate;
      coords.resize(guest.corners());
      for(unsigned int v=0; v<coords.size(); ++v) {
        if(forward)
          coords[v] = host.global(guest.corner(v));
        else
          coords[v] = host.local(guest.corner(v));
      }
    }

    /** \brief Combine the consecutive mapping of two geometries into
        a combined geometry.
     */
    template< typename NewType, typename Geometry, typename HostGeometry >
    NewType combineGeometries(const Geometry & guest, const HostGeometry & host,
                               const bool forward = true)
    {
      typedef typename Geometry::GlobalCoordinate Coordinate;
      std::vector<Coordinate> coords(guest.corners());
      combineGeometries(guest,host,forward,coords);
      return NewType(guest.type(),coords);
    }
    template< typename Geometry, typename HostGeometry >
    Geometry combineGeometries(const Geometry & guest, const HostGeometry & host,
                               const bool forward = true)
    {
      return combineGeometries<Geometry,Geometry,HostGeometry>(guest,host,forward);
    }

    /** \brief Combine the consecutive mapping of two geometries into
        a combined geometry.
     */
    template< typename Geometry, typename HostGeometry, typename TargetGeometry >
    void
    combineGeometries(const Geometry & guest, const HostGeometry & host,
                      Dune::shared_ptr<TargetGeometry> & result,
                      const bool forward = true)
    {
      typedef typename Geometry::GlobalCoordinate Coordinate;
      std::vector<Coordinate> coords(guest.corners());
      combineGeometries(guest,host,forward,coords);
      result = Dune::shared_ptr<Geometry>(new Geometry(guest.type(),coords));
    }


    /** \brief Wrapper for the Dune::GenericGeometry::BasicGeometry
        with virtual base class.

        \tparam DF              Domain field type
        \tparam localDim        Dimension of domain
        \tparam globalDim       Dimension of range
     */
    template < typename DF, int localDim, int globalDim >
    class DefaultGeometry :
      public GeometryInterface<DF,localDim,globalDim>
    {
#if DUNE_VERSION_NEWER(DUNE_GEOMETRY,2,3)
      // We don't need the caching, but the uncached MultiLinearGeometry is not defined at all vertices
      typedef Dune::CachedMultiLinearGeometry<DF, localDim, globalDim> Imp;
#else
      typedef Dune::GenericGeometry::
      DefaultGeometryTraits<DF,localDim,globalDim> T;
      typedef Dune::GenericGeometry::
      BasicGeometry<localDim, T> Imp;
#endif

      //! Base class
      typedef GeometryInterface<DF,localDim,globalDim> Base;


      template<typename Geometry>
      static
      std::vector<typename Base::GlobalCoordinate>
      getCorners(const Geometry & geometry){
        std::vector<typename Base::GlobalCoordinate> corners(geometry.corners());
        for(size_t c=0; c<corners.size(); ++c)
          corners[c] = geometry.corner(c);
        return corners;
      }

    public:

      //! Constructor
      DefaultGeometry
        (const Dune::GeometryType type,
        const std::vector<typename Base::GlobalCoordinate> & coords)
        : imp_(type,coords)
      {}

      //! Constructor
      template <typename SomeGeometry>
      DefaultGeometry
        (const SomeGeometry & geometry)
        : imp_(geometry.type(),getCorners(geometry))
      {}

      //! \brief Interface functions @{
      Dune::GeometryType type() const { return imp_.type(); }
      virtual int corners() const { return imp_.corners(); }

      virtual typename Imp::GlobalCoordinate corner
        (const int i) const { return imp_.corner(i); }

      virtual typename Imp::GlobalCoordinate global
        (const typename Imp::LocalCoordinate & local) const
      { return imp_.global(local); }

      virtual typename Imp::LocalCoordinate   local
        (const typename Imp::GlobalCoordinate & global) const
      { return imp_.local(global); }

      virtual typename Imp::GlobalCoordinate center() const
      { return imp_.center(); }

      virtual bool affine() const { return imp_.affine(); }

      virtual typename Imp::ctype integrationElement
        (const typename Imp::LocalCoordinate & local) const
      { return imp_.integrationElement(local); }

      virtual typename Imp::ctype volume() const
      { return imp_.volume(); }

#if DUNE_VERSION_NEWER(DUNE_GEOMETRY,2,3)
      typename Imp::JacobianTransposed jacobianTransposed
#else
      const typename Imp::JacobianTransposed & jacobianTransposed
#endif
        (const typename Imp::LocalCoordinate & local) const
      { return imp_.jacobianTransposed(local); }

#if DUNE_VERSION_NEWER(DUNE_GEOMETRY,2,3)
      typename Imp::JacobianInverseTransposed jacobianInverseTransposed
#else
      const typename Imp::JacobianInverseTransposed & jacobianInverseTransposed
#endif
        (const typename Imp::LocalCoordinate & local) const
      { return imp_.jacobianInverseTransposed(local); }
      //! @}

    private:
      Imp imp_;
    };

  }
}

#endif
