// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef SUBENTITY_HH
#define SUBENTITY_HH

#include <dune/geometry/type.hh>
#include <dune/common/shared_ptr.hh>

#include "cuboid.hh"
#include "geometry.hh"

namespace Dune {
  namespace UDG {

    // forward decl
    template<int dim, class ct>
    class InnerDomain;

    template <class Grid>
    class EntityPartBase {
    public:

      enum { dim=Grid::dimension };
      enum { dimworld=Grid::dimensionworld };
      typedef typename Grid::ctype ct;
      typedef typename Grid::template Codim<0>::Entity Entity;
      typedef Dune::FieldVector<ct, dimworld> Point;
      typedef Dune::Cuboid<dimworld, ct> BoundingBoxGeometry;
      typedef Dune::GridCuboid<Grid> BoundingBox;

      typedef SubTriangulationGeometry<dim,dimworld,ct> GlobalGeometry;
      typedef SubTriangulationGeometry<dim,dim,ct> LocalGeometry;

      // local -> global
      const GlobalGeometry& geometry () const { return global_g; };
      // local -> bbox
      const LocalGeometry& geometryInCell () const { return rect_g; };
      // domain information
      const int domainTag() const { return domain_tag; }
      const int domainIndex() const { return domain_index; }

      /** \brief Returns entity for which this entity part was constructed.

          \notice Returned entity may in general be different from entity returned
          by boundingBox().entity() .
       */
      const Entity & entity() const { return entity_; }

      /** \brief Returns bounding box with whom this sub entity is
          associated.
       */
      const BoundingBox & boundingBox() const { return *bbox; }

      EntityPartBase() : bbox((BoundingBox*)(0)) {}

      /** \brief Set entity for which this entity part was constructed.  */
      void entity(const Entity & e_)
      { entity_ = e_; }

      /** \brief Set bounding box associated with this entity.  */
      void boundingBox(const BoundingBox & bbox_)
      { bbox = Dune::shared_ptr<BoundingBox>( new BoundingBox(bbox_) ); }


    protected:
      int domain_tag;
      int domain_index;
      GlobalGeometry global_g;
      LocalGeometry rect_g;
      Entity entity_;
      Dune::shared_ptr<BoundingBox> bbox;

      template<int n>
      void setupGeometries(const BoundingBoxGeometry & _r,
                           const Dune::GeometryType & t,
                           Dune::FieldMatrix<ct, n, dim> & c);
    };

    template <class Grid,
        int dim=Grid::dimension,
        int dimworld=Grid::dimensionworld>
    class EntityPart;

    template <class Grid>
    class EntityPart< Grid, 2, 2 > : public EntityPartBase< Grid > {
      typedef EntityPartBase<Grid> Base;

      using Base::domain_index;
      using Base::domain_tag;

    public:
      enum { dim=2 };
      enum { dimworld=2 };

      typedef typename Base::ct ct;
      typedef typename Base::Point Point;
      typedef typename Base::BoundingBoxGeometry BoundingBoxGeometry;

      // 2d interface
      // Quadrilaterals
      EntityPart(const BoundingBoxGeometry & _r, int tag = 1, int index = 0);
      EntityPart(const BoundingBoxGeometry & _r,
                 const Point & A, const Point & B, int tag = 1, int index = 0);
      EntityPart(const BoundingBoxGeometry & _r,
                 const Point & A, const Point & B,
                 const Point & C, const Point & D, int tag = 1, int index = 0);
      // Triangles
      EntityPart(const BoundingBoxGeometry & _r,
                 const Point & A, const Point & B, const Point & C, int tag = 1, int index = 0);
      // CurveLinear Triangle
      EntityPart(const BoundingBoxGeometry & _r,
                 const Point & A, const Point & B, const Point & C,
                 const InnerDomain<2,ct> & i, int tag = 1, int index = 0);
      // 3d interface
      EntityPart(const BoundingBoxGeometry & _r,
                 const Point & A, const Point & B,
                 const Point & C, const Point & D,
                 const Point & E, int tag = 1, int index = 0) {
        DUNE_THROW(Dune::Exception, "Accessing 3D interface on 2D object");
      }
      EntityPart(const BoundingBoxGeometry & _r,
                 const Point & A, const Point & B,
                 const Point & C, const Point & D,
                 const Point & E, const Point & F, int tag = 1, int index = 0) {
        DUNE_THROW(Dune::Exception, "Accessing 3D interface on 2D object");
      }
      EntityPart(const BoundingBoxGeometry & _r,
                 const Point & A, const Point & B,
                 const Point & C, const Point & D,
                 const Point & E, const Point & F,
                 const Point & g, const Point & H, int tag = 1, int index = 0) {
        DUNE_THROW(Dune::Exception, "Accessing 3D interface on 2D object");
      }
    };

    template <class Grid>
    class EntityPart< Grid, 3, 3 > : public EntityPartBase< Grid > {
    private:
      typedef EntityPartBase<Grid> Base;

      using Base::domain_index;
      using Base::domain_tag;

    public:

      enum { dim=3 };
      enum { dimworld=3 };

      typedef typename Base::ct ct;
      typedef typename Base::Point Point;
      typedef typename Base::BoundingBoxGeometry BoundingBoxGeometry;

      // 2d interface
      EntityPart(const BoundingBoxGeometry & _r,
                 const Point & A, const Point & B,
                 const Point & C, int tag = 1, int index = 0) {
        DUNE_THROW(Dune::Exception, "Accessing 2D interface on 3D object");
      }
      // 3d interface
      // Cube
      EntityPart(const BoundingBoxGeometry & _r, int tag = 1, int index = 0);
      EntityPart(const BoundingBoxGeometry & _r,
                 const Point & A, const Point & B, int tag = 1, int index = 0);
      EntityPart(const BoundingBoxGeometry & _r,
                 const Point & A, const Point & B,
                 const Point & C, const Point & D,
                 const Point & E, const Point & F,
                 const Point & G, const Point & H, int tag = 1, int index = 0);
      // Simplex
      EntityPart(const BoundingBoxGeometry & _r,
                 const Point & A, const Point & B,
                 const Point & C, const Point & D, int tag = 1, int index = 0);
      // Pyramid
      EntityPart(const BoundingBoxGeometry & _r,
                 const Point & A, const Point & B,
                 const Point & C, const Point & D,
                 const Point & E, int tag = 1, int index = 0);
      // Prism
      EntityPart(const BoundingBoxGeometry & _r,
                 const Point & A, const Point & B,
                 const Point & C, const Point & D,
                 const Point & E, const Point & F, int tag = 1, int index = 0);
    };

  } // end UDG
} // end Dune

#include "subentity_impl.cc"

#endif // ENTITY_HH
