// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
/*
   TODO:
   - EntityPart(BoundingBox)
    -> EntityPartBase(BoundingBox)
   - EntityPart(BoundingBox,Point,Point)
    -> EntityPartBase(BoundingBox,Point,Point)
   - Remove dependency for innerdomain
 */

#include "subentity.hh"
#include <dune/udg/2dtriangulation/innerdomain.hh>

#include <dune/grid/yaspgrid.hh>
#if HAVE_UG
#include <dune/grid/uggrid.hh>
#endif
#if HAVE_ALUGRID
#include <dune/grid/alugrid.hh>
#endif

namespace Dune {
  namespace UDG {

    template<class Grid>
    template<int n>
    void EntityPartBase<Grid>::
    setupGeometries(const BoundingBoxGeometry & _r,
                    const Dune::GeometryType & t,
                    Dune::FieldMatrix<ct,n,dim> & c)
    {
      global_g = GlobalGeometry(t, c);

      Point bbox_a = _r.global(Point(0));
      Point bbox_b = _r.global(Point(1));

      for (int d=0; d<dim; d++)
      {
        c[0][d] -= bbox_a[d];
        for (int i=0; i<n; i++)
        {
          c[i][d] *= 1.0/(bbox_b[d] - bbox_a[d]);
        }
      }

      rect_g = LocalGeometry(t, c);
    }

    //////////////////////////////////////////////////////////////////////////
    // 2D
    //////////////////////////////////////////////////////////////////////////

    // Triangle
    template<class G>
    EntityPart< G, 2,2 >::EntityPart(const BoundingBoxGeometry & _r,
                                     const Point & A, const Point & B,
                                     const Point & C, int tag, int index)
    {
      domain_tag = tag; domain_index = index;
      Dune::FieldMatrix<ct, 3, 2> c;

      c[0] = A;
      c[1] = B-A;
      c[2] = C-A;

      this->setupGeometries(_r, GeometryTypes::triangle, c);
    }

    // isoTriangle
    template<class G>
    EntityPart< G, 2,2 >::EntityPart(const BoundingBoxGeometry & _r,
                                     const Point & A, const Point & B,
                                     const Point & C,
                                     const InnerDomain<2,ct> & i, int tag, int index)

    {
      domain_tag = tag; domain_index = index;
      Dune::FieldMatrix<ct, 6, 2> _c;

      Dune::FieldVector<ct, 3> b;
      Dune::FieldVector<ct, 3> c;
      int side = 0;
      i.polynom(B, C, side, b, c);

      _c[0] = A;
      _c[1][0] = b[0]-A[0];
      _c[1][1] = c[0]-A[1];
      _c[2][0] = b[0]+b[1]+b[2]-A[0];
      _c[2][1] = c[0]+c[1]+c[2]-A[1];
      _c[3] = 0;
      _c[4][0] = -b[2];
      _c[4][1] = -c[2];
      _c[5] = 0;

      this->setupGeometries(_r, GeometryTypes::triangle, _c);
    }

    // BBox
    template<class G>
    EntityPart< G, 2,2 >::EntityPart(const BoundingBoxGeometry & _r, int tag, int index)

    {
      domain_tag = tag; domain_index = index;
      Dune::FieldMatrix<ct, 3, 2> c;

      Point o = _r.origin();
      Point h = _r.height();

      c[0] = o;
      c[1][0] = h[0];
      c[1][1] = 0;
      c[2][0] = 0;
      c[2][1] = h[1];

      this->setupGeometries(_r, GeometryTypes::quadrilateral, c);
    }

    // Rectangle
    template<class G>
    EntityPart< G, 2,2 >::EntityPart(const BoundingBoxGeometry & _r,
                                     const Point & A, const Point & B, int tag, int index)

    {
      domain_tag = tag; domain_index = index;
      Dune::FieldMatrix<ct, 3, 2> c;

      c[0] = A;
      c[1][0] = B[0]-A[0];
      c[1][1] = 0;
      c[2][0] = 0;
      c[2][1] = B[1]-A[1];

      this->setupGeometries(_r, GeometryTypes::quadrilateral, c);
    }

    // Quadrilateral
    template<class G>
    EntityPart< G, 2,2 >::EntityPart(const BoundingBoxGeometry & _r,
                                     const Point & A, const Point & B,
                                     const Point & C, const Point & D, int tag, int index)

    {
      domain_tag = tag; domain_index = index;
      Dune::FieldMatrix<ct, 6, 2> c;

      c[0] = A;
      c[1] = B-A;
      c[2] = C-A;
      c[3] = 0;
      c[4] = A-B+D-C;
      c[5] = 0;

      this->setupGeometries(_r, GeometryTypes::quadrilateral, c);
    }

    //////////////////////////////////////////////////////////////////////////
    // 3D
    //////////////////////////////////////////////////////////////////////////

    // BBox
    template<class G>
    EntityPart< G, 3,3 >::EntityPart(const BoundingBoxGeometry & _r, int tag, int index)

    {
      domain_tag = tag; domain_index = index;
      Dune::FieldMatrix<ct, 4, 3> c;

      Point o = _r.origin();
      Point h = _r.height();

      c[0] = o;
      c[1][0] = h[0];
      c[1][1] = 0;
      c[1][2] = 0;
      c[2][0] = 0;
      c[2][1] = h[1];
      c[2][2] = 0;
      c[3][0] = 0;
      c[3][1] = 0;
      c[3][2] = h[2];

      this->setupGeometries(_r, GeometryTypes::hexahedron, c);
    }

    // Rectangle
    template<class G>
    EntityPart< G, 3,3 >::EntityPart(const BoundingBoxGeometry & _r,
                                     const Point & A, const Point & B, int tag, int index)

    {
      domain_tag = tag; domain_index = index;
      Dune::FieldMatrix<ct, 4, 3> c;

      c[0] = A;
      c[1][0] = B[0]-A[0];
      c[1][1] = 0;
      c[1][2] = 0;
      c[2][0] = 0;
      c[2][1] = B[1]-A[1];
      c[2][2] = 0;
      c[3][0] = 0;
      c[3][1] = 0;
      c[3][2] = B[2]-A[2];

      this->setupGeometries(_r, GeometryTypes::hexahedron, c);
    }

    // Hexahedron
    template<class Grid>
    EntityPart< Grid, 3,3 >::EntityPart(const BoundingBoxGeometry & _r,
                                        const Point & A, const Point & B,
                                        const Point & C, const Point & D,
                                        const Point & E, const Point & F,
                                        const Point & G, const Point & H, int tag, int index)

    {
      domain_tag = tag; domain_index = index;
      Dune::FieldMatrix<ct, 20, 3> c;

      c[0] = A;
      c[1] = B-A;
      c[2] = C-A;
      c[3] = E-A;
      c[4] = 0;
      c[5] = D+A-B-C;
      c[6] = F+A-B-E;
      c[7] = 0;
      c[8] = G+A-C-E;
      c[9] = 0;
      c[10] = 0;
      c[11] = 0;
      c[12] = 0;
      c[13] = 0;
      c[14] = H-A+B+C+E-D-F-G;
      c[15] = 0;
      c[16] = 0;
      c[17] = 0;
      c[18] = 0;
      c[19] = 0;

      this->setupGeometries(_r, GeometryTypes::hexahedron, c);
    }

    // Prism
    template<class G>
    EntityPart< G, 3,3 >::EntityPart(const BoundingBoxGeometry & _r,
                                     const Point & A, const Point & B,
                                     const Point & C, const Point & D,
                                     const Point & E, const Point & F, int tag, int index)

    {
      domain_tag = tag; domain_index = index;
      Dune::FieldMatrix<ct, 20, 3> c(0.0);

      c[0] = A;
      c[1] = B-A;
      c[2] = C-A;
      c[3] = D-A;
      c[4] = 0;
      c[5] = 0;
      c[6] = E+A-B-D;
      c[7] = 0;
      c[8] = F+A-C-D;

      this->setupGeometries(_r, GeometryTypes::prism, c);
    }

    // Pyramid
    template<class G>
    EntityPart< G, 3,3 >::EntityPart(const BoundingBoxGeometry & _r,
                                     const Point & A, const Point & B,
                                     const Point & C, const Point & D,
                                     const Point & E, int tag, int index)

    {
      domain_tag = tag; domain_index = index;
      Dune::FieldMatrix<ct, 20, 3> c(0.0);

      c[0] = A;
      c[1] = B-A;
      c[2] = D-A;
      c[3] = E-A;
      c[4] = 0;
      c[5] = C+A-B-D;

      this->setupGeometries(_r, GeometryTypes::pyramid, c);
    }

    // Simplex
    template<class G>
    EntityPart< G, 3,3 >::EntityPart(const BoundingBoxGeometry & _r,
                                     const Point & A, const Point & B,
                                     const Point & C, const Point & D, int tag, int index)

    {
      domain_tag = tag; domain_index = index;
      Dune::FieldMatrix<ct, 4, 3> c;

      c[0] = A;
      c[1] = B-A;
      c[2] = C-A;
      c[3] = D-A;

      this->setupGeometries(_r, GeometryTypes::tetrahedron, c);
    }

    //////////////////////////////////////////////////////////////////////////
    // Instantiation
    //////////////////////////////////////////////////////////////////////////

    // -- 2d --

    extern template
    EntityPart< Dune::YaspGrid<2> >::EntityPart(const BoundingBoxGeometry & _r,
                                                const Point & A, const Point & B, const Point & C, int tag, int index);

    extern template
    EntityPart< Dune::YaspGrid<2> >::EntityPart(const BoundingBoxGeometry & _r,
                                                const Point & A, const Point & B, const Point & C,
                                                const InnerDomain<2,double> & i, int tag, int index);
    extern template
    EntityPart< Dune::YaspGrid<2> >::EntityPart(const BoundingBoxGeometry & _r, int tag, int index);

    extern template
    EntityPart< Dune::YaspGrid<2> >::EntityPart(const BoundingBoxGeometry & _r,
                                                const Point & A, const Point & B, int tag, int index);
    extern template
    EntityPart< Dune::YaspGrid<2> >::EntityPart(const BoundingBoxGeometry & _r,
                                                const Point & A, const Point & B,
                                                const Point & C, const Point & D, int tag, int index);

    // -- 3d --

    extern template
    EntityPart< Dune::YaspGrid<3> >::EntityPart(const BoundingBoxGeometry & _r, int tag, int index);

    extern template
    EntityPart< Dune::YaspGrid<3> >::EntityPart(const BoundingBoxGeometry & _r,
                                                const Point & A, const Point & B, int tag, int index);

    extern template
    EntityPart< Dune::YaspGrid<3> >::EntityPart(const BoundingBoxGeometry & _r,
                                                const Point & A, const Point & B,
                                                const Point & C, const Point & D,
                                                const Point & E, const Point & F,
                                                const Point & G, const Point & H, int tag, int index);

    extern template
    EntityPart< Dune::YaspGrid<3> >::EntityPart(const BoundingBoxGeometry & _r,
                                                const Point & A, const Point & B,
                                                const Point & C, const Point & D,
                                                const Point & E, const Point & F, int tag, int index);

    extern template
    EntityPart< Dune::YaspGrid<3> >::EntityPart(const BoundingBoxGeometry & _r,
                                                const Point & A, const Point & B,
                                                const Point & C, const Point & D,
                                                const Point & E, int tag, int index);

    extern template
    EntityPart< Dune::YaspGrid<3> >::EntityPart(const BoundingBoxGeometry & _r,
                                                const Point & A, const Point & B,
                                                const Point & C, const Point & D, int tag, int index);

#if HAVE_UG
    // -- 2d --

    extern template
    EntityPart< Dune::UGGrid<2> >::EntityPart(const BoundingBoxGeometry & _r,
                                              const Point & A, const Point & B, const Point & C, int tag, int index);
    extern template
    EntityPart< Dune::UGGrid<2> >::EntityPart(const BoundingBoxGeometry & _r,
                                              const Point & A, const Point & B, const Point & C,
                                              const InnerDomain<2,double> & i, int tag, int index);
    extern template
    EntityPart< Dune::UGGrid<2> >::EntityPart(const BoundingBoxGeometry & _r, int tag, int index);

    extern template
    EntityPart< Dune::UGGrid<2> >::EntityPart(const BoundingBoxGeometry & _r,
                                              const Point & A, const Point & B, int tag, int index);
    extern template
    EntityPart< Dune::UGGrid<2> >::EntityPart(const BoundingBoxGeometry & _r,
                                              const Point & A, const Point & B,
                                              const Point & C, const Point & D, int tag, int index);

    // -- 3d --

    extern template
    EntityPart< Dune::UGGrid<3> >::EntityPart(const BoundingBoxGeometry & _r, int tag, int index);

    extern template
    EntityPart< Dune::UGGrid<3> >::EntityPart(const BoundingBoxGeometry & _r,
                                              const Point & A, const Point & B, int tag, int index);

    extern template
    EntityPart< Dune::UGGrid<3> >::EntityPart(const BoundingBoxGeometry & _r,
                                              const Point & A, const Point & B,
                                              const Point & C, const Point & D,
                                              const Point & E, const Point & F,
                                              const Point & G, const Point & H, int tag, int index);

    extern template
    EntityPart< Dune::UGGrid<3> >::EntityPart(const BoundingBoxGeometry & _r,
                                              const Point & A, const Point & B,
                                              const Point & C, const Point & D,
                                              const Point & E, const Point & F, int tag, int index);

    extern template
    EntityPart< Dune::UGGrid<3> >::EntityPart(const BoundingBoxGeometry & _r,
                                              const Point & A, const Point & B,
                                              const Point & C, const Point & D,
                                              const Point & E, int tag, int index);

    extern template
    EntityPart< Dune::UGGrid<3> >::EntityPart(const BoundingBoxGeometry & _r,
                                              const Point & A, const Point & B,
                                              const Point & C, const Point & D, int tag, int index);
#endif

#if HAVE_ALUGRID
    // -- 3d --
    extern template
    EntityPart< Dune::ALUGrid< 3, 3, cube, nonconforming > >::EntityPart(const BoundingBoxGeometry & _r, int tag, int index);

    extern template
    EntityPart< Dune::ALUGrid< 3, 3, cube, nonconforming > >::EntityPart(const BoundingBoxGeometry & _r,
                                                     const Point & A, const Point & B, int tag, int index);

    extern template
    EntityPart< Dune::ALUGrid< 3, 3, cube, nonconforming > >::EntityPart(const BoundingBoxGeometry & _r,
                                                     const Point & A, const Point & B,
                                                     const Point & C, const Point & D,
                                                     const Point & E, const Point & F,
                                                     const Point & G, const Point & H, int tag, int index);

    extern template
    EntityPart< Dune::ALUGrid< 3, 3, cube, nonconforming > >::EntityPart(const BoundingBoxGeometry & _r,
                                                     const Point & A, const Point & B,
                                                     const Point & C, const Point & D,
                                                     const Point & E, const Point & F, int tag, int index);

    extern template
    EntityPart< Dune::ALUGrid< 3, 3, cube, nonconforming > >::EntityPart(const BoundingBoxGeometry & _r,
                                                     const Point & A, const Point & B,
                                                     const Point & C, const Point & D,
                                                     const Point & E, int tag, int index);

    extern template
    EntityPart< Dune::ALUGrid< 3, 3, cube, nonconforming > >::EntityPart(const BoundingBoxGeometry & _r,
                                                     const Point & A, const Point & B,
                                                     const Point & C, const Point & D, int tag, int index);

    // -- 2d --

    extern template
    EntityPart< Dune::ALUGrid< 2, 2, simplex, nonconforming > >::EntityPart(const BoundingBoxGeometry & _r, int tag, int index);

    extern template
    EntityPart< Dune::ALUGrid< 2, 2, simplex, nonconforming > >::EntityPart(const BoundingBoxGeometry & _r,
                                                        const Point & A, const Point & B, const Point & C, int tag, int index);
#endif

  } // end UDG
} // end Dune
