// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:

#ifndef INTERSECTION_HH
#define INTERSECTION_HH

#include <dune/udg/common/cuboid.hh>
#include "geometry.hh"
#include <vector>
#include <cmath>
#include <limits>

namespace Dune {
  namespace UDG  {

    namespace BoundaryFlags {
      enum {
        internal = 99,
        processor = 100
                    // internal = -1,
                    // processor = -2
      };
    };


    template <class Grid>
    class IntersectionPartBase
    {
    private:
      enum { dim=Grid::dimension };
      enum { dimworld=Grid::dimensionworld };
      typedef typename Grid::ctype ct;
      typedef typename Grid::template Codim<0>::Entity Entity;

      typedef Dune::Cuboid<dimworld, ct> BoundingBox;
      typedef Dune::GridCuboid<Grid> GridCuboid;

    public:
      typedef SubTriangulationGeometry<dim-1,dimworld,ct> GlobalGeometry;
      typedef SubTriangulationGeometry<dim-1,dim,ct> LocalGeometry;

    private:
      GlobalGeometry global_g;
      LocalGeometry inside_g;
      LocalGeometry outside_g;
      Entity inside_p;
      Entity outside_p;
      //  NormalVectorFactory normal;
      int bid;
      int domain_tag;
      int inside_domain_index;
      int outside_domain_index;
      GridCuboid inBBox;
      GridCuboid outBBox;
      ct face_area;
      ct inside_volume;
      ct outside_volume;

    public:
      typedef Dune::FieldVector< ct, dimworld > NormalVector;
      typedef Dune::FieldVector< ct, dimworld > Point;
    protected:
      template<int n>
      void setupGeometries(const BoundingBox & _in,
                           const BoundingBox & _out,
                           const Dune::GeometryType & t,
                           const Dune::FieldMatrix<ct, n, dim> & c);

      IntersectionPartBase(const GridCuboid & _r, int _bid, int tag,
                           const int _inside_domain_index, const int _outside_domain_index) :
        inside_p(_r.entity()), outside_p(_r.entity()), bid(_bid), domain_tag(tag),
        inside_domain_index(_inside_domain_index), outside_domain_index(_outside_domain_index),
        inBBox(_r), outBBox(_r), face_area(0), inside_volume(0), outside_volume(0)
      {}

      IntersectionPartBase(const GridCuboid & _r1, const GridCuboid & _r2, int tag,
                           const int _inside_domain_index, const int _outside_domain_index) :
        inside_p(_r1.entity()), outside_p(_r2.entity()), domain_tag(tag),
        inside_domain_index(_inside_domain_index), outside_domain_index(_outside_domain_index),
        inBBox(_r1), outBBox(_r2), face_area(0), inside_volume(0), outside_volume(0)
      {}

    public:

      //! \todo doc me!
      ct area() const {return face_area;}
      ct & area() {return face_area;}

      //! \todo doc me!
      ct insideVolume() const {return inside_volume;}
      ct & insideVolume() {return inside_volume;}

      //! \todo doc me!
      ct outsideVolume() const {return outside_volume;}
      ct & outsideVolume() {return outside_volume;}

      //! @copydoc IntersectionIteratorInterface::boundaryId
      int boundarySegmentIndex () const {
        return bid;
      }
      //! \todo doc me!
      int domainTag () const {
        return domain_tag;
      }
      //! \todo doc me!
      int insideDomainIndex () const {
        return inside_domain_index;
      }
      //! \todo doc me!
      int outsideDomainIndex () const {
        return outside_domain_index;
      }
      //! @copydoc IntersectionIteratorInterface::boundary
      bool boundary() const {
        return outside_domain_index != inside_domain_index;
      }
      //! @copydoc IntersectionIteratorInterface::boundary
      bool internalBoundary() const {
        return bid == BoundaryFlags::internal;
      }
      //! @copydoc IntersectionIteratorInterface::neighbor
      bool neighbor() const {
        return outside_domain_index != -1;
      }
      //! @copydoc IntersectionIteratorInterface::outside
      const Entity & outside() const {
        return outside_p;
      }
      //! @copydoc IntersectionIteratorInterface::inside
      const Entity & inside() const {
        return inside_p;
      }
      //! @copydoc IntersectionIteratorInterface::geometryInInside
      const LocalGeometry& geometryInInside () const
      {
        return inside_g;
      }
      //! @copydoc IntersectionIteratorInterface::geometryInOutside
      const LocalGeometry& geometryInOutside () const
      {
        return outside_g;
      }
      //! @copydoc IntersectionIteratorInterface::intersectionGlobal
      const GlobalGeometry& geometry () const
      {
        return global_g;
      }
      const GridCuboid & boundingBoxInside() const
      {
        return inBBox;
      }
      const GridCuboid & boundingBoxOutside() const
      {
        return outBBox;
      }


      //! @copydoc IntersectionIteratorInterface::intersectionSelfLocal
      const LocalGeometry& intersectionSelfLocal () const DUNE_DEPRECATED
      {
        return inside_g;
      }
      //! @copydoc IntersectionIteratorInterface::intersectionNeighborLocal
      const LocalGeometry& intersectionNeighborLocal () const DUNE_DEPRECATED
      {
        return outside_g;
      }
      //! @copydoc IntersectionIteratorInterface::intersectionGlobal
      const GlobalGeometry& intersectionGlobal () const DUNE_DEPRECATED
      {
        return global_g;
      }
    };

    template <class Grid,
        int dim=Grid::dimension,
        int dimworld=Grid::dimensionworld>
    class IntersectionPart;

    template <class Grid>
    class IntersectionPart< Grid,2,2 > : public IntersectionPartBase< Grid > {
    private:
      enum { dim=2 };
      enum { dimworld=2 };
      typedef typename Grid::ctype ct;
    public:
      typedef Dune::Cuboid<dimworld, ct> BoundingBox;
      typedef Dune::GridCuboid<Grid> GridCuboid;

      typedef SubTriangulationGeometry<dim-1,dimworld,ct> GlobalGeometry;
      typedef SubTriangulationGeometry<dim-1,dim,ct> LocalGeometry;

      typedef Dune::FieldVector< ct, dimworld > NormalVector;
      typedef Dune::FieldVector< ct, dimworld > Point;
      typedef Dune::FieldVector< ct, dimworld -1 > FacePoint;
    private:

      /* needed for normal() */
      Dune::FieldMatrix<ct,3,2> c;
      int side;
      NormalVector N;
      inline ct sqr(ct x) const { return x*x; }
      const NormalVector normal (ct t) const;

    public:
      /* 2D Interface */
      IntersectionPart(const GridCuboid & _r, Point A, Point B, int _bid,
                       const NormalVector & n, int tag = 3, const int inside_domain = 0, const int outside_domain = -1);
      template<template<int,typename> class Domain>
      IntersectionPart(const GridCuboid & _r, Point A, Point B, int _bid,
                       const Domain<dim,ct> & _i, int tag = 3, const int inside_domain = 0, const int outside_domain = -1);
      IntersectionPart(const GridCuboid & _r1, const GridCuboid & _r2,
                       Point A, Point B, const NormalVector & n, int tag = 3, const int inside_domain = 0, const int outside_domain = -1);
      /* 3D Interface */
      IntersectionPart(const GridCuboid & _r,
                       Point A, Point B, Point C, Point D, int _bid,
                       const NormalVector & n, int tag = 3, const int inside_domain = 0, const int outside_domain = -1) : IntersectionPartBase< Grid >(_r, _bid, tag, inside_domain, outside_domain) {
        DUNE_THROW(Dune::Exception, "Accessing 3D interface on 2D object");
      }
      IntersectionPart(const GridCuboid & _r,
                       Point A, Point B, Point C, int _bid,
                       const NormalVector & n, int tag = 3, const int inside_domain = 0, const int outside_domain = -1) : IntersectionPartBase< Grid >(_r, _bid, tag, inside_domain, outside_domain) {
        DUNE_THROW(Dune::Exception, "Accessing 3D interface on 2D object");
      }
      IntersectionPart(const GridCuboid & _in,
                       const GridCuboid & _out,
                       Point A, Point B, Point C, Point D,
                       const NormalVector & n, int tag = 3, const int inside_domain = 0, const int outside_domain = -1 ) : IntersectionPartBase< Grid >(_in,_out, tag, inside_domain, outside_domain) {
        DUNE_THROW(Dune::Exception, "Accessing 3D interface on 2D object");
      }
      IntersectionPart(const GridCuboid & _in,
                       const GridCuboid & _out,
                       Point A, Point B, Point C,
                       const NormalVector & n, int tag = 3, const int inside_domain = 0, const int outside_domain = -1) : IntersectionPartBase< Grid >(_in,_out, tag, inside_domain, outside_domain) {
        DUNE_THROW(Dune::Exception, "Accessing 3D interface on 2D object");
      }

      //! @copydoc IntersectionIteratorInterface::unitOuterNormal
      const NormalVector unitOuterNormal (ct t) const { return normal(t); };
    };


    template <class Grid>
    class IntersectionPart< Grid,3,3 > : public IntersectionPartBase< Grid > {
    private:
      enum { dim=3 };
      enum { dimworld=3 };
      typedef typename Grid::ctype ct;
      typedef Dune::Cuboid<dimworld, ct> BoundingBox;

    public:
      typedef Dune::GridCuboid<Grid> GridCuboid;
      typedef SubTriangulationGeometry<dim-1,dimworld,ct> GlobalGeometry;
      typedef SubTriangulationGeometry<dim-1,dim,ct> LocalGeometry;

      typedef Dune::FieldVector< ct, dimworld > NormalVector;
      typedef Dune::FieldVector< ct, dimworld > Point;
      typedef Dune::FieldVector< ct, dimworld-1 > FacePoint;
    private:
      Dune::FieldMatrix<ct,6,3> c;
      int side;
      NormalVector N;
      inline ct sqr(ct x) const { return x*x; }
      const NormalVector normal (FacePoint t) const;

    public:
      /* 2D Interface */
      IntersectionPart(const GridCuboid & _r, Point A, Point B, int _bid,
                       const NormalVector & n, int tag = 3, const int inside_domain = 0, const int outside_domain = -1) :
        IntersectionPartBase< Grid >(_r,_bid, tag, inside_domain, outside_domain) {
        DUNE_THROW(Dune::Exception, "Accessing 2D interface on 3D object");
      }

      IntersectionPart(const GridCuboid & _r1, const GridCuboid & _r2,
                       Point A, Point B, const NormalVector & n, int tag = 3, const int inside_domain = 0, const int outside_domain = -1) :
        IntersectionPartBase< Grid >(_r1,_r2, tag, inside_domain, outside_domain) {
        DUNE_THROW(Dune::Exception, "Accessing 2D interface on 3D object");
      }

      /* 3D Interface */
      IntersectionPart(const GridCuboid & _r,
                       Point A, Point B, Point C, Point D, int _bid,
                       const NormalVector & n, int tag = 3, const int inside_domain = 0, const int outside_domain = -1);
      IntersectionPart(const GridCuboid & _r,
                       Point A, Point B, Point C, int _bid,
                       const NormalVector & n, int tag = 3, const int inside_domain = 0, const int outside_domain = -1);
      IntersectionPart(const GridCuboid & _in,
                       const GridCuboid & _out,
                       Point A, Point B, Point C, Point D,
                       const NormalVector & n, int tag = 3, const int inside_domain = 0, const int outside_domain = -1);
      IntersectionPart(const GridCuboid & _in,
                       const GridCuboid & _out,
                       Point A, Point B, Point C,
                       const NormalVector & n, int tag = 3, const int inside_domain = 0, const int outside_domain = -1);

      //! @copydoc IntersectionIteratorInterface::unitOuterNormal
      const NormalVector unitOuterNormal (FacePoint t) const { return normal(t); };
    };

  } // end namespace UDG
} // end namespace DUNE

#include "subintersection_impl.cc"

#endif // INTERSECTION_HH
