// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:

// only include config.h if we compile an intersection.o
#ifndef INTERSECTION_HH
#include <config.h>
#endif

#include <cmath>
#include <dune/udg/common/subintersection.hh>

namespace Dune {
  namespace UDG  {

    static const Dune::GeometryType _Line(1);
    static const Dune::GeometryType _Triangle(Dune::GeometryType::simplex, 2);
    static const Dune::GeometryType _Quadrilateral(Dune::GeometryType::cube, 2);

    template<class Grid>
    template<int n>
    void IntersectionPartBase<Grid>::
    setupGeometries(const BoundingBox & _in,
                    const BoundingBox & _out,
                    const Dune::GeometryType & t,
                    const Dune::FieldMatrix<ct,n,dim> & _c)
    {
      global_g = GlobalGeometry(t, _c);

      Dune::FieldMatrix<ct,n,dim> c2;
      Point o;
      Point h;

      o = _in.origin();
      h = _in.height();
      c2 = _c;
      for (int d=0; d<dim; d++) {
        c2[0][d] -= o[d];
        for (int i=0; i<n; i++) {
          c2[i][d] *= 1.0/h[d];
          if (!std::isfinite(c2[i][d]))
          {
            std::ostringstream msg;
            msg << t << std::endl;
            msg << o << std::endl;
            msg << h << std::endl;
            int corners =
              Dune::ReferenceElements<double,dim>::general(t).size(dim);
            for (int c=0; c<corners; c++) {
              Point p =
                Dune::ReferenceElements<double,dim>::general(t).position(c,dim);
              msg << c << ": " << _in.global(p) << std::endl;
            }
            DUNE_THROW(Dune::MathError, "Degenerized Transformation Matrix"
                       << std::endl << msg.str());
          }
        }
      }
      inside_g = LocalGeometry(t,c2);

      o = _out.origin();
      h = _out.height();
      c2 = _c;
      for (int d=0; d<dim; d++) {
        c2[0][d] -= o[d];
        for (int i=0; i<n; i++) {
          c2[i][d] *= 1.0/h[d];
          if (!std::isfinite(c2[i][d]))
          {
            std::ostringstream msg;
            msg << t << std::endl;
            msg << o << std::endl;
            msg << h << std::endl;
            int corners =
              Dune::ReferenceElements<double,dim>::general(t).size(dim);
            for (int c=0; c<corners; c++) {
              Point p =
                Dune::ReferenceElements<double,dim>::general(t).position(c,dim);
              msg << c << ": " << _out.global(p) << std::endl;
            }
            msg << "_c: " << _c << std::endl;
            msg << "c2: " << c2 << std::endl;
            DUNE_THROW(Dune::MathError, "Degenerized Transformation Matrix"
                       << std::endl << msg.str());
          }
        }
      }
      outside_g = LocalGeometry(t,c2);
    }

    /////////////////////////////////////////////

    template <class G>
    template <template<int,typename> class Domain>
    IntersectionPart< G, 2,2 >::IntersectionPart
    (
      const GridCuboid & _r,
      Point A, Point B, int _bid,
      const Domain<dim,ct> & _i,
      int tag, const int inside_domain, const int outside_domain
    ) :
      IntersectionPartBase< G >(_r, _bid,tag,inside_domain,outside_domain)
    {
      Dune::FieldVector<ct,3> a;
      Dune::FieldVector<ct,3> b;
      _i.polynom(A, B, side, a, b);
      for (int i=0; i<3; i++) {
        c[i][0] = a[i]; c[i][1] = b[i];
      }

      setupGeometries(_r, _r, _Line, c);
    }

    template <class G>
    IntersectionPart< G, 2,2 >::IntersectionPart
    (
      const GridCuboid & _r,
      Point A, Point B, int _bid,
      const NormalVector & n,
      int tag, const int inside_domain, const int outside_domain
    ) :
      IntersectionPartBase< G >(_r, _bid,tag,inside_domain,outside_domain), side(1), N(n)
    {
      c[0] = A;
      c[1] = B-A;
      c[2] = 0;

      setupGeometries(_r, _r, _Line, c);

      if (normal(0) * n < 0) side = -1;
    }

    template <class G>
    IntersectionPart< G, 2,2 >::IntersectionPart
    (
      const GridCuboid & in,
      const GridCuboid & out,
      Point A, Point B,
      const Dune::FieldVector< ct, dimworld > & n,
      int tag, const int inside_domain, const int outside_domain
    ) :
      IntersectionPartBase< G >(in,out,tag,inside_domain,outside_domain), side(1), N(n)
    {
      c[0] = A;
      c[1] = B-A;
      c[2] = 0;

      setupGeometries(in, out, _Line, c);

      if (normal(0) * n < 0) side = -1;
    }

    template <class G>
    const typename IntersectionPart< G, 2,2 >::NormalVector
    IntersectionPart< G, 2,2 >::normal (ct T) const
    {
      if (side == 0) return N;

      Dune::FieldVector< ct, dimworld > v;
      // v_0 = dy/dt / |v|;
      v[0] = - (c[1][1]+2*c[2][1]*T);
      assert(std::isfinite(v[0]));
      // v_1 = dx/dt / |v|;
      v[1] = c[1][0]+2*c[2][0]*T;
      assert(std::isfinite(v[1]));
      ct l = sqrt(sqr(v[0])+sqr(v[1]));
      assert(std::isfinite(l));
      assert(l != 0);
      v *= -side / l;

      return v;
    }

    /////////////////////////////////////////////

    template <class G>
    IntersectionPart< G, 3,3 >::IntersectionPart
    (
      const GridCuboid & _in,
      const GridCuboid & _out,
      Point A, Point B, Point C,
      const NormalVector & n,
      int tag, const int inside_domain, const int outside_domain
    ) :
      IntersectionPartBase< G >(_in, _out,tag,inside_domain,outside_domain), side(1), N(n)
    {
      c[0] = A;
      c[1] = B-A;
      c[2] = C-A;
      c[3] = 0;
      c[4] = 0;
      c[5] = 0;

      setupGeometries(_in, _out, _Triangle, c);

#ifndef UDG_INTERSECTION_TRUST_NORMAL
      if (normal(FacePoint(0)) * n < 0) side = -1;
      N = normal(FacePoint(0));
#endif
      side = 0;
    }

    template <class G>
    IntersectionPart< G, 3,3 >::IntersectionPart
    (
      const GridCuboid & _in,
      Point A, Point B, Point C, int _bid,
      const NormalVector & n,
      int tag, const int inside_domain, const int outside_domain
    ) :
      IntersectionPartBase< G >(_in, _bid,tag,inside_domain,outside_domain), side(1), N(n)
    {
      c[0] = A;
      c[1] = B-A;
      c[2] = C-A;
      c[3] = 0;
      c[4] = 0;
      c[5] = 0;

      setupGeometries(_in, _in, _Triangle, c);
#ifndef UDG_INTERSECTION_TRUST_NORMAL
      if (normal(FacePoint(0)) * n < 0) side = -1;
      N = normal(FacePoint(0));
#endif
      side = 0;
    }

    template <class G>
    IntersectionPart< G, 3,3 >::IntersectionPart
    (
      const GridCuboid & _in,
      const GridCuboid & _out,
      Point A, Point B, Point C, Point D,
      const NormalVector & n,
      int tag, const int inside_domain, const int outside_domain
    ) :
      IntersectionPartBase< G >(_in, _out,tag,inside_domain,outside_domain), side(1), N(n)
    {
      c[0] = A;
      c[1] = B-A;
      c[2] = C-A;
      c[3] = 0;
      c[4] = A-B+D-C;
      c[5] = 0;

      setupGeometries(_in, _out, _Quadrilateral, c);

#ifndef UDG_INTERSECTION_TRUST_NORMAL
      if (normal(FacePoint(0)) * n < 0) side = -1;
      N = normal(FacePoint(0));
#endif
      side = 0;
    }

    template <class G>
    IntersectionPart< G, 3,3 >::IntersectionPart
    (
      const GridCuboid & _in,
      Point A, Point B, Point C, Point D, int _bid,
      const NormalVector & n,
      int tag, const int inside_domain, const int outside_domain
    ) :
      IntersectionPartBase< G >(_in, _bid,tag,inside_domain,outside_domain), side(1), N(n)
    {
      c[0] = A;
      c[1] = B-A;
      c[2] = C-A;
      c[3] = 0;
      c[4] = A-B+D-C;
      c[5] = 0;

      setupGeometries(_in, _in, _Quadrilateral, c);
#ifndef UDG_INTERSECTION_TRUST_NORMAL
      if (normal(FacePoint(0)) * n < 0) side = -1;
      N = normal(FacePoint(0));
#endif
      side = 0;
    }

    template <class G>
    const typename IntersectionPart< G, 3,3 >::NormalVector
    IntersectionPart< G, 3,3 >::normal (FacePoint t) const
    {
      if (side == 0) return N;

      Dune::FieldVector< ct, dimworld > v;
      v[0] = (c[1][1]+2*c[3][1]*t[0]+c[4][1]*t[1])
             *(c[2][2]+c[4][2]*t[0]+2*c[5][2]*t[1])
             -(c[1][2]+2*c[3][2]*t[0]+c[4][2]*t[1])
             *(c[2][1]+c[4][1]*t[0]+2*c[5][1]*t[1]);
      assert(std::isfinite(v[0]));
      v[1] = (c[1][2]+2*c[3][2]*t[0]+c[4][2]*t[1])
             *(c[2][0]+c[4][0]*t[0]+2*c[5][0]*t[1])
             -(c[1][0]+2*c[3][0]*t[0]+c[4][0]*t[1])
             *(c[2][2]+c[4][2]*t[0]+2*c[5][2]*t[1]);
      assert(std::isfinite(v[1]));
      v[2] = (c[1][0]+2*c[3][0]*t[0]+c[4][0]*t[1])
             *(c[2][1]+c[4][1]*t[0]+2*c[5][1]*t[1])
             -(c[1][1]+2*c[3][1]*t[0]+c[4][1]*t[1])
             *(c[2][0]+c[4][0]*t[0]+2*c[5][0]*t[1]);
      assert(std::isfinite(v[2]));
      ct l = sqrt(sqr(v[0])+sqr(v[1])+sqr(v[2]));
      assert(std::isfinite(l));
      assert(l != 0);
      v *= -side / l;
      return v;
#if 0
      v[0] = c[1][1]*c[2][2] + c[1][2]*c[2][1]
             + (c[1][1]*c[4][2] + c[1][2]*c[4][1]
                + (2*c[3][2] + 2*c[5][2])*c[2][1] + (2*c[3][1] + 2*c[5][1])*c[2][2]
                + ((2*c[3][1] + 2*c[5][1])*c[4][2]
                   + (2*c[3][2] + 2*c[5][2])*c[4][1])*t[0])*t[0]
             + (c[4][1]*c[2][2] + c[4][2]*c[2][1] + 2*c[4][1]*c[4][2]*t[0])*t[1];
      assert(std::isfinite(v[0]));
      v[1] = c[1][2]*c[2][0] + c[1][0]*c[2][2]
             + (c[1][2]*c[4][0] + c[1][0]*c[4][2]
                + (2*c[3][0] + 2*c[5][0])*c[2][2] + (2*c[3][2] + 2*c[5][2])*c[2][0]
                + ((2*c[3][2] + 2*c[5][2])*c[4][0]
                   + (2*c[3][0] + 2*c[5][0])*c[4][2])*t[0])*t[0]
             + (c[4][2]*c[2][0] + c[4][0]*c[2][2] + 2*c[4][2]*c[4][0]*t[0])*t[1];
      assert(std::isfinite(v[1]));
      v[2] = c[1][0]*c[2][1] + c[1][1]*c[2][0]
             + (c[1][0]*c[4][1] + c[1][1]*c[4][0]
                + (2*c[3][1] + 2*c[5][1])*c[2][0] + (2*c[3][0] + 2*c[5][0])*c[2][1]
                + ((2*c[3][0] + 2*c[5][0])*c[4][1]
                   + (2*c[3][1] + 2*c[5][1])*c[4][0])*t[0])*t[0]
             + (c[4][0]*c[2][1] + c[4][1]*c[2][0] + 2*c[4][0]*c[4][1]*t[0])*t[1];
      assert(std::isfinite(v[2]));
      ct l = sqrt(sqr(v[0])+sqr(v[1])+sqr(v[2]));
      assert(std::isfinite(l));
      assert(l != 0);
      v *= -side / l;
      return v;
#endif
    }

  } // end namespace UDG
} // end namespace DUNE
