// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_UDG_COMMON_THREADUTIL_HH
#define DUNE_UDG_COMMON_THREADUTIL_HH

#include <thread>
#include <mutex>

/** \brief A base for local storage elements */
class LocalStorageBase
{
public:
  //! Constructor
  LocalStorageBase() : busyFlag(false) {}

  //! Access busy flag
  bool & busy() { return busyFlag; }

  //! Access busy flag
  void release() { busyFlag = false; }

  //! Access busy flag
  const bool & busy() const { return busyFlag; }

protected:
  //! Flag indicating that storage is accessed
  bool busyFlag;
};

/** \brief A local storage device for c++11 threads

    \tparam S Storage type
 */
template<typename S>
class ThreadLocalStorageBase
{

  //! Storages for all threads
  typedef std::map<std::thread::id,S> Storages;

public:
  //! Storage type
  typedef S Storage;

  /** \brief Constructor

      \param _nThreads (Maximum) number of threads.
   */
  ThreadLocalStorageBase(const int _nThreads)
    : nThreads(_nThreads) {}

  /** \brief Create a new empty storage element */
  virtual Storage emptyStorage() const = 0;

  /** \brief Find entry in storage */
  Storage & findInCache(std::mutex & f_lock) const
  {
    std::thread::id id = std::this_thread::get_id();

    Storage * pentry;
    {
      std::lock_guard<std::mutex> lock(f_lock);

      typename Storages::iterator cacheIt =
        storages.find(id);

      // If no entry yet then insert an empty entry
      if(cacheIt == storages.end()) {

        // Empty storages from last assembly
        if(storages.size()>=nThreads+1) {
          bool busy = true;
          while(busy) {
            busy = false;
            for(typename Storages::iterator
                cit = storages.begin();
                cit != storages.end(); ++cit)
              busy |= cit->second.busy();
          }
          std::cout << "Empty thread storages!" << std::endl;
          storages.clear();
        }
        std::pair<std::thread::id,Storage> newPair =
          (std::make_pair(id,this->emptyStorage()));
        cacheIt = storages.insert(newPair).first;
      }
      pentry = &(cacheIt->second);
      pentry->busy() = true;
      return *pentry;
    }

  }

private:
  const size_t nThreads;
  mutable Storages storages;
};


/** \brief Cache for intersection parts

    \tparam IPL Intersection part list type
    \tparam Size Size type of index set in grid
 */
template<typename IPL, typename Size>
struct IntersectionCache : public LocalStorageBase
{
  //! Index of the currently cached cell
  Size cachedCellIndex;
  //! List of cached intersection parts
  typedef IPL IntersectionPartList;
  IPL intersectionCache;
};

/** A thread cache for PDELab type grid function */
template<typename IPL, typename Size>
class IntersectionThreadCache :
  public ThreadLocalStorageBase<
      IntersectionCache<IPL,Size>
      >
{
public:

  //! Storage type
  typedef IntersectionCache<IPL,Size> Storage;
  //! Base
  typedef ThreadLocalStorageBase<Storage>
  Base;
  //! List to store intersection parts
  typedef IPL IntersectionPartList;

  //! Constructor
  IntersectionThreadCache
    (const Size _nCells,const int _nThreads)
    : Base(_nThreads), nCells(_nCells){}

  /** \brief Create a new empty storage element */
  virtual Storage emptyStorage() const
  {
    Storage storage;
    storage.cachedCellIndex = nCells+1;
    return storage;
  }

private:
  const Size nCells;
};


#endif
