// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
// $Id: adapter.hh 572 2009-12-20 22:03:44Z christi $

#ifndef DUNE_FUNCTION_ADAPTER_HH
#define DUNE_FUNCTION_ADAPTER_HH

#include "functions.hh"

namespace Dune
{

  /**
     @brief Access a Function as a GridFunction

     This class will provide the GridFunction methods evallocal() and
     evalalllocal() for any dune-disc Function class, even if that class is
     itself not a GridFunction.  This is done by using the Geometry of the
     Entity given to evallocal() or evalalllocal() to transform the local
     coordinates to global coordinates and then calling eval() with the
     transformed corrdinates.
   */
  template<class G, class RT, int m=1>
  class GridFunctionAdapter :
    virtual public GridFunction<G,RT,m>,
    virtual public GridFunctionDefault<G,RT,m>,
    virtual public FunctionDefault<typename G::ctype,RT,G::dimension,m>
  {
    //! get domain field type from the grid
    typedef typename G::ctype DT;

    //! get entity from the grid
    typedef typename G::template Codim<0>::Entity Entity;

    //! get domain dimension from the grid
    enum {n=G::dimension};

    //! get world dimension from the grid
    enum {dimw=G::dimensionworld};

    //! make copy constructor private
    GridFunctionAdapter (const GridFunctionAdapter&);
  public:
    //! create Adaptor from a Function
    GridFunctionAdapter(const FunctionBase<DT,RT,n,m>& f) : _fnkt(f) {};

    // forward the evaluation to _fnkt
    virtual RT eval (int comp, const Dune::FieldVector<DT,n>& x) const
    {
      return _fnkt.eval(comp, x);
    }

    //! VTK output
    template<class GV>
    typename VTKWriter<GV>::VTKFunctionPtr
    vtkFunction (VTKWriter<GV>& vtkwriter, std::string s) const
    {
      return typename VTKWriter<GV>::VTKFunctionPtr(
               new VTKGridFunctionWrapper<GV,RT,m>(*this,s));
    }

  private:
    const FunctionBase<DT,RT,n,m>& _fnkt;
  };

  /**
     Access a DifferentiableFunction as a GridFunction
   */
  template<class G, class RT, int m=1>
  class DifferentiableGridFunctionAdapter :
    virtual public GridFunction<G,RT,m>,
    virtual public DifferentiableGridFunction<G,RT,m>,
    virtual public GridFunctionDefault<G,RT,m>,
    virtual public DifferentiableGridFunctionDefault<G,RT,m>,
    virtual public FunctionDefault<typename G::ctype,RT,G::dimension,m>
  {
    //! get domain field type from the grid
    typedef typename G::ctype DT;

    //! get entity from the grid
    typedef typename G::template Codim<0>::Entity Entity;

    //! get domain dimension from the grid
    enum {n=G::dimension};

    //! get world dimension from the grid
    enum {dimw=G::dimensionworld};

    //! make copy constructor private
    DifferentiableGridFunctionAdapter (const DifferentiableGridFunctionAdapter&);
  public:
    //! create Adaptor from a Function
    DifferentiableGridFunctionAdapter(const DifferentiableFunction<DT,RT,n,m>& f) : _fnkt(f) {};

    // forward the evaluation to _fnkt
    virtual RT eval (int comp, const Dune::FieldVector<DT,n>& x) const
    {
      return _fnkt.eval(comp, x);
    }

    // forward the derivative to _fnkt
    virtual RT derivative (int comp, const Dune::FieldVector<int,n>& d, const Dune::FieldVector<DT,n>& x) const
    {
      return _fnkt.derivative(comp, d, x);
    }

    // forward to _fnkt
    virtual int order () const
    {
      return _fnkt.order();
    }

    //! VTK output
    template<class GV>
    typename VTKWriter<GV>::VTKFunctionPtr
    vtkFunction (VTKWriter<GV>& vtkwriter, std::string s) const
    {
      return typename VTKWriter<GV>::VTKFunctionPtr(
               new VTKGridFunctionWrapper<GV,RT,m>(*this,s));
    }

  private:
    const DifferentiableFunction<DT,RT,n,m>& _fnkt;
  };

  /**
     Evaluate a GridLevelFunction as a GridLeafFunction
   */
  template<class G, class RT, int m=1>
  class GridLevelLeafFunctionAdapter :
    virtual public GridFunction<G,RT,m>,
    virtual public GridFunctionDefault<G,RT,m>,
    virtual public FunctionDefault<typename G::ctype,RT,G::dimension,m>
  {
    //! get domain field type from the grid
    typedef typename G::ctype DT;

    //! get entity from the grid
    typedef typename G::template Codim<0>::Entity Entity;

    //! get entitypointer from the grid
    typedef typename G::template Codim<0>::EntityPointer EntityPointer;

    //! get domain dimension from the grid
    enum {n=G::dimension};

    //! get world dimension from the grid
    enum {dimw=G::dimensionworld};

    //! make copy constructor private
    GridLevelLeafFunctionAdapter (const GridLevelLeafFunctionAdapter&);
  public:
    //! create Adaptor from a Function
    GridLevelLeafFunctionAdapter(const GridFunction<G,RT,m>& f, int l) : _fnkt(f), _level(l) {};

    //! @copydoc Dune::FunctionBase::eval()
    // forward the evaluation to _fnkt
    virtual RT eval (int comp, const Dune::FieldVector<DT,n>& x) const
    {
      return _fnkt.eval(comp, x);
    }

    //! @copydoc Dune::GridFunction::evallocal()
    RT evallocal (int comp, const Entity& e, const Dune::FieldVector<DT,n>& xi) const
    {
      assert(e.level() >= _level);
      if (e.level() > _level)
      {
        return evallocal(comp, *(e.father()), e.geometryInFather().global(xi));
      }
      return _fnkt.evallocal(comp, e, xi);
    }

    //! VTK output
    template<class GV>
    typename VTKWriter<GV>::VTKFunctionPtr
    vtkFunction (VTKWriter<GV>& vtkwriter, std::string s) const
    {
      return typename VTKWriter<GV>::VTKFunctionPtr(
               new VTKGridFunctionWrapper<GV,RT,m>(*this,s));
    }

  private:
    const GridFunction<G,RT,m>& _fnkt;
    int _level;
  };

  /**
     Evaluate a GridFunction<IS> as a GridLeafFunction
   */
  template<class G, class IS, class RT, int m=1>
  class GridLeafFunctionAdapter :
    virtual public GridFunction<G,RT,m>,
    virtual public GridFunctionDefault<G,RT,m>,
    virtual public FunctionDefault<typename G::ctype,RT,G::dimension,m>
  {
    //! get domain field type from the grid
    typedef typename G::ctype DT;

    //! indexSet type for the export/evaluation
    typedef typename G::LeafIndexSet ISX;

    //! get entity from the grid
    typedef typename G::template Codim<0>::Entity Entity;

    //! get entitypointer from the grid
    typedef typename G::template Codim<0>::EntityPointer EntityPointer;

    //! get domain dimension from the grid
    enum {n=G::dimension};

    //! get world dimension from the grid
    enum {dimw=G::dimensionworld};

    //! make copy constructor private
    GridLeafFunctionAdapter (const GridLeafFunctionAdapter&);
  public:
    //! create Adaptor from a Function
    GridLeafFunctionAdapter(const GridFunction<G,RT,m>& f, const IS & is) : _fnkt(f), _is(is) {};

    //! @copydoc Dune::FunctionBase::eval()
    // forward the evaluation to _fnkt
    virtual RT eval (int comp, const Dune::FieldVector<DT,n>& x) const
    {
      return _fnkt.eval(comp, x);
    }

    //! @copydoc Dune::GridFunction::evallocal()
    RT evallocal (int comp, const Entity& e, const Dune::FieldVector<DT,n>& xi) const
    {
      if (! _is.contains(e))
      {
        return evallocal(comp, *(e.father()), e.geometryInFather().global(xi));
      }
      return _fnkt.evallocal(comp, e, xi);
    }

    //! VTK output
    template<class GV>
    typename VTKWriter<GV>::VTKFunctionPtr
    vtkFunction (VTKWriter<GV>& vtkwriter, std::string s) const
    {
      return typename VTKWriter<GV>::VTKFunctionPtr(
               new VTKGridFunctionWrapper<GV,RT,m>(*this,s));
    }

  private:
    const GridFunction<G,RT,m>& _fnkt;
    const IS & _is;
  };

  /**
     Access a Functions Gradient as a Function itself
   */
  template<class G, class RT>
  class GridFunctionGradient :
    virtual public GridFunction<G,RT,G::dimension>,
    virtual public GridFunctionDefault<G,RT,G::dimension>,
    virtual public FunctionDefault<typename G::ctype,RT,G::dimension,G::dimension>
  {
    //! get domain field type from the grid
    typedef typename G::ctype DT;

    //! get entity from the grid
    typedef typename G::template Codim<0>::Entity Entity;

    //! get domain dimension from the grid
    enum {n=G::dimension};

    //! set result dimension
    enum {m=G::dimension};

    //! get world dimension from the grid
    enum {dimw=G::dimensionworld};

    //! make copy constructor private
    GridFunctionGradient (const GridFunctionGradient&);
  public:
    //! create Adaptor from a Function
    GridFunctionGradient(const DifferentiableGridFunction<G,RT,1>& f) : _fnkt(f) {};

    // forward the evaluation to _fnkt
    virtual RT evallocal (int comp, const Entity& e, const Dune::FieldVector<DT,n>& xi) const
    {
      Dune::FieldVector<int,n> d(0);
      d[comp] = 1;
      return _fnkt.derivativelocal(0, d, e, xi);
    }

    // forward the evaluation to _fnkt
    virtual RT eval (int comp, const Dune::FieldVector<DT,n>& x) const
    {
      Dune::FieldVector<int,n> d(0);
      d[comp] = 1;
      return _fnkt.derivative(0, d, x);
    }

  private:
    const DifferentiableGridFunction<G,RT,1>& _fnkt;
  };

  /**
     Access a "- Function" as a Function itself
   */
  template<class G, class RT, int m>
  class GridFunctionMinus :
    virtual public GridFunction<G,RT,m>,
    virtual public GridFunctionDefault<G,RT,m>,
    virtual public FunctionDefault<typename G::ctype,RT,G::dimension,m>
  {
    //! get domain field type from the grid
    typedef typename G::ctype DT;

    //! get entity from the grid
    typedef typename G::template Codim<0>::Entity Entity;

    //! get domain dimension from the grid
    enum {n=G::dimension};

    //! get world dimension from the grid
    enum {dimw=G::dimensionworld};

    //! make copy constructor private
    GridFunctionMinus (const GridFunctionMinus&);
  public:
    //! create Adaptor from a Function
    GridFunctionMinus(const GridFunction<G,RT,m>& f) : _fnkt(f) {};

    // forward the evaluation to _fnkt
    virtual RT evallocal (int comp, const Entity& e, const Dune::FieldVector<DT,n>& xi) const
    {
      return - _fnkt.evallocal(comp, e, xi);
    }

    // forward the evaluation to _fnkt
    virtual RT eval (int comp, const Dune::FieldVector<DT,n>& x) const
    {
      return - _fnkt.eval(comp, x);
    }

  private:
    const GridFunction<G,RT,m>& _fnkt;
  };

} // end namespace Dune

#endif // DUNE_FUNCTION_ADAPTER_HH
