// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DG_FUNCTION_HH
#define DG_FUNCTION_HH

#include <dune/istl/bvector.hh>
#include <dune/grid/common/quadraturerules.hh>
#include "functions.hh"
#include "monomialshapefunctions.hh"
#include "orthonormalshapefunctions.hh"

/**
 * @file
 * @brief a class for discontinous galerkin element functions
 * @author Christian Engwer
 */

namespace Dune {

  /** @addtogroup DISC_Functions
   *
   * @{
   */

  //! class for DG finite element functions on a grid
  /*! This class implements the interface of a
        DifferentiableGridFunction with piecewise discontinous elements
        using a orthonormal basis. It is assumed that all elements use the
        same shapefunction set.

        In addition to the DifferentiableGridFunction
        interface. Dereferencing delivers the coefficient vector.

     \param GV The GridView associating the values of the funciton with
     according entities
     \param RT The type used for the component values of the function
     \param o The order of the monomial shapefunctions
     \param SFSC The type of the ShapeFunctionSetContainer
   */
  template<class GV, class RT, int o,
      template<class,class,int,int> class SFSC = MonomialShapeFunctionSetContainer>
  class DGFunction :
    virtual public GridFunction<typename GV::Grid,RT,1>,
    virtual public GridFunctionDefault<typename GV::Grid,RT,1>,
    virtual public GridFunctionGlobalEvalDefault<GV,RT,1>,
    virtual public FunctionDefault<typename GV::Grid::ctype,RT,GV::Grid::dimension,1>,
    virtual public H1Function<typename GV::Grid::ctype,RT,GV::Grid::dimension,1>,
    virtual public GridFunctionGlobalDifferentiationDefault<GV,RT,1>,
    virtual public L2Function<typename GV::Grid::ctype,RT,GV::Grid::dimension,1>
  {
    typedef typename GV::Grid G;
    typedef typename GV::IndexSet IS;

  protected:
    //! get domain field type from the grid
    typedef typename G::ctype DT;

    //! get entity from the grid
    typedef typename G::template Codim<0>::Entity Entity;

    //! get domain dimension from the grid
    enum {n=G::dimension};

    //! make copy constructor private
    DGFunction (const DGFunction&);


  public:
    //! type of ShapeFunctionSetContainer
    typedef SFSC<DT,RT,G::dimension,o> ShapeFunctionSetContainer;
    //! type of ShapeFunctionSet
    typedef typename ShapeFunctionSetContainer::ShapeFunctionSet ShapeFunctionSet;
  protected:
    //! size of local vector blocks
    static const int BlockSize = ShapeFunctionSetContainer::maxsize;
  public:
    typedef RT ResultType;
    typedef DT DomainType;
    typedef
    FieldVector<ResultType,BlockSize> BlockType;
    typedef BlockVector< BlockType > RepresentationType;

    //! allocate a vector with the data
    DGFunction (const GV& gv)
      : GridFunctionGlobalEvalDefault<GV,RT,1>(gv)
        , GridFunctionGlobalDifferentiationDefault<GV,RT,1>(gv)
        , gv_(gv)
    {
      coeff_.resize(gv.indexSet().size(0), false);
    }

    //! deallocate the vector
    ~DGFunction () {}

    //! evaluate single component comp in the entity e at local coordinates xi
    /*! Evaluate the function in an entity at local coordinates.
       @param[in]  comp   number of component to be evaluated
       @param[in]  e      reference to grid entity of codimension 0
       @param[in]  xi     point in local coordinates of the reference element of e
       \return            value of the component
     */
    virtual RT evallocal (int comp, const Entity& e, const Dune::FieldVector<DT,n>& xi) const
    {
      assert(comp == 0);

      RT value = 0;
      const ShapeFunctionSet & s = shapefnkts_(e.type(),o);
      int eid = gv_.indexSet().index(e);
      for (int i=0; i<BlockSize; ++i)
      {
        value += coeff_[eid][i] * s[i].evaluateFunction(0, xi);
      }
      return value;
    }

    virtual RT derivativelocal (int comp, const Dune::FieldVector<int,n>& d,
                                const Entity& e, const Dune::FieldVector<DT,n>& xi) const
    {
      assert(comp == 0);

      const ShapeFunctionSet & s = shapefnkts_(e.type(),o);

      typedef Dune::FieldVector<DT,n> Gradient;
      typedef Dune::FieldMatrix<DT,n,n> JacobianInverseMatrix;
      Gradient grad(0);
      Gradient grad_phi_ei;
      JacobianInverseMatrix invj;
      invj = e.geometry().jacobianInverseTransposed(xi);
      int eid = gv_.indexSet().index(e);

      for (int i=0; i<BlockSize; ++i)
      {
        Gradient tmp(0.0);
        for (int d=0; d<n; d++)
          tmp[d] = s[i].evaluateDerivative(0, d, xi);
        grad_phi_ei = 0;
        invj.umv(tmp,grad_phi_ei);
        for (int d=0; d<n; d++)
          grad[d] += grad_phi_ei[d] * coeff_[eid][i];
      }

      for (int u=0; u<n; u++)
        if (d[u] == 1) return grad[u];
      DUNE_THROW(NotImplemented,"Can not evaluate this derivative: d=" << d);
    }

    //! interpolate nodal values from a grid function
    /*! L2 Projection of a C0 function

       @param[in]  u    a continuous function
     */
    void interpolate (const C0Function<DT,RT,n,1>& u)
    {
      typedef typename GV::template Codim<0>::Iterator Iterator;

      Iterator it = gv_.template begin<0> ();
      const Iterator endit = gv_.template end<0> ();

      for(; it != endit; ++it)
      {
        FieldVector<RT,BlockSize> l2_b(0.0);
        FieldMatrix<RT,BlockSize,BlockSize> l2_A(0.0);

        const ShapeFunctionSet & s = shapefnkts_(it->type(),o);
        int eid = gv_.IndexSet().index(*it);
        coeff_[eid] = 0;

        // Get quadrature rule
        typedef QuadratureRule<DT,n> Quad;
        typedef typename QuadratureRule<DT,n>::const_iterator QuadIterator;
        const Quad& quad =
          QuadratureRules<DT,n>::rule(it->type(),
                                      o*getGeometryOrder(it->type()));

        for (QuadIterator q=quad.begin(); q!=quad.end(); ++q)
        {
          FieldVector<DT,n> c = q->position();
          FieldVector<DT,n> cg = it->geometry().global(c);
          for(int i=0; i<BlockSize; i++)
          {
            // b_i = \int\limits_{\Omega} exact * Phi_i dV
            RT det = it->geometry().integrationElement(c);
            RT value = u.eval(0, cg);
            RT phi_i = s[i].evaluateFunction(0, c);
            l2_b[i] += q->weight() * phi_i * value * det;
            for(int j=0; j<BlockSize; j++)
            {
              // A_{i,j} = \int\limits_{\Omega} Phi_i * Phi_j dV
              RT phi_j = s[j].evaluateFunction(0, c);
              l2_A[i][j] += q->weight() * phi_i * phi_j * det;
            }
          }
        }
        // solve localy
        fm_solve(l2_A, coeff_[eid], l2_b);
      }
    }

    //! return const reference to coefficient vector
    /*! Dereferencing a finite element function returns the
       coefficient representation of the finite element function.
       This is the const version.
     */
    const RepresentationType& operator* () const
    {
      return coeff_;
    }

    //! return reference to coefficient vector
    /*! Dereferencing a finite element function returns the
       coefficient representation of the finite element function.
       This is the non-const version.
     */
    RepresentationType& operator* ()
    {
      return coeff_;
    }

    /** empty method to maintain symmetry
        For vertex data nothing is required in preAdapt but for other
        finite element functions this method is necessary.
     */
    void preAdapt ()
    {
      DUNE_THROW(NotImplemented, "preAdapt()");
    }

    /** @brief Initiate update process

       Call this method after the grid has been adapted. The
       representation is now updated to the new grid and the finite
       element function can be used on the new grid. However the data
       is not initialized.  The old representation (with respect to
       the old grid) can still be accessed if it has been saved. It
       is deleted in endUpdate().
     */
    void postAdapt ()
    {
      DUNE_THROW(NotImplemented, "postAdapt()");
    }

    //! VTK output
    typename VTKWriter<GV>::VTKFunctionPtr
    vtkFunction (std::string s) const
    {
      return typename VTKWriter<GV>::VTKFunctionPtr(
               new VTKGridFunctionWrapper<GV,RT,1>(*this,s));
    }

  protected:
    int getGeometryOrder(const Dune::GeometryType & t)
    {
      if (t.isLine()) return 1;
      if (t.isSimplex()) return 1;
      if (t.isCube()) return t.dim();
      if (t.isPyramid()) return 2;
      if (t.isPrism()) return 2;
      DUNE_THROW(Dune::Exception,"Unknown GeometryType " << t);
    }

    // The GridView.  Don't use a reference here since Grid::levelView returns
    // a temporary
    const GV gv_;

    // reference to local shapefunctionset
    ShapeFunctionSetContainer shapefnkts_;

    // and a dynamically allocated vector
    RepresentationType coeff_;
  };

  /**
     \brief DG finite element function on the leaf grid

     \param G The grid
     \param RT The type used for the component values of the function
     \param o The order of the monomial shapefunctions
   */
  template<class G, class RT, int o,
      template<class,class,int,int> class SFSC = MonomialShapeFunctionSetContainer>
  class LeafDGFunction : public DGFunction<typename G::LeafGridView,RT,o,SFSC>
  {
  public:
    LeafDGFunction (const G& grid) :
      GridFunctionGlobalEvalDefault<typename G::LeafGridView,RT,1>(grid.leafView()),
      GridFunctionGlobalDifferentiationDefault<typename G::LeafGridView,RT,1>(grid.leafView()),
      DGFunction<typename G::LeafGridView,RT,o,SFSC>(grid.leafView())
    {}
  };

  /**
     \brief DG finite element function on a given level grid

     \param G The grid
     \param RT The type used for the component values of the function
     \param o The order of the monomial shapefunctions
   */
  template<class G, class RT, int o,
      template<class,class,int,int> class SFSC = MonomialShapeFunctionSetContainer>
  class LevelDGFunction : public DGFunction<typename G::LevelGridView,RT,o,SFSC>
  {
  public:
    LevelDGFunction (const G& grid, int level) :
      GridFunctionGlobalEvalDefault<typename G::LevelGridView,RT,1>(grid.levelView(level)),
      GridFunctionGlobalDifferentiationDefault<typename G::LevelGridView,RT,1>(grid.levelView(level)),
      DGFunction<typename G::LevelGridView,RT,o,SFSC>(grid.levelView(level))
    {}
  };

  /** @} */

}

#endif // DG_FUNCTION_HH
