// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_SIMPLEXSHAPEFUNCTIONS_HH
#define DUNE_SIMPLEXSHAPEFUNCTIONS_HH

#include <iostream>
#include <dune/common/fvector.hh>
#include <dune/common/exceptions.hh>
#include <dune/geometry/type.hh>

#include "cubeshapefunctions.hh"

/**
 * @file
 * @brief  define Lagrange type shape functions for simplicial elements
 * @author Sreejith Pulloor Kuttanikkad
 */

namespace Dune
{
  /** @addtogroup DISC_Shapefnkt
   *
   * @{
   */



  /***********************************************************
  * The interface for  Simplex shape functions of arbitrary
  * order and element type
  * now support only triangle and tetrahedron
  ***********************************************************/



  /***********************************************************
  * P0 shape functions for the simplex
  ***********************************************************/

  /*!
   * A class for piecewise constant shape functions in a simplex.
   * Evaluation is done at the centroid(Centre of gravity) of the simplex
   */


  template<typename C, typename T, int d>
  class P0SimplexShapeFunction
  {
  public:

    // compile time sizes
    enum { dim=d };        // maps from R^d
    enum { comps=1 };          // to R^1

    enum { m=1 };     // total number of basis functions

    // export types
    typedef C CoordType;
    typedef T ResultType;
    typedef P0SimplexShapeFunction ImplementationType;

    //! make a shape function object
    P0SimplexShapeFunction ()
    {
      for (int j=0; j<dim; j++)
        pos[j] = 1/(dim+1);

    }


    //! evaluate shape function in local coordinates
    ResultType evaluateFunction (int comp, const FieldVector<CoordType,d>& x) const
    {
      return 1;
    }

    //! evaluate gradient in local coordinates
    ResultType evaluateDerivative (int comp, int dir, const FieldVector<CoordType,d>& x) const
    {
      return 0;
    }

    //! consecutive number of associated dof within element
    int localindex (int comp) const
    {
      return 0;
    }

    //! codim of associated dof
    int codim () const
    {
      return 0;
    }

    //! entity (of codim) of associated dof
    int entity () const
    {
      return 0;
    }

    //! consecutive number of dof within entity
    int entityindex () const
    {
      return 0;
    }

    //! interpolation point associated with shape function
    const FieldVector<CoordType,dim>& position () const
    {
      return pos;
    }

  private:
    FieldVector<CoordType,d> pos;
  };


  template<typename C, typename T, int d, typename S>
  class P0SimplexShapeFunctionSet
  {
  public:

    // compile time sizes
    enum { dim=d };        // maps from R^d
    enum { comps=1 };          // to R^1

    enum { m=1 };     // total number of basis functions

    // export types
    typedef C CoordType;
    typedef T ResultType;
    typedef S value_type;
    typedef typename S::ImplementationType Imp;     // Imp is either S or derived from S

    //! make a shape function object
    P0SimplexShapeFunctionSet ()
    {
      sf = Imp();     // assignment of derived class objects defined in wrapper
    }

    //! return total number of shape functions
    int size () const
    {
      return 1;
    }

    //! total number of shape functions associated with entity in codim
    int size (int entity, int codim) const
    {
      if (codim==0) return 1;else return 0;
    }

    //! random access to shape functions
    const value_type& operator[] (int i) const
    {
      return sf;     // ok derived class reference goes for base class reference
    }

    //! return order
    int order () const
    {
      return 0;
    }

    //! return type of element
    GeometryType type () const
    {
      static GeometryType simplex(GeometryType::simplex, dim);
      return simplex;
    }

  private:
    S sf;
  };



  //! P0 shape functions in the simplex without virtual functions
  template<typename C, typename T, int d>
  class P0SimplexShapeFunctionSetContainer
  {
  public:
    // compile time sizes
    enum { dim=d };
    enum { comps=1 };
    enum { maxsize=1 };

    // exported types
    typedef C CoordType;
    typedef T ResultType;
    typedef P0SimplexShapeFunctionSet<C,T,d,P0SimplexShapeFunction<C,T,d> > value_type;

    const value_type& operator() (GeometryType type, int order) const
    {
      if (type.isSimplex()) return p0simplex;
      DUNE_THROW(NotImplemented, "type not implemented yet");
    }
  private:
    value_type p0simplex;
  };



  /***********************************************************
  * P1 shape functions for the simplex
  ***********************************************************/


  /*!
   * A class for piecewise linear shape functions in a simplex geometry(triangle/tetrehedron).
   * Evaluation is done at the quadrature points
   */
  template<typename C, typename T, int d>
  class P1SimplexShapeFunction
  {

  public:

    enum {dim=d}; // dim=2 or 3
    enum {comps=1};
    enum {m=dim+1}; // no of basis functions=3 or 4

    typedef C CoordType;
    typedef T ResultType;
    typedef P1SimplexShapeFunction ImplementationType;
    //triangle shape fn object
    // i'th shape function
    P1SimplexShapeFunction(int i)
    {
      number=i;

      if(i==0)
      {
        a[0]=1;
        for (int j=0; j<dim; ++j) {
          b[j]=-1;
          pos[j]=0;
        }
      }
      else
      {
        a[0]=0;
        for (int j=0; j<dim; ++j) {
          b[j]=0;
          pos[j]=0;
        }
        b[i-1]=1;
        pos[i-1]=1;
      }
    }
    //! must be defaultconstructible
    P1SimplexShapeFunction ()
    {}

    //! evaluate shape function in local coordinates
    ResultType evaluateFunction (int comp, const FieldVector<CoordType,d>& x) const
    {

      ResultType phi=a[0];
      for (int j=0; j<dim; ++j)
      {
        phi +=b[j]*x[j];
      }

      return phi;
    }


    ResultType evaluateDerivative (int comp, int dir, const FieldVector<CoordType,d>& x) const

    {

      ResultType deriv = b[dir];
      return deriv;
    }



    //! consecutive number of associated dof within element
    int localindex (int comp) const
    {
      return number;
    }

    //! codim of associated dof
    int codim () const
    {
      return dim;
    }

    //! entity (of codim) of associated dof
    int entity () const
    {
      return number;
    }

    //! consecutive number of dof within entity
    int entityindex () const
    {
      return 0;
    }

    //! interpolation point associated with shape function
    const FieldVector<CoordType,dim>& position () const
    {
      return pos;
    }

  private:
    int number;
    ResultType a[dim], b [dim];
    FieldVector<CoordType,d> pos;

  };


  template<typename C, typename T, int d, typename S>
  class P1SimplexShapeFunctionSet
  {

  public:
    enum {dim=d}; // dim=2 or 3
    enum {comps=1};
    enum {m=dim+1}; // no of basis functions=3 or 4

    typedef C CoordType;
    typedef T ResultType;
    typedef S value_type;
    typedef typename S::ImplementationType Imp; // Imp is either S or derived from S
    //! make a shape fn object
    P1SimplexShapeFunctionSet()
    {
      for (int i=0; i<m; i++)
        sf[i] = Imp(i); // assignment of derived class objects defined in wrapper
    }


    //! return total number of shape functions
    int size () const
    {
      return m;
    }
    //! total number of shape functions associated with entity in codim
    int size (int entity, int codim) const
    {
      if (codim==dim) return 1;else return 0;
    }
    //! random access to shape functions
    const value_type& operator[] (int i) const
    {
      return sf[i]; // ok derived class reference goes for base class reference
    }

    //! return order
    int order () const
    {
      return 1;
    }

    //! return type of element
    GeometryType type () const
    {
      static GeometryType simplex(GeometryType::simplex, dim);
      return simplex;
    }
  private:
    S sf[m];
  };

  //! These are P1 shape functions in the simplex without virtual functions
  template<typename C, typename T, int d>
  class P1SimplexShapeFunctionSetContainer
  {

  public:

    enum {dim=d};
    enum {comps=1};
    enum {maxsize=dim+1};

    // exported types
    typedef C CoordType;
    typedef T ResultType;
    typedef P1SimplexShapeFunctionSet<C,T,d,P1SimplexShapeFunction<C,T,d> > value_type;

    const value_type& operator() (GeometryType type, int order) const
    {
      if (type.isSimplex()) return p1simplex;
      DUNE_THROW(NotImplemented,"type not yet implemented");
    }
  private:
    value_type p1simplex;
  };

}
#endif
