// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_LEGENDRESHAPEFUNCTIONS_HH
#define DUNE_LEGENDRESHAPEFUNCTIONS_HH

#include <iostream>
#include <vector>

#include <dune/common/fvector.hh>
#include <dune/common/exceptions.hh>
#include <dune/common/stdstreams.hh>
#include <dune/common/misc.hh>
#include <dune/grid/common/grid.hh>

#include "shapefunctions.hh"

/**
   @file
   @brief  define legendre type shape functions
 */

namespace Dune
{

  /**
     @addtogroup DISC_Shapefnkt
     @{
   */

  enum {
    /**
       Default value for maximum order for the
       LegendreShapeFunctionSetContainer
     */
    LegendreShapeFunctionDefaultMaxOrder = 6
  };

  template<typename C,typename T, int dim>
  class LegendreShapeFunction;
  template<typename C,typename T, int dim>
  class LegendreShapeFunctionSet;
  template<typename C,typename T, int dim,
      int maxOrder=LegendreShapeFunctionDefaultMaxOrder>
  class LegendreShapeFunctionSetContainer;

  /**
     \class LegendreShapeFunction
     \brief multivariate monomials of the form x^a.y^b
     monomial order/total degree == p ; p <= (a+b)
   */
  template<typename C,typename T, int d>
  class LegendreShapeFunction : public ShapeFunction<C,T,d,1>
  {
  public:
    enum { dim=d /*!< Dimension of the ShapeFunction */ };
    typedef C CoordType;
    typedef T ResultType;

    /**
       Create the n-th LegendreShapeFunction of the ShapeFunctionSet of
       the given order.

       \param n this is the n-th shape functions in the set
       \param exp exponent a,b of x^a * y^b
     */
    LegendreShapeFunction(int n, const FieldVector<int, dim> & exp) :
      n_(n), exp_(exp)
    {

      for(int c=0; c<dim; ++c)
      {
        cof_[c].resize(exp[c]+1,0.0);

        switch(exp[c])
        {
        case 0 :
          cof_[c][0]= 1.0;
          break;
        case 1 :
          cof_[c][1]= 1.0;
          break;
        case 2 :
          cof_[c][0]=-0.5;
          cof_[c][2]= 1.5;
          break;
        case 3 :
          cof_[c][1]=-1.5;
          cof_[c][3]= 2.5;
          break;
        case 4 :
          cof_[c][0]= 0.375;
          cof_[c][2]=-3.75;
          cof_[c][4]= 4.375;
          break;
        case 5 :
          cof_[c][1]= 15.0/8.0;
          cof_[c][3]=-70.0/8.0;
          cof_[c][5]= 63.0/8.0;
          break;
        case 6 :
          cof_[c][0]=-  5.0/16.0;
          cof_[c][2]= 105.0/16.0;
          cof_[c][4]=-315.0/16.0;
          cof_[c][6]= 231.0/16.0;
          break;
        default :
          DUNE_THROW(Dune::NotImplemented,"Max order for Legendre Pol. is " << LegendreShapeFunctionDefaultMaxOrder);
          break;
        }
      }
    }

    /** \copydoc ShapeFunction::evaluateFunction */
    virtual ResultType
    evaluateFunction (int, const FieldVector<CoordType,dim>& x) const
    {

      FieldVector<CoordType,dim> p(-0.5);
      p += x;
      p *= 2.0;

      CoordType phi = 1.0;
      // phi= x^(alpha-beta) * y^beta
      // The compiler should be able to eliminate this loop
      for (int c=0; c<dim; c++)
      {
        CoordType tmp = 0.0;
        for(int i=0; i<exp_[c]+1; ++i)
        {
          tmp += cof_[c][i] * power(p[c],i);
        }

        phi *= tmp;
      }
      return phi;
    }

    /** \copydoc ShapeFunction::evaluateDerivative */
    virtual ResultType
    evaluateDerivative(int, int dir, const FieldVector<CoordType,dim>& x) const
    {

      FieldVector<CoordType,dim> p(-0.5);
      p += x;
      p *= 2.0;

      CoordType phi = 1.0;
      // phi= x^(alpha-beta) * y^beta
      // The compiler should be able to eliminate this loop
      for (int c=0; c<dim; c++)
      {
        CoordType tmp = 0.0;
        for(int i=0; i<exp_[c]+1; ++i)
        {
          if (c==dir)
            tmp += cof_[c][i] * (CoordType)i  * power(p[c],i-1);
          else
            tmp += cof_[c][i] * power(p[c],i);
        }

        phi *= tmp;
      }
      return phi * 2.0;
    }

    /** \copydoc ShapeFunction::localindex */
    virtual int localindex (int comp) const { return n_; };

    /** \copydoc ShapeFunction::codim */
    virtual int codim () const { return 0; };

    /** \copydoc ShapeFunction::entity */
    virtual int entity () const { return 0; };

    /** \copydoc ShapeFunction::entityindex */
    virtual int entityindex () const { return 0; };

    /** pretty print this shape function */
    void print (std::ostream& s) const
    {
      static const char names[] =
      { 'x', 'y', 'z', 'u', 'v', 'w', 'q', 'r', 's' };
      assert( dim < 9 ); /* avoid buffer overrun un names */;
      bool all_zero = true;
      for (int c=0; c<dim; c++)
      {
        s << "(";
        for(int i=0; i<exp_[c]+1; ++i)
        {
          if (cof_[c][i] != 0.0) {
            if (i == 0)
              s << cof_[c][i];
            else if (i == 1)
              s << cof_[c][i] << " " << names[c];
            else
              s << cof_[c][i] << " " << names[c] << "^" << i;

            if (i != exp_[c])
              s << " + ";

            all_zero = false;
          }
        }

        if (all_zero) {
          s << 1;
        }
        all_zero = true;
        s << ")";
      }
    };
  private:
    /** this is the n-th shape functions in the set */
    int n_;
    /** exponents a,b of x^a * y^b */
    FieldVector<int, dim> exp_;
    /** coefficients of monoms **/
    FieldVector<std::vector<CoordType>,dim> cof_;
    /** calculate the power x^p */
    ResultType power(CoordType x, int p) const {
      if (p <= 0)
        return 1.0;
      return x*power(x, p-1);
    }
  };

  template<typename C,typename T, int d>
  inline std::ostream& operator<<(std::ostream& os,
                                  const LegendreShapeFunction<C,T,d>& s)
  {
    s.print(os);
    return os;
  }

  /**
     \class LegendreShapeFunctionSetCreator
     \brief Template Metaprogramm creating the entries for
            LegendreShapeFunctionSet
     \internal
   */
  template <typename C,typename T, int dim, int c>
  struct LegendreShapeFunctionSetCreator
  {
    typedef LegendreShapeFunction<C,T,dim> ShapeFunction;
    typedef FieldVector<int, dim> expV;
    typedef std::vector< ShapeFunction* > sfV;
    static void create (int bound, int& cnt, expV& exp, sfV& SFnts)
    {
      for (int a=0; a<=bound; a++)
      {
        exp[c-1] = bound-a;
        LegendreShapeFunctionSetCreator<C,T,dim,c+1>::create(a,cnt,exp,SFnts);
      }
    }
  };

  /**
     \copydoc LegendreShapeFunctionSetCreator
     \brief Specializes the beginning of the recursion
     \internal
   */
  template <typename C,typename T, int dim>
  struct LegendreShapeFunctionSetCreator<C,T,dim,0>
  {
    typedef LegendreShapeFunction<C,T,dim> ShapeFunction;
    typedef FieldVector<int, dim> expV;
    typedef std::vector< ShapeFunction* > sfV;
    static int create (int order, sfV& SFnts)
    {
      expV exp(0);
      int cnt = 0;
      for (int a=0; a<=order; a++)
      {
        LegendreShapeFunctionSetCreator<C,T,dim,1>::create(a,cnt,exp,SFnts);
      }
      return cnt;
    }
  };

  /**
     \copydoc LegendreShapeFunctionSetCreator
     \brief Specializes the end of the recursion
     \internal
   */
  template <typename C,typename T, int dim>
  struct LegendreShapeFunctionSetCreator<C,T,dim,dim>
  {
    typedef LegendreShapeFunction<C,T,dim> ShapeFunction;
    typedef FieldVector<int, dim> expV;
    typedef std::vector< ShapeFunction* > sfV;
    static void create (int bound, int& cnt, expV& exp, sfV& SFnts)
    {
      exp[dim-1] = bound;
      SFnts[cnt] = new ShapeFunction(cnt, exp);
      for (int i=0; i<dim; i++)
        dvverb << exp[i] << " ";
      dvverb << std::endl;
      cnt++;
    }
  };

  /**
     \class LegendreShapeFunctionSet
     \brief a ShapeFunctionSet of monomial shapefunctions

     The shape functions map from C^d to T.

     template parameters are:
     <ul>
     <li>C: coordinate type
     <li>T: result type
     <li>d: dimension of the function space
     </ul>

     The order of the shape functions is a parameter to the constructor
   */
  template<typename C,typename T, int d>
  class LegendreShapeFunctionSet : public ShapeFunctionSet<C, T, d, 1>
  {
  public:
    enum { dim=d /*!< Dimension of the ShapeFunctionSet */ };
    typedef C CoordType;
    typedef T ResultType;
    typedef LegendreShapeFunction<C,T,dim> ShapeFunction;

    /**
       Construct a monomial shape function set of given order
     */
    LegendreShapeFunctionSet(int order) :
      order_(order), n_(sz(order)), shapeFunctions(n_)
    {
      dvverb << "Constructing legendre shape function set of order "
             << order << std::endl;
#ifndef  NDEBUG
      size_t i =
#endif
      LegendreShapeFunctionSetCreator<C,T,dim,0>::
      create(order,shapeFunctions);
      assert (n_==i);
    }

    virtual ~LegendreShapeFunctionSet()
    {
      for (size_t i=0; i < n_; i++)
      {
        if (shapeFunctions[i]) delete shapeFunctions[i];
      }
    }

    //! \copydoc ShapeFunctionSet::size
    virtual int size() const
    {
      assert(shapeFunctions.size() == n_);
      return n_;
    }

    //! \copydoc ShapeFunctionSet::size(int, int)
    virtual int size (int entity, int codim) const
    {
      if (codim == 0) return n_;
      return 0;
    };

    //! \copydoc ShapeFunctionSet::operator[]
    virtual const ShapeFunction& operator[](int i) const
    {
      assert (i < (int)n_);
      assert (n_ == shapeFunctions.size());
      return *shapeFunctions[i];
    }

    //! \copydoc ShapeFunctionSet::order
    virtual int order() const
    {
      return order_;
    }

    //! \copydoc ShapeFunctionSet::type
    virtual GeometryType type () const
    {
      //! we dont' have a special geometry, so we return unitsquare
      static GeometryType t(GeometryType::cube, d);
      return t;
    };

  private:
    /** order of the ShapeFunctionSet */
    const int order_;
    /** total number of shapefunctions in this set */
    const size_t n_;
    /** \brief calculate total number of shapefunctions in a set of given order */
    static int sz(int order)
    {
      int s;
      s = 1;
      for (int i=1; i<=dim; i++) {
        s = s * (order + i) / i;
      }
      return s;
    }
    /** vector which holds the shape functions */
    std::vector< ShapeFunction* > shapeFunctions;
  };

  /**
     \class LegendreShapeFunctionSetSize
     \brief calculates the size of a LegendreShapeFunctionSet
     \internal
   */
  template <int dim, int order>
  struct LegendreShapeFunctionSetSize
  {
    enum
    {
      maxSize =
        LegendreShapeFunctionSetSize<dim-1,order>::maxSize
        * (order+dim) / dim
    };
  };

  /**
     \copydoc LegendreShapeFunctionSetSize
     \brief Specializeses the end of the recursion
     \internal
   */
  template <int order>
  struct LegendreShapeFunctionSetSize<1,order>
  {
    enum { maxSize = order + 1 };
  };

  /**
     \class LegendreShapeFunctionSetContainer
     \brief a ShapeFunctionSetContainer of monomial shapefunctions

     The shape functions map from C^d to T.

     template parameters are:
     <ul>
     <li>C: coordinate type
     <li>T: result type
     <li>d: dimension of the function space
     <li>maxOrder: container will contain shapefunctions up to order maxOrder
     </ul>
   */
  template<typename C,typename T, int d, int maxOrder>
  class LegendreShapeFunctionSetContainer :
    public ShapeFunctionSetContainer<C,T,d,1,LegendreShapeFunctionSetSize<d,maxOrder>::maxSize >
  {
  public:
    enum { dim = d /*!< Dimension of the ShapeFunctionSetContainer */ };
    enum { maxsize = LegendreShapeFunctionSetSize<d,maxOrder>::maxSize };
    typedef C CoordType;
    typedef T ResultType;
    typedef LegendreShapeFunctionSet<C,T,dim> ShapeFunctionSet;

    /**
       \brief Create a ShapeFunctionSetContainer of monomial shapefunctions
     */
    LegendreShapeFunctionSetContainer() :
      shapeFunctionSets(maxsize)
    {
      for (int order=0; order<=maxOrder; order++)
      {
        shapeFunctionSets[order] =
          new ShapeFunctionSet(order);
      }
      for (int order=maxOrder+1; order<maxsize; order++)
        shapeFunctionSets[order] = 0;
    }

    virtual ~LegendreShapeFunctionSetContainer()
    {
      for (int i=0; i < maxsize; i++)
      {
        if (shapeFunctionSets[i]) delete shapeFunctionSets[i];
      }
    }

    /** \copydoc ShapeFunctionSetContainer::operator()(GeometryType type, int order) */
    virtual const ShapeFunctionSet& operator() (GeometryType type, int order) const
    {
      assert(order <= maxOrder);
      return *shapeFunctionSets[order];
    }
  private:
    /** vector which holds the shape functions */
    std::vector< ShapeFunctionSet* > shapeFunctionSets;
  };

  /** @} */

} //end of namespace


#endif // DUNE_LEGENDRESHAPEFUNCTIONS_HH
