// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_ORTHONORMALSHAPEFUNCTIONS_HH
#define DUNE_ORTHONORMALSHAPEFUNCTIONS_HH

#include <iostream>
#include <vector>
#include <map>
#include <dune/common/fvector.hh>
#include <dune/common/exceptions.hh>
#include <dune/common/stdstreams.hh>
#include <dune/common/misc.hh>
#include <dune/grid/common/grid.hh>

#include "shapefunctions.hh"

/**
   @file
   @brief  define monomial type shape functions
   @author Christian Engwer
 */

#define PMAX 8

namespace Dune
{

  /**
     @addtogroup DISC_Shapefnkt
     @{
   */

  template<typename C,typename T, int dim>
  class OrthonormalShapeFunction;
  template<typename C,typename T, int dim>
  class OrthonormalShapeFunctionSet;
  template<typename C,typename T, int dim,
      int maxOrder=PMAX>
  class OrthonormalShapeFunctionSetContainer;

  /**
     \class OrthonormalShapeFunctionSetSize
     \brief calculates the size of a OrthonormalShapeFunctionSet
     \internal
   */
  template <int dim, int order>
  struct OrthonormalShapeFunctionSetSize
  {
    enum
    {
      maxSize =
        OrthonormalShapeFunctionSetSize<dim-1,order>::maxSize
        * (order+dim) / dim
    };
  };

  /**
     \copydoc OrthonormalShapeFunctionSetSize
     \brief Specializeses the end of the recursion
     \internal
   */
  template <int order>
  struct OrthonormalShapeFunctionSetSize<1,order>
  {
    enum { maxSize = order + 1 };
  };

  /**
     \class OrthonormalShapeFunction
     \brief multivariate monomials of the form x^a.y^b
     monomial order/total degree == p ; (a+b) <= p
   */
  template<typename C,typename T, int d>
  class OrthonormalShapeFunction : public ShapeFunction<C,T,d,1>
  {
  public:
    enum { dim=d /*!< Dimension of the ShapeFunction */ };
    typedef C CoordType;
    typedef T ResultType;
    typedef FieldVector<C,dim> VectorType;
    typedef FieldVector<T,dim> GradientType;
    typedef OrthonormalShapeFunctionSet<C,T,dim> ShapeFunctionSet;

    /**
       Create the n-th OrthonormalShapeFunction of the ShapeFunctionSet of
       the given order.

       \param n this is the n-th shape functions in the set
       \param s the ShapeFunctionSet
     */
    OrthonormalShapeFunction(int n, const ShapeFunctionSet & s) :
      n_(n), s_(s) {}

    /** \copydoc ShapeFunction::evaluateFunction */
    virtual ResultType
    evaluateFunction (int, const FieldVector<CoordType,dim>& x) const
    {
      return s_.evaluateFunction(n_, x);
    }

    /** \copydoc ShapeFunction::evaluateDerivative */
    virtual ResultType
    evaluateDerivative(int, int dir, const FieldVector<CoordType,dim>& x) const
    {
      GradientType g;
      s_.evaluateGradient(n_, x, g);
      return g[dir];
    }

    /** \copydoc ShapeFunction::localindex */
    virtual int localindex (int comp) const { return n_; };

    /** \copydoc ShapeFunction::codim */
    virtual int codim () const { return 0; };

    /** \copydoc ShapeFunction::entity */
    virtual int entity () const { return 0; };

    /** \copydoc ShapeFunction::entityindex */
    virtual int entityindex () const { return 0; };

  private:
    /** this is the n-th shape functions in the set */
    int n_;
    /** this is the ShapeFunctionSet this shape function belongs to */
    const ShapeFunctionSet & s_;
  };

  /**
     \class OrthonormalShapeFunctionSet
     \brief a ShapeFunctionSet of orthonormal shapefunctions

     The shape functions map from C^d to T.

     template parameters are:
     <ul>
     <li>C: coordinate type
     <li>T: result type
     <li>d: dimension of the function space
     </ul>

     The order of the shape functions is a parameter to the constructor
   */
  template<typename C,typename T, int d>
  class OrthonormalShapeFunctionSet : public ShapeFunctionSet<C, T, d, 1>
  {
  public:
    enum { dim=d /*!< Dimension of the ShapeFunctionSet */ };
    typedef C CoordType;
    typedef T ResultType;
    typedef FieldVector<C,dim> VectorType;
    typedef FieldVector<T,dim> GradientType;
    typedef OrthonormalShapeFunction<C,T,dim> ShapeFunction;

    /**
       Construct a monomial shape function set of given order
     */
    OrthonormalShapeFunctionSet(const GeometryType & type, int order) :
      type_(type), order_(order), n_(sz(order)), shapeFunctions(n_)
    {
      for (int i=0; i < n_; i++)
      {
        shapeFunctions[i] = new ShapeFunction(i,*this);
      }
    }

    virtual ~OrthonormalShapeFunctionSet()
    {
      for (int i=0; i < n_; i++)
      {
        if (shapeFunctions[i]) delete shapeFunctions[i];
      }
    }

    //! \copydoc ShapeFunctionSet::size
    virtual int size() const
    {
      assert ((size_t)n_ == shapeFunctions.size());
      return n_;
    }

    //! \copydoc ShapeFunctionSet::size(int, int)
    virtual int size (int entity, int codim) const
    {
      if (codim == 0) return n_;
      return 0;
    };

    //! \copydoc ShapeFunctionSet::operator[]
    virtual const ShapeFunction& operator[](int i) const
    {
      assert (i < n_);
      assert ((size_t)n_ == shapeFunctions.size());
      return *shapeFunctions[i];
    }

    //! \copydoc ShapeFunctionSet::order
    virtual int order() const
    {
      return order_;
    }

    //! \copydoc ShapeFunctionSet::type
    virtual GeometryType type () const
    {
      return type_;
    };

    /** \copydoc ShapeFunction::evaluateFunction */
    virtual ResultType
    evaluateFunction (int i, const FieldVector<CoordType,dim>& x) const
    {
      if (type_.isTriangle())
        return eval_triangle_2d (i,x);
      if (type_.isQuadrilateral())
        return eval_quadrilateral_2d (i,x);
      if (type_.isTetrahedron())
        return eval_tetrahedron_3d (i,x);
      if (type_.isPyramid())
        return eval_pyramid_3d (i,x);
      if (type_.isPrism())
        return eval_prism_3d (i,x);
      if (type_.isHexahedron())
        return eval_hexahedron_3d (i,x);
      DUNE_THROW(Exception, "unhandled geometry type: " << type_);
    }

    /** \copydoc ShapeFunction::evaluateDerivative */
    virtual void
    evaluateGradient(int i, const FieldVector<CoordType,dim>& x, GradientType & g) const
    {
      if (type_.isTriangle())
        return grad_triangle_2d (i,x,g);
      if (type_.isQuadrilateral())
        return grad_quadrilateral_2d (i,x,g);
      if (type_.isTetrahedron())
        return grad_tetrahedron_3d (i,x,g);
      if (type_.isPyramid())
        return grad_pyramid_3d (i,x,g);
      if (type_.isPrism())
        return grad_prism_3d (i,x,g);
      if (type_.isHexahedron())
        return grad_hexahedron_3d (i,x,g);
      DUNE_THROW(Exception, "unhandled geometry type: " << type_);
    }

  private:
    const GeometryType type_;
    /** order of the ShapeFunctionSet */
    const int order_;
    /** total number of shapefunctions in this set */
    const int n_;
    /** \brief calculate total number of shapefunctions in a set of given order */
    static int sz(int order)
    {
      int s;
      s = 1;
      for (int i=1; i<=dim; i++) {
        s = s * (order + i) / i;
      }
      return s;
    }
    /** vector which holds the shape functions */
    std::vector< ShapeFunction* > shapeFunctions;

    double eval_triangle_2d ( int i, const VectorType & xi ) const;
    double eval_quadrilateral_2d ( int i, const VectorType & xi ) const;
    double eval_tetrahedron_3d ( int i, const VectorType & xi ) const;
    double eval_pyramid_3d ( int i, const VectorType & xi ) const;
    double eval_prism_3d ( int i, const VectorType & xi ) const;
    double eval_hexahedron_3d ( int i, const VectorType & xi ) const;
    void grad_triangle_2d ( int i, const VectorType & xi, GradientType & grad ) const;
    void grad_quadrilateral_2d ( int i, const VectorType & xi, GradientType & grad ) const;
    void grad_tetrahedron_3d ( int i, const VectorType & xi, GradientType & grad ) const;
    void grad_pyramid_3d ( int i, const VectorType & xi, GradientType & grad ) const;
    void grad_prism_3d ( int i, const VectorType & xi, GradientType & grad ) const;
    void grad_hexahedron_3d ( int i, const VectorType & xi, GradientType & grad ) const;

  };

  /**
     \class OrthonormalShapeFunctionSetContainer
     \brief a ShapeFunctionSetContainer of orthonormal shapefunctions

     The shape functions map from C^d to T.

     template parameters are:
     <ul>
     <li>C: coordinate type
     <li>T: result type
     <li>d: dimension of the function space
     <li>maxOrder: container will contain shapefunctions up to order maxOrder
     </ul>
   */
  template<typename C,typename T, int d, int maxOrder>
  class OrthonormalShapeFunctionSetContainer :
    public ShapeFunctionSetContainer<C,T,d,1,OrthonormalShapeFunctionSetSize<d,maxOrder>::maxSize >
  {
  public:
    enum { dim = d /*!< Dimension of the ShapeFunctionSetContainer */ };
    typedef C CoordType;
    typedef T ResultType;
    typedef FieldVector<C,dim> VectorType;
    typedef FieldVector<T,dim> GradientType;
    typedef OrthonormalShapeFunctionSet<C,T,dim> ShapeFunctionSet;

    /**
       \brief Create a ShapeFunctionSetContainer of monomial shapefunctions
     */
    OrthonormalShapeFunctionSetContainer()
    {
      assert (maxOrder <= PMAX);
    }

    virtual ~OrthonormalShapeFunctionSetContainer()
    {
      typename
      std::map< std::pair<GeometryType,int>, ShapeFunctionSet* >::iterator
      it = shapeFunctionSets.begin();
      for (; it != shapeFunctionSets.end(); ++it)
      {
        delete it->second;
      }
    }

    /** \copydoc ShapeFunctionSetContainer::operator()(GeometryType type, int order) */
    virtual const ShapeFunctionSet& operator() (GeometryType type, int order) const
    {
      assert(order <= maxOrder);
      std::pair<GeometryType, int> key(type, order);
      typename
      std::map< std::pair<GeometryType,int>, ShapeFunctionSet* >::iterator
      it = shapeFunctionSets.find(key);
      if (it == shapeFunctionSets.end())
      {
        ShapeFunctionSet *sf = new ShapeFunctionSet(type, order);
        shapeFunctionSets[key] = sf;
        return *sf;
      }
      return *(it->second);
    }

  private:
    /** vector which holds the shape functions */
    mutable std::map< std::pair<GeometryType,int>, ShapeFunctionSet* > shapeFunctionSets;
  };

  /** @} */

} //end of namespace

#include "orthonormalshapefunctionsset.cc"

#undef PMAX

#endif // DUNE_ORTHONORMALSHAPEFUNCTIONS_HH
