// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#include "udgfunction.hh"

/**
 * @file
 * @brief a class for discontinous galerkin element functions
 * @author Christian Engwer
 */

namespace Dune {

  //////////////////////////////////////////////////////////////////////////
  //
  //  Interpolation
  //

  //
  //  Interpolation of C0Functions
  //

  template<class GV, class T, class RT, int o,
      template<class,class,int,int> class SFSC, class M>
  void UDGFunction<GV, T, RT, o, SFSC, M>::
  localInterpolate( // map BBox -> global
    const BBoxGeometry &bboxgeometry,
    // map Integration Part -> BBox
    const IPGeometry &ipgeometry,
    FieldMatrix<RT,BlockSize,BlockSize>& Aee,
    FieldVector<RT,BlockSize> &Be,
    const C0Function<DT,RT,n,1> &u)
  {
    int ipGeometryOrder = getGeometryOrder(ipgeometry.type());
    // get shapefunction set
    const ShapeFunctionSet &base = shapefnkts_(bboxgeometry.type(), o);
    assert(BlockSize == base.size());

    // in the matrix I have phi_i*phi_j, each may have order o
    // in the vector I have phi_i*u, u is to be projected onto a function of order o
    // so this is 2*o in both cases
    enum { quadOrder = o * 2 };
    // the following is explained in doc/quadrature-with-udg.tex
    // [integrandOrder + 1] * ipGeomOrder - 1
    int order = (quadOrder+1) * ipGeometryOrder - 1;

    // coordinates within integration part
    FieldVector<DT,n> ci(0.0);
    // coordinates within bbox
    FieldVector<DT,n> cb(0.0);
    // global coordinates
    FieldVector<DT,n> cg(0.0);

    // Get quadrature rule
    typedef QuadratureRule<DT,n> Quad;
    typedef typename Quad::const_iterator QuadIterator;
    const Quad quad = QuadratureRules<DT, n>::
                      rule(ipgeometry.type(), order);

    // for quadrature points
    const QuadIterator qend = quad.end();
    for (QuadIterator q=quad.begin(); q != qend; ++q) {
      // coordinates within integration part
      ci = q->position();
      // coordinates within bbox
      cb = ipgeometry.global(ci);
      // global coordinates
      cg = bboxgeometry.global(cb);

      // total integration element
      RT det = bboxgeometry.integrationElement(cb)
               * ipgeometry.integrationElement(ci);
      // evaluate the C0 function
      RT value = u.eval(0, cg);

      for(int i=0; i<BlockSize; i++) {
        // b_i = \int\limits_{\Omega} exact * Phi_i dV
        RT phi_i = base[i].evaluateFunction(0, cb);
        Be[i] += q->weight() * phi_i*value * det;

        for(int j=0; j<BlockSize; j++) {
          // A_{i,j} = \int\limits_{\Omega} Phi_i * Phi_j dV
          RT phi_j = base[j].evaluateFunction(0, cb);
          Aee[i][j] += q->weight() * phi_i*phi_j * det;
        } // j
      } // i
    } // for quadrature points
  }

  template<class GV, class T, class RT, int o,
      template<class,class,int,int> class SFSC, class M>
  void UDGFunction<GV, T, RT, o, SFSC, M>::
  interpolate (const C0Function<DT,RT,n,1>& u)
  {
    // base functions are not orthogonal
    // => need to solve linear equation system
    FieldVector<RT,BlockSize> l2_b(0.0);
    FieldMatrix<RT,BlockSize,BlockSize> l2_A(0.0);

    typedef typename GV::template Codim<0>::Iterator Iterator;

    Iterator it = gv_.template begin<0> ();
    const Iterator endit = gv_.template end<0> ();

    // for elements
    for(; it != endit; ++it) {
      l2_b = 0.0;
      l2_A = 0.0;

      // get element id
      int eid = mapper.map(*it);
      // initialize elements coefficients
      coeff_[eid] = 0;

      // create list of integration parts
      std::list<EntityPart> iParts;
      iParts.clear();
      t_.create_geometries(it, iParts);

      // use zero coefficients if element is complety outside the domain
      if (iParts.empty())
        continue;

      // sum over all integration parts
      typename std::list<EntityPart>::iterator ipend = iParts.end();
      typename std::list<EntityPart>::iterator ipit = iParts.begin();

      // TODO: ignore cells which are to small??? (is this neccesary?)

      // for integration parts
      for(assert(ipit == iParts.begin()); ipit != ipend; ++ipit) {
        localInterpolate(t_.BBox(it), ipit->rectLocal(), l2_A, l2_b, u);
      } // for integration parts
        // solve localy
      l2_A.solve(coeff_[eid], l2_b);
    } // for elements
  }

  //
  //  Interpolation of GridFunctions
  //

  template<class GV, class T, class RT, int o,
      template<class,class,int,int> class SFSC, class M>
  void UDGFunction<GV, T, RT, o, SFSC, M>::
  localInterpolate(const Entity &e,
                   // map BBox -> global
                   const BBoxGeometry &bboxgeometry,
                   // map Integration Part -> BBox
                   const IPGeometry &ipgeometry,
                   FieldMatrix<RT,BlockSize,BlockSize>& Aee,
                   FieldVector<RT,BlockSize> &Be,
                   const GridFunction<G,RT,1> &u)
  {
    int ipGeometryOrder = getGeometryOrder(ipgeometry.type());
    // get shapefunction set
    const ShapeFunctionSet &base = shapefnkts_(bboxgeometry.type(), o);
    assert(BlockSize == base.size());

    // in the matrix I have phi_i*phi_j, each may have order o
    // in the vector I have phi_i*u, u is to be projected onto a function of order o
    // so this is 2*o in both cases
    enum { quadOrder = o * 2 };
    // the following is explained in doc/quadrature-with-udg.tex
    // [integrandOrder + 1] * ipGeomOrder - 1
    int order = (quadOrder+1) * ipGeometryOrder - 1;

    // coordinates within integration part
    FieldVector<DT,n> ci(0.0);
    // coordinates within bbox
    FieldVector<DT,n> cb(0.0);
    // coordinates within reference element
    FieldVector<DT,n> ce(0.0);
    // global coordinates
    FieldVector<DT,n> cg(0.0);

    // Get quadrature rule
    typedef QuadratureRule<DT,n> Quad;
    typedef typename Quad::const_iterator QuadIterator;
    const Quad quad = QuadratureRules<DT, n>::
                      rule(ipgeometry.type(), order);

    // for quadrature points
    const QuadIterator qend = quad.end();
    for (QuadIterator q=quad.begin(); q != qend; ++q) {
      // coordinates within integration part
      ci = q->position();
      // coordinates within bbox
      cb = ipgeometry.global(ci);
      // global coordinates
      cg = bboxgeometry.global(cb);
      // reference element coordinates
      ce = e.geometry().local(cg);

      // total integration element
      RT det = bboxgeometry.integrationElement(cb)
               * ipgeometry.integrationElement(ci);
      // evaluate the C0 function
      RT value = u.evallocal(0, e, ce);

      for(int i=0; i<BlockSize; i++) {
        // b_i = \int\limits_{\Omega} exact * Phi_i dV
        RT phi_i = base[i].evaluateFunction(0, cb);
        Be[i] += q->weight() * phi_i*value * det;

        for(int j=0; j<BlockSize; j++) {
          // A_{i,j} = \int\limits_{\Omega} Phi_i * Phi_j dV
          RT phi_j = base[j].evaluateFunction(0, cb);
          Aee[i][j] += q->weight() * phi_i*phi_j * det;
        } // j
      } // i
    } // for quadrature points
  }

  template<class GV, class T, class RT, int o,
      template<class,class,int,int> class SFSC, class M>
  void UDGFunction<GV, T, RT, o, SFSC, M>::
  interpolate (const GridFunction<G,RT,1>& u)
  {
    // base functions are not orthogonal
    // => need to solve linear equation system
    FieldVector<RT,BlockSize> l2_b(0.0);
    FieldMatrix<RT,BlockSize,BlockSize> l2_A(0.0);

    typedef typename GV::template Codim<0>::Iterator Iterator;

    Iterator it = gv_.template begin<0> ();
    const Iterator endit = gv_.template end<0> ();

    // for elements
    for(; it != endit; ++it) {
      l2_b = 0.0;
      l2_A = 0.0;

      // get element id
      int eid = mapper.map(*it);
      // initialize elements coefficients
      coeff_[eid] = 0;

      // create list of integration parts
      std::list<EntityPart> iParts;
      iParts.clear();
      t_.create_geometries(it, iParts);

      // use zero coefficients if element is complety outside the domain
      if (iParts.empty())
        continue;

      // sum over all integration parts
      typename std::list<EntityPart>::iterator ipend = iParts.end();
      typename std::list<EntityPart>::iterator ipit = iParts.begin();

      // TODO: ignore cells which are to small??? (is this neccesary?)

      // for integration parts
      for(assert(ipit == iParts.begin()); ipit != ipend; ++ipit) {
        localInterpolate(*it, t_.BBox(it), ipit->rectLocal(), l2_A, l2_b, u);
      } // for integration parts
        // solve localy
      l2_A.solve(coeff_[eid], l2_b);
    } // for elements
  }

} // namespace Dune
