// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef SCALED_DG_FUNCTION_HH
#define SCALED_DG_FUNCTION_HH

#include <dune/grid/common/mcmgmapper.hh>

#include <dune/udg/common/cuboid.hh>
#include "dgfunction.hh"
#include "../sparsitysetup.hh"

/**
 * @file
 * @brief a class for discontinous galerkin element functions
 * @author Christian Engwer

   \todo implement element->bbox mapping to speedup evallocal

 */

namespace Dune {

  /** @addtogroup DISC_Functions
   *
   * @{
   */

  //! class for DG finite element functions on a grid
  /*! This class implements the interface of a
     DifferentiableGridFunction with piecewise discontinous elements
     using a monomial basis. It is assumed that all elements use the
     same shapefunction set.

     In addition to the DifferentiableGridFunction
     interface. Dereferencing delivers the coefficient vector.

     \param GV The GridView associating the values of the funciton with
              according entities
     \param RT The type used for the component values of the function
     \param o The order of the monomial shapefunctions
   */
  template<class GV, class T, class RT, int o,
      template<class,class,int,int> class SFSC = MonomialShapeFunctionSetContainer,
      class M = MultipleCodimMultipleGeomTypeMapper< GV, CellLayout > >
  class UDGFunction :
    public DGFunction<GV,RT,o,SFSC>
  {
    typedef typename GV::Grid G;
    typedef typename GV::IndexSet IS;
    typedef M Mapper;
    typedef typename GV::Grid Grid;
    typedef typename GV::IndexSet IndexSet;
    typedef MultipleCodimMultipleGeomTypeMapper< GV, CellLayout > MCMGMapper;

    //! get domain field type
    typedef typename DGFunction<GV,RT,o,SFSC>::DT DT;

    //! get entity type
    typedef typename DGFunction<GV,RT,o,SFSC>::Entity Entity;

    //! get domain dimension
    enum { n = DGFunction<GV,RT,o,SFSC>::n };

    //! bbox type
    typedef Dune::Cuboid<n, DT> Cuboid;
    typedef Dune::GridCuboid<G> BBoxGeometry;

    //! make copy constructor private
    UDGFunction (const UDGFunction&);

  public:
    //! type of ShapeFunctionSetContainer
    typedef typename DGFunction<GV,RT,o,SFSC>::ShapeFunctionSetContainer
    ShapeFunctionSetContainer;
    //! type of ShapeFunctionSet
    typedef typename DGFunction<GV,RT,o,SFSC>::ShapeFunctionSet
    ShapeFunctionSet;
  private:
    //! size of local vector blocks
    static const int BlockSize = DGFunction<GV,RT,o,SFSC>::BlockSize;
  public:
    typedef RT ResultType;
    typedef DT DomainType;
    typedef T SubTriangulation;
    typedef FieldVector<ResultType,BlockSize> BlockType;
    typedef BlockVector< BlockType > RepresentationType;

    //! allocate a vector with the data
    UDGFunction (const GV& gv, const T& t)
      : GridFunctionGlobalEvalDefault<GV,RT,1>(gv)
        , GridFunctionGlobalDifferentiationDefault<GV,RT,1>(gv)
        , DGFunction<GV,RT,o,SFSC>(gv), t_(t)
        , CubeType(Dune::GeometryType::cube,n)
        , mcmgmapper(gv)
        , mapper(mcmgmapper)
    {}

    UDGFunction (const GV& gv, const T& t, const Mapper & m)
      : GridFunctionGlobalEvalDefault<GV,RT,1>(gv)
        , GridFunctionGlobalDifferentiationDefault<GV,RT,1>(gv)
        , DGFunction<GV,RT,o,SFSC>(gv), t_(t)
        , CubeType(Dune::GeometryType::cube,n)
        , mcmgmapper(gv)
        , mapper(m)
    {}

    //! deallocate the vector
    ~UDGFunction () {}

    //! evaluate single component comp in the entity e at local coordinates xi
    /*! Evaluate the function in an entity at local coordinates.
       @param[in]  comp   number of component to be evaluated
       @param[in]  e      reference to grid entity of codimension 0
       @param[in]  xi     point in local coordinates of the reference element of e
       \return            value of the component
     */
    virtual RT evallocal (int comp, const Entity& e, const Dune::FieldVector<DT,n>& xlocal) const
    {
      assert(comp == 0);

      FieldVector<DT,n> xbbox = bboxlocal(e, xlocal);

      RT value = 0;
      // get the shapefunctions, they live on the bounding box, thus
      // the geometry type is a cube
      const ShapeFunctionSet & s = this->shapefnkts_(CubeType,o);
      int eid = mapper.map(e);
      for (int i=0; i<BlockSize; ++i)
      {
        value += this->coeff_[eid][i] * s[i].evaluateFunction(0, xbbox);
      }
      return value;
    }

    virtual RT derivativelocal (int comp, const Dune::FieldVector<int,n>& d,
                                const Entity& e, const Dune::FieldVector<DT,n>& xi) const
    {
      assert(comp == 0);

      const ShapeFunctionSet & s = this->shapefnkts_(e.geometry().type(),o);

      FieldVector<DT,n> xr = bboxlocal(e, xi);

      typedef Dune::FieldVector<DT,n> Gradient;
      typedef Dune::FieldMatrix<DT,n,n> JacobianInverseMatrix;
      Gradient grad(0);
      Gradient grad_phi_ei;
      JacobianInverseMatrix invj;
      Cuboid bbox = t_.BBox(e);
      invj = bbox.jacobianInverseTransposed(xr);
      int eid = mapper.map(e);

      for (int i=0; i<BlockSize; ++i)
      {
        Gradient tmp(0.0);
        for (int d=0; d<n; d++)
          tmp[d] = s[i].evaluateDerivative(0, d, xr);
        grad_phi_ei = 0;
        invj.umv(tmp,grad_phi_ei);
        for (int d=0; d<n; d++)
          grad[d] += grad_phi_ei[d] * this->coeff_[eid][i];
      }

      for (int u=0; u<n; u++)
        if (d[u] == 1) return grad[u];
      DUNE_THROW(NotImplemented,"Can not evaluate this derivative: d=" << d);
    }

  private:
    // map local -> BBox
    const FieldVector<DT,n> bboxlocal(const Entity& e, const FieldVector<DT,n> & xi) const
    {
      #warning slow code!!!
      return t_.BBox(e).local(e.geometry().global(xi));
    }

    // map BBox -> global
    typedef Dune::UDG::EntityPart<G> EntityPart;
    // map Integration Part -> BBox
    typedef ::SubTriangulationGeometry<n, n, DT> IPGeometry;

    void localInterpolate( // map BBox -> global
      const BBoxGeometry &bboxgeometry,
      // map Integration Part -> BBox
      const IPGeometry &ipgeometry,
      FieldMatrix<RT,BlockSize,BlockSize>& Aee,
      FieldVector<RT,BlockSize> &Be,
      const C0Function<DT,RT,n,1> &u);
    void localInterpolate(const Entity &e,
                          // map BBox -> global
                          const BBoxGeometry &bboxgeometry,
                          // map Integration Part -> BBox
                          const IPGeometry &ipgeometry,
                          FieldMatrix<RT,BlockSize,BlockSize>& Aee,
                          FieldVector<RT,BlockSize> &Be,
                          const GridFunction<G,RT,1> &u);

  public:
    //! interpolate nodal values from a C0 function
    /*! L2 Projection of a C0 function

       @param[in]  u    a continuous function
     */
    void interpolate (const C0Function<DT,RT,n,1>& u);
    //! interpolate nodal values from a C0 function
    /*! L2 Projection of a C0 function

       @param[in]  u    a continuous function
     */
    void interpolate (const GridFunction<G,RT,1>& u);

    //! VTK output
    typename VTKWriter<GV>::VTKFunction *
    vtkFunction (std::string s) const
    {
      return new VTKGridFunctionWrapper<GV,RT,1>(*this,s);
    }

  private:
    const SubTriangulation & t_;
    Dune::GeometryType CubeType;
    MCMGMapper mcmgmapper;
    const Mapper & mapper;

  protected:
    // import the gridview set from the DGFunction base
    using DGFunction<GV,RT,o,SFSC>::gv_;
    // import the coefficient vector from the DGFunction base
    using DGFunction<GV,RT,o,SFSC>::coeff_;
    // import the shape functions from the DGFunction base
    using DGFunction<GV,RT,o,SFSC>::shapefnkts_;
  };

  /**
     \brief DG finite element function on the leaf grid

     \param G The grid
     \param RT The type used for the component values of the function
     \param o The order of the monomial shapefunctions
   */
  template<class G, class T, class RT, int o,
      template<class,class,int,int> class SFSC = MonomialShapeFunctionSetContainer>
  class LeafUDGFunction :
    public UDGFunction<typename G::LeafGridView,T,RT,o,SFSC>
  {
  public:
    LeafUDGFunction (const G& grid, const T& t)
      : GridFunctionGlobalEvalDefault<typename G::LeafGridView,RT,1>(grid.leafView())
        , GridFunctionGlobalDifferentiationDefault<typename G::LeafGridView,RT,1>(grid.leafView())
        , UDGFunction<typename G::LeafGridView,T,RT,o,SFSC>(grid.leafView(),t)
    {}
  };

  /**
     \brief DG finite element function on a given level grid

     \param G The grid
     \param RT The type used for the component values of the function
     \param o The order of the monomial shapefunctions
   */
  template<class G, class T, class RT, int o,
      template<class,class,int,int> class SFSC = MonomialShapeFunctionSetContainer>
  class LevelUDGFunction :
    public UDGFunction<typename G::LevelGridView,T,RT,o,SFSC>
  {
  public:
    LevelUDGFunction (const G& grid,  const T& t, int level)
      : GridFunctionGlobalEvalDefault<typename G::LevelGridView,RT,1>(grid.levelView(level))
        , GridFunctionGlobalDifferentiationDefault<typename G::LevelGridView,RT,1>(grid.levelView(level))
        , UDGFunction<typename G::LevelGridView,T,RT,o,SFSC>(grid.levelView(level),t)
    {}
  };

  /** @} */

}

#include "udgfunction.cc"

#endif // SCALED_DG_FUNCTION_HH
