// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_UDG_IO_VTKFUNCTION_HH
#define DUNE_UDG_IO_VTKFUNCTION_HH

#include <dune/udg/common/entitypart.hh>
#include <dune/udg/common/intersectionpart.hh>
#include <dune/udg/vtriangulation.hh>
#include <dune/localfunctions/common/interfaceswitch.hh>
#include <dune/pdelab/gridfunctionspace/localfunctionspace.hh>
#include <dune/pdelab/gridfunctionspace/subspace.hh>

namespace Dune {

  namespace UDG {

    // Base type for VTK functions for refined VTK writer
    template< typename GV >
    class UnfittedVTKFunction : public Dune::VTKFunction<GV>
    {
    public:
      typedef typename Dune::UDG::IntersectionPartInterface<GV>
      ::IntersectionPartPointer IntersectionPartPointer;

      typedef typename Dune::UDG::EntityPartInterface<GV>
      ::EntityPartPointer EntityPartPointer;

      virtual bool evaluateOn( const EntityPartPointer & part) const = 0;
      virtual bool evaluateOn( const IntersectionPartPointer & part) const = 0;
    };

    template<typename GFS, typename U, typename ST>
    class UnfittedVTKGridFunction :
      public UnfittedVTKFunction<typename GFS::Traits::GridViewType>
    {
    protected:
      typedef typename GFS::Traits::GridViewType GridView;
      typedef typename GridView::ctype DF;
      enum {dim=GridView::dimension};
      typedef typename GridView::template Codim<0>::Entity Entity;

      typedef typename GFS::Traits::FiniteElementType FiniteElement;
      typedef FiniteElementInterfaceSwitch<FiniteElement> FESwitch;
      typedef typename FESwitch::Basis Basis;
      typedef typename BasisInterfaceSwitch<Basis>::RangeField RangeField;
      typedef typename BasisInterfaceSwitch<Basis>::Range Range;
      typedef typename Entity::Geometry::GlobalCoordinate Domain;
      typedef UnfittedVTKFunction<GridView> Base;

      static int getLayer(const ST & st, const int domainIndex){
        typedef MultiLayerVirtualSubTriangulation<GridView> MultiLayerST;
        const MultiLayerST * mst = dynamic_cast<const MultiLayerST*>(&st);
        if(!mst) return 0;
        return mst->layerOfDomain(domainIndex);
      }

    public:

      UnfittedVTKGridFunction
        (std::string _label, const GFS & _gfs, const U & _u,
        const ST & _st, const int _domainIndex = 0, bool scaleToBBox = true)
        : label_(_label), lfs_(_gfs), lfsCache_(lfs_), lfsOut_(_gfs), lfsOutCache_(lfsOut_),
          u_view(_u), st_(_st), bbox_(0), bboxOut_(0),
          xl_(_gfs.maxLocalSize()), xlOut_(_gfs.maxLocalSize()), yb_(_gfs.maxLocalSize()),
          domainIndex_(_domainIndex), layer_(getLayer(_st,_domainIndex)),
          pgfs_(stackobject_to_shared_ptr(_gfs)), warningPrinted_(false), scaleToBBox_(scaleToBBox)
      {}

      UnfittedVTKGridFunction
        (std::string _label, const Dune::shared_ptr<const GFS> & _gfs, const U & _u,
        const ST & _st, const int _domainIndex = 0, bool scaleToBBox = true)
        : label_(_label), lfs_(*_gfs), lfsCache_(lfs_), lfsOut_(*_gfs), lfsOutCache_(lfsOut_),
          u_view(_u), st_(_st), bbox_(0), bboxOut_(0),
          xl_(_gfs->maxLocalSize()), xlOut_(_gfs->maxLocalSize()), yb_(_gfs->maxLocalSize()),
          domainIndex_(_domainIndex), layer_(getLayer(_st,_domainIndex)), pgfs_(_gfs),
          warningPrinted_(false), scaleToBBox_(scaleToBBox)
      {}

      virtual int ncomps () const
      {
        return 1;
      }

      virtual bool evaluateOn( const typename Base::EntityPartPointer & part) const
      {
        part->resetLayer();
        part->setLayer(layer_);
        if(part->domainIndex() != domainIndex_) {
          part->resetLayer();
          return false;
        }
        bbox_ = &(part->boundingBox());
        lfs_.bind(bbox_->entity());
        entity_ = bbox_->entity();
        lfs_.bind(entity_);
        lfsCache_.update();
        u_view.bind(lfsCache_);
        u_view.read(xl_);
        u_view.unbind();
        evalOnIntersection = false;
        part->resetLayer();
        return true;
      }

      //! \todo Remove std::cout warning as the mechanism seems to work flawless
      virtual bool evaluateOn( const typename Base::IntersectionPartPointer & part) const
      {
        part->setLayer(layer_);

        if(part->insideDomainIndex() != domainIndex_ && part->outsideDomainIndex() != domainIndex_) {
          part->resetLayer();
          return false;
        }
        if (!warningPrinted_)
        {
          std::cout << "Warning: Called UnfittedVTKGridFunction on interface" << std::endl;
          warningPrinted_ = true;
        }
        bbox_ = &(part->boundingBoxInside());
        entity_ = bbox_->entity();
        lfs_.bind(entity_);
        lfsCache_.update();
        u_view.bind(lfsCache_);
        u_view.read(xl_);
        u_view.unbind();

        evalOnIntersection = false;
        if(part->neighbor()) {
          bboxOut_ = &(part->boundingBoxOutside());
          entityOut_ = bboxOut_->entity();
          lfsOut_.bind(entityOut_);
          lfsOutCache_.update();
          u_view.bind(lfsOutCache_);
          u_view.read(xlOut_);
          u_view.unbind();
          evalOnIntersection = true;
        }
        part->resetLayer();
        return true;
      }

      virtual double evaluate (int comp, const Entity& e, const Dune::FieldVector<DF,dim>& xi) const
      {
        assert(bbox_);
        double y = 0;
        Domain gx = e.geometry().global(xi);

        { // Evaluate inside
          Domain local;
          if (scaleToBBox_)
            local = bbox_->local(gx);
          else
            local = entity_.geometry().local(gx);
          FESwitch::basis(lfs_.finiteElement()).evaluateFunction(local,yb_);
          for (unsigned int i=0; i<yb_.size(); i++)
            y += xl_[i] * yb_[i];
        }

        if(evalOnIntersection) { // Evaluate outside and average
          y *= 0.5;
          assert(bboxOut_);
          Domain localOut;
          if (scaleToBBox_)
            localOut = bboxOut_->local(gx);
          else
            localOut = entityOut_.geometry().local(gx);
          FESwitch::basis(lfsOut_.finiteElement()).evaluateFunction(localOut,yb_);
          for (unsigned int i=0; i<yb_.size(); i++)
            y += 0.5 * xlOut_[i] * yb_[i];
        }
        return y;
      }

      virtual std::string name () const
      {
        return label_;
      }

    protected:
      typedef Dune::PDELab::LocalFunctionSpace<GFS> LFS;
      typedef Dune::PDELab::LFSIndexCache<LFS> LFSCache;
      typedef typename U::template ConstLocalView<LFSCache> UView;

      std::string label_;
      mutable bool evalOnIntersection;
      mutable LFS lfs_;
      mutable LFSCache lfsCache_;
      mutable LFS lfsOut_;
      mutable LFSCache lfsOutCache_;
      mutable UView u_view;
      const ST & st_;
      mutable const typename ST::BoundingBox * bbox_;
      mutable Entity entity_;
      mutable const typename ST::BoundingBox * bboxOut_;
      mutable Entity entityOut_;
      mutable std::vector<RangeField> xl_;
      mutable std::vector<RangeField> xlOut_;
      mutable std::vector<Range> yb_;
      int domainIndex_;
      int layer_;
      Dune::shared_ptr<const GFS> pgfs_;

      mutable bool warningPrinted_;
      bool scaleToBBox_;
    };


    template<typename GFS, typename U, typename ST>
    class UnfittedVTKGridFunctionGradient :
      public UnfittedVTKGridFunction<GFS,U,ST>
    {
    protected:
      typedef UnfittedVTKGridFunction<GFS,U,ST> Base;
      enum {dim = Base::dim};
      typedef typename Base::Entity Entity;
      typedef typename Base::DF DF;
      typedef typename Base::Domain Domain;
      typedef typename Base::RangeField RangeField;
      typedef typename Base::FESwitch FESwitch;
      typedef BasisInterfaceSwitch<typename Base::Basis> BasisSwitch;
    public:

      UnfittedVTKGridFunctionGradient
        (std::string _label, const GFS & _gfs, const U & _u,
        const ST & _st, const int _domainIndex = 0, bool scaleToBBox = true)
        : Base(_label, _gfs, _u, _st, _domainIndex, scaleToBBox)
      {}

      UnfittedVTKGridFunctionGradient
        (std::string _label, const Dune::shared_ptr<GFS> & _gfs, const U & _u,
        const ST & _st, const int _domainIndex = 0, bool scaleToBBox = true)
        : Base(_label, _gfs, _u, _st, _domainIndex, scaleToBBox)
      {}

      virtual int ncomps () const
      {
        return dim;
      }

      virtual bool evaluateOn( const typename Base::EntityPartPointer & part) const
      {
        if (!Base::evaluateOn(part))
          return false;
        if (this->scaleToBBox_)
          jac_ = this->bbox_->jacobianInverseTransposed(Domain(.5));
        else
          jac_ = this->entity_.geometry().jacobianInverseTransposed(Domain(.5));
        return true;
      }

      virtual bool evaluateOn( const typename Base::IntersectionPartPointer & part) const
      {
        if (!Base::evaluateOn(part))
          return false;
        if (this->scaleToBBox_)
          jac_ = this->bbox_->jacobianInverseTransposed(Domain(.5));
        else
          jac_ = this->entity_.geometry().jacobianInverseTransposed(Domain(.5));

        if(this->evalOnIntersection) {
          if (this->scaleToBBox_)
            jacOut_ = this->bboxOut_->jacobianInverseTransposed(Domain(.5));
          else
            jacOut_ = this->entityOut_.geometry().jacobianInverseTransposed(Domain(.5));
        }
        return true;
      }

      virtual double evaluate (int comp, const Entity& e, const Dune::FieldVector<DF,dim>& xi) const
      {
        assert(this->bbox_);
        double y = 0;
        Domain gx = e.geometry().global(xi);

        { // Evaluate inside
          Domain local;
          if (this->scaleToBBox_)
            local = this->bbox_->local(gx);
          else
            local = this->entity_.geometry().local(gx);
          std::vector<Dune::FieldMatrix<RangeField,1,dim> > gradpsi(this->lfs_.size());
          FESwitch::basis(this->lfs_.finiteElement()).evaluateJacobian(local, gradpsi);
          Dune::FieldMatrix<RangeField,1,dim> tmp;
          for (unsigned int i = 0; i < gradpsi.size(); i++) {
            for (unsigned int r = 0; r < tmp.N(); ++r)
              jac_.mv(gradpsi[i][r], tmp[r]);
            gradpsi[i] = tmp;
          }

          for (unsigned int i=0; i<gradpsi.size(); i++)
            y += this->xl_[i] * gradpsi[i][0][comp];
        }

        if(this->evalOnIntersection) { // Evaluate outside and average
          y *= 0.5;
          assert(this->bboxOut_);
          Domain local;
          if (this->scaleToBBox_)
            local = this->bboxOut_->local(gx);
          else
            local = this->entityOut_.geometry().local(gx);
          std::vector<Dune::FieldMatrix<RangeField,1,dim> > gradpsi(this->lfsOut_.size());
          FESwitch::basis(this->lfsOut_.finiteElement()).evaluateJacobian(local, gradpsi);
          Dune::FieldMatrix<RangeField,1,dim> tmp;
          for (unsigned int i = 0; i < gradpsi.size(); i++) {
            for (unsigned int r = 0; r < tmp.N(); ++r)
              jacOut_.mv(gradpsi[i][r], tmp[r]);
            gradpsi[i] = tmp;
          }
          for (unsigned int i=0; i<gradpsi.size(); i++)
            y += 0.5 * this->xlOut_[i] * gradpsi[i][0][comp];
        }
        return y;
      }
    private:
      mutable Dune::FieldMatrix<DF, dim, dim> jac_;
      mutable Dune::FieldMatrix<DF, dim, dim> jacOut_;
    };



    namespace UnfittedVectorVTKGridFunctionImp {

      /** \brief TMP to fill a vector of UnfittedVTKGridFunction based
          using the corresponding sub grid view.
       */
      template<typename GFS, typename U, typename ST, typename F, int k>
      struct FillVTKFunctions
      {
        static void fill
          (const std::string label, const GFS & gfs,
          const U & u, const ST & st, const int domainIndex,
          F & functions)
        {
          typedef Dune::PDELab::GridFunctionSubSpace<GFS,Dune::TypeTree::TreePath<GFS::CHILDREN-k> > SubGFS;
          Dune::shared_ptr<SubGFS> subgfs ( new SubGFS(gfs) );

          typedef Dune::UDG::UnfittedVTKGridFunction<SubGFS,U,ST> UnfittedVTKGridFunction;
          Dune::shared_ptr< UnfittedVTKFunction<typename GFS::Traits::GridViewType> >
          vtkfunction( new UnfittedVTKGridFunction(label,subgfs,u,st,domainIndex) );
          functions.push_back(vtkfunction);
          FillVTKFunctions<GFS,U,ST,F,k-1>::fill(label,gfs,u,st,domainIndex,functions);
        }
      };

      /** \brief TMP to fill a vector of UnfittedVTKGridFunction based
          using the corresponding sub grid view.
       */
      template<typename GFS, typename U, typename ST, typename F>
      struct FillVTKFunctions<GFS, U, ST, F, 0>
      {
        static void fill
          (const std::string label, const GFS & gfs,
          const U & u, const ST & st, const int domainIndex,
          F & functions)
        {}
      };

    }

    /** \brief Unfitted vtk function for a vector valued grid function  */
    template<typename GFS, typename U, typename ST>
    class UnfittedVectorVTKGridFunction :
      public UnfittedVTKFunction<typename GFS::Traits::GridViewType>
    {
      typedef UnfittedVTKFunction<typename GFS::Traits::GridViewType>
      Base;

      typedef Dune::shared_ptr<Base> VTKFunctionPointer;

      typedef typename GFS::Traits::GridViewType GridView;
      typedef typename GridView::template Codim<0>::Entity Entity;
      typedef typename GridView::ctype DF;
      enum {dim=GridView::dimension};

    public:

      UnfittedVectorVTKGridFunction
        (std::string _label, const GFS & _gfs, const U & _u,
        const ST & _st, const int _domainIndex = 0)
        : label_(_label)
      {
        functions_.reserve(GFS::CHILDREN);
        UnfittedVectorVTKGridFunctionImp::
        FillVTKFunctions<GFS,U,ST,std::vector<VTKFunctionPointer>,GFS::CHILDREN>::
        fill(_label,_gfs,_u,_st,_domainIndex,functions_);
      }

      virtual bool evaluateOn( const typename Base::EntityPartPointer & part) const
      {
        bool eval = true;
        for(size_t f=0; f<functions_.size(); ++f)
          eval = functions_[f]->evaluateOn(part) && eval;
        return eval;
      }

      virtual bool evaluateOn( const typename Base::IntersectionPartPointer & part) const
      {
        bool eval = true;
        for(size_t f=0; f<functions_.size(); ++f)
          eval = functions_[f]->evaluateOn(part) && eval;
        return eval;
      }

      virtual double evaluate
        (int comp, const Entity& e, const Dune::FieldVector<DF,dim>& xi) const
      {
        return functions_[comp]->evaluate(0,e,xi);
      }

      virtual int ncomps () const
      {
        return GFS::CHILDREN;
      }

      virtual std::string name () const
      {
        return label_;
      }

    private:
      std::string label_;
      std::vector<VTKFunctionPointer> functions_;
    };



    template<typename UGF>
    class UnfittedVTKGeometryFunctionAdapter :
      public UnfittedVTKFunction<typename UGF::Traits::GridViewType>
    {
    protected:
      typedef typename UGF::Traits::GridViewType GridView;
      typedef typename UGF::Traits::SubTriangulation SubTriangulation;
      typedef typename GridView::ctype DF;
      enum {dim=GridView::dimension};

      typedef typename UGF::Traits::RangeFieldType RangeField;
      typedef typename UGF::Traits::RangeType Range;
      typedef typename UGF::Traits::DomainType Domain;
      typedef UnfittedVTKFunction<GridView> Base;
      typedef typename UGF::Traits::SubTriangulation::EntityPart EntityPart;
      typedef typename UGF::Traits::SubTriangulation::IntersectionPart IntersectionPart;
      typedef typename UGF::Traits::SubTriangulation::Entity Entity;

      static int getLayer(const SubTriangulation & st, const int domainIndex){
        typedef MultiLayerVirtualSubTriangulation<GridView> MultiLayerST;
        const MultiLayerST * mst = dynamic_cast<const MultiLayerST*>(&st);
        if(!mst) return 0;
        return mst->layerOfDomain(domainIndex);
      }

    public:

      UnfittedVTKGeometryFunctionAdapter
        (const UGF & ugf, const int domainIndex, const std::string label) :
        ugf_(ugf), domainIndex_(domainIndex), label_(label) {}

      virtual int ncomps () const
      {return UGF::Traits::dimRange;}

      virtual bool evaluateOn( const typename Base::EntityPartPointer & part) const
      {
        evalOnInside_ = evalOnOutside_ = false;
        evalOnPart_ = true;

        const int layer = getLayer(ugf_.subTriangulation(),domainIndex_);
        part->setLayer(layer);
        if(part->domainIndex() != domainIndex_) {
          part->resetLayer();
          return false;
        }

        entityPart_ = part.get();
        part->resetLayer();
        return true;
      }

      virtual bool evaluateOn( const typename Base::IntersectionPartPointer & part) const
      {
        const int layer = getLayer(ugf_.subTriangulation(),domainIndex_);
        part->setLayer(layer);

        evalOnPart_    = false;
        evalOnInside_  = part->insideDomainIndex() != domainIndex_;
        evalOnOutside_ = part->outsideDomainIndex() != domainIndex_;

        part->resetLayer();
        if(!evalOnInside_ && !evalOnOutside_)
          return false;

        intersectionPart_ = part.get();
        return true;
      }

      virtual double evaluate (int comp, const Entity& e, const Dune::FieldVector<DF,dim>& xi) const
      {
        const int layer = getLayer(ugf_.subTriangulation(),domainIndex_);

        Domain gx = e.geometry().global(xi);

        const typename UGF::Traits::BoundingBox * inside_bbox(0);
        const typename UGF::Traits::BoundingBox * outside_bbox(0);
        if(evalOnPart_) {
          entityPart_->setLayer(layer);
          inside_bbox = &( entityPart_->boundingBox() );
        }
        if(evalOnInside_) {
          intersectionPart_->setLayer(layer);
          inside_bbox = &( intersectionPart_->boundingBoxInside() );
        }
        if(evalOnOutside_) {
          intersectionPart_->setLayer(layer);
          outside_bbox = &( intersectionPart_->boundingBoxOutside() );
        }

        typename UGF::Traits::RangeType y_in(0);
        typename UGF::Traits::RangeType y_out(0);
        if(inside_bbox)
          ugf_.evaluate(*inside_bbox,inside_bbox->local(gx),y_in);
        if(outside_bbox)
          ugf_.evaluate(*outside_bbox,outside_bbox->local(gx),y_out);

        typename UGF::Traits::RangeType y(y_in + y_out);
        if(inside_bbox && outside_bbox)
          y /= 2.0;


        if(evalOnPart_)
          entityPart_->resetLayer();
        if( evalOnInside_ || evalOnOutside_ )
          intersectionPart_->resetLayer();

        return y[comp];
      }

      virtual std::string name () const
      {
        return label_;
      }

    private:
      const UGF & ugf_;
      const int domainIndex_;
      const std::string label_;
      mutable EntityPart* entityPart_;
      mutable IntersectionPart* intersectionPart_;
      mutable bool evalOnPart_;
      mutable bool evalOnInside_;
      mutable bool evalOnOutside_;
    };


    template<typename ST>
    class UnfittedMultiDomainVTKGridFunction :
      public UnfittedVTKFunction<typename ST::GridView>
    {
      typedef typename ST::GridView GridView;
      typedef typename GridView::ctype DF;
      enum {dim=GridView::dimension};
      typedef typename GridView::template Codim<0>::Entity Entity;

      typedef UnfittedVTKFunction<GridView> Base;

    public:

      UnfittedMultiDomainVTKGridFunction
        (const std::string _label) : label_(_label), components_(0), warningPrinted_(false)
      {}

      UnfittedMultiDomainVTKGridFunction* add
        (const Base * func)
      {
        if(components_ != func->ncomps() && components_ > 0)
        {DUNE_THROW(Dune::Exception,"Components of all domain functions must be equal!");}

        components_ = func->ncomps();
        functions_.push_back(BasePointer(func));
        return this;
      }

      virtual int ncomps () const
      {
        return components_;
      }

      virtual bool evaluateOn( const typename Base::EntityPartPointer & part) const
      {
        currentFunction_=-1;
        currentFunctionOut_=-1;
        for(size_t i=0; i<functions_.size(); ++i) {
          if(! functions_[i]->evaluateOn(part) ) continue;
          if(currentFunction_ >= 0)
          {DUNE_THROW(Dune::Exception,"Domains of given function overlap!");}

          currentFunction_ = i;
        }
        return currentFunction_ >= 0;
      }


      virtual bool evaluateOn( const typename Base::IntersectionPartPointer & part) const
      {
        currentFunction_=-1;
        currentFunctionOut_=-1;
        for(size_t i=0; i<functions_.size(); ++i) {
          if(! functions_[i]->evaluateOn(part) ) continue;
          if(currentFunction_ >= 0 && currentFunctionOut_ >= 0)
          {DUNE_THROW(Dune::Exception,"Domains of given function overlap!");}

          currentFunctionOut_ = currentFunction_;
          currentFunction_ = i;

        }
        return currentFunction_ >= 0;
      }

      virtual double evaluate (int comp, const Entity& e, const Dune::FieldVector<DF,dim>& xi) const
      {
        assert(currentFunction_>=0);
        double y = functions_[currentFunction_]->evaluate(comp,e,xi);
        if(currentFunctionOut_ >= 0)
        {
          if (!warningPrinted_)
          {
            std::cout << "Warning: Evaluating UnfittedMultiDomainVTKGridFunction on interface"
                      << " between two domains - averaging function values..." << std::endl;
            warningPrinted_ = true;
          }
          y *= 0.5;
          y += 0.5*functions_[currentFunctionOut_]->evaluate(comp,e,xi);
        }
        return y;
      }

      virtual std::string name () const
      {
        return label_;
      }

    private:
      const std::string label_;
      typedef Dune::shared_ptr<const Base> BasePointer;
      std::vector<BasePointer> functions_;
      mutable int currentFunction_;
      mutable int currentFunctionOut_;
      int components_;

      mutable bool warningPrinted_;
    };


    template<typename GV>
    class HostCellIndexUnfittedVTKGridFunction
      : public UnfittedVTKFunction<GV>
    {
      typedef typename GV::ctype DF;
      enum {n=GV::dimension};
      typedef typename GV::template Codim<0>::Entity Entity;
      typedef UnfittedVTKFunction<GV> Base;
    public:

      HostCellIndexUnfittedVTKGridFunction(const GV & _gv, const int layer = 0)
        : gv_(_gv), layer_(layer) {}

      virtual int ncomps () const
      {
        return 1;
      }

      virtual bool evaluateOn( const typename Base::EntityPartPointer & part) const
      {
        (part.get())->setLayer(layer_);
        hostIndex = gv_.indexSet().index(part->boundingBox().entity());
        domainIndex = part->domainIndex();
        (part.get())->resetLayer();
        return true;
      }

      virtual bool evaluateOn( const typename Base::IntersectionPartPointer & part) const
      {
        return false;
      }

      virtual double evaluate (int comp, const Entity& e, const Dune::FieldVector<DF,n>& xi) const
      {
        return hostIndex;
      }

      virtual std::string name () const
      {
        std::stringstream sname;
        sname << "HostIndex";

        if(layer_ >=0)
          sname << "Layer" << layer_;
        return sname.str();
      }

    private:
      GV gv_;
      const int layer_;
      mutable int hostIndex;
      mutable int domainIndex;
    };


    template<typename GV>
    class DomainTagUnfittedVTKGridFunction
      : public UnfittedVTKFunction<GV>
    {
      typedef typename GV::ctype DF;
      enum {n=GV::dimension};
      typedef typename GV::template Codim<0>::Entity Entity;
      typedef UnfittedVTKFunction<GV> Base;
    public:

      DomainTagUnfittedVTKGridFunction(const GV & _gv, const int layer = 0)
        : gv_(_gv), layer_(layer) {}

      virtual int ncomps () const
      {
        return 1;
      }

      virtual bool evaluateOn( const typename Base::EntityPartPointer & part) const
      {
        (part.get())->setLayer(layer_);
        tag = part->domainTag();
        (part.get())->resetLayer();
        return true;
      }

      virtual bool evaluateOn( const typename Base::IntersectionPartPointer & part) const
      {
        return false;
      }

      virtual double evaluate (int comp, const Entity& e, const Dune::FieldVector<DF,n>& xi) const
      {
        return tag;
      }

      virtual std::string name () const
      {
        std::stringstream sname;
        sname << "DomainTag";

        if(layer_ >=0)
          sname << "Layer" << layer_;
        return sname.str();
      }

    private:
      const GV & gv_;
      mutable int tag;
      const int layer_;
    };

    template<typename GV>
    class DomainIndexUnfittedVTKGridFunction
      : public UnfittedVTKFunction<GV>
    {
      typedef typename GV::ctype DF;
      enum {n=GV::dimension};
      typedef typename GV::template Codim<0>::Entity Entity;
      typedef UnfittedVTKFunction<GV> Base;
    public:

      DomainIndexUnfittedVTKGridFunction
        (const GV & _gv, const bool insideDomain = true, const int layer = 0)
        : gv_(_gv), insideDomain_(insideDomain), layer_(layer) {}

      virtual int ncomps () const
      {
        return 1;
      }

      virtual bool evaluateOn( const typename Base::EntityPartPointer & part) const
      {
        (part.get())->setLayer(layer_);
        domainIndex = part->domainIndex();
        (part.get())->resetLayer();
        return true;
      }

      virtual bool evaluateOn( const typename Base::IntersectionPartPointer & part) const
      {
        (part.get())->setLayer(layer_);
        if(insideDomain_)
          domainIndex = part->insideDomainIndex();
        else
          domainIndex = part->outsideDomainIndex();
        (part.get())->resetLayer();
        return true;
      }

      virtual double evaluate (int comp, const Entity& e, const Dune::FieldVector<DF,n>& xi) const
      {
        return domainIndex;
      }

      virtual std::string name () const
      {
        std::stringstream sname;
        if(insideDomain_)
          sname << "InsideDomainIndex";
        else
          sname << "OutsideDomainIndex";

        if(layer_ >=0)
          sname << "Layer" << layer_;
        return sname.str();
      }

    private:
      const GV & gv_;
      bool insideDomain_;
      const int layer_;
      mutable int domainIndex;
    };

  }
}
#endif
