// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DG_UDG_LOCALSTIFFNESS_HH
#define DG_UDG_LOCALSTIFFNESS_HH

#include <dune/common/fvector.hh>
#include <dune/common/fmatrix.hh>

#include <dune/udg/common/subentity.hh>
#include <dune/udg/common/subintersection.hh>
#include <dune/udg/common/geometry.hh>

namespace Dune {

  template <class T>
  class UDGLocalStiffness
  {
    // dimension of grid
    enum
    { dim = T::Grid::dimension /* !<dimension of the grid we are using */  };
    enum
    { dimw = T::Grid::dimensionworld /* !<dimension of the world */  };
  public:
    typedef T Traits;
    enum  { Components = Traits::Components };
    // Grid
    typedef typename Traits::Grid Grid;
    enum
    { dimension = Traits::Grid::dimension /* !<dimension of the grid we are using */  };
    enum
    { dimensionworld = Traits::Grid::dimensionworld /* !<dimension of the world */  };
    // ! Coordinate Type
    typedef typename Traits::ctype ctype;
    typedef typename Traits::matvec_t matvec_t;

    typedef Dune::FieldVector < matvec_t, dimw > Gradient;
    typedef Dune::FieldVector < matvec_t, dimw - 1 > FaceCoord;
    typedef Dune::FieldMatrix < matvec_t, dimw, dimw > JacobianInverseMatrix;

    // Geometries
    typedef Dune::Cuboid < dim, ctype > Geometry;
    typedef ::SubTriangulationGeometry < dim, dim, ctype > LocalGeometry;
    typedef ::SubTriangulationGeometry < dim - 1, dim, ctype > FaceGeometry;
    typedef ::SubTriangulationGeometry < dim - 1, dim, ctype > FaceLocalGeometry;
    typedef Dune::UDG::EntityPart < Grid > EntityPart;
    typedef Dune::UDG::IntersectionPart < Grid > IntersectionPart;
    typedef typename Grid::template Codim<0>::EntityPointer EntityPointer;

    // Local Matrix and Vector Type
    enum { BlockSize = Traits::BlockSize } DUNE_DEPRECATED;
    typedef typename Traits::LocalVectorBlock LocalVectorBlock;
    typedef typename Traits::LocalMatrixBlock LocalMatrixBlock;

    // Parameter Classes
    typedef typename Traits::BoundaryConditions BoundaryConditions;     // DUNE_DEPRECATED
    typedef typename Traits::BoundaryParameters BoundaryParameters;
    typedef typename Traits::ElementParameters ElementParameters;
    inline ElementParameters * createElementParameters(const EntityPointer & e) const
    {
      return asImp().createElementParameters(e);
    }
    inline BoundaryParameters * createBoundaryParameters(const IntersectionPart & b) const
    {
      return asImp().createBoundaryParameters(b);
    }
    inline const BoundaryConditions & getBoundaryConditions () const     // DUNE_DEPRECATED
    {
      return asImp().getBoundaryConditions();
    }

    // Shapefunctions
    typedef typename Traits::ShapeFunctionSet ShapeFunctionSet;
    inline const ShapeFunctionSet & getShapeFunctionSet (Dune::GeometryType t) const
    {
      return asImp().getShapeFunctionSet(t);
    }

    /* Assemble local Matrix entries */
    void assembleVolumeTerm (ElementParameters &ep,
                             /* the element e itself (not bounding box or anything) */
                             const EntityPointer &e,
                             /* the element id of e (since it is already calculated in the assembler) */
                             int eid,
                             /* map shapefnkt's elements -> global */
                             const Geometry & basegeometry,
                             /* map local -> shapefnkt's elements */
                             const LocalGeometry & localgeometry,
                             LocalMatrixBlock & Aee,
                             LocalVectorBlock & Be) const
    {
      return asImp().assembleVolumeTerm(ep, e, eid, basegeometry, localgeometry, Aee, Be);
    }

    void assembleFaceTerm (const BoundaryParameters &bp,
                           /* the element id of e (since it is already calculated in the assembler) */
                           int eid,
                           /* map e's shapefnkt's elements -> global */
                           const Geometry & geomInside,
                           /* the element id of f (since it is already calculated in the assembler) */
                           int fid,
                           /* map f's shapefnkt's elements -> global */
                           const Geometry & geomOutside,
                           /* for a dune grid this would be an
                              IntersectionIterator */
                           const IntersectionPart & it,
                           ctype edgenorm,
                           LocalMatrixBlock & Aee,
                           LocalMatrixBlock & Aef,
                           LocalMatrixBlock & Afe,
                           LocalVectorBlock & Be) const
    {
      return asImp().assembleFaceTerm(bp, eid, geomInside, fid, geomOutside, it, edgenorm, Aee, Aef, Afe, Be);
    }

    void assembleBoundaryTerm (const BoundaryParameters &bp,
                               /* the element e itself (not bounding box or anything) */
                               int eid,
                               /* map shapefnkt's elements -> global */
                               const Geometry & geometry,
                               /* for a dune grid this would be an
                                  IntersectionIterator */
                               const IntersectionPart & it,
                               ctype edgenorm,
                               LocalMatrixBlock & Aee,
                               LocalVectorBlock & Be) const
    {
      return asImp().assembleBoundaryTerm(bp, eid, geometry, it, edgenorm, Aee, Be);
    }

  private:
    typedef typename Traits::LocalAssembler Imp;
    Imp & asImp () { return static_cast<Imp&>(*this); }
    const Imp & asImp () const { return static_cast<const Imp&>(*this); }
  };

  template <class Traits>
  class UDGLocalStiffnessInterface : public UDGLocalStiffness<Traits>
  {
  public:
    typedef typename UDGLocalStiffness<Traits>::Grid Grid;
    typedef typename UDGLocalStiffness<Traits>::ctype ctype;
    typedef typename UDGLocalStiffness<Traits>::EntityPointer EntityPointer;
    typedef typename UDGLocalStiffness<Traits>::Geometry Geometry;
    typedef typename UDGLocalStiffness<Traits>::LocalGeometry LocalGeometry;
    typedef typename UDGLocalStiffness<Traits>::IntersectionPart IntersectionPart;
    typedef typename UDGLocalStiffness<Traits>::LocalMatrixBlock LocalMatrixBlock;
    typedef typename UDGLocalStiffness<Traits>::LocalVectorBlock LocalVectorBlock;
    typedef typename UDGLocalStiffness<Traits>::ShapeFunctionSet ShapeFunctionSet;

    typedef typename UDGLocalStiffness<Traits>::BoundaryConditions BoundaryConditions;
    typedef typename UDGLocalStiffness<Traits>::BoundaryParameters BoundaryParameters;
    typedef typename UDGLocalStiffness<Traits>::ElementParameters ElementParameters;

  private:
    const ShapeFunctionSet & getShapeFunctionSet (Dune::GeometryType t) const;

    void assembleVolumeTerm (ElementParameters &ep,
                             const EntityPointer &e,
                             int eid,
                             const Geometry & basegeometry,
                             const LocalGeometry & localgeometry,
                             LocalMatrixBlock & Aee,
                             LocalVectorBlock & Be) const;

    void assembleFaceTerm (const BoundaryParameters &fp,
                           int eid,
                           const Geometry & geomInside,
                           int fid,
                           const Geometry & geomOutside,
                           const IntersectionPart & it,
                           ctype edgenorm,
                           LocalMatrixBlock & Aee,
                           LocalMatrixBlock & Aef,
                           LocalMatrixBlock & Afe,
                           LocalVectorBlock & Be) const;

    void assembleBoundaryTerm (const BoundaryParameters &bp,
                               int eid,
                               const Geometry & geometry,
                               const IntersectionPart & it,
                               ctype edgenorm,
                               LocalMatrixBlock & Aee,
                               LocalVectorBlock & Be) const;

    const BoundaryConditions & getBoundaryConditions () const;
  public:
    ElementParameters * createElementParameters(const EntityPointer & e) const
    {
      return static_cast<const typename Traits::LocalAssembler&>(*this).getBoundaryConditions().elementParameters(e);
    }

    BoundaryParameters * createBoundaryParameters(const IntersectionPart & b) const
    {
      return static_cast<const typename Traits::LocalAssembler&>(*this).getBoundaryConditions().boundaryParameters(b);
    }

  };

  template <typename T>
  class LocalStiffnessNewInterface
  {
    typedef char one;
    typedef long two;

    template <typename C> static one test(const Dune::UDGLocalStiffness<typename C::Traits> *) ;
    template <typename C> static two test(...);
  public:
    // true if barton-nackman
    enum { value = sizeof(test<T>(0)) == sizeof(one) };
  };

  template <typename T, bool x>
  struct LocalStiffnessTypedefHelper
  {
    typedef T type;
  };

  template <typename T>
  struct LocalStiffnessTypedefHelper<T,true>
  {
    typedef UDGLocalStiffness<typename T::Traits> type;
  };

  template <typename T>
  struct LocalStiffnessTypedef
  {
    typedef typename LocalStiffnessTypedefHelper<T, LocalStiffnessNewInterface<T>::value>::type type;
  };

} // namespace Dune

#endif // DG_UDG_LOCALSTIFFNESS_HH
