// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_MC33TRIANGULATION_HH
#define DUNE_MC33TRIANGULATION_HH

#if defined HAVE_DUNE_MC or defined HAVE_DUNE_TPMC

#include <vector>   // provides std::vector
#include <list>     // provides std::list
#include <utility>  // provides std::pair
#include <limits>   // provides std::numeric_limits
#include <cmath>    // provides std::pow
#include <deque>    // provides std::deque
#include <mutex>    // provides std::mutex

#include <dune/common/version.hh>
#if !DUNE_VERSION_NEWER(DUNE_COMMON,2,4)
#include <dune/common/static_assert.hh>
#endif
#include <dune/common/fvector.hh>
#include <dune/common/parametertree.hh>

#include <dune/tpmc/tpmcrefinement.hh>

#include <dune/udg/vtriangulation.hh>
#include <dune/udg/common/entitypart.hh>
#include <dune/udg/common/intersectionpart.hh>
#include <dune/udg/mc33triangulation/domaindefinition.hh>
#include <dune/udg/mc33triangulation/domainrefinement.hh>
#include <dune/udg/mc33triangulation/domaincell.hh>
#include <dune/udg/mc33triangulation/merger.hh>
#include <dune/udg/mc33triangulation/constsublayermerger.hh>
#include <dune/udg/mc33triangulation/triangulationlayer.hh>
#include <dune/udg/mc33triangulation/parts.hh>
#include <dune/udg/common/threadutil.hh>

#ifndef NTHREADS
#define NTHREADS 1
#endif


namespace Dune
{
  namespace UDG
  {

    /**
     * \brief Multi layer sub triangulation class for MC33 engine.
     *
     * \tparam DD Type of a domain definition.
     */
    template< typename DD >
    class MarchingCube33SubTriangulation :
      public MultiLayerVirtualSubTriangulation<typename DD::GridView>
    {
    private:
      //! Type of fundamental mesh grid view
      typedef typename DD::GridView GV;
      //! Type of level set mesh grid view
      typedef typename DD::LevelSetGridView LGV;

    public:
      //! Type of domain definition
      typedef DD DomainDefinition;

      //! Type of fundamental mesh grid view
      typedef GV GridView;
      //! Type of level set mesh grid view
      typedef LGV LevelSetGridView;

      //! Type of a triangulation layer
      typedef MarchingCube33SubTriangulationLayer<DD> TriangulationLayer;

      //! Flag indicating that this sub triangulation uses the MC33 engine
      static const bool isMarchingCube33 = true;

      //! Interface for sub triangulation part of an entity
      typedef typename Dune::UDG::EntityPartInterface<GridView> EntityPartInterface;
      //! List of entity part pointers
      typedef typename EntityPartInterface::EntityPartList EntityPartList;
      //! Geometry engine for entity part
      typedef typename Dune::UDG::DefaultEntityPartGeometry<GridView> EntityPartGeometry;
      //! Implementation of entity part
      typedef typename Dune::UDG::DefaultEntityPart<EntityPartGeometry> SubEntityPart;
      typedef typename Dune::UDG::MultiLayerEntityPart<SubEntityPart> EntityPart;
      //! Shared pointer to entity part
      typedef typename EntityPartInterface::EntityPartPointer EntityPartPointer;

      //! Geometry engine for sub triangulation part of an intersection
      typedef Dune::UDG::AffineIntersectionPartGeometry<GridView>
      IntersectionPartGeometry;
      //! Sub triangulation part of an intersection
      typedef Dune::UDG::DefaultIntersectionPart<IntersectionPartGeometry>
      SubIntersectionPart;
      typedef Dune::UDG::MultiLayerIntersectionPart<SubIntersectionPart>
      IntersectionPart;
      //! Interface to sub triangulation part of an intersection
      typedef Dune::UDG::IntersectionPartInterface<GridView>
      IntersectionPartInterface;
      //! List for intersection parts
      typedef typename IntersectionPart::IntersectionPartList IntersectionPartList;
      //! Intersection part pointer of list
      typedef typename IntersectionPartInterface::IntersectionPartPointer
      IntersectionPartPointer;

      //! Type of bounding box of a domain cell for scaling of the local
      //! functions
      typedef GridCuboid< typename GV::Grid > BoundingBox;

      //! Continuous field type
      typedef typename GV::ctype DomainField;

    private:
      //! Global dimension
      enum { dim = GV::dimension };
#if DUNE_VERSION_NEWER(DUNE_COMMON,2,4)
      static_assert( (int(dim)==int(GV::dimensionworld)), "Manifold grids not supported!");
#else
      dune_static_assert( (int(dim)==int(GV::dimensionworld)), "Manifold grids not supported!");
#endif

      //! Coordinate types
      typedef Dune::FieldVector<DomainField,dim> Coordinate;
      typedef std::vector<Coordinate> Coordinates;

      //! Index types
      typedef typename GV::IndexSet::IndexType Index;
      typedef std::list<Index> IndexList;
      typedef std::pair<Index,Index> IndexPair;
      typedef typename GV::IndexSet::IndexType Size;
      static Index invalidIndex(){ return std::numeric_limits<Index>::max(); }

      //! Grid view types
      typedef typename LGV::Traits::template Codim<0>::Entity CellEntity;
      typedef typename GV::Traits::template Codim<0>::Entity GVCellEntity;

      //! Domain refinement types
      typedef Dune::UDG::DomainRefinement<DomainDefinition> DomainRefinement;
      typedef typename DomainRefinement::DomainRefinementPart DomainRefinementPart;
      typedef typename DomainRefinementPart::Geometry Geometry;
      typedef typename DomainRefinementPart::DomainTag DomainTag;

      //! Domain cell types
      typedef MarchingCube33SubTriangulationImp::DomainCell<GV> DomainCell;
      typedef MarchingCube33SubTriangulationImp::DomainCellFace<GV> DomainCellFace;
      typedef typename DomainCell::RawBoundingBox RawBoundingBox;
      typedef typename std::list<DomainCell> DomainCellList;

      //! Type of cache for intersection parts individualized by threads
      typedef IntersectionThreadCache<IntersectionPartList,Index>
      IntersectionCache;

      //! Lookup types
      typedef typename TriangulationLayer::LookupMap LookupMap;
      typedef typename LookupMap::LookupItem LookupItem;

      //! Type for multi layer information about domain cells at an intersection
      struct MultiIntersectionDomainCells
      {
        MultiIntersectionDomainCells(const Index layers)
          : insideDomainCells(layers,0), outsideDomainCells(layers,0)
        {}

        //! Check if intersection is at an exterior boundary
        //! \todo This method involves all layers in its decision. Is this
        //!       intended, i.e. do all layers must have the same exterior
        //!       boundary? Can insideDomainCells[l]==outsideDomainCells[l]
        //!       theoretically yield different values on different layers?
        //!       See also createMultiPartsForWholeCell(...) where a similar
        //!       aspect is questioned by a TODO.
        bool isBoundary() const {
          bool boundary = true;
          for(Index l=0; l<insideDomainCells.size(); ++l)
            boundary &= insideDomainCells[l]==outsideDomainCells[l];
          return boundary;
        }

        //! Store adjacent domain cells on each layer
        std::vector<const DomainCell*> insideDomainCells;
        std::vector<const DomainCell*> outsideDomainCells;
      };

      /**
       * \brief Create a single multi layer entity part and multi layer
       *        intersection parts from a fundamental mesh cell which is
       *        entirely contained within a single domain.
       *
       * \note This function should only be called for a cell for which
       *       domainDefinition().numberOfInterfacesInCell(...,
       *       domainDefinition().numberOfLayers()-1) returns true.
       *
       * \todo The routine which checks if the fundamental mesh cell is
       *       in no domain might be implemented wrong. The entity part
       *       creation is skipped if the fundamental mesh cell is not
       *       contained in a domain >on at least one layer<. Why is
       *       this valid? Shouldn't it be >on all layers<? Or do we
       *       have a (silent) prerequisite that each layer decomposes
       *       the same geometrical object (e.g. the full fundamental
       *       mesh domain) in domains?
       *       See also MultiIntersectionDomainCells::isBoundary where
       *       a similar aspect is questioned by a TODO.
       */
      void createMultiPartsForWholeCell
      (
        const GVCellEntity & cell,
        EntityPartList & entityparts,
        IntersectionPartList & intersectionCache
      ) const;


      // Create intersection part for given vertices, domain indices,
      // bounding boxes and normal vector.
      // \todo DocMe right.
      SubIntersectionPart* createIntersectionPart
      (
        const BoundingBox & inBBox,
        const BoundingBox & outBBox,
        const DomainTag domainTag,
        const int insideDomainIndex,
        const int outsideDomainIndex,
        const Coordinates & c,
        const Coordinate & normal,
        const DomainField & area,
        const DomainField & insideVolume,
        const DomainField & outsideVolume
      ) const;

      // Create intersection part for given vertices, normal vector
      // and domain cells.
      // \todo DocMe right.
      SubIntersectionPart* createIntersectionPart
      (
        const DomainCell & insideDomainCell,
        const DomainCell & outsideDomainCell,
        const Coordinates & c,
        const Coordinate & normal,
        const bool boundary
      ) const;

      // // Create an multi layer intersection part for the given domain
      // // cell face, global coordinates, and normal vector.
      // IntersectionPart* createMultiIntersectionPart
      // (
      //  const Index indexInPart,
      //  const Index partIndex,
      //  const Index cellIndex,
      //  const Index lCellIndex,
      //  const Coordinates & c,
      //  const Coordinate & normal
      //  ) const;


      /**
       * \brief Create a multi layer intersection part from the adjacent domain
       *        cells on each layer and the intersection part's corners in global
       *        coordinates as well as its outer unit normal.
       *
       * \todo Why is isBoundary in the code dependent on domainCells.isBoundary()
       *       which involves all layers in its decision?
       *       See the implementation of MultiIntersectionDomainCells::isBoundary.
       */
      IntersectionPart* createMultiIntersectionPart
      (
        const MultiIntersectionDomainCells & domainCells,
        const Coordinates & c,
        const Coordinate & normal
      ) const;

      /**
       * \brief Create a multi layer entity part from a level set mesh cell,
       *        a given refinement part's part index as well as its corners in
       *        local coordinates of the level set mesh cell, and the associated
       *        domain cells on each layer.
       *
       * \note The level set mesh cell may be replaced by a fundamental mesh
       *       cell if the multi layer entity part shall be created from a
       *       whole fundamental mesh cell. But this shall only be exploited
       *       by createMultiPartsForWholeCell(...) which is intended for that
       *       purpose.
       */
      EntityPart* createMultiEntityPart
      (
        const CellEntity & lCell,
        const Index partIndex,
        const Coordinates & c,
        const std::vector<const LookupItem*> & domainCells
      ) const;

      /**
       * \brief Create an entity part from a level set mesh cell, a given
       *        refinement part's part index as well as its corners in local
       *        coordinates of the level set mesh cell, and the fundamental
       *        mesh host cell, bounding box, domain index, tag, domain tag
       *        and volume of the associated domain cell.
       *
       * \note The level set mesh cell may be replaced by a fundamental mesh
       *       cell if the multi layer entity part shall be created from a
       *       whole fundamental mesh cell. But this shall only be exploited
       *       by createMultiPartsForWholeCell(...) which is intended for that
       *       purpose.
       *
       * \todo Remove tag parameter from entity part and this method.
       */
      SubEntityPart* createEntityPart
      (
        const CellEntity & homeInLGV,
        const int partIndex,
        Coordinates c,
        const GVCellEntity & hostInGV,
        const RawBoundingBox & rawbbox,
        const Index & domainIndex,
        const int tag,
        const DomainTag & domainTag,
        const DomainField & volume
      ) const;

      /**
       * \brief Create an entity part from a level set mesh cell, a given
       *        refinement part's part index as well as its corners in local
       *        coordinates of the level set mesh cell, and the associated
       *        domain cell.
       *
       * \note The level set mesh cell may be replaced by a fundamental mesh
       *       cell if the multi layer entity part shall be created from a
       *       whole fundamental mesh cell. But this shall only be exploited
       *       by createMultiPartsForWholeCell(...) which is intended for that
       *       purpose.
       */
      SubEntityPart* createEntityPart
      (
        const CellEntity & homeInLGV,
        const int partIndex,
        const Coordinates & c,
        const DomainCell & domainCell
      ) const;


      // Compute intersection parts for the given refinement part
      // created for the given cells and store it in the cache.
      // \todo DocMe right.
      void cacheIntersectionParts
        (const CellEntity & lCell,
        const std::vector<const LookupItem*> lookupItems,
        const DomainRefinementPart & part,
        IntersectionPartList &intersectionCache) const;


    public:
      /**
       * \brief Constructor which does NOT take ownership of domainDef.
       *
       * \param domainDef
       *        Definition of the domains.
       * \param parameters
       *        Optional Dune::ParameterTree (substructure) which contains
       *        the following optional parameters.
       *        force_refinement:
       *          Switch for refinement of fundamental mesh cells which
       *          are entirely filled on each layer by a single domain.
       *        use_intersections:
       *          To make the code more efficient, this switch can be
       *          deactivated if intersection parts are not required.
       *        remove_identical_domaincell_face_corners_eps:
       *          Numerical epsilon which is used to decide whether two corners
       *          of a domain refinement part's face are identical which might
       *          prevent the creation of a domain cell face corresponding to
       *          this face.
       *        min_edge_length_eps:
       *          Numerical epsilon which yields a minimum edge length during
       *          domain refinement.
       *        linearize_parts:
       *          Switch for iterative domain refinement procedure which
       *          triggers linearization before refinement using the next
       *          interface.
       *        finalize_verbose:
       *          Switch for verbose output during finalization of the
       *          sub triangulation.
       */
      MarchingCube33SubTriangulation
        (DomainDefinition & domainDef,
         const Dune::ParameterTree parameters = Dune::ParameterTree())
        : pDomainDef_(Dune::stackobject_to_shared_ptr(domainDef)),
          domainDef_(*pDomainDef_),
          gv_(domainDef.gridView()), lgv_(domainDef.levelSetGridView()),
          parameters_(parameters),
          finalized_(false), intersectionCache_(gv_.size(0),NTHREADS)
      { init(); }

      /**
       * \brief Constructor which takes ownership of domainDef.
       *
       * \copydetails MarchingCube33SubTriangulation<DD>::
       *              MarchingCube33SubTriangulation
       *              (DomainDefinition&,const Dune::ParameterTree)
       */
      MarchingCube33SubTriangulation
        (const Dune::shared_ptr<DomainDefinition> & domainDef,
         const Dune::ParameterTree parameters = Dune::ParameterTree())
        : pDomainDef_(domainDef), domainDef_(*pDomainDef_),
          gv_(domainDef->gridView()), lgv_(domainDef->levelSetGridView()),
          parameters_(parameters),
          finalized_(false), intersectionCache_(gv_.size(0),NTHREADS)
      { init(); }

      /**
       * \brief Enable/disable an enforced refinement of fundamental mesh cells
       *        which are entirely filled on each layer by a single domain.
       *
       * \note Naturally, setting this flag is inconsequential if the grid
       *       views representing the fundamental mesh / the level set mesh
       *       (i.e. domainDefinition().gridView() and
       *       domainDefinition().levelSetGridView()) are the same.
       */
      bool & forceRefinement() { return forceRefinement_; };

      /**
       * \brief To make the code more efficient, this flag can be deactivated
       *        if intersection parts are not required.
       */
      bool & useIntersections() { return useIntersections_; }

      /**
       * \brief Interface function providing the intersection parts for a
       *        given fundamental mesh cell.
       */
      void create_intersection_parts(const GVCellEntity & e,
                                     IntersectionPartList & intersectionparts) const;

      /**
       * \brief Interface function providing the entity parts for a given
       *        fundamental mesh cell.
       */
      void create_entity_parts(const GVCellEntity & e,
                               EntityPartList & entityparts) const;


#ifndef DOXYGEN
      //! \todo DocMe.
      //! \todo Why does this method depend on DOXYGEN?
      void debugFaceConnections
        (const DomainField minInside = -std::numeric_limits<DomainField>::max(),
        const DomainField maxInside = std::numeric_limits<DomainField>::max()) const;
#endif

      /** \brief Interface function allowing access to the bounding
          boxes of the domain cell.

          This interface requires that the flag uniqueHostCells of the
          setupSubTriangulation member function was set to true.

          \todo DocMe right.
       */
      const BoundingBox BBox
        (const unsigned int domainIndex, const GVCellEntity & host) const;


      /**
       * \brief Provide bounding box for a given entity part.
       *
       * \note This function even works if the entity part was created for a
       *       different instance of MarchingCube33SubTriangulation as long as
       *       both sub triangulations have the same level set mesh grid view
       *       and identical interfaces on identical layers. They may however
       *       differ with regard to the fundamental mesh grid view, the defined
       *       domains and the merging of the domain cells within these domains.
       *
       * \note Notice that for a multi layer sub triangulation the bounding
       *       box will depend on the current layer set in the entity part.
       *
       * \todo In principal the bounding box could be provided for an arbitary
       *       layer independent from the current layer set in the entity part.
       *       Maybe add a parameter "int layer" to allow access in this way.
       * \todo The first \note seems limited to the case that a different
       *       instance of MarchingCube33SubTriangulation needs to have a
       *       finer fundamental mesh grid view than this instance.
       *       Otherwise the determination of gvEntity in the code will fail.
       */
      const BoundingBox boundingBox(const EntityPartInterface & ep) const;

      /**
       * \brief Check whether a given fundamental mesh cell is host cell of
       *        entity parts with a given domain index.
       *
       * \note This function may return false even if some entity parts with the
       *       given domain index are geometrically positioned in the given cell
       *       (which is therefore their fundamental mesh "home cell") but are
       *       merged to - and therefore associated with - another fundamental
       *       mesh cell (their "host cell").
       */
      bool isHostCell
        (const GVCellEntity & hostCandidate, const int & domainIndex) const;

      /**
       * \brief Computes an edge norm for the domain cell intersection which
       *        contains a given intersection part using the volume of the
       *        adjacent domain cells as
       *        0.5 * {harmonic average of inside and outside volume}^{1/dim}.
       *
       * \note Notice that for a multi layer sub triangulation the edge norm
       *       will depend on the current layer set in the entity part.
       *
       * \todo This is an interface function of VirtualSubTriangulation. Is
       *       such an interface function useful? A DUNE-PDELab local operator
       *       would need to know the sub triangulation to access the function.
       */
      DomainField getEdgeNorm(const IntersectionPartInterface & ipart) const
      {
        const DomainField iv = ipart.insideVolume();
        const DomainField ov = ipart.outsideVolume();
        const DomainField av = 1.0 / (1.0 / iv + 1.0 / ov);
        return std::pow(av, 1.0 / DomainField(dim));
      }

      /**
       * \brief Unused interface function.
       */
      bool isInDomain(const GVCellEntity & e, const int & domain_index) const
      { DUNE_THROW(Dune::NotImplemented,"Not implemented"); return false; }

      /**
       * \brief Constant access to fundamental mesh grid view.
       */
      const GridView & gridView() const
      { return gv_; }

      /**
       * \brief Return number of domains defined for this sub triangulation.
       */
      Index nDomains() const { return domainDef_.numberOfDomains(); }

      /**
       * \brief Return number of interfaces defined for this sub triangulation.
       */
      Index nInterfaces() const { return domainDef_.numberOfInterfaces(); }

      /**
       * \brief Setup the sub triangulation according to domain definition.
       *
       * \note This function should be called directly after the interfaces
       *       and domains have been defined and before any other interface
       *       function.
       */
      void init();

      /**
       * \brief Reinitialize the given triangulation layer with index
       *        layerIndex, or all layers if layerIndex == -1.
       */
      void reinit(const int layerIndex = -1)
      {
        if(layerIndex==-1)
          for(Index l=0; l<domainDef_.numberOfLayers(); ++l)
            layers_[l].init();
        else
          layers_[layerIndex].init();
      }

      /**
       * \brief Merge the domain cells using a given merger engine.
       *
       * \note This function should only be called between intialize()
       *       and finalize().
       */
      template< typename Merger>
      void mergeDomainCells(Merger & merger);

      /**
       * \brief Merge the domain cells of a given layer using a given
       *        merger engine.
       *
       * \note This function should only be called between intialize()
       *       and finalize().
       */
      template< typename Merger>
      void mergeDomainCells(const int layer, Merger & merger);

      /**
       * \brief Make sub triangulation ready for usage.
       *
       * \note This function should be called directly after intialize() or
       *       the last call of mergeDomainCells(...) and before any other
       *       interface function.
       */
      void finalize(const int layerIndex = -1, bool verbose = true);

      /**
       * \brief Get the layer in which the domain of the given index is
       *        defined.
       */
      Index layerOfDomain(const Index domainIndex) const
      {
        return domainDef_.domainLayer(domainIndex);
      }

      /**
       * \brief Access to the domain definition.
       *
       * \note Changes to the domain defintion might invalidate the
       *       data-structures of the triangulation layers which then
       *       have to be reinitialized using reinit(...).
       */
      DomainDefinition & domainDefinition()
      {
        return domainDef_;
      }

      /**
       * \brief Constant access to the domain definition.
       */
      const DomainDefinition & domainDefinition() const
      {
        return domainDef_;
      }

      /**
       * \brief Constant access to the triangulation layers.
       */
      const TriangulationLayer & triangulationLayer(const Index layer) const
      {
        return layers_[layer];
      }

    private:
      //! Domain definition
      Dune::shared_ptr<DomainDefinition> pDomainDef_;
      DomainDefinition & domainDef_;

      //! Grid view for finite element grid, i.e. the fundamental mesh
      const GV & gv_;
      //! Grid view for interface reconstruction, i.e. the level set mesh
      const LGV & lgv_;

      //! Optional parameters
      const Dune::ParameterTree parameters_;

      //! Triangulation layers
      std::deque<TriangulationLayer> layers_;

      //! Flag indicating whether finalize was called
      bool finalized_;

      //! Flag indicating whether refinement of fundamental mesh cells which
      //! are entirely filled on each layer by a single domain is enforced
      bool forceRefinement_;

      //! Flag indicating whether intersection parts are required
      bool useIntersections_;

      //! Cache for intersection parts
      // mutable IntersectionPartList intersectionCache_;
      // mutable Index intersectionCacheCellIndex_;
      IntersectionCache intersectionCache_;
      mutable std::mutex stmutex_;
    };

  }
}


#include "mc33triangulation/mc33triangulation.cc"
#include "mc33triangulation/mc33triangulation_debug.cc"

#endif  // HAVE_DUNE_MC

#endif  // DUNE_MC33TRIANGULATION_HH
