// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_UDG_MC33_CONST_SUB_LAYER_MERGER_HH
#define DUNE_UDG_MC33_CONST_SUB_LAYER_MERGER_HH
namespace Dune
{
  namespace UDG
  {

    /** \brief Merger which restores a partition of domain cells on a
        given layer after interfaces on higher layers have changed.
     */
    template<typename GV>
    class ConstantSubLayerMerger
    {
    public:
      typedef typename GV::IndexSet::IndexType Index;
      typedef std::pair<Index,Index> IndexPair;
      typedef std::map<IndexPair,Index> HostCellIndexMap;

      ConstantSubLayerMerger
      (
        const HostCellIndexMap & hostCellIndexMap,
        const Index projectionsOfPartIndex
      ) :
        hostCellIndexMap_(hostCellIndexMap),
        projectionsOfPartIndex_(projectionsOfPartIndex)
      {}

      template<typename Layers>
      void mergeAllLayers(Layers & layers)
      {}

      /** \brief Globally merge domain cells to fulfill the minimum
          volume criterion.
       */
      template<typename HostCellMap>
      void merge(HostCellMap & hostCellMap)
      {
        std::cout << "Merge domain cells to restore partition!" << std::endl;

        // First, iterate all domain cells and reset their host cell if necessary
        for(Index hostCellIndex=0; hostCellIndex<hostCellMap.size(); ++hostCellIndex) {

          typedef typename HostCellMap::value_type DomainCellList;
          DomainCellList & domainCells = hostCellMap[hostCellIndex];

          typename DomainCellList::iterator domainCellIt = domainCells.begin();
          while(domainCellIt != domainCells.end()) {

            if(domainCellIt->lgvMembers.empty()) {
              domainCellIt++;
              continue;
            }

            const Index lCellIndex = domainCellIt->lgvMembers.begin()->first;
            Index partIndex = *(domainCellIt->lgvMembers.begin()->second.begin());
            for(Index l=0; l<projectionsOfPartIndex_; ++l)
              partIndex /= DomainRefinementImp::maxPartsInLayer;

            typename HostCellIndexMap::const_iterator hostCellIndexIt =
              hostCellIndexMap_.find(std::make_pair(lCellIndex,partIndex));

            // The cell in lgv was formerly intersected but is now
            // whole or vice versa.
            if(hostCellIndexIt == hostCellIndexMap_.end())
              hostCellIndexIt = hostCellIndexMap_.lower_bound(std::make_pair(lCellIndex,0));
            assert(hostCellIndexIt != hostCellIndexMap_.end());

            // Save iterator and increment
            typename DomainCellList::iterator currentIt = domainCellIt;
            domainCellIt++;

            // Set new host cell of domain cell if necessary
            if(hostCellIndexIt->second != hostCellIndex) {

              DomainCellList & destinationList = hostCellMap[hostCellIndexIt->second];

              // Set correct host cell
              assert(!destinationList.empty());
              currentIt->hostCell = destinationList.begin()->hostCell;

              // Move to correct host cell list
              destinationList.splice(destinationList.begin(),domainCells,currentIt);
            }

          }

        }

        // Second, merge all domain cells with the same host cell
        for(Index hostCellIndex=0; hostCellIndex<hostCellMap.size(); ++hostCellIndex) {
          typedef typename HostCellMap::value_type DomainCellList;
          DomainCellList & domainCells = hostCellMap[hostCellIndex];
          typedef typename DomainCellList::value_type DomainCell;
          DomainCell::mergeAdjacentDomainCells(domainCells);
        }

      }

    private:
      const std::map<IndexPair,Index> & hostCellIndexMap_;
      const Index projectionsOfPartIndex_;
    };

  }
}

#endif
