// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_UDG_MC33_DEFAULTLEVELSETBACKEND_HH
#define DUNE_UDG_MC33_DEFAULTLEVELSETBACKEND_HH
#include "levelsetbackend.hh"
#include "mc33triangulationimp.hh"
#include <dune/udg/common/geometryinterface.hh>
#include <dune/functions/common/functionconcepts.hh>

namespace Dune {
  namespace UDG {

    /** \brief A backend allowing access to the level set function
        values.
     */
    template< typename DomainDefinition, typename LevelSetFunction, typename Switch=void >
    class DefaultLevelSetBackend :
      public virtual LevelSetBackendNodeStorageInterface<DomainDefinition>
    {
    public:
      //! Base class
      typedef LevelSetBackendInterface<DomainDefinition> Base;
      //! Traits class
      typedef typename Base::Traits Traits;

    private:
      typedef typename Traits::LevelSetGridView LevelSetGridView;
      typedef typename Traits::GridView GridView;
      typedef typename Traits::LevelSetValues LevelsetValues;
      typedef typename Traits::Coordinate Coordinate;
      typedef typename Coordinate::value_type DomainField;
      typedef std::vector<DomainField> LevelSetValues;
      typedef typename Traits::CoordinateComparator CoordinateComparator;
      enum { dim = Coordinate::dimension };
      typedef typename Traits::CellEntity CellEntity;
      typedef typename Traits::GVCellEntity GVCellEntity;
      typedef typename Traits::Index Index;
      typedef typename Traits::RefinementPart RefinementPart;

    protected:

      //! Helper function to get sign from vector of values
      int getSign
        (const typename Traits::LevelSetValues & values) const
      {
        // Flag to indicate an interface
        bool interfaceFound = false;
        typename Traits::DomainField lastValue(0);

        for(typename Traits::Index v=0; v<values.size(); ++v) {
          if(values[v] == domainDef_->invalidValue()) continue;

          // Check for interface intersecting this cell
          interfaceFound |= ( lastValue * values[v] ) < 0;
          lastValue = values[v];
        }
        if( interfaceFound )
          return 0;
        else
          return int( ((1.0+1e-10)*lastValue)/std::abs(lastValue) );
      }

      // Make all small corner values positive
      void normalizeValuesInVertices
        (const Dune::GeometryType type,
        LevelSetValues & values) const
      {
        // Reference element of cell to refine
        auto ref =
          Dune::referenceElement<DomainField,dim>(type);

        const int nEdges = ref.size(dim-1);
        for(int e=0; e<nEdges; ++e) {
          const int lIndex = ref.subEntity(e,dim-1,0,dim);
          const int rIndex = ref.subEntity(e,dim-1,1,dim);

          DomainField & lValue = values[lIndex];
          DomainField & rValue = values[rIndex];
          if(lValue * rValue > 0) continue;

          const bool leftSmaller = std::abs(lValue) < std::abs(rValue);
          DomainField & smallValue = leftSmaller ? lValue : rValue;
          DomainField & bigValue   = leftSmaller ? rValue : lValue;

          const DomainField minValue = std::numeric_limits<DomainField>::min();
          if(smallValue == 0.0 || std::abs(smallValue/bigValue) < 1e-8)
            smallValue = std::max(std::abs(smallValue),minValue);

        }
      }


      // Interpolate the given interface with the given level set
      // function on the given cell in lgv. This function may call
      // itself recursively if it detects that this cell has a hanging
      // corner in an adjacent coarser cell.
      void interpolateCellCorners
        (const GVCellEntity & cell, const CellEntity & lCell,
        std::vector<bool> & signInfo)
      {
        // Loop over cell vertices
#if DUNE_VERSION_NEWER(DUNE_GRID,2,4)
        Index nVertices = lCell.subEntities(dim);
#else
        Index nVertices = lCell.template count<dim>();
#endif
        auto lf = localFunction(*lsf_);
        lf.bind(cell);
        LevelSetValues cellValues(nVertices);
        for(Index v=0; v < nVertices; ++v) {

          // Vertex entity
          typedef typename LevelSetGridView::Traits::
          template Codim<dim>::Entity NodeEntity;
          const NodeEntity ve = lCell.template subEntity<dim>(v);
          const Index vertexIndex = domainDef_->levelSetGridView().indexSet().index(ve);

          // Value at node
          DomainField & value = levelSetValues_[vertexIndex];

          // Do not compute the same value twice
          if(value != domainDef_->invalidValue()) {
            cellValues[v] = value;
            continue;
          }

          // Evaluate level set function locally at vertex
          const Coordinate global = ve.geometry().center();
          const Coordinate local  = lCell.geometry().local(global);
          value = lf(local);

          // Do not allow zero values
          if(value==0) value = std::numeric_limits<DomainField>::min();

          // Store value locally for hanging nodes
          cellValues[v] = value;

        } // v

        // Update sign info for this cell
        const Index cellIndex = domainDef_->gridView().indexSet().index(cell);
        for(Index v=0; v < nVertices; ++v) {
          assert(cellValues[v]!=0);
          const bool negativeValue = cellValues[v]<0;
          typedef typename std::vector<bool>::reference BitReference;
          BitReference negativeBit = signInfo[2*cellIndex];
          negativeBit = negativeValue || negativeBit;
          BitReference positiveBit = signInfo[2*cellIndex+1];
          positiveBit = !negativeValue || positiveBit;
        }

      }


      // Interpolate given level set function in vertex positions of
      // level set grid view for the given interface. The
      // interpolation values may be modified to make all edges comply
      // to the minimum cut-off condition.
      void interpolate(std::vector<bool> & signInfo)
      {
        // Allocate level set values
        levelSetValues_.clear();
        levelSetValues_.resize
          (domainDef_->levelSetGridView().size(dim),domainDef_->invalidValue());

        // Allocate sign info flags
        signInfo.clear();
        signInfo.resize(domainDef_->gridView().size(0)*2,false);

        typedef typename DomainDefinition::CoordinateComparator Comparator;
        Comparator& comp = const_cast<Comparator&>(domainDef_->coordinateComparator());

        // Iterate level set grid view
        typedef typename GridView::Traits::template Codim<0>::Iterator CellIterator;
        CellIterator it = domainDef_->gridView().template begin<0>();
        const CellIterator eit = domainDef_->gridView().template end<0>();
        for(; it != eit; ++it) {

          const GVCellEntity & cell = *it;

          // Iterate all sub cells of cell in lgv_
          using namespace MarchingCube33SubTriangulationImp;
          SubGridViewIterator<GridView,LevelSetGridView> hit
            (domainDef_->gridView(),domainDef_->levelSetGridView(),cell);
          for(; hit.valid(); ++hit) {

            const CellEntity & lCell = *hit;

            // Inform the coordinate comparator about cell geometry (if
            // not done yet).
            if(!comp.valid())
              comp.updateReferenceMap(lCell.geometry());

            // Interpolate values in corners of cell
            interpolateCellCorners(cell,lCell,signInfo);

          } // hit

        } // it

        // Finish setup of coordinate comparator (if not done yet).
        if(!comp.valid()) comp.finalize();
      }

    public:

      //! Constructor
      DefaultLevelSetBackend
        (const DomainDefinition & domainDef, const LevelSetFunction & lsf, const Index itfIndex)
        : domainDef_(&domainDef), lsf_(&lsf), itfIndex_(itfIndex)
      {
        static_assert(
            Dune::Functions::Concept::isGridViewFunction<LevelSetFunction, DomainField(Coordinate),
                                                         LevelSetGridView>(),
            "expecting a grid view function on the level set grid view");
      }

      //! Clear all cached values in the backend
      void reset(std::vector<bool> & signInfo)
      {interpolate(signInfo);}

      void printStatistics() const {}

      int cellCornerValues
        (const CellEntity & lCell,
        LevelSetValues & values) const
      {
        // Reference element of lgv cell in which to refine.
        auto ref =
          Dune::referenceElement<DomainField,dim>(lCell.type());

        // Values in corners of lCell
        values.resize(ref.size(dim));
        for(Index v=0; v<values.size(); ++v) {
          typedef typename LevelSetGridView::Traits::template Codim<dim>::Entity NodeEntity;
          const NodeEntity node = lCell.template subEntity<dim>(v);
          const Index nodeIndex = domainDef_->levelSetGridView().indexSet().index(node);
          values[v] = levelSetValues_[nodeIndex];

          assert(values[v]); // Zero values should have been removed
        }
        return getSign(values);
      }


      int partCornerValues
      (
        const CellEntity & lCell,
        const RefinementPart & part,
        LevelSetValues & values
      ) const
      {
        values.resize(part.geometry.corners());

        // Compute bi-linear interpolation values in embedding
        // (which has to be either a cube or a simplex)
        using namespace MarchingCube33SubTriangulationImp;
        typedef typename RefinementPart::Geometry Geometry;
        const Geometry localInEmbedding =
          combineGeometries(part.geometry,*(part.embedding),false);
        computeValuesInVertices
          (part.embedding->type(),(*(part.embeddingValues))[itfIndex_],localInEmbedding,values);
        normalizeValuesInVertices(localInEmbedding.type(),values);

        return getSign(values);
      }

      //! Access to level set function values
      LevelSetValues & levelSetFunctionValues()
      { return levelSetValues_; }
      //! Const access to level set function values
      const LevelSetValues & levelSetFunctionValues() const
      { return levelSetValues_; }


    protected:
      const DomainDefinition * domainDef_;
      const LevelSetFunction * lsf_;
      const Index itfIndex_;

      // Values of the level set function of this domain. The index is
      // given by the index of the corresponding vertex in lgv in the
      // index set of lgv.
      LevelSetValues levelSetValues_;
    };



    // A stub which is used if the level set function does not fulfill
    // the unfitted geometry function interface
    template< typename DomainDefinition, typename LevelSetFunction>
    class DefaultLevelSetBackend
    <DomainDefinition,LevelSetFunction,
     typename std::enable_if<LevelSetFunction::unfitted_geometry_function>::type> :
      public LevelSetBackendNodeStorageStub<DomainDefinition>{
    public:
      typedef typename LevelSetBackendNodeStorageStub<DomainDefinition>::Traits Traits;
      DefaultLevelSetBackend
        (const DomainDefinition & , const LevelSetFunction & , const typename Traits::Index){}
    };

  }
}

#endif
