// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_MC33TRIANGULATION_DOMAINCELL_HH
#define DUNE_MC33TRIANGULATION_DOMAINCELL_HH

#include <dune/udg/mc33triangulation/mc33triangulationimp.hh>
#include "domaincellface.hh"

namespace MarchingCube33SubTriangulationImp
{

  /**
   * \brief Data structure to represent a raw bounding box.
   *
   * \note Freshly initialized objects of this type are in an
   *       invalid state.
   *
   * \todo Avoid raw bounding box objects in an invalid state,
   *       e.g. call update on initialization.
   */
  template <typename DomainField, int dim>
  struct RawBoundingBox
  {
    //! Type of bounding box corners
    typedef Dune::FieldVector<DomainField,dim> Coordinate;

    //! Initializes a new bounding box object
    RawBoundingBox() :
      lower(std::numeric_limits<DomainField>::max()),
      higher(-(std::numeric_limits<DomainField>::max()-1.0))
    {}

    //! Extend bounding box such that it contains a given point
    void update(const Coordinate & x){
      for(int d=0; d<dim; ++d) {
        lower[d] = std::min(x[d],lower[d]);
        higher[d] = std::max(x[d],higher[d]);
      }
    }

    //! Extend bounding box such that it contains a given bounding box
    void update(const RawBoundingBox & b){
      this->update(b.lower);
      this->update(b.higher);
    }

    //! Lower and higher corner defining the bounding box
    Coordinate lower; Coordinate higher;
  };


  /**
   * \brief Data structure identifying a domain cell. A domain cell
   *        is in principle defined by a number of cells in the
   *        level set mesh grid view, a domain index and a (possibly
   *        exclusive) host cell in the fundamental mesh grid view.
   *
   * Notice that not all BUT SOME of its member cells in the level
   * set mesh grid view have to be contained within the host cell
   * in the fundamental mesh.
   *
   * Notice that all entity parts of the same domain within a given
   * cell in the level set mesh grid view have to be contained in the
   * same domain cell.
   *
   * \tparam GV Type of the fundamental mesh grid view.
   *
   * \todo Perhaps use invalidIndex() to tag an exterior boundary
   *       ghost-domaincell.
   */
  template <typename GV>
  class DomainCell
  {
  public:
    // Grid view types
    typedef typename GV::Traits::template Codim<0>::Entity CellEntity;
    typedef typename GV::Traits::template Codim<0>::Entity::EntitySeed CellEntitySeed;
    typedef typename GV::IndexSet::IndexType Index;
    typedef typename GV::ctype DomainField;
    typedef std::list<Index> IndexList;
    typedef GV GridView;
    enum { dim = GV::dimension };

    // Currently unused, see \todo
    // static Index invalidIndex(){ return std::numeric_limits<Index>::max(); }

    // Space coordinate types
    typedef Dune::FieldVector<DomainField,dim> Coordinate;
    typedef std::vector<Coordinate> Coordinates;

    // Face types
#if DUNE_VERSION_NEWER(DUNE_GEOMETRY,2,3)
    typedef Dune::CachedMultiLinearGeometry<DomainField, dim-1, dim> FaceGeometry;
#else
    //! Geometry traits type
    typedef typename Dune::GenericGeometry::
    DefaultGeometryTraits<DomainField,dim-1,dim> FaceGeometryTraits;
    //! Geometry type
    typedef Dune::GenericGeometry::
    BasicGeometry<dim-1, FaceGeometryTraits > FaceGeometry;
#endif
    typedef MarchingCube33SubTriangulationImp::DomainCellFace<GV> DomainCellFace;
    typedef std::pair<Index,Index> IndexPair;

    //! Type of raw bounding box
    typedef typename MarchingCube33SubTriangulationImp::
    RawBoundingBox<DomainField,dim> RawBoundingBox;

    //! Type of helper for comparing global coordinates
    typedef NormalizedCoordinateComparator<Coordinate> CoordinateComparator;

    /**
     * \brief Constructor.
     *
     * \todo Please doc me!
     */
    DomainCell
      (const CellEntity & _hostCell, const Index _id,
      const CoordinateComparator & _ncComp) :
      hostCell(_hostCell.seed()), volume(0), id(_id), ncComp(&_ncComp)
    {}

    /**
     * \brief Compare domain cell to a given domain cell by the id
     *        of both domain cells.
     */
    bool operator==(const DomainCell & other) const
    {
      return other.id == id;
    }

    /**
     * \brief Merge domain cell with a given domain cell without
     *        any regard to connectivity.
     *
     * This function performs the merging without removing matching
     * faces on both domain cells. Therefore, after the call is
     * complete, such faces will be contained as doublettes (except
     * that normal vectors will have different sign and possibly the
     * interior lgv cell index might not match). If you want to erase
     * such faces use DomainCell::mergeConnected(...) instead.
     *
     * \param[in] orig The domain cell to be merged with.
     *                 NOTICE that the data structures of this
     *                 cell will be corrupted in the course of
     *                 merging and should not be used afterwards.
     */
    void blindMerge(DomainCell & orig)
    {
      // This domain cell shall be destination of the merging
      DomainCell & dest = *this;

      #ifdef ALL_MC33TRIANGULATION_VERBOSE
      std::cout << "Blind merge domain cell id:" << orig.id << " to id:"
                << dest.id << " " << std::endl;
      #endif

      // Merge face areas
      {
        typedef typename DomainCell::IntersectionAreaMap AreaMap;
        typedef typename AreaMap::iterator AreaMapIt;
        AreaMap & origAreas = orig.areas;
        for(AreaMapIt ait = origAreas.begin(); ait != origAreas.end(); ++ait) {
          if(ait->first == orig.id) {
            dest.areas[dest.id] += ait->second;
          }
          else {
            std::pair<AreaMapIt,bool> res = dest.areas.insert(*ait);
            if(!res.second)
              res.first->second += ait->second;
          }
        }
        dest.areas.erase(orig.id);
      }

      #ifdef ALL_MC33TRIANGULATION_VERBOSE
      typedef typename DomainCell::IntersectionAreaMap AreaMap;
      typedef typename AreaMap::iterator AreaMapIt;
      for(AreaMapIt ait = dest.areas.begin(); ait != dest.areas.end(); ++ait) {
        std::cout << "\t Neighbor id:" << ait->first
                  << " area: " << ait->second << std::endl;
      }
      #endif

      // Merge domain cell faces
      {
        typedef typename DomainCell::DomainCellFaceListMap DomainCellFaceListMap;
        DomainCellFaceListMap & faceListMap = orig.faceListMap;

        typedef typename DomainCellFaceListMap::iterator DomainCellFaceListToupleIt;
        DomainCellFaceListToupleIt fltit = faceListMap.begin();
        for(; fltit != faceListMap.end(); ++fltit) {

          // Make sure key in dest.faceListMap exist
          typedef typename DomainCell::DomainCellFaceList DomainCellFaceList;
          DomainCellFaceListToupleIt destFaceListToupleIt =
            dest.faceListMap.insert(std::make_pair(fltit->first,DomainCellFaceList())).first;

          // Iterate all faces in orig
          DomainCellFaceList & origFaceList = fltit->second;
          typename DomainCellFaceList::iterator fit = origFaceList.begin();
          for(; fit != origFaceList.end(); ++fit) {

            // Update the insideDomainCell pointer
            fit->insideDomainCell = &dest;

            // If face of disappearing cell has a neighbor, update its
            // area map:
            typedef typename DomainCellFace::DomainCellFacePointers FacePointers;
            typename FacePointers::const_iterator mirrorFaceIt = fit->allMirrorFaces().begin();
            for(; mirrorFaceIt != fit->allMirrorFaces().end(); ++mirrorFaceIt) {

              // Remove area entry of disappearing domain cell
              typedef typename DomainCell::IntersectionAreaMap AreaMap;
              AreaMap & neighborAreas = (*mirrorFaceIt)->insideDomainCell->areas;
              neighborAreas.erase(orig.id);

              // Add new area entry into area map of neighbor
              const DomainField area = (*mirrorFaceIt)->area;
              typedef typename AreaMap::iterator AreaMapIt;
              std::pair<AreaMapIt,bool> iRes =
                neighborAreas.insert(std::make_pair(dest.id,area));
              if(!iRes.second) iRes.first->second += area;
            }

          }

          // Merge faces to dest
          DomainCellFaceList & destFaceList = destFaceListToupleIt->second;
          assert(destFaceList.size() == 0);
          destFaceList.swap(origFaceList);
        }
      }

      // Merge gvMembers
      dest.gvMembers.insert(orig.gvMembers.begin(),orig.gvMembers.end());

      // Merge lgvMembers
      {
        typedef typename DomainCell::LGVMemberMap LGVMemberMap;
        for(typename LGVMemberMap::iterator it = orig.lgvMembers.begin();
            it != orig.lgvMembers.end(); ++it) {
          assert(it->second.size());

          if(!dest.lgvMembers.insert(*it).second) {
            typedef typename LGVMemberMap::mapped_type IndexList;
            IndexList & destList = dest.lgvMembers[it->first];
            IndexList & origList = it->second;
            destList.splice(destList.begin(),origList,origList.begin(),origList.end());
          }
        } // it
      }

      // Merge bounding boxes and add volume
      dest.bbox.update(orig.bbox);
      dest.volume += orig.volume;
    }

    /**
     * \brief Cut-off all domain cell faces which connect the domain cell
     *        to a domain cell of a given id. A cut-off face connects the
     *        domain cell to itself.
     *
     * \param id All faces connected to the domain cell of this id will be
     *           removed.
     * \return   The number of cut-off faces.
     */
    int cutOffFacesById(const Index id)
    {
      // Face types
      typedef std::list<DomainCellFace> DomainCellFaceList;
      typedef typename DomainCell::DomainCellFaceListMap DomainCellFaceListMap;
      typedef typename DomainCellFaceListMap::iterator DomainCellFaceListToupleIt;
      typedef typename DomainCell::DomainCellFaceList DomainCellFaceList;
      typedef typename DomainCellFaceList::iterator DomainCellFaceIt;

      int matches(0);
      DomainCellFaceListToupleIt faceListToupleIt = this->faceListMap.begin();
      for(; faceListToupleIt != this->faceListMap.end(); ++faceListToupleIt) {

        DomainCellFaceIt faceIt = faceListToupleIt->second.begin();
        while(faceIt != faceListToupleIt->second.end()) {

          typedef typename DomainCellFace::DomainCellFacePointers FacePointers;
          typename FacePointers::iterator mirrorFaceIt = faceIt->allMirrorFaces().begin();
          for(; mirrorFaceIt != faceIt->allMirrorFaces().end(); ++mirrorFaceIt) {

            // Cut-off face (i.e. connect domain cell to itself)
            if((*mirrorFaceIt)->insideDomainCell->id == id) {
              (*mirrorFaceIt) = &(*faceIt);
              matches++;
            }
          }
          ++faceIt;

        } // faceIt

      } // faceListToupleIt
      return matches;
    }

    /**
     * \brief Remove all domain cell faces which have been cut-off.
     *        A cut-off face connects the domain cell to itself.
     *
     * \return The number of removed faces.
     */
    int removeCutOffFaces()
    {
      // Face types
      typedef std::list<DomainCellFace> DomainCellFaceList;
      typedef typename DomainCell::DomainCellFaceListMap DomainCellFaceListMap;
      typedef typename DomainCellFaceListMap::iterator DomainCellFaceListToupleIt;
      typedef typename DomainCell::DomainCellFaceList DomainCellFaceList;
      typedef typename DomainCellFaceList::iterator DomainCellFaceIt;

      int matches(0);
      DomainCellFaceListToupleIt faceListToupleIt = this->faceListMap.begin();
      for(; faceListToupleIt != this->faceListMap.end(); ++faceListToupleIt) {

        DomainCellFaceIt faceIt = faceListToupleIt->second.begin();
        while(faceIt != faceListToupleIt->second.end()) {

          bool foundCutOff = false;

          // Remove cut off faces from mirror list
          typedef typename DomainCellFace::DomainCellFacePointers FacePointers;
          typename FacePointers::iterator mirrorFaceIt = faceIt->allMirrorFaces().begin();
          while( mirrorFaceIt != faceIt->allMirrorFaces().end() ) {

            if((*mirrorFaceIt) == &(*faceIt)) {
              foundCutOff = true;
              faceIt->allMirrorFaces().erase(mirrorFaceIt);
              mirrorFaceIt = faceIt->allMirrorFaces().begin();
              matches++;
              continue;
            }
            ++mirrorFaceIt;
          }

          // Remove this face if it has no remaining mirrors
          if(foundCutOff && faceIt->allMirrorFaces().size() == 0) {
            faceIt = faceListToupleIt->second.erase(faceIt);
            continue;
          }
          ++faceIt;

        } // faceIt

      } // faceListToupleIt
      return matches;
    }

    /**
     * \brief Merge domain cell with a given domain cell using
     *        established connectivity information.
     *
     * If the domain cell and the given domain cell are different
     * domain cells from the same domain which have common faces,
     * these faces are removed and DomainCell::blindMerge(...) is
     * called. Otherwise no modifications are done.
     *
     * \param[in] orig The domain cell to be merged with.
     *                 NOTICE that the data structures of this
     *                 cell will be corrupted in the course of
     *                 merging and should not be used afterwards.
     *
     * \return         True only if the merging was performed.
     */
    bool mergeConnected(DomainCell & orig)
    {
      // This domain cell shall be destination of the merging
      DomainCell & dest = *this;

      // Only merge domain cells from the same domain
      if(orig.domainIndex != dest.domainIndex)
        return false;

      // Don't merge domain cell with itself
      if(orig.id == dest.id)
        return false;

      // Cut-off all domain cell faces which connect the domain
      // cell to orig
      const int destMatches = dest.cutOffFacesById(orig.id);
      if(destMatches==0) return false;

      // Cut-off all domain cell faces which connect orig to the
      // domain cell
      const int origMatches = orig.cutOffFacesById(dest.id);
      assert(origMatches == destMatches);

      // Remove all domain cell faces of the domain cell and orig
      // which have been cut-off
      const int destCutOff = dest.removeCutOffFaces();
      const int origCutOff = orig.removeCutOffFaces();

      // Assert that removing was successful
      assert(destCutOff == destMatches);
      assert(origCutOff == origMatches);

      // Remove warnings when not debugging
      ((void)(origMatches));
      ((void)(destCutOff));
      ((void)(origCutOff));

      #ifdef ALL_MC33TRIANGULATION_VERBOSE
      std::cout << "Merge domain cell id:" << orig.id << " to id:" << dest.id
                << std::endl;
      #endif

      // Call DomainCell::blindMerge
      dest.blindMerge(orig);
      return true;
    }


    /** \brief This function finds and initializes the connectivity
        relation to the given domain cell.

        After this call is completed, this domain cell may hold a
        pointer to the domain cell which was passed to this function
        and vice versa. Therefore, after this call is completed,
        neither of the two domain cells should be copied any more (or
        otherwise relocated in memory).

        \param orig     The domain cell with which to connect.
        \return         The number of matching domain cell faces.
     */
    template <typename DomainCell>
    int
    connectFaces(DomainCell & orig)
    {
      // This domain cell shall be called "destination"
      DomainCell & dest = *this;

      // Directly return with value 0 if the two domain cells are identical
      if(orig.id == dest.id)
        return 0;

      // Face types
      typedef typename DomainCell::DomainCellFace DomainCellFace;
      typedef std::list<DomainCellFace> DomainCellFaceList;

      // Connect all matching domain cell faces
      int matches(0);

      typedef typename DomainCell::DomainCellFaceListMap DomainCellFaceListMap;
      typedef typename DomainCellFaceListMap::iterator DomainCellFaceListToupleIt;
      DomainCellFaceListToupleIt origFaceListToupleIt = orig.faceListMap.begin();
      for(; origFaceListToupleIt != orig.faceListMap.end(); ++origFaceListToupleIt) {
        DomainCellFaceListToupleIt destFaceListToupleIt = dest.faceListMap.begin();
        for(; destFaceListToupleIt != dest.faceListMap.end(); ++destFaceListToupleIt) {

          typedef typename DomainCell::DomainCellFaceList DomainCellFaceList;
          typedef typename DomainCellFaceList::iterator DomainCellFaceIt;

          DomainCellFaceIt destFaceIt = destFaceListToupleIt->second.begin();
          for(; destFaceIt != destFaceListToupleIt->second.end(); ++destFaceIt) {

            DomainCellFaceIt origFaceIt = origFaceListToupleIt->second.begin();
            for(; origFaceIt != origFaceListToupleIt->second.end(); ++origFaceIt) {

              // Skip pair if both are already connected
              if(origFaceIt->mirrorFace() && destFaceIt->mirrorFace())
                continue;

              // Check for mutual inclusion
              if( origFaceIt->isOverlapping(*destFaceIt) || destFaceIt->isOverlapping(*origFaceIt) ) {

                // assert(!(origFaceIt->mirrorFace) ||
                //        destFaceIt->insideDomainCell->id ==
                //        origFaceIt->mirrorFace->insideDomainCell->id);

                // assert(!(destFaceIt->mirrorFace) ||
                //        origFaceIt->insideDomainCell->id ==
                //        destFaceIt->mirrorFace->insideDomainCell->id);


                origFaceIt->setMirrorFace( &(*destFaceIt) );
                destFaceIt->setMirrorFace( &(*origFaceIt) );

                // Update area maps
                typedef typename DomainCell::IntersectionAreaMap AreaMap;
                typedef typename AreaMap::iterator AreaMapIt;
                typedef typename DomainCell::DomainField DomainField;
                typedef typename DomainCell::Index Index;
                const DomainField area = std::min(origFaceIt->area,destFaceIt->area);
                const Index destID = destFaceIt->insideDomainCell->id;
                const Index origID = origFaceIt->insideDomainCell->id;

                #ifdef ALL_MC33TRIANGULATION_VERBOSE
                std::cout << "Connected domain cell id:" << origID << " to id:" << destID<<" "
                          << std::endl;
                #endif

                {
                  AreaMap & origAreas = origFaceIt->insideDomainCell->areas;
                  std::pair<AreaMapIt,bool> res = origAreas.insert(std::make_pair(destID,area));
                  if(!res.second)
                    res.first->second += area;
                  origAreas[origID] -= area;
                }
                {
                  AreaMap & destAreas = destFaceIt->insideDomainCell->areas;
                  std::pair<AreaMapIt,bool> res = destAreas.insert(std::make_pair(origID,area));
                  if(!res.second)
                    res.first->second += area;
                  destAreas[destID] -= area;
                }

                matches++;
              }


            } // origFaceIt
          } // destFaceIt

        } // destFaceListToupleIt
      } // origFaceListToupleIt

      return matches;
    }


    /** \brief Provides a map which maps the domain cell indices of each
        adjacent domain cell to a pointer to that adjacent domain cell.

        \param[out] connections            Map to be filled
        \param[in]  ignoreForeignDomains   If the flag ignoreForeignDomains is set, then
                                           only connections to domain cells with the same
                                           domain index will be added.
     */
    template <typename DomainCell>
    void neighbors
      (std::map<typename DomainCell::Index, DomainCell*> & connections,
      bool ignoreForeignDomains = false)
    {
      DomainCell & domainCell = *this;

      // Iterate all faces of domain cell
      typedef typename DomainCell::DomainCellFaceListMap DomainCellFaceListMap;
      DomainCellFaceListMap & faceListMap = domainCell.faceListMap;
      typedef typename DomainCellFaceListMap::iterator DomainCellFaceListToupleIt;
      DomainCellFaceListToupleIt fltit = faceListMap.begin();
      for(; fltit != faceListMap.end(); ++fltit) {

        typedef typename DomainCell::DomainCellFaceList DomainCellFaceList;
        DomainCellFaceList & faceList = fltit->second;
        typename DomainCellFaceList::iterator fit = faceList.begin();
        for(; fit != faceList.end(); ++fit) {

          // new solution:
          typedef typename DomainCellFace::DomainCellFacePointers FacePointers;
          typename FacePointers::iterator mirrorFaceIt = fit->allMirrorFaces().begin();
          for(; mirrorFaceIt != fit->allMirrorFaces().end(); ++mirrorFaceIt) {
            DomainCell * neighbor = (*mirrorFaceIt)->insideDomainCell;
            if(ignoreForeignDomains && neighbor->domainIndex != domainCell.domainIndex)
              continue;
            connections.insert(std::make_pair(neighbor->id, neighbor));
          }

          // // old solution:
          // if(fit->mirrorFace()){
          //   DomainCell * neighbor = fit->mirrorFace->insideDomainCell;
          //   if(ignoreForeignDomains && neighbor->domainIndex != domainCell.domainIndex)
          //     continue;
          //   connections.insert(std::make_pair(neighbor->id, neighbor));
          // }
        }
      }
    }


    /**
     * \brief Given a container of domain cells, merge those
     *        domain cells which are connected and from the
     *        same domain.
     */
    template <typename DomainCellContainer>
    static void mergeAdjacentDomainCells
      (DomainCellContainer & localDomainCells)
    {
      typedef typename DomainCellContainer::iterator Iterator;
      using namespace MarchingCube33SubTriangulationImp;

      // Merge domain cells
      bool merged = true;
      while(merged) { merged = false;

                      Iterator it = localDomainCells.begin();
                      for(; it != localDomainCells.end(); ++it) {

                        Iterator next(it); ++next;
                        while(next != localDomainCells.end()) {

                          // Either merge or proceed increment next
                          if(it->mergeConnected(*next)) {
                            merged = true;
                            next = localDomainCells.erase(next);
                          }
                          else
                            next++;
                        }
                      } // it
      }
    }

    /**
     * \brief Output the current list of outward faces of this domain cell.
     */
    void outputFaceList() const
    {
      typedef std::list<DomainCellFace> DomainCellFaceList;
      typedef std::map<IndexPair,DomainCellFaceList> DomainCellFaceListMap;

      std::cout << "Output of faces of domain cell id:" << id << std::endl;

      // Iterate elements of map which contains the outward faces of this domain cell
      typedef typename DomainCellFaceListMap::const_iterator DomainCellFaceListToupleIt;
      DomainCellFaceListToupleIt fltit = faceListMap.begin();
      for (; fltit != faceListMap.end(); ++fltit)
      {
        // Get current key and face list
        const IndexPair & key = fltit->first;
        const DomainCellFaceList & faceList = fltit->second;

        // Output current key
        std::cout << "\tKey: " << key.first << "," << key.second << std::endl;

        // Output current face list
        typedef typename DomainCellFaceList::const_iterator DomainCellFaceIt;
        DomainCellFaceIt fit = faceList.begin();
        for(; fit != faceList.end(); ++fit)
        {
          std:: cout << "\t\tCenter " << fit->geometry.center() << std::endl;
          if(fit->mirrorFace())
            std::cout << "\t\tOutside id:"
                      << fit->mirrorFace()->insideDomainCell->id << std::endl;
          else
            std::cout << "\t\tOutside id:" << id << std::endl;
        }
      }
    }

  public:
    //! Map which contains for all cells in the level set mesh grid
    //! view, that contribute parts to this domain cell, the indices
    //! of the refinement parts belonging to this domain cell
    typedef std::map<Index,IndexList> LGVMemberMap;
    LGVMemberMap lgvMembers;

    //! Map which contains entity pointers for all cells in the
    //! fundamental mesh grid view whose children in the level
    //! set mesh grid view contribute parts to this domain cell
    typedef std::map<Index,CellEntitySeed> MemberCellMap;
    MemberCellMap gvMembers;

    //! Domain index of this domain cell
    Index domainIndex;

    //! Host cell of this domain cell
    CellEntitySeed hostCell;

    //! Raw bounding box of this domain cell
    RawBoundingBox bbox;

    //! Volume of this domain cell
    DomainField volume;

    //! Unique id of domain cell
    Index id;

    //! Map containing all the outward faces of this domain cell. The
    //! key is a pair consisting of the <lgv cell index, entity part
    //! index> which contains the face.
    typedef std::list<DomainCellFace> DomainCellFaceList;
    typedef std::map<IndexPair,DomainCellFaceList> DomainCellFaceListMap;
    DomainCellFaceListMap faceListMap;

    //! Pointer to helper for comparing global coordinates
    const CoordinateComparator * ncComp;

    //! Map index of neighboring domain cell to intersection area with
    //! this domain cell
    typedef std::map<Index,DomainField> IntersectionAreaMap;
    IntersectionAreaMap areas;
  };

}
#endif
