// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_MC33TRIANGULATION_DOMAINCELLFACE_HH
#define DUNE_MC33TRIANGULATION_DOMAINCELLFACE_HH

#if DUNE_VERSION_NEWER(DUNE_GEOMETRY,2,3)
#include <dune/geometry/multilineargeometry.hh>
#else
#include <dune/geometry/genericgeometry/geometry.hh>
#endif

namespace MarchingCube33SubTriangulationImp
{

  /** Declaration of DomainCell required as DomainCellFace keeps a
      pointer on its owner DomainCell object. */
  template < typename GV > class DomainCell;


  template< typename GV >
  class DomainCellFace {
  public:
    //! Index type
    typedef typename GV::IndexSet::IndexType Index;
    //! Dimension of computational domain
    enum { dim = GV::dimension };
    //! Field type of domain coordinate.
    typedef typename GV::ctype DomainField;
    //! Domain coordinate.
    typedef Dune::FieldVector<DomainField, dim> Coordinate;
    //! Face coordinate.
    typedef Dune::FieldVector<DomainField, dim-1> FaceCoordinate;

    //! Cell entity pointer in grid view.
    typedef typename GV::Traits::template Codim<0>::Entity CellEntity;

#if DUNE_VERSION_NEWER(DUNE_GEOMETRY,2,3)
    typedef Dune::CachedMultiLinearGeometry<DomainField, dim-1, dim> FaceGeometry;
#else
    //! Geometry traits type
    typedef typename Dune::GenericGeometry::
    DefaultGeometryTraits<DomainField,dim-1,dim> FaceGeometryTraits;
    //! Geometry type
    typedef Dune::GenericGeometry::
    BasicGeometry<dim-1, FaceGeometryTraits > FaceGeometry;
#endif

    //! Type of list of mirror faces
    typedef std::list<DomainCellFace*> DomainCellFacePointers;

  private:
    /** \brief Returns the indices of the corners adjacent to the
        corner with local index c within this face.

        This function should not be called for dim==2.
     */
    std::pair<int,int> adjacentIndices(const int c, const FaceGeometry & g) const
    {
      if(g.corners()==3)
        return std::make_pair(int((c+1)%3),int((c+2)%3));
      else if(g.corners()==4) {
        if(c==0 || c==3) return std::make_pair(int(1),int(2));
        else return std::make_pair(int(0),int(3));
      }
      else{DUNE_THROW(Dune::Exception,"Illegal number of face corners!");}
    }

    /** \brief Compute cross product assuming 3D vector parameters */
    static Coordinate crossProd(const Coordinate & a, const Coordinate & b)
    {
      Coordinate c;
      c[0] = a[1]*b[2] - a[2]*b[1];
      c[1] = a[2]*b[0] - a[0]*b[2];
      c[2] = a[0]*b[1] - a[1]*b[0];
      return c;
    }

    /** \brief Checks whether the given face is linear i.e. contained
        in a linear plane. */
    static bool isLinear
      (const DomainCellFace & face)
    {
      if(face.geometry.corners() <= 3) return true;

      const typename DomainCell<GV>::CoordinateComparator &
      ncComp = *(face.insideDomainCell->ncComp);

      const Coordinate A = ncComp.inRef(face.geometry.corner(0));
      const Coordinate B = ncComp.inRef(face.geometry.corner(1));
      const Coordinate C = ncComp.inRef(face.geometry.corner(2));
      const Coordinate P = ncComp.inRef(face.geometry.corner(3));

      const Coordinate vb = (B-A);
      const Coordinate vc = (C-A);
      const Coordinate vp = (P-A);

      // Orthogonal vector to triangle plane
      Coordinate vk = crossProd(vb,vc); vk /= vk.two_norm();

      // Check whether P in plane spanned by ABC
      Coordinate vpProj(vp); vpProj.axpy(-(vk*vp),vk);
      if(!ncComp.isEqualInRef(vpProj,vp)) return false;

      return true;
    }

    /** \brief This function checks whether this face overlaps with
        the given face and both are assumed to be linear. If the given
        face is not a triangle you may only check for overlap with
        either its criss-triangle part (triangle==0) or its
        cross-triangle part (triangle==1).
     */
    bool overlapsTriangle
      (const DomainCellFace & oppositeFace, int triangle = 0) const
    {
      const DomainCellFace & face = *this;

      const typename DomainCell<GV>::CoordinateComparator &
      ncComp = *(face.insideDomainCell->ncComp);
      const DomainField & eps = ncComp.numericalEqualityEps();

      // Map to local coords
      const FaceGeometry & oppFaceGeom = oppositeFace.geometry;
      const Coordinate A = ncComp.inRef(oppFaceGeom.corner(0+triangle));
      const Coordinate B = ncComp.inRef(oppFaceGeom.corner(1+triangle));
      const Coordinate C = ncComp.inRef(oppFaceGeom.corner(2+triangle));

      const Coordinate vb = (B-A);
      const Coordinate vc = (C-A);

      const DomainField vbvc   = (vb*vc);
      const DomainField vb2   = (vb*vb);
      const DomainField vc2   = (vc*vc);
      const DomainField cosBC2 = vb2*vc2 - vbvc*vbvc;

      // Orthogonal vector to triangle plane
      Coordinate vk = crossProd(vb,vc); vk /= vk.two_norm();

      bool isOutside = false;
      for(int c=0; c<face.geometry.corners(); ++c) {
        const Coordinate P = ncComp.inRef(face.geometry.corner(c));
        const Coordinate vp = (P-A);

        // Check whether P in plane spanned by ABC
        Coordinate vpProj(vp); vpProj.axpy(-(vk*vp),vk);
        if(!ncComp.isEqualInRef(vpProj,vp)) return false;

        const DomainField baryCoord1 = ( (vp*vb)*vc2 - (vp*vc)*vbvc ) / cosBC2;
        const DomainField baryCoord2 = ( (vp*vc)*vb2 - (vp*vb)*vbvc ) / cosBC2;

        // Return true immediately if p really inside (not only on edge)
        if( baryCoord1 > eps && baryCoord2 > eps && baryCoord2 + baryCoord1 < 1.0-eps )
          return true;

        isOutside |= baryCoord1 < -eps || baryCoord2 < -eps || (baryCoord1 + baryCoord2 > 1.0+eps );

      }
      return !isOutside;
    }

    /** \brief This function checks whether this face overlaps with
        the given face and both are assumed to be lines in 2D space.
     */
    bool overlapsLine
      (const DomainCellFace & b) const
    {
      const DomainCellFace & a = *this;
      const typename DomainCell<GV>::CoordinateComparator &
      ncComp = *(a.insideDomainCell->ncComp);
      const DomainField & eps = ncComp.numericalEqualityEps();

      const Coordinate A = ncComp.inRef(a.geometry.corner(0));
      const Coordinate B = ncComp.inRef(a.geometry.corner(1));
      const Coordinate va = (B-A);
      const DomainField va2 = va*va; assert(va2);

      // Orthogonal vector on line
      Coordinate vk; vk[0] = -va[1]; vk[1] = va[0]; vk /= vk.two_norm();

      bool isOutside = false;
      for(Index c=0; c<2; ++c) {
        const Coordinate P = ncComp.inRef(b.geometry.corner(c));
        const Coordinate vp = (P-A);

        // Check whether P in line spanned by AB
        Coordinate vpProj(vp); vpProj.axpy(-(vk*vp),vk);
        if(!ncComp.isEqualInRef(vpProj,vp)) return false;

        const DomainField baryCoord = (vp * va) / va2;

        // Return true immediately if p really inside (not only on edge)
        if(baryCoord > eps && baryCoord < 1.0-eps) return true;

        isOutside |= baryCoord < -eps || baryCoord > 1.0+eps;
      }

      return !isOutside;
    }

    /** \brief This function checks whether this face's vertices are
        contained within the face b.

        \return True, if either all vertices of this face are not
        outside of b, or if at least one vertex is completely (>eps)
        within b. Otherwise returns false.
     */
    bool containsFaceVertices
      (const DomainCellFace & b) const
    {
      //const DomainCellFace & a = *this;

      if(dim==2) {
        return overlapsLine(b);
      }
      else if(dim==3) {
        bool overlaps(false);

        // Check whether a overlaps with b (with its cross-triangle
        // part if b is quadrilateral).
        overlaps |= overlapsTriangle(b,0);

        // If this is a quadrilateral we have to check the
        // criss-triangle part of b too.
        if(b.geometry.corners()>3) {
          assert(b.geometry.corners()==4);
          overlaps |= overlapsTriangle(b,1);
        }
        return overlaps;
      }
      else{DUNE_THROW(Dune::NotImplemented,"Wrong dimension");}

    }

    /** \brief This function assumes that this face and face b are in
        the same plane and share at least one corner. Based on this
        assumption it determines whether the faces overlap.

        \param[in] b            The domain cell face to be checked.
     */
    bool isAngleOverlapping
      (const DomainCellFace & b) const
    {
      const DomainCellFace & a = *this;

      const typename DomainCell<GV>::CoordinateComparator &
      ncComp = *(a.insideDomainCell->ncComp);
      const DomainField & eps = ncComp.numericalEqualityEps();

      if(dim==2) {
        std::vector<Coordinate> aCoords(2);
        std::vector<Coordinate> bCoords(2);
        for(int c=0; c<2; ++c) {
          aCoords[c] = ncComp.inRef(a.geometry.corner(c));
          bCoords[c] = ncComp.inRef(b.geometry.corner(c));
        }

        const DomainField la = (aCoords[0]-aCoords[1]).two_norm();
        for(int ia=0; ia<2; ia++)
          for(int ib=0; ib<2; ib++)
            if( (aCoords[ia]-bCoords[ib]).two_norm() > la + eps)
              return false;

        return true;
      }
      else if(dim==3) {

        const FaceGeometry & ga = a.geometry;
        const FaceGeometry & gb = b.geometry;

        for(int ca=0; ca<ga.corners(); ++ca) {
          const Coordinate aCorner = ncComp.inRef(ga.corner(ca));

          for(int cb=0; cb<gb.corners(); ++cb) {
            const Coordinate bCorner = ncComp.inRef(gb.corner(cb));

            if(ncComp.isEqualInRef(aCorner,bCorner)) {
              const std::pair<int,int> adjacentInA = adjacentIndices(ca,ga);
              const std::pair<int,int> adjacentInB = adjacentIndices(cb,gb);

              const Coordinate aCornerLeft = ga.corner(adjacentInA.first);
              Coordinate aEdgeLeft = ncComp.inRef(aCornerLeft) - aCorner;
              assert(aEdgeLeft.two_norm2());
              aEdgeLeft /= aEdgeLeft.two_norm();
              const Coordinate aCornerRight = ga.corner(adjacentInA.second);
              Coordinate aEdgeRight = ncComp.inRef(aCornerRight) - aCorner;
              assert(aEdgeRight.two_norm2());
              aEdgeRight /= aEdgeRight.two_norm();
              Coordinate aEdge = aEdgeLeft+aEdgeRight;
              if(aEdge.two_norm() < eps) continue;
              aEdge /= aEdge.two_norm();

              const Coordinate bCornerLeft = gb.corner(adjacentInB.first);
              Coordinate bEdgeLeft = ncComp.inRef(bCornerLeft) - bCorner;
              assert(bEdgeLeft.two_norm2());
              bEdgeLeft /= bEdgeLeft.two_norm();
              const Coordinate bCornerRight = gb.corner(adjacentInB.second);
              Coordinate bEdgeRight = ncComp.inRef(bCornerRight) - bCorner;
              assert(bEdgeRight.two_norm2());
              bEdgeRight /= bEdgeRight.two_norm();
              Coordinate bEdge = bEdgeLeft+bEdgeRight;
              if(bEdge.two_norm() < eps) continue;
              bEdge /= bEdge.two_norm();

              const DomainField wBcBr = - 1.0 * (bEdge * bEdgeRight) + 1.0;
              const DomainField wAcAr = - 1.0 * (aEdge * aEdgeRight) + 1.0;
              const DomainField wBcAc = - 1.0 * (bEdge * aEdge) + 1.0;
              if( wBcAc < wBcBr || wBcAc < wAcAr)
                return true;

              return false;
            }

          }

        }
      }
      DUNE_THROW(Dune::Exception,"This should not be called for "
                 "faces without common corners!");
    }


    /** \brief Checks whether these two normal vectors match.

        The vectors are considered to be matched if they have max. 3
        degrees of deviation and opposite sign (if sign=1.0) or same
        sign (if sign=-1.0).

        \todo Detailed part of documentation seems outdated (the part
              with the maximum deviation in degrees).

        \todo This method uses a hardcoded numerical epsilon for its
              decision. We should change this as this might cause errors?
     */
    template< typename Normal>
    bool
    matchingUnitNormalVectors
      (const Normal & b,
      const typename Normal::value_type sign) const
    {
      const typename DomainCell<GV>::CoordinateComparator &
      ncComp = *(this->insideDomainCell->ncComp);

      Normal a = this->unitOuterNormal;
      a *= -sign;
      return ncComp.isEqualInRef(b,a,1e-5);
    }

  public:
    //! Constructor
    DomainCellFace
      (DomainCell<GV> * inside, const Index _insideCellIndex, const Index _insideLCellIndex,
      const Index _insidePartIndex, const Index _indexInPart, const FaceGeometry & _faceGeometry,
      const Coordinate & _unitOuterNormal) :
      insideDomainCell(inside),
      insideCellIndex(_insideCellIndex),
      insideLCellIndex(_insideLCellIndex),
      insidePartIndex(_insidePartIndex),
      indexInPart(_indexInPart),
      geometry(_faceGeometry),
      unitOuterNormal(_unitOuterNormal),
      area(_faceGeometry.volume())
    { linear = isLinear(*this); }


    /** \brief Checks whether this domain cell face overlaps with a
        given domain cell face.


        \param[in] b            The domain cell face to be checked.
        \param[in] min_cos      Max. matching value for scalar product of normal vectors.
                                If the value is higher, the faces are considered to be
                                not matched.
     */
    bool isOverlapping
      (const DomainCellFace & b) const
    {
      const DomainCellFace & a = *this;
      const typename DomainCell<GV>::CoordinateComparator &
      ncComp = *(a.insideDomainCell->ncComp);

      // Check matching normal vectors ( this ignores the sign, as the
      // normal vector computation for pyramids does not guarentee
      // correct sign).
      if(!matchingUnitNormalVectors(b.unitOuterNormal,1.0) &&
         !matchingUnitNormalVectors(b.unitOuterNormal,-1.0))
        return false;

      // If all corners of one face match, then it is completely
      // contained within the other face.
      const int nCommonCorners = ncComp.commonCorners(a.geometry,b.geometry);
      if(nCommonCorners == std::min(a.geometry.corners(),b.geometry.corners()))
        return true;

      // Non-linear faces have to be matched exactly
      assert(dim==3 || (a.linear&&b.linear));
      if(!a.linear || !b.linear) return false;

      // If faces have common corner, check angle condition
      if(nCommonCorners >= 1) {
        if(!isAngleOverlapping(b))
          return false;
      }
      // If the faces have no common corner, check whether b contains
      // vertices of this face.
      else if(!containsFaceVertices(b))
        return false;

      return true;
    }

    /** \brief Set the given domain cell face as mirror face

        If this face already has a mirror, then this function does
        some consistency checks.
     */
    void setMirrorFace(DomainCellFace * mirror)
    {
      // If mirror faces list is empty (this is the most likely case)
      // just add the given face as mirror and return
      mirrorFaces.push_back(mirror);
      if(mirrorFaces.size()==1) return;

      // Otherwise choose the face with the biggest area as mirror
      // (and the same inside domain index if possible).
      const Index myDomainIndex = insideDomainCell->domainIndex;

      typename DomainCellFacePointers::iterator chosen = mirrorFaces.begin();
      typename DomainCellFacePointers::iterator it = mirrorFaces.begin();
      ++it;
      for(; it != mirrorFaces.end(); ++it) {
        const Index itFaceDomainIndex=(*it)->insideDomainCell->domainIndex;
        const Index chosenDomainIndex=(*chosen)->insideDomainCell->domainIndex;
        if( chosenDomainIndex == myDomainIndex && itFaceDomainIndex != myDomainIndex)
          continue;

        if((*chosen)->area < (*it)->area)
          chosen = it;
      }

      // Sort the chosen element as the first
      if(chosen != mirrorFaces.begin())
        mirrorFaces.splice(mirrorFaces.begin(),mirrorFaces,chosen);

#ifndef NDEBUG
#ifdef ALL_MC33TRIANGULATION_VERBOSE
      typename std::list<DomainCellFace*>::iterator fit = mirrorFaces.begin();
      for(; fit!=mirrorFaces.end(); ++fit) {
        typename std::list<DomainCellFace*>::iterator nfit = fit;
        nfit++;
        for(; nfit != mirrorFaces.end(); ++nfit) {
          if((*nfit)->insideLCellIndex == (*fit)->insideLCellIndex) {

            if((*fit)->insideDomainCell->domainIndex
               != (*nfit)->insideDomainCell->domainIndex) {
              std::cerr << "Found face with non-unique mirror "
                        << "at " << geometry.center()  << std::endl;
              std::cerr << "Domain indices "
                        << (*fit)->insideDomainCell->domainIndex << " and "
                        << (*nfit)->insideDomainCell->domainIndex
                        << " are ambiguous." << std::endl;
            }
          }
          else{
            if((*fit)->insideDomainCell->id != (*nfit)->insideDomainCell->id) {
              std::cerr << "Found face with non-unique mirror "
                        << "at " << geometry.center()  << std::endl;
            }
          }

        }
      }
#endif
#endif
    }

    //! Check whether this face has at least one mirror face
    const DomainCellFace* mirrorFace() const
    {
      return bool(mirrorFaces.size()) ? mirrorFaces.front() : 0;
    }

    //! Return list of mirror faces
    DomainCellFacePointers & allMirrorFaces()
    {
      return mirrorFaces;
    }

    //! Return list of mirror faces
    const DomainCellFacePointers & allMirrorFaces() const
    {
      return mirrorFaces;
    }

  public:
    //! Inside domain cell
    DomainCell<GV> * insideDomainCell;
    //! Indices of gv_ cells containing the corresponding entity parts
    const Index insideCellIndex;
    //! Indices of lgv_ cells containing the corresponding entity parts
    const Index insideLCellIndex;
    //! Index of refinement part on inside
    const Index insidePartIndex;
    //! The local index of this face in the reference element of the
    //! part for which it was created
    const Index indexInPart;
    //! Geometry of this face
    const FaceGeometry geometry;
    //! Global coordinate of outer unit normal
    const Coordinate unitOuterNormal;
    //! Area of this face
    const DomainField area;
    //! Indicates that this face is not curved
    bool linear;

  private:
    //! Mirror faces
    DomainCellFacePointers mirrorFaces;
  };

}
#endif
