// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_UDG_DOMAINDEFINITION_HH
#define DUNE_UDG_DOMAINDEFINITION_HH

#include "levelsetbackends.hh"

namespace Dune {
  namespace UDG {

    /** \brief Domain definitions via level set functions for given
        grid views.

        This class represents a partition of a domain given by the
        grid view of type GV into sub domains defined by level set
        functions in nodes of a grid view of type LGV which has to be
        a refinement of the former grid view.

        \tparam GV      Grid view on computational domain
        \tparam LGV     Grid view for interpolation of level set functions.
     */
    template < typename GV, typename LGV = typename GV::Grid::LeafGridView>
    class DomainDefinition
    {
    public:
      //! Grid view
      typedef GV GridView;
      //! Level set grid view
      typedef LGV LevelSetGridView;
      //! Index type
      typedef typename GV::IndexSet::IndexType Index;
      //! Coordinate field type
      typedef typename GV::ctype DomainField;
      //! Dimension of grid domain
      enum { dim = GV::dimension };
      //! Random access container for the level set values.
      typedef std::vector<DomainField> LevelSetValues;
      //! Space coordinate.
      typedef Dune::FieldVector<DomainField, dim> Coordinate;
      //! Criteria defining whether the inside of a domain is located
      //! relative to the sign of the level set function.
      enum InsideDomainCriteria { whereLevelSetFunctionIsNegative = 0,
                                  whereLevelSetFunctionIsPositive = 1,
                                  whereLevelSetFunctionIsPositiveOrNegative = 2};
      //! Class for a unified and normalized comparison of coordinates.
      typedef MarchingCube33SubTriangulationImp::
      NormalizedCoordinateComparator<Coordinate> CoordinateComparator;

      //! Flag for template switch.
      static const bool isMarchingCube33 = true;

      /** \brief Tag holding domain information relative to interfaces
          separating the domains.

          Indicates whether a domain is on the positive or negative side
          of an interface. Semantics: interfaceTag[<interface index>] =
          [+1,-1]
       */
      typedef std::vector<int> DomainTag;

      /** \brief Options indicating which level set function backend to use

          unique : Backend guarantees unique values in part corners

          nonunique : Function must guarantee (sufficiently) unique
          values in part corners

          ugf : Level set values are interpolated from unfitted
          geometry function
       */
      enum LevelSetFunctionBackendOption { unique, nonunique, ugf };
      //! Backend for the access to the level set function
      typedef LevelSetBackendInterface<DomainDefinition>
      LevelSetBackend;
      //! Backend for the access to the level set function
      typedef LevelSetBackendNodeStorageInterface<DomainDefinition>
      LevelSetBackendNodeStorage;

    private:
      // Entity types
      typedef typename LGV::template Codim<0>::Entity LCellEntity;
      typedef typename GV::template Codim<0>::Entity GVCellEntity;

      // Data structure collecting information about an interface
      // separating domains
      struct Interface {
        Interface() : index(invalidIndex()) {}

        // Unique index of this interface
        Index index;

        // Backend to access the values of the level set function
        Dune::shared_ptr<LevelSetBackend> levelSetBackend;

        // Bit vector indicating the placement of the cell relative to
        // the interface:
        // signInfo[cellIndex*2] == true :
        //      If cell contains nodes in lgv with negative value
        // signInfo[cellIndex*2+1] == true :
        //      If cell contains nodes in lgv with positive value
        std::vector<bool> signInfo;

        // Layer of this interface
        Index layer;
      };

      // Data structure collecting information about a computational
      // sub domain
      struct Domain {

        Domain()
          : index(invalidIndex())
        {}

        // Unique index of this domain
        Index index;
        // Indicates whether this domain is on the positive or negative
        // side of an interface (or on both). Semantics:
        // interfaceTag[<interface index>] = [+1,0,-1]
        DomainTag tag;
        // Layer of this domain
        Index layer;
      };



    public:

      //! Constructor
      DomainDefinition(const GV & gv, const LGV & lgv) :
        gv_(gv), lgv_(lgv), numberOfLayers_(0) {}

      //! Constructor
      DomainDefinition(const GV & gv) :
        gv_(gv), lgv_(gv.grid().leafGridView()), numberOfLayers_(0) {}


      /** \brief Define interface via a given level set function

          Notice: If in the constructor _nInterfaces was set to N, then
          this function MUST be called fo all interfaces with index in
          [0,N-1].

          Notice: This function may be called multiple times in case the
          interface changes during the computation.

          \param index          Index of interface (must start with zero)
          \param lsf            Level set function (must fulfill
                                Dune::PDELab::GridFunction interface).
          \param layer          The layer of this interface
          \param backendOpt     The backend to use for the evaluation
                                 of the level set function values
       */
      template< typename LevelSetFunction >
      void addInterface
        (const Index index,
        const LevelSetFunction & lsf,
        const Index layer,
        const LevelSetFunctionBackendOption backendOpt = unique
        )
      {
        // Setup interface object
        if(interfaces_.size() <= index)
          interfaces_.resize(index+1);
        interfaces_[index].index = index;
        interfaces_[index].layer = layer;

        // Choose level set function backend
        LevelSetBackend* levelSetBackend;
        if(backendOpt == unique) {
          typedef Dune::UDG::UniqueLevelSetBackend
          <DomainDefinition,LevelSetFunction>
          UniqueLevelSetBackend;
          levelSetBackend = new UniqueLevelSetBackend(*this,lsf,index);
        }
        else if(backendOpt == nonunique) {
          typedef Dune::UDG::DefaultLevelSetBackend
          <DomainDefinition,LevelSetFunction>
          DefaultLevelSetBackend;
          levelSetBackend = new DefaultLevelSetBackend(*this,lsf,index);
        }
        else if(backendOpt == ugf) {
          typedef Dune::UDG::UnfittedGeometryFunctionLevelSetBackend
          <DomainDefinition,LevelSetFunction>
          UGFLevelSetBackend;
          levelSetBackend = new UGFLevelSetBackend(*this,lsf,index);
        }
        else{DUNE_THROW(Dune::NotImplemented,"Wrong ls backend!");}

        // Initialize new level set backend
        levelSetBackend->reset(interfaces_[index].signInfo);
        // Set new level set backend
        interfaces_[index].levelSetBackend.reset(levelSetBackend);

        // Update maximum number of layers
        numberOfLayers_ = std::max(layer+1,numberOfLayers_);

        // Increment domain tag size for domains
        for(Index d=0; d<domains_.size(); ++d)
          domains_[d].tag.resize(interfaces_.size(),0);
      }


      /** \brief Define domain relative to a given interface

          Notice: If interfaces_.size() == N, then for each domain, this
          function has to be called N times, relative to the interfaces in
          [0,N-1].

          \param domainIndex    Index of the domain to be defined
          \param itfIndex       Index of interface relative to which the
                                domain is to be defined
          \param criteria       Criteria of type InsideDomainCriteria indicating
                                whether this domain is defined where the level
                                set function is positive or negative.
          \param layer          The layer to which to add this domain.
       */
      void addDomain
        (const Index domainIndex, const Index itfIndex,
        const InsideDomainCriteria criteria,
        const Index layer)
      {
        // Create domain object
        if(domainIndex <= domains_.size())
          domains_.resize(domainIndex+1);

        // Set domain index
        Domain & domain = domains_[domainIndex];
        domain.index = domainIndex;
        domain.layer = layer;

        // Enforce validity of interface index
        if(interfaces_.size() <= itfIndex)
          interfaces_.resize(itfIndex+1);

        // Update maximum number of layers
        numberOfLayers_ = std::max(layer+1,numberOfLayers_);

        // Enforce domain tag size to match number of interfaces.
        for(Index d=0; d<domains_.size(); ++d)
          domains_[d].tag.resize(interfaces_.size(),0);

        // Set interface side criteria
        if(criteria == whereLevelSetFunctionIsPositive)
          domain.tag[itfIndex] = 1;
        else if(criteria == whereLevelSetFunctionIsNegative)
          domain.tag[itfIndex] = -1;
        else if(criteria == whereLevelSetFunctionIsPositiveOrNegative)
          domain.tag[itfIndex] = 0;
        else
        {DUNE_THROW(Dune::NotImplemented,"Invalid InsideDomainCriteria");}
      }

      /** \brief Define domain

          \param domainIndex    Index of the domain to be defined
          \param layer          The layer to which to add this domain.
       */
      void addDomain
        (const Index domainIndex,
        const Index layer)
      {
        // Create domain object
        if(domainIndex <= domains_.size())
          domains_.resize(domainIndex+1);

        // Set domain index
        Domain & domain = domains_[domainIndex];
        domain.index = domainIndex;
        domain.layer = layer;

        // Update maximum number of layers
        numberOfLayers_ = std::max(layer+1,numberOfLayers_);

        // Enforce domain tag size to match number of interfaces.
        for(Index d=0; d<domains_.size(); ++d)
          domains_[d].tag.resize(interfaces_.size(),0);
      }

      /** \brief Access the fundamental mesh cell which contains
          the given cell in the level set mesh grid view. */
      GVCellEntity gvEntity(const LCellEntity & lCell) const
      {
        GVCellEntity ep(lCell);
        while(!gv_.indexSet().contains(ep))
          ep = ep.father();
        return ep;
      }

      /** \brief Checks whether the given domain tag corresponds to
         some existing domain.

         \param[in] testtag The domain tag to be tested is allowed to
         have values in [-1,0,1], where zero is a wildcard equalling
         both 1 and -1. The tag should be zero for all interfaces which
         do not belong to the given layer.

         \return The domain's domain index or the invalid index. If
         wildcards are used the domain corresponding to the given tag
         may not be unique and the eligible domain with the lowest
         domain index is returned.
       */
      Index domainTag2Index(const DomainTag & testtag, const Index layer) const
      {
        for(Index d=0; d<domains_.size(); ++d) {

          if(domains_[d].layer != layer) continue;

          DomainTag tag = domains_[d].tag;

          bool inside = true;
          for(Index i=0; i<interfaces_.size(); ++i) {
            inside = inside &&
                     (testtag[i]==0 || tag[i]==0 || testtag[i] == tag[i]);
          }

          if(inside) return d;
        }
        return invalidIndex();
      }

      /** \brief Query domain tag for given domain index */
      DomainTag domainIndex2Tag(const Index domainIndex){
        return domains_[domainIndex].tag;
      }

      /** \brief Compute domain tag for an interface between the two
          given indices. */
      DomainTag domainIndex2Tag
        (const Index insideIndex, const Index outsideIndex)
      {
        DomainTag insideTag  = domainIndex2Tag(insideIndex);
        DomainTag outsideTag = domainIndex2Tag(outsideIndex);
        DomainTag tag(insideTag.size());
        for(Index i=0; i<tag.size(); ++i)
          tag[i] = (insideTag[i]+outsideTag[i])/2;
        return tag;
      }

      /** \brief Finish definition of domains.

          This function should be called after all domains and interfaces have been defined.

          \return True, if the definitions are consistent and non-conflicting.
       */
      bool finalize()
      {
        for(Index d=0; d<domains_.size(); ++d) {
          if(domains_[d].index == invalidIndex())
            return false;
        } // d

        for(Index itf=0; itf<interfaces_.size(); ++itf) {
          if(interfaces_[itf].index == invalidIndex())
            return false;
        }

        for(Index itf=1; itf<interfaces_.size(); ++itf)
          if(interfaces_[itf-1].layer > interfaces_[itf].layer)
            return false;

        return true;
      }


      /** \brief Determine sign of level set function of the given
          interface index on the given cell in the domain grid view.

          \return Returns +1 if the level set function is positive in
          all nodes of all cells in the level set grid view which are
          contained by the cell with the given index in the domain
          grid view and vice versa. Returns 0 if it contains nodes of
          both signs.
       */
      int signOfLevelSetFunction
        (const Index itfIndex, const Index cellIndex) const
      {
        int sign = 0;
        sign -= int(interfaces_[itfIndex].signInfo[2*cellIndex]);
        sign += int(interfaces_[itfIndex].signInfo[2*cellIndex+1]);
        return sign;
      }

      /** \brief Return the number of interfaces intersecting the
          given cell in the domain grid view.
       */
      const Index numberOfInterfacesInCell
        (const Index cellIndex, const Index layer) const
      {
        Index nInterfaces(0);
        for(Index i=0; i<interfaces_.size(); ++i) {

          if(layer < interfaces_[i].layer) break;

          // Rhs maps {-1,0,1} -> {0,1,0}
          nInterfaces += ( signOfLevelSetFunction(i,cellIndex) + 3 ) % 2;
        }

        return nInterfaces;
      }

      //! Access to grid view
      const GV & gridView() const { return gv_; }
      //! Access to level set grid view
      const LGV & levelSetGridView() const { return lgv_; }

      //! Access to level set function backend
      const LevelSetBackend & levelSetFunctionBackend(const Index itfIndex) const
      { return *(interfaces_[itfIndex].levelSetBackend); }
      //! Access to level set function backend
      LevelSetBackend & levelSetFunctionBackend(const Index itfIndex)
      { return *(interfaces_[itfIndex].levelSetBackend); }

      //! Number of interfaces
      const Index numberOfInterfaces() const { return interfaces_.size(); }
      //! Number of domains
      const Index numberOfDomains() const { return domains_.size(); }
      //! Number of domains
      const Index numberOfLayers() const { return numberOfLayers_; }

      //! Access the coordinate comparator
      const CoordinateComparator & coordinateComparator() const
      { return ncComp_; }
      //! Signal value for an invalid index
      static Index invalidIndex(){ return std::numeric_limits<Index>::max(); }
      //! Layer of the interface with given index
      Index interfaceLayer(const Index itfIndex) const
      { return interfaces_[itfIndex].layer; }
      Index domainLayer(const Index domainIndex) const
      { return domains_[domainIndex].layer; }

      //! True if the given interface starts a new layer
      bool inNewLayer(const Index itfIndex) const
      {
        if(itfIndex == 0 && interfaces_[itfIndex].layer > 0)
          return true;
        else if(itfIndex>0 && interfaces_[itfIndex].layer > interfaces_[itfIndex-1].layer)
          return true;
        else return false;
      }

      // Invalid integer value
      static int invalidInt(){ return std::numeric_limits<int>::max(); }
      // Invalid level set value
      static DomainField invalidValue(){ return std::numeric_limits<DomainField>::max(); }

    private:

      // Grid view for finite element grid
      const GV gv_;
      // Grid view for interface reconstruction
      const LGV lgv_;
      // Domain information array indexed by layers
      std::vector<Domain> domains_;
      // Interface informtion
      std::vector<Interface> interfaces_;
      // Coordinate comparator
      CoordinateComparator ncComp_;
      // Number of layers
      Index numberOfLayers_;
    };

  }
}

#endif
