// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_UDG_DOMAINREFINEMENT_HH
#define DUNE_UDG_DOMAINREFINEMENT_HH

#include <list>

#if DUNE_VERSION_NEWER(DUNE_GEOMETRY,2,3)
#include <dune/geometry/multilineargeometry.hh>
#else
#include <dune/geometry/genericgeometry/geometry.hh>
#endif

#include <dune/tpmc/tpmc.hh>

#include <dune/udg/mc33triangulation/mc33triangulationimp.hh>
#include <dune/udg/common/geometryinterface.hh>


namespace Dune
{
  namespace UDG
  {

    struct DomainRefinementImp
    {
      //! Maximum number of allowed refinement parts per refinement
      enum { maxPartsInLayer = 128 };
    };

    /**
     * \brief This class represents a refinement part obtained by the
     *        marching cubes method plus domain information.
     */
    template< typename Index, typename DomainField, int dim >
    class DomainRefinementPart
    {
    public:
#if DUNE_VERSION_NEWER(DUNE_GEOMETRY,2,3)
      typedef Dune::CachedMultiLinearGeometry<DomainField, dim, dim> Geometry;
#else
      //! Type of geometry object representing the refinement part
      typedef Dune::GenericGeometry::BasicGeometry<
        dim,
        Dune::GenericGeometry::DefaultGeometryTraits<DomainField,dim,dim>
        > Geometry;
#endif
      //! Type of a pointer to geometry object representing the refinement part
      typedef Dune::shared_ptr<const Geometry> GeometryPointer;

      //! Type of space coordinates
      typedef Dune::FieldVector<DomainField,dim> Coordinate;
      //! Type of a vector of space coordinates
      typedef std::vector<Coordinate> Coordinates;

      //! Type of a vector storing the level set functions' values in the
      //! corners of the refinement part
      typedef std::vector<std::vector<DomainField> > CornerValues;
      //! Type of a pointer to a vector storing the level set functions'
      //! values in the corners of the refinement part
      typedef Dune::shared_ptr< CornerValues > CornerValuesPointer;

      //! Type of tag holding the refinement part's domain information
      //! with respect to the interfaces separating the domains
      //! \todo Rename type to InterfaceTag?
      typedef std::vector<int> DomainTag;

      //! Constructor
      DomainRefinementPart
      (
        const Geometry & _geometry,
        const Index & _index,
        const DomainTag & _domainTag,
        const GeometryPointer & _embedding,
        const CornerValuesPointer & _embeddingValues
      )
        : geometry(_geometry),
          index(_index),
          embedding(_embedding),
          embeddingValues(_embeddingValues),
          domainTag(_domainTag)
      {}

    public:
      //! Geometry object representing the refinement part
      Geometry geometry;

      //! The refinement part's unique index within level set mesh home cell
      Index index;

      //! Pointer to geometry object representing the refinement part via its
      //! embedding in the level set mesh home cell (this pointer should not
      //! be accessed outside of DomainRefinement)
      GeometryPointer embedding;

      //! Corner values of embedding (this pointer should not be
      //! accessed outside of DomainRefinement)
      //! \todo Please doc me verbose!
      CornerValuesPointer embeddingValues;

      /**
       * \brief Tag holding the refinement part's domain information
       *        with respect to the interfaces separating the domains.
       *
       * Indicates for each interface (represented by its index) whether
       * it intersects the refinement part or the refinement part lies
       * on the positive/negative side of the level set function which
       * represents the interface. Semantics:
       *
       *   domain-/interfaceTag[<interface index>] \in \{+1,0,-1\}
       *
       * where +1/-1 represents the positive/negative side and 0 indicates
       * an intersection.
       *
       * \todo Rename to interfaceTag?
       */
      DomainTag domainTag;
    };


    /**
     * \todo Please doc me!
     *
     * \tparam DD Type of a domain definition.
     */
    template< typename DD >
    class DomainRefinement
    {
    private:
      //! Short cut to type of level set mesh grid view
      typedef typename DD::LevelSetGridView LGV;

    public:
      //! Type of domain definition
      typedef DD DomainDefinition;
      //! Index type
      typedef typename LGV::IndexSet::IndexType Index;
      //! Coordinate field type
      typedef typename LGV::ctype DomainField;
      //! Dimension of grid domain
      enum { dim = LGV::dimension };
      //! A part of the domain refinement
      typedef Dune::UDG::DomainRefinementPart<Index,DomainField,dim>
      DomainRefinementPart;
      //! Domain tag
      typedef typename DomainRefinementPart::DomainTag DomainTag;
      //! The refinement parts with domain information
      typedef std::list< DomainRefinementPart > DomainPartList;

    private:
      // Domain coordinate
      typedef Dune::FieldVector<DomainField,dim> Coordinate;
      // Vector of coordinates
      typedef std::vector<Coordinate> Coordinates;
      // Cell entity type of grid view
      typedef typename LGV::Traits::template Codim<0>::Entity CellEntity;

      // Codim 0 geometry
      typedef typename DomainRefinementPart::Geometry Geometry;
      // The refinement type of dune-mc
      typedef Dune::TpmcRefinement<DomainField,dim> Refinement;

      // Random access container for the level set values in grid view nodes
      typedef std::vector<DomainField> LevelSetValues;

      // Container for cell parition
      typedef Dune::shared_ptr<const Geometry> GeometryPointer;
      typedef std::list< std::pair< GeometryPointer, std::vector<int> > > CellPartition;

      // Create the local refinement of the part given by its geometry
      // (within its lgv_ father cell) for the given values of the
      // intersection with the given index.
      Index createPartRefinement
        (const Index interfaceIndex,
        const CellEntity & lCell,
        const LevelSetValues & partValues,
        const DomainRefinementPart & part,
        DomainPartList & parts) const
      {
        Index partIndex = part.index;

        // Set the minimum edge length (relative) for this interface. It
        // is chosen such that all interfaces can actually achieve this
        // length.
        const DomainField minEdge =
          minEdgeLengthEps_ * std::pow(DomainField(0.5),DomainField(interfaceIndex));

        // Create a mapping for this part to the reference domain
        // for numerical coordinate comparison.
        const Geometry & normalizedReference =
          domainDef_.coordinateComparator().referenceMap();
        const bool forward = false;
        using namespace MarchingCube33SubTriangulationImp;
        const Geometry globalPartGeometry =
          combineGeometries(part.geometry,lCell.geometry());
        const Geometry normalizedPartGeometry =
          combineGeometries(globalPartGeometry,normalizedReference,forward);

        // Create threshold functor for geometry object
        typedef MarchingCubes::NonDegeneratingThresholdFunctor
        <DomainField,Geometry> ThresholdFunctor;
        ThresholdFunctor threshFunc(normalizedPartGeometry,minEdge);

        // Create refinement object
        typedef TpmcRefinement<DomainField, dim, ThresholdFunctor> Refinement;
        Refinement refinement(partValues.begin(), partValues.end(), threshFunc);

        // Compute refinement for both sides
        for(int s=0; s<2; ++s) {
          // Set domain tag according to current side
          DomainTag domainTag( part.domainTag );
          domainTag[interfaceIndex] = s*2-1;

          // Do not create parts if they are not required by any
          // domain
          if(domainDef_.domainTag2Index(domainTag,layer_) == domainDef_.invalidIndex())
            continue;

          tpmc::ReconstructionType rt = s ? tpmc::ReconstructionType::ExteriorDomain
                                          : tpmc::ReconstructionType::InteriorDomain;
          // Add refinements
          for(const auto& rp : refinement.volume(rt)) {

            // Map part geometry to cell geometry (if already
            // recursive)
            Geometry newPartGeometry =
              combineGeometries<Geometry>(rp,part.geometry);
            // TODO Check this when we have cubes again

            // Add to list of new parts
            const DomainRefinementPart & candidate =
              DomainRefinementPart
                (newPartGeometry,partIndex,domainTag,part.embedding,part.embeddingValues);

            // Increment local running index
            partIndex++;

            parts.push_back(candidate);
          }

        }

        // Return the part index in the current layer
        return partIndex;
      }



      // Create refinement for the given entity parts and the given
      // interface index. If the a refinement with minimal edge length
      // bigger than minEdge could not be created, this function
      // returns false.
      void createPartsRefinement
        (const CellEntity & lCell,
        DomainPartList & parts) const
      {
        assert(parts.size()==1);

        // Iterate all interfaces via their indices
        for(Index i=0; i<domainDef_.numberOfInterfaces(); ++i) {

          // Only refine up to the given layer
          if(domainDef_.interfaceLayer(i) > layer_) break;

          // List for the newly created parts and groups
          DomainPartList newParts;
          typedef typename DomainPartList::iterator Iterator;

          // Shift bits of part indices if interface in new layer
          if( domainDef_.inNewLayer(i) ) {

            for(Iterator pit = parts.begin(); pit != parts.end(); ++pit) {
              // Init part index for children of this part
              assert((std::numeric_limits<Index>::max()-DomainRefinementImp::maxPartsInLayer)
                     /DomainRefinementImp::maxPartsInLayer>pit->index);
              pit->index *= DomainRefinementImp::maxPartsInLayer;
            }
          }

          // Linearize all parts if requested
          if(linearizeParts_)
            linearizeParts(parts);

          // Iterate parts
          Index partIndex(0);
          for(Iterator pit = parts.begin(); pit != parts.end(); ++pit) {

            partIndex = std::max(partIndex,pit->index);
            pit->index = partIndex;

            // // Set part index (only in current layer)
            // pit->index = (pit->index/DomainRefinementImp::maxPartsInLayer)
            //   *DomainRefinementImp::maxPartsInLayer+partIndex;

            // Do not refine part if its domain tag indicates that there
            // is no intersection with the given interface.
            if(pit->domainTag[i]!=0) {
              partIndex++;
              newParts.push_back(*pit);
              continue;
            }

            // Compute values in part vertices
            LevelSetValues partValues;
            using namespace MarchingCube33SubTriangulationImp;
            if(i == 0)
              partValues = (*(pit->embeddingValues))[i];
            else
              domainDef_.levelSetFunctionBackend(i).partCornerValues(lCell,*pit,partValues);

            // Create the local refinement of pit
            partIndex = createPartRefinement(i,lCell,partValues,*pit,newParts);

          } // pit

          // Set the new parts to be the old ones (realize recursion)
          parts.swap(newParts);

        } // interface index
      }

      Geometry createGeometry
        (const Dune::GeometryType & type) const
      {
        // Reference element of cell to refine
        auto ref =
          Dune::referenceElement<DomainField,dim>(type);

        // Setup array of local vertex coordinates and indices for
        // reference element
        Coordinates refCoords(ref.size(dim));
        for(int v=0; v<ref.size(dim); ++v)
          refCoords[v] = ref.position(v,dim);

        typedef typename DomainRefinementPart::Geometry Geometry;
        const Geometry refGeometry = Geometry(type,refCoords);
        return refGeometry;
      }

      void createLinearPartitionFromCube
        (CellPartition & partition) const
      {
        auto type = Dune::GeometryTypes::cube(dim);

        // Reference element of cell to refine
        auto ref =
          Dune::referenceElement<DomainField,dim>(type);

        // Start and stop node (for all simplices)
        const Coordinate start = ref.position(0,dim);

        // Init permutation
        std::vector<int> perm(dim);
        for(int d=0; d<dim; ++d) perm[d] = (1<<d);

        // Create geometries for all permutations
        auto simplexType = Dune::GeometryTypes::simplex(dim);
        Coordinates nodes(dim+1);
        nodes[0] = start;
        std::vector<int> nodeIndices(dim+1);
        nodeIndices[0]=0;
        do {

          int nodeIndex = 0;
          for(int d=0; d<dim; ++d) {
            nodeIndex += perm[d];
            nodes[d+1] = ref.position(nodeIndex,dim);
            nodeIndices[d+1]=nodeIndex;
          }
          GeometryPointer pGeometry(new Geometry(simplexType,nodes));
          partition.push_back( std::make_pair(pGeometry,nodeIndices) );

        } while(std::next_permutation(perm.begin(),perm.end()));
      }

      void createLinearPartitionFromPrism
        (CellPartition & partition) const
      {
        if(dim == 2) return;

        auto type = Dune::GeometryTypes::prism;
        const int simplices[][4] = { {1,2,4,3}, {2,4,5,3}, {0,1,2,3} };
        createLinearPartitionFromList(partition,type,simplices,3);
      }

      void createLinearPartitionFromPyramid
        (CellPartition & partition) const
      {
        if(dim == 2) return;

        auto type = Dune::GeometryTypes::pyramid;
        const int simplices[][4] = { {0,1,2,4}, {1,2,3,4} };
        createLinearPartitionFromList(partition,type,simplices,2);
      }

      void createLinearPartitionFromList
      (
        CellPartition & partition,
        const Dune::GeometryType & type,
        const int simplices[][4],
        const int nSimplices
      ) const
      {
        // Reference element of cell to refine
        auto ref =
          Dune::referenceElement<DomainField,dim>(type);

        auto simplex_type = Dune::GeometryTypes::simplex(dim);

        const int sNodes = 4;
        for(int s=0; s<nSimplices; ++s) {
          std::vector<int> nodeIndices(simplices[s],simplices[s]+sNodes);
          Coordinates nodes(sNodes);
          for(int c=0; c<sNodes; ++c)
            nodes[c] = ref.position(nodeIndices[c],dim);
          GeometryPointer pGeometry(new Geometry(simplex_type,nodes));
          partition.push_back( std::make_pair(pGeometry,nodeIndices) );
        }
      }

      void createLinearPartitionFromSimplex
        (CellPartition & partition) const
      {
        auto type = Dune::GeometryTypes::simplex(dim);

        // Reference element of cell to refine
        auto ref =
          Dune::referenceElement<DomainField,dim>(type);

        Coordinates nodes(ref.size(dim));
        std::vector<int> nodeIndices(ref.size(dim));
        for(int n=0; n<ref.size(dim); ++n) {
          nodes[n] = ref.position(n,dim);
          nodeIndices[n] = n;
        }
        GeometryPointer pGeometry(new Geometry(type,nodes));
        partition.push_back( std::make_pair(pGeometry,nodeIndices) );
      }

      void getLinearPartition
        (const Dune::GeometryType & type, CellPartition & partition) const
      {
        if(type.isSimplex()) {
          partition = linearSimplexPartition_;
        }
        // Compute disassembly into simplices
        else if(type.isCube()) {
          partition = linearCubePartition_;
        }
        else if(type.isPrism()) {
          partition = linearPrismPartition_;
        }
        else if(type.isPyramid()) {
          partition = linearPyramidPartition_;
        }
        else{DUNE_THROW(Dune::NotImplemented,"Not Implemented for this geom type");}
      }

      void linearizeParts(DomainPartList & parts) const
      {
        // Determine minimal index of all parts in list
        Index partIndex(0);
        // typedef typename DomainPartList::iterator Iterator;
        // for(Iterator pit = parts.begin(); pit != parts.end(); ++pit)
        //   partIndex = std::min(partIndex,pit->index);

        // Iterate all parts
        DomainPartList newParts;
        typedef typename DomainPartList::iterator Iterator;
        for(Iterator pit = parts.begin(); pit != parts.end(); ++pit) {

          partIndex = std::max(pit->index,partIndex);
          pit->index = partIndex;

          // Just copy linear parts. Make sure that partIndex is increased
          // so that the next part gets a new unique index
          if(pit->geometry.type().isSimplex()) {
            newParts.push_back(*pit);
            ++partIndex;
            continue;
          }

          // Query for partition
          CellPartition partition;
          getLinearPartition(pit->geometry.type(),partition);
          assert(linearCubePartition_.size() > 0);

          typedef typename CellPartition::const_iterator GeometryIt;
          typedef typename DomainRefinementPart::CornerValuesPointer CornerValuesPointer;
          typedef typename CornerValuesPointer::element_type CornerValues;

          // Compute the values of the part via bi-linear interpolation
          const Geometry localInEmbedding =
            combineGeometries(pit->geometry,*(pit->embedding),false);

          const Index nInterfaces = domainDef_.numberOfInterfaces();
          CornerValues partValues(nInterfaces);
          for(Index i=0; i<nInterfaces; ++i) {
            partValues[i].resize(pit->geometry.corners());
            const CornerValues & embeddingValues = (*(pit->embeddingValues));

            using namespace MarchingCube33SubTriangulationImp;
            computeValuesInVertices
              (pit->embedding->type(),embeddingValues[i],localInEmbedding,partValues[i]);
          }

          // Iterate all partition elements
          for(GeometryIt it = partition.begin(); it != partition.end(); ++it) {
            CornerValuesPointer nodeValues(new CornerValues(nInterfaces));

            // Iterate all interfaces
            for(Index i=0; i<nInterfaces; ++i) {

              // Iterate all nodes of partition element
              const std::vector<int> & nodeIndices = it->second;
              (*nodeValues)[i].resize(nodeIndices.size());
              for(Index n=0; n<nodeIndices.size(); ++n) {
                const Index nodeIndex = nodeIndices[n];
                (*nodeValues)[i][n] = partValues[i][nodeIndex];
              }
            }

            // Store new part
            GeometryPointer linGeometryRef = it->first;
            const Geometry linGeometry = combineGeometries(*linGeometryRef,pit->geometry);
            const GeometryPointer embedding(new Geometry(linGeometry));

            const DomainTag & tag = pit->domainTag;
            newParts.push_back
              (DomainRefinementPart(linGeometry,partIndex,tag,embedding,nodeValues));

            // Increment local running index
            partIndex++;

          } // it

        }
        // Exchange new and old list
        parts.swap(newParts);
      }

    public:
      /**
       * \brief Constructor.
       *
       * \todo Please doc me!
       *
       * \param domainDef        Domain definition providing information related
       *                         to the interfaces which separate the domains.
       * \param layer            Index of the layer up to which shall be refined.
       * \param minEdgeLengthEps Numerical epsilon which yields a minimum edge
       *                         length during domain refinement.
       * \param linearizeParts   Switch for domain refinement procedure.
       *                         While iterating the interfaces which separate
       *                         the domains, decomposition of the current set
       *                         of domain refinement parts into simplices
       *                         (-> linearization) before refinement using
       *                         the next interface can be deactivated.
       *                         This results in less domain refinement parts
       *                         and we dont have a geometrical bias in our
       *                         geometry reconstruction.
       *                         Setting linearizeParts = false shall be used
       *                         with care as it is not yet tested very well.
       *
       * \todo Comment on the interpretation of minEdgeLengthEps value.
       */
      DomainRefinement(const DD & domainDef, const Index layer,
                       const DomainField & minEdgeLengthEps,
                       const bool linearizeParts)
        : domainDef_(domainDef), lgv_(domainDef.levelSetGridView()),
          layer_(layer),
          minEdgeLengthEps_(minEdgeLengthEps), linearizeParts_(linearizeParts)
      {
        if(linearizeParts_ && linearCubePartition_.size() == 0)
          createLinearPartitionFromCube(linearCubePartition_);
        if(linearizeParts_ && linearSimplexPartition_.size() == 0)
          createLinearPartitionFromSimplex(linearSimplexPartition_);
        if(linearizeParts_ && linearPrismPartition_.size() == 0)
          createLinearPartitionFromPrism(linearPrismPartition_);
        if(linearizeParts_ && linearPyramidPartition_.size() == 0)
          createLinearPartitionFromPyramid(linearPyramidPartition_);
      }

      /**
       * \brief Creates the actual domain refinement parts and the domain tag
       *        for a given level set mesh cell.
       *
       * \todo Please doc my parameters!
       */
      void createDomainRefinement
        (const CellEntity & lCell, DomainPartList & parts, DomainTag & domainTag) const
      {
        parts.clear();

        // Query the values in the nodes of this cell for each
        // interface and compute (partial) domain tag.
        typedef typename DomainRefinementPart::CornerValuesPointer CornerValuesPointer;
        typedef typename CornerValuesPointer::element_type CornerValues;
        CornerValuesPointer values(new CornerValues(domainDef_.numberOfInterfaces()));
        bool isIntersected = false;

        domainTag.clear();
        domainTag.resize(domainDef_.numberOfInterfaces(),0);
        for(Index i=0; i<domainTag.size(); ++i) {

          const int sign = domainDef_.levelSetFunctionBackend(i).cellCornerValues(lCell,(*values)[i]);
          const bool pureSign = bool(sign);
          isIntersected |= !pureSign;

          // Write domain tag entry
          if(domainDef_.interfaceLayer(i) <= layer_)
            domainTag[i] = sign;
        }

        // If the (partial) domain tag can not be associated to any
        // domain, then nothing is to be done
        if(domainDef_.domainTag2Index(domainTag,layer_) == domainDef_.invalidIndex())
          return;

        // Create initial part for quasi-recursion
        const Geometry cellGeometry = createGeometry(lCell.geometry().type());
        const GeometryPointer embedding = Dune::stackobject_to_shared_ptr(cellGeometry);
        const Index partIndex(0);
        parts.push_back(DomainRefinementPart(cellGeometry,partIndex,domainTag,embedding,values));

        // Create sub triangulation if intersected
        if(isIntersected) createPartsRefinement(lCell, parts);

        // Release memory of embedding objects
        typedef typename DomainPartList::iterator Iterator;
        for(Iterator pit = parts.begin(); pit != parts.end(); ++pit) {
          pit->embedding.reset();
          pit->embeddingValues.reset();
        }
      }

    private:
      //! Domain definition
      const DomainDefinition & domainDef_;

      //! Grid view for interface reconstruction, i.e. the level set mesh
      const LGV & lgv_;

      //! Index of the layer up to which shall be refined
      const Index layer_;

      //! Numerical epsilon which yields a minimum edge length during
      //! domain refinement
      //! \todo Comment on the interpretation of its value
      const DomainField minEdgeLengthEps_;

      //! Switch for iterative domain refinement procedure which triggers
      //! linearization before refinement using the next interface
      const bool linearizeParts_;

      //! Static members used for linearization by each instance
      static CellPartition linearCubePartition_;
      static CellPartition linearSimplexPartition_;
      static CellPartition linearPrismPartition_;
      static CellPartition linearPyramidPartition_;
    };

    // Definition of static member
    template<class DD>
    typename DomainRefinement<DD>::CellPartition
    DomainRefinement<DD>::linearCubePartition_;

    // Definition of static member
    template<class DD>
    typename DomainRefinement<DD>::CellPartition
    DomainRefinement<DD>::linearSimplexPartition_;

    // Definition of static member
    template<class DD>
    typename DomainRefinement<DD>::CellPartition
    DomainRefinement<DD>::linearPrismPartition_;

    // Definition of static member
    template<class DD>
    typename DomainRefinement<DD>::CellPartition
    DomainRefinement<DD>::linearPyramidPartition_;

  }
}

#endif
