// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_UDG_MC33_LEVELSETBACKEND_HH
#define DUNE_UDG_MC33_LEVELSETBACKEND_HH
#include <dune/udg/mc33triangulation/domainrefinement.hh>

namespace Dune {
  namespace UDG {

    template< typename DomainDefinition >
    struct LevelSetBackendTraits
    {
      //! Grid view
      typedef typename DomainDefinition::GridView GridView;
      //! Level set grid view
      typedef typename DomainDefinition::LevelSetGridView LevelSetGridView;
      //! Random access container for the level set values.
      typedef typename DomainDefinition::Coordinate Coordinate;
      //! Field value of domain
      typedef typename Coordinate::value_type DomainField;
      //! Level set value vector
      typedef std::vector<DomainField> LevelSetValues;
      //! Normalized global comparison of coordinates
      typedef typename DomainDefinition::CoordinateComparator CoordinateComparator;
      //! Dimension of domain
      enum { dim = Coordinate::dimension };
      //! Geometric grid view cell entity
      typedef typename LevelSetGridView::template Codim<0>::Entity CellEntity;
      //! Geometric grid view cell entity
      typedef typename GridView::template Codim<0>::Entity GVCellEntity;
      //! Unsigned index type
      typedef typename DomainDefinition::Index Index;
      //! Refinement part
      typedef DomainRefinementPart<Index,DomainField,dim> RefinementPart;
    };

    template< typename DomainDefinition >
    class LevelSetBackendInterface
    {
    public:
      typedef LevelSetBackendTraits<DomainDefinition> Traits;

      //! Clear all cached values in the backend
      virtual void reset(std::vector<bool> & signInfo) = 0;

      //! Print some statistics (if available)
      virtual void printStatistics() const = 0;

      //! Evaluate values in corners of given cell
      virtual int cellCornerValues
        (const typename Traits::CellEntity & lCell,
        typename Traits::LevelSetValues & values) const = 0;

      //! Evaluate values in corners of given refinement part
      virtual int partCornerValues
      (
        const typename Traits::CellEntity & lCell,
        const typename Traits::RefinementPart & part,
        typename Traits::LevelSetValues & values
      ) const = 0;
    };

    template< typename DomainDefinition >
    class LevelSetBackendNodeStorageInterface :
      public virtual LevelSetBackendInterface<DomainDefinition>
    {
    public:
      typedef LevelSetBackendTraits<DomainDefinition> Traits;

      virtual typename Traits::LevelSetValues &
      levelSetFunctionValues() = 0;

      virtual const typename Traits::LevelSetValues &
      levelSetFunctionValues() const = 0;

    };

    /** \brief An implementation stub for template switches */
    template< typename DomainDefinition>
    class LevelSetBackendNodeStorageStub :
      public virtual LevelSetBackendNodeStorageInterface<DomainDefinition>
    {
    public:
      typedef LevelSetBackendTraits<DomainDefinition> Traits;

      //! Clear all cached values in the backend
      virtual void reset(std::vector<bool> & signInfo){}

      //! Print some statistics (if available)
      virtual void printStatistics() const {}

      //! Evaluate values in corners of given cell
      virtual int cellCornerValues
        (const typename Traits::CellEntity & lCell,
        typename Traits::LevelSetValues & values) const
      { DUNE_THROW(Dune::Exception,"Should not be called!");}

      //! Evaluate values in corners of given refinement part
      virtual int partCornerValues
      (
        const typename Traits::CellEntity & lCell,
        const typename Traits::RefinementPart & part,
        typename Traits::LevelSetValues & values
      ) const
      { DUNE_THROW(Dune::Exception,"Should not be called!");}

      virtual typename Traits::LevelSetValues &
      levelSetFunctionValues()
      { DUNE_THROW(Dune::Exception,"Should not be called!");}

      virtual const typename Traits::LevelSetValues &
      levelSetFunctionValues() const
      { DUNE_THROW(Dune::Exception,"Should not be called!");}
    };

  }
}

#endif
