// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_UDG_MC33_LEVELSETCACHE_HH
#define DUNE_UDG_MC33_LEVELSETCACHE_HH

namespace Dune {
  namespace UDG {


    /** \brief Cache for off-corner level set values */
    template< typename DomainDefinition >
    class LevelSetValueCache
    {
    public:
      //! Grid view
      typedef typename DomainDefinition::GridView GridView;
      //! Level set grid view
      typedef typename DomainDefinition::LevelSetGridView LevelSetGridView;
      //! Random access container for the level set values.
      typedef typename DomainDefinition::LevelSetValues LevelsetValues;
      //! Random access container for the level set values.
      typedef typename DomainDefinition::Coordinate Coordinate;
      //! Field value of domain
      typedef typename Coordinate::value_type DomainField;
      //! Normalized global comparison of coordinates
      typedef typename DomainDefinition::CoordinateComparator CoordinateComparator;
      //! Dimension of domain
      enum { dim = Coordinate::dimension };
      //! Geometric grid view cell entity
      typedef typename LGV::template Codim<0>::Entity CellEntity;
      //! Unsigned index type
      typedef typename DomainDefinition::Index Index;

      struct CacheEntry {
        CacheEntry
          (const Coordinate & coordInRef, const DomainField & value)
          : coordInRef_(coordInRef), value_(value)
            Coordinate coordInRef_;
        DomainField value_;
      };
      typedef std::map<DomainField,CacheEntry> CacheLine;

      class Cache {
      public:
        Cache(const CoordinateComparator & comp)
          : comp_(comp),
            cells(dim-1), nCells(1)
        {
          high = comp_.inRef(comp_.domainMaximum());
          low = comp_.inRef(comp_.domainMinimum());
          length = high - low;

          Coordinate Hg(comp_.domainMaximum()-comp_.domainMinimum());
          DomainField hg(comp_.minDistance());
          for(Index d=0; d<dim-1; ++d) {
            cells[d] = Index(Hg[d] / hg + 1.0);
            nCells *= cells[d];
            h[d] = length[d] / DomainField(cells[d]);
          }
          cachelines.resize(nCells);
        }

        Index flatindex(const std::vector<Index> & index)
        {
          Index i(index[0]);
          for(Index d=1; d<dim-1; ++d)
            i += index[d] * cells[d-1];
          return i;
        }

        bool findInLine
          (const Coordinate & ref, const Index flatindex, DomainField & value)
        {
          const DomainField eps = comp_.numericalEqualityEps();
          const DomainField lowKey = ref[dim-1]-2.0*eps;
          typename CacheLine::iterator lineIt =
            cachelines[flatindex].lower_bound(lowKey);
          while(lineIt != cachelines.end() && lineIt->first < lowKey+2.0*eps) {
            const CacheEntry & entry = lineIt->second;
            if(comp_.isEqualInRef(entry.coordInRef_,ref)) {
              value = entry.value_;
              return true;
            }
            lineIt++;
          }
          return false;
        }

        void addToLine
          (const Coordinate & ref, const Index flatindex, DomainField & value)
        {
          const DomainField key = ref[dim-1];
          CacheEntry entry(ref,value);
          cachelines[flatindex].insert(std::make_pair(key,entry));
        }

        bool find(const Coordinate & global, DomainField & value)
        {
          hit = false;
          const Coordinate ref = comp_.inRef(global);

          Coordinate cIndex;
          std::vector<Index> index(dim-1);
          for(Index d=0; d<dim-1; ++d) {
            cIndex[d] = (ref[d]-low[d])/h[d];
            index[d] = Index(cIndex[d]);
            index[d] = std::max(index[d],0);
            index[d] = std::min(index[d],cells[d]-1);
          }

          { // Check whether coordinate in this line
            const Index flat = flatindex(index);
            if(findInLine(ref,flat,value)) return true;
          }

          // Check whether coordinate in any neighbor line
          for(Index d=0; d<dim-1; ++d) {
            std::vector<Index> nIndex(index);
            if((cIndex[d] - DomainField(index[d])) < 0.5)
              nIndex[d] = std::max(nIndex[d]-1,0);
            else
              nIndex[d] = std::min(nIndex[d]+1,cells[d]-1);

            const Index flat = flatindex(nIndex);
            if(findInLine(ref,flat,value)) return true;
          }

          { // Cache miss! Add to cache
            const Index flat = flatindex(index);
            addToLine(ref,flat,value);
          }
          return false;
        }

        void clear()
        {
          cachelines.clear();
        }

      private:
        CoordinateComparator comp_;
        std::vector<Index> cells;
        Index nCells;
        Coordinate h;
        Coordinate low;
        Coordinate high;
        Coordinate length;
        std::vector<CacheLine> cachelines;
        CacheLine emptyLine;
      };

    private:

      //! Compute local index of node entity corresponding to local
      //! coordinate
      int localCubeIndex
        (const CellEntity & lCell, const Coordinate & local) const
      {
        Coordinate corner;
        int index(0);
        for(int d=0; d<dim; ++d) {
          const int comp = int(local[d] > 0.5);
          corner[d] = DomainField(comp);
          index += (1<<d) * comp;
        }
        corner = lCell.geometry().global(corner);
        const Coordinate global = lCell.geometry().global(local);
        if(domainDef_.coordinateComparator().isEqual(corner,global))
          return index;
        else
          return -1;
      }

      //! Compute local index of node entity corresponding to local
      //! coordinate
      int localSimplexIndex
        (const CellEntity & lCell, const Coordinate & local) const
      {
        Coordinate corner;
        int index(0);
        for(int d=0; d<dim; ++d) {
          const int comp = int(local[d] > 0.5);
          corner[d] = DomainField(comp);
          index += (d+1) * comp;
        }
        corner = lCell.geometry().global(corner);
        const Coordinate global = lCell.geometry().global(local);
        if(domainDef_.coordinateComparator().isEqual(corner,global))
          return index;
        else
          return -1;
      }

      //! Compute local index of node entity corresponding to local
      //! coordinate (works only for cubes).
      int localIndex
        (const CellEntity & lCell, const Coordinate & local) const
      {
        if(lCell.type().isCube()) return localCubeIndex(lCell,local);
        else if(lCell.type().isSimplex()) return localSimplexIndex(lCell,local);
        else { DUNE_THROW(Dune::NotImplemented,"Not implemented for this type"); }
      }


    public:

      //! Constructor
      LevelSetValueCache(const DomainDefinition & domainDef)
        : domainDef_(domainDef) {}

      template< typename RefinementPart >
      int levelSetFunctionValues
        (const Index itfIndex,
        const CellEntity & lCell,
        const RefinementPart & part,
        LevelSetValues & values) const
      {
        // Access ls values in lgv nodes
        const LevelSetValues & lsValues = domainDef_.levelSetFunctionValues(itfIndex);

        // Reference element of lgv cell in which to refine.
        const Dune::ReferenceElement<DomainField,dim> & ref =
          Dune::ReferenceElements<DomainField,dim>::general(part.geometry.type());

        // Flag to indicate an interface
        bool interfaceFound = false;
        DomainField lastValue(0);

        // Values in corners of lCell
        values.resize(ref.size(dim));
        for(Index v=0; v<values.size(); ++v) {

          const Coordinate localInLCell = part.corner(v);
          const int localNodeIndex = localIndex(lCell,localInLCell);
          if(localNodeIndex >= 0) {
            typedef typename LGV::Traits::template Codim<dim>::Entity
            NodeEntity;
            const NodeEntity node = lCell.template subEntity<dim>(localNodeIndex);
            const Index nodeIndex = lgv_.indexSet().index(node);
            values[v] = lsValues[nodeIndex];
          }
          else{
            const Coordinate global = lCell.geometry().global(localInLCell);

          }
          assert(values[v]); // Zero values should have been removed

          if(values[v] == invalidValue()) continue;

          // Check for interface intersecting this cell
          interfaceFound |= ( lastValue * values[v] ) < 0;
          lastValue = values[v];
        }

        if( interfaceFound )
          return 0;
        else
          return int( ((1.0+1e-10)*lastValue)/std::abs(lastValue) );
      }




    private:
      const DomainDefinition & domainDef_;
    };


  }
}

#endif
