// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_MC33TRIANGULATION_CC
#define DUNE_MC33TRIANGULATION_CC

#include <vector>    // provides std::vector
#include <iostream>  // provides std::cout
#include <set>       // provides std::set

#include <dune/common/shared_ptr.hh>    // provides Dune::shared_ptr
#include <dune/common/exceptions.hh>    // provides DUNE_THROW, Dune::Exception

#include <dune/geometry/type.hh>               // provides Dune::GeometryType
#include <dune/geometry/referenceelements.hh>  // provides Dune::ReferenceElement(s)

#include <dune/udg/common/intersectionpart.hh>  // provides Dune::UDG::internal
#include <dune/udg/common/cuboid.hh>            // provides Dune::CuboidMode

#include <dune/udg/mc33triangulation.hh>


namespace Dune
{
  namespace UDG
  {

    template< typename DD >
    void
    MarchingCube33SubTriangulation<DD>::
    createMultiPartsForWholeCell
    (
      const GVCellEntity & cell,
      EntityPartList & entityparts,
      IntersectionPartList & intersectionCache
    ) const
    {
      const Index cellIndex = gv_.indexSet().index(cell);

      // Collect a domain cell lookup item on each layer
      bool outside(false);
      std::vector<const LookupItem*> cellLookupItems(domainDef_.numberOfLayers());
      for(Index layer=0; layer < domainDef_.numberOfLayers(); ++layer) {
        cellLookupItems[layer] = layers_[layer].lookupMap().find(cellIndex);

        if(!cellLookupItems[layer]) outside = true;

        // assert that all layers have the same exterior boundary by ensuring
        // that once outside is set this property is fulfilled on each higher
        // layer
        assert(!outside || !cellLookupItems[layer]);

        // assert(!outside || !cellLookupItems[layer]->insideDomainCell);
        // if(!cellLookupItems[layer]->insideDomainCell)
        //   outside = true;
      }
      // Return if this cell is in no domain
      if(outside) return;


      // Create multi entity part from whole fundamental mesh cell
      using namespace MarchingCube33SubTriangulationImp;
      const Index partIndex = 0;
      Coordinates coordinates = getReferenceCorners(cell.geometry());
      EntityPartPointer ePart ( createMultiEntityPart(cell,partIndex,coordinates,cellLookupItems) );
      entityparts.push_back(ePart);


      if(useIntersections_) {

        // Visit neighbors and create intersection parts
        typename GV::IntersectionIterator iit = gv_.ibegin(cell);
        const typename GV::IntersectionIterator eiit = gv_.iend(cell);
        for(; iit != eiit; ++iit) {

          // Collect the intersection domain cells on all layers
          bool isRealIntersection = false;
          MultiIntersectionDomainCells intersectionDomainCells(cellLookupItems.size());
          for(Index layer=0; layer<cellLookupItems.size(); ++layer) {

            // Set inside domain cell
            const DomainCell* insideDomainCell =
              cellLookupItems[layer]->insideDomainCell;

            intersectionDomainCells.insideDomainCells[layer] =
              insideDomainCell;
            intersectionDomainCells.outsideDomainCells[layer] =
              insideDomainCell;

            // Check for neighbor
            if(!iit->boundary()) {
              GVCellEntity neighbor = iit->outside();
              const Index neighborCellIndex = gv_.indexSet().index(neighbor);

              const typename LookupItem::FacesLookupMap& faceLookupMap =
                cellLookupItems[layer]->facesLookup;

              typename LookupItem::FacesLookupMap::const_iterator faceIt =
                faceLookupMap.find(neighborCellIndex);
              if(faceIt != faceLookupMap.end()) {
                isRealIntersection = true;
                intersectionDomainCells.outsideDomainCells[layer]
                  = faceIt->second;
              }

            }
            // All boundaries are created
            else isRealIntersection = true;

          } // layer

          // Only create intersection parts if it is adjacent to two
          // different domain cells on at least one layer or if it is
          // a boundary of the fundamental mesh.
          if(!isRealIntersection) continue;

          // Get the intersection's corners in global coordinates and its outer unit normal
          Coordinates faceVertices = getGeometryCorners(iit->geometry());
          const Coordinate unitOuterNormal = iit->centerUnitOuterNormal();

          // Create multi intersection part
          IntersectionPart* intersection
            ( createMultiIntersectionPart(intersectionDomainCells,faceVertices,unitOuterNormal) );
          if(intersection) {
            IntersectionPartPointer intersectionPointer(intersection);
            intersectionCache.push_back( intersectionPointer );
          }

        } // iit

      } // useIntersections_

    }


    template< typename DD >
    typename MarchingCube33SubTriangulation<DD>::SubIntersectionPart*
    MarchingCube33SubTriangulation<DD>::
    createIntersectionPart
    (
      const BoundingBox & inBBox,
      const BoundingBox & outBBox,
      const DomainTag domainTag,
      const int insideDomainIndex,
      const int outsideDomainIndex,
      const Coordinates & c,
      const Coordinate & normal,
      const DomainField & area,
      const DomainField & insideVolume,
      const DomainField & outsideVolume
    ) const
    {
      int intersectionTag(0);
      for(int i=0; i<int(domainTag.size()); ++i) {
        const int neg = int(domainTag[i]==-1);
        const int pos = int(domainTag[i]==+1);
        intersectionTag = intersectionTag | (pos << (2*i));
        intersectionTag = intersectionTag | (neg << (2*i+1));
      }

      // Setup intersection geometry
      Dune::GeometryType type = Dune::geometryTypeFromVertexCount(dim-1,c.size());
      typename IntersectionPart::FaceGeometryImp faceGeometry(type,c);
      IntersectionPartGeometry intersectionGeometry
        (faceGeometry,inBBox,outBBox,normal);

      // TODO: not parallel yet
      const int bid = Dune::UDG::internal;
      return new SubIntersectionPart
               (intersectionGeometry,intersectionTag,domainTag,bid,
               insideDomainIndex, outsideDomainIndex,area,insideVolume,outsideVolume);
    }

    template< typename DD >
    typename MarchingCube33SubTriangulation<DD>::SubIntersectionPart*
    MarchingCube33SubTriangulation<DD>::
    createIntersectionPart
    (
      const DomainCell & insideDomainCell,
      const DomainCell & outsideDomainCell,
      const Coordinates & c,
      const Coordinate & normal,
      const bool boundary
    ) const
    {

      const int insideDomainIndex = insideDomainCell.domainIndex;
      const int outsideDomainIndex = boundary ? -1 : outsideDomainCell.domainIndex;

      const DomainTag faceDomainTag = boundary ?
                                      domainDef_.domainIndex2Tag(insideDomainIndex) :
                                      domainDef_.domainIndex2Tag(insideDomainIndex,outsideDomainIndex) ;

      typedef typename DomainCell::IntersectionAreaMap AreaMap;
      typename AreaMap::const_iterator areaIt =
        insideDomainCell.areas.find(outsideDomainCell.id);
#ifndef NDEBUG
      if(areaIt == insideDomainCell.areas.end()) {
        for(typename AreaMap::const_iterator it=insideDomainCell.areas.begin();
            it != insideDomainCell.areas.end();
            ++it)
          std::cout << "Inside area entry " <<it->first << " " << it->second << std::endl;
        for(typename AreaMap::const_iterator it=outsideDomainCell.areas.begin();
            it != outsideDomainCell.areas.end();
            ++it)
          std::cout << "Outside area entry " <<it->first << " " << it->second << std::endl;
        insideDomainCell.outputFaceList();
      }
      assert(areaIt != insideDomainCell.areas.end());
#endif

      const DomainField area = areaIt->second;
      const DomainField insideVolume  = insideDomainCell.volume;
      const DomainField outsideVolume = outsideDomainCell.volume;

      BoundingBox inBBox(gv_.grid().entity(insideDomainCell.hostCell),
                         insideDomainCell.bbox.lower,
                         insideDomainCell.bbox.higher,
                         Dune::CuboidMode::corners);

      BoundingBox outBBox(gv_.grid().entity(outsideDomainCell.hostCell),
                          outsideDomainCell.bbox.lower,
                          outsideDomainCell.bbox.higher,
                          Dune::CuboidMode::corners);

      SubIntersectionPart* intersection =
        createIntersectionPart(inBBox,outBBox, faceDomainTag,
                               insideDomainIndex,outsideDomainIndex,
                               c,normal,
                               area,insideVolume,outsideVolume);

      return intersection;
    }


    template< typename DD >
    typename MarchingCube33SubTriangulation<DD>::IntersectionPart*
    MarchingCube33SubTriangulation<DD>::
    createMultiIntersectionPart
    (
      const MultiIntersectionDomainCells & domainCells,
      const Coordinates & c,
      const Coordinate & normal
    ) const
    {

      const bool isBoundary = domainCells.isBoundary();

      // Allocate multi layer intersection part
      IntersectionPart * multiIntersection =
        new IntersectionPart(domainDef_.numberOfLayers());

      for(Index layer = 0; layer < domainDef_.numberOfLayers(); ++layer) {

        const DomainCell * insideDomainCell = domainCells.insideDomainCells[layer];

        const DomainCell * outsideDomainCell = domainCells.outsideDomainCells[layer];

        Dune::shared_ptr<SubIntersectionPart> intersection
          ( createIntersectionPart(*insideDomainCell,*outsideDomainCell,c,normal,isBoundary) );
        multiIntersection->insertLayer(layer,intersection);

      }

      multiIntersection->resetLayer();

      #warning Deactivated "one sided assembler constraint" because the UDG assembler itself does the filtering.
      // // One sided assembler constraint
      // if( multiIntersection->outsideDomainIndex() != -1 &&
      //     multiIntersection->insideDomainIndex() > multiIntersection->outsideDomainIndex()){
      //   delete multiIntersection;
      //   return 0;
      // }

      return multiIntersection;
    }


    template< typename DD >
    typename MarchingCube33SubTriangulation<DD>::EntityPart*
    MarchingCube33SubTriangulation<DD>::
    createMultiEntityPart
    (
      const CellEntity & lCell,
      const Index partIndex,
      const Coordinates & c,
      const std::vector<const LookupItem*> & lookupItems
    ) const
    {
      // Allocate multi layer entity part
      EntityPart * multiEntity = new EntityPart(domainDef_.numberOfLayers());

      for(Index layer = 0; layer < domainDef_.numberOfLayers(); ++layer) {
        Dune::shared_ptr<SubEntityPart> subEntityPart
          (createEntityPart(lCell,partIndex,c,*(lookupItems[layer]->insideDomainCell)));
        multiEntity->insertLayer(layer,subEntityPart);
      }

      multiEntity->resetLayer();
      return multiEntity;
    }

    template< typename DD >
    typename MarchingCube33SubTriangulation<DD>::SubEntityPart*
    MarchingCube33SubTriangulation<DD>::
    createEntityPart
    (
      const CellEntity & homeInLGV,
      const int partIndex,
      Coordinates c,
      const GVCellEntity & hostInGV,
      const RawBoundingBox & rawbbox,
      const Index & domainIndex,
      const int tag,
      const DomainTag & domainTag,
      const DomainField & volume
    ) const
    {
      // Map to global coordinates
      for(Index i=0; i<c.size(); ++i) c[i] = homeInLGV.geometry().global(c[i]);

      assert(domainIndex != invalidIndex());

      BoundingBox bbox(hostInGV, rawbbox.lower, rawbbox.higher, Dune::CuboidMode::corners);

      Dune::GeometryType type = Dune::geometryTypeFromVertexCount(dim,c.size());
      typename EntityPart::GeometryImp geometry(type,c);
      EntityPartGeometry partGeometry(geometry,homeInLGV,bbox);

      return new SubEntityPart(partIndex,partGeometry,tag,domainTag,domainIndex,volume);
    }


    template< typename DD >
    typename MarchingCube33SubTriangulation<DD>::SubEntityPart*
    MarchingCube33SubTriangulation<DD>::
    createEntityPart
    (
      const CellEntity & homeInLGV,
      const int partIndex,
      const Coordinates & c,
      const DomainCell & domainCell
    ) const
    {
      // Create entity part interface object
      const GVCellEntity hostCell = gv_.grid().entity(domainCell.hostCell);
      const RawBoundingBox & bbox = domainCell.bbox;
      const Index domainIndex = domainCell.domainIndex;
      const DomainTag domainTag = domainDef_.domainIndex2Tag(domainIndex);
      const DomainField volume = domainCell.volume;

      assert(c.size()!=5);
      return createEntityPart
               (homeInLGV,partIndex,c,hostCell,bbox,domainIndex,domainCell.id,domainTag,volume);
    }


    template< typename DD >
    void
    MarchingCube33SubTriangulation<DD>::
    cacheIntersectionParts
      (const CellEntity & lCell,
      const std::vector<const LookupItem*> lookupItems,
      const DomainRefinementPart & part,
      IntersectionPartList & intersectionCache) const
    {
      // Create generic reference element for this part
      auto ref =
        Dune::referenceElement<DomainField,dim>(part.geometry.type());

      const Geometry globalGeometry = combineGeometries(part.geometry,lCell.geometry());

      // Determine all face indices of face to be created
      std::set<Index> indicesInPart;
      for(Index layer=0; layer<lookupItems.size(); ++layer) {

        const typename LookupItem::FacesLookupMap& faceLookupMap
          = lookupItems[layer]->facesLookup;

        typename LookupItem::FacesLookupMap::const_iterator faceIndexIt =
          faceLookupMap.begin();
        for(; faceIndexIt != faceLookupMap.end(); ++faceIndexIt)
          indicesInPart.insert(faceIndexIt->first);
      }


      // Create intersection parts
      typename std::set<Index>::const_iterator indexInPartIt =
        indicesInPart.begin();
      for(; indexInPartIt != indicesInPart.end(); ++indexInPartIt) {

        // Determine the intersection domain cells on all layers
        MultiIntersectionDomainCells intersectionDomainCells(lookupItems.size());
        for(Index layer=0; layer<lookupItems.size(); ++layer) {

          // Set inside domain cell
          intersectionDomainCells.insideDomainCells[layer] =
            lookupItems[layer]->insideDomainCell;

          const typename LookupItem::FacesLookupMap& faceLookupMap =
            lookupItems[layer]->facesLookup;

          typename LookupItem::FacesLookupMap::const_iterator faceIt =
            faceLookupMap.find(*indexInPartIt);

          if(faceIt != faceLookupMap.end())
            intersectionDomainCells.outsideDomainCells[layer]
              = faceIt->second;
          else
            intersectionDomainCells.outsideDomainCells[layer]
              = lookupItems[layer]->insideDomainCell;
        }


        // Extract face vertices
        Coordinates faceVertices;
        const Index indexInInside = *indexInPartIt;
        using namespace MarchingCube33SubTriangulationImp;
        extractFaceCoords(indexInInside,globalGeometry,faceVertices);

        // Check normal vector match
        const Coordinate localCenter = ref.position(indexInInside,1);
        Coordinate unitOuterNormal = computeNormal(globalGeometry,indexInInside,localCenter);
        const DomainField normalLength = unitOuterNormal.two_norm();
        assert(normalLength > 0);
        unitOuterNormal *= 1.0 / normalLength;

        IntersectionPart* intersection
          ( createMultiIntersectionPart(intersectionDomainCells,faceVertices,unitOuterNormal) );

        if(intersection) {
          IntersectionPartPointer intersectionPointer(intersection);
          intersectionCache.push_back( intersectionPointer );
        }

      }
    }

    template< typename DD >
    void
    MarchingCube33SubTriangulation<DD>::
    create_intersection_parts(const GVCellEntity & e,
                              IntersectionPartList & intersectionparts) const
    {
      assert(finalized_); // You forgot to call finalize()
      if(!useIntersections_) {
        DUNE_THROW(Dune::Exception,"Intersections have been disabled!");
      }

      // Reset intersection part cache
      const Index cellIndex = gv_.indexSet().index(e);

      typename IntersectionCache::Storage * threadIntersectionCache
        = &(intersectionCache_.findInCache(stmutex_));
      IntersectionPartList * intersectionPartList =
        &(threadIntersectionCache->intersectionCache);

      while(threadIntersectionCache->cachedCellIndex != cellIndex) {
        threadIntersectionCache->release();

        EntityPartList tmp;
        create_entity_parts(e,tmp);

        // Find again to reserve cache
        threadIntersectionCache
          = &(intersectionCache_.findInCache(stmutex_));
        intersectionPartList = &(threadIntersectionCache->intersectionCache);
      }

      // We simply copy the intersection parts from the cache.
      intersectionparts = *intersectionPartList;
      threadIntersectionCache->release();
    }


    template< typename DD >
    void
    MarchingCube33SubTriangulation<DD>::
    create_entity_parts
      (const GVCellEntity & e, EntityPartList & entityparts) const
    {
      assert(finalized_); // You forgot to call finalize()
      const Index cellIndex = gv_.indexSet().index(e);

      // Reset intersection part cache
      typename IntersectionCache::Storage & threadIntersectionCache
        = intersectionCache_.findInCache(stmutex_);
      IntersectionPartList & intersectionPartList =
        threadIntersectionCache.intersectionCache;
      intersectionPartList.clear();

      // Check whether this cell contains only one domain and is
      // therefore eligible for providing only a single entity part
      const Index nLayers = domainDef_.numberOfLayers();
      if( (!forceRefinement_) && domainDef_.numberOfInterfacesInCell(cellIndex,nLayers-1)==0) {
        createMultiPartsForWholeCell(e,entityparts,intersectionPartList);
        threadIntersectionCache.cachedCellIndex = cellIndex;
        threadIntersectionCache.release();
        return;
      }

      // Iterate the cell's child-cells in level set mesh
      using namespace MarchingCube33SubTriangulationImp;
      SubGridViewIterator<GV,LGV> hit(gv_,lgv_,e);
      for(; hit.valid(); ++hit) {

        const CellEntity & lCell = *hit;
        const Index lCellIndex = lgv_.indexSet().index(lCell);

        // Create domain refinement for current child cell
        typedef typename DomainRefinement::DomainPartList PartList;
        PartList parts;
        DomainTag lCellDomainTag;
        layers_.back().createDomainRefinement(lCell,parts,lCellDomainTag);

        // Iterate parts of current child cell's domain refinement
        typedef typename PartList::const_iterator Iterator;
        Iterator pit = parts.begin();
        for(; pit != parts.end(); ++pit) {

          // Find and store a domain cell lookup item on each layer
          std::vector<const LookupItem*> lookupItems(domainDef_.numberOfLayers());
          for(Index layer = 0; layer < domainDef_.numberOfLayers(); ++layer) {
            typedef typename TriangulationLayer::LookupMap LookupMap;
            const LookupMap & lookupMap = layers_[layer].lookupMap();

            typedef typename LookupMap::LookupItem LookupItem;
            const LookupItem * lookupItem = lookupMap.find(cellIndex,lCellIndex,pit->index);
            assert(lookupItem);

            lookupItems[layer] = lookupItem;
          }

          // Compute and cache the intersection parts for current refinement part
          if(useIntersections_)
            cacheIntersectionParts(lCell,lookupItems,*pit,intersectionPartList);

          // Aquire the current refinement part's corners in local coordinates
          // of child cell
          using namespace MarchingCube33SubTriangulationImp;
          Coordinates c = getGeometryCorners(pit->geometry);

          // Special switch to avoid integration on pyramids
          if(c.size() == 5) {
            Coordinates s0(4);
            s0[0] = c[0]; s0[1] = c[1]; s0[2] = c[2]; s0[3] = c[4];
            Coordinates s1(4);
            s1[0] = c[1]; s1[1] = c[2]; s1[2] = c[3]; s1[3] = c[4];

            EntityPartPointer p0(createMultiEntityPart(lCell,pit->index,s0,lookupItems));
            EntityPartPointer p1(createMultiEntityPart(lCell,pit->index,s1,lookupItems));
            entityparts.push_back(p0);
            entityparts.push_back(p1);
          }
          else{
            EntityPartPointer p0(createMultiEntityPart(lCell,pit->index,c,lookupItems));
            entityparts.push_back(p0);
          }

        } // pit

      } // hit (lCell)

      threadIntersectionCache.cachedCellIndex = cellIndex;
      threadIntersectionCache.release();
    }


    template< typename DD >
    const typename MarchingCube33SubTriangulation<DD>::BoundingBox
    MarchingCube33SubTriangulation<DD>::
    BBox(const unsigned int domainIndex, const GVCellEntity & host) const
    {
      DUNE_THROW(Dune::Exception,"Should not be called!");
      const DomainCell * current(0);
      return BoundingBox(gv_.grid().entity(current->hostCell), current->bbox.lower,
                         current->bbox.higher, Dune::CuboidMode::corners);
    }

    template< typename DD >
    const typename MarchingCube33SubTriangulation<DD>::BoundingBox
    MarchingCube33SubTriangulation<DD>::
    boundingBox(const EntityPartInterface & ep) const
    {
      // Query for domain cell lookup map
      const size_t layer = ep.currentLayer();
      typedef typename TriangulationLayer::LookupMap LookupMap;
      typedef typename LookupMap::LookupItem LookupItem;
      const LookupMap & lookupMap = layers_[layer].lookupMap();

      // Determine corresponding home entity contained in either grid view
      CellEntity partEntity ( ep.entity() );

      // Determine corresponding home entity in fundamental mesh grid view
      GVCellEntity gvEntity = domainDef_.gvEntity(partEntity);
      const Index cellIndex = gv_.indexSet().index(gvEntity);

      // Determine corresponding domain cell lookup item
      const LookupItem * lookupItem;
      if(lgv_.indexSet().contains(partEntity)) {

        const Index lCellIndex = lgv_.indexSet().index(partEntity);
        lookupItem = lookupMap.find(cellIndex,lCellIndex,ep.partIndex());
        assert(lookupItem);
      }
      else{
        lookupItem = lookupMap.find(cellIndex);

        // This case only occurs if fundamental mesh grid view and level set
        // mesh grid view do not coincide and the entity part is a whole
        // fundamental mesh cell with respect to its original instance of
        // MarchingCube33SubTriangulation but not a whole fundamental mesh
        // cell with respect to this instance of MarchingCube33SubTriangulation
        if(!lookupItem) {
          typedef typename CellEntity::HierarchicIterator HierarchicIterator;
          HierarchicIterator hit( partEntity.hbegin(lgv_.grid().maxLevel()) );
          const HierarchicIterator ehit( partEntity.hend(lgv_.grid().maxLevel()) );
          for(; hit != ehit; ++hit) {
            while(!lgv_.indexSet().contains(*hit)) continue;

            const Index lCellIndex = lgv_.indexSet().index(*hit);
            const Index partIndex = 0; // At this position it must be zero
            lookupItem = lookupMap.find(cellIndex,lCellIndex,partIndex);
            assert(lookupItem);
            break;
          }
        }
      }

      // Extract domain cell which contains the entity part
      const DomainCell * domainCell = lookupItem->insideDomainCell;

      // Create and return bounding box
      return BoundingBox(gv_.grid().entity(domainCell->hostCell), domainCell->bbox.lower,
                         domainCell->bbox.higher, Dune::CuboidMode::corners);
    }


    template< typename DD >
    bool
    MarchingCube33SubTriangulation<DD>::
    isHostCell(const GVCellEntity & hostCandidate, const int & domainIndex) const
    {
      for(Index l=0; l<layers_.size(); ++l)
        if(layers_[l].isHostCell(hostCandidate,domainIndex))
          return true;
      return false;
    }


    template< typename DD >
    void
    MarchingCube33SubTriangulation<DD>::
    init()
    {
      // Initalize non-const members, optionally using the parameters
      forceRefinement_ = false;
      useIntersections_ = true;
      forceRefinement_
        = parameters_.get<bool>("force_refinement",forceRefinement_);
      useIntersections_
        = parameters_.get<bool>("use_intersections",useIntersections_);

      // Get settings needed for initialization, optionally from parameters
      double removeIdenticalDomainCellFaceCornersEps = 1e-9;
      double minEdgeLengthEps = 1e-2;
      bool linearizeParts = true;
      removeIdenticalDomainCellFaceCornersEps
        = parameters_.get<double>("remove_identical_domaincell_face_corners_eps",
                                  removeIdenticalDomainCellFaceCornersEps);
      minEdgeLengthEps
        = parameters_.get<double>("min_edge_length_eps",minEdgeLengthEps);
      linearizeParts
        = parameters_.get<bool>("linearize_parts",linearizeParts);

      typename IntersectionCache::Storage & threadIntersectionCache
        = intersectionCache_.findInCache(stmutex_);
      threadIntersectionCache.cachedCellIndex = invalidIndex();
      threadIntersectionCache.release();

      // Assert that the domain and interface definitions as provided by
      // the user are complete and consistent.
      if(!domainDef_.finalize())
      {DUNE_THROW(Dune::Exception,"Domains or Interfaces are not well defined");}

      // Initialize triangulation layers
      assert(domainDef_.numberOfLayers());

      for(Index l=0; l<domainDef_.numberOfLayers(); ++l) {
        std::cout << "Initialize triangulation layer "
                  << l << std::endl;
        layers_.push_back
          (TriangulationLayer(domainDef_,l,
                              removeIdenticalDomainCellFaceCornersEps,
                              minEdgeLengthEps,linearizeParts));
        layers_[l].init();
      }
    }

    template< typename DD >
    template< typename Merger >
    void
    MarchingCube33SubTriangulation<DD>::
    mergeDomainCells(Merger & merger)
    {
      merger.mergeAllLayers(layers_);
      for(Index l=0; l<layers_.size(); ++l)
        merger.merge(layers_[l].hostCellMap());
    }

    template< typename DD >
    template< typename Merger >
    void
    MarchingCube33SubTriangulation<DD>::
    mergeDomainCells(const int layer, Merger & merger)
    {
      merger.merge(layers_[layer].hostCellMap());
    }

    template< typename DD >
    void
    MarchingCube33SubTriangulation<DD>::
    finalize(const int layerIndex, bool verbose)
    {
      // Get verbosity settings, optionally from parameters
      verbose = parameters_.get<bool>("finalize_verbose",verbose);

      // Write mapping : home cell -> possible host cells
      if(layerIndex < 0) {
        for(Index l=0; l<layers_.size(); ++l) {
          std::cout << "Write lookup map of layer " << l << std::endl;
          layers_[l].writeLookupMap(verbose);
        }
      }
      else{
        std::cout << "Write lookup map of layer " << layerIndex
                  << std::endl;
        layers_[layerIndex].writeLookupMap(verbose);
      }

      if(verbose)
        for(Index i=0; i<domainDef_.numberOfInterfaces(); ++i)
          domainDef_.levelSetFunctionBackend(i).printStatistics();

      finalized_ = true;
      std::cout << "Finished intialization of triangulation" << std::endl;
    }

  } // UDG
} // Dune


#endif
