// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_MC33TRIANGULATION_DEBUG_CC
#define DUNE_MC33TRIANGULATION_DEBUG_CC

#include <dune/common/dynmatrix.hh>


namespace Dune
{
  namespace UDG
  {

    template< typename DD >
    void
    MarchingCube33SubTriangulation<DD>::
    debugFaceConnections
      (const DomainField minInside, const DomainField maxInside) const
    {
      if(!forceRefinement_) {
        DUNE_THROW(Dune::Exception,"This function requires forceRefinement to be set");
      }

      // Index: host cell index; Value: Number of faces with this
      // inside - outside combination.
      typedef Dune::DynamicMatrix<int> CounterMatrix;
      typedef std::vector< CounterMatrix > FaceCounter;
      const Index nDomains = domainDef_.numberOfDomains();
      FaceCounter internalCounter(layers_.back().hostCellMap().size(),CounterMatrix(nDomains,nDomains+1,0));

      std::cout << "Begin intersection part check:" << std::endl;

      for(Index c=0; c<layers_.back().hostCellMap().size(); ++c) {

        const DomainCellList & domainCells = layers_.back().hostCellMap()[c];
        typedef typename DomainCellList::const_iterator Iterator;
        Iterator it = domainCells.begin();
        for(; it != domainCells.end(); ++it) {

          typedef typename DomainCell::DomainCellFaceListMap DomainCellFaceListMap;
          const DomainCellFaceListMap & faces = it->faceListMap;
          typedef typename DomainCellFaceListMap::const_iterator DomainCellFaceListIt;
          DomainCellFaceListIt flit = faces.begin();
          for(; flit != faces.end(); ++flit) {

            typedef std::list<DomainCellFace> DomainCellFaceList;
            const DomainCellFaceList & faceList = flit->second;
            typename DomainCellFaceList::const_iterator fit = faceList.begin();
            for(; fit != faceList.end(); ++fit) {
              const Index insideDomainIndex = fit->insideDomainCell->domainIndex;

              if(fit->mirrorFace()) {
                const Index outsideDomainIndex =
                  fit->mirrorFace()->insideDomainCell->domainIndex;

                // Such faces are ignored by the interface calls to be
                // consistent with the PDELab UDG assembler.
                if(insideDomainIndex > outsideDomainIndex)
                  continue;

                internalCounter[c][insideDomainIndex][outsideDomainIndex]+=1;
              }
              else{

                bool outside = false;
                for(int d=0; d<dim; ++d)
                  outside |=
                    (fit->geometry.center())[d] < minInside ||
                    (fit->geometry.center())[d] > maxInside;
                if(!outside) {
                  std::cout << "!T! Unmatched face at (rel+-1e-6) : "<< fit->geometry.center()
                            << std::endl;
                  std::cout << "(Inside lCell,part,index in part) : "
                            << fit->insideLCellIndex << "," << fit->insidePartIndex
                            << "," << fit->indexInPart
                            << std::endl;
                }

                internalCounter[c][insideDomainIndex][nDomains]+=1;
              }
            }
          } // flit

        } // it

      } // c


      // Interface check
      FaceCounter interfaceCounter(layers_.back().hostCellMap().size(),CounterMatrix(nDomains,nDomains+1,0));
      typedef typename GV::Traits::template Codim<0>::Iterator GVCellIterator;
      GVCellIterator it = gv_.template begin<0>();
      const GVCellIterator eit = gv_.template end<0>();
      for(; it != eit; ++it) {
        IntersectionPartList intersections;
        create_intersection_parts(*it,intersections);
        for(typename IntersectionPartList::const_iterator fit=intersections.begin();
            fit != intersections.end(); ++fit) {
          const IntersectionPart & iPart = static_cast<IntersectionPart&>(**fit);
          const Index insideHostIndex = gv_.indexSet().index(*(iPart.boundingBoxInside().entity()));
          const int insideDomainIndex = iPart.insideDomainIndex();
          if(iPart.neighbor()) {
            const int outsideDomainIndex = iPart.outsideDomainIndex();
            interfaceCounter[insideHostIndex][insideDomainIndex][outsideDomainIndex]++;
          }
          else{
            interfaceCounter[insideHostIndex][insideDomainIndex][nDomains]++;
          }
        }
      }

      // Compare counters
      Index matches(0);
      Index mismatches(0);
      CounterMatrix interfaceSum(nDomains,nDomains+1,0);
      CounterMatrix internalSum(nDomains,nDomains+1,0);
      for(Index c=0; c<layers_.back().hostCellMap().size(); ++c) {

        interfaceSum += interfaceCounter[c];
        internalSum += internalCounter[c];
        CounterMatrix diff (interfaceCounter[c]);
        diff -= internalCounter[c];
        if(diff.infinity_norm() != 0) {
          std::cout << "!T! Mismatch in host cell "<< c << std::endl;
          std::cout << "Interface:"<< std::endl;
          std::cout << interfaceCounter[c] << std::endl;
          std::cout << "Internal:"<< std::endl;
          std::cout << internalCounter[c] << std::endl;
          mismatches++;
        }
        else{
          matches++;
        }

      }
      std::cout << "Faces matched in "<< matches <<" of "<< mismatches+matches
                << " host cells." << std::endl;
      std::cout << "Sum of connections for interface: " << std::endl
                << interfaceSum << std::endl;
      std::cout << "Sum of internal connections: " << std::endl
                << internalSum << std::endl;
      std::cout << "Testing output: " << std::endl;
      for(Index r=0; r<nDomains; ++r)
        for(Index c=r; c<nDomains; ++c)
          std::cout <<"!T! Faces between domains "<< r <<"-"<< c <<" (abs+-0.5): "
                    << interfaceSum[r][c] << std::endl;
      std::cout << "Finished intersection part check!" << std::endl;

    }

  }
}
#endif
