// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_MC33TRIANGULATIONIMP_HH
#define DUNE_MC33TRIANGULATIONIMP_HH

#if DUNE_VERSION_NEWER(DUNE_GEOMETRY,2,3)
#include <dune/geometry/multilineargeometry.hh>
#else
#include <dune/geometry/genericgeometry/geometry.hh>
#endif
#include <dune/udg/common/utilities.hh>

namespace MarchingCube33SubTriangulationImp
{

  /**
   * \brief Compute the vertices of a given geometry object in local
   *        coordinates of the corresponding reference element.
   *
   * \todo Isn't actually Geometry::LocalCoordinate the correct type
   *       here in general?
   */
  template <typename Geometry>
  std::vector<typename Geometry::GlobalCoordinate>
  getReferenceCorners(const Geometry & g)
  {
    std::vector<typename Geometry::GlobalCoordinate> c(g.corners());
    const int dim = Geometry::GlobalCoordinate::dimension;
    typedef typename Geometry::GlobalCoordinate::value_type DomainField;
    auto ref =
      Dune::referenceElement<DomainField,dim>(g.type());

    for(int i=0; i<g.corners(); ++i)
      c[i] = ref.position(i,dim);
    return c;
  }

  /**
   * \brief Compute the vertices of a given geometry object in
   *        global coordinates of the geometry object.
   */
  template <typename Geometry>
  std::vector<typename Geometry::GlobalCoordinate>
  getGeometryCorners(const Geometry & g)
  {
    std::vector<typename Geometry::GlobalCoordinate> c(g.corners());
    for(int i=0; i<g.corners(); ++i)
      c[i] = g.corner(i);
    return c;
  }

  // Remove all coordinates from the vector for which the indicator
  // vector provides a value < 0.5.
  template< typename Coordinate >
  void
  reduceByMatchingFactors
    (std::vector<Coordinate> & corners,
    const std::vector<typename Coordinate::value_type> & indicators)
  {
    size_t newsize(0);
    for(size_t i=0; i<indicators.size(); ++i) {
      if(indicators[i] > 0.5)
        corners[newsize++] = corners[i];
    }
    corners.resize(newsize);
  }

  // Pre-Declaration
  template< typename Coordinate >
  size_t
  computeMatchingFactors
    (std::vector<Coordinate> & corners,
    std::vector<typename Coordinate::value_type> & indicators,
    typename Coordinate::value_type eps);


  // Remove corners from given array which are doublettes
  template< typename Coordinate >
  void
  removeDegeneratedCorners
    (std::vector<Coordinate> & corners, typename Coordinate::value_type eps)
  {
    typedef typename Coordinate::value_type Field;
    std::vector<Field> indicators;
    if(!computeMatchingFactors(corners,indicators,eps))
      return;

    reduceByMatchingFactors(corners,indicators);
  }

  // Remove corners from given array which are doublettes
  template< typename Cell, typename Coordinate >
  void
  removeGlobalDegeneratedCorners
    (const Cell & cell, std::vector<Coordinate> & corners,
    typename Coordinate::value_type eps)
  {
    // Make copy and then map to local coordinates
    std::vector<Coordinate> copy(corners);
    for(size_t c=0; c<corners.size(); ++c)
      copy[c] = cell.geometry().local(corners[c]);

    typedef typename Coordinate::value_type Field;
    std::vector<Field> indicators;
    if(!computeMatchingFactors(copy,indicators,eps))
      return;

    reduceByMatchingFactors(corners,indicators);
  }


  // Evaluate Lagrangian Q1 function on cube at local coordinate x.
  template<typename Domain, typename Values>
  typename Values::value_type
  evaluateOnCube(const Domain & x, const Values & values)
  {
    typedef typename Values::value_type RF;
    RF val(0);

    const int dim = Domain::dimension;
    const size_t size = 1 << dim;
    for (size_t i=0; i<size; i++) {
      RF sf_val(1);

      for (int j=0; j<dim; j++)
        // if j-th bit of i is set multiply with in[j], else with 1-in[j]
        sf_val *= (i & (1<<j)) ? x[j] :  1-x[j];

      // Value is linear combination of all sf values with coeffs
      val += sf_val * values[i];
    }
    return val;
  }

  // Evaluate Lagrangian P1 function on simplex at local coordinate x.
  template<typename Domain, typename Values>
  typename Values::value_type
  evaluateOnSimplex(const Domain & x, const Values & values)
  {
    typedef typename Values::value_type RF;
    RF val(0);

    const int dim = Domain::dimension;
    val = values[0];
    for (int i=0; i<dim; i++) {
      val -= values[0] * x[i];
      val += values[i+1] * x[i];
    }
    return val;
  }


  // Compute the values of the level set function (defined by values
  // hostValues in the corners of the host geometry) in the corners of
  // the guest geometry (which has to be embedded in the host
  // geometry). The host geometry may either be a cube or a
  // simplex. The guest geometry may be of arbitrary topological type.
  template< typename GuestGeometry, typename Values >
  void computeValuesInVertices(const Dune::GeometryType & hostType,
                               const Values & hostValues,
                               const GuestGeometry & guest, Values & guestValues)
  {
    guestValues.resize(guest.corners());

    if(hostType.isCube()) {
      for(int v=0; v<guest.corners(); ++v)
        guestValues[v] = evaluateOnCube(guest.corner(v),hostValues);
    }
    else if(hostType.isSimplex()) {
      for(int v=0; v<guest.corners(); ++v)
        guestValues[v] = evaluateOnSimplex(guest.corner(v),hostValues);
    }
    else{
      DUNE_THROW(Dune::NotImplemented,"Can not evaluated level set function on "
                 "given geometry type.");
    }
  }


  /**
   * \brief Given a geometry object and one of its faces via its
   *        index with respect to the geometry object's corresponding
   *        reference element, extract the face's vertices in global
   *        coordinates of the geometry object as a container of
   *        coordinates.
   */
  template <typename Geometry, typename Coordinates>
  void extractFaceCoords
    (const int faceIndex, const Geometry & geom, Coordinates & faceVertices)
  {
    typedef typename Coordinates::value_type Coordinate;
    typedef typename Coordinate::value_type DomainField;
    const size_t dim = Coordinate::dimension;

    auto ref =
      Dune::referenceElement<DomainField,dim>(geom.type());
    assert(faceIndex < ref.size(1));
    // Create face geometry
    const int nFaceVertices = ref.size(faceIndex,1,dim);
    using namespace MarchingCube33SubTriangulationImp;
    faceVertices.resize(nFaceVertices);
    for(int v=0; v<nFaceVertices; ++v) {
      const int vIndex = ref.subEntity(faceIndex,1,v,dim);
      faceVertices[v] = geom.corner(vIndex);
    }
  }


  // Compute a normal vector on the face given by its local index
  // within the geometry 'geom' at the position given by the cell
  // local coordinate 'local'.
  template<typename Geometry, typename Coordinate>
  Coordinate computeNormal
    (const Geometry & geom, const int faceIndex, const Coordinate & local)
  {
    typedef typename Coordinate::value_type DomainField;
    const size_t dim = Coordinate::dimension;

    auto refElement
      = Dune::referenceElement< DomainField, dim >( geom.type() );
    assert( faceIndex < refElement.size( 1 ) );

    const typename Geometry::JacobianInverseTransposed &jit
      = geom.jacobianInverseTransposed( local );
    const Coordinate &refNormal = refElement.integrationOuterNormal( faceIndex );

    Coordinate normal;
    jit.mv( refNormal, normal );
    normal *= geom.integrationElement( local );
    return normal;
  }


  /** \brief This class provides static functionality to determine a
      local index of a corner in a reference element for a given local
      coordinate.
   */
  class Coordinate2LocalIndexHelper
  {
  private:
    //! Compute local index of node entity corresponding to local
    //! coordinate
    template<typename CellEntity, typename Coordinate, typename Comparator>
    static int localCubeIndex
      (const Comparator & ccomp, const CellEntity & lCell, const Coordinate & local)
    {
      Coordinate corner;
      typedef typename Coordinate::value_type DomainField;
      const int dim = Coordinate::dimension;
      int index(0);
      for(int d=0; d<dim; ++d) {
        const int comp = int(local[d] > 0.5);
        corner[d] = DomainField(comp);
        index += (1<<d) * comp;
      }
      corner = lCell.geometry().global(corner);
      const Coordinate global = lCell.geometry().global(local);
      if(ccomp.isEqual(corner,global))
        return index;
      else
        return -1;
    }

    //! Compute local index of node entity corresponding to local
    //! coordinate
    template<typename CellEntity, typename Coordinate, typename Comparator>
    static int localSimplexIndex
      (const Comparator & ccomp, const CellEntity & lCell, const Coordinate & local)
    {
      Coordinate corner;
      typedef typename Coordinate::value_type DomainField;
      const int dim = Coordinate::dimension;
      int index(0);
      for(int d=0; d<dim; ++d) {
        const int comp = int(local[d] > 0.5);
        corner[d] = DomainField(comp);
        index += (d+1) * comp;
      }
      corner = lCell.geometry().global(corner);
      const Coordinate global = lCell.geometry().global(local);
      if(ccomp.isEqual(corner,global))
        return index;
      else
        return -1;
    }
  public:

    //! Compute local index of node entity corresponding to local
    //! coordinate (works only for cubes).
    template<typename CellEntity, typename Coordinate, typename Comparator>
    static int localIndex
      (const Comparator & ccomp, const CellEntity & lCell, const Coordinate & local)
    {
      if(lCell.type().isCube()) return localCubeIndex(ccomp,lCell,local);
      else if(lCell.type().isSimplex()) return localSimplexIndex(ccomp,lCell,local);
      else { DUNE_THROW(Dune::NotImplemented,"Not implemented for this type"); }
    }
  };


  /** \brief Helper class to compare global coordinates.

      This class uses a mapping that is scaled to the smallest cell of
      one (or more) grid view(s). This reference map together with a
      numerical epsilon in the reference domain is used to determine
      whether two coordinate in the global domain are equal.

      It also provides some convenience functions providing efficient
      ways to determine wheter a global coordinate matches
      (numerically) at least one coordinate in a given container.

   */
  template< typename Coordinate >
  class NormalizedCoordinateComparator
  {
  public:

    enum { dim = Coordinate::dimension };
    typedef typename Coordinate::value_type DomainField;
    typedef Coordinate GlobalCoordinate;

#if DUNE_VERSION_NEWER(DUNE_GEOMETRY,2,3)
    typedef Dune::CachedMultiLinearGeometry<DomainField, dim, dim> Geometry;
#else
    typedef typename Dune::GenericGeometry::
    DefaultGeometryTraits<DomainField,dim,dim> GeometryTraits;
    typedef Dune::GenericGeometry::
    BasicGeometry<dim, GeometryTraits > Geometry;
#endif

    /** \brief Constructor */
    NormalizedCoordinateComparator() :
      centerOfMass_(0),
      mapCoords_(dim+1,centerOfMass_),
      simplexType_(Dune::GeometryTypes::simplex(dim)),
      referenceMap_(simplexType_, mapCoords_),
      volume_(0),
      finalized_(false),
      H_(std::numeric_limits<DomainField>::min()),
      h_(std::numeric_limits<DomainField>::max()),
      minCoord_(std::numeric_limits<DomainField>::max()),
      maxCoord_(std::numeric_limits<DomainField>::min()),
      eps_(1e-7)
    {}

    /** Access minimum distance between two points that should be
        considered equal.
     */
    DomainField & numericalEqualityEps()
    { return eps_;}

    /** Const access minimum distance between two points that should be
        considered equal.
     */
    const DomainField & numericalEqualityEps() const
    { return eps_;}

    /** \brief Update reference map with regard to this cell geometry.

        This function should be called for every grid cell before
        using any other function of this class.
     */
    template< typename CellGeometry >
    void updateReferenceMap(const CellGeometry & cellGeometry)
    {
      // Reference element of cell to refine
      auto ref =
        Dune::referenceElement<DomainField,dim>(cellGeometry.type());

      // Iterate edges
      for(int e=0; e<ref.size(dim-1); ++e) {
        const int lVertexIndex = ref.subEntity(e,dim-1,0,dim);
        const int rVertexIndex = ref.subEntity(e,dim-1,1,dim);

        // Compute global coords of edge nodes
        const Coordinate lVertexCoord = cellGeometry.corner(lVertexIndex);
        const Coordinate rVertexCoord = cellGeometry.corner(rVertexIndex);
        const Coordinate edge = lVertexCoord - rVertexCoord;

        for(int d=0; d<dim; ++d) {
          minCoord_[d] = std::min(minCoord_[d],lVertexCoord[d]);
          minCoord_[d] = std::min(minCoord_[d],rVertexCoord[d]);
          maxCoord_[d] = std::max(maxCoord_[d],lVertexCoord[d]);
          maxCoord_[d] = std::max(maxCoord_[d],rVertexCoord[d]);
        }

        // Update minimun edge length
        const DomainField length = edge.two_norm();
        h_ = std::min(h_,length);
        H_ = std::max(H_,length);
      }

      // Update center of mass
      const DomainField cellVolume = cellGeometry.volume();
      centerOfMass_.axpy(cellVolume,cellGeometry.center());
      volume_ += cellVolume;
    }

    /** \brief Finalize the setup of this object.

        This should be called after updateReferenceMap was called of
        each cell of the grid and before calling any other member
        function.
     */
    void finalize(bool verbose=true)
    {
      finalized_ = true;

      // Compute center of mass of cells
      centerOfMass_ /= DomainField(volume_);

      // Write transformation map to normalized domain
      for(int d=0; d<dim; ++d)
        mapCoords_[0][d] -= 0.5*H_;
      for(int d=0; d<dim; ++d) {
        mapCoords_[d+1] = mapCoords_[0];
        mapCoords_[d+1][d] += H_;
      }

      referenceMap_ = Geometry(simplexType_,mapCoords_);

      if(verbose) {
        std::cout << "Initialized normalized coordinate comparator:" << std::endl;
        std::cout << "\tReference simplex: " << mapCoords_[0] << std::endl;
        for (size_t i = 1; i < mapCoords_.size(); ++i)
          std::cout << "\t                 : " << mapCoords_[i] << std::endl;
        std::cout << "\tH                : " << H_ << std::endl
                  << "\tCenter of mass   : " << centerOfMass_ << std::endl;

      }
    }

    /** \brief Returns a metric values for two coordinates v and
        w. For err:=|v-w|_2 we get:

        If err == 0             :      0,
        if err == eps           :    0.5,
        if err -> infinity      : -> 1.0.
     */
    static DomainField metricInRef
      (const Coordinate & v, const Coordinate & w, const DomainField eps)
    {
      const double nVal = (v-w).two_norm();
      return nVal / (nVal + eps);
    }

    /** \brief Returns a metric values for two coordinates v and
        w. For err:=|v-w|_2 we get:

        If err == 0             :      0,
        if err == eps           :    0.5,
        if err -> infinity      : -> 1.0.
     */
    DomainField metricInRef
      (const Coordinate & v, const Coordinate & w) const
    {
      const double nVal = (v-w).two_norm();
      return nVal / (nVal + eps_);
    }

    //! Directly compare two scalars
    bool isEqual(const DomainField & a, const DomainField & b) const
    { return std::abs(a-b) < eps_; }

    //! Directly compare two coordinates
    bool isEqualInRef(Coordinate globalA, Coordinate globalB) const
    {
      return metricInRef(globalA,globalB) < 0.5;
    }

    //! Directly compare two coordinates
    bool isEqualInRef(Coordinate globalA, Coordinate globalB,
                      const DomainField eps) const
    {
      return metricInRef(globalA,globalB,eps) < 0.5;
    }

    //! Compare two global coordinates in the reference domain
    bool isEqual(Coordinate globalA, Coordinate globalB) const
    {
      globalA = referenceMap_.local(globalA);
      globalB = referenceMap_.local(globalB);
      return metricInRef(globalA,globalB) < 0.5;
    }

    /** \brief Returns true if the given global corner is one of ther
        corners in the given global geometry.

        \param[in] corner       Global corner to compare
        \param[in] gb           Global geometry to compare
        \param[in] refMap       Reference map for normalization of coordinates
        \param[in] eps          Minimum distance for two equal coordinates
                                in reference domain.
        \param[in] subEntity    The local index of the sub entity of
                                gb which to search
        \param[in] codim        The codim of the sub entity in which
                                to search
     */
    template < typename TestGeometry, typename RefGeometry, typename Corner >
    static bool hasCorner
      (const Corner & corner, const TestGeometry & gb,
      const RefGeometry & refMap, DomainField eps,
      const int subEntity = 0, const int codim = 0)
    {
      auto ref =
        Dune::referenceElement<DomainField,TestGeometry::mydimension>(gb.type());

      assert(subEntity < ref.size(codim));
      const int nCorners = ref.size(subEntity,codim,TestGeometry::mydimension);

      // Scale eps to number of coordinates
      eps /= DomainField(nCorners);

      DomainField matchingValue = 1.0;
      const Coordinate aLocalCenter = refMap.local(corner);
      for(int sc=0; sc<nCorners; ++sc) {
        const int cb = ref.subEntity(subEntity,codim,sc,TestGeometry::mydimension);
        const Coordinate bLocalCenter = refMap.local(gb.corner(cb));
        matchingValue *= metricInRef(aLocalCenter,bLocalCenter,eps);
      }
      return bool(matchingValue < 0.4);
    }

    /** \brief Returns the number of corners that the two global
        geometries share.

        \param[in] ga           Global geometry to compare
        \param[in] gb           Global geometry to compare
     */
    template < typename AGeometry, typename BGeometry >
    int commonCorners
      (const AGeometry & ga, const BGeometry & gb) const
    {
      int nCommon = 0;
      for(int ca=0; ca<ga.corners(); ++ca) {
        const Coordinate aCoord = ga.corner(ca);
        nCommon += int( hasCorner(aCoord,gb,referenceMap_,eps_) );
      }
      return nCommon;
    }

    //! Map global coordinate to normalized reference domain.
    const Coordinate inRef(const Coordinate & c) const
    { return referenceMap_.local(c); }

    //! Check whether this object has been set up properly and may be
    //! used.
    bool valid() const { return finalized_; }

    /** \brief Constant access to reference map

        Notice that using the global() member of this map, actually
        maps a coordinate in the reference map into the global domain.
     */
    const Geometry & referenceMap() const { return referenceMap_; }

    //! Returns coordinate which is smaller than every coordinate
    //! within the domain in every component
    const Coordinate domainMinimum() const { return minCoord_; }
    //! Returns coordinate which is bigger than every coordinate
    //! within the domain in every component
    const Coordinate domainMaximum() const { return maxCoord_; }
    //! Minimume distance between two connected nodes in grid view
    const DomainField minDistance() const { return h_; }
  private:
    Coordinate centerOfMass_;
    std::vector<Coordinate> mapCoords_;
    const Dune::GeometryType simplexType_;
    Geometry referenceMap_;
    DomainField volume_;
    bool finalized_;
    DomainField H_;
    DomainField h_;
    Coordinate minCoord_;
    Coordinate maxCoord_;
    DomainField eps_;
  };



  // Compute for each coordinate a matching factor indicating whether
  // it is a doublette within the given vector of coordinates. The
  // coloring is such that only the first n-1 of n doublettes will
  // have values < 0.5.
  template< typename Coordinate >
  size_t
  computeMatchingFactors
    (std::vector<Coordinate> & corners,
    std::vector<typename Coordinate::value_type> & indicators,
    typename Coordinate::value_type eps)
  {
    size_t matches(0);

    typedef typename Coordinate::value_type Field;
    indicators = std::vector<Field>(corners.size(),1.0);
    for(size_t c=0; c<corners.size(); ++c) {
      for(size_t cc=c+1; cc<corners.size(); ++cc) {
        typedef NormalizedCoordinateComparator<Coordinate> Comp;
        indicators[c] *= Comp::metricInRef(corners[c],corners[cc],eps);
      }
      matches += size_t(indicators[c] < 0.5);
    }
    return matches;
  }


  /**
   * \brief An non-STL type iterator which allows to iterate on the
   *        subset of child-cells of a cell e in a grid view gv which
   *        are contained in a refined grid view subgv.
   */
  template <typename GV, typename SubGV>
  class SubGridViewIterator
  {
  public:
    typedef typename GV::template Codim<0>::Entity Entity;
    typedef typename SubGV::template Codim<0>::Entity SubEntity;
    typedef typename Entity::HierarchicIterator HierarchicIterator;

    SubGridViewIterator
      (const GV & gv, const SubGV & subgv, const Entity & e)
      : gv_(gv), subgv_(subgv), e_(e),
        hit(e.hbegin(gv.grid().maxLevel())),
        ehit(e.hend(gv.grid().maxLevel()))
    {
      noHierarchy = subgv_.indexSet().contains(e);

      if(noHierarchy) hit = ehit;

      if(!noHierarchy && !subgv_.indexSet().contains(*hit))
        while(++hit != ehit && !subgv_.indexSet().contains(*hit)) ;
    }

    SubGridViewIterator & operator++()
    {
      if(!noHierarchy)
        while(++hit != ehit && !subgv_.indexSet().contains(*hit)) ;
      else
        noHierarchy = false;
      return *this;
    }

    const SubEntity & operator*()
    {
      if(noHierarchy) return e_;
      else return *hit;
    }

    bool valid() const
    {
      return noHierarchy || hit != ehit;
    }

  private:
    const GV & gv_;
    const SubGV & subgv_;
    const Entity & e_;
    HierarchicIterator hit;
    HierarchicIterator ehit;
    bool noHierarchy;
  };


}
#endif
