// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_MC33TRIANGULATION_MERGER_HH
#define DUNE_MC33TRIANGULATION_MERGER_HH

#include <dune/udg/mc33triangulation/mc33triangulationimp.hh>

namespace MarchingCube33SubTriangulationImp {

  /** \brief Merger for MarchingCube33SubTriangulation class which
      does not perform any merging */
  class NoMergingMerger {

  public:

    //! Constructor
    NoMergingMerger() {}

    /** \brief Globally merge domain cells to fulfill the minimum
       volume criterion.
     */
    template<typename HostCellMap>
    void merge(HostCellMap & ){}

    template<typename Layers>
    void mergeAllLayers(Layers & layers){}

  };

  /** \brief Base class for mergers which merge according to the
      maximum intersection area criterion.
   */
  template< typename GV >
  class MaxAreaMergerBase {

  protected:
    // Index types
    typedef typename GV::IndexSet::IndexType Index;
    typedef std::list<Index> IndexList;

    // Continuous field type
    typedef typename GV::ctype DomainField;

    MaxAreaMergerBase( const GV & _gv) : gv_(_gv) {}

    // Merge the smaller of the two domain cells encapsuled in the
    // candidate pair to the bigger one and remove it from its
    // containing list.
    template< typename DomainCell, typename HostCellMap >
    void mergeAndRemove(DomainCell & a, DomainCell & b, HostCellMap & hostCellMap){

      // Find smaller domain cell
      DomainCell* smaller = &a;
      DomainCell* bigger  = &b;
      if(a.volume > b.volume) {
        smaller = &b;
        bigger  = &a;
      }

      // Merge cells
      if(!bigger->mergeConnected(*smaller))
      {DUNE_THROW(Dune::Exception,"At this point the merge must suceed!");}

      typename DomainCell::Index smallerIndex = gv_.indexSet().index(gv_.grid().entity(smaller->hostCell));
      hostCellMap[smallerIndex].remove(*smaller);
    }

    typedef typename std::pair<Index, DomainField> IndexAreaPair;
    static bool lessMapValue(const IndexAreaPair& a, const IndexAreaPair & b){
      if(a.second < b.second)
        return true;
      return false;
    }

    // Merge this domain cell to a neighbor domain cell
    template< typename HostCellMap, typename DomainCell >
    DomainCell* mergeWithNeighbors
      (HostCellMap & hostCellMap, DomainCell & domainCell)
    {
      // Eligible domain cells for merging
      typedef std::map<Index, DomainCell*> CandidateMap;
      typedef std::map<Index, DomainField> CandidateAreaMap;
      CandidateMap allCandidates;
      const bool ignoreForeignDomains = true;
      domainCell.neighbors(allCandidates,ignoreForeignDomains);
      CandidateAreaMap allAreas = domainCell.areas;

      // Extract the inner-domain intersection areas
      CandidateAreaMap allCandidatesAreas;
      for(typename CandidateMap::iterator ait= allCandidates.begin();
          ait != allCandidates.end(); ++ait) {
        allCandidatesAreas[ait->first] = allAreas[ait->first];
      }

      // If no sufficient candidate was found, merge to the cell with
      // biggest intersection area (if available).
      if(allCandidates.size()) {

        const Index maxIndex =
          std::max_element(allCandidatesAreas.begin(),allCandidatesAreas.end(),lessMapValue)->first;

        DomainCell* maxDomainCell = allCandidates[maxIndex];
        return maxDomainCell;
      }

      return (DomainCell*)(0);
    }

  protected:
    const GV & gv_;
  };

}
#endif
