// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_MC33TRIANGULATION_PARALLELMERGER_HH
#define DUNE_MC33TRIANGULATION_PARALLELMERGER_HH

namespace MarchingCube33SubTriangulationImp {
  namespace ParallelCompatibilityMergerImp {

    template<class GV>
    class CellIndexDataHandle :
      public Dune::CommDataHandleIF<
          CellIndexDataHandle<GV>,
          typename GV::IndexSet::IndexType
          >
    {
    public:
      typedef typename GV::IndexSet::IndexType Index;
      typedef std::vector< std::map<Index,Index> >
      Indices;

      CellIndexDataHandle (const GV & gv, Indices & indices)
        : gv_(gv), indices_(indices)
      {}

      //! \brief returns true if data for this codim should be
      //! communicated
      bool contains (int dim, int codim) const
      {
        return dim==GV::dimension && codim == 0;
      }

      //! \brief returns true if size per entity of given dim and codim
      //! is a constant
      bool fixedsize (int dim, int codim) const
      {
        return true;
      }

      /*! \brief how many objects of type DataType have to be sent for a given entity

         Note: Only the sender side needs to know this size.
       */
      template<class EntityType>
      size_t size (EntityType& e) const
      {
        return 2;
      }

      //! \brief pack data from user to message buffer
      template<class MessageBuffer, class EntityType>
      void gather (MessageBuffer& buff, const EntityType& e) const
      {
        const Index rank(gv_.comm().rank());
        buff.write(rank);
        const typename GV::IndexSet::IndexType index =
          gv_.indexSet().index(e);
        buff.write(index);
      }

      /*! \brief unpack data from message buffer to user

         n is the number of objects sent by the sender
       */
      template<class MessageBuffer, class EntityType>
      void scatter (MessageBuffer& buff, const EntityType& e, size_t n)
      {
        assert(n==2);
        typename GV::IndexSet::IndexType rank;
        buff.read(rank);

        typename GV::IndexSet::IndexType remoteIndex;
        buff.read(remoteIndex);

        const typename GV::IndexSet::IndexType index =
          gv_.indexSet().index(e);

        indices_[remoteIndex].insert(std::make_pair(rank,index));
      }

    private:
      const GV & gv_;
      Indices & indices_;
    };

    template<class GV, class M>
    class BurnmarkMapsDataHandle :
      public Dune::CommDataHandleIF<
          BurnmarkMapsDataHandle<GV,M>,
          typename GV::IndexSet::IndexType
          >
    {
      // Data structure to be communicated
      typedef M BurnmarkMaps;
      typedef typename M::value_type BurnmarkMap;

      typedef typename GV::IndexSet::IndexType Index;
      typedef std::pair<Index,Index> IndexPair;
      typedef std::vector< std::map<Index,Index> > Indices;

    public:

      BurnmarkMapsDataHandle
        (const GV & gv,
        const Indices & lgvRemoteIndices,
        const BurnmarkMaps & burnmarkMaps,
        BurnmarkMaps & remoteBurnmarkMaps)
        : gv_(gv),
          lgvRemoteIndices_(lgvRemoteIndices),
          burnmarkMaps_(burnmarkMaps),
          remoteBurnmarkMaps_(remoteBurnmarkMaps)
      {}

      //! \brief returns true if data for this codim should be
      //! communicated
      bool contains (int dim, int codim) const
      {
        return dim==GV::dimension && codim == 0;
      }

      //! \brief returns true if size per entity of given dim and codim
      //! is a constant
      bool fixedsize (int dim, int codim) const
      {
        return false;
      }

      /*! \brief how many objects of type DataType have to be sent for a given entity

         Note: Only the sender side needs to know this size.
       */
      template<class EntityType>
      size_t size (EntityType& e) const
      {
        const Index cellIndex = gv_.indexSet().index(e);
        const Index nParts = burnmarkMaps_[cellIndex].size();
        if(nParts)
          return 3*nParts+1;
        else
          return 0;
      }

      //! \brief pack data from user to message buffer
      template<class MessageBuffer, class EntityType>
      void gather (MessageBuffer& buff, const EntityType& e) const
      {
        const Index cellIndex = gv_.indexSet().index(e);

        typename BurnmarkMap::const_iterator pit =
          burnmarkMaps_[cellIndex].begin();
        const typename BurnmarkMap::const_iterator epit =
          burnmarkMaps_[cellIndex].end();

        // Write rank if there is any data
        if(pit!=epit) {
          Index rank(gv_.comm().rank());
          buff.write(rank);
        }

        for(; pit != epit; ++pit) {
          buff.write(pit->first.first);
          buff.write(pit->first.second);
          buff.write(pit->second.second);
        }
      }

      /*! \brief unpack data from message buffer to user

         n is the number of objects sent by the sender
       */
      template<class MessageBuffer, class EntityType>
      void scatter (MessageBuffer& buff, const EntityType& e, size_t n)
      {
        if(n==0) return;

        // Determine index in processor grid
        const Index cellIndex = gv_.indexSet().index(e);

        // Determine number of parts to communicate
        assert((n-1)%3==0);
        const Index nParts = n/3;

        // Read first buffer entry holding the rank if there is any data
        Index rank; buff.read(rank);

        for(size_t p=0; p<nParts; ++p) {

          // Read buffer entries for this part
          Index lCellIndex; buff.read(lCellIndex);
          Index partIndex; buff.read(partIndex);
          Index domainCellId; buff.read(domainCellId);

          // Convert the index of the lCell in remote grid view to the
          // index in this grid view
          typename std::map<Index,Index>::const_iterator lCellRemoteIndexIt =
            lgvRemoteIndices_[lCellIndex].find(rank);
          assert(lCellRemoteIndexIt != lgvRemoteIndices_[lCellIndex].end());
          lCellIndex = lCellRemoteIndexIt->second;

          // Insert entry in burnmark map
          const IndexPair key = std::make_pair(lCellIndex,partIndex);
          const IndexPair value = std::make_pair(Index(gv_.comm().rank()),domainCellId);
#ifndef NDEBUG
          std::pair<typename BurnmarkMap::iterator,bool>
          insertionRes =
#endif
          remoteBurnmarkMaps_[cellIndex].insert
            (std::make_pair(key,value));

          // As domain cells are a real partition this should
          // always insert.
          assert(insertionRes.second);
        } // p

      }

    private:
      const GV & gv_;
      const Indices & lgvRemoteIndices_;
      const BurnmarkMaps & burnmarkMaps_;
      BurnmarkMaps & remoteBurnmarkMaps_;
    };
  }

  template<typename GV, typename LGV>
  class ParallelCompatibilityMerger :
    public PDELabCompatibilityMerger<GV>
  {

  private:
    // Base class
    typedef PDELabCompatibilityMerger<GV> Base;

    // Index types
    typedef typename GV::IndexSet::IndexType Index;
    typedef std::list<Index> IndexList;
    typedef std::pair<Index,Index> IndexPair;

    // Continuous field type
    typedef typename GV::ctype DomainField;

    // Grid view types
    typedef typename GV::Traits::template Codim<0>::Entity CellEntity;

    // Map a <lCellIndex,partIndex> pair to a domain cell index
    typedef std::map<IndexPair,IndexPair> BurnmarkMap;
    // Map a cell index in gv to a BurnmarkMap
    typedef std::vector<BurnmarkMap> BurnmarkMaps;

    // Find and store indices of cells at processor boundary
    void findProcessorBoundaryCells(std::vector<bool> & processorCellIndices)
    {
      processorCellIndices.resize(gv_.size(0),false);

      typedef typename GV::Traits::template Codim<0>::Iterator ElementIterator;
      typedef typename GV::IntersectionIterator IntersectionIterator;
      typedef typename IntersectionIterator::Intersection Intersection;

      // Iterate grid cells and intersections
      for (ElementIterator it = gv_.template begin<0>();
           it!=gv_.template end<0>(); ++it) {

        for (IntersectionIterator iit = gv_.ibegin(*it);
             iit!=gv_.iend(*it); ++iit) {

          // Check if processor boundary
          if ((!iit->boundary()) && (!iit->neighbor())) {
            const Index cellIndex = gv_.indexSet().index(*it);
            processorCellIndices[cellIndex] = true;
          }
        } // iit

      } // it
    }

    // Find and store pointers to domain cells
    template< typename HostCellMap, typename DomainCellPointerList >
    void findProcessorDomainCells
      (HostCellMap & hostCellMap,
      const std::vector<bool> & processorCellIndices,
      DomainCellPointerList & processorDomainCells)
    {
      // Iterate host cell lists
      typedef typename HostCellMap::value_type DomainCellList;
      typedef typename DomainCellList::value_type DomainCell;
      for(Index c=0; c<hostCellMap.size(); ++c) {

        // Iterate domain cells with this host
        DomainCellList & domainCells = hostCellMap[c];
        typedef typename DomainCellList::iterator Iterator;
        Iterator it = domainCells.begin();
        const Iterator eit = domainCells.end();
        for(; it!=eit; ++it) {
          DomainCell & domainCell = *it;

          // Iterate all home cells in gv
          typename DomainCell::MemberCellMap::const_iterator mit =
            domainCell.gvMembers.begin();
          typename DomainCell::MemberCellMap::const_iterator emit =
            domainCell.gvMembers.end();
          for(; mit != emit; ++mit) {
            const Index homeCellIndex = mit->first;
            if(processorCellIndices[homeCellIndex]) {
              processorDomainCells.push_back(&domainCell);
              break;
            }
          } // mit

        } // it

      } // c
    }

    // Store the burnmarks for
    template< typename DomainCellPointerList >
    void accumulateBurnmarkMap
      (const std::vector<bool> & processorCellIndices,
      DomainCellPointerList & processorDomainCells,
      BurnmarkMaps & burnmarkMaps)
    {
      // Iterate processor domain cells
      typename DomainCellPointerList::const_iterator
      domainCellPtIt = processorDomainCells.begin();
      const typename DomainCellPointerList::const_iterator
      domainCellPtEit = processorDomainCells.end();
      for(; domainCellPtIt != domainCellPtEit; ++domainCellPtIt) {

        // Extract domain cell
        typedef typename DomainCellPointerList::value_type
        DomainCellPt;
        typedef typename Dune::TypeTraits<DomainCellPt>::PointeeType
        DomainCell;
        const DomainCellPt domainCell = *domainCellPtIt;

        // Iterate all home cells in gv
        typename DomainCell::MemberCellMap::const_iterator mit =
          domainCell->gvMembers.begin();
        typename DomainCell::MemberCellMap::const_iterator emit =
          domainCell->gvMembers.end();
        for(; mit != emit; ++mit) {

          // Skip all non-processor cells
          const Index homeCellIndex = mit->first;
          if(!processorCellIndices[homeCellIndex])
            continue;

          // Iterate lgv cells contained in home cell
          SubGridViewIterator<GV,LGV> hit(gv_,lgv_,*(mit->second));
          for(; hit.valid(); ++hit) {
            const Index lCellIndex = lgv_.indexSet().index(*hit);

            // Search for entity parts of this lCell in domain cell
            typename DomainCell::LGVMemberMap::const_iterator
            partIndexListIt = domainCell->lgvMembers.find(lCellIndex);

            // Domain cell might not contain this cell in lgv
            if(partIndexListIt == domainCell->lgvMembers.end()) continue;

            // Iterate part indices in this lCell and write domain
            // index map
            const IndexList & partIndices = partIndexListIt->second;
            for(typename IndexList::const_iterator partIndexIt =
                  partIndices.begin(); partIndexIt != partIndices.end();
                ++partIndexIt) {

              const IndexPair key = std::make_pair(lCellIndex,*partIndexIt);
              const IndexPair value =
                std::make_pair(Index(gv_.comm().rank()),domainCell->id);
#ifndef NDEBUG
              std::pair<typename BurnmarkMap::iterator,bool>
              insertionRes =
#endif
              burnmarkMaps[homeCellIndex].insert
                (std::make_pair(key,value));
              // As domain cells are a real partition this should
              // always insert.
              assert(insertionRes.second);

            } // partIndexIt
          } // hit

        } // mit

      }
    }


  public:

    //! Constructor
    ParallelCompatibilityMerger
      (const GV & _gv, const LGV & _lgv)
      : Base(_gv), gv_(_gv), lgv_(_lgv),
        invalidIndex_(std::numeric_limits<Index>::max())
    {}

    template<typename Layers>
    void mergeAllLayers(Layers & layers)
    {}

    /** \brief Globally merge domain cells to fulfill the minimum
        volume criterion.
     */
    template<typename HostCellMap>
    void merge(HostCellMap & hostCellMap)
    {
      // Write processor boundary map (only if it is actually
      // parallel)
      const bool isParallel = gv_.comm().size()>1;

      // Do nothing if domain not decomposed
      if(!isParallel) return;

      // Indicate for each cell index in gv_ whether the correwsponding
      // cell is on the processor boundary
      std::vector<bool> processorCellIndices;

      // Find and store cell indices on processor boundary
      findProcessorBoundaryCells(processorCellIndices);

      // List of all domain cells which have a home cell on the
      // processor boundary
      typedef typename HostCellMap::value_type DomainCellList;
      typedef typename DomainCellList::value_type DomainCell;
      std::list<DomainCell*> processorDomainCells;
      findProcessorDomainCells
        (hostCellMap,processorCellIndices,processorDomainCells);

      // Determine max gv size
      Index procGVSize( gv_.size(0) );
      Index maxGVSize = gv_.comm().max(procGVSize);

      using namespace ParallelCompatibilityMergerImp;

      // Communicate cell indices in gv_
      typedef std::vector< std::map<Index,Index> > Indices;
      Indices gvRemoteIndices(maxGVSize);
      CellIndexDataHandle<GV> dh(gv_,gvRemoteIndices);
      gv_.communicate(dh,Dune::All_All_Interface,
                      Dune::ForwardCommunication);

      // Determine max lgv size
      Index procLGVSize( lgv_.size(0) );
      Index maxLGVSize = lgv_.comm().max(procLGVSize);

      // Communicate cell indices in lgv_
      Indices lgvRemoteIndices(maxLGVSize);
      CellIndexDataHandle<LGV> ldh(lgv_,lgvRemoteIndices);
      lgv_.communicate(ldh,Dune::All_All_Interface,
                       Dune::ForwardCommunication);

      // Accumulate burn mark information about entity parts in
      // processor home cells
      BurnmarkMaps burnmarkMaps(gv_.size(0));
      accumulateBurnmarkMap
        (processorCellIndices,
        processorDomainCells,
        burnmarkMaps);

      // Generate copy of map for communication
      BurnmarkMaps remoteBurnmarkMaps(gv_.size(0));
      BurnmarkMapsDataHandle<GV,BurnmarkMaps>
      bdh(gv_,lgvRemoteIndices,
          burnmarkMaps,remoteBurnmarkMaps);
      gv_.communicate(bdh,Dune::All_All_Interface,
                      Dune::ForwardCommunication);

    }

  private:
    const GV & gv_; // Current fundamental grid view
    const LGV & lgv_; // Current geometry grid view
    const Index invalidIndex_;  // Index identifying invalid values

  };


}
#endif
