// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_MC33TRIANGULATION_PARTS_HH
#define DUNE_MC33TRIANGULATION_PARTS_HH

namespace Dune {
  namespace UDG {

    /** \brief Multi layer entity part containing an hierarchy of
        entity parts such that each entity part contains all entity
        parts on higher layers.
     */
    template< typename EP >
    class MultiLayerEntityPart :
      public EntityPartInterface<typename EP::Traits::GridView>
    {
    public:

      //! Type traits
      typedef typename EP::Traits Traits;

      //! Intersection part pointer
      typedef Dune::shared_ptr<EP> ImpPointer;

      //! Geometry engine
      typedef typename EP::GeometryImp GeometryImp;

      //! Constructor
      MultiLayerEntityPart(const int nLayers)
        : imps_(nLayers), current_(0), layer_(0)
      {}

      //! Insert intersection part into the current layer
      void insertLayer(const size_t layer, ImpPointer & epp)
      {
        imps_[layer] = epp;
        current_ = imps_[layer].get();
        layer_=layer;
      }

      //! Return the current layer index
      size_t currentLayer() const
      { return layer_; }

      //! Set to the intersection part in the current layer
      void setLayer(const size_t layer) const
      {
        layer_=layer;
        assert(layer >= 0 && layer < imps_.size());
        current_ = imps_[layer].get();
      }

      //! Set to the intersection part in the max layer
      size_t resetLayer() const
      {
        current_ = imps_.back().get();
        layer_ = imps_.size()-1;
        return layer_;
      }

      /** \brief Access the geometry mapping from the local domain of
         the entity part to the local domain of the bounding box. */
      virtual const typename Traits::Geometry& geometryInCell () const
      { return current_->geometryInCell(); }

      /** \brief Access the geometry mapping from the local domain of
         the entity part to the local domain of the bounding box. */
      virtual const Dune::shared_ptr<typename Traits::Geometry>
      geometryInCellPointer () const
      { return current_->geometryInCellPointer(); }

      //! Access the global geometry
      virtual const typename Traits::Geometry& geometry () const
      { return current_->geometry(); }

      //! Access the entity part's home entity in level set mesh or in
      //! fundamental mesh if the entity part was constructed from a whole
      //! fundamental mesh cell
      virtual const typename Traits::Entity & entity() const
      { return current_->entity(); }

      //! Access the bounding box
      virtual const typename Traits::BoundingBox & boundingBox() const
      { return current_->boundingBox(); }

      /** \brief Legacy function to be removed */
      virtual int domainTag () const
      { return current_->domainTag(); }

      /** \brief Domain index of this entity part */
      virtual int domainIndex () const
      { return current_->domainIndex(); }

      /** \brief Index of this part within the entity */
      virtual int partIndex() const
      { return current_->partIndex(); }

      /** \brief Sign vector indicating the placement of this
          entity part relative to the interfaces separating
          the sub triangulation's domains.

          For each interface of the sub triangulation, this sign
          vector has one entry holding an element of {-1,0,1},
          where 0 indicates that this entity part is intersected
          by the corresponding interface.
       */
      virtual const typename Traits::DomainTag & newDomainTag () const
      { return current_->newDomainTag(); }

      /** \brief Volume of the domain cell */
      virtual typename Traits::DomainField volume() const
      { return current_->volume(); }

    private:
      std::vector<ImpPointer> imps_;
      mutable EP * current_;
      mutable size_t layer_;
    };



    template< typename IP >
    class MultiLayerIntersectionPart :
      public IntersectionPartInterface<typename IP::Traits::GridView>
    {
    public:

      //! Type traits
      typedef typename IP::Traits Traits;

      //! Intersection part pointer
      typedef Dune::shared_ptr<IP> ImpPointer;

      //! Geometry engine
      typedef typename IP::FaceGeometryImp FaceGeometryImp;

      //! Constructor
      MultiLayerIntersectionPart(const int nLayers)
        : imps_(nLayers), current_(0), layer_(0)
      {}

      //! Insert intersection part into the current layer
      void insertLayer(const size_t layer, ImpPointer & ipp)
      {
        imps_[layer] = ipp;
        current_ = imps_[layer].get();
        layer_=layer;
      }

      //! Return the current layer index
      size_t currentLayer() const
      { return layer_; }

      //! Set to the intersection part in the current layer
      void setLayer(const size_t layer) const
      {
        layer_=layer;
        assert(layer >= 0 && layer < imps_.size());
        current_ = imps_[layer].get();
      }

      //! Set to the intersection part in the max layer
      size_t resetLayer() const
      {
        current_ = imps_.back().get();
        layer_ = imps_.size()-1;
        return layer_;
      }

      //! Access the host entity on the outside of intersection part
      //! \note This is not comparable to method entity() of entity part
      virtual const typename Traits::Entity & outside() const
      { return current_->outside(); }

      //! Access the host entity on the inside of intersection part
      //! \note This is not comparable to method entity() of entity part
      virtual const typename Traits::Entity & inside() const
      { return current_->inside(); }

      /** \brief Access the geometry mapping from the local
         intersection domain to the local domain of the bounding box on
         inside of this intersection part. */
      virtual const typename Traits::FaceGeometry& geometryInInside () const
      { return current_->geometryInInside(); }

      /** \brief Access the geometry mapping from the local
         intersection domain to the local domain of the bounding box on
         outside of this intersection part. */
      virtual const typename Traits::FaceGeometry& geometryInOutside () const
      { return current_->geometryInOutside(); }

      //! Access the global geometry
      virtual const typename Traits::FaceGeometry& geometry () const
      { return current_->geometry(); }

      //! Access the bounding box on the inside
      virtual const typename Traits::BoundingBox & boundingBoxInside() const
      { return current_->boundingBoxInside(); }

      //! Access the bounding box on the outside
      virtual const typename Traits::BoundingBox & boundingBoxOutside() const
      { return current_->boundingBoxOutside(); }

      //! Return unit outer normal at local face coordinate
      virtual typename Traits::GlobalCoordinate unitOuterNormal
        (const typename Traits::LocalCoordinate x) const
      { return current_->unitOuterNormal(x); }

      /** \brief Accumulated area of all the intersection parts
          separating the two domain cell patches on both sides of this
          intersection part.
       */
      virtual typename Traits::DomainField area() const
      { return current_->area(); }

      /** \brief Volume of the domain cell on the inside of this
          intersection part */
      virtual typename Traits::DomainField insideVolume() const
      { return current_->insideVolume(); }

      /** \brief Volume of the domain cell on the outside of this
          intersection part */
      virtual typename Traits::DomainField outsideVolume() const
      { return current_->outsideVolume(); }

      /** \brief Sign vector indicating the placement of this
          intersection part relative to the interfaces separating
          the sub triangulation's domains.

          For each interface of the sub triangulation, this sign
          vector has one entry holding an element of {-1,0,1},
          where 0 indicates that this intersection part is directly
          on the corresponding interface.
       */
      virtual const typename Traits::DomainTag & newDomainTag () const
      { return current_->newDomainTag(); }

      /** \brief Legacy function to be removed */
      virtual int domainTag () const
      { return current_->domainTag(); }

      /** \brief Domain index on the inside of this intersection part */
      virtual int insideDomainIndex () const
      { return current_->insideDomainIndex(); }
      /** \brief Domain index on the outside of this intersection part */
      virtual int outsideDomainIndex () const
      { return current_->outsideDomainIndex(); }

      /** \brief Indicates whether this intersection part is on a
          domain boundary. */
      virtual bool boundary() const
      { return current_->boundary(); }
      /** \brief Indicates whether this intersection part is on an
           internal boundary. */
      virtual bool internalBoundary() const
      { return current_->internalBoundary(); }

      /** \brief Indicates whether this intersection part has a domain
          cell patch neighbor on the outside. */
      virtual bool neighbor() const
      { return current_->neighbor(); }

    private:
      std::vector<ImpPointer> imps_;
      mutable IP * current_;
      mutable size_t layer_;
    };

  }
}
#endif
