// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_MC33TRIANGULATION_PDELABMERGER_HH
#define DUNE_MC33TRIANGULATION_PDELABMERGER_HH

namespace MarchingCube33SubTriangulationImp {

  template<typename GV>
  class PDELabCompatibilityMerger
  {

  private:

    // Index types
    typedef typename GV::IndexSet::IndexType Index;
    typedef std::list<Index> IndexList;

    // Continuous field type
    typedef typename GV::ctype DomainField;

    // Grid view types
    typedef typename GV::Traits::template Codim<0>::Entity CellEntity;

    // Check whether the given list contains domain cells which are
    // too small
    template<typename DomainCellList, typename IndexMap>
    bool containsDoubletteCells
      (const DomainCellList & domainCells, IndexMap & doublettes) const
    {
      bool found = false;
      typedef typename DomainCellList::const_iterator Iterator;
      Iterator it = domainCells.begin();
      for(; it!=domainCells.end(); ++it) {

        typedef std::pair<typename IndexMap::iterator,bool> Result;
        const Result res = doublettes.insert(std::make_pair(it->domainIndex,1));
        if(!res.second) {
          found = true;
          res.first->second += 1;
        }
      }
      return found;
    }

    template<typename DomainCellList>
    bool containsDomain
      (const DomainCellList & domainCells, Index domainIndex) const
    {
      typedef typename DomainCellList::const_iterator Iterator;
      Iterator it = domainCells.begin();
      for(; it!=domainCells.end(); ++it) {
        if(it->domainIndex==domainIndex)
          return true;
      }
      return false;
    }


    template< typename DomainCellList >
    void forceMerge(DomainCellList & domainCells, const Index domainIndex)
    {
      typedef typename DomainCellList::iterator Iterator;
      Iterator it = domainCells.begin();
      Iterator first = domainCells.end();
      for(; it!=domainCells.end(); ++it) {

        // Only process domain cells of specified domain index
        if(it->domainIndex != domainIndex)
          continue;

        if(first == domainCells.end()) {
          first = it;
          continue;
        }

        it->blindMerge(*first);
        domainCells.erase(first);
        first = it;
      }
    }

    // For the given domain cell try to find a home cell which is not
    // yet host cell to a domain cell of the same domain index. If
    // such a cell is found, then set the host cell of this domain
    // cell to the respective home cell and copy this domain cell to
    // the appropriate position in the host cell map
    template< typename HostCellMap, typename DomainCellList, typename DomainCellIt >
    bool findFreeHostAmongHomeCells
      (const Index domainIndex, HostCellMap & hostCellMap,
      DomainCellList & domainCells, DomainCellIt & domainCellIt)
    {
      typedef typename DomainCellList::value_type DomainCell;
      typedef typename DomainCell::MemberCellMap CellMap;
      typedef typename CellMap::iterator CellMapIt;

      // Iterate all home cells of this domain cell
      CellMapIt cmit = domainCellIt->gvMembers.begin();
      for(; cmit != domainCellIt->gvMembers.end(); ++cmit) {

        // Find index of home cell in grid view
        const Index memberIndex = gv_.indexSet().index(*(cmit->second));

        // Check whether this home cell is already host cell to a
        // domain cell of given domain index.
        if(!containsDomain(hostCellMap[memberIndex],domainIndex)) {

          // Update host cell pointer
          domainCellIt->hostCell = cmit->second;

          // Copy to appropriate list in host cell map
          DomainCellList & dest = hostCellMap[memberIndex];
          dest.splice(dest.begin(),domainCells,domainCellIt);
          return true;
        }
      }
      return false;
    }

  protected:

    // Enforce the PDELab compatibility condition by locally merging
    // all domain cells with the same host cell and the same domain
    // index even if this leads to domain cells without inner
    // connectivity.
    template<typename HostCellMap>
    void enforceCompatibilityLocally
      (HostCellMap & hostCellMap)
    {
      // Collect the small domain cells
      for(Index c=0; c<hostCellMap.size(); ++c) {

        // Check for and store host cells that host more than one
        // domain cell of the same domain index.
        typedef std::map<Index,Index> IndexMap;
        IndexMap doublettes;
        if(containsDoubletteCells(hostCellMap[c],doublettes)) {

          typedef typename HostCellMap::value_type DomainCellList;
          DomainCellList & domainCells = hostCellMap[c];

          // Check whether doublettes still exist.
          typedef typename IndexMap::iterator IndexMapIt;
          IndexMapIt imit = doublettes.begin();
          for(; imit != doublettes.end(); ++imit) {
            // For each remaining doublette, force merging of the
            // domain cells of same domain index.
            if(imit->second > 1) {
              forceMerge(domainCells,imit->first);
              std::cout << "Forcing merge of disconnected domain cells "
              "with host cell index " << c << std::endl;
            }
          } // imit
        } // if doublettes
      } // c
    }

    // Try to redistribute domain cells among the available host cells
    // such that each host cell hosts only one domain cell of the same
    // domain index. This function does not give any guarantees with
    // regard to whether or not this aim will be achieved.
    template<typename HostCellMap>
    bool redistributeHostCells
      (HostCellMap & hostCellMap)
    {
      bool success = true;

      // Iterate host cell indices
      for(Index c=0; c<hostCellMap.size(); ++c) {

        // Check for and store host cells that host more than one
        // domain cell of the same domain index.
        typedef std::map<Index,Index> IndexMap;  // domainIndex --> number of domain cells which have this domainIndex
        IndexMap doublettes;
        if(containsDoubletteCells(hostCellMap[c],doublettes)) {

          typedef typename HostCellMap::value_type DomainCellList;
          DomainCellList & domainCells = hostCellMap[c];

          // Try to find a home cell suitable to be a host
          typename DomainCellList::iterator dcit = domainCells.begin();
          while(dcit != domainCells.end()) {

            const Index domainIndex = dcit->domainIndex;
            assert(doublettes.find(domainIndex) != doublettes.end());
            assert(doublettes[domainIndex] >=1);
            if(doublettes[domainIndex] > 1) {

              // Search for a new host among the home cells of the
              // domain cell.
              const bool foundNewHost =
                findFreeHostAmongHomeCells
                  (domainIndex,hostCellMap,domainCells,dcit);

              if(foundNewHost) {
                doublettes[domainIndex]--;
                dcit = domainCells.begin();
                continue;
              }
            }
            dcit++;
          } // dcit

          // Check whether condition was fulfilled
          doublettes.clear();
          success &= !containsDoubletteCells(hostCellMap[c],doublettes);

        } // contains doublettes
      } // c
      return success;
    }

  public:

    //! Constructor
    PDELabCompatibilityMerger(const GV & _gv) : gv_(_gv)
    {}


    template<typename Layers>
    void mergeAllLayers(Layers & layers)
    {}

    /** \brief Globally merge domain cells to fulfill the minimum
        volume criterion.
     */
    template<typename HostCellMap>
    void merge(HostCellMap & hostCellMap)
    {
      // First try to redistribute the domain cells which violate the
      // pdelab condition among their home cells.
      redistributeHostCells(hostCellMap);
      // Then force local merging of the domain cells which could not
      // be redistributed successfully.
      enforceCompatibilityLocally(hostCellMap);
    }

  private:
    const GV & gv_; // Current fundamental grid view
  };


}
#endif
