// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_MC33TRIANGULATION_TRIANGULATIONLAYER_HH
#define DUNE_MC33TRIANGULATION_TRIANGULATIONLAYER_HH


namespace Dune
{
  namespace UDG
  {

    /**
     * \brief Class representing two maps which associate each domain
     *        refinement part of a level set mesh cell, respectively
     *        each fundamental mesh cell, with domain cells.
     *
     * Map 1) associates each (index in lgv, part index) pair, i.e. each
     * domain refinement part of a level set mesh cell, with the domain
     * cell which contains the part and domain cells adjacent to the part.
     *
     * Map 2) associates indices of fundamental mesh cells, whose children
     * in the level set mesh grid view contribute domain refinement parts
     * to exactly one domain cell, with this domain cell and its adjacent
     * domain cells which have a common face on the fundamental mesh
     * skeleton.
     *
     * \tparam DomainCell Type of a domain cell.
     */
    template< typename DomainCell >
    class SubTriangulationLayerLookupMap
    {
    public:
      //! Index types
      typedef typename DomainCell::Index Index;
      typedef std::pair<Index,Index> IndexPair;

      //! Type containing the association of
      //!   1) a domain refinement part of a level set mesh cell
      //! respectively
      //!   2) a fundamental mesh cell, whose children in the level set mesh
      //!      grid view contribute domain refinement parts to a domain cell,
      //! with domain cells
      struct LookupItem
      {
        LookupItem() : insideDomainCell(0) {}

        //! Store domain cell which contains
        //!   1) the level set mesh cell's domain refinement part
        //! respectively
        //!   2) domain refinement parts of the fundamental mesh cell's
        //!      children in the level set mesh grid view
        const DomainCell* insideDomainCell;

        //! Store a map which associates
        //!   1) a domain cell face's local index in the reference element
        //!      of the refinement part for which it was created
        //! respectively
        //!   2) the index of a domain cell face's neighbor fundamental mesh
        //!      cell (i.e. the index of the fundamental mesh cell which is
        //!      parent of the level set mesh cell which owns the domain cell
        //!      face's corresponding domain refinement part on the outside)
        //! with the adjacent domain cell
        typedef std::map<Index,const DomainCell*> FacesLookupMap;
        FacesLookupMap facesLookup;
      };

      //! Type of a list of domain cells
      typedef std::list<DomainCell> DomainCellList;

    private:
      //! Type of the map with (index in lgv, part index) access, i.e. map 1)
      typedef std::map<IndexPair,LookupItem> PairLookupMapImp;
      //! Type of a map with [index in gv][domainCell.id] access, used for map 2)
      typedef std::map<Index,LookupItem> IndexLookupMapImp;
      typedef std::vector<IndexLookupMapImp> CellMap;
      //! Type of field used by domain cell type.
      typedef typename DomainCell::DomainField DomainField;

    public:
      /**
       * \brief This function generates a mapping from the map
       *        with (index in lgv, part index) access, which
       *        itself associates (index in lgv, part index)
       *        pairs with the indices of the host cells in
       *        the fundamental mesh grid view.
       */
      void generateHostCellIndexMap
        (const typename DomainCell::GridView & gv,
        std::map<IndexPair,Index> & hostCellIndexMap,
        const Index projectionsOnPartIndex = 0) const
      {
        typename PairLookupMapImp::const_reverse_iterator it = lookupMap_.rbegin();
        const typename PairLookupMapImp::const_reverse_iterator eit = lookupMap_.rend();
        for(; it != eit; ++it) {
          const Index lCellIndex = it->first.first;
          Index partIndex = it->first.second;

          for(Index l=0; l<projectionsOnPartIndex; ++l)
            partIndex /= DomainRefinementImp::maxPartsInLayer;

          const Index hostCellIndex =
            gv.indexSet().index(*(it->second.insideDomainCell->hostCell));

          const IndexPair key = std::make_pair(lCellIndex,partIndex);
          hostCellIndexMap.insert(hostCellIndexMap.begin(),std::make_pair(key,hostCellIndex));
        }
      }

      /**
       * \brief Evaluate map 2) which associates indices of fundamental
       *        mesh cells, whose children in the level set mesh grid view
       *        contribute domain refinement parts to exactly one domain
       *        cell, with this domain cell and its adjacent domain cells
       *        which have a common face on the fundamental mesh skeleton.
       *
       * Such fundamental mesh cells are for example those which can be
       * identified with domain cells as they are not intersected by domain
       * interfaces, i.e. those which will make up one entity part in the
       * sub triangulation.
       *
       * \return Pointer to an object holding the fundamental mesh cell's
       *         association with domain cells. This is the null pointer for
       *         fundamental mesh cells that are not contained in any domain
       *         cell or in multiple domain cells.
       */
      const LookupItem* find
        (const Index cellIndex) const
      {
        const IndexLookupMapImp & cellLookupMap = cellMap_[cellIndex];
        if(cellLookupMap.size()!=1) return 0;
        return &(cellLookupMap.begin()->second);
      }

      /**
       * \brief Evaluate map 1) which associates (index in lgv, part index)
       *        pairs, i.e. domain refinement parts of a level set mesh cell,
       *        with the domain cell which contains the part and domain cells
       *        adjacent to the part.
       *
       * \return Pointer to an object holding the domain refinement part's
       *         association with domain cells. This is the null pointer for
       *         invalid/unassociated pairs.
       *
       * \todo Can cellIndex be removed?
       */
      const LookupItem* find
        (const Index cellIndex, const Index lCellIndex, const Index partIndex) const
      {
        typename PairLookupMapImp::const_iterator itemIt =
          lookupMap_.find(std::make_pair(lCellIndex,partIndex));

        if(itemIt == lookupMap_.end())
          return 0;

        return &(itemIt->second);
      }

      /**
       * \brief Evaluate map 1) which associates (index in lgv, part index)
       *        pairs, i.e. domain refinement parts of a level set mesh cell,
       *        with the domain cell which contains the part and domain cells
       *        adjacent to the part.
       *
       * \return Pointer to an object holding the domain refinement part's
       *         association with domain cells. This is the null pointer for
       *         invalid/unassociated pairs.
       *
       * \todo Can cellIndex be removed?
       */
      const LookupItem* find(const Index cellIndex, const IndexPair & pair) const
      {
        return find(cellIndex,pair.first,pair.second);
      }

      /**
       * \brief Given a map mapping each fundamental mesh cell via its index in
       *        gv to the list of domain cells which are (by merging) associated
       *        with the fundamental mesh cell, this function generates both
       *        maps that are represented by this class.
       *
       * \note This function should be called before any other interface function.
       *
       * \todo Shouldn't we count the number of domain cells by
       *       nDomainCells += domainCells.size()?
       */
      void write
        (const std::vector<DomainCellList> & hostCellMap, const bool verbose = true)
      {
        lookupMap_.clear();
        cellMap_.clear();
        cellMap_.resize(hostCellMap.size());

        DomainField maxVolume(std::numeric_limits<DomainField>::min());
        DomainField minVolume(std::numeric_limits<DomainField>::max());
        nDomainCells_ = 0;

        // Iterate all fundamental mesh cells via their indices in gv,
        // i.e. iterate all potential host cells
        for(Index i=0; i<hostCellMap.size(); ++i) {

          // Get list of domain cells associated with current host cell
          const DomainCellList & domainCells = hostCellMap[i];

          // Count number of domain cells
          if(!domainCells.empty())
            nDomainCells_++;

          // Store part of map with (index in lgv, part index) access
          // which is associated with current host cell
          PairLookupMapImp localLookupMap;

          // Iterate list of domain cells associated with current host cell
          typename DomainCellList::const_iterator domainCellIt = domainCells.begin();
          for(; domainCellIt != domainCells.end(); ++domainCellIt) {

            maxVolume = std::max(maxVolume,domainCellIt->volume);
            minVolume = std::min(minVolume,domainCellIt->volume);

            const DomainCell & domainCell = *domainCellIt;

            // Iterate all fundamental mesh cells which are part of current domain cell
            typedef typename DomainCell::MemberCellMap MemberCellMap;
            typename MemberCellMap::const_iterator cellIt = domainCellIt->gvMembers.begin();
            for(; cellIt != domainCellIt->gvMembers.end(); ++cellIt) {
              const Index cellIndex = cellIt->first;

              // Find lookup item for current domain cell
              IndexLookupMapImp & cellLookupMap = cellMap_[cellIndex];
              typename IndexLookupMapImp::iterator lookupItemIt =
                cellLookupMap.find(domainCell.id);

              // If entry for current domain cell exists, then skip...
              if(lookupItemIt != cellLookupMap.end()) continue;

              // Create new lookup item
              const std::pair<Index,LookupItem> & entry =
                std::make_pair(domainCell.id,LookupItem());
              lookupItemIt = cellLookupMap.insert(entry).first;
              LookupItem & item = lookupItemIt->second;
              item.insideDomainCell = &domainCell;
            }

            // Iterate all face lists of domain cell
            typedef typename DomainCell::DomainCellFaceListMap DomainCellFaceListMap;
            const DomainCellFaceListMap & faceListMap = domainCell.faceListMap;
            typedef typename DomainCellFaceListMap::const_iterator DomainCellFaceListTupleIt;
            DomainCellFaceListTupleIt fltit = faceListMap.begin();
            for(; fltit != faceListMap.end(); ++fltit) {

              // Add empty lookup item if key still unused
              const IndexPair lookupKey = fltit->first;
              typename PairLookupMapImp::iterator currentLookupItemIt =
                localLookupMap.insert(std::make_pair(lookupKey,LookupItem())).first;

              // Set inside domain cell
              currentLookupItemIt->second.insideDomainCell = &domainCell;

              // Iterate all faces in list
              typedef typename DomainCell::DomainCellFaceList DomainCellFaceList;
              const DomainCellFaceList & faceList = fltit->second;
              typename DomainCellFaceList::const_iterator fit = faceList.begin();
              for(; fit != faceList.end(); ++fit) {

                typedef typename DomainCellFaceList::value_type DomainCellFace;
                // typedef typename DomainCellFace::DomainCellFacePointers FacePointers;  // unused typedef

                // Determine outside cell
                const DomainCellFace* mirrorFace = fit->mirrorFace();
                const DomainCell* outsideDomainCell;
                Index outsideCellIndex;
                if(mirrorFace) {
                  outsideDomainCell = mirrorFace->insideDomainCell;
                  outsideCellIndex =  mirrorFace->insideCellIndex;
                }
                else{
                  outsideDomainCell = &domainCell;
                  outsideCellIndex = fit->insideCellIndex;
                }

                // Write entry in current lookup map
                const int indexInPart = fit->indexInPart;
                currentLookupItemIt->second.facesLookup[indexInPart] = outsideDomainCell;

                // If this is on a cell face in the fundamental mesh grid view
                if(fit->insideCellIndex != outsideCellIndex) {

                  // Access corresponding lookup item in cell map
                  IndexLookupMapImp & cellLookupMap = cellMap_[fit->insideCellIndex];
                  typename IndexLookupMapImp::iterator lookupItemIt =
                    cellLookupMap.find(domainCell.id);
                  assert(lookupItemIt!=cellLookupMap.end());
                  LookupItem & lookupItem = lookupItemIt->second;

                  // Access faces lookup in lookup item
                  typename LookupItem::FacesLookupMap & facesLookup =
                    lookupItem.facesLookup;
                  // typename LookupItem::FacesLookupMap::const_iterator faceLookupIt =
                  //   facesLookup.insert(std::make_pair(outsideCellIndex, outsideDomainCell)).first;  // unused variable
                  facesLookup.insert(std::make_pair(outsideCellIndex, outsideDomainCell));  // has to be done nevertheless
                }

              } // fit
            } // fltit

            // Iterate all level set mesh cells which are part of current domain cell
            typename DomainCell::LGVMemberMap::const_iterator mtit =
              domainCell.lgvMembers.begin();
            for(; mtit != domainCell.lgvMembers.end(); ++mtit) {

              // Iterate all refinement parts contained in level set mesh cell
              typedef typename DomainCell::LGVMemberMap::mapped_type IndexList;
              typename IndexList::const_iterator partIndexIt = mtit->second.begin();
              for(; partIndexIt != mtit->second.end(); ++partIndexIt) {

                // Add empty lookup item if key still unused
                const IndexPair lookupKey = std::make_pair(mtit->first,*partIndexIt);
                typename PairLookupMapImp::iterator currentLookupItemIt =
                  localLookupMap.insert(std::make_pair(lookupKey,LookupItem())).first;

                // Set inside domain cell
                currentLookupItemIt->second.insideDomainCell = &domainCell;
              } // partIndexIt

            } // mtit

          } // domainCellIt

          // Write to global lookup map
          lookupMap_.insert(localLookupMap.begin(),localLookupMap.end());

        } // i

        if(verbose) {
          std::cout << "\tDomain cells: " << numberOfDomainCells()
                    <<" on "<< hostCellMap.size() << " grid cells." << std::endl;
          std::cout << "\tDomain cell volumes:  Max. volume = "<< maxVolume
                    << "  Min. volume = " << minVolume << std::endl;
        }
      }

      /**
       * \brief Return number of domain cells.
       */
      Index numberOfDomainCells() const { return nDomainCells_; }

    protected:
      //! Store the map with (index in lgv, part index) access, i.e. map 1)
      PairLookupMapImp lookupMap_;

      //! Store a map with [index in gv][domainCell.id] access, used in order
      //! to implement map 2): for each fundamental mesh cell, whose children
      //! in the level set mesh grid view contribute domain refinement parts
      //! to domain cells, the id of each such a domain cell is mapped to the
      //! item holding the fundamental mesh cell's association with domain
      //! cells
      CellMap cellMap_;

      //! Store number of domain cells
      Index nDomainCells_;
    };


    /**
     * \brief Class representing a layer of the sub triangulation.
     *
     * \tparam DD Type of a domain definition.
     */
    template< typename DD >
    class MarchingCube33SubTriangulationLayer
    {
    private:
      //! Type of fundamental mesh grid view
      typedef typename DD::GridView GV;
      //! Type of level set mesh grid view
      typedef typename DD::LevelSetGridView LGV;

    public:
      //! Type of domain definition
      typedef DD DomainDefinition;

      //! Type of fundamental mesh grid view
      typedef GV GridView;
      //! Type of level set mesh grid view
      typedef LGV LevelSetGridView;

    private:
      //! Global dimension
      enum { dim = GV::dimension };

      //! Continuous field type
      typedef typename GV::ctype DomainField;

      //! Coordinate types
      typedef Dune::FieldVector<DomainField,dim> Coordinate;
      typedef std::vector<Coordinate> Coordinates;

      //! Index types
      typedef typename GV::IndexSet::IndexType Index;
      typedef std::list<Index> IndexList;
      typedef std::pair<Index,Index> IndexPair;
      typedef typename GV::IndexSet::IndexType Size;
      static Index invalidIndex(){ return std::numeric_limits<Index>::max(); }

      //! Grid view types
      typedef typename LGV::Traits::template Codim<0>::Entity CellEntity;
      typedef typename GV::Traits::template Codim<0>::Entity GVCellEntity;

      //! Domain refinement types
      typedef Dune::UDG::DomainRefinement<DomainDefinition> DomainRefinement;
      typedef typename DomainRefinement::DomainRefinementPart DomainRefinementPart;
      typedef typename DomainRefinementPart::Geometry Geometry;
      typedef typename DomainRefinementPart::DomainTag DomainTag;

      //! Domain cell types
      typedef MarchingCube33SubTriangulationImp::DomainCell<GV> DomainCell;
      typedef MarchingCube33SubTriangulationImp::DomainCellFace<GV> DomainCellFace;
      typedef typename DomainCell::RawBoundingBox RawBoundingBox;
      typedef typename std::list<DomainCell> DomainCellList;

    public:
      //! Type of maps mapping each domain refinement part of a level set mesh
      //! cell, respectively each fundamental mesh cell, to domain cells
      typedef SubTriangulationLayerLookupMap<DomainCell> LookupMap;

      //! Type of map mapping a cell index in gv to the list of domain
      //! cells which are (by merging) associated with this cell
      typedef std::vector<DomainCellList> HostCellMap;

      //! \todo Please doc me!
      typedef std::map<IndexPair,const DomainCell*> DomainCellLookup;

      /**
       * \brief Creates a new layer of the sub triangulation.
       *
       * \param domainDef
       *        Domain definition providing the domains on this layer.
       * \param layerIndex
       *        Index of the layer which shall be created.
       * \param removeIdenticalDomainCellFaceCornersEps
       *        Numerical epsilon which is used to decide whether two corners
       *        of a domain refinement part's face are identical which might
       *        prevent the creation of a domain cell face corresponding to
       *        this face.
       * \param minEdgeLengthEps
       *        Numerical epsilon which yields a minimum edge length during
       *        domain refinement.
       * \param linearizeParts
       *        Switch for iterative domain refinement procedure which triggers
       *        linearization before refinement using the next interface.
       */
      MarchingCube33SubTriangulationLayer(DomainDefinition & domainDef, const Index layerIndex,
                                          const double removeIdenticalDomainCellFaceCornersEps,
                                          const double minEdgeLengthEps, const bool linearizeParts)
        : domainDef_(domainDef), layerIndex_(layerIndex),
          gv_(domainDef.gridView()), lgv_(domainDef.levelSetGridView()),
          removeIdenticalDomainCellFaceCornersEps_(removeIdenticalDomainCellFaceCornersEps),
          minEdgeLengthEps_(minEdgeLengthEps), linearizeParts_(linearizeParts)
      {}

      /**
       * \brief Creates the actual domain refinement parts and the domain tag
       *        for a given level set mesh cell.
       */
      void createDomainRefinement
        (const CellEntity & lCell,
         typename DomainRefinement::DomainPartList & parts,
         DomainTag & domainTag) const
      {
        const int maxLayerIndex = domainDef_.numberOfLayers()-1;
        DomainRefinement refinement
          (domainDef_,maxLayerIndex,minEdgeLengthEps_,linearizeParts_);
        refinement.createDomainRefinement(lCell,parts,domainTag);
      }

      /**
       * \brief Given the current state of the triangulation layer, write the
       *        maps mapping each domain refinement part of a level set mesh
       *        cell, respectively each fundamental mesh cell, to domain cells.
       *
       * \note This function should be called directly after all domain cell
       *       merging is finished.
       */
      void writeLookupMap(const bool verbose = true)
      {
        if(verbose)
          std::cout << "Final status of MC33 sub triangulation layer "
                    << layerIndex_ << std::endl;
        lookupMap_.write(hostCellMap_,verbose);
      }

      //       // Return the domain cell which is positioned on the inside of
      //       // the given face. The given face is assumed to be created for
      //       // the top layer of the current domain definition and the domain
      //       // tag is assumed to represent the signs of the cell in lgv on
      //       // the inside of the given face.
      //       const DomainCell * embeddedDomainCell
      //       (Index partIndex,
      //        const Index cellIndex,
      //        const Index lCellIndex) const
      //       {
      //         // Here we create a lookup table and search for the
      //         // appropriate domain cell.
      //         #warning Local lookup table should become a cache

      //         std::list<const DomainCell*>  domainCells;
      //         findIntersectingDomainCells(cellIndex,domainCells);

      //         DomainCellLookup domainCellLookup;
      //         fillDomainCellLookup(domainCells,lCellIndex,domainCellLookup);

      //         return embeddedDomainCell(partIndex,domainCells,domainCellLookup,lCellIndex);
      //       }

      //       const DomainCell * embeddedDomainCell
      //       (Index partIndex,
      //        const std::list<const DomainCell*> & domainCells,
      //        const DomainCellLookup & domainCellLookup,
      //        const Index lCellIndex) const
      //       {
      //         // Compute appropriate part index
      //         // const Index topLayers = domainDef_.numberOfLayers()-1 - layerIndex_;
      //         // for(Index l=0; l<topLayers; ++l)
      //         //   partIndex /= DomainRefinement::maxPartsInLayer;

      //         typename DomainCellLookup::const_iterator lookupEntry =
      //           domainCellLookup.find(std::make_pair(lCellIndex,partIndex));

      //         // This check is necessary to find domain cells which were
      //         // created for the lower layer at a time when it was not
      //         // insersected.
      //         if(domainDef_.numberOfLayers() > 1 && lookupEntry == domainCellLookup.end())
      //           lookupEntry = domainCellLookup.lower_bound(std::make_pair(lCellIndex,0));

      //         assert(lookupEntry != domainCellLookup.end());
      //         const DomainCell & domainCell = *(lookupEntry->second);

      //         return &domainCell;
      //       }

      // Connect the existing domain cells i.e. find partners for the
      // existing domain cell faces. After this call, domain face cell
      // objects hold pointers on their mirror faces and should not be
      // copied anymore.
      void connectAdjacentDomainCells(DomainCellList & localDomainCells) const
      {
        typedef typename DomainCellList::iterator Iterator;
        Index matches(0);
        Iterator it = localDomainCells.begin();
        for(; it != localDomainCells.end(); ++it) {
          Iterator next = it; next++;
          for(; next != localDomainCells.end(); ++next)
            matches += next->connectFaces(*it);
        }
      }

      // Connect the domain cells in domainCellsA with domain cells in
      // domainCellsB, i.e. find partners for the existing domain cell
      // faces. After this call, domain face cell objects hold
      // pointers on their mirror faces and should not be copied
      // anymore.
      void connectAdjacentDomainCells
        (DomainCellList & domainCellsA, DomainCellList & domainCellsB) const
      {
        typedef typename DomainCellList::iterator Iterator;
        Index matches(0);
        Iterator ita = domainCellsA.begin();
        for(; ita != domainCellsA.end(); ++ita) {
          Iterator itb = domainCellsB.begin();
          for(; itb != domainCellsB.end(); ++itb)
            matches += ita->connectFaces(*itb);
        }
      }

      /**
       * \brief Determine whether a point is on a given fundamental mesh cell's
       *        boundary. The point x is given in local coordinates of the cell
       *        and is supposed to be located either on the boundary or inside
       *        the cell.
       *
       * \todo This method uses a hardcoded numerical epsilon for its
       *       decision. We should change this as this might cause errors?
       */
      bool isOnCellBoundary(const GVCellEntity & cell, const Coordinate & x) const
      {
        const DomainField eps = 1e-7;
        if(cell.type().isCube()) {
          for(int d=0; d<dim; ++d)
            if(x[d] < eps || x[d] > 1.0-eps)
              return true;
        }
        else if(cell.type().isSimplex()) {
          DomainField sum(0);
          for(int d=0; d<dim; ++d) {
            sum += x[d];
            if(x[d] < eps)
              return true;
          }
          if(sum > 1.0-eps)
            return true;
        }
        else{
          DUNE_THROW(Dune::NotImplemented,"This is only implemented for simplices and cubes.");
        }
        return false;
      }

      /**
       * \brief Given a fundamental mesh cell and a list of domain cells in this
       *        fundamental mesh cell, force a merge of all domain cells in the
       *        given list without checking that this is reasonable.
       */
      void blindMergeDomainCellsOnGVCell
        (const GVCellEntity & cell, std::list<DomainCell> & localDomainCells)
      {
        typedef typename DomainCellList::iterator Iterator;
        using namespace MarchingCube33SubTriangulationImp;

        // Manually check that list is not empty because otherwise
        // faces will not be defined
        if(localDomainCells.empty()) return;

        // Merge all domain cells without removing matching faces
        Iterator it = localDomainCells.begin();
        while(localDomainCells.size() > 1) {
          localDomainCells.back().blindMerge(*it);
          it = localDomainCells.erase(it);
        }

        // Only one domain cell is left
        DomainCell & domainCell = *(localDomainCells.begin());

        // Reset all intersection area of this domain cell and get
        // reference to (exterior?) boundary area field
        domainCell.areas.clear();
        domainCell.areas[domainCell.id] = 0;
        DomainField & domainCellArea = domainCell.areas[domainCell.id];

        // Remove all faces with centers not on fundamental mesh cell
        // boundary and collect intersection area of this domain cell
        typedef typename DomainCell::DomainCellFaceListMap DomainCellFaceListMap;
        DomainCellFaceListMap & faceListMap = domainCell.faceListMap;
        typedef typename DomainCellFaceListMap::iterator DomainCellFaceListTupleIt;
        DomainCellFaceListTupleIt fltit = faceListMap.begin();
        while(fltit != faceListMap.end()) {

          typedef std::list<DomainCellFace> DomainCellFaceList;
          typename DomainCellFaceList::iterator fit = fltit->second.begin();
          while(fit != fltit->second.end()) {
            const Coordinate localInCell =
              cell.geometry().local(fit->geometry.center());

            if(!isOnCellBoundary(cell,localInCell)) {
              fit = fltit->second.erase(fit);
              continue;
            }
            else{
              if(fit->mirrorFace()) {
                typedef typename DomainCellFace::DomainCellFacePointers FacePointers;
                typename FacePointers::const_iterator mirrorFaceIt = fit->allMirrorFaces().begin();
                for(; mirrorFaceIt != fit->allMirrorFaces().end(); ++mirrorFaceIt)
                  domainCell.areas[(*mirrorFaceIt)->insideDomainCell->id] += (*mirrorFaceIt)->area;
              }
              else
                domainCellArea += fit->area;
            }

            ++fit;
          }

          ++fltit;
        }

        // Remove all empty face lists
        fltit = faceListMap.begin();
        while(fltit != faceListMap.end()) {
          if(fltit->second.empty()) {
            faceListMap.erase(fltit);
            fltit = faceListMap.begin();
          }
          else ++fltit;
        }
      }

      // Connect domain cells in domainCells to domain cells in the
      // lgv_ neighbors of lCell. Create domain cells for neighbor
      // cells in lgv_ if necessary i.e. if their LGVDomainCells entry
      // is an empty list. Skip lgv_ neighbors whose index is
      // contained in the processedLCellIndices set.
      void connectDomainCellsInAdjacentLGVCells
        (const CellEntity & lCell, DomainCellList & domainCells,
        std::vector<DomainCellList> & lGVDomainCells,
        const std::vector<bool> & processedLCellIndices) const
      {
        // Connect local domain cells to the domain cells in the
        // neighbors of this lCell in lgv_
        typename LGV::IntersectionIterator iit = lgv_.ibegin(lCell);
        const typename LGV::IntersectionIterator eiit = lgv_.iend(lCell);
        for(; iit != eiit; ++iit) {
          if(iit->boundary() || !iit->neighbor()) continue;

          // Access neighbor domain cells
          const CellEntity neighborLCell = iit->outside();
          const Index neighborLCellIndex = lgv_.indexSet().index(neighborLCell);

          // If neighbor was already processed, connections have
          // already been established.
          if(processedLCellIndices[neighborLCellIndex])
            continue;

          // Access domain cells in neighbor (create, if not existing yet)
          DomainCellList & neighborDomainCells = lGVDomainCells[neighborLCellIndex];
          if(neighborDomainCells.size()==0) {
            const GVCellEntity neighborCell = domainDef_.gvEntity(neighborLCell);
            createDomainCells(neighborDomainCells,neighborCell,neighborLCell);
          }

          // Connect domain cells
          connectAdjacentDomainCells(domainCells,neighborDomainCells);
        }
      }

      /**
       * \brief Create domain cells in a level set mesh cell and add them to a
       *        list of domain cells.
       *
       * \todo Please document the parameters!
       */
      void createDomainCells
        (std::list<DomainCell> & localDomainCells,
         const GVCellEntity & cell, const CellEntity & lCell) const
      {
        const Index cellIndex = gv_.indexSet().index(cell);
        const Index lCellIndex = lgv_.indexSet().index(lCell);

        // Create refinement parts and connectivity vertex groups
        typedef typename DomainRefinement::DomainPartList PartList;
        PartList parts;
        DomainTag lCellDomainTag;
        createDomainRefinement(lCell,parts,lCellDomainTag);

        // Iterate refinement parts
        typedef typename PartList::const_iterator Iterator;
        Iterator pit = parts.begin();
        for(; pit != parts.end(); ++pit) {

          // Add new domain cell for current refinement part
          localDomainCells.push_back
            ( DomainCell(cell,domainCellIdCounter_++,domainDef_.coordinateComparator()) );
          DomainCell & domainCell = localDomainCells.back();
          domainCell.lgvMembers[lCellIndex].push_back( pit->index );
          domainCell.gvMembers.insert( std::make_pair(cellIndex,cell.seed()) );
          domainCell.domainIndex = domainDef_.domainTag2Index(pit->domainTag,layerIndex_);
          DomainField & allFaceArea = domainCell.areas[domainCell.id];

          // Determine corners of current refinement part in global coordinates
          // and update bounding box
          const Index nCorners( pit->geometry.corners() );
          Coordinates globalCoords(nCorners);
          for(Index c=0; c<nCorners; ++c) {

            // Compute coordinate local in level set mesh cell
            const Coordinate & localInLCell = pit->geometry.corner(c);

            // Compute global coordinate
            globalCoords[c] = lCell.geometry().global(localInLCell);

            // Update bounding box
            domainCell.bbox.update(globalCoords[c]);
          }

          // Create global geometry object for current refinement part
          typedef typename DomainRefinementPart::Geometry Geometry;
          const Geometry globalGeometry = Geometry(pit->geometry.type(),globalCoords);

          // Set domain cell volume
          domainCell.volume = globalGeometry.volume();

          // Initialize the domain cell's face list map with a map entry which
          // associates the current refinement part with an empty list of domain
          // cell faces
          typedef typename DomainCell::DomainCellFaceListMap DomainCellFaceListMap;
          typedef typename DomainCell::DomainCellFaceList DomainCellFaceList;
          typedef typename DomainCellFaceListMap::iterator DomainCellFaceListTupleIt;
          DomainCellFaceListTupleIt domainCellFaceListTupleIt
            = domainCell.faceListMap.insert(std::make_pair(std::make_pair(lCellIndex,pit->index),
                                                           DomainCellFaceList())).first;

          // Get reference element of current refinement part
          auto ref =
            Dune::referenceElement<DomainField,dim>(pit->geometry.type());

          // Iterate faces of current refinement part
          const int nFaces = ref.size(1);
          for(int f=0; f<nFaces; ++f) {

            // Get the current face's vertices in global coordinates
            using namespace MarchingCube33SubTriangulationImp;
            Coordinates faceVertices;
            extractFaceCoords(f,globalGeometry,faceVertices);

            // Remove identical corners
            removeGlobalDegeneratedCorners
              (lCell,faceVertices,removeIdenticalDomainCellFaceCornersEps_);
            if(faceVertices.size() < dim) continue;

            // Create global geometry object for current face
            typedef typename DomainCell::FaceGeometry FaceGeometry;
            Dune::GeometryType faceType =
              Dune::geometryTypeFromVertexCount(dim-1,faceVertices.size());
            const FaceGeometry faceGeometry = FaceGeometry(faceType,faceVertices);

            // Compute unit outer normal for current face
            const Coordinate localCenter = ref.position(f,1);
            Coordinate unitOuterNormal = computeNormal(globalGeometry,f,localCenter);
            const DomainField normalLength = unitOuterNormal.two_norm();
            assert(normalLength > 0);
            unitOuterNormal /= normalLength;

            // Create domain cell face from current face and add it to list
            // of domain cell faces associated with current refinement part
            domainCellFaceListTupleIt->second.push_back
              (DomainCellFace
                (&domainCell,cellIndex,lCellIndex,pit->index,f,faceGeometry,unitOuterNormal));
            DomainCellFace & domainCellFace = domainCellFaceListTupleIt->second.back();
            allFaceArea += domainCellFace.area;

          } // f

        } // pit

        // Connect and merge all domain cells in this level set mesh cell
        connectAdjacentDomainCells(localDomainCells);
        DomainCell::mergeAdjacentDomainCells(localDomainCells);
      }

      /**
       * \brief Check whether a given fundamental mesh cell is host cell of
       *        entity parts with a given domain index.
       *
       * \note This function may return false even if some entity parts with the
       *       given domain index are geometrically positioned in the given cell
       *       (which is therefore their fundamental mesh "home cell") but are
       *       merged to - and therefore associated with - another fundamental
       *       mesh cell (their "host cell").
       */
      bool isHostCell
        (const GVCellEntity & hostCandidate, const int & domainIndex) const
      {
        const Index hostIndex = gv_.indexSet().index(hostCandidate);
        const DomainCellList & domainCells = hostCellMap_[hostIndex];
        for (typename DomainCellList::const_iterator it = domainCells.begin();
             it != domainCells.end(); ++it)
          if(int(it->domainIndex) == domainIndex)
          {
            assert(gv_.indexSet().index(gv_.grid().entity(it->hostCell))==gv_.indexSet().index(hostCandidate));
            return true;
          }
        return false;
      }

      /**
       * \brief Setup the layer of the sub triangulation according to
       *        domain definition.
       *
       * \note This function should be called directly after the interfaces
       *       and domains have been defined and before any other interface
       *       function.
       */
      void init()
      {
        // Initalize non-const members
        domainCellIdCounter_ = 0;
        hostCellMap_.clear();

        // Allocate map of domain cells
        hostCellMap_.resize(gv_.size(0));

        // Temporary map associating level set mesh cells to their
        // local domain cells
        std::vector<DomainCellList> lGVDomainCells(lgv_.size(0));

        // Store the cell indices in level set grid view which were
        // already processed
        std::vector<bool> processedLCellIndices(lgv_.size(0),false);

        // Iterate cells in fundamental mesh grid view for normalization
        typedef typename GV::Traits::template Codim<0>::Iterator GVCellIterator;
        GVCellIterator it = gv_.template begin<0>();
        const GVCellIterator eit = gv_.template end<0>();
        for(; it != eit; ++it) {

          // Get current fundamental mesh cell
          const GVCellEntity & cell = *it;
          const Index cellIndex = gv_.indexSet().index(cell);

          // Get domain cells in current fundamental mesh cell
          DomainCellList & localDomainCells = hostCellMap_[cellIndex];

          // Iterate all child cells in level set mesh grid view
          using namespace MarchingCube33SubTriangulationImp;
          SubGridViewIterator<GV,LGV> hit(gv_,lgv_,*it);
          for(; hit.valid(); ++hit) {

            // Get current child cell
            const CellEntity & lCell = *hit;
            const Index lCellIndex = lgv_.indexSet().index(lCell);

            // Mark current child cell as processed
            processedLCellIndices[lCellIndex] = true;

            // Create domain cells in current child cell (if required)
            DomainCellList & domainCells = lGVDomainCells[lCellIndex];
            if(domainCells.size() == 0)
              createDomainCells(domainCells,cell,lCell);

            // Connect current child cell's domain cells to domain cells
            // in level set mesh neighbors of current child cell
            connectDomainCellsInAdjacentLGVCells
              (lCell,domainCells,lGVDomainCells,processedLCellIndices);

            // Transfer domain cells in current child cell to the list
            // of domain cells in current fundamental mesh cell
            localDomainCells.splice
              (localDomainCells.begin(),
               domainCells,domainCells.begin(),domainCells.end());

          } // hit

          // Merge the domain cells in current fundamental mesh cell (use
          // a special implementation ("blind merge") for fundamental mesh
          // cells without intersections)
          if(domainDef_.numberOfInterfacesInCell(cellIndex,layerIndex_) == 0)
            blindMergeDomainCellsOnGVCell(cell,localDomainCells);
          else
            DomainCell::mergeAdjacentDomainCells(localDomainCells);

        } // it
      }

      /**
       * \brief Access to the maps mapping each domain refinement part of a
       *        level set mesh cell, respectively each fundamental mesh cell,
       *        to domain cells.
       */
      const LookupMap & lookupMap() const { return lookupMap_; }

      /**
       * \brief Access to the map mapping a cell index in fundamental mesh
       *        grid view to the list of domain cells which are (by merging)
       *        associated with this cell.
       */
      HostCellMap & hostCellMap()
      { return hostCellMap_; }

      /**
       * \brief Constant access to the map mapping a cell index in fundamental
       *        mesh grid view to the list of domain cells which are (by merging)
       *        associated with this cell.
       */
      const HostCellMap & hostCellMap() const
      { return hostCellMap_; }

    private:
      //! Domain definition
      DomainDefinition & domainDef_;

      //! Index of this layer in domain definition
      const Index layerIndex_;

      //! Grid view for finite element grid, i.e. the fundamental mesh
      const GV & gv_;
      //! Grid view for interface reconstruction, i.e. the level set mesh
      const LGV & lgv_;

      //! Map mapping a cell index in fundamental mesh grid view to the list
      //! of domain cells which are (by merging) associated with this cell
      std::vector<DomainCellList> hostCellMap_;

      //! Maps mapping each domain refinement part of a level set mesh
      //! cell, respectively each fundamental mesh cell, to domain cells
      LookupMap lookupMap_;

      //! Numerical epsilon which is used to decide whether two corners of
      //! a domain refinement part's face are identical which might prevent
      //! the creation of a domain cell face corresponding to this face
      //! \todo Comment on the interpretation of its value
      const double removeIdenticalDomainCellFaceCornersEps_;
      //! Numerical epsilon which yields a minimum edge length during
      //! domain refinement
      //! \todo Comment on the interpretation of its value
      const double minEdgeLengthEps_;

      //! Switch for iterative domain refinement procedure which triggers
      //! linearization before refinement using the next interface
      const bool linearizeParts_;

      //! Counter to provide a unique id for each domain cell
      mutable Index domainCellIdCounter_;
    };

  }
}


#endif
