// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_UDG_MC33_UGFLEVELSETBACKEND_HH
#define DUNE_UDG_MC33_UGFLEVELSETBACKEND_HH
#include "levelsetbackend.hh"
#include "mc33triangulationimp.hh"
#include "../pdelab/subtriangulation.hh"

namespace Dune {
  namespace UDG {

    // A stub which is used if the level set function does not fulfill
    // the unfitted geometry function interface
    template< typename DomainDefinition, typename LevelSetFunction, typename Switch=void>
    class UnfittedGeometryFunctionLevelSetBackend :
      public LevelSetBackendNodeStorageStub<DomainDefinition>{
    public:
      typedef typename LevelSetBackendNodeStorageStub<DomainDefinition>::Traits Traits;
      UnfittedGeometryFunctionLevelSetBackend
        (const DomainDefinition & , const LevelSetFunction & , const typename Traits::Index){}
    };


    /** \brief A backend allowing access to the level set function
        values which guarantees unique level set values in identical
        corners.

        This backend provides the level set function values in the
        corners of a given refinement parts. It guarantees the
        following: If two parts share a corner which, according to the
        coordinate comparator of the domain definition class are
        identical, the provided level set function value in this
        corner will be identical (not only numerical but bit-wise).

        To realize this functionality, this class uses a cache like
        data storage. The memory usage for a grid view of size n^dim
        is grows like: n^(dim-1) + (number of nodes in triangulation
        which are not corners in the grid view).
     */
    template< typename DomainDefinition, typename LevelSetFunction >
    class UnfittedGeometryFunctionLevelSetBackend
    <DomainDefinition,LevelSetFunction,
     typename std::enable_if<LevelSetFunction::unfitted_geometry_function>::type> :
      public virtual LevelSetBackendNodeStorageInterface<DomainDefinition>
    {
    public:
      //! Base class
      typedef LevelSetBackendNodeStorageInterface<DomainDefinition> Base;
      //! Traits class
      typedef typename Base::Traits Traits;

    private:
      typedef typename Traits::GridView GridView;
      typedef typename Traits::LevelSetGridView LevelSetGridView;
      typedef typename Traits::LevelSetValues LevelsetValues;
      typedef typename Traits::Coordinate Coordinate;
      typedef typename Coordinate::value_type DomainField;
      typedef std::vector<DomainField> LevelSetValues;
      typedef typename Traits::CoordinateComparator CoordinateComparator;
      enum { dim = Coordinate::dimension };
      typedef typename LevelSetGridView::template Codim<0>::Entity CellEntity;
      typedef typename GridView::template Codim<0>::Entity GVCellEntity;
      typedef typename LevelSetFunction::Traits::SubTriangulation SubTriangulation;
      typedef typename SubTriangulation::EntityPart EntityPart;
      typedef typename SubTriangulation::EntityPartList::value_type EntityPartPointer;
      typedef typename Traits::Index Index;
      typedef typename Traits::RefinementPart RefinementPart;
      typedef CoordinateValueCache<CoordinateComparator,Index> Cache;

    protected:
      //! Helper function to get sign from vector of values
      int getSign
        (const typename Traits::LevelSetValues & values) const
      {
        // Flag to indicate an interface
        bool interfaceFound = false;
        typename Traits::DomainField lastValue(0);

        for(typename Traits::Index v=0; v<values.size(); ++v) {
          if(values[v] == domainDef_->invalidValue()) continue;

          // Check for interface intersecting this cell
          interfaceFound |= ( lastValue * values[v] ) < 0;
          lastValue = values[v];
        }
        if( interfaceFound )
          return 0;
        else
          return int( ((1.0+1e-10)*lastValue)/std::abs(lastValue) );
      }

      Index globalCornerIndex(const CellEntity & lCell, const Index localIndex) const
      {
        assert(localIndex >=0 && int(localIndex) < lCell.geometry().corners() );
        typedef typename LevelSetGridView::Traits::
        template Codim<dim>::Entity NodeEntity;
        const NodeEntity ve = lCell.template subEntity<dim>(localIndex);
        const Index vertexIndex = domainDef_->levelSetGridView().indexSet().index(ve);
        return vertexIndex;
      }

      void updateSignInfo
        (const Index cellIndex, const DomainField value, std::vector<bool> & signInfo) const
      {
        assert(value!=0);
        const bool negativeValue = value<0;
        typedef typename std::vector<bool>::reference BitReference;
        BitReference negativeBit = signInfo[2*cellIndex];
        negativeBit = negativeValue || negativeBit;
        BitReference positiveBit = signInfo[2*cellIndex+1];
        positiveBit = !negativeValue || positiveBit;
      }

      void interpolatePartCorners
      (
        const GVCellEntity & homeCell,
        const EntityPartPointer & part,
        std::vector<bool> & newSignInfo
      )
      {
        const Index cellIndex = domainDef_->gridView().indexSet().index(homeCell);

        // Access cell in geometry grid view for this part
        typedef typename EntityPart::Traits::Entity Entity;
        const Entity lCell( part->entity() );
        const Index lCellIndex = domainDef_->levelSetGridView().indexSet().index(lCell);

        // Create PDELab style geometry
        const Dune::PDELab::UnfittedGeometryWrapper<EntityPart> partGeom(homeCell,part);

        // Reference element of cell to refine
        const Dune::ReferenceElement<DomainField,dim> & ref =
          Dune::ReferenceElements<DomainField,dim>::general(part->geometry().type());

        DomainField lastValue(0);
        for(int c=0; c<ref.size(dim); ++c) {
          const Coordinate local = ref.position(c,dim);

          // Determine level set function value
          typename LevelSetFunction::Traits::RangeFieldType value;
          {
            typename LevelSetFunction::Traits::RangeType rvalue;
            lsf_->evaluate(partGeom,local,rvalue);
            value = rvalue;
          }

          // Avoid zero values
          value = value ? value : std::numeric_limits<DomainField>::min();

          // Determine local index in lCell
          const Coordinate global = part->geometry().global(local);
          const Coordinate localInCell = lCell->geometry().local(global);
          using namespace MarchingCube33SubTriangulationImp;
          const int localNodeIndex = Coordinate2LocalIndexHelper::
                                     localIndex(domainDef_->coordinateComparator(),lCell,localInCell);

          // Put to cache if node of part not in cell corner or to the
          // grid storage otherwise. In both cases value will be
          // overwritten if a value for this position has already been
          // computed!
          if(localNodeIndex < 0) {
            cache_.find(global,value);
          }
          else{
            const Index vertexIndex = globalCornerIndex(lCell,localNodeIndex);
            DomainField & currentValue = levelSetValues_[vertexIndex];
            if(currentValue == domainDef_->invalidValue())
              currentValue = value;
            else
              value = currentValue;
          }

          // Update the sign info for this cell
          updateSignInfo(cellIndex,value,newSignInfo);

          // Store occurence of intersection for this lCell
          typedef typename std::vector<bool>::reference BitReference;
          BitReference isIntersected = lCellIntersections_[lCellIndex];
          isIntersected = value*lastValue < 0 || isIntersected;
          lastValue = value;
        } // c
      }

    public:

      //! Constructor
      UnfittedGeometryFunctionLevelSetBackend
        (const DomainDefinition & domainDef, const LevelSetFunction & lsf, const Index itfIndex)
        : domainDef_(&domainDef), itfIndex_(itfIndex), lsf_(&lsf),
          cache_(domainDef_->coordinateComparator())
      {}

      //! Clear all cached values in the backend
      void reset(std::vector<bool> & signInfo)
      {
        // First modify a copy of sign info. This is useful if the
        // given unitted geometry function is working on the current
        // sub triangulation.
        std::vector<bool> newSignInfo;

        // Allocate level set values
        levelSetValues_.clear();
        levelSetValues_.resize
          (domainDef_->levelSetGridView().size(dim),domainDef_->invalidValue());

        // Store whether an lCell is intersected
        lCellIntersections_.clear();
        lCellIntersections_.resize(domainDef_->levelSetGridView().size(0),false);

        // Allocate sign info flags
        newSignInfo.clear();
        newSignInfo.resize(domainDef_->gridView().size(0)*2,false);

        cache_.reset();

        typedef typename DomainDefinition::CoordinateComparator Comparator;
        Comparator& comp = const_cast<Comparator&>(domainDef_->coordinateComparator());
        std::vector<bool> compUpdateChecklist(domainDef_->levelSetGridView().size(0),false);

        // Access sub triangulation from level set function
        typedef typename LevelSetFunction::Traits::SubTriangulation SubTriangulation;
        const SubTriangulation & subTriangulation = lsf_->subTriangulation();

        // Iterate all cells in grid view
        const GridView & gv = domainDef_->gridView();
        typedef typename GridView::template Codim<0>::
        template Partition<Dune::Interior_Partition>::Iterator CellIterator;
        const CellIterator cend =  gv.template end<0,Dune::Interior_Partition>();
        CellIterator cit =  gv.template begin<0,Dune::Interior_Partition>();
        for(; cit != cend; ++cit) {

          // Create entity parts
          typedef typename SubTriangulation::EntityPartList EntityParts;
          EntityParts entityParts;
          const bool oldFRSetting =
            const_cast<SubTriangulation&>(subTriangulation).forceRefinement();
          const_cast<SubTriangulation&>(subTriangulation).forceRefinement() = true;
          subTriangulation.create_entity_parts(*cit,entityParts);
          const_cast<SubTriangulation&>(subTriangulation).forceRefinement() = oldFRSetting;

          // Iterate entity parts
          typename EntityParts::const_iterator eit  = entityParts.begin();
          typename EntityParts::const_iterator eeit = entityParts.end();
          for(; eit != eeit; ++eit) {

            // Inform the coordinate comparator about cell geometry (if
            // not done yet).
            typedef typename EntityPart::Traits::Entity EntityPointer;
            const GVCellEntity lCell( (*eit)->entity() );
            const Index lCellIndex( domainDef_->levelSetGridView().indexSet().index(lCell) );
            if( !comp.valid() && !compUpdateChecklist[lCellIndex] )
              comp.updateReferenceMap(lCell.geometry());

            // Interpolate in all corners of part
            const EntityPartPointer & part = *eit;
            interpolatePartCorners(*cit,part,newSignInfo);
          } // eit

        } // cit

        // Exchange new and old sign info
        signInfo.swap(newSignInfo);

        // Finish setup of coordinate comparator (if not done yet).
        if(!comp.valid()) comp.finalize();
      }


      int cellCornerValues
        (const CellEntity & lCell,
        LevelSetValues & values) const
      {
        // Values in corners of lCell
        values.resize(lCell.geometry().corners());
        for(Index v=0; v<values.size(); ++v) {
          const Index nodeIndex = globalCornerIndex(lCell,v);
          values[v] = levelSetValues_[nodeIndex];
          assert(values[v]); // Zero values should have been removed
        }
        const int sign = getSign(values);
        const Index lCellIndex( domainDef_->levelSetGridView().indexSet().index(lCell) );
        const bool intersection = lCellIntersections_[lCellIndex];
        return intersection ? 0 : sign;
      }


      int partCornerValues
      (
        const CellEntity & lCell,
        const RefinementPart & part,
        LevelSetValues & values
      ) const
      {
        values.resize(part.geometry.corners());

        for(Index v=0; v<values.size(); ++v) {
          // Determine local index of corner in cell
          const Coordinate localInLCell = part.geometry.corner(v);
          using namespace MarchingCube33SubTriangulationImp;
          const int localNodeIndex = Coordinate2LocalIndexHelper::
                                     localIndex(domainDef_->coordinateComparator(),lCell,localInLCell);
          if(localNodeIndex < 0) {
            const Coordinate global = lCell.geometry().global(localInLCell);
            const bool hit = cache_.find(global,values[v]);
            if(!hit) {
              cache_.find(global,values[v]);
              DUNE_THROW(Dune::Exception,"This cache should always hit!");
            }
          }
          else{
            const Index vertexIndex = globalCornerIndex(lCell,localNodeIndex);
            values[v] = levelSetValues_[vertexIndex];
          }
          assert(values[v]); // Zero values should have been removed
        }
        return getSign(values);
      }

      void printStatistics() const
      {
        cache_.statistics();
      }

      //! Access to level set function values
      LevelSetValues & levelSetFunctionValues()
      { return levelSetValues_; }
      //! Const access to level set function values
      const LevelSetValues & levelSetFunctionValues() const
      { return levelSetValues_; }

    private:
      const DomainDefinition * domainDef_;
      const Index itfIndex_;
      const LevelSetFunction * lsf_;
      mutable Cache cache_;
      LevelSetValues levelSetValues_;
      std::vector<bool> lCellIntersections_;
    };


  }
}

#endif
