// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_UDG_MC33_UNIQUELEVELSETBACKEND_HH
#define DUNE_UDG_MC33_UNIQUELEVELSETBACKEND_HH

#include <map>
#include "levelsetbackend.hh"
#include "mc33triangulationimp.hh"

namespace Dune {
  namespace UDG {

    template< typename CoordinateComparator, typename Index>
    class CoordinateValueCache {
    public:
      //! Coordinate type
      typedef typename CoordinateComparator::GlobalCoordinate Coordinate;
      //! Dimension
      enum { dim = Coordinate::dimension };
      //! Field type
      typedef typename CoordinateComparator::DomainField DomainField;

    private:
      //! Cache entry for coordinate value pair
      struct CacheEntry {
        CacheEntry
          (const Coordinate & coordInRef, const DomainField & value)
          : coordInRef_(coordInRef), value_(value){}
        Coordinate coordInRef_;
        DomainField value_;
      };
      //! A line of cache entries
      typedef std::map<DomainField,CacheEntry> CacheLine;

      //! Compute a flat index from the dim-1 coordinate index
      Index flatindex(const std::vector<Index> & index) const
      {
        Index i(index[0]);
        for(Index d=1; d<dim-1; ++d)
          i += index[d] * cells[d-1];
        return i;
      }

      //! Try to find the coordinate for a given cache line
      bool findInLine
        (const Coordinate & ref, const Index flatindex, DomainField & value) const
      {
        const DomainField eps = comp_->numericalEqualityEps();
        const DomainField lowKey = ref[dim-1]-2.0*eps;
        typename CacheLine::const_iterator lineIt =
          cachelines[flatindex].lower_bound(lowKey);
        while(lineIt != cachelines[flatindex].end() && lineIt->first < lowKey+4.0*eps) {
          const CacheEntry & entry = lineIt->second;
          if(comp_->isEqualInRef(entry.coordInRef_,ref)) {
            value = entry.value_;
            hits++;
            return true;
          }
          lineIt++;
        }
        return false;
      }

      //! Add the given coordinate to the given cache line
      void addToLine
        (const Coordinate & ref, const Index flatindex, DomainField & value)
      {
        DomainField key = ref[dim-1];
        if(value==0) value = std::numeric_limits<DomainField>::min();
        CacheEntry entry(ref,value);
        typedef typename CacheLine::iterator LineIt;
        bool inserted = false;
        while(!inserted) {
          std::pair<LineIt,bool> res =
            cachelines[flatindex].insert(std::make_pair(key,entry));
          inserted = res.second;

          //! If it was not inserted make sure it was not blocked
          if(!inserted) {
            const CacheEntry & entry = res.first->second;
            if(comp_->isEqualInRef(entry.coordInRef_,ref)) {
              std::cerr<< "Warning: Reinsertion!" << std::endl;
              inserted = true;
            }
            else
              key += (std::numeric_limits<DomainField>::epsilon()*100.0)*key;
          }
        }
        misses++;
      }

    public:
      //! Constructor
      CoordinateValueCache(const CoordinateComparator & comp)
        : comp_(&comp),
          cells(dim-1), nCells(1)
      {}

      //! Reset the cache, empty all entries
      void reset(){
        cachelines.clear();

        high = comp_->inRef(comp_->domainMaximum());
        low = comp_->inRef(comp_->domainMinimum());
        length = high - low;

        Coordinate Hg(comp_->domainMaximum()-comp_->domainMinimum());
        DomainField hg(comp_->minDistance());
        for(Index d=0; d<dim-1; ++d) {
          cells[d] = Index(Hg[d] / hg + 1.0);
          nCells *= cells[d];
          h[d] = length[d] / DomainField(cells[d]);
        }
        cachelines.resize(nCells);
        hits = 0;
        misses = 0;
      }

      /** \brief Find the given global coordinate in the cache.

          If the coordinate is found, value will be set to the cache
          value and the return value is true. Otherwise, the given
          value will be stored in the cache and the return value is
          false.
       */
      bool find(const Coordinate & global, DomainField & value)
      {
        const Coordinate ref = comp_->inRef(global);

        Coordinate cIndex;
        std::vector<Index> index(dim-1);
        for(Index d=0; d<dim-1; ++d) {
          cIndex[d] = (ref[d]-low[d])/h[d];
          cIndex[d] = std::max(cIndex[d],0.0);
          index[d] = Index(cIndex[d]);
          index[d] = std::min(index[d],cells[d]-1);
        }

        { // Check whether coordinate in this line
          const Index flat = flatindex(index);
          if(findInLine(ref,flat,value)) return true;
        }

        // Check whether coordinate in any neighbor line
        for(Index d=0; d<dim-1; ++d) {
          std::vector<Index> nIndex(index);
          if((cIndex[d] - DomainField(index[d])) < 0.5)
            nIndex[d] = std::max(nIndex[d],Index(1))-1;
          else
            nIndex[d] = std::min(nIndex[d]+1,cells[d]-1);

          const Index flat = flatindex(nIndex);
          if(findInLine(ref,flat,value)) return true;
        }

        { // Cache miss! Add to cache
          const Index flat = flatindex(index);
          addToLine(ref,flat,value);
        }
        return false;
      }

      /** \brief Print some statistics */
      void statistics() const
      {
        DomainField ratio = DomainField(hits)/DomainField(hits+misses);
        std::cout << "Level set cache: " << std::endl;
        std::cout << "\tHit ratio: "<< ratio << std::endl;
        std::cout << "\t  Storage: "<< misses << std::endl;
      }

    private:
      const CoordinateComparator * comp_;
      std::vector<Index> cells;
      Index nCells;
      Coordinate h;
      Coordinate low;
      Coordinate high;
      Coordinate length;
      std::vector<CacheLine> cachelines;
      CacheLine emptyLine;
      mutable Index hits;
      mutable Index misses;
    };


    /** \brief A backend allowing access to the level set function
        values which guarantees unique level set values in identical
        corners.

        This backend provides the level set function values in the
        corners of a given refinement parts. It guarantees the
        following: If two parts share a corner which, according to the
        coordinate comparator of the domain definition class are
        identical, the provided level set function value in this
        corner will be identical (not only numerical but bit-wise).

        To realize this functionality, this class uses a cache like
        data storage. The memory usage for a grid view of size n^dim
        is grows like: n^(dim-1) + (number of nodes in triangulation
        which are not corners in the grid view).
     */
    template< typename DomainDefinition, typename LevelSetFunction, typename Switch=void>
    class UniqueLevelSetBackend :
      public DefaultLevelSetBackend<DomainDefinition,LevelSetFunction>
    {
    public:
      //! Base class
      typedef DefaultLevelSetBackend<DomainDefinition,LevelSetFunction> Base;
      //! Traits class
      typedef typename Base::Traits Traits;

    private:
      typedef typename Traits::LevelSetGridView LevelSetGridView;
      typedef typename Traits::LevelSetValues LevelsetValues;
      typedef typename Traits::Coordinate Coordinate;
      typedef typename Coordinate::value_type DomainField;
      typedef std::vector<DomainField> LevelSetValues;
      typedef typename Traits::CoordinateComparator CoordinateComparator;
      enum { dim = Coordinate::dimension };
      typedef typename LevelSetGridView::template Codim<0>::Entity CellEntity;
      typedef typename Traits::Index Index;
      typedef typename Traits::RefinementPart RefinementPart;
      typedef CoordinateValueCache<CoordinateComparator,Index> Cache;

    public:
      using DefaultLevelSetBackend<DomainDefinition,LevelSetFunction>::getSign;

      //! Constructor
      UniqueLevelSetBackend
        (const DomainDefinition & domainDef, const LevelSetFunction & lsf, const Index itfIndex)
        : Base(domainDef,lsf,itfIndex), domainDef_(&domainDef),
          cache_(domainDef_->coordinateComparator())
      {}

      //! Clear all cached values in the backend
      void reset(std::vector<bool> & signInfo){
        Base::reset(signInfo);
        cache_.reset();
      }

      void printStatistics() const
      {
        cache_.statistics();
      }

      int partCornerValues
      (
        const CellEntity & lCell,
        const RefinementPart & part,
        LevelSetValues & values
      ) const
      {
        // Compute values by interpolation
        Base::partCornerValues(lCell,part,values);

        // Feed cache with values or update values from cache values
        // if available
        for(Index v=0; v<values.size(); ++v) {
          const Coordinate localInLCell = part.geometry.corner(v);
          using namespace MarchingCube33SubTriangulationImp;
          const int localNodeIndex = Coordinate2LocalIndexHelper::
                                     localIndex(domainDef_->coordinateComparator(),lCell,localInLCell);
          if(localNodeIndex < 0) {
            const Coordinate global = lCell.geometry().global(localInLCell);
            cache_.find(global,values[v]);
          }
          else{
            typedef typename LevelSetGridView::Traits::template Codim<dim>::Entity NodeEntity;
            const NodeEntity node = lCell.template subEntity<dim>(localNodeIndex);
            const Index nodeIndex = domainDef_->levelSetGridView().indexSet().index(node);
            values[v] = Base::levelSetValues_[nodeIndex];
          }
          assert(values[v]); // Zero values should have been removed
        }
        return getSign(values);
      }

    private:
      const DomainDefinition * domainDef_;
      mutable Cache cache_;
    };


    // A stub which is used if the level set function does not fulfill
    // the unfitted geometry function interface
    template< typename DomainDefinition, typename LevelSetFunction>
    class UniqueLevelSetBackend
    <DomainDefinition,LevelSetFunction,
     typename std::enable_if<LevelSetFunction::unfitted_geometry_function>::type> :
      public LevelSetBackendNodeStorageStub<DomainDefinition>{
    public:
      typedef typename LevelSetBackendNodeStorageStub<DomainDefinition>::Traits Traits;
      UniqueLevelSetBackend
        (const DomainDefinition & , const LevelSetFunction & , const typename Traits::Index){}
    };

  }
}

#endif
