// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_MC33TRIANGULATION_VOLUMEMERGER_HH
#define DUNE_MC33TRIANGULATION_VOLUMEMERGER_HH

#include <dune/udg/mc33triangulation/merger.hh>

namespace MarchingCube33SubTriangulationImp {

  template<typename GV>
  class MinVolumeMerger : public MaxAreaMergerBase<GV> {

  private:
    typedef MaxAreaMergerBase<GV> Base;

    // Index types
    typedef typename GV::IndexSet::IndexType Index;
    typedef std::list<Index> IndexList;

    // Continuous field type
    typedef typename GV::ctype DomainField;

    // Grid view types
    typedef typename GV::Traits::template Codim<0>::Entity CellEntity;

    // Check whether the given list contains domain cells which are
    // too small
    template<typename DomainCellList>
    bool containsSmallDomainCells(const DomainCellList & domainCells) const
    {
      typedef typename DomainCellList::const_iterator Iterator;
      Iterator it = domainCells.begin();
      for(; it!=domainCells.end(); ++it) {
        if(it->volume < minEntityPartVolume_)
          return true;
      }
      return false;
    }

  public:

    //! Constructor
    MinVolumeMerger(const GV & _gv, DomainField _minEntityPartVolume, bool _verbose = true) :
      Base(_gv), minEntityPartVolume_(_minEntityPartVolume), verbose_(_verbose)
    {}

    template<typename Layers>
    void mergeAllLayers(Layers & layers)
    {}

    /** \brief Globally merge domain cells to fulfill the minimum
       volume criterion.
     */
    template<typename HostCellMap>
    void merge(HostCellMap & hostCellMap)
    {
      smallDomainCellHosts_.clear();

      if (verbose_)
        std::cout << "Merge domain cells to achieve minimum volume of "
                  << minEntityPartVolume_ << std::endl;

      // Collect the small domain cells
      for(Index c=0; c<hostCellMap.size(); ++c) {
        if(containsSmallDomainCells(hostCellMap[c]))
          smallDomainCellHosts_.push_back(c);
      }

      bool anyMerge;
      do {
        anyMerge = false;

        // Try to merge as much as possible
        typename IndexList::iterator sit = smallDomainCellHosts_.begin();
        for(; sit != smallDomainCellHosts_.end(); ++sit) {
          typedef typename HostCellMap::value_type DomainCellList;
          DomainCellList & domainCells = hostCellMap[*sit];

          bool smallCells = false;
          // Iterate all domain cells
          typename DomainCellList::iterator dcit = domainCells.begin();
          while(dcit != domainCells.end()) {

            // Check domain cell size and try to merge if necessary
            if(dcit->volume < minEntityPartVolume_) {
              smallCells = true;

              typedef typename DomainCellList::value_type DomainCell;
              DomainCell * mergeCandidate = this->mergeWithNeighbors(hostCellMap,*dcit);
              if( mergeCandidate ) {
                this->mergeAndRemove(*dcit,*(mergeCandidate),hostCellMap);
                anyMerge = true;
                dcit = domainCells.begin();
                continue;
              }
            }

            dcit++;
          }

          // Remove from list if it does not contain small cells
          // anymore
          if(!smallCells) {
            sit = smallDomainCellHosts_.erase(sit);
            if(sit != smallDomainCellHosts_.begin())
              sit--;
          }

        } // sit
      } while(anyMerge); // merging loop

      // TODO: Add some handling of small cells that could not be
      // merged.
    }

  private:
    IndexList smallDomainCellHosts_;
    const DomainField minEntityPartVolume_;
    const bool verbose_;
  };


}
#endif
