// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_MCTRIANGULATION_HH
#define DUNE_MCTRIANGULATION_HH

#include <list>
#include <vector>
#include <memory>
#include <map>

#if ( defined HAVE_DUNE_MC or defined HAVE_DUNE_TPMC )

#include <dune/udg/vtriangulation.hh>

#include <dune/udg/common/cuboid.hh>
#include <dune/udg/common/entitypart.hh>
#include <dune/udg/common/intersectionpart.hh>

#include <dune/common/bitsetvector.hh>
#include <dune/common/version.hh>
#if !DUNE_VERSION_NEWER(DUNE_COMMON,2,4)
#include <dune/common/static_assert.hh>
#endif
#include <dune/common/float_cmp.hh>
#include <dune/grid/common/mcmgmapper.hh>

#include <dune/udg/common/utilities.hh>
#include <dune/udg/disc/functions.hh>
#include <dune/udg/disc/p1function.hh>
#include <dune/tpmc/marchingcubes.hh>
#include "mctriangulation/vertexcoloring.hh"

/**
   \todo interpolate sollte static werden und das AnalyticLevelSet
   sollte ausserhalb von MarchingCubeSubTriangulation leben.

   \todo umstellen auf dune-mc

   \todo Muss BBox(..) virtual sein?

 */

//! Layout class used for the default MarchingCubeSubTriangulation Mapper
/**
 * This Layout just selects all Entities from one particular Dimension -- this
 * would usually the Grid dimension
 *
 * @tparam dim The dimension of the Geometries to select
 */
template<int dim>
struct MarchingCubeSubTriangulationLayout {
  bool contains (Dune::GeometryType gt) {
    return dim == gt.dim();
  }
};

//! Class implementing a triangulation
/**
 * @tparam GridViewT Type of GridView for the degrees of freedom
 * @tparam M         Mapper used to count the elements
 */
template<class GridViewT,
    typename M = Dune::MultipleCodimMultipleGeomTypeMapper
                 <GridViewT, MarchingCubeSubTriangulationLayout> >
class MarchingCubeSubTriangulation : public VirtualSubTriangulation<GridViewT>
{
public:
  bool force_refinement;

  typedef GridViewT GridView;
  typedef typename GridView::Grid Grid;

  enum { dimension=Grid::dimension /*!<dimension of the grid we are using*/ };
  enum { dimensionworld=Grid::dimensionworld /*!<dimension of the world*/ };
  typedef typename Grid::ctype ctype;
  Dune::GeometryType CubeType;
  //! LevelSet is defined by a projection from a C0 function.
  typedef Dune::C0GridFunction<Grid, ctype, 1> AnalyticLevelSet;
  typedef Dune::BitSetVector<2> DomainBitSetVector;
  typedef std::vector<DomainBitSetVector> DomainInformation;

  //! LevelSet is stored as a P1 Function on the leaf grid
  typedef Dune::LeafP1Function<Grid, ctype> LevelSet;

  //! The native unfitted geometry classes
  /** @{ */
  typedef typename Dune::UDG::EntityPartInterface<GridView> EntityPartInterface;
  typedef typename EntityPartInterface::EntityPartList EntityPartList;
  typedef typename Dune::UDG::DefaultEntityPartGeometry<GridView> EntityPartGeometry;
  typedef typename Dune::UDG::DefaultEntityPart<EntityPartGeometry> EntityPart;
  typedef typename EntityPartInterface::EntityPartPointer EntityPartPointer;

  typedef Dune::UDG::AffineIntersectionPartGeometry<GridView>
  IntersectionPartGeometry;
  typedef Dune::UDG::DefaultIntersectionPart<IntersectionPartGeometry>
  IntersectionPart;
  typedef Dune::UDG::IntersectionPartInterface<GridView> IntersectionPartInterface;
  typedef typename IntersectionPart::IntersectionPartList IntersectionPartList;
  typedef typename IntersectionPartInterface::IntersectionPartPointer IntersectionPartPointer;
  typedef typename IntersectionPartList::iterator IntersectionPartIterator;
  typedef typename IntersectionPartList::const_iterator IntersectionPartConstIterator;
  typedef Dune::Cuboid<dimension, ctype> Cuboid;
  typedef Dune::GridCuboid<Grid> BoundingBox;
  /** @} */

  //! The mapper type
  typedef M Mapper;

  //! List used to store domain indices as return type
  typedef std::list<int> DomainIndexList;
  //! Tupel to store domain indices (corresponding to an
  //! intersection)
  typedef Dune::FieldVector<int,2> DomainIndexTupel;

protected:

  enum { dim = Grid::dimension /*!< dimension of the grid we are using */ };
  enum { nv = 1<<dim /*!< number of vertices in the cube of dimension */ };
  enum { dimw = Grid::dimensionworld /*!< dimension of the world */ };

  //! access types exported by the grid view
  /** @{ */
  typedef typename GridView::IntersectionIterator IntersectionIterator;
  typedef typename Grid::LeafIntersectionIterator LeafIntersectionIterator;
  typedef typename IntersectionIterator::Intersection Intersection;
  typedef typename LeafIntersectionIterator::Intersection LeafIntersection;
  typedef typename GridView::template Codim<0>::Iterator ElementIterator;
  typedef typename GridView::template Codim<dim>::Iterator VertexIterator;
  typedef typename Grid::template Codim<0>::Entity Entity;
  typedef typename Grid::HierarchicIterator HierarchicIterator;
  /** @} */

  // Vector types defined for convenience
  typedef Dune::FieldVector<ctype, dimw> Point;
  typedef Dune::FieldVector<ctype, dimw-1> FacePoint;
  typedef Dune::FieldVector<ctype, dim> GridPoint;
  typedef Dune::FieldVector<ctype, dim-1> GridFacePoint;
  typedef Dune::FieldVector<ctype, dimw> Normal;
  typedef std::vector< Dune::FieldVector<double,dim> > Vertices;
  typedef std::vector< Dune::FieldVector<double,dim-1> > FaceVertices;

  //! Container to store the level set functions. Necessary because
  //! P1 functions may not be stored in STL vectors.
  class LevelSetContainer {
  private:
    typedef LevelSet* Pointer;
    std::vector<Pointer> p_levelsets;
    const size_t size;
  public:
    LevelSetContainer(const Grid & g, const int _size)
      : size(_size)
    {
      for(size_t i=0; i<size; ++i)
        p_levelsets.push_back(new LevelSet(g));
    }

    LevelSet & operator[](const size_t i)
    {
      assert(i<size);
      return *(p_levelsets[i]);
    }

    const LevelSet & operator[](const size_t i) const
    {
      assert(i<size);
      return *(p_levelsets[i]);
    }

    int getSize() const
    {
      return size;
    }

    LevelSetContainer & operator=(const LevelSetContainer & l)
    {
      assert(size == l.getSize());
      for(int i=0; i<size; ++i)
        *(*(p_levelsets[i])) = *(l[i]);
      return *this;
    }

    ~LevelSetContainer()
    {
      for(size_t i=0; i<size; ++i)
        delete p_levelsets[i];
    }
  };

  //! The levelsets describing the geometries
  mutable LevelSetContainer levelsets;

  //! The gridview describing the grid on which the basis and test
  //! space are defined with regard to the geometric constraints
  GridView & gridview;

  //! Store which domains intersect with each grid cell
  typedef DomainBitSetVector::value_type BitSet;
  typedef DomainBitSetVector::reference BitSetRef;
  typedef DomainBitSetVector::const_reference ConstBitSetRef;
  DomainInformation domain_flags;
  std::vector<DomainBitSetVector> bbox_patterns;
  mutable std::vector<bool> sticky_bits;

  std::vector<DomainBitSetVector> inside_patterns;

  //! Store the bounding box for each grid cell in the given grid
  //! view. First index is the domain. Second index is the cell
  //! index (from mapper).
  std::vector< BoundingBox > bboxes;
  //! Store a unique index for each bounding box
  std::vector<std::vector< int > > bbox_indices;

  //! Cache for edge norms
  typedef std::map<int,ctype> CMap;
  mutable std::vector< CMap > edge_norms;

  //!
  double eqEpsilon;
  static const double default_eqEpsilon() { return 1e-8; }
  static const double face_epsilon() { return 1e-6; }
  static const double levelset_epsilon() { return 1e-10; }

  //!
  unsigned int n_levelsets;
  unsigned int n_domains;
  std::vector<bool> is_virtual_interface;

  //! Helper Functions
  /** @{ */

  double evaluateLevelset(const LevelSet & ls, const Entity& ep, const Point & x) const
  {
    if(!ep.isLeaf()) {
      int level = gridview.grid().maxLevel();
      HierarchicIterator hit = ep.hbegin(level);
      HierarchicIterator hend = ep.hend(level);
      for (; hit!=hend; ++hit)
      {
        if (hit->isLeaf())
          return ls.evallocal(0, *hit, x);
      }
    }
    return ls.evallocal(0, ep, x);

  }

  //! Returns a Dune::GeometryType corresponding to the vertices in
  //! sub_coords. In 1D it will return a cube rather than a simplex.
  template<class VertexType, int this_dim>
  Dune::GeometryType get_subentity_type(std::vector<VertexType> & sub_coords) const
  {
    // Get geometry type of subentity
    Dune::GeometryType subentity_type;
    const int n_nodes = sub_coords.size();
    if(n_nodes == ( 1 << this_dim ))
      subentity_type.makeCube(this_dim);
    else if(n_nodes == this_dim+1)
      subentity_type.makeSimplex(this_dim);
    else if(n_nodes == 5)
      subentity_type.makePyramid();
    else
      DUNE_THROW(Dune::Exception, "Strange Element");
    return subentity_type;
  }

  //! Check whether sub entity with domain pattern dbsv might be in
  //! any domain. For this test to be sufficient, all the bit sets
  //! of the corresponding DomainBitSetVector have to be checked
  inline bool in_any_domain(const DomainBitSetVector & dbsv,
                            const int p, const int max_index) const
  {
    bool r = true;
    // const int debugo = get_domain_tag(dbsv);
    for(int l=0; l<=max_index; ++l) {

      BitSet tmp = dbsv[l];
      tmp ^= inside_patterns[p][l];
      const bool is_it = !tmp.none();

      r = r && is_it;
    }
    return r;
  }

  inline bool in_any_domain(const DomainBitSetVector & given_domain,
                            int max_index) const
  {
    bool inside = false;
    for(unsigned int p=0; p<inside_patterns.size(); ++p) {
      inside = in_any_domain(given_domain,p,max_index) || inside;
    }
    return inside;
  }

  //! Check whether domain is of interest. This is the case if the
  //! domain flags of the cell with mapper index == id indicate that
  //! it contains part of one of the domains.
  inline bool in_any_domain(const int id) const
  {
    return in_any_domain(domain_flags[id],n_levelsets-1);
  }


  //! Check whether entity with index id is intersected by any of
  //! the levelsets
  inline bool is_intersected(const int id) const
  {
    bool r = false;
    for(unsigned int l=0; l<n_levelsets; ++l)
      r = is_intersected(id,l) || r;
    return r;
  }

  //! Check whether entity with index id is intersected by any of
  //! the levelsets
  inline bool is_intersected(const int id, const int index) const
  {
    BitSet d = domain_flags[id][index];
    d.flip();
    return d.none();
  }

  //! The marching cube reconstructs volume of positive domain. If
  //! bitset indicates the domain of interest, then get_sign will
  //! return the appropriate sign factor.
  inline ctype get_sign(const ConstBitSetRef & bitset) const
  {
    return bitset[0] ? -1.0 : 1.0;
  }



  //! Get the index of the domain associated with the
  //! DomainBitSetVector v. If v belongs to no domain, then -1 is
  //! returned.
  int get_domain_index(const DomainBitSetVector & v) const
  {
    for(unsigned int p=0; p<n_domains; ++p) {
      if(in_any_domain(v,p,n_levelsets-1))
        return p;
    }
    return -1;
  }

  //! Get the interface index corresponding to the
  //! DomainBitSetVector of a given boundary
  const int get_interface_index(const DomainBitSetVector & v) const
  {
    for(unsigned int l=0; l<n_levelsets; ++l) {
      BitSet pattern = v[l];
      pattern.flip();
      if(pattern.none())
        return l;
    }
    return -1;
  }
  inline const int get_interface_index(const int domain_tag) const
  {
    return get_interface_index(get_domain_flags(domain_tag));
  }


  //! Convert domain tag to domain flags
  inline DomainBitSetVector get_domain_flags(const int domain_tag) const
  {
    DomainBitSetVector v(n_levelsets);
    for(unsigned int l=0; l<n_levelsets; ++l) {
      for(int p=0; p<2; ++p) {
        const int current_bit = 1 << (2*l+p);
        v[l][p] = current_bit & domain_tag;
      }
    }
    return v;
  }



  template <int vertexdim, int typedim>
  void normalize_sub_entity(const Dune::GeometryType & type,
                            std::vector<ctype> & v,
                            const std::vector<Dune::FieldVector<ctype,vertexdim> > & c
                            )
  const
  {
    const ctype sub_normalization_epsilon = default_eqEpsilon();

    const Dune::ReferenceElement<ctype,typedim> &
    refelem =
      Dune::ReferenceElements<ctype,typedim>::general(type);

    const int edge_codim = type.dim()-1;
    for (int i=0; i<refelem.size(edge_codim); i++) {
      const unsigned int l = refelem.subEntity (i, edge_codim, 0, typedim);
      const unsigned int r = refelem.subEntity (i, edge_codim, 1, typedim);

      double & vl = v[l];
      double & vr = v[r];

      if(vl==vr)
        continue;

      assert(v.size() == c.size());
      assert(l<c.size() && r < c.size());

      const Dune::FieldVector<ctype,vertexdim> & pl = c[l];
      const Dune::FieldVector<ctype,vertexdim> & pr = c[r];

      Dune::FieldVector<ctype,vertexdim> x;
      for (int d=0; d<vertexdim; d++)
        x[d] = pl[d] - (vl * (pr[d] - pl[d])) / (vr - vl);

      const ctype dist_xl = (x-pl).two_norm2();
      const ctype dist_xr = (x-pr).two_norm2();

      const ctype square_epsilon
        = sub_normalization_epsilon * sub_normalization_epsilon;

      if(dist_xl < square_epsilon && dist_xl < dist_xr)
        vl = 0.0;

      if(dist_xr < square_epsilon && dist_xr < dist_xl)
        vr = 0.0;
    }
  }

  //! Check if levelset value if zero for all vertices in c. Then
  //! this is not an interior face, but a sticky boundary
  bool is_sticky_boundary(const LeafIntersection & it, const FaceVertices & c,
                          const LevelSet & current_levelset) const
  {
    // Get levelset values in vertices given by c. These coordinates
    // are understood to be local to the levelset entity in ls_entity
    ctype v(0);
    for (unsigned int i=0; i<c.size(); i++) {
      const GridPoint c_g = it.geometryInInside().global(c[i]);
      v += std::abs(current_levelset.evallocal(0, *(it.inside()), c_g));
    }
    return v < levelset_epsilon();
  }

  //! Check if all the vertices in c are located on a face of e which
  //! is neighbor to a cell with a different ancestor on the dof
  //! gridview. Then this is a sticky boundary and we have to
  //! determine the id of the neighbor. If sticky boundaries coincide
  //! with the grid boundary, this function will return false!! This
  //! is intended.
  bool is_sticky_boundary(const Entity & e, const Vertices & c,
                          int & n_id, Entity & f) const
  {
    using namespace Dune;
    LeafIntersectionIterator iit = e.ileafbegin();
    LeafIntersectionIterator iend = e.ileafend();
    for (; iit != iend; ++iit)
    {
      const GridPoint center = iit->geometryInInside().global(GridFacePoint(0.5));
      for(int d=0; d<dim; ++d) {
        const bool not_zero =
          !FloatCmp::eq<ctype,FloatCmp::absolute>(center[d],0.0,eqEpsilon);
        const bool not_one =
          !FloatCmp::eq<ctype,FloatCmp::absolute>(center[d],1.0,eqEpsilon);
        if(not_zero && not_one)
          continue;

        const GridPoint globalcenter = iit->geometry().global(GridFacePoint(0.5));

        bool is_inside = true;
        for(unsigned int i=0; i<c.size(); ++i)
          is_inside =
            is_inside && FloatCmp::eq<ctype,FloatCmp::absolute>
              (globalcenter[d],c[i][d],eqEpsilon);

        if(is_inside) {

          // Sticky boundary on grid boundary
          if(!iit->neighbor())
            return false;

          f = iit->outside();
          const Entity a_f = ancestor(f);
          const Entity a_e = ancestor(e);
          if(a_f != a_e) {
            f = a_f;
            n_id = mapper.map(a_f);
            return true;
          }
          else
            return false;
        }
      }       // d
    }
    return false;
  }

  /**
     Calculate the bounding box bbox for a given element e and set the
     domains bitset to indicate the domains intersected by e. The entity
     e may be geometrically resolved by a number of smaller entities
     hierarchically embedded in e and given in childlist. Notice, that
     domain[d] will be set true even if the domain boundary to domain
     number d does only touch but not truely intersect with e.
   */
  void calcBBox
  (
    const Entity & e,
    const DomainBitSetVector & this_bbox_pattern,
    std::list<EntityPointer> & childlist,
    BoundingBox& bbox,
    DomainBitSetVector & domains
  ) const;

  void calcBBox
  (
    const Entity & e,
    const DomainBitSetVector & this_bbox_pattern,
    BoundingBox& bbox,
    DomainBitSetVector & domains
  ) const
  {
    std::list<Entity> childlist;
    children(e, childlist);
    calcBBox(e,this_bbox_pattern,childlist,bbox,domains);
  }

  //! Disable copy constructor
  MarchingCubeSubTriangulation(const MarchingCubeSubTriangulation &);
  //! Get entity in grid view which contains entity e
  Entity ancestor(const Entity & e) const;
  //! Check equality of two point with in eqEpsilon error range
  bool pointEq(const Point & a, const Point & b) const;
  Normal normal(const Entity & e, const Point & p, const int l) const;
  //!
  void children(const Entity & e,
                std::list<Entity> & children) const;
  /** @} */

  //! Functions to create the native geometry objects for a given
  //! vector of vertices and the corresponding bounding boxes
  /** @{ */
  EntityPartPointer
  createEntityPart(const Entity & entity,
                   const BoundingBox & dof_bbox,
                   const std::vector< Dune::FieldVector<double,dim> > & c,
                   const int domain_tag, const int domain_index,
                   const double volume) const;

  IntersectionPartPointer
  createIntersection(const BoundingBox & bbox, const BoundingBox & n_bbox,
                     const std::vector< Dune::FieldVector<double,dim> > & c,
                     const int bid,
                     const Normal & n,
                     const int intersection_tag,
                     const int inside_domain_index,
                     const int outside_domain_index) const;

  IntersectionPartPointer
  createIntersection(const BoundingBox & bbox,
                     const std::vector< Dune::FieldVector<double,dim> > & c,
                     const int bid,
                     const Normal & n,
                     const int intersection_tag,
                     const int inside_domain_index,
                     const int outside_domain_index) const;
  /** @} */

private:
  //! Store an automatically deleted pointer to the mapper when we
  //! created it ourselves
  const std::auto_ptr<Mapper> mapperP;

protected:
  //! Mapper used to get indices of the given grid view
  Mapper &mapper;

  void bbox_intersection(Point & gmax, Point & gmin, const Point & max, const Point & min) const
  {
    for(int d=0; d<dim; ++d) {
      gmax[d] = gmax[d] < max[d] ? gmax[d] : max[d];
      gmin[d] = gmin[d] > min[d] ? gmin[d] : min[d];
    }
  }

  void edge_norm_indices
    (const IntersectionPart & intersection, int & e_bbox_index, int & f_bbox_index) const
  {
    const int e_domain_index = intersection.insideDomainIndex();
    e_bbox_index = getBBoxIndex( e_domain_index,*( intersection.boundingBoxInside().entity() ) );

    const int f_domain_index = intersection.outsideDomainIndex();
    if(f_domain_index == -1) {
      const int bid = intersection.boundarySegmentIndex();
      const bool unfitted_boundary =
        bid == Dune::UDG::BoundaryFlags::internal ||
        bid == Dune::UDG::BoundaryFlags::processor;

      if( unfitted_boundary )
        f_bbox_index = e_bbox_index;
      else
        f_bbox_index = getNumberOfBBoxes() + bid;
    }
    else
      f_bbox_index = getBBoxIndex(f_domain_index,*(intersection.boundingBoxOutside().entity()) );
  }

public:

  //! Fuzzy factor to avoid the creation of sub entities of sizes
  //! near zero
  double edgeEpsilon;
  static const double default_edgeEpsilon() { return 1e-3; }

  //! Rebuilds the bboxes and domains arrays
  void updateCaches();

  //! This computes the domain index from a domain tag, as might be
  //! provided by a sub entity. Domain tag must refer to one and
  //! only one domain. NO checking is done.
  inline int getDomainIndex(const int & domain_tag) const
  {
    DomainBitSetVector v = get_domain_flags(domain_tag);
    return get_domain_index(v);
  }

  //! Check whether codim 0 entity intersects with domain of index
  //! domain_index.
  inline bool isInDomain(const Entity & e, const int & domain_index) const
  {
    const int eid = mapper.map(e);
    return isInDomain(eid,domain_index);
  }
  inline bool isInDomain(const int & eid, const int & domain_index) const
  {
    const DomainBitSetVector & flags = domain_flags[eid];
    return in_any_domain(flags,domain_index,n_levelsets-1);
  }

  //! Provides the domain indices of domains which intersect with
  //! cell of mapper index == eid. It is a simple application of
  //! isInDomain(..) and provided for convenience.
  inline DomainIndexList getDomainIndices(const Entity & e) const
  {
    const int eid = mapper.map(e);
    return getDomainIndices(eid);
  }
  inline DomainIndexList getDomainIndices(const int & eid) const
  {
    DomainIndexList ei;
    for(unsigned int p=0; p<n_domains; ++p)
      if(isInDomain(eid,p))
        ei.push_back(p);
    return ei;
  }

  //! Compute edge norms according to (merged) elements of the
  //! sub triangulation.
  void cacheEdgeNorms() const;

  //! Returns a consecutive index for the bounding box corresponding
  //! to the given domain and entity. Unless the bounding boxes were
  //! merged, this index is equal to the index returned by the cell
  //! mapper.
  inline int getBBoxIndex(const int domain_index, const int eid) const
  {
    return bbox_indices[domain_index][eid];
  }
  inline int getBBoxIndex(const int domain_index, const Entity & e) const
  {
    const int eid = mapper.map(e);
    return getBBoxIndex(domain_index,eid);
  }

  //! This function checks whether a valid bbox index
  //! exists.
  inline bool hasBBoxIndex(const int domain_index, const int eid) const
  {
    return bbox_indices[domain_index][eid] >= 0;
  }
  inline bool hasBBoxIndex(const int domain_index, const Entity & e) const
  {
    return hasBBoxIndex(domain_index,mapper.map(e));
  }

  //! Returns the total number of bboxes including all domains.
  int getNumberOfBBoxes() const
  {
    return bboxes.size();
  }

  //! Compute norms for a list of sub intersections, such that the
  //! lengths of intersections adjacent to the same cell or boundary
  //! are accumulated.
  ctype getEdgeNorm(const int bbox_index_e, const int bbox_index_f) const
  {
    assert(edge_norms.size() > 0);
    const ctype edge_norm = edge_norms[bbox_index_e][bbox_index_f];
    assert(std::isfinite(edge_norm));
    assert(edge_norm > 0);
    return edge_norm;
  }
  ctype getEdgeNorm(const IntersectionPartInterface & intersection) const
  {
    int e_index(0);
    int f_index(0);
    const IntersectionPart & intersectionPart
      (static_cast<const IntersectionPart&>(intersection));
    edge_norm_indices(intersectionPart,e_index,f_index);
    return getEdgeNorm(e_index,f_index);
  }

  //! This algorithm merges the bounding boxes according to some
  //! algorithm which is provided by a static member function of
  //! Merge Algorithm.
  template<class MergeAlgorithm>
  void mergeBBoxes(const ctype min_volume_ratio, const bool squarify = true)
  {
    // The merger requires the edge norms
    cacheEdgeNorms();

    MergeAlgorithm::mergeBBoxes(*this, bboxes, bbox_indices,
                                gridview, mapper,
                                min_volume_ratio);

    //! Transform all bounding boxes to be perfect squares
    if(squarify) {
      for(std::size_t i=0; i<bboxes.size(); ++i) {
        BoundingBox & bbox = bboxes[i];

        // Extract bbox parameters
        Dune::FieldVector<ctype,dim> origin = bbox.BoundingBox::Base::origin();
        Dune::FieldVector<ctype,dim> height = bbox.BoundingBox::Base::height();
        Entity ep = bbox.entity();

        // Compute max length of bbox
        ctype max_height(0);
        for(int d=0; d<dim; ++d)
          max_height = std::max(height[d],max_height);

        // Create new quadratic bbox with max length
        height = max_height;
        bbox = BoundingBox(ep,origin,height);
      }
    }

    // Recompute edge norms after bboxes have changed
    cacheEdgeNorms();
  }

  //! Assignment operator
  MarchingCubeSubTriangulation & operator=
    (const MarchingCubeSubTriangulation & mcst)
  {
    std::cout << "Copy MC Sub Triangulation" << std::endl;
    edgeEpsilon = mcst.edgeEpsilon;
    eqEpsilon = mcst.eqEpsilon;

    levelsets = mcst.levelsets;

    domain_flags = mcst.domain_flags;
    bbox_patterns = mcst.bbox_patterns;
    sticky_bits = mcst.sticky_bits;

    inside_patterns = mcst.inside_patterns;
    bboxes = mcst.bboxes;
    bbox_indices = mcst.bbox_indices;

    n_domains = mcst.n_domains;
    n_levelsets = mcst.n_levelsets;
    is_virtual_interface = mcst.is_virtual_interface;

    gridview = mcst.gridview;

    force_refinement = mcst.force_refinement;

    edge_norms = mcst.edge_norms;

    return *this;
  }

  const int get_domain_tag(const DomainBitSetVector & v) const
  {
    int tag(0);
    const int nbits = 2;
    for(unsigned int l=0; l<n_levelsets; ++l) {
      for(int b=0; b<nbits; ++b) {
        const int trigger = v[l][b];
        tag +=  (1 << (l*nbits+b)) * trigger;
      }
    }
    return tag;
  }

  //! The domain_patterns holds only the elements of bbox_patterns
  //! which are visible. This allows a faster processing without a
  //! lot of if-branching in the code
  void update_domain_patterns()
  {
    // Create inside pattern
    inside_patterns.clear();
    inside_patterns.resize(n_domains);
    for(unsigned int p=0; p<n_domains; ++p) {

      inside_patterns[p].resize(n_levelsets,false);
      for(unsigned int l=0; l<n_levelsets; ++l) {

        // Extend inside pattern
        inside_patterns[p][l] = bbox_patterns[p][l];
        inside_patterns[p][l].flip();
      }
    }
  }

  //! This adds a domain of interest for which sub entities are
  //! created. The size of pattern must be equal to the number of
  //! level set functions given in the constructor. The bits denote
  //! for each level set function whether its negative (first bit)
  //! or positive (second bit) part is contained in this domain.
  int setDomainPattern(const DomainBitSetVector & pattern)
  {
    bbox_patterns.push_back(pattern);
    n_domains = bbox_patterns.size();

    assert(pattern.size() == n_levelsets);

    update_domain_patterns();

    return bbox_patterns.size()-1;
  }

  //! An interface may be set invisible. Then it will be ignored
  //! when calling get_boundaries()
  void setInterfaceVisibility(const int interface_index, const bool visible)
  {
    is_virtual_interface[interface_index] = !visible;
  }

  const LevelSet & levelset(const int l) const {
    return levelsets[l];
  }

  /**
   * @param g The GridView to triangulate
   *
   * The mapper used is created via new from g's grid and indexset and
   * deleted when the object is destroyed
   */
  MarchingCubeSubTriangulation(GridView & g) :
    force_refinement(false),
    CubeType(Dune::GeometryType::cube,dim),
    levelsets(g.grid(),1), gridview(g),
    eqEpsilon(default_eqEpsilon()), n_levelsets(1),
    n_domains(0),
    is_virtual_interface(n_levelsets,false),
    mapperP(new Mapper(g)), mapper(*mapperP),
    edgeEpsilon(default_edgeEpsilon())
  {
    DomainBitSetVector bbox_pattern(1);
    bbox_pattern[0][1]=1;
    bbox_pattern[0][0]=0;
    setDomainPattern(bbox_pattern);
  }

  /**
   * @param g The GridView to triangulate
   * @param l Number of level set functions
   *
   * The mapper used is created via new from g's grid and indexset and
   * deleted when the object is destroyed.
   */
  MarchingCubeSubTriangulation(GridView & g, int l) :
    force_refinement(false),
    CubeType(Dune::GeometryType::cube,dim),
    levelsets(g.grid(),l), gridview(g),
    eqEpsilon(default_eqEpsilon()), n_levelsets(l),
    n_domains(0),
    is_virtual_interface(n_levelsets,false),
    mapperP(new Mapper(g)), mapper(*mapperP),
    edgeEpsilon(default_edgeEpsilon()) {}


  /**
   * @param g The GridView to triangulate
   * @param m The mapper to use.
   *
   * With this constructor a reference to the mapper m is stored in the
   * object.  The mapper is not deleted when the object is destroyed.
   */
  MarchingCubeSubTriangulation(GridView & g, Mapper &m) :
    force_refinement(false),
    CubeType(Dune::GeometryType::cube,dim),
    levelsets(g.grid(),1), gridview(g),
    eqEpsilon(default_eqEpsilon()), n_levelsets(1),
    n_domains(0),
    is_virtual_interface(n_levelsets,false),
    mapper(m),
    edgeEpsilon(default_edgeEpsilon())
  {
    DomainBitSetVector bbox_pattern(1);
    bbox_pattern[0][1]=1;
    bbox_pattern[0][0]=0;
    setDomainPattern(bbox_pattern);
  }

  ~MarchingCubeSubTriangulation() {};

  const GridView & gridView() const { return gridview; }

  // Enable / disable cube refinement
  bool & forceRefinement() { return force_refinement; };

  virtual
  const BoundingBox BBox(const unsigned int key, const Entity & e) const;

private:

  //! Coarsen levelset grid to reduce assembling time if adapt is
  //! true. Otherwise "just" update caches
  void postinterpolate(bool adapt=false)
  {

    // #define YES_I_UPDATED_THIS_ALGORITHM_TO_WORK_WITH_A_GRID_VIEW
#ifdef YES_I_UPDATED_THIS_ALGORITHM_TO_WORK_WITH_A_GRID_VIEW
    bool todo = adapt;
    while (todo)
    {
      todo = false;

      grid.preAdapt();
      levelset.preAdapt();

      for (LeafIterator it=grid.template leafbegin<0>();
           it!=grid.template leafend<0>(); ++it)
      {
        int inside = 0;
        int outside = 0;
        Dune::GeometryType type = it->type();
        int corners =
          Dune::ReferenceElements<ctype,dim>::general(type).size(dim);
        for (int c=0; c<corners; c++) {
          Point p =
            Dune::ReferenceElements<ctype,dim>::general(type).position(c,dim);
          if (isInside(it, p))
            inside++;
          else
            outside++;
        }
        if ((inside == 0 || outside == 0) && ! mapper.contains(*it))
        {
          grid.mark(-1,it);
        }
      }

      /* work around a uggrid bug */
      unsigned int old = mapper.size();
      todo = grid.adapt();
      mapper.update();
      todo = (mapper.size() != old);

      // resize the levelset
      Dune::P1FunctionManager<Grid, ctype> manager(grid);
      levelset.postAdapt(manager);
      grid.postAdapt();

      levelset.interpolate(u);
    }
#endif
    updateCaches();
  }

  //! Get sub entities in local coordinates. These are recursive
  //! functions.
  /** @{ */

  //! Begin of recursion
  template<bool get_boundaries>
  void get_subtriangulation
  (
    const Entity & ls_entity,
    const Dune::GeometryType & type,
    const Vertices & c,
    std::list< std::vector<Vertices> > & all_volumes,
    std::list< DomainBitSetVector > & all_domains,
    std::list< std::vector<Vertices> > & all_boundaries,
    std::list< DomainBitSetVector > & all_boundaries_domains
  ) const;

  //! Recursion step
  template<bool get_boundaries>
  void get_subtriangulation
  (
    const Entity & ls_entity,
    const Dune::GeometryType & type,
    const Vertices & c,
    std::list< std::vector<Vertices> > & all_volumes,
    std::list< DomainBitSetVector > & all_domains,
    std::list< std::vector<Vertices> > & all_boundaries,
    std::list< DomainBitSetVector > & all_boundaries_domains,
    const int interface_index,
    DomainBitSetVector & current_domain
  ) const;


  void get_subentity_boundaries
  (
    const Entity & ls_entity,
    const Dune::GeometryType & type,
    const Vertices & v_coords,
    std::list< std::vector<Vertices> > & all_boundaries,
    std::list< DomainBitSetVector > & all_boundaries_domains,
    DomainBitSetVector current_domain,
    const int interface_index
  ) const;

  void get_boundary_subtriangulation
  (
    const Entity & ls_entity,
    const Dune::GeometryType & type,
    const Vertices & v_coords,
    std::list< std::vector<Vertices> > & all_boundaries,
    std::list< DomainBitSetVector > & all_boundaries_domains,
    DomainBitSetVector & current_domain,
    const int interface_index
  ) const;

  //! Begin of recursion
  void get_edge_subtriangulation
  (
    const int id,
    const Entity & ls_entity,
    const LeafIntersection & it,
    const Dune::GeometryType & type,
    const FaceVertices & c,
    std::list< std::vector<FaceVertices> > & all_c_x,
    std::list< DomainBitSetVector > & all_c_d
  ) const;

  // Recursion step
  void get_edge_subtriangulation
  (
    const int id,
    const Entity & ls_entity,
    const LeafIntersection & it,
    const Dune::GeometryType & type,
    const FaceVertices & c,
    std::list< std::vector<FaceVertices> > & all_c_x,
    std::list< DomainBitSetVector > & all_c_d,
    const int interface_index,
    DomainBitSetVector & current_domain
  ) const;

  /** @} */

  //! Apply marching cube algorithm to obtain volume, interior face
  //! and boundary sub entities
  /** @{ */

  void face_marching_cubes
  (
    const Entity & ls_entity,
    const LeafIntersection & it,
    const Dune::GeometryType & type,
    const FaceVertices & c,
    std::vector< FaceVertices > & c_x,
    const LevelSet & current_levelset,
    const ctype sign
  ) const;

  template <int mcdim>
  void marching_cubes
  (
    const Entity & ls_entity,
    const Dune::GeometryType & type,
    const std::vector< Dune::FieldVector<double,dim> > & c,
    std::vector< std::vector< Dune::FieldVector<double,dim> > > & c_x,
    const LevelSet & current_levelset,
    const ctype sign
  ) const ;

  void boundary_marching_cubes
  (
    const Entity & ls_entity,
    const Dune::GeometryType & type,
    const std::vector< Dune::FieldVector<double,dim> > & c,
    std::vector< std::vector< Dune::FieldVector<double,dim> > > & c_x,
    const LevelSet & current_levelset
  ) const;
  /** @} */


  //! Check which faces can be handled on the dof level, create them
  //! and set the corresponding processed[] flag
  void create_coarse_edges
  (
    const Entity & e,
    IntersectionPartList & edges,
    std::map<int,bool> & processed
  ) const;

  //! Check whether the interface intersections with barycenters
  //! center_a and center_b are the same intersection with regard to
  //! different level set functions.
  bool is_same_intersection(Point center_a, const Point center_b,
                            const DomainBitSetVector & d_a,
                            const DomainBitSetVector & d_b) const
  {
    center_a -= center_b;
    if(center_a.two_norm() < face_epsilon()) {
      for(unsigned int i=0; i<d_a.size(); ++i) {
        BitSet ba = d_a[i]; ba.flip();
        BitSet bb = d_b[i]; bb.flip();
        if(ba.none() != bb.none())
          return true;
      }
    }
    return false;
  }

  //! Determine the face normal vector as well as the domain indices
  //! inside and outside of the interface.
  void determine_intersection(const Entity & e,
                              const DomainBitSetVector & domain,
                              const Point center,
                              Normal & face_normal,
                              int & inside_index,
                              int & outside_index) const
  {
    // Find out the indices of all interfaces coinciding in this
    // intersection.
    std::vector<int> interface_indices;
    for(unsigned int i=0; i<domain.size(); ++i) {
      BitSet bd = domain[i]; bd.flip();
      if(bd.none())
        interface_indices.push_back(i);
    }

    assert(interface_indices.size() != 0);

    face_normal = normal(e,center,interface_indices[0]);
    face_normal *= 1.0 / face_normal.two_norm();

    DomainBitSetVector i_domain(domain);
    DomainBitSetVector o_domain(domain);
    i_domain[interface_indices[0]][1] = true;
    i_domain[interface_indices[0]][0] = false;
    o_domain[interface_indices[0]][0] = true;
    o_domain[interface_indices[0]][1] = false;

    for(unsigned int i = 1; i<interface_indices.size(); ++i) {
      Normal i_normal = normal(e,center,interface_indices[i]);
      if(i_normal * face_normal >= 0) {
        i_domain[interface_indices[i]][1] = true;
        i_domain[interface_indices[i]][0] = false;
        o_domain[interface_indices[i]][0] = true;
        o_domain[interface_indices[i]][1] = false;
      }
      else{
        i_domain[interface_indices[i]][0] = true;
        i_domain[interface_indices[i]][1] = false;
        o_domain[interface_indices[i]][1] = true;
        o_domain[interface_indices[i]][0] = false;
      }
    }

    inside_index = get_domain_index(i_domain);
    outside_index = get_domain_index(o_domain);
  }

public:

  //! Interpolate a given function on levelset grid
  /** @{ */
  template<class F>
  void interpolate(const F & u, bool adapt=false, int index = 0, bool update = true)
  {
    typedef typename Grid::template Codim<0>::LeafIterator LeafIterator;
    levelsets[index].interpolate(u);
    if(update)
      postinterpolate(adapt);
  }
  template<class F, class VC>
  void interpolate(const F & u, VC & vcseed, bool adapt=false, int index = 0, bool update = true)
  {
    typedef typename Grid::template Codim<0>::LeafIterator LeafIterator;
    levelsets[index].interpolate(u);
    Dune::VertexColoring<typename Grid::LeafGridView> vertexcoloring(gridview.grid().leafView());
    if(update)
      postinterpolate(adapt);
  }
  /** @} */

  //! Access to geometry and domain information
  /** @{ */
  const LevelSet & geom(unsigned int index = 0) const { return levelsets[index]; }
  LevelSet & geom(unsigned int index = 0) { return levelsets[index]; }
  unsigned int nGeoms() const { return n_levelsets; }
  int getDomainIndex(const DomainBitSetVector & dinfo) const { return get_domain_index(dinfo); }
  inline const std::vector<DomainBitSetVector> & domainInfo() const { return domain_flags; }
  const DomainBitSetVector::const_reference & domainInfo(DomainBitSetVector::size_type eid, const int l=0) const
  { return domain_flags[eid][l]; }
  /** @} */

  // create Entities
private:
  void create_rectangles(const Entity & e,
                         std::list<BoundingBox> & rectangles, bool force_ref) const;
  void create_edges(const Entity & e,
                    IntersectionPartList & edges) const;
  void create_boundaries(const Entity & e,
                         IntersectionPartList & boundaries) const;
public:
  void create_entity_parts(const Entity & e,
                           EntityPartList & geometries) const;

  void create_entity_parts(const Entity& e,
                           const Entity& e_fine,
                           EntityPartList& geometries) const;

  void create_intersection_parts(const Entity & e,
                                 IntersectionPartList & intersections) const
  {
    create_edges(e,intersections);
    create_boundaries(e,intersections);
  }

};

#include "mctriangulation/mctriangulation.cc"

#endif // defined HAVE_DUNE_MC or defined HAVE_DUNE_TPMC

#endif

/*
   Local variables:
   c-file-style:"bsd"
   c-indent-level: 4
   c-basic-offset: 4
   indent-tabs-mode: nil
   End:
 */
