// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_MCTRIANGULATION_HH
#include <config.h>
#endif
#include <dune/udg/mctriangulation.hh>
#include <dune/geometry/quadraturerules.hh>
#include <algorithm>
#include <map>

class ThresholdFunctor {
public:
  //! Function to test whether a point is inside the isosurface.
  static bool isInside(double d)
  {
    return d > 0;
  }
  //! distance to iso surface
  static double getDistance(double d)
  {
    return d;
  }
  //! isInside for distance value
  static bool isLower(double d)
  {
    return d > 0;
  }
};

#ifndef MC_DROP_THRESHOLD
#define MC_DROP_THRESHOLD (1e-10)
#endif

template<class GridView, typename M>
const typename MarchingCubeSubTriangulation<GridView, M>::BoundingBox
MarchingCubeSubTriangulation<GridView, M>::BBox(const unsigned int key, const EntityPointer & e) const
{
  return BBox(key,*e);
}

template<class GridView, typename M>
const typename MarchingCubeSubTriangulation<GridView, M>::BoundingBox
MarchingCubeSubTriangulation<GridView, M>::BBox(const unsigned int key, const Entity & e) const
{
  int index;

#ifndef NDEBUG
  const bool contains =
#endif
  mapper.contains(e, index);
  assert(contains);
  assert(key < bbox_patterns.size());

  const int bbox_index = bbox_indices[key][index];
  assert(bbox_index != -1);

  return bboxes[bbox_index];
  //return BoundingBox(e.template subEntity<0>(0), bboxes[bbox_index]);

}

template<class GridView, typename M>
const typename MarchingCubeSubTriangulation<GridView, M>::EntityPointer
MarchingCubeSubTriangulation<GridView, M>::ancestor(const EntityPointer & e) const
{
  int index;
  if (mapper.contains(*e, index))
    return e;
  else
    return ancestor(e->father());
}

namespace std {

  template<class K, int n>
  Dune::FieldVector<K,n>
  min(const Dune::FieldVector<K,n> & a, const Dune::FieldVector<K,n> & b)
  {
    Dune::FieldVector<K,n> c;
    for (int i=0; i<n; i++)
      c[i] = std::min(a[i], b[i]);
    return c;
  }

  template<class K, int n>
  Dune::FieldVector<K,n>
  max(const Dune::FieldVector<K,n> & a, const Dune::FieldVector<K,n> & b)
  {
    Dune::FieldVector<K,n> c;
    for (int i=0; i<n; i++)
      c[i] = std::max(a[i], b[i]);
    return c;
  }

}

template<class GridView, typename M>
void
MarchingCubeSubTriangulation<GridView, M>::
cacheEdgeNorms() const
{
  // Resize cache
  const int n_bboxes = getNumberOfBBoxes();
  edge_norms.resize(0);
  edge_norms.resize(n_bboxes);

  // Iterate all cells
  typedef typename GridView::template Codim<0>::
  template Partition<Dune::All_Partition>::Iterator AllIterator;

  AllIterator e = gridview.template begin<0,Dune::All_Partition>();
  AllIterator eend = gridview.template end<0,Dune::All_Partition>();
  for (; e != eend; ++e)
  {
    const int eid = mapper.map(*e);
    const DomainIndexList & e_domains = getDomainIndices(eid);

    // Shortcut in case the cell is not inside any domain.
    if(e_domains.size() == 0)
      continue;

    // If cell is within multiple domains, then we have to create the
    // edges and boundaries.
    IntersectionPartList edges;
    create_edges(EntityPointer(e), edges);
    create_boundaries(EntityPointer(e), edges);
    IntersectionPartConstIterator fit = edges.begin();
    IntersectionPartConstIterator efit = edges.end();
    for (; fit != efit; ++fit)
    {
      int e_bbox_index(0);
      int f_bbox_index(0);
      IntersectionPart & intersectionPart(static_cast<IntersectionPart&>(**fit));

      edge_norm_indices(intersectionPart,e_bbox_index,f_bbox_index);

      const int f_domain_index = intersectionPart.outsideDomainIndex();
      if(e_bbox_index == f_bbox_index && f_domain_index != -1)
        continue;

      const ctype volume =
        intersectionPart.geometry().integrationElement(GridFacePoint(0.5));

      // Write contribution
      CMap & norm_map = edge_norms[e_bbox_index];
      if(norm_map.find(f_bbox_index) == norm_map.end())
        norm_map[f_bbox_index] = 0;
      norm_map[f_bbox_index] += volume;

    } // fit

  } // e

  // Check edge norms
#ifdef DUNE_UDG_CHECK_EDGES

  // vector which hold the domain index for each bbox
  std::vector< int > domains(bboxes.size());
  for(int d = 0; d < bbox_indices.size(); ++d) {
    for(int id = 0; id < bbox_indices[d].size(); ++id) {
      const int b_index = bbox_indices[d][id];
      if(b_index >= 0)
        domains[b_index] = d;
    }
  }

  std::vector< CMap > sums(n_domains);
  for(int id=0; id < edge_norms.size(); ++id) {
    const int d = domains[id];
    typename CMap::iterator it = edge_norms[id].begin();
    typename CMap::iterator eit = edge_norms[id].end();
    for(; it!=eit; ++it) {
      const int f_id = it->first;
      const int f_v = it->second;

      if(f_id >= bboxes.size()) {
        if(sums[d].count(-1) == 0)
          sums[d][-1] = 0;
        sums[d][-1] += edge_norms[id][f_id];
        continue;
      }

      const int f_d = domains[f_id];

      if(edge_norms[f_id].count(id) == 0 || f_id > id) {
        if(sums[d].count(f_d) == 0)
          sums[d][f_d] = 0;
        sums[d][f_d] += edge_norms[id][f_id];
      }

    }
  }

  std::cout << "Accumulated edge norms:" << std::endl;
  for(int s=0; s < sums.size(); ++s) {
    typename CMap::iterator it = sums[s].begin();
    typename CMap::iterator eit = sums[s].end();
    for(; it!=eit; ++it) {
      std::cout << " Between domain " << s << " and " << it->first << " : "
                << it->second << std::endl;
    }
  }
#endif

}


template<class GridView, typename M>
void
MarchingCubeSubTriangulation<GridView, M>::calcBBox
(
  const EntityPointer & e,
  const DomainBitSetVector & this_bbox_pattern,
  std::list<EntityPointer> & childlist,
  BoundingBox& bbox,
  DomainBitSetVector & domains
) const
{
  const int id = mapper.map(*e);

  ThresholdFunctor tf;

  // Bitset vector with domain information
  domains.resize(n_levelsets);
  domains.unsetAll();

  // Filter out the level set functions which are relevant for the
  // calculation of the bounding box.
  std::vector<int> restricting_levelsets;
  std::vector<double> restricting_sign;
  {
    DomainBitSetVector pattern = this_bbox_pattern;
    for(unsigned int l=0; l<n_levelsets; ++l) {
      pattern[l].flip();
      if(!pattern[l].none()) {
        restricting_levelsets.push_back(l);
        const double sign = pattern[l][0] ? 1.0 : -1.0;
        restricting_sign.push_back(sign);
      }
    }
  }

  Point all_min(-std::numeric_limits<ctype>::max());
  Point all_max(std::numeric_limits<ctype>::max());

  bool default_bbox = true;

  std::vector<Point> min(restricting_levelsets.size());
  std::vector<Point> max(restricting_levelsets.size());
  for(unsigned int r=0; r<max.size(); ++r) {
    min[r]=std::numeric_limits<ctype>::max();
    max[r]=-std::numeric_limits<ctype>::max();
  }

  // Iterate children on levelset level
  typename std::list<EntityPointer>::const_iterator child=childlist.begin();
  for (; child!=childlist.end(); ++child) {

    Dune::GeometryType type = (*child)->type();

#ifndef NO_MC_NORMALIZING
    ctype refSz = edgeEpsilon * (1<<(child->level() - e->level()));
#endif

    const Dune::ReferenceElement<ctype,dim> & rElement
      = Dune::ReferenceElements<ctype,dim>::general(type);

    unsigned int edges = rElement.size(dim-1);
    unsigned int vertices = rElement.size(dim);

    std::vector<Point> coords(vertices, Point(0.0));
    std::vector<std::vector<double> > values(vertices);
    std::vector<int> ids(vertices, 0);

    // Get vertex values, coordinates, and indices
    for (unsigned int v=0; v<vertices; v++)
    {
      coords[v] = rElement.position(v, dim);
      ids[v] = levelsets[0].mapper().map(*(*child),v,dim);

      values[v].resize(n_levelsets);
      for(unsigned int l=0; l<n_levelsets; ++l)
        values[v][l] = levelsets[l].evallocal(0, *(*child), coords[v]);
    }

    for(unsigned int r=0; r<restricting_levelsets.size(); ++r) {

      // Iterate edges
      for (unsigned int e=0; e<edges; e++) {

        // local vertex indices
        int c = rElement.subEntity (e, dim-1, 0, dim);
        int n = rElement.subEntity (e, dim-1, 1, dim);

        Point a = coords[c];
        Point b = coords[n];

        Point x;

        // Get true number of levelset function
        const int l = restricting_levelsets[r];
        const double l_sign = restricting_sign[r];

        // Take values from first levelset which defines the bounding
        // box
        double fa = l_sign * values[c][l];
        double fb = l_sign * values[n][l];

        ///////////////////////////////////////////////////////
        // Compute interface intersection point x on edge
        for (int d=0; d<dim; d++)
          x[d] = a[d] - (fa * (b[d] - a[d])) / (fb - fa);
        ///////////////////////////////////////////////////////

#ifndef NO_MC_NORMALIZING
        // Distances of intersection point to edge point
        ctype dist_xa = (x-a).two_norm();
        ctype dist_xb = (x-b).two_norm();

        if (dist_xa <= refSz && dist_xa > 0 && dist_xb <= 1.0)
        {
          int i = ids[c];
#ifndef NDEBUG
          std::cout << "Normalizing vertex " << i << ", levelset nearly 0 (" << fa << ")" << std::endl;
#endif
          (*(levelsets[l]))[i] = 0.0;
          fa = 0.0;
          values[c][l] = 0.0;
          x = a;
        }
        else if (dist_xb <= refSz && dist_xb > 0 && dist_xa <= 1.0)
        {
          int i = ids[n];
#ifndef NDEBUG
          std::cout << "Normalizing vertex " << i << ", levelset nearly 0 (" << fb << ")" << std::endl;
#endif
          (*(levelsets[l]))[i] = 0.0;
          fb = 0.0;
          values[n][l] = 0.0;
          x = b;
        }
#endif //NO_MC_NORMALIZING

        // Set sticky bit for sticky boundary
        if(fb==0.0 || fa==0.0) {
          sticky_bits[id]=true;
        }

        // Domain information
        bool oa = fa <= 0.0;
        bool ob = fb <= 0.0;

        // Update bounding box corners with this edge.
        if ( oa != ob )
        {
          x = (*child)->geometry().global(x);
          min[r] = std::min(x,min[r]);
          max[r] = std::max(x,max[r]);
          default_bbox = false;
        }
        if (!oa && !ob)
        {
          a = (*child)->geometry().global(a);
          min[r] = std::min(a,min[r]);
          max[r] = std::max(a,max[r]);
          b = (*child)->geometry().global(b);
          min[r] = std::min(b,min[r]);
          max[r] = std::max(b,max[r]);
          default_bbox = false;
        }

      } // edges

    } // l

    for (unsigned int v=0; v<vertices; v++)
    {
      for(unsigned int l=0; l<n_levelsets; ++l) {
        const bool outside = values[v][l] < 0.0;
        const bool inside = values[v][l] > 0.0;
        domains[l][0] = outside || domains[l][0];
        domains[l][1] = inside || domains[l][1];
      }
    }
  } // children


  for(unsigned int r=0; r<max.size(); ++r)
    bbox_intersection(all_max,all_min,max[r],min[r]);


#ifdef NOBBOX
  bbox = BoundingBox(e);
#else

  // Check if bbox was set at all
  if (default_bbox) {
    all_min = 0.0;
    all_min = e->geometry().global(all_min);
    all_max = 1.0;
    all_max = e->geometry().global(all_max);
    bbox = BoundingBox(e);
  }
  else{
    bbox = BoundingBox(e, all_min, all_max, Dune::CuboidMode::corners);
  }

  // Compute volume
  double V = 1;
  for (int i=0; i<dim; i++)
    V *= all_max[i] - all_min[i];

  // Check if bounding box is too small
  if (V <= MC_DROP_THRESHOLD)
  {
    std::cout << "warning " << mapper.map(*e) << " size " << V << "\n";
    std::cout << "min " << all_min << "\n";
    std::cout << "max " << all_max << std::endl;
  }

#endif // NOBBOX
}

template<class GridView, typename M>
void
MarchingCubeSubTriangulation<GridView, M>::updateCaches()
{
  bboxes.clear();
  bboxes.resize(bbox_patterns.size() * mapper.size(),
                BoundingBox(EntityPointer(gridview.template begin<0>())) );
  bbox_indices.resize(bbox_patterns.size());
  for(unsigned int p=0; p<bbox_patterns.size(); ++p)
    bbox_indices[p].resize(mapper.size());

  domain_flags.resize(mapper.size());
  for(unsigned int i=0; i<domain_flags.size(); ++i) {
    domain_flags[i].clear();
    domain_flags[i].resize(n_levelsets, false);
  }

  sticky_bits.resize(mapper.size(),false);

  DomainBitSetVector domains(n_levelsets, false);

  // loop over all elementes e
  typedef typename GridView::template Codim<0>::template Partition<Dune::All_Partition>
  ::Iterator AllIterator;
  AllIterator e = gridview.template begin<0,Dune::All_Partition>();
  AllIterator eend = gridview.template end<0,Dune::All_Partition>();
  for (; e != eend; ++e) {
    std::list<EntityPointer> childlist;
    children(EntityPointer(e), childlist);
    int eid = mapper.map(*e);

    for(unsigned int p=0; p<bbox_patterns.size(); ++p) {
      const int current_bbox_index = p + bbox_patterns.size() * eid;
      calcBBox(EntityPointer(e),bbox_patterns[p],childlist,bboxes[current_bbox_index],
               domains);
      bbox_indices[p][eid] = current_bbox_index;
    }

    domain_flags[eid] = domains;
  }
}

template<class GridView, typename M>
bool
MarchingCubeSubTriangulation<GridView, M>::pointEq(const Point & a,
                                                   const Point & b) const
{
  ctype sum = 0;
  for (int d=0; d<dim; d++)
    sum += sqr(a[d] - b[d]);
  return (sqrt(sum) <= eqEpsilon);
}

/** \todo update to use jacobian typedef */
template<class GridView, typename M>
typename MarchingCubeSubTriangulation<GridView, M>::Normal
MarchingCubeSubTriangulation<GridView, M>::normal(const EntityPointer & e,
                                                  const Point & p,
                                                  const int l) const
{
  // ,,guess'' normal vector
  Normal n;
  Dune::FieldVector<double, dim> tmp(0);
  for (int d=0; d<dim; d++)
  {
    Dune::FieldVector<int,dim> D(0);
    D[d] = 1;
    tmp[d] = -levelsets[l].derivativelocal(0, D, *e, p);
  }
  /** \todo update to use jacobian typedef */
  Dune::FieldMatrix<double , dim, dim> invj_e;
  invj_e = e->geometry().jacobianInverseTransposed(p);
  n = 0;
  invj_e.umv(tmp,n);
  return n;
}

template<class GridView, typename M>
void
MarchingCubeSubTriangulation<GridView, M>::children(const EntityPointer & e,
                                                    std::list<EntityPointer> & childlist) const
{
  if (e->isLeaf())
  {
    childlist.push_back(e);
    return;
  }
  int level = gridview.grid().maxLevel();
  HierarchicIterator hit = e->hbegin(level);
  HierarchicIterator hend = e->hend(level);
  for (; hit!=hend; ++hit)
  {
    if (hit->isLeaf())
      childlist.push_back(EntityPointer(hit));
  }
}

/* ---===--- Factory Methods ---===--- */

template<class GridView, typename M>
typename MarchingCubeSubTriangulation<GridView, M>::EntityPartPointer
MarchingCubeSubTriangulation<GridView, M>::createEntityPart
(
  const EntityPointer & entity,
  const BoundingBox & dof_bbox,
  const std::vector< Dune::FieldVector<double,dim> > & c,
  const int domain_tag, const int domain_index,
  const double volume
) const
{
  typename EntityPart::GeometryImp geometry(Dune::geometryTypeFromVertexCount(dim, c.size()), c);
  EntityPartGeometry partGeometry(geometry,entity,dof_bbox);

  // Setup compatible domaint tag
  typename IntersectionPart::Traits::DomainTag
  domainTag(levelsets.getSize(),0);
  for(int l=0; l<levelsets.getSize(); ++l) {
    domainTag[l] += int(bool((1 << l*2) | domain_tag));
    domainTag[l] -= int(bool((1 << (l*2+1)) | domain_tag));
  }

  // Part index for this triangulation is always zero as the domain
  // index together with the entity uniquely identifies the bounding box.
  const int partIndex = 0;

  EntityPart * ePart =
    new EntityPart(partIndex,partGeometry,domain_tag,domainTag,domain_index,volume);

  return EntityPartPointer(ePart);
}

template<class GridView, typename M>
typename MarchingCubeSubTriangulation<GridView, M>::IntersectionPartPointer
MarchingCubeSubTriangulation<GridView, M>::createIntersection
(
  const BoundingBox & bbox, const BoundingBox & n_bbox,
  const std::vector< Dune::FieldVector<double,dim> > & c,
  const int bid,
  const Normal & n,
  const int intersection_tag,
  const int inside_domain_index,
  const int outside_domain_index
) const
{
  // Setup intersection geometry
  typename IntersectionPart::FaceGeometryImp faceGeometry(
      Dune::geometryTypeFromVertexCount(dim - 1, c.size()), c);
  IntersectionPartGeometry intersectionGeometry(faceGeometry,bbox,n_bbox,n);

  // Setup compatible domaint tag
  typename IntersectionPart::Traits::DomainTag
  domainTag(levelsets.getSize(),0);
  for(int l=0; l<levelsets.getSize(); ++l) {
    domainTag[l] += int(bool((1 << l*2) | intersection_tag));
    domainTag[l] -= int(bool((1 << (l*2+1)) | intersection_tag));
  }

  return IntersectionPartPointer
           (new IntersectionPart
             (intersectionGeometry,intersection_tag,domainTag,bid,
             inside_domain_index, outside_domain_index,0,0,0));
}

template<class GridView, typename M>
typename MarchingCubeSubTriangulation<GridView, M>::IntersectionPartPointer
MarchingCubeSubTriangulation<GridView, M>::createIntersection
(
  const BoundingBox & bbox,
  const std::vector< Dune::FieldVector<double,dim> > & c,
  const int bid,
  const Normal & n,
  const int intersection_tag,
  const int inside_domain_index,
  const int outside_domain_index
) const
{
  return createIntersection
           (bbox,bbox,c,bid,n,intersection_tag,
           inside_domain_index,outside_domain_index);
}

/* ---===--- INTERFACE ---===--- */

template<class GridView, typename M>
void
MarchingCubeSubTriangulation<GridView, M>::
create_rectangles(const EntityPointer & e,
                  std::list<BoundingBox> & rectangles, bool force_ref) const
{
  int id = mapper.map(*e);

  // Check if entity is intersected by some level set
  const bool not_intersected = !is_intersected(id);
  const bool is_inside = in_any_domain(id);
  const bool is_sticky = sticky_bits[id];

  if (is_inside)
  {

    if( (not_intersected && (! force_ref) && !is_sticky) ) {
      rectangles.push_back(BoundingBox(e));
      return ;
    }

    std::list<EntityPointer> childlist;
    children(e,childlist);
    typename std::list<EntityPointer>::const_iterator i=childlist.begin();
    typename std::list<EntityPointer>::const_iterator end=childlist.end();
    for (; i!=end; ++i)
    {
      rectangles.push_back(BoundingBox(*i));
    }

  } // is not outside
}

template<class GridView, typename M>
void MarchingCubeSubTriangulation<GridView, M>::
get_subentity_boundaries
(
  const EntityPointer & ls_entity,
  const Dune::GeometryType & type,
  const Vertices & v_coords,
  std::list< std::vector<Vertices> > & all_boundaries,
  std::list< DomainBitSetVector > & all_boundaries_domains,
  DomainBitSetVector current_domain,
  const int interface_index
) const
{
  std::vector<Vertices> boundaries;
  boundary_marching_cubes(ls_entity,type,v_coords,boundaries,levelsets[interface_index]);
  current_domain[interface_index]=true;

  // If there is more than one level set
  const unsigned int next_interface_index = interface_index + 1;
  if(n_levelsets > next_interface_index) {

    // Compute intersections with remaining level set functions
    for(unsigned int i=0; i<boundaries.size(); ++i) {

      Dune::GeometryType boundary_type = get_subentity_type<GridPoint,dim-1>(boundaries[i]);

      current_domain[next_interface_index][0]=true;
      get_boundary_subtriangulation(ls_entity,boundary_type,boundaries[i],
                                    all_boundaries,all_boundaries_domains,
                                    current_domain,next_interface_index);

      current_domain[next_interface_index].flip();

      get_boundary_subtriangulation(ls_entity,boundary_type,boundaries[i],
                                    all_boundaries,all_boundaries_domains,
                                    current_domain,next_interface_index);

      current_domain[next_interface_index]=false;
    }

  }
  else{
    all_boundaries.push_back(boundaries);
    all_boundaries_domains.push_back(current_domain);
  }
}

template<class GridView, typename M>
void MarchingCubeSubTriangulation<GridView, M>::
get_boundary_subtriangulation
(
  const EntityPointer & ls_entity,
  const Dune::GeometryType & type,
  const Vertices & v_coords,
  std::list< std::vector<Vertices> > & all_boundaries,
  std::list< DomainBitSetVector > & all_boundaries_domains,
  DomainBitSetVector & current_domain,
  const int interface_index
) const
{
  const BitSetRef & current_flags = current_domain[interface_index];
  const ctype sign = get_sign(current_flags);

  std::vector<Vertices> boundaries;

  const unsigned int next_interface_index = interface_index + 1;
  if(next_interface_index < n_levelsets) {

    marching_cubes<dim-1>(ls_entity,type,v_coords,boundaries,levelsets[interface_index],sign);

    for(unsigned int i=0; i<boundaries.size(); ++i) {

      Dune::GeometryType boundary_type = get_subentity_type<GridPoint,dim-1>(boundaries[i]);

      current_domain[next_interface_index][0]=true;
      get_boundary_subtriangulation(ls_entity,boundary_type,boundaries[i],
                                    all_boundaries,all_boundaries_domains,
                                    current_domain,next_interface_index);

      current_domain[next_interface_index].flip();

      get_boundary_subtriangulation(ls_entity,boundary_type,boundaries[i],
                                    all_boundaries,all_boundaries_domains,
                                    current_domain,next_interface_index);

      current_domain[next_interface_index]=false;
    }

  }
  else{
    all_boundaries.push_back(boundaries);
    all_boundaries_domains.push_back(current_domain);
    marching_cubes<dim-1>(ls_entity,type,v_coords,all_boundaries.back(),levelsets[interface_index],sign);
  }
}

template<class GridView, typename M>
template<bool get_boundaries>
void MarchingCubeSubTriangulation<GridView, M>::
get_subtriangulation
(
  const EntityPointer & ls_entity,
  const Dune::GeometryType & type,
  const Vertices & v_coords,
  std::list< std::vector<Vertices> > & all_volumes,
  std::list< DomainBitSetVector > & all_volumes_domains,
  std::list< std::vector<Vertices> > & all_boundaries,
  std::list< DomainBitSetVector > & all_boundaries_domains
) const
{
  DomainBitSetVector current_domain(n_levelsets,false);

  if(get_boundaries) {
    const int interface_index = 0;
    if(!is_virtual_interface[interface_index])
      get_subentity_boundaries(ls_entity,type,v_coords,all_boundaries,
                               all_boundaries_domains,current_domain,interface_index);

    const unsigned int next_interface_index = interface_index + 1;
    if(next_interface_index == n_levelsets)
      return;
  } // get_boundaries

  current_domain[0][0]=true;

  if(in_any_domain(current_domain,0))
    get_subtriangulation<get_boundaries>(ls_entity,type,v_coords,
                                         all_volumes,all_volumes_domains,
                                         all_boundaries,all_boundaries_domains,
                                         0,current_domain);

  current_domain[0].flip();

  if(in_any_domain(current_domain,0))
    get_subtriangulation<get_boundaries>(ls_entity,type,v_coords,
                                         all_volumes,all_volumes_domains,
                                         all_boundaries,all_boundaries_domains,
                                         0,current_domain);
}


template<class GridView, typename M>
template<bool get_boundaries>
void MarchingCubeSubTriangulation<GridView, M>::
get_subtriangulation
(
  const EntityPointer & ls_entity,
  const Dune::GeometryType & type,
  const Vertices & v_coords,
  std::list< std::vector<Vertices> > & all_volumes,
  std::list< DomainBitSetVector > & all_domains,
  std::list< std::vector<Vertices> > & all_boundaries,
  std::list< DomainBitSetVector > & all_boundaries_domains,
  const int interface_index,
  DomainBitSetVector & current_domain
) const
{
  const BitSetRef & current_flags = current_domain[interface_index];
  const ctype sign = get_sign(current_flags);

  const unsigned int next_interface_index = interface_index + 1;

  const bool is_recursion_leaf = next_interface_index == n_levelsets;

  std::vector<Vertices> volumes;

  // Recursion if true
  if(!is_recursion_leaf) {

    // Decompose sub entity given by coordinates in c and store its
    // children in volumes
    marching_cubes<dim>(ls_entity,type,v_coords,volumes,levelsets[interface_index],sign);

    // Iterate all children of the sub entity
    for(unsigned int i=0; i<volumes.size(); ++i) {

      // Geometry type of child
      Dune::GeometryType volumes_type = get_subentity_type<GridPoint,dim>(volumes[i]);

      // Set flags to domain of negative level set function
      current_domain[next_interface_index][0]=true;

      if(in_any_domain(current_domain,next_interface_index))
        get_subtriangulation<get_boundaries>(ls_entity,volumes_type,volumes[i],
                                             all_volumes,all_domains,all_boundaries,all_boundaries_domains,
                                             next_interface_index,current_domain);

      // Set flags to domain of positive level set function
      current_domain[next_interface_index].flip();

      if(in_any_domain(current_domain,next_interface_index))
        get_subtriangulation<get_boundaries>(ls_entity,volumes_type,volumes[i],
                                             all_volumes,all_domains,all_boundaries,all_boundaries_domains,
                                             next_interface_index,current_domain);

      // Reset flags
      current_domain[next_interface_index]=false;
    } // i
  }
  else{
    all_volumes.push_back(volumes);
    all_domains.push_back(current_domain);
    marching_cubes<dim>(ls_entity,type,v_coords,all_volumes.back(),levelsets[interface_index],sign);

    // Remove if empty
    if(all_volumes.back().empty()) {
      all_volumes.pop_back();
      all_domains.pop_back();
    }
  }

  // Compute interface in volumes of levelset with index
  // next_interface_index.
  if(get_boundaries) {

    if(next_interface_index >= n_levelsets)
      return;

    if(!is_virtual_interface[next_interface_index]) {

      // Iterate all children of the sub entity
      for(unsigned int i=0; i<volumes.size(); ++i) {

        // Geometry type of child
        Dune::GeometryType volumes_type = get_subentity_type<GridPoint,dim>(volumes[i]);

        get_subentity_boundaries(ls_entity,volumes_type,volumes[i],all_boundaries,
                                 all_boundaries_domains,current_domain,next_interface_index);
      } // i
    } // not virtual
  } // get_boundaries
}



template<class GridView, typename M>
void MarchingCubeSubTriangulation<GridView, M>::
face_marching_cubes
(
  const EntityPointer & ls_entity,
  const LeafIntersection & it,
  const Dune::GeometryType & type,
  const FaceVertices & c,
  std::vector< FaceVertices > & c_x,
  const LevelSet & current_levelset,
  const ctype sign
) const
{
  // Create Marching Cubes object
  Dune::MarchingCubes33<double, dim-1, ThresholdFunctor> mc;

  // Get levelset values in vertices given by c. These coordinates
  // are understood to be local to the levelset entity in ls_entity
  std::vector< double > v;
  for (unsigned int i=0; i<c.size(); i++) {
    const GridPoint c_g = it.geometryInInside().global(c[i]);
    v.push_back(sign * evaluateLevelset(current_levelset,ls_entity,c_g));
  }

  normalize_sub_entity<dim-1,dim-1>(type,v,c);

  // Create marching cubes subtriangulation
  int key = mc.getKey(v, v.size(), USEMC33);
  mc.getElements(v, v.size(), key, false, c_x); // false => cells
}


template<class GridView, typename M>
template<int mcdim>
void MarchingCubeSubTriangulation<GridView, M>::
marching_cubes
(
  const EntityPointer & ls_entity,
  const Dune::GeometryType & type,
  const std::vector< Dune::FieldVector<double,dim> > & c,
  std::vector< std::vector< Dune::FieldVector<double,dim> > > & c_x,
  const LevelSet & current_levelset,
  const ctype sign
) const
{
  // Create Marching Cubes object
  Dune::MarchingCubes33<double, mcdim, ThresholdFunctor> mc;

  // Get levelset values in vertices given by c. These coordinates
  // are understood to be local to the levelset entity in ls_entity
  std::vector<ctype> v;
  for (unsigned int i=0; i<c.size(); i++) {
    v.push_back(sign * evaluateLevelset(current_levelset,ls_entity,c[i]));
  }

  normalize_sub_entity<dim,mcdim>(type,v,c);

  bool outside = true;
  for (unsigned int i=0; i<v.size(); i++)
    outside = outside && !(v[i]);
  if(outside && sign > 0.0) {
    c_x.push_back(c);
    return;
  }

  // Create marching cubes subtriangulation
  int key = mc.getKey(v, v.size(), USEMC33);
  mc.getElements(v, v.size(), key, false, c_x); // false => cells
}

template<class GridView, typename M>
void MarchingCubeSubTriangulation<GridView, M>::
boundary_marching_cubes
(
  const EntityPointer & ls_entity,
  const Dune::GeometryType & type,
  const std::vector< Dune::FieldVector<double,dim> > & c,
  std::vector< std::vector< Dune::FieldVector<double,dim> > > & c_x,
  const LevelSet & current_levelset
) const
{
  // Create Marching Cubes object
  Dune::MarchingCubes33<double, dim, ThresholdFunctor> mc;

  // Get levelset values in vertices given by c. These coordinates
  // are understood to be local to the levelset entity in ls_entity
  std::vector< ctype > v;
  for (unsigned int i=0; i<c.size(); i++)
    v.push_back(evaluateLevelset(current_levelset,ls_entity,c[i]));

  normalize_sub_entity<dim,dim>(type,v,c);

  // Create marching cubes subtriangulation
  int key = mc.getKey(v, v.size(), USEMC33);
  mc.getElements(v, v.size(), key, true, c_x); // false => faces
}

template<class GridView, typename M>
void MarchingCubeSubTriangulation<GridView, M>::
create_entity_parts(const EntityPointer& e,
                    EntityPartList& geometries) const
{
  // Get list of levelset grid cell entities in dof grid cell
  // represented as bounding boxes
  std::list<BoundingBox> rectangles;
  create_rectangles(e, rectangles, force_refinement);
  for (typename std::list<BoundingBox>::iterator r=rectangles.begin();
       r!=rectangles.end(); ++r)
    create_entity_parts(e,r->entity(),geometries);
}

template<class GridView, typename M>
void MarchingCubeSubTriangulation<GridView, M>::
create_entity_parts(const EntityPointer& e,
                    const EntityPointer& e_fine,
                    EntityPartList& geometries) const
{
  // Geometry of levelset grid cell
  const typename Grid::template Codim<0>::Geometry &
  geometry = e_fine->geometry();
  // Reference element of levelset grid cell
  const Dune::ReferenceElement<double,dim> &
  refelem =
    Dune::ReferenceElements<double,dim>::general(geometry.type());

  // Get vertex positions local in levelset grid cell
  std::vector< Dune::FieldVector<double,dim> > c;
  for (int i=0; i<refelem.size(dim); i++)
    c.push_back(refelem.position (i, dim));

  // Compute vertex positions of sub entities in levelset grid cell
  typedef std::list<std::vector< Vertices > > VertexList;
  VertexList all_c_x;
  std::list< DomainBitSetVector> all_c_d;

  get_subtriangulation<false>
    (e_fine,geometry.type(),c,all_c_x,all_c_d,all_c_x,all_c_d);

  assert(all_c_x.size() == all_c_d.size());

  //Check whether any decompositioning was done. If not, then
  //modify vertices to assure construction of a linear sub entity.
  // if(all_c_x.size() == 1){
  //   std::vector<Vertices> & vertices = all_c_x.front();
  //   if(vertices.size() == 1){

  //     bool unchanged(true);
  //     for(int d=0; d<dim && unchanged; ++d){
  //       for (size_t i=0; i<vertices[0].size(); i++)
  //         if(!Dune::FloatCmp::eq<ctype>(vertices[0][i][d],c[i][d],1e-10)){
  //           unchanged = false;
  //           break;
  //         }
  //     }
  //     if(unchanged){
  //       Point A = c[0]; Point B = c[c.size()-1];
  //       vertices[0].clear();
  //       vertices[0].push_back(A);
  //       vertices[0].push_back(B);
  //     }
  //   }
  // }

  // Create sub entities from coordinates
  while(!all_c_x.empty()) {
    std::vector<Vertices> & vertices = all_c_x.front();
    DomainBitSetVector & domain = all_c_d.front();
    const int domain_index = get_domain_index(domain);
    const int domain_tag = get_domain_tag(domain);

    // Bounding box of shape function support
    const BoundingBox & bbox = BBox(domain_index,e);

    // Create actual sub entities
    for (size_t i=0; i<vertices.size(); i++)
    {
      for (size_t j=0; j<vertices[i].size(); j++)
        vertices[i][j] = geometry.global(vertices[i][j]);

      EntityPartPointer entity_part =
        createEntityPart(e_fine, bbox, vertices[i], domain_tag, domain_index,0);
      geometries.push_back(entity_part);
    }

    all_c_x.pop_front();
    all_c_d.pop_front();
  } // while
}

template<class GridView, typename M>
void MarchingCubeSubTriangulation<GridView, M>::
create_coarse_edges(const EntityPointer & e,
                    IntersectionPartList & edges,
                    std::map<int,bool> & processed) const
{
  const int id = mapper.map(*e);

  if(!is_intersected(id) && !sticky_bits[id]) {
    bool unprocessed_faces = false;

    const int domain_index = get_domain_index(domain_flags[id]);
    const int domain_tag = get_domain_tag(domain_flags[id]);

    BoundingBox bbox = BBox(domain_index,e);

    IntersectionIterator iit = e->ilevelbegin();
    IntersectionIterator iend = e->ilevelend();
    for (; iit != iend; ++iit) {
      if (!iit->neighbor())
        continue;

      const int n_id = mapper.map(*(iit->outside()));

      if(!is_intersected(n_id)) {

        BoundingBox n_bbox = BBox(domain_index,iit->outside());

        typedef Dune::ReferenceElement<double,dim-1> ReferenceFace;
        typedef Dune::ReferenceElements<double,dim-1> ReferenceFaces;
        const ReferenceFace & refelem =
          ReferenceFaces::general(iit->geometry().type());

        const typename Grid::template Codim<1>::Geometry &
        geometry = iit->geometry();

        std::vector< Dune::FieldVector<double,dim> > c;
        for (int i=0; i<refelem.size(dim-1); i++) {
          const GridFacePoint lc = refelem.position(i, dim-1);
          c.push_back( geometry.global(lc) );
        }

        static const GridFacePoint zeroCoord(0.0);
        Normal n = iit->outerNormal(zeroCoord);

        const int bid = Dune::UDG::internal;
        edges.push_back
          (createIntersection(bbox, n_bbox, c, bid, n, domain_tag, domain_index, domain_index));

        processed[n_id] = true;
      }
      else
        unprocessed_faces = true;
    }
    if(!unprocessed_faces)
      return ;
  }
}

template<class GridView, typename M>
void MarchingCubeSubTriangulation<GridView, M>::
get_edge_subtriangulation
(
  const int id,
  const EntityPointer & ls_entity,
  const LeafIntersection & it,
  const Dune::GeometryType & type,
  const FaceVertices & c,
  std::list< std::vector<FaceVertices> > & all_c_x,
  std::list< DomainBitSetVector > & all_c_d
) const
{
  DomainBitSetVector current_domain(n_levelsets,false);
  current_domain[0][0]=true;

  if(in_any_domain(current_domain,0))
    get_edge_subtriangulation(id,ls_entity,it,type,c,all_c_x,all_c_d,0,current_domain);

  current_domain[0].flip();

  if(in_any_domain(current_domain,0))
    get_edge_subtriangulation(id,ls_entity,it,type,c,all_c_x,all_c_d,0,current_domain);
}


template<class GridView, typename M>
void MarchingCubeSubTriangulation<GridView, M>::
get_edge_subtriangulation
(
  const int id,
  const EntityPointer & ls_entity,
  const LeafIntersection & it,
  const Dune::GeometryType & type,
  const FaceVertices & c,
  std::list< std::vector<FaceVertices> > & all_c_x,
  std::list< DomainBitSetVector > & all_c_d,
  const int interface_index,
  DomainBitSetVector & current_domain
) const
{
  const BitSetRef & current_flags = current_domain[interface_index];
  const ctype sign = get_sign(current_flags);
  std::vector<FaceVertices> c_x;

  const unsigned int next_interface_index = interface_index + 1;

  // Recursion if true
  if(next_interface_index < n_levelsets) {

    // Decompose sub entity given by coordinates in c and store its
    // children in c_x
    face_marching_cubes(ls_entity,it,type,c,c_x,levelsets[interface_index],sign);

    // Iterate all children of the sub entity
    for(unsigned int i=0; i<c_x.size(); ++i) {

      // Geometry type of child
      const Dune::GeometryType c_x_type = get_subentity_type<GridFacePoint,dim-1>(c_x[i]);

      // Set flags to domain of negative level set function
      current_domain[next_interface_index][0]=true;

      if(in_any_domain(current_domain,next_interface_index))
        get_edge_subtriangulation(id,ls_entity,it,c_x_type,c_x[i],
                                  all_c_x,all_c_d,next_interface_index,current_domain);

      // Set flags to domain of positive level set function
      current_domain[next_interface_index].flip();

      if(in_any_domain(current_domain,next_interface_index))
        get_edge_subtriangulation(id,ls_entity,it,c_x_type,c_x[i],
                                  all_c_x,all_c_d,next_interface_index,current_domain);

      // Reset flags
      current_domain[next_interface_index]=false;
    } // i
  }
  else{

    // Check if this is not really a face but rather a sticky boundary
    // with regard to the current level set function
    if(sticky_bits[id] && is_sticky_boundary(it,c,levelsets[interface_index]))
      return ;

    // Leaf of recursion tree
    all_c_x.push_back(c_x);
    all_c_d.push_back(current_domain);
    face_marching_cubes(ls_entity,it,type,c,all_c_x.back(),levelsets[interface_index],sign);
  }
}

template<class GridView, typename M>
void
MarchingCubeSubTriangulation<GridView, M>::create_edges(const EntityPointer & e,
                                                        IntersectionPartList & edges) const
{
  const int id = mapper.map(*e);

  // Check whether this cell is within a domain of interest at all
  if(!in_any_domain(id))
    return;

  // Create coarse edges if possible to reduce quadrature costs
  std::map<int,bool> processed;
  if (! force_refinement)
    create_coarse_edges(e,edges,processed);

  // The remaining egdes have to be created via Marching Cubes on the
  // Levelset level
  std::list<BoundingBox> rectangles;
  create_rectangles(e, rectangles, true);

  for (typename std::list<BoundingBox>::iterator r=rectangles.begin();
       r!=rectangles.end(); ++r) {

    LeafIntersectionIterator iit = r->entity()->ileafbegin();
    LeafIntersectionIterator iend = r->entity()->ileafend();
    for (; iit != iend; ++iit)
    {
      if (iit->neighbor())
      {
        EntityPointer ances = ancestor(iit->outside());

        int n_id;
        assert(mapper.contains(*ances, n_id));

        n_id = mapper.map(*ances);

        const bool already_processed = !( processed.find(n_id) == processed.end() );
        const bool is_real_face = e != ances;
        const bool any_domain = in_any_domain(n_id);

        // If this is true, the face may still be a sticky boundary,
        // but we won't find out until marching cubes is called, which
        // in such a case will return nothing
        if( is_real_face && !already_processed && any_domain)
        {
          // Intersection -> Global
          const typename Grid::template Codim<1>::Geometry &
          geometry = iit->geometry();
          // // Intersection -> Local in self
          // const typename Grid::template Codim<1>::Geometry &
          //   localgeometry = iit->geometryInInside();

          // Reference element of intersection
          const Dune::ReferenceElement<double,dim-1> &
          refelem =
            Dune::ReferenceElements<double,dim-1>::general(geometry.type());

          static const GridFacePoint zeroCoord(0.0);
          const Normal normal = iit->outerNormal(zeroCoord);

          // Local vertex coords
          const int n_refelem_vertices = refelem.size(dim-1);
          std::vector< Dune::FieldVector<double,dim-1> > c(n_refelem_vertices);
          for (int i=0; i<refelem.size(dim-1); i++)
            c[i] =refelem.position(i, dim-1);

          // Create sub entities
          std::list<std::vector<FaceVertices> > all_c_x;
          std::list< DomainBitSetVector > all_c_d;
          get_edge_subtriangulation(id,*(iit->inside()), *iit,
                                    geometry.type(),c,all_c_x,all_c_d);

          while(!all_c_x.empty()) {

            std::vector<FaceVertices> & vertices = all_c_x.front();

            if(!vertices.empty()) {

              DomainBitSetVector & domain = all_c_d.front();
              const int domain_index = get_domain_index(domain);
              const int domain_tag = get_domain_tag(domain);

              BoundingBox bbox = BBox(domain_index,e);
              BoundingBox n_bbox = BBox(domain_index,ances);

              // Create actual sub entities
              for (size_t i=0; i<vertices.size(); i++)
              {
                Vertices global_vertices(vertices[i].size());
                for (size_t j=0; j<vertices[i].size(); j++)
                  global_vertices[j] = geometry.global(vertices[i][j]);

                const int bid= Dune::UDG::internal;
                edges.push_back(createIntersection(bbox,n_bbox,global_vertices,bid,normal,domain_tag,
                                                   domain_index, domain_index));
              }
            }

            all_c_x.pop_front();
            all_c_d.pop_front();
          } // while


        } // really interior
      } // no boundary
    } // iit
  } // r
}


template<class GridView, typename M>
void
MarchingCubeSubTriangulation<GridView, M>::create_boundaries
  (const EntityPointer& e,
  IntersectionPartList& boundaries) const
{

  std::list<BoundingBox> rectangles;
  create_rectangles(e, rectangles, true);

  const int id = mapper.map(*e);

  // Check whether this cell is within a domain of interest at all
  if(!in_any_domain(id))
    return;

  for (typename std::list<BoundingBox>::iterator r=rectangles.begin();
       r!=rectangles.end(); ++r) {

    ///////////////////////////////////////////////////////
    // Iterate intersections to process the grid boundaries
    LeafIntersectionIterator iit = r->entity()->ileafbegin();
    LeafIntersectionIterator iend = r->entity()->ileafend();
    for (; iit != iend; ++iit)
    {
      const bool is_boundary = !iit->neighbor();
      if (is_boundary)
      {
        std::vector< Dune::FieldVector<double,dim-1> > c;
        const typename Grid::template Codim<1>::Geometry &
        geometry = iit->geometry();

        // get normal vector
        static const GridFacePoint zeroCoord(0.0);
        Normal normal = iit->outerNormal (zeroCoord);
        int bid = Dune::UDG::BoundaryFlags::processor;
        if (iit->boundary())
          bid = iit->indexInInside(); // iit->boundaryId();

        const Dune::ReferenceElement<double,dim-1> &
        refelem =
          Dune::ReferenceElements<double,dim-1>::general(geometry.type());

        for (int i=0; i<refelem.size(dim-1); i++)
          c.push_back(refelem.position(i, dim-1));

        // Create sub entities
        std::list<std::vector<FaceVertices> > all_c_x;
        std::list< DomainBitSetVector > all_c_d;
        get_edge_subtriangulation(id,*(iit->inside()), *iit,
                                  geometry.type(),c,all_c_x,all_c_d);

        while(!all_c_x.empty()) {
          std::vector<FaceVertices> & vertices = all_c_x.front();

          if(!vertices.empty()) {

            DomainBitSetVector & domain = all_c_d.front();
            const int domain_index = get_domain_index(domain);
            const int domain_tag = get_domain_tag(domain);

            BoundingBox bbox = BBox(domain_index,e);

            // Create actual sub entities
            for (size_t i=0; i<vertices.size(); i++)
            {
              Vertices global_vertices(vertices[i].size());
              for (size_t j=0; j<vertices[i].size(); j++)
                global_vertices[j] = geometry.global(vertices[i][j]);

              boundaries.push_back(createIntersection(bbox,global_vertices,bid,normal,domain_tag,
                                                      domain_index, -1));
            }
          }

          all_c_x.pop_front();
          all_c_d.pop_front();
        } // while
      } // is_boundary
    } // iit

    if(!is_intersected(id) && !sticky_bits[id])
      continue;

    ///////////////////////////////////////////////////////
    // Process interface boundaries
    std::vector< Dune::FieldVector<double,dim> > c;
    const typename Grid::template Codim<0>::Geometry &
    geometry = r->entity()->geometry();

    const Dune::ReferenceElement<double,dim> &
    refelem =
      Dune::ReferenceElements<double,dim>::general(geometry.type());

    for (int i=0; i<refelem.size(dim); i++)
      c.push_back(refelem.position (i, dim));

    // Compute vertex positions of sub entities in levelset grid cell
    typedef std::list<std::vector< Vertices > > VertexList;
    VertexList all_volumes;
    std::list< DomainBitSetVector> all_volumes_domains;
    VertexList all_boundaries;
    std::list< DomainBitSetVector> all_boundaries_domains;
    get_subtriangulation<true>(r->entity(),geometry.type(),c,
                               all_volumes,all_volumes_domains,
                               all_boundaries,all_boundaries_domains);

    // The intermediate containers, where the intersecions are stored
    // for post-processing
    std::list< Vertices > intersections;
    std::list< DomainBitSetVector > intersection_domains;
    std::list< Normal > intersection_centers;

    while(!all_boundaries.empty()) {

      std::vector< Vertices > & boundary_coords = all_boundaries.front();
      const int interface_index = get_interface_index(all_boundaries_domains.front());

      // Check if levelset is border to some domain and store their
      // index. This loop guarantees that domain_index <
      // neighbor_domain_index. TODO: This behavior is not sufficient
      // for coinciding level set interfaces.
      int domain_index = -1;
      for(int p=n_domains-1; p>=0; --p) {
        BitSet flags = bbox_patterns[p][interface_index];
        flags.flip();
        if(flags.any()) {
          domain_index = p;
        }
      }

      // This check allows to assemble boundaries of interfaces which
      // are not a neighbor to any domain
      if(domain_index == -1)
        domain_index = get_domain_index(all_boundaries_domains.front());

      for (size_t i=0; i<boundary_coords.size(); i++)
      {

        // Compute center of intersection
        Point center(0);
        for (size_t j=0; j<boundary_coords[i].size(); j++)
        {
          center += boundary_coords[i][j];
          boundary_coords[i][j] = geometry.global(boundary_coords[i][j]);
        }
        center *= 1.0 / (ctype)boundary_coords[i].size();
        // Store raw intersection data
        intersections.push_back(boundary_coords[i]);
        intersection_centers.push_back(center);
        intersection_domains.push_back(all_boundaries_domains.front());
      } // i

      all_boundaries_domains.pop_front();
      all_boundaries.pop_front();
    } // not empty


    while(! intersections.empty() ) {
      const Vertices & intersection = intersections.front();
      DomainBitSetVector & domain = intersection_domains.front();
      const Point & center = intersection_centers.front();

      typename std::list<Vertices>::iterator intersection_it = intersections.begin();
      typename std::list<Vertices>::iterator intersection_eit = intersections.end();
      typename std::list<DomainBitSetVector>::iterator domain_it = intersection_domains.begin();
      typename std::list<Point>::iterator center_it = intersection_centers.begin();

      // Skip self
      if(intersection_it != intersection_eit) {
        ++intersection_it; ++center_it; ++domain_it;
      }

      // Compare with all other intersections in list
      for(; intersection_it != intersection_eit; ++intersection_it, ++domain_it, ++center_it)
      {
        if(is_same_intersection(center,*center_it,domain,*domain_it))
        {
          // Merge domain information
          for(unsigned int i=0; i<domain.size(); ++i) {
            BitSet bb = (*domain_it)[i]; bb.flip();
            if( bb.none() )
              domain[i]=true;
          }

          // Remove doublettes from lists
          intersection_it = --intersections.erase(intersection_it);
          domain_it = --intersection_domains.erase(domain_it);
          center_it = --intersection_centers.erase(center_it);

        }
      }

      Normal face_normal;
      int inside_domain_index, outside_domain_index;
      determine_intersection(r->entity(),domain,center,face_normal,inside_domain_index,outside_domain_index);

      if(outside_domain_index == -1) {
        int domain_tag = get_domain_tag(domain);

        const int bid = Dune::UDG::BoundaryFlags::internal;
        const BoundingBox inside_box = BBox(inside_domain_index,e);
        boundaries.push_back(createIntersection(inside_box, intersection, bid,
                                                face_normal,domain_tag,
                                                inside_domain_index, outside_domain_index));
      }
      else{
        int domain_tag = get_domain_tag(domain);

        // Expensive checking for sticky boundary is only
        // performed if sticky bit is set
        int sticky_id = -1;
        EntityPointer sticky_entity = e;
        if(sticky_bits[id])
          is_sticky_boundary(r->entity(),intersection,sticky_id,sticky_entity);

        const BoundingBox inside_box = BBox(inside_domain_index,e);
        const EntityPointer f = sticky_id >= 0 ? sticky_entity : e;
        const BoundingBox outside_box = BBox(outside_domain_index,f);
        const int bid = Dune::UDG::internal;
        boundaries.push_back(createIntersection(inside_box, outside_box,intersection, bid,
                                                face_normal, domain_tag,
                                                inside_domain_index, outside_domain_index));
      }

      intersections.pop_front();
      intersection_centers.pop_front();
      intersection_domains.pop_front();
    }



  } // r
}

/*
   Local variables:
   c-file-style:"stroustrup"
   c-indent-level: 2
   c-basic-offset: 2
   indent-tabs-mode: nil
   End:
 */
