// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_BBOX_MERGER_HH
#define DUNE_BBOX_MERGER_HH

#include <vector>
#include <list>
#include <map>

#include <dune/common/fvector.hh>
#include <dune/common/float_cmp.hh>

namespace Dune {
  namespace UDG  {

    template <class GV, class MCST>
    class BBoxMerge
    {
    private:
      //! Grid Types
      /** @{ */
      typedef GV GridView;
      typedef typename GridView::Grid Grid;
      typedef typename Grid::template Codim<0>::Entity Entity;
      enum { dim=Grid::dimension };
      enum { dimw=Grid::dimensionworld };
      typedef typename Grid::ctype ctype;
      typedef Dune::FieldVector<ctype, dim> Point;
      typedef typename GridView::template Codim<0>::Iterator ElementIterator;
      typedef typename GridView::IntersectionIterator IntersectionIterator;
      typedef typename IntersectionIterator::Intersection Intersection;
      /** @} */


      //! Triangulation Types
      /** @{ */
      typedef MCST Triangulation;
      typedef typename Triangulation::BoundingBox BBox;
      typedef typename Triangulation::Cuboid Cuboid;
      typedef std::vector<BBox> BBoxVector;
      typedef typename Triangulation::Mapper Mapper;
      typedef typename Triangulation::IntersectionPart IntersectionPart;
      typedef typename IntersectionPart::IntersectionPartList IntersectionPartList;
      typedef typename IntersectionPartList::iterator IntersectionPartIterator;
      /** @} */

      typedef std::list<int> IList;
      typedef typename std::list<int>::iterator IListIterator;

      typedef std::pair<int,ctype> ictype;
      typedef std::vector<ictype> ICVector;

      static inline ctype get_volume(const Point & h)
      {
        ctype tmp(1);
        for(int d=0; d<dim; ++d)
          tmp *= h[d];
        return tmp;
      }

      static inline bool is_ge( const ctype a, const ctype b)
      {
        return Dune::FloatCmp::ge<ctype,Dune::FloatCmp::absolute>(a,b);
      }

      // Compare the values of both pairs. If they are approximately
      // equal, then compare indices.
      static bool comp_map_pair(ictype a, ictype b)
      {
        if(Dune::FloatCmp::eq<ctype>(a.second,b.second))
          return a.first > b.first;
        return a.second > b.second;
      }

      static void merge_bboxes(BBox & a, const BBox & b)
      {
        Point al = a.Cuboid::origin();
        Point ar = al + a.Cuboid::height();
        Point bl = b.Cuboid::origin();
        Point br = bl + b.Cuboid::height();

        for(int d=0; d<dim; ++d) {
          al[d] = std::min(al[d],bl[d]);
          ar[d] = std::max(ar[d],br[d]);
          ar[d] -= al[d];
        }

        a = BBox(a.entity(),al,ar);
      }

      enum { unbound = -1 };

      static int trace(int id, std::vector<int> responses)
      {
        const size_t limiter = responses.size();
        size_t counter = 0;

        while(responses[id] != unbound && counter++ < limiter)
          id = responses[id];

        assert(counter < limiter);
        return id;
      }


    public:

      //! This function merges the bounding boxes as created by the MC Sub
      //! Triangulation class. The provided mapper is expected to map an
      //! EntityPointer to the correct index in the corresponding bboxes
      //! array.
      static void mergeBBoxes
      (
        Triangulation & triangulation,
        BBoxVector & bboxes,
        std::vector< std::vector<int> > & bbox_indices,
        const GridView & gridview,
        const Mapper & mapper,
        const ctype merge_ratio
      )
      {
        assert(gridview.size(0) == mapper.size());

        BBoxVector tmp_bboxes;
        std::vector< std::vector<int> > tmp_bbox_indices(bbox_indices);

        for(unsigned int domain_index = 0; domain_index < bbox_indices.size(); ++domain_index) {

          // Store for each cell, the indices of the cells which want to
          // merge with it.
          std::vector< std::list<int> > requests(mapper.size());
          // Store for each cell, the indices of the cells which it wants to
          // merge with.
          std::vector< int > responses(mapper.size(),unbound);

          // Check if cell is outside of domain
          std::vector<bool> is_outside(mapper.size(),false);

          typedef std::vector<int>::iterator BBoxIndexIterator;
          BBoxIndexIterator it = bbox_indices[domain_index].begin();
          const BBoxIndexIterator eit = bbox_indices[domain_index].end();
          for(; it != eit; ++it) {

            const BBox & e_bbox = bboxes[*it];
            const Entity & e = e_bbox.entity();
            const int eid = mapper.map(e);
            ctype e_volume = get_volume(e_bbox.height());
            ctype main_volume = e_bbox.entity().geometry().volume();

            // Check whether this bbox is inside of this domain
            const bool is_in_domain = triangulation.isInDomain(e,domain_index);
            is_outside[eid] = ! is_in_domain;
            if(!is_in_domain)
              continue;

            // Check if this bbox is big enough : yes? -> proceed
            if( is_ge(e_volume/main_volume, merge_ratio) )
              continue;

            ///////////////////////////////////////////////////////>>
            // Get intersection areas to adjacent bboxes and sort them
            // accordingly
            IntersectionPartList edges;
            triangulation.create_intersection_parts(e, edges);

            typedef std::map<int,ctype> CMap;
            CMap areas_map;

            IntersectionPartIterator fit = edges.begin();
            IntersectionPartIterator efit = edges.end();
            for (; fit != efit; ++fit) {

              const IntersectionPart & iPart =
                static_cast<IntersectionPart&>(**fit);

              const unsigned int bbox_domain_index =
                triangulation.getDomainIndex(iPart.domainTag());

              if( bbox_domain_index != domain_index )
                continue;

              if (! iPart.neighbor())
                continue;

              if( iPart.boundingBoxOutside().entity() == iPart.boundingBoxInside().entity() )
                continue;

              const Entity f = iPart.boundingBoxOutside().entity();
              const int fid = mapper.map(f);

              const ctype edge_norm =
                triangulation.getEdgeNorm(iPart);

              areas_map[fid] = edge_norm;

            } // fit

            ICVector areas;

            typename std::map<int,ctype>::const_iterator mit
              = areas_map.begin();
            for(; mit != areas_map.end() ; ++mit) {
              areas.push_back(ictype(mit->first,mit->second));
            }

            std::sort(areas.begin(),areas.end(),comp_map_pair);
            //////////////////////////////////////////////////////<<

            bool size_sufficient = false;
            std::list<int> local_requests;
            typedef typename ICVector::iterator ICVectorIterator;
            ICVectorIterator ait = areas.begin();
            const ICVectorIterator eait = areas.end();
            for(; ait != eait; ++ait) {
              const int fid = ait->first;

              // Do not bind neighbor boxes which are already bound
              if(!requests[fid].empty() && ! (responses[fid]==unbound))
                continue;

              const int f_bbox_index = bbox_indices[domain_index][fid];
              const BBox & f_bbox = bboxes[f_bbox_index];
              const ctype f_volume = get_volume(f_bbox.height());

              local_requests.push_back(fid);

              // Check if size is sufficient now
              e_volume += f_volume;
              if( is_ge(e_volume/main_volume, merge_ratio) ) {
                size_sufficient = true;
                break;
              }

            } // ait


            // If the size is not sufficient, then request the next "best"
            // cell for merging and discard any older requests.
            if(!size_sufficient) {
              ait = areas.begin();
              if(ait != areas.end()) {
                const int fid = ait->first;
                local_requests.clear();
                local_requests.push_back(fid);
              }
            }

            // Write local requests to global arrays and trace
            // interconnections
            for(; !local_requests.empty(); local_requests.pop_front()) {
              const int leader = trace(local_requests.front(),responses);
              if(leader != eid) {

                requests[leader].push_back(eid);
                responses[eid] = leader;

                // At this point some other cells may have requested this cell
                // for merging. Those "followers" have to be by passed to the
                // new "leader".
                for(; !requests[eid].empty(); requests[eid].pop_front()) {
                  const int follower = requests[eid].front();
                  if(leader != follower)
                    requests[leader].push_back(follower);
                }

              }
            }

          } // it

          // Perform the actual merging
          it = bbox_indices[domain_index].begin();
          for(; it != eit; ++it) {

            const int e_bbox_index = *it;
            const BBox & e_bbox = bboxes[e_bbox_index];
            const Entity & e = e_bbox.entity();
            const int eid = mapper.map(e);

            assert(bbox_indices[domain_index][eid] == *it);

            // If a cell does not intersect with the domain, then no
            // bounding box is created and the index is invalidated.
            if(is_outside[eid]) {
              tmp_bbox_indices[domain_index][eid] = -1;
              continue;
            }

            IListIterator lit = requests[eid].begin();
            const IListIterator elit = requests[eid].end();

            // If this box is requested for merge by some other box, then
            // create an instance of a BBox and a valid index.
            if(lit != elit) {
              tmp_bboxes.push_back(BBox(bboxes[e_bbox_index]));
              tmp_bbox_indices[domain_index][eid] =  tmp_bboxes.size()-1;
            }
            // Otherwise check if the cell is self sufficient
            else if(responses[eid] == unbound) {
              tmp_bboxes.push_back(BBox(bboxes[e_bbox_index]));
              tmp_bbox_indices[domain_index][eid] =  tmp_bboxes.size()-1;
            }

            // Merging
            for(; lit != elit; ++lit) {
              const int f_bbox_index = bbox_indices[domain_index][*lit];
              merge_bboxes(tmp_bboxes.back(),bboxes[f_bbox_index]);
              tmp_bbox_indices[domain_index][*lit] =  tmp_bboxes.size()-1;
            }

          } // it

        } // domain_index


        ///////////////////////////////////////////////////////>>
        // Copy tmp array to original array and modify indices accordingly
        bboxes.resize(tmp_bboxes.size(),tmp_bboxes.front());
        std::vector<int> new_index(bboxes.size(),-1);
        unsigned int b_index=0;
        for(int i=0; i<mapper.size(); ++i) {
          for(unsigned int domain_index = 0; domain_index < bbox_indices.size(); ++domain_index) {
            const int tmp_b_index = tmp_bbox_indices[domain_index][i];

            // Default is invalid!
            bbox_indices[domain_index][i] = -1;

            // If valid index
            if(tmp_b_index != -1) {

              // Check if bbox was already added
              if(new_index[tmp_b_index] < 0) {
                bboxes[b_index] = tmp_bboxes[tmp_b_index];
                bbox_indices[domain_index][i] = b_index;
                new_index[tmp_b_index] = b_index;
                b_index++;
              }
              else{
                bbox_indices[domain_index][i] = new_index[tmp_b_index];
              }
            }


          }
        } // i
        assert(b_index == tmp_bboxes.size());
        ///////////////////////////////////////////////////////<<

      }

    };

  } // end namespace UDG
} // end namespace DUNE

#endif
