// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef UDG_VERTEXCOLORING_HH
#define UDG_VERTEXCOLORING_HH

#include <dune/common/reservedvector.hh>
#include <dune/grid/common/grid.hh>
#include <dune/geometry/referenceelements.hh>
#include "marching.hh"

namespace Dune {

#ifndef DUNE_COMMON_POWER_HH
  template <int A, int B>
  struct StaticPower
  {
    enum { value = StaticPower<A, B-1>::value * A };
  };

  template <int A>
  struct StaticPower<A, 0>
  {
    enum { value = 1 };
  };
#endif

  // A DataHandle class to exchange vertex coloring seeds
  template <class GV>
  class SeedExchange
    : public Dune::CommDataHandleIF<SeedExchange<GV>, bool>
  {
    typedef typename GV::IndexSet IndexSet;
    typedef typename IndexSet::IndexType IndexType;
  public:
    //! export type of data for message buffer
    typedef bool DataType;

    //! returns true if data for this codim should be communicated
    bool contains (int dim, int codim) const
    {
      // communicated vertex data
      return (codim==dim);
    }

    //! returns true if size per entity of given dim and codim is a constant
    bool fixedsize (int dim, int codim) const
    {
      return true;
    }

    /*! how many objects of type DataType have to be sent for a given entity

       Note: Only the sender side needs to know this size.
     */
    template<class EntityType>
    size_t size (EntityType& e) const
    {
      return 1;
    }

    //! pack data from user to message buffer
    template<class MessageBuffer, class EntityType>
    void gather (MessageBuffer& buff, const EntityType& e) const
    {
      bool b = connected[gv.indexSet().index(e)];
      buff.write(b);
    }

    /*! unpack data from message buffer to user

       n is the number of objects sent by the sender
     */
    template<class MessageBuffer, class EntityType>
    void scatter (MessageBuffer& buff, const EntityType& e, size_t n)
    {
      bool b;
      buff.read(b);
      if (b == true)
      {
        todo.push_back(gv.indexSet().index(e));
        //                connected[gv.indexSet().index(e)] = true;
      }
    }

    //! constructor
    SeedExchange (const GV& gv_, std::vector<bool>& c_, std::vector<IndexType> & t_)
      : gv(gv_), connected(c_), todo(t_)
    {}

  private:
    const GV & gv;
    std::vector<bool> & connected;
    std::vector<IndexType> & todo;
  };

  template<class GV>
  class VertexColoring
  {
  public:
    typedef GV GridView;
    typedef typename GridView::Grid Grid;
    typedef typename Grid::ctype ctype;
    typedef typename GridView::IndexSet IndexSet;
    typedef typename IndexSet::IndexType IndexType;
    enum { dim = GridView::Grid::dimension };
    typedef typename GridView::template Codim<0>::Iterator CellIterator;
    typedef typename GridView::template Codim<dim>::Iterator VertexIterator;
    enum { maxNeighbors = StaticPower<3,dim>::value };
    typedef ReservedVector<IndexType,maxNeighbors> NeighborList;

    class DefaultParameters
    {
    public:
      bool isSeed(const typename GridView::Intersection & intersection) const
      {
        return true;
      };
    };

    VertexColoring(const GridView & gridview) :
      gridview(gridview) {};

    template <class Vector, class Seeds>
    void colorize(Vector & v, const DefaultParameters & ) {}

    template <class Vector, class Seeds>
    void colorize(Vector & v, const Seeds & seeds)
    {
      // get indexset (we use all vertizes => indexset is suffient,
      // mapper not required)
      const IndexSet & indexset = gridview.indexSet();

      // store connectivity graph and search seeds
      if (gridview.comm().rank() == 0)
        std::cout << "vertexcoloring: creating connectivity graph" << std::endl;
      std::vector<NeighborList> couplings(indexset.size(dim));
      std::vector<IndexType> todo;
      std::vector<bool> connected(indexset.size(dim), false);
      std::vector<bool> inside(indexset.size(dim), false);

      for (CellIterator e = gridview.template begin<0>();
           e != gridview.template end<0>(); ++e)
      {
        GeometryType type = e->type();
        const ReferenceElement<ctype,dim> & rElement =
          ReferenceElements<ctype,dim>::general(type);

#if VERTEX_COLOR_DIAGONAL==1
        int vertizes = rElement.size(dim);
        // find connections
        for (int i=0; i<vertizes; i++)
        {
          IndexType id_i = indexset.subIndex(*e, i,dim);
          double v_i = v[id_i];
          inside[id_i] = ! marching.isInside(v_i);
          if (! inside[id_i])
            continue;
          for (int j=i+1; j<vertizes; j++)
          {
            IndexType id_j = indexset.subIndex(*e, j,dim);
            double v_j = v[id_j];
            if (marching.isInside(v_j))
              continue;
            // ACHTUNG, hier kann die gleiche Kannte mehrfach eingefügt werden!!!
            if (contains(couplings[id_i], id_j))
            {
              assert(contains(couplings[id_j], id_i));
              continue;
            }
            couplings[id_i].push_back(id_j);
            couplings[id_j].push_back(id_i);
          }
        }
#elif VERTEX_COLOR_DIAGONAL==2
        int faces = rElement.size(1);
        for (int f=0; f<faces; f++)
        {
          int vertizes = rElement.size(f,1, dim);
          for (int k=0; k<vertizes; k++)
          {
            int i = rElement.subEntity (f, 1, k, dim);
            IndexType id_i = indexset.subIndex(*e, i,dim);
            double v_i = v[id_i];
            inside[id_i] = ! marching.isInside(v_i);
            if (! inside[id_i])
              continue;
            for (int l=k+1; l<vertizes; l++)
            {
              int j = rElement.subEntity (f, 1, l, dim);
              IndexType id_j = indexset.subIndex(*e, j,dim);
              double v_j = v[id_j];
              if (marching.isInside(v_j))
                continue;
              // ACHTUNG, hier kann die gleiche Kannte mehrfach eingefügt werden!!!
              if (contains(couplings[id_i], id_j))
              {
                assert(contains(couplings[id_j], id_i));
                continue;
              }
              couplings[id_i].push_back(id_j);
              couplings[id_j].push_back(id_i);
            }
          }
        }
#else
        int edges = rElement.size(dim-1);
        for (int ed=0; ed<edges; ed++)
        {
          int i = rElement.subEntity (ed, dim-1, 0, dim);
          IndexType id_i = indexset.subIndex(*e, i,dim);
          double v_i = v[id_i];
          inside[id_i] = ! marching.isInside(v_i);
          if (! inside[id_i])
            continue;
          int j = rElement.subEntity (ed, dim-1, 1, dim);
          IndexType id_j = indexset.subIndex(*e, j,dim);
          double v_j = v[id_j];
          if (marching.isInside(v_j))
            continue;
          // ACHTUNG, hier kann die gleiche Kannte mehrfach eingefügt werden!!!
          if (contains(couplings[id_i], id_j))
          {
            assert(contains(couplings[id_j], id_i));
            continue;
          }
          couplings[id_i].push_back(id_j);
          couplings[id_j].push_back(id_i);
        }
#endif

        // find seeds
        typedef typename GridView::IntersectionIterator IntersectionIterator;
        for (IntersectionIterator it=gridview.ibegin(*e);
             it != gridview.iend(*e); ++it)
        {
          if (! it->boundary())
            continue;
          if (seeds.isSeed(*it))
          {
            int f = it->indexInInside();
            int facevertizes = rElement.size(f,1,dim);
            for (int x=0; x<facevertizes; x++)
            {
              int lid_x = rElement.subEntity(f,1,x,dim);
              IndexType id_x = indexset.subIndex(*e, lid_x,dim);
              todo.push_back(id_x);
            }
          }
        }
      }

      int rerun = 1;
      // for MPI_SIZE
      while (rerun > 0)
      {
        rerun = 0;
        // communicate overlap and add additional seeds
        SeedExchange<GV> seedex(gridview, connected, todo);
        gridview.communicate(seedex,Dune::All_All_Interface,Dune::ForwardCommunication);

        // recursive colorization
        if (gridview.comm().rank() == 0)
          std::cout << "vertexcoloring: starting recursion" << std::endl;
        while (! todo.empty())
        {
          // get entry id
          int i = todo.back();
          todo.pop_back();
          // did we already see this entry?
          if (connected[i] == true) continue;
          // we found a new entry... rerun is necessary
          rerun = 1;
          // register current entry
          connected[i] = true;
          // search all couplings and put them on the stack
          for (typename NeighborList::size_type n=0;
               n<couplings[i].size(); ++n)
          {
            int j=couplings[i][n];
            if (i != j && connected[j] == false && couplings[j].size() != 0)
            {
              todo.push_back(j);
            }
          }
        }

        // globally sync rerun
        rerun = gridview.comm().sum(rerun);
      }

      // eliminated vertizes, which are not part of the domain
      if (gridview.comm().rank() == 0)
        std::cout << "vertexcoloring: updating levelset" << std::endl;
      for (size_t i=0; i<inside.size(); ++i)
      {
        if (inside[i] && !connected[i])
        {
          v[i] *= -v[i];
        }
      }

      if (gridview.comm().rank() == 0)
        std::cout << "vertexcoloring: done" << std::endl;
    }
  private:
    bool contains(const NeighborList &l, IndexType i)
    {
      for (typename NeighborList::size_type n=0;
           n<l.size(); ++n)
        if (i == l[n])
          return true;
      return false;
    }

    const GridView & gridview;
    Dune::UDG::Marching<ctype, dim> marching;
  };

}

#endif // UDG_VERTEXCOLORING_HH
