// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_NOTRIANGULATION_HH
#define DUNE_NOTRIANGULATION_HH

#include <list>
#include <vector>
#include <memory>
#include <map>

#include "vtriangulation.hh"
#include <dune/udg/common/cuboid.hh>
#include <dune/udg/common/entitypart.hh>
#include <dune/udg/common/intersectionpart.hh>

#include <dune/udg/disc/functions.hh>
#include <dune/udg/disc/p1function.hh>
#include <dune/geometry/type.hh>

template<class GridViewT>
class NoSubTriangulation : public VirtualSubTriangulation<GridViewT> {
public:
  typedef GridViewT GridView;
  typedef typename GridView::Grid Grid;
  enum { dimension=Grid::dimension /*!<dimension of the grid we are using*/ };
  enum { dimensionworld=Grid::dimensionworld /*!<dimension of the world*/ };
  typedef typename Grid::ctype ctype;
  Dune::GeometryType CubeType;
  //! LevelSet is defined by a projection from a C0 function.
  typedef Dune::C0GridFunction<Grid, ctype, 1> AnalyticLevelSet;

  typedef typename Dune::UDG::EntityPartInterface<GridView> EntityPartInterface;
  typedef typename EntityPartInterface::EntityPartList EntityPartList;
  typedef typename Dune::UDG::DefaultEntityPartGeometry<GridView> EntityPartGeometry;
  typedef typename Dune::UDG::DefaultEntityPart<EntityPartGeometry> EntityPart;
  typedef typename EntityPartInterface::EntityPartPointer EntityPartPointer;

  typedef Dune::UDG::AffineIntersectionPartGeometry<GridView>
  IntersectionPartGeometry;
  typedef Dune::UDG::DefaultIntersectionPart<IntersectionPartGeometry>
  IntersectionPart;
  typedef Dune::UDG::IntersectionPartInterface<GridView>
  IntersectionPartInterface;
  typedef typename IntersectionPart::IntersectionPartPointer
  IntersectionPartPointer;
  typedef typename IntersectionPart::IntersectionPartList IntersectionPartList;
  typedef Dune::GridCuboid<Grid> BoundingBox;

protected:
  /* --- typedefs --- */

  enum { dim = Grid::dimension /*!< dimension of the grid we are using */ };
  enum { nv = 1<<dim /*!< number of vertices in the cube of dimension */ };
  enum { dimw = Grid::dimensionworld /*!< dimension of the world */ };
  //! access types exported by the grid
  /** @{ */
  typedef typename GridView::template Codim<0>::Iterator ElementIterator;
  typedef typename GridView::template Codim<dim>::Iterator VertexIterator;
  typedef typename GridView::IntersectionIterator IntersectionIterator;
  typedef typename Grid::template Codim<0>::Entity Entity;
  /** @} */
  //! a geometrical point
  typedef Dune::FieldVector<ctype, dimw> Point;
  typedef Dune::FieldVector<ctype, dimw> Normal;

  //! disable copy constructor
  NoSubTriangulation(const NoSubTriangulation &);
public:
  NoSubTriangulation(const GridView & _g, int i=0) : g(_g), rect_sub(false) {};
  const GridView & gridView() const { return g; }
  virtual ~NoSubTriangulation() {};
  bool & forceRefinement() { return rect_sub; };
  const BoundingBox BBox(const unsigned int domain_index, const Entity & e) const
  {
    assert(domain_index == 0);
    return BoundingBox(e);
  }
  // // quadratureParts
  // void quadrature_parts(const EntityPointer & e,
  //   std::list<EntityPart> & parts) const
  //   {
  //       parts.push_back(EntityPart(BoundingBox(e)));
  //   }
  // void quadrature_parts(const IntersectionIterator & it,
  //   std::list<IntersectionPart> & parts) const;

  //! Check whether codim 0 entity intersects with domain of index
  //! domain_index.
  inline bool isInDomain(const Entity & e, const int & domain_index) const
  {
    return (domain_index == 0);
  }
  ctype getEdgeNorm(const IntersectionPartInterface & intersection) const
  {
    static const Dune::FieldVector<ctype, dim-1> zero(0);
    return intersection.geometry().integrationElement(zero);
  }

  // create Entities
  void create_entity_parts(const Entity & e,
                           EntityPartList & geometries) const
  {
    const typename Entity::Geometry& g=e.geometry();

    std::vector<typename EntityPart::GeometryImp::GlobalCoordinate>
    corners(g.corners());
    for(int c=0; c<g.corners(); ++c) corners[c] = g.corner(c);

    typename EntityPartGeometry::GeometryImp geometry(g.type(),corners);
    EntityPartGeometry partGeometry(geometry,e,BoundingBox(e));
    typename IntersectionPart::Traits::DomainTag domainTag(1);
    EntityPartPointer ePart( new EntityPart(0,partGeometry,0,domainTag,0,g.volume()) );
    geometries.push_back(ePart);
  }

  void create_intersection_parts
    (const Entity & e,
    IntersectionPartList & intersections) const
  {
    BoundingBox E = e;
    for (IntersectionIterator i=g.ibegin(e); i!= g.iend(e); ++i)
    {
      Dune::FieldVector<ctype, dim-1> zero(0);
      const int bid = i->boundarySegmentIndex();
      typename IntersectionPart::Traits::DomainTag domainTag(1);

      switch (i->geometry().corners())
      {
      case 2 :
        assert(dim==2);
        {
          Dune::FieldVector<ctype, dim> A = i->geometry().corner(0);
          Dune::FieldVector<ctype ,dim> B = i->geometry().corner(1);
          typename IntersectionPart::FaceGeometryImp
          faceGeometry = IntersectionPart::createFaceGeometry(2,&A,&B);

          if (i->neighbor()) {
            BoundingBox N(i->outside());
            IntersectionPartGeometry intersectionGeometry
              (faceGeometry,E,N,i->outerNormal(zero));
            IntersectionPartPointer iPart
              (new IntersectionPart(intersectionGeometry,0,domainTag,bid,0,0,0,0,0) );
            intersections.push_back(iPart);
          }
          else
          {
            IntersectionPartGeometry intersectionGeometry
              (faceGeometry,E,E,i->outerNormal(zero));
            IntersectionPartPointer iPart
              (new IntersectionPart(intersectionGeometry,0,domainTag,bid,0,0,0,0,0) );
            intersections.push_back(iPart);
          }
        }
        break;
      case 3 :
        assert(dim==3);
        {
          Dune::FieldVector<ctype, dim> A = i->geometry().corner(0);
          Dune::FieldVector<ctype ,dim> B = i->geometry().corner(1);
          Dune::FieldVector<ctype, dim> C = i->geometry().corner(2);
          typename IntersectionPart::FaceGeometryImp
          faceGeometry = IntersectionPart::createFaceGeometry(3,&A,&B,&C);

          if (i->neighbor()) {
            BoundingBox N(i->outside());
            IntersectionPartGeometry intersectionGeometry
              (faceGeometry,E,N,i->outerNormal(zero));
            IntersectionPartPointer iPart
              (new IntersectionPart(intersectionGeometry,0,domainTag,bid,0,0,0,0,0) );
            intersections.push_back(iPart);
          }
          else
          {
            IntersectionPartGeometry intersectionGeometry
              (faceGeometry,E,E,i->outerNormal(zero));
            IntersectionPartPointer iPart
              (new IntersectionPart(intersectionGeometry,0,domainTag,bid,0,0,0,0,0) );
            intersections.push_back(iPart);
          }
        }
        break;
      case 4 :
        assert(dim==3);
        {
          Dune::FieldVector<ctype, dim> A = i->geometry().corner(0);
          Dune::FieldVector<ctype ,dim> B = i->geometry().corner(1);
          Dune::FieldVector<ctype, dim> C = i->geometry().corner(2);
          Dune::FieldVector<ctype ,dim> D = i->geometry().corner(3);
          typename IntersectionPart::FaceGeometryImp
          faceGeometry = IntersectionPart::createFaceGeometry(3,&A,&B,&C,&D);

          if (i->neighbor()) {
            BoundingBox N(i->outside());
            IntersectionPartGeometry intersectionGeometry
              (faceGeometry,E,N,i->outerNormal(zero));
            IntersectionPartPointer iPart
              (new IntersectionPart(intersectionGeometry,0,domainTag,bid,0,0,0,0,0) );
            intersections.push_back(iPart);
          }
          else
          {
            IntersectionPartGeometry intersectionGeometry
              (faceGeometry,E,E,i->outerNormal(zero));
            IntersectionPartPointer iPart
              (new IntersectionPart(intersectionGeometry,0,domainTag,bid,0,0,0,0,0) );
            intersections.push_back(iPart);
          }
        }
        break;
      default :
        DUNE_THROW(Dune::Exception, "Impossible Geometry. " <<
                   "In " << dim-1
                                                            << "D there is no known geometry with " <<
                   i->geometry().corners() << " corners");
      }
    }
  }
  // void create_edges(const EntityPointer & e,
  //                   std::list<IntersectionPart> & edges) const
  //   {
  //     BoundingBox E = e;
  //     for (IntersectionIterator i=g.ibegin(*e); i!= g.iend(*e); ++i)
  //     {
  //       if (i->neighbor()) {
  //         BoundingBox N = i->outside();
  //         Dune::FieldVector<ctype, dim-1> zero(0);
  //         switch (i->geometry().corners())
  //         {
  //         case 2:
  //           assert(dim==2);
  //           {
  //             Dune::FieldVector<ctype, dim> A = i->geometry().corner(0);
  //             Dune::FieldVector<ctype ,dim> B = i->geometry().corner(1);;
  //             edges.push_back(
  //               IntersectionPart(E,N,A,B,i->outerNormal(zero)));
  //           }
  //           break;
  //         case 3:
  //           assert(dim==3);
  //           {
  //             Dune::FieldVector<ctype, dim> A = i->geometry().corner(0);
  //             Dune::FieldVector<ctype ,dim> B = i->geometry().corner(1);
  //             Dune::FieldVector<ctype, dim> C = i->geometry().corner(2);
  //             edges.push_back(
  //               IntersectionPart(E,N,A,B,C,i->outerNormal(zero)));
  //           }
  //           break;
  //         case 4:
  //           assert(dim==3);
  //           {
  //             Dune::FieldVector<ctype, dim> A = i->geometry().corner(0);
  //             Dune::FieldVector<ctype ,dim> B = i->geometry().corner(1);
  //             Dune::FieldVector<ctype, dim> C = i->geometry().corner(2);
  //             Dune::FieldVector<ctype ,dim> D = i->geometry().corner(3);
  //             edges.push_back(
  //               IntersectionPart(E,N,A,B,C,D,i->outerNormal(zero)));
  //           }
  //           break;
  //         default:
  //           DUNE_THROW(Dune::Exception, "Impossible Geometry. " <<
  //                      "In " << dim-1
  //                      << "D there is no known geometry with " <<
  //                      i->geometry().corners() << " corners");
  //         }
  //       }
  //     }
  //   }
  // void create_boundaries(const EntityPointer & e,
  //                        std::list<IntersectionPart> & boundaries) const
  //   {
  //     BoundingBox E = e;
  //     for (IntersectionIterator i=g.ibegin(*e); i!= g.iend(*e); ++i)
  //       if (i->boundary()) {
  //         Dune::FieldVector<ctype, dim-1> zero(0);
  //         switch (i->geometry().corners())
  //         {
  //         case 2:
  //           assert(dim==2);
  //           {
  //             Dune::FieldVector<ctype, dim> A = i->geometry().corner(0);
  //             Dune::FieldVector<ctype ,dim> B = i->geometry().corner(1);
  //             boundaries.push_back(
  //               IntersectionPart(E,A,B,i->boundarySegmentIndex(),i->outerNormal(zero)));
  //           }
  //           break;
  //         case 3:
  //           assert(dim==3);
  //           {
  //             Dune::FieldVector<ctype, dim> A = i->geometry().corner(0);
  //             Dune::FieldVector<ctype ,dim> B = i->geometry().corner(1);
  //             Dune::FieldVector<ctype, dim> C = i->geometry().corner(2);
  //             boundaries.push_back(
  //               IntersectionPart(E,A,B,C,i->boundarySegmentIndex(),i->outerNormal(zero)));
  //           }
  //           break;
  //         case 4:
  //           assert(dim==3);
  //           {
  //             Dune::FieldVector<ctype, dim> A = i->geometry().corner(0);
  //             Dune::FieldVector<ctype ,dim> B = i->geometry().corner(1);
  //             Dune::FieldVector<ctype, dim> C = i->geometry().corner(2);
  //             Dune::FieldVector<ctype ,dim> D = i->geometry().corner(3);
  //             boundaries.push_back(
  //               IntersectionPart(E,A,B,C,D,i->boundarySegmentIndex(),i->outerNormal(zero)));
  //           }
  //           break;
  //         default:
  //           DUNE_THROW(Dune::Exception, "Impossible Geometry. " <<
  //                      "In " << dim-1
  //                      << "D there is no known geometry with " <<
  //                      i->geometry().corners() << " corners");
  //         }
  //       }
  //   }
private:
  const GridView & g;
  bool rect_sub;
};

#endif
