// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_UDG_UDG_ASSEMBLER_HH
#define DUNE_UDG_UDG_ASSEMBLER_HH

#include <dune/pdelab/gridoperator/common/assemblerutilities.hh>
#include <dune/typetree/typetree.hh>
#include <dune/udg/pdelab/assembler/ulocalfunctionspace.hh>
#include <dune/pdelab/common/elementmapper.hh>
#include <dune/udg/notriangulation.hh>


namespace Dune
{
  namespace UDG
  {

    /**
     * \brief The assembler for UDG subtriangulations.
     *
     * \note This version of the assembler assumes that all intersection
     *       parts initially occur two-sided; using fixed criteria, it
     *       selects one of both sides for a unique visit, except when
     *       the LOP chooses the member flag doSkeletonTwoSided = true.
     *
     * \note Analogous to the logic of the PDELab assembler, the flag
     *       doSkeletonTwoSided of the LOP decides whether the *_skeleton
     *       methods of the LOP are called two-sided. The decision affects
     *       all intersection parts forwarded to the *_skeleton methods,
     *       not just those which lie on a domain skeleton (also see next
     *       note).
     *
     * \note UDG applications with multiple domains may need to visit both
     *       sides of intersection parts on an interior domain boundary in
     *       order to assemble boundary integral terms of both adjacent
     *       domains as usual (i.e. only using inside entities with respect
     *       to each domain). Furthermore, in UDG applications which use a
     *       multi layer subtriangulation, an intersection part can have
     *       multiple roles, possibly requiring a different orientation on
     *       each layer.
     *       For those applications it seems unappropriate to allow the
     *       assembler to choose one fixed side for each intersection part.
     *       They can avoid this by using LOPs that choose the member flag
     *       doSkeletonTwoSided = true.
     *       The *_skeleton methods of the LOPs then get called two-sided
     *       for each intersection part (independent of its role) and have
     *       to take care of its orientation themselves.
     *
     * \tparam GFSU                Type of GridFunctionSpace for ansatz
     *                             functions.
     * \tparam GFSV                Type of GridFunctionSpace for test
     *                             functions.
     * \tparam nonoverlapping_mode Indicates whether assembling is done
     *                             for overlap cells.
     *
     * \todo Has the non-overlapping mode ever been tested?
     */
    template <typename GFSU, typename GFSV,
              bool nonoverlapping_mode = false,
              typename ST = NoSubTriangulation<typename GFSU::Traits::GridViewType> >
    class UDGAssembler
    {
    public:
      //! Types related to fundamental mesh grid view
      //! @{
      typedef typename GFSU::Traits::GridViewType GV;
      typedef typename GV::Traits::template Codim<0>::Iterator ElementIterator;
      typedef typename GV::Traits::template Codim<0>::Entity Element;
      typedef typename GV::IntersectionIterator IntersectionIterator;
      typedef typename IntersectionIterator::Intersection Intersection;
      //! @}

      //! Subtriangulation types
      //! @{
      typedef typename ST::IntersectionPartIterator IntersectionPartIterator;
      typedef typename ST::EntityPartIterator EntityPartIterator;
      typedef ST SubTriangulation;
      //! @}

      //! Grid function space types
      //! @{
      typedef GFSU TrialGridFunctionSpace;
      typedef GFSV TestGridFunctionSpace;
      //! @}

      //! Size type as used in grid function space
      typedef typename GFSU::Traits::SizeType SizeType;

      //! Static check on whether this is a Galerkin method
      static const bool isGalerkinMethod = std::is_same<GFSU,GFSV>::value;

      //! \todo Please doc me!
      UDGAssembler (const GFSU& gfsu_, const GFSV& gfsv_, const ST& st_)
        : gfsu(gfsu_), gfsv(gfsv_), lfsu(gfsu_), lfsv(gfsv_),
          lfsun(gfsu_), lfsvn(gfsv_), sub_triangulation(st_)
      {}

      //! Get the trial grid function space
      const GFSU& trialGridFunctionSpace () const
      {
        return gfsu;
      }

      //! Get the test grid function space
      const GFSV& testGridFunctionSpace () const
      {
        return gfsv;
      }

      //! \todo Please doc me!
      //! \todo Please hide me from public
      //! \todo At (*) we assume that each layer decomposes the same fixed
      //!       subset of the fundamental mesh domain into a set of domains
      //!       associated with the layer such that each layer has the same
      //!       exterior boundary. Is this intended, i.e. do all layers must
      //!       have the same exterior boundary?
      template <typename LocalAssemblerEngine, typename Mapper, typename IntersectionPart>
      bool visitCouplingFace (const LocalAssemblerEngine& assembler_engine,
                              Mapper& cell_mapper,
                              const IntersectionPart& part) const
      {
        // Extract integration requirements from the local assembler
        const bool require_uv_skeleton = assembler_engine.requireUVSkeleton();
        const bool require_v_skeleton = assembler_engine.requireVSkeleton();
        const bool require_skeleton_two_sided = assembler_engine.requireSkeletonTwoSided();

        // Skip intersection part if it lies on an exterior boundary
        // or if no coupling face integration is needed at all (*)
        if (!( part.neighbor() && (require_uv_skeleton || require_v_skeleton) ))
          return false;

        // Iterate layers of the subtriangulation
        const int maxLayer = part.resetLayer();
        for (int l = 0; l <= maxLayer; ++l)
        {
          // Set layer of intersection part to current layer
          part.setLayer(l);

          // Compute unique id for adjacent fundamental mesh host cells on current layer
          typedef typename GV::IndexSet::IndexType Index;
          const Index ids = cell_mapper.map(part.boundingBoxInside().entity());
          const Index idn = cell_mapper.map(part.boundingBoxOutside().entity());
          // const Index ids = cell_mapper.map(*(ig.inside()));
          // const Index idn = cell_mapper.map(*(ig.outside()));

          // Get domain indices on current layer
          const int insideDomainIndex = part.insideDomainIndex();
          const int outsideDomainIndex = part.outsideDomainIndex();

          // Check if inside fundamental mesh host cell on current layer is a ghost
          const bool isExterior
            = part.boundingBoxInside().entity().partitionType() != Dune::InteriorEntity;

          // Reset layer of intersection part
          part.resetLayer();

          // Visit domain skeletons only once / twice
          const bool domain_skeleton_side_one = (insideDomainIndex == outsideDomainIndex && ids > idn);
          const bool domain_skeleton_side_two = (insideDomainIndex == outsideDomainIndex && ids < idn);
          // Skip side two if intersection parts for the LOP's
          // *_skeleton methods shall be uniquely visited
          if (domain_skeleton_side_two && !require_skeleton_two_sided) return false;
          bool visit_face = (domain_skeleton_side_one ||
                             (domain_skeleton_side_two && require_skeleton_two_sided));

          // Visit interior domain boundaries only once / twice
          const bool domain_boundary_side_one = (insideDomainIndex < outsideDomainIndex);
          const bool domain_boundary_side_two = (insideDomainIndex > outsideDomainIndex);
          // Skip side two if intersection parts for the LOP's
          // *_skeleton methods shall be uniquely visited
          if (domain_boundary_side_two && !require_skeleton_two_sided) return false;
          visit_face |= (domain_boundary_side_one ||
                         (domain_boundary_side_two && require_skeleton_two_sided));

          // In non-overlapping mode also visit the face if inside
          // fundamental mesh host cell on current layer is a ghost
          visit_face |= (nonoverlapping_mode && isExterior);

          if (visit_face) return true;
        }

        return false;
      }

      //! \todo Please doc me!
      //! \todo Please hide me from public
      //! \todo We assume that each layer decomposes the same fixed
      //!       subset of the fundamental mesh domain into a set of domains
      //!       associated with the layer such that each layer has the same
      //!       exterior boundary. Is this intended, i.e. do all layers must
      //!       have the same exterior boundary?
      template <typename LocalAssemblerEngine, typename IntersectionPart>
      bool visitExteriorBoundaryFace (const LocalAssemblerEngine& assembler_engine,
                                      const IntersectionPart& part) const
      {
        // Extract integration requirements from the local assembler
        const bool require_uv_boundary = assembler_engine.requireUVBoundary();
        const bool require_v_boundary = assembler_engine.requireVBoundary();

        // Skip intersection part if no exterior boundary face integration is
        // needed at all
        if (!(require_uv_boundary || require_v_boundary))
          return false;

        // An intersection part on an exterior boundary is an intersection part
        // on a domain boundary which has no domain cell patch neighbor on the
        // outside
        return (!part.neighbor() && part.boundary());
      }

      //! \todo Please doc me!
      template <class LocalAssemblerEngine>
      void assemble (LocalAssemblerEngine& assembler_engine) const
      {
        typedef typename LocalAssemblerEngine::LocalAssembler::Traits::TrialGridFunctionSpaceConstraints CU;
        typedef typename LocalAssemblerEngine::LocalAssembler::Traits::TestGridFunctionSpaceConstraints CV;

        const CU& cu = assembler_engine.trialConstraints();
        const CV& cv = assembler_engine.testConstraints();

        typedef PDELab::LFSIndexCache<LFSU,CU> LFSUCache;
        typedef PDELab::LFSIndexCache<LFSV,CV> LFSVCache;

        const bool needs_constraints_caching = assembler_engine.needsConstraintsCaching(cu,cv);

        LFSUCache lfsu_cache(lfsu,cu,needs_constraints_caching);
        LFSVCache lfsv_cache(lfsv,cv,needs_constraints_caching);
        LFSUCache lfsun_cache(lfsun,cu,needs_constraints_caching);
        LFSVCache lfsvn_cache(lfsvn,cv,needs_constraints_caching);

        // Notify assembler engine about oncoming assembly
        assembler_engine.preAssembly();

        // Map each cell to unique id
        PDELab::ElementMapper<GV> cell_mapper(gfsu.gridView());

        // Extract integration requirements from the local assembler
        const bool require_uv_volume = assembler_engine.requireUVVolume();
        const bool require_v_volume = assembler_engine.requireVVolume();
        const bool require_uv_skeleton = assembler_engine.requireUVSkeleton();
        const bool require_v_skeleton = assembler_engine.requireVSkeleton();
        const bool require_uv_boundary = assembler_engine.requireUVBoundary();
        const bool require_v_boundary = assembler_engine.requireVBoundary();
        const bool require_uv_post_skeleton = assembler_engine.requireUVVolumePostSkeleton();
        const bool require_v_post_skeleton = assembler_engine.requireVVolumePostSkeleton();
        const bool create_subtriangulation = require_uv_volume || require_v_volume ||
                                             require_uv_skeleton || require_v_skeleton ||
                                             require_uv_boundary || require_v_boundary ||
                                             require_uv_post_skeleton || require_v_post_skeleton;

        // Traverse fundamental mesh grid view
        for (ElementIterator it = gfsu.gridView().template begin<0>();
             it != gfsu.gridView().template end<0>(); ++it)
        {

          if (assembler_engine.assembleCell(*it))
            continue;

          // Create entity parts and intersection parts on current element
          if (create_subtriangulation)
            sub_triangulation.create(*it);

          // Check whether volume integrals are required
          if (require_uv_volume || require_v_volume)
          {
            // Traverse entity parts
            EntityPartIterator epit = sub_triangulation.begin();
            EntityPartIterator eepit = sub_triangulation.end();
            for (; epit != eepit; ++epit)
            {
              // Bind local test function space to element
              lfsv.bind(epit->subEntity(),true);
              lfsv_cache.update();

              // Notify assembler engine about bind
              assembler_engine.onBindLFSV(*epit,lfsv_cache);

              // Volume integration
              assembler_engine.assembleVVolume(*epit,lfsv_cache);

              if (require_uv_volume)
              {
                // Bind local trial function space to element
                lfsu.bind(epit->subEntity(),true);
                lfsu_cache.update();

                // Notify assembler engine about bind
                assembler_engine.onBindLFSUV(*epit,lfsu_cache,lfsv_cache);

                // Load coefficients of local functions
                assembler_engine.loadCoefficientsLFSUInside(lfsu_cache);

                // Volume integration
                assembler_engine.assembleUVVolume(*epit,lfsu_cache,lfsv_cache);

                // Notify assembler engine about unbinds
                assembler_engine.onUnbindLFSUV(*epit,lfsu_cache,lfsv_cache);
              }

              // Notify assembler engine about unbinds
              assembler_engine.onUnbindLFSV(*epit,lfsv_cache);
            }
          }

          // Skip if no intersection part iterator is needed
          if (require_uv_skeleton || require_v_skeleton ||
              require_uv_boundary || require_v_boundary)
          {
            // Traverse intersection parts
            IntersectionPartIterator ipit = sub_triangulation.ibegin();
            IntersectionPartIterator eipit = sub_triangulation.iend();
            for (; ipit != eipit; ++ipit)
            {
              // (Unique) visit of intersection part
              if (visitCouplingFace(assembler_engine,cell_mapper,ipit->intersection()))
              {
                // Rebind local test space
                lfsv.bind(ipit->intersection(),true);
                lfsv_cache.update();

                // Notify assembler engine about binds
                assembler_engine.onBindLFSVInside(*ipit,lfsv_cache);

                // Bind local test space to neighbor element
                lfsvn.bind(ipit->intersection(),false);
                lfsvn_cache.update();

                // Notify assembler engine about binds
                assembler_engine.onBindLFSVOutside(*ipit,lfsv_cache,lfsvn_cache);

                // Skeleton integration
                assembler_engine.assembleVSkeleton(*ipit,lfsv_cache,lfsvn_cache);

                if (require_uv_skeleton)
                {
                  // Rebind local trial space
                  lfsu.bind(ipit->intersection(),true);
                  lfsu_cache.update();

                  // Notify assembler engine about binds
                  assembler_engine.onBindLFSUVInside(*ipit,lfsu_cache,lfsv_cache);

                  // Load coefficients of local functions
                  assembler_engine.loadCoefficientsLFSUInside(lfsu_cache);

                  // Bind local trial space to neighbor element
                  lfsun.bind(ipit->intersection(),false);
                  lfsun_cache.update();

                  // Notify assembler engine about binds
                  assembler_engine.onBindLFSUVOutside(*ipit,lfsu_cache,lfsv_cache,lfsun_cache,lfsvn_cache);

                  // Load coefficients of local functions
                  assembler_engine.loadCoefficientsLFSUOutside(lfsun_cache);

                  // Skeleton integration
                  assembler_engine.assembleUVSkeleton(*ipit,lfsu_cache,lfsv_cache,lfsun_cache,lfsvn_cache);

                  // Notify assembler engine about unbinds
                  assembler_engine.onUnbindLFSUVOutside(*ipit,lfsu_cache,lfsv_cache,lfsun_cache,lfsvn_cache);
                  assembler_engine.onUnbindLFSUVInside(*ipit,lfsu_cache,lfsv_cache);
                }

                // Notify assembler engine about unbinds
                assembler_engine.onUnbindLFSVOutside(*ipit,lfsv_cache,lfsvn_cache);
                assembler_engine.onUnbindLFSVInside(*ipit,lfsv_cache);
              }
              else
                // If intersection part is on an exterior domain boundary
                // and boundary integration is required
                if (visitExteriorBoundaryFace(assembler_engine,ipit->intersection()))
                {
                  // Rebind local function spaces and read coefficents
                  lfsv.bind(ipit->intersection(),true);
                  lfsv_cache.update();

                  // Notify assembler engine about binds
                  assembler_engine.onBindLFSVInside(*ipit,lfsv_cache);

                  // Boundary integration
                  assembler_engine.assembleVBoundary(*ipit,lfsv_cache);

                  if (require_uv_boundary)
                  {
                    // Rebind local function spaces and read coefficents
                    lfsu.bind(ipit->intersection(),true);
                    lfsu_cache.update();

                    // Notify assembler engine about binds
                    assembler_engine.onBindLFSUVInside(*ipit,lfsu_cache,lfsv_cache);

                    // Load coefficients of local functions
                    assembler_engine.loadCoefficientsLFSUInside(lfsu_cache);

                    // Boundary integration
                    assembler_engine.assembleUVBoundary(*ipit,lfsu_cache,lfsv_cache);

                    // Notify assembler engine about unbinds
                    assembler_engine.onUnbindLFSUVInside(*ipit,lfsu_cache,lfsv_cache);
                  }

                  // Notify assembler engine about unbinds
                  assembler_engine.onUnbindLFSVInside(*ipit,lfsv_cache);
                }
            }  // ipit
          }  // do skeleton and boundary

          // Check whether volume integrals shall be assembled after skeleton integrals
          if (require_uv_post_skeleton || require_v_post_skeleton)
          {
            // Traverse entity parts
            EntityPartIterator epit = sub_triangulation.begin();
            EntityPartIterator eepit = sub_triangulation.end();
            for (; epit != eepit; ++epit)
            {
              // Bind local test function space to element
              lfsv.bind(epit->subEntity(),true);
              lfsv_cache.update();

              // Notify assembler engine about binds
              assembler_engine.onBindLFSV(*epit,lfsv_cache);

              // Volume integration
              assembler_engine.assembleVVolumePostSkeleton(*epit,lfsv_cache);

              if (require_uv_post_skeleton)
              {
                // Bind local trial function space to element
                lfsu.bind(epit->subEntity(),true);
                lfsu_cache.update();

                // Notify assembler engine about binds
                assembler_engine.onBindLFSUV(*epit,lfsu_cache,lfsv_cache);

                // Load coefficients of local functions
                assembler_engine.loadCoefficientsLFSUInside(lfsu_cache);

                // Volume integration
                assembler_engine.assembleUVVolumePostSkeleton(*epit,lfsu_cache,lfsv_cache);

                // Notify assembler engine about unbinds
                assembler_engine.onUnbindLFSUV(*epit,lfsu_cache,lfsv_cache);
              }

              // Notify assembler engine about unbinds
              assembler_engine.onUnbindLFSV(*epit,lfsv_cache);
            }
          }
        }  // it

        // Notify assembler engine that assembly is finished
        assembler_engine.postAssembly(gfsu,gfsv);
      }

    private:
      //! Global grid function spaces on the fundamental mesh grid view
      const GFSU& gfsu;
      const GFSV& gfsv;

      //! Local function space types
      typedef Dune::PDELab::UnfittedLocalFunctionSpace<GFSU,Dune::PDELab::TrialSpaceTag> LFSU;
      typedef Dune::PDELab::UnfittedLocalFunctionSpace<GFSV,Dune::PDELab::TestSpaceTag> LFSV;

      //! Local function spaces in inside fundamental mesh host cell
      mutable LFSU lfsu;
      mutable LFSV lfsv;

      //! Local function spaces in outside/neighbor fundamental mesh host cell
      mutable LFSU lfsun;
      mutable LFSV lfsvn;

      //! UDG subtriangulation yielding entity parts and intersection parts
      const ST sub_triangulation;
    };

  }
}


#endif
