// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_UDG_PDELAB_LOCALFUNCTIONSPACE_HH
#define DUNE_UDG_PDELAB_LOCALFUNCTIONSPACE_HH

#include <vector>

#include <dune/common/stdstreams.hh>

#include <dune/geometry/referenceelements.hh>

#include <dune/localfunctions/common/interfaceswitch.hh>
#include <dune/localfunctions/common/localkey.hh>

#include <dune/typetree/typetree.hh>

#include <dune/pdelab/gridfunctionspace/tags.hh>
#include <dune/pdelab/gridfunctionspace/localvector.hh>
#include <dune/pdelab/gridfunctionspace/localfunctionspace.hh>

namespace Dune {
  namespace PDELab {

    //! \addtogroup GridFunctionSpace
    //! \ingroup PDELab
    //! \{

    template<typename GFS>
    struct gfs_to_ulfs {

      //! The DOFIndex type that will be used in the resulting UnfittedLocalFunctionSpace tree.
      typedef typename build_dof_index_type<GFS>::type DOFIndex;
    };

    namespace {

      /** Bounding box extractor to determine bounding box from both
          intersection part and entity part via a common interface
          @{
       */
      template< typename PartType, typename Switch = void >
      struct BoundingBoxExtractor
      {
        typedef typename PartType::Traits::BoundingBox BoundingBox;
        static const BoundingBox & get(const PartType & part, const bool)
        { return part.boundingBox(); }
      };

      template<typename PartType>
      struct BoundingBoxExtractor
      <PartType,
          typename std::enable_if<AlwaysTrue<typename PartType::Traits::FaceGeometry>::value>::type>
      {
        typedef typename PartType::Traits::BoundingBox BoundingBox;
        static const BoundingBox & get(const PartType & part, const bool inside)
        { return inside ? part.boundingBoxInside() : part.boundingBoxOutside(); }
      };
      /** @} */

      template<typename Part>
      struct UnfittedComputeSizeVisitor
        : public TypeTree::TreeVisitor
          , public TypeTree::DynamicTraversal
      {

        template<typename Node, typename TreePath>
        void pre(Node& node, TreePath treePath)
        {
          node.offset = offset;
        }

        template<typename Node, typename TreePath>
        void post(Node& node, TreePath treePath)
        {
          node.n = offset - node.offset;
        }

        template<typename Node, typename TreePath>
        void leaf(Node& node, TreePath treePath)
        {
          node.offset = offset;
          Node::FESwitch::setStore(node.pfe, node.pgfs->finiteElementMap().find(ep_,inside_));
          node.n = Node::FESwitch::basis(*node.pfe).size();
          offset += node.n;
        }

        UnfittedComputeSizeVisitor
          (const Part& ep, const bool inside, std::size_t _offset = 0)
          : ep_(ep), inside_(inside), offset(_offset)
        {}

        const Part& ep_;
        const bool inside_;
        std::size_t offset;

      };


      template<typename Part>
      struct UnfittedFillIndicesVisitor
        : public TypeTree::TreeVisitor
          , public TypeTree::DynamicTraversal
      {

        template<typename Node, typename TreePath>
        void leaf(Node& node, TreePath treePath)
        {
          const int layer = node.pgfs->finiteElementMap().triangulationLayer();
          ep_.setLayer(layer);
          typedef BoundingBoxExtractor<Part> BBoxExtractor;
          const typename BBoxExtractor::BoundingBox & bbox = BBoxExtractor::get(ep_,inside_);
          ep_.resetLayer();

          // setup DOFIndices for this finite element
          node.dofIndices(bbox.entity(),node._dof_indices->begin()+node.offset,
                          node._dof_indices->begin()+node.offset+node.n);
        }

        template<typename Node, typename Child, typename TreePath, typename ChildIndex>
        void afterChild(const Node& node, const Child& child, TreePath treePath, ChildIndex childIndex)
        {
          for (std::size_t i = 0; i<child.n; ++i)
          {
            (*node._dof_indices)[child.offset+i].treeIndex().push_back(childIndex);
          }
        }

        UnfittedFillIndicesVisitor(const Part & ep, const bool inside)
          : ep_(ep), inside_(inside)
        {}

        const Part& ep_;
        const bool inside_;
      };

    } // end empty namespace

    //=======================================
    // local function space base: base class
    //=======================================

    //! traits mapping global function space information to local function space
    template<typename GFS, typename DI>
    struct UnfittedLocalFunctionSpaceBaseTraits
    {
      //! \brief Type of the underlying grid function space
      typedef GFS GridFunctionSpaceType;

      //! \brief Type of the underlying grid function space
      typedef GFS GridFunctionSpace;

      //! \brief Type of the grid view that the underlying grid function space is defined on.
      typedef typename GFS::Traits::GridViewType GridViewType;

      //! \brief Type of the grid view that the underlying grid function space is defined on.
      typedef typename GFS::Traits::GridViewType GridView;

      //! \brief Type of codim 0 entity in the grid
      typedef typename GridViewType::Traits::template Codim<0>::Entity Element;

      //! \brief Type to store indices from Backend
      typedef typename GFS::Traits::SizeType SizeType;

      //! \brief Type of container to store indices
      typedef typename std::vector<SizeType> IndexContainer;

      //! \brief Type of DOFIndex associated with this UnfittedLocalFunctionSpace.
      typedef DI DOFIndex;

      //! \brief Type of container to store multiindices.
      typedef typename std::vector<DI> DOFIndexContainer;

    };

    template <typename GFS, typename DOFIndex>
    class UnfittedLocalFunctionSpaceBaseNode
    {
      typedef typename GFS::Traits::BackendType B;

      template<typename>
      friend struct PropagateGlobalStorageVisitor;

      template<typename>
      friend struct UnfittedComputeSizeVisitor;

      template<typename>
      friend struct UnfittedFillIndicesVisitor;

    public:
      typedef UnfittedLocalFunctionSpaceBaseTraits<GFS,DOFIndex> Traits;

      //! \brief construct from global function space
      UnfittedLocalFunctionSpaceBaseNode (shared_ptr<const GFS> gfs)
        : pgfs(gfs)
          , _dof_index_storage()
          , _dof_indices(&_dof_index_storage)
          , n(0)
      {}

      //! \brief get current size
      typename Traits::IndexContainer::size_type size () const
      {
        return n;
      }

      std::size_t subSpaceDepth() const
      {
        return 0;
      }

      //! \brief get maximum possible size (which is maxLocalSize from grid function space)
      typename Traits::IndexContainer::size_type maxSize () const
      {
        return _dof_indices->size();
      }

      //! \brief get size of an appropriate local vector object
      /**
         this is the number of dofs of the complete local function
         space tree, i.e. the size() of the root node. The local
         vector objects must always have this size and the localIndex
         method maps into the range [0,localVectorSize()[
       */
      typename Traits::IndexContainer::size_type localVectorSize () const
      {
        return _dof_indices->size();
      }

      //! \brief map index in this local function space to root local function space
      typename Traits::IndexContainer::size_type localIndex (typename Traits::IndexContainer::size_type index) const
      {
        return offset+index;
      }

      //! \brief Maps given index in this local function space to its corresponding global DOFIndex.
      /**
       * \param index  The local index value from the range 0,...,size()-1
       * \returns      A const reference to the associated, globally unique DOFIndex. Note that the returned
       *               object may (and must) be copied if it needs to be stored beyond the time of the next
       *               call to bind() on this UnfittedLocalFunctionSpace (e.g. when the DOFIndex is used as a DOF
       *               identifier in a constraints container).
       */
      const typename Traits::DOFIndex& dofIndex(typename Traits::IndexContainer::size_type index) const
      {
        return (*_dof_indices)[offset + index];
      }

      //! \brief print debug information about this local function space
      void debug () const
      {
        std::cout << n << " indices = (";
        for (typename Traits::IndexContainer::size_type k=0; k<n; k++)
          std::cout << (*_dof_indices)[localIndex(k)] << " ";
        std::cout << ")" << std::endl;
      }

      //! Returns the GridFunctionSpace underlying this UnfittedLocalFunctionSpace.
      const GFS& gridFunctionSpace() const
      {
        return *pgfs;
      }

    protected:
      //! \brief bind local function space to entity
      /**

         This is a generic implementation of the bind function. It is
         parametrized with the NodeType, which the type of the derived
         UnfittedLocalFunctionSpaceNode. Handing the NodeType as a parammeter
         avoid the need for the CRTP construct, but all derived
         classes have to add a method bind, which forward to this
         method.

         \param node reference to the derived node, the address must be the same as this
         \param e entity to bind to
       */
      template<typename NodeType, typename Part>
      void bind (NodeType& node, const Part& ep, const bool inside);

      template<typename NodeType>
      void setup(NodeType& node)
      {
        _dof_index_storage.resize(gridFunctionSpace().ordering().maxLocalSize());
        TypeTree::applyToTree(node,PropagateGlobalStorageVisitor<>());
      }

      shared_ptr<GFS const> pgfs;
      typename Traits::DOFIndexContainer _dof_index_storage;
      typename Traits::DOFIndexContainer* _dof_indices;
      typename Traits::IndexContainer::size_type n;
      typename Traits::IndexContainer::size_type offset;
    };

    template <typename GFS, typename DOFIndex>
    template <typename NodeType, typename Part>
    void UnfittedLocalFunctionSpaceBaseNode<GFS,DOFIndex>::bind
      (NodeType& node, const Part& ep, const bool inside)
    {
      assert(&node == this);

      // compute sizes
      UnfittedComputeSizeVisitor<Part> csv(ep,inside);
      TypeTree::applyToTree(node,csv);

      // initialize iterators and fill indices
      UnfittedFillIndicesVisitor<Part> fiv(ep,inside);
      TypeTree::applyToTree(node,fiv);
    }

    //=======================================
    // local function space base: power implementation
    //=======================================

    //! traits for multi component local function space
    template<typename GFS, typename DOFIndex, typename N>
    struct PowerCompositeUnfittedLocalFunctionSpaceTraits : public UnfittedLocalFunctionSpaceBaseTraits<GFS,DOFIndex>
    {
      //! type of local function space node
      typedef N NodeType;
    };

    //! Tag denoting a PowerUnfittedLocalFunctionSpace
    struct PowerUnfittedLocalFunctionSpaceTag {};

    // local function space for a power grid function space
    template<typename GFS, typename DOFIndex, typename ChildLFS, std::size_t k>
    class PowerUnfittedLocalFunctionSpaceNode :
      public UnfittedLocalFunctionSpaceBaseNode<GFS,DOFIndex>,
      public TypeTree::PowerNode<ChildLFS,k>
    {
      typedef UnfittedLocalFunctionSpaceBaseNode<GFS,DOFIndex> BaseT;
      typedef TypeTree::PowerNode<ChildLFS,k> TreeNode;

      template<typename>
      friend struct PropagateGlobalStorageVisitor;

      template<typename>
      friend struct ClearSizeVisitor;

      template<typename>
      friend struct UnfittedComputeSizeVisitor;

      template<typename>
      friend struct UnfittedFillIndicesVisitor;

    public:
      typedef PowerCompositeUnfittedLocalFunctionSpaceTraits<GFS,DOFIndex,PowerUnfittedLocalFunctionSpaceNode> Traits;

      typedef PowerUnfittedLocalFunctionSpaceTag ImplementationTag;

      //! \brief initialize with grid function space
      template<typename Transformation>
      PowerUnfittedLocalFunctionSpaceNode (shared_ptr<const GFS> gfs,
                                           const Transformation& t,
                                           const std::array<shared_ptr<ChildLFS>,k>& children)
        : BaseT(gfs)
          , TreeNode(children)
      {}

      template<typename Transformation>
      PowerUnfittedLocalFunctionSpaceNode (const GFS& gfs,
                                           const Transformation& t,
                                           const std::array<shared_ptr<ChildLFS>,k>& children)
        : BaseT(stackobject_to_shared_ptr(gfs))
          , TreeNode(children)
      {}

      //! \brief bind local function space to entity
      template < typename Part >
      void bind (const Part& e, const bool inside)
      {
        // call method on base class, this avoid the barton neckman trick
        BaseT::bind(*this,e,inside);
      }

    };


    // transformation template, we need a custom template in order to
    // inject the DOFIndex type into the UnfittedLocalFunctionSpace
    template<typename SourceNode, typename Transformation>
    struct power_gfs_to_ulfs_template
    {
      template<typename TC>
      struct result
      {
        typedef PowerUnfittedLocalFunctionSpaceNode
        <SourceNode,typename Transformation::DOFIndex,TC,SourceNode::CHILDREN> type;
      };
    };

    // register PowerGFS -> UnfittedLocalFunctionSpace transformation
    template<typename PowerGridFunctionSpace, typename Params>
    Dune::TypeTree::TemplatizedGenericPowerNodeTransformation<
        PowerGridFunctionSpace,
        gfs_to_ulfs<Params>,
        power_gfs_to_ulfs_template<PowerGridFunctionSpace,gfs_to_ulfs<Params> >::template result
        >
    registerNodeTransformation(PowerGridFunctionSpace* pgfs, gfs_to_ulfs<Params>* t, PowerGridFunctionSpaceTag* tag);


    //=======================================
    // local function space base: composite implementation
    //=======================================

    //! Tag denoting a CompositeUnfittedLocalFunctionSpace
    struct CompositeUnfittedLocalFunctionSpaceTag {};

    // local function space for a power grid function space
#if DUNE_VERSION_NEWER(DUNE_TYPETREE,2,4)
    template<typename GFS, typename DOFIndex, typename... Children>
    class CompositeUnfittedLocalFunctionSpaceNode
      : public UnfittedLocalFunctionSpaceBaseNode<GFS,DOFIndex>
      , public TypeTree::CompositeNode<Children...>
    {
      typedef UnfittedLocalFunctionSpaceBaseNode<GFS,DOFIndex> BaseT;
      typedef TypeTree::CompositeNode<Children...> NodeType;
#else
    template<typename GFS, typename DOFIndex, DUNE_TYPETREE_COMPOSITENODE_TEMPLATE_CHILDREN>
    class CompositeUnfittedLocalFunctionSpaceNode
      : public UnfittedLocalFunctionSpaceBaseNode<GFS,DOFIndex>
        , public DUNE_TYPETREE_COMPOSITENODE_BASETYPE
    {
      typedef UnfittedLocalFunctionSpaceBaseNode<GFS,DOFIndex> BaseT;
      typedef DUNE_TYPETREE_COMPOSITENODE_BASETYPE NodeType;
#endif

      template<typename>
      friend struct PropagateGlobalStorageVisitor;

      template<typename>
      friend struct ClearSizeVisitor;

      template<typename>
      friend struct UnfittedComputeSizeVisitor;

      template<typename>
      friend struct UnfittedFillIndicesVisitor;

    public:
      typedef PowerCompositeUnfittedLocalFunctionSpaceTraits<GFS,DOFIndex,CompositeUnfittedLocalFunctionSpaceNode> Traits;

      typedef CompositeUnfittedLocalFunctionSpaceTag ImplementationTag;

      template<typename Transformation>
      CompositeUnfittedLocalFunctionSpaceNode (shared_ptr<const GFS> gfs,
                                               const Transformation& t,
#if DUNE_VERSION_NEWER(DUNE_TYPETREE,2,4)
                                               std::shared_ptr<Children>... children)
        : BaseT(gfs)
          , NodeType(children...)
#else
                                               DUNE_TYPETREE_COMPOSITENODE_STORAGE_CONSTRUCTOR_SIGNATURE)
        : BaseT(gfs)
          , NodeType(DUNE_TYPETREE_COMPOSITENODE_CHILDVARIABLES)
#endif
      {}

      template<typename Transformation>
      CompositeUnfittedLocalFunctionSpaceNode (const GFS& gfs,
                                               const Transformation& t,
#if DUNE_VERSION_NEWER(DUNE_TYPETREE,2,4)
                                               std::shared_ptr<Children>... children)
        : BaseT(stackobject_to_shared_ptr(gfs))
          , NodeType(children...)
#else
                                               DUNE_TYPETREE_COMPOSITENODE_STORAGE_CONSTRUCTOR_SIGNATURE)
        : BaseT(stackobject_to_shared_ptr(gfs))
          , NodeType(DUNE_TYPETREE_COMPOSITENODE_CHILDVARIABLES)
#endif
      {}

      //! \brief bind local function space to entity
      template < typename Part >
      void bind (const Part& e, const bool inside)
      {
        // call method on base class, this avoid the barton neckman trick
        BaseT::bind(*this,e,inside);
      }

    };

#if HAVE_VARIADIC_TEMPLATES

    // transformation template, we need a custom template in order to inject the DOFIndex type into the UnfittedLocalFunctionSpace
    template<typename SourceNode, typename Transformation>
    struct variadic_composite_gfs_to_ulfs_template
    {
      template<typename ... TC>
      struct result
      {
        typedef CompositeUnfittedLocalFunctionSpaceNode<SourceNode,typename Transformation::DOFIndex,TC ...> type;
      };
    };

    // register CompositeGFS -> UnfittedLocalFunctionSpace transformation (variadic version)
    template<typename CompositeGridFunctionSpace, typename Params>
    Dune::TypeTree::TemplatizedGenericCompositeNodeTransformation<
        CompositeGridFunctionSpace,
        gfs_to_ulfs<Params>,
        variadic_composite_gfs_to_ulfs_template<CompositeGridFunctionSpace,gfs_to_ulfs<Params> >::template result
        >
    registerNodeTransformation(CompositeGridFunctionSpace* cgfs, gfs_to_ulfs<Params>* t, CompositeGridFunctionSpaceTag* tag);

#else

    // transformation template, we need a custom template in order to inject the DOFIndex type into the UnfittedLocalFunctionSpace
    template<typename SourceNode, typename Transformation>
    struct composite_gfs_to_ulfs_template
    {
      template<typename TC0,
          typename TC1,
          typename TC2,
          typename TC3,
          typename TC4,
          typename TC5,
          typename TC6,
          typename TC7,
          typename TC8,
          typename TC9>
      struct result
      {
        typedef CompositeUnfittedLocalFunctionSpaceNode<SourceNode,typename Transformation::DOFIndex,TC0,TC1,TC2,TC3,TC4,TC5,TC6,TC7,TC8,TC9> type;
      };
    };

    // register CompositeGFS -> UnfittedLocalFunctionSpace transformation (non-variadic version)
    template<typename CompositeGridFunctionSpace, typename Params>
    Dune::TypeTree::TemplatizedGenericCompositeNodeTransformation<
        CompositeGridFunctionSpace,
        gfs_to_ulfs<Params>,
        composite_gfs_to_ulfs_template<CompositeGridFunctionSpace,gfs_to_ulfs<Params> >::template result
        >
    registerNodeTransformation(CompositeGridFunctionSpace* cgfs, gfs_to_ulfs<Params>* t, CompositeGridFunctionSpaceTag* tag);

#endif

    //=======================================
    // local function space base: single component implementation
    //=======================================

    //! Tag denoting a LeafUnfittedLocalFunctionSpace
    struct LeafUnfittedLocalFunctionSpaceTag {};


    //! traits for single component local function space
    template<typename GFS, typename DOFIndex, typename N>
    struct LeafUnfittedLocalFunctionSpaceTraits : public PowerCompositeUnfittedLocalFunctionSpaceTraits<GFS,DOFIndex,N>
    {
      //! Type of local finite element
      typedef typename GFS::Traits::FiniteElementType FiniteElementType;

      typedef typename GFS::Traits::FiniteElementType FiniteElement;

      //! \brief Type of constraints engine
      typedef typename GFS::Traits::ConstraintsType ConstraintsType;

      typedef typename GFS::Traits::ConstraintsType Constraints;

    };

    //! single component local function space
    template<typename GFS, typename DOFIndex>
    class LeafUnfittedLocalFunctionSpaceNode
      : public UnfittedLocalFunctionSpaceBaseNode<GFS,DOFIndex>
        , public TypeTree::LeafNode
    {
      typedef UnfittedLocalFunctionSpaceBaseNode<GFS,DOFIndex> BaseT;

      template<typename>
      friend struct PropagateGlobalStorageVisitor;

      template<typename>
      friend struct ClearSizeVisitor;

      template<typename>
      friend struct UnfittedComputeSizeVisitor;

      template<typename>
      friend struct UnfittedFillIndicesVisitor;

    public:
      typedef LeafUnfittedLocalFunctionSpaceTraits<GFS,DOFIndex,LeafUnfittedLocalFunctionSpaceNode> Traits;

      typedef LeafUnfittedLocalFunctionSpaceTag ImplementationTag;

    private:
      typedef FiniteElementInterfaceSwitch<
          typename Traits::FiniteElementType
          > FESwitch;

    public:

      //! \brief initialize with grid function space
      template<typename Transformation>
      LeafUnfittedLocalFunctionSpaceNode (shared_ptr<const GFS> gfs, const Transformation& t)
        : BaseT(gfs)
      {}

      template<typename Transformation>
      LeafUnfittedLocalFunctionSpaceNode (const GFS& gfs, const Transformation& t)
        : BaseT(stackobject_to_shared_ptr(gfs))
      {}

      //! get finite element
      const typename Traits::FiniteElementType& finiteElement () const
      {
        return *pfe;
      }

      //! \brief get constraints engine
      const typename Traits::ConstraintsType& constraints () const
      {
        return this->pgfs->constraints();
      }

      //! Calculates the multiindices associated with the given entity.
      template<typename Entity, typename DOFIndexIterator>
      void dofIndices(const Entity& e, DOFIndexIterator it, DOFIndexIterator endit)
      {
        // get layout of entity
        const typename FESwitch::Coefficients &coeffs =
          FESwitch::coefficients(*pfe);

        typedef typename GFS::Traits::GridViewType GV;
        GV gv = this->gridFunctionSpace().gridView();

        const Dune::ReferenceElement<double,GV::Grid::dimension>& refEl =
          Dune::ReferenceElements<double,GV::Grid::dimension>::general(this->pfe->type());

        for (std::size_t i = 0; i < std::size_t(coeffs.size()); ++i, ++it)
        {
          // get geometry type of subentity
          Dune::GeometryType gt = refEl.type(coeffs.localKey(i).subEntity(),
                                            coeffs.localKey(i).codim());

          // evaluate consecutive index of subentity
          typename GV::IndexSet::IndexType index = gv.indexSet().subIndex(e,
                                                                          coeffs.localKey(i).subEntity(),
                                                                          coeffs.localKey(i).codim());

          // store data
          GFS::Ordering::Traits::DOFIndexAccessor::store(*it,gt,index,coeffs.localKey(i).index());

          // make sure we don't write past the end of the iterator range
          assert(it != endit);
        }
      }

      /** \brief write back coefficients for one element to container */
      template<typename GC, typename LC>
      void insert_constraints (const LC& lc, GC& gc) const
      {
      // LC and GC are maps of maps
      typedef typename LC::const_iterator local_col_iterator;
      typedef typename LC::value_type::second_type::const_iterator local_row_iterator;
      typedef typename GC::iterator global_col_iterator;
      typedef typename GC::value_type::second_type global_row_type;

      for (local_col_iterator cit=lc.begin(); cit!=lc.end(); ++cit)
        {

          // look up entry in global map, if not found, insert an empty one.
          global_col_iterator gcit = gc.insert(std::make_pair(std::ref(this->dofIndex(cit->first)),global_row_type())).first;

          // copy row to global container with transformed indices
          for (local_row_iterator rit=(cit->second).begin(); rit!=(cit->second).end(); ++rit)
            gcit->second[this->dofIndex(rit->first)] = rit->second;
        }
      }

      //! \brief bind local function space to entity
      template< typename Part >
      void bind (const Part& e, const bool inside)
      {
        // call method on base class, this avoid the barton neckman trick
        BaseT::bind(*this,e,inside);
      }

    private:
      typename FESwitch::Store pfe;
    };

    // Register LeafGFS -> UnfittedLocalFunctionSpace transformation
    template<typename GridFunctionSpace, typename Params>
    Dune::TypeTree::GenericLeafNodeTransformation<
        GridFunctionSpace,
        gfs_to_ulfs<Params>,
        LeafUnfittedLocalFunctionSpaceNode<GridFunctionSpace,typename gfs_to_ulfs<Params>::DOFIndex>
        >
    registerNodeTransformation(GridFunctionSpace* gfs, gfs_to_ulfs<Params>* t, LeafGridFunctionSpaceTag* tag);

    //=======================================
    // local function facade
    //=======================================

    template <typename GFS, typename TAG=AnySpaceTag>
    class UnfittedLocalFunctionSpace;

    /**
       \brief Create a local function space from a global function space

       The local function space can be tagged with on of the tags
       defined in localfunctionspacetags.hh. This allows to
       destinguish between trial and test space.

       If no TAG is specified the AnySpaceTag is used, which basicly
       states, that it is not clear, whether this is a trial of a test
       space.

       \extends UnfittedLocalFunctionSpaceBaseNode
     */
    template <typename GFS, typename TAG>
    class UnfittedLocalFunctionSpace :
      public Dune::TypeTree::TransformTree<GFS,gfs_to_ulfs<GFS> >::Type
    {
      typedef typename Dune::TypeTree::TransformTree<GFS,gfs_to_ulfs<GFS> >::Type BaseT;
      typedef typename BaseT::Traits::IndexContainer::size_type I;
      typedef typename BaseT::Traits::IndexContainer::size_type LocalIndex;

      template<typename>
      friend struct PropagateGlobalStorageVisitor;

      template<typename>
      friend struct ClearSizeVisitor;

      template<typename>
      friend struct UnfittedComputeSizeVisitor;

      template<typename>
      friend struct UnfittedFillIndicesVisitor;

      template<typename LFS, typename C, typename Tag>
      friend class LFSIndexCacheBase;

    public:
      typedef typename BaseT::Traits Traits;

      UnfittedLocalFunctionSpace(const GFS & gfs)
        : BaseT(TypeTree::TransformTree<GFS,gfs_to_ulfs<GFS> >::transform(gfs))
      {
        this->setup(*this);
      }

      UnfittedLocalFunctionSpace(const UnfittedLocalFunctionSpace & lfs)
        : BaseT(lfs)
      {
        // We need to reset the global pointers in the new LFS tree,
        // as they are still pointing to the global_storage of the
        // old tree.
        this->_dof_indices = &(this->_dof_index_storage);
        this->setup(*this);
      }

      LocalIndex localIndex (typename Traits::IndexContainer::size_type index) const
      {
        return LocalIndex(BaseT::localIndex(index));
      }

    private:
      // we don't support getChild yet, so let's hide it!
      template<int i>
      void getChild () const;
      template<int i>
      void child () const;
    };

    // specialization for AnySpaceTag
    template <typename GFS>
    class UnfittedLocalFunctionSpace<GFS, AnySpaceTag> :
      public Dune::TypeTree::TransformTree<GFS,gfs_to_ulfs<GFS> >::Type
    {
      typedef typename Dune::TypeTree::TransformTree<GFS,gfs_to_ulfs<GFS> >::Type BaseT;

      template<typename>
      friend struct PropagateGlobalStorageVisitor;

      template<typename>
      friend struct ClearSizeVisitor;

      template<typename>
      friend struct UnfittedComputeSizeVisitor;

      template<typename>
      friend struct UnfittedFillIndicesVisitor;

      template<typename LFS, typename C, typename Tag, bool>
      friend class LFSIndexCacheBase;

    public:
      UnfittedLocalFunctionSpace(const GFS & gfs)
        : BaseT(TypeTree::TransformTree<GFS,gfs_to_ulfs<GFS> >::transform(gfs))
      {
        this->_dof_indices = &(this->_dof_index_storage);
        this->setup(*this);
      }

      UnfittedLocalFunctionSpace(const UnfittedLocalFunctionSpace & lfs)
        : BaseT(lfs)
      {
        // We need to reset the global pointers in the new LFS tree,
        // as they are still pointing to the global_storage of the
        // old tree.
        this->_dof_indices = &(this->_dof_index_storage);
        this->setup(*this);
      }

    };

    //! \} group GridFunctionSpace


  }
}

#endif
