// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_UDG_CUTFEMMULTIPHASEOPERATOR_HH
#define DUNE_UDG_CUTFEMMULTIPHASEOPERATOR_HH

#include <dune/udg/pdelab/twophaseoperator.hh>

namespace Dune {
  namespace UDG {
    template<typename LOP, typename TR = double>
    class CutFEMMultiPhaseLocalOperatorWrapper :
      public Dune::PDELab::LocalOperatorDefaultFlags,
      public Dune::PDELab::InstationaryLocalOperatorDefaultMethods<TR>,
      public Dune::PDELab::FullVolumePattern,
      public Dune::PDELab::FullSkeletonPattern

    {
    private:
      LOP & lop;

    public:
      // pattern assembly flags
      enum { doPatternVolume = LOP::doPatternVolume };
      enum { doPatternSkeleton = LOP::doPatternSkeleton };
      enum { doPatternBoundary = LOP::doPatternBoundary };
      enum { doPatternVolumePostSkeleton = LOP::doPatternVolumePostSkeleton };

      // residual assembly flags
      enum { doAlphaVolume = LOP::doAlphaVolume };
      enum { doAlphaVolumePostSkeleton = LOP::doAlphaVolumePostSkeleton };
      enum { doLambdaVolume = LOP::doLambdaVolume };
      enum { doLambdaVolumePostSkeleton = LOP::doLambdaVolumePostSkeleton };
      enum { doAlphaSkeleton = LOP::doAlphaSkeleton };
      enum { doLambdaSkeleton = LOP::doLambdaSkeleton };
      enum { doAlphaBoundary = LOP::doAlphaBoundary };
      enum { doLambdaBoundary = LOP::doLambdaBoundary };

      // select twoSided option for real skeleton terms
      enum { doSkeletonTwoSided = LOP::doSkeletonTwoSided };

      CutFEMMultiPhaseLocalOperatorWrapper(LOP & lop_) :
        lop(lop_)
      {}

      void setTime(TR t_)
      {
        lop.setTime(t_);
      }

      // volume integral depending on test and ansatz functions
      template<typename EG, typename LFSU, typename X, typename LFSV, typename R>
      void alpha_volume (const EG& eg, const LFSU& lfsu, const X& x, const LFSV& lfsv, R& r) const
      {
        typedef typename LFSU::template Child<0>::Type DLFSU;
        typedef typename LFSV::template Child<0>::Type DLFSV;
        const int domainIndex = eg.subEntity().domainIndex();
        const DLFSU& dlfsu = lfsu.child(domainIndex);
        const DLFSV& dlfsv = lfsv.child(domainIndex);
        Dune::PDELab::LocalAssemblerCallSwitch<LOP,LOP::doAlphaVolume CALLSWITCHLINEARFLAG>::
        alpha_volume(lop,eg,dlfsu,x,dlfsv,r);
      }

      template<typename EG, typename LFSV, typename R>
      void lambda_volume (const EG& eg, const LFSV& lfsv, R& r) const
      {
        typedef typename LFSV::template Child<0>::Type DLFSV;
        const int domainIndex = eg.subEntity().domainIndex();
        const DLFSV& dlfsv = lfsv.child(domainIndex);
        Dune::PDELab::LocalAssemblerCallSwitch<LOP,LOP::doLambdaVolume CALLSWITCHLINEARFLAG>::
        lambda_volume(lop,eg,dlfsv,r);
      }

      template<typename EG, typename LFSU, typename X, typename LFSV, typename M>
      void jacobian_volume (const EG& eg, const LFSU& lfsu, const X& x, const LFSV& lfsv,
                            M & mat) const
      {
        typedef typename LFSU::template Child<0>::Type DLFSU;
        typedef typename LFSV::template Child<0>::Type DLFSV;
        const int domainIndex = eg.subEntity().domainIndex();
        const DLFSU& dlfsu = lfsu.child(domainIndex);
        const DLFSV& dlfsv = lfsv.child(domainIndex);
        Dune::PDELab::LocalAssemblerCallSwitch<LOP,LOP::doAlphaVolume CALLSWITCHLINEARFLAG>::
        jacobian_volume(lop,eg,dlfsu,x,dlfsv,mat);
      }

      template<typename IG, typename LFSU, typename X, typename LFSV, typename R>
      void alpha_skeleton (const IG& ig,
                           const LFSU& lfsu_s, const X& x_s, const LFSV& lfsv_s,
                           const LFSU& lfsu_n, const X& x_n, const LFSV& lfsv_n,
                           R& r_s, R& r_n) const
      {
        typedef typename LFSU::template Child<0>::Type DLFSU;
        typedef typename LFSV::template Child<0>::Type DLFSV;

        const int in_phase  = ig.intersection().insideDomainIndex();
        const int out_phase = ig.intersection().outsideDomainIndex();

        const DLFSU& dlfsu_s = lfsu_s.child(in_phase);
        const DLFSV& dlfsv_s = lfsv_s.child(in_phase);

        const DLFSU& dlfsu_n = lfsu_n.child(out_phase);
        const DLFSV& dlfsv_n = lfsv_n.child(out_phase);

        if(in_phase != out_phase) {
          Dune::PDELab::LocalAssemblerCallSwitch<LOP,LOP::doAlphaSkeleton CALLSWITCHLINEARFLAG>::
          alpha_skeleton(lop, ig, dlfsu_s, x_s, dlfsv_s, dlfsu_n, x_n, dlfsv_n, r_s, r_n);
        }
      }

      template<typename IG, typename LFSV, typename R>
      void lambda_skeleton(const IG& ig,const LFSV& lfsv_s, const LFSV& lfsv_n, R& r_s, R& r_n) const
      {
        typedef typename LFSV::template Child<0>::Type DLFSV;

        const int in_phase  = ig.intersection().insideDomainIndex();
        const int out_phase = ig.intersection().outsideDomainIndex();

        const DLFSV& dlfsv_s = lfsv_s.child(in_phase);
        const DLFSV& dlfsv_n = lfsv_n.child(out_phase);

        if(in_phase != out_phase) {
          Dune::PDELab::LocalAssemblerCallSwitch<LOP,LOP::doLambdaSkeleton CALLSWITCHLINEARFLAG>::
          lambda_skeleton(lop, ig, dlfsv_s, dlfsv_n, r_s, r_n);
        }
      }


      template<typename IG, typename LFSU, typename X, typename LFSV, typename M>
      void jacobian_skeleton
        ( const IG& ig,
        const LFSU& lfsu_s, const X& x_s, const LFSV& lfsv_s,
        const LFSU& lfsu_n, const X& x_n, const LFSV& lfsv_n,
        M & mat_ss, M & mat_sn,
        M & mat_ns, M & mat_nn) const
      {
        typedef typename LFSU::template Child<0>::Type DLFSU;
        typedef typename LFSV::template Child<0>::Type DLFSV;

        const int in_phase  = ig.intersection().insideDomainIndex();
        const int out_phase = ig.intersection().outsideDomainIndex();

        const DLFSU& dlfsu_s = lfsu_s.child(in_phase);
        const DLFSV& dlfsv_s = lfsv_s.child(in_phase);

        const DLFSU& dlfsu_n = lfsu_n.child(out_phase);
        const DLFSV& dlfsv_n = lfsv_n.child(out_phase);

        if(in_phase != out_phase) {
          Dune::PDELab::LocalAssemblerCallSwitch<LOP,LOP::doAlphaSkeleton CALLSWITCHLINEARFLAG>::
          jacobian_skeleton(lop, ig, dlfsu_s, x_s, dlfsv_s, dlfsu_n, x_n, dlfsv_n,
                            mat_ss, mat_sn, mat_ns, mat_nn);
        }
      }

      template<typename IG, typename LFSU, typename X, typename LFSV, typename R>
      void alpha_boundary (const IG& ig,
                           const LFSU& lfsu_s, const X& x_s, const LFSV& lfsv_s,
                           R& r_s) const
      {
        typedef typename LFSU::template Child<0>::Type DLFSU;
        typedef typename LFSV::template Child<0>::Type DLFSV;

        const int in_phase  = ig.intersection().insideDomainIndex();
        const int out_phase = ig.intersection().outsideDomainIndex();

        const DLFSU& dlfsu_s = lfsu_s.child(in_phase);
        const DLFSV& dlfsv_s = lfsv_s.child(in_phase);

        if(out_phase == -1) {
          Dune::PDELab::LocalAssemblerCallSwitch<LOP,LOP::doAlphaBoundary CALLSWITCHLINEARFLAG>::
          alpha_boundary(lop, ig, dlfsu_s, x_s, dlfsv_s, r_s);
        }
      }

      template<typename IG, typename LFSV, typename R>
      void lambda_boundary (const IG& ig, const LFSV& lfsv_s, R& r_s) const
      {
        typedef typename LFSV::template Child<0>::Type DLFSV;

        const int in_phase  = ig.intersection().insideDomainIndex();
        const int out_phase = ig.intersection().outsideDomainIndex();

        const DLFSV& dlfsv_s = lfsv_s.child(in_phase);

        if(out_phase == -1) {
          Dune::PDELab::LocalAssemblerCallSwitch<LOP,LOP::doLambdaBoundary CALLSWITCHLINEARFLAG>::
          lambda_boundary(lop, ig, dlfsv_s, r_s);
        }
      }


      template<typename IG, typename LFSU, typename X, typename LFSV, typename M>
      void jacobian_boundary
        ( const IG& ig, const LFSU& lfsu_s, const X& x_s, const LFSV& lfsv_s,
        M & mat_ss) const
      {
        typedef typename LFSU::template Child<0>::Type DLFSU;
        typedef typename LFSV::template Child<0>::Type DLFSV;

        const int in_phase  = ig.intersection().insideDomainIndex();
        const int out_phase = ig.intersection().outsideDomainIndex();

        const DLFSU& dlfsu_s = lfsu_s.child(in_phase);
        const DLFSV& dlfsv_s = lfsv_s.child(in_phase);

        if(out_phase == -1) {
          Dune::PDELab::LocalAssemblerCallSwitch<LOP,LOP::doAlphaBoundary CALLSWITCHLINEARFLAG>::
          jacobian_boundary(lop, ig, dlfsu_s, x_s, dlfsv_s, mat_ss);
        }
      }

      double getMinH() const { return lop.getMinH(); }
      double getMaxH() const { return lop.getMaxH(); }
    };
  }
}

#endif // DUNE_UDG_CUTFEMMULTIPHASEOPERATOR_HH
