// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef UNFITTED_LFE_HH
#define UNFITTED_LFE_HH

#include <dune/localfunctions/common/localkey.hh>
#include <dune/udg/common/entitypart.hh>
#include <dune/udg/common/intersectionpart.hh>

namespace Unfitted {

  template<class LFE,class SE>
  class UnfittedBasis
  {

  public:
    typedef typename LFE::Traits::LocalBasisType::Traits::DomainFieldType DF;
    typedef typename LFE::Traits::LocalBasisType::Traits::JacobianType JacobianType;
    typedef SE EntityPart;
    typedef typename EntityPart::Traits::GridView GridView;
    typedef typename EntityPart::Traits::Geometry Geometry;
    enum { dim = Geometry::dimension, dimw = Geometry::dimension };
    typedef Dune::UDG::DefaultGeometry<DF,dim,dim> DefaultGeometry;

    typedef Dune::FieldMatrix<DF,dimw,dim> JacobianInverseTransposed;
    typedef Dune::FieldVector<DF,dim> Coordinate;

    //! \brief export type traits for function signature
    typedef typename LFE::Traits::LocalBasisType::Traits EmbeddedTraits;

    struct Traits {
      typedef typename EmbeddedTraits::DomainFieldType DomainField;
      static const std::size_t dimDomainLocal = EmbeddedTraits::dimDomain;
      static const std::size_t dimDomainGlobal = EmbeddedTraits::dimDomain;
      typedef Dune::FieldVector<DomainField, dimDomainLocal> DomainLocal;
      typedef Dune::FieldVector<DomainField, dimDomainGlobal> DomainGlobal;

      typedef typename EmbeddedTraits::RangeFieldType RangeField;
      static const std::size_t dimRange = EmbeddedTraits::dimRange;
      typedef typename EmbeddedTraits::RangeType Range;

      typedef typename EmbeddedTraits::JacobianType Jacobian;

      static const std::size_t diffOrder = 1;
    };


    void reset() const
    {
      global = identity;
      global2bbox = identity;
      jac = default_jac;
    }

    //! \todo UnfittedBasis only works for bounding boxes that are affine
    //!       linear transformations of the reference bounding box, as
    //!       jac(Coordinate(0.5)) is chosen in the rebindTo*() methods
    //!       and then used in evaluateJacobian().
    UnfittedBasis(const LFE& lfe_, const int domain_index_, const int layer_,
                  bool scaleToBBox_ = true)
        : // A reference to the local finite element basis
        lb(lfe_.localBasis())
        , default_jac(0.0)
        ,
        // An identity mapping of type LocalGeometry
        identity(get_identity())
        , domain_index(domain_index_)
        , layer(layer_)
        , scaleToBBox(scaleToBBox_)
    {
      for(int i=0; i<dim; ++i)
        default_jac[i][i] = 1.0;
      reset();
    }

    template <class UEG>
    void rebindToEntityPart(const UEG & ueg) const
    {
      ueg.subEntity().resetLayer();
      // local coordinates of subEntity to global coordinates
      global = Dune::stackobject_to_shared_ptr(static_cast<const Geometry&>(ueg.subEntity().geometry()));

      ueg.subEntity().setLayer(layer);
      // local coordinates of bounding box to global coordinates
      if (scaleToBBox) {
        global2bbox.reset(new DefaultGeometry(ueg.subEntity().boundingBox()));
        jac = ueg.subEntity().boundingBox().jacobianInverseTransposed(Coordinate(0.5));
      } else {
        const auto& geo = ueg.subEntity().boundingBox().entity().geometry();
        global2bbox =
            std::make_shared<DefaultGeometry>(geo);
        jac = geo.jacobianInverseTransposed(Coordinate(.5));
      }
      ueg.subEntity().resetLayer();
    }

    template <class UIG>
    void rebindToIntersectionInside(const UIG & uig) const
    {
      uig.intersection().resetLayer();
      uig.intersection().setLayer(layer);
      // local coordinates of inside bounding box to global coordinates
      global.reset( new DefaultGeometry(uig.intersection().boundingBoxInside()) );

      // local coordinates of inside bounding box to global coordinates
      if (scaleToBBox) {
        global2bbox.reset(new DefaultGeometry(uig.intersection().boundingBoxInside()));
        jac = uig.intersection().boundingBoxInside().jacobianInverseTransposed(Coordinate(0.5));
      } else {
        const auto& geo = uig.intersection().inside().geometry();
        global2bbox =
            std::make_shared<DefaultGeometry>(geo);
        jac = geo.jacobianInverseTransposed(Coordinate(.5));
      }

      uig.intersection().resetLayer();
    }

    template <class UIG>
    void rebindToIntersectionOutside(const UIG & uig) const
    {
      uig.intersection().resetLayer();
      uig.intersection().setLayer(layer);
      // local coordinates of outside bounding box to global coordinates
      global.reset( new DefaultGeometry(uig.intersection().boundingBoxOutside()) );

      // local coordinates of outside bounding box to global coordinates
      if (scaleToBBox) {
        global2bbox.reset(new DefaultGeometry(uig.intersection().boundingBoxOutside()));
        jac = uig.intersection().boundingBoxOutside().jacobianInverseTransposed(Coordinate(0.5));
      } else {
        const auto& geo = uig.intersection().outside().geometry();
        global2bbox =
            std::make_shared<DefaultGeometry>(geo);
        jac = geo.jacobianInverseTransposed(Coordinate(.5));
      }

      uig.intersection().resetLayer();
    }

    template <class UIG>
    void rebindToIntersection(const UIG & uig) const
    {
      reset();
      uig.intersection().setLayer(layer);

      if (uig.intersection().insideDomainIndex() == domain_index) {
        rebindToIntersectionInside(uig);
      } else if (uig.intersection().outsideDomainIndex() == domain_index) {
        rebindToIntersectionOutside(uig);
      }

      uig.intersection().resetLayer();
    }

    //! \brief number of shape functions
    unsigned int size () const
    {
      return lb.size();
    }

    //! \brief Evaluate all shape functions
    inline void evaluateFunction (const typename Traits::DomainLocal& in,
                                  std::vector<typename Traits::Range>& out) const
    {
      // transform local coordinates of subEntity or inside/outside bounding box to global coordinates
      typename Traits::DomainLocal u_global = global->global(in);
      // transform global coordinates to local coordinates of (inside/outside) bounding box
      typename Traits::DomainLocal u_in = global2bbox->local(u_global);
      // evaluate
      lb.evaluateFunction(u_in,out);
    }

    //! return given derivative of all components
    template<unsigned int k>
    inline void evaluate (const std::array<int,k>& directions,
                          const typename Traits::DomainLocal& in,
                          std::vector<typename Traits::Range>& out) const
    {
      DUNE_THROW(Dune::NotImplemented,"This is not implemented for unfitted finite elements");
    }

    //! \brief Evaluate Jacobian of all shape functions
    inline void
    evaluateJacobian (const typename Traits::DomainLocal& in,         // position
                      std::vector<typename Traits::Jacobian>& out) const      // return value
    {
      // transform local coordinates of subEntity or inside/outside bounding box
      // to local coordinates of (inside/outside) bounding box
      typename Traits::DomainLocal u_in = global2bbox->local(global->global(in));

      // evaluate
      lb.evaluateJacobian(u_in,out);
      typename Traits::Jacobian tmp;
      for (unsigned int i=0; i<lb.size(); i++) {
        for(unsigned int r=0; r<tmp.N(); ++r)
          jac.mv(out[i][r],tmp[r]);

        out[i] = tmp;
      }
    }

    //! \brief Polynomial order of the shape functions
    unsigned int order () const
    {
      return lb.order();
    }

  private:

    // Create parameter matrix to create a LocalGeometry object which
    // represents the identity mapping.
    static DefaultGeometry* get_identity(){
      const Dune::ReferenceElement<DF,dim> & ref =
        Dune::ReferenceElements<DF,dim>::simplex();

      const int nCorners = ref.size(dim);
      std::vector<typename Geometry::GlobalCoordinate> coords(nCorners);
      for(int c=0; c<nCorners; ++c)
        coords[c] = ref.position(c,dim);

      return new DefaultGeometry(ref.type(0,0),coords);
    }

    UnfittedBasis (const UnfittedBasis &);

    const typename LFE::Traits::LocalBasisType & lb;
    mutable Dune::FieldMatrix<DF,dimw,dim> jac;
    mutable Dune::FieldMatrix<DF,dimw,dim> default_jac;

    mutable Dune::shared_ptr<const Geometry> global;
    // local coordinates of bbox to global coordinates, but only
    // used in the direction "global to bbox", therefore its name:
    mutable Dune::shared_ptr<const Geometry> global2bbox;
    const Dune::shared_ptr<const DefaultGeometry> identity;

    const int domain_index;
    const int layer;
    const bool scaleToBBox;
  };


  template<class LFE>
  class UnfittedCoefficients
  {
  private:
    const typename LFE::Traits::LocalCoefficientsType & lc;
    UnfittedCoefficients (const UnfittedCoefficients &);
  public:
    //! \brief Standard constructor
    UnfittedCoefficients (const LFE & lfe_)
      : lc(lfe_.localCoefficients())
    {}

    template <class UEG>
    void rebind(const UEG & ueg) const
    {}

    //! number of coefficients
    std::size_t size () const
    {
      return lc.size();
    }

    //! get i'th index
    const Dune::LocalKey& localKey (std::size_t i) const
    {
      return lc.localKey(i);
    }

  };


  template<class LFE, class SE>
  class UnfittedInterpolation
  {
  private:
    typedef typename LFE::Traits::LocalBasisType::Traits::DomainFieldType DF;
    typedef SE EntityPart;
    typedef typename EntityPart::Traits::Geometry Geometry;
    enum { dim = Geometry::dimension, dimw = Geometry::dimension };
    typedef Dune::UDG::DefaultGeometry<DF, dim, dim> DefaultGeometry;
    const LFE& lfe;
    mutable std::shared_ptr<DefaultGeometry> element2global;
    mutable std::shared_ptr<DefaultGeometry> bbox2global;
    UnfittedInterpolation (const UnfittedInterpolation &);

    template <class F>
    struct WrappedFunction {
      using Traits = typename F::Traits;
      template <class X, class Y>
      void evaluate(const X& x, Y& y) const
      {
        f.evaluate(element2global->local(bbox2global->global(x)), y);
      }
      const F& f;
      std::shared_ptr<DefaultGeometry> element2global;
      std::shared_ptr<DefaultGeometry> bbox2global;
    };

  public:
    UnfittedInterpolation (const LFE & lfe_) : lfe(lfe_)
    {}

    template <class UEG>
    void rebind(const UEG & ueg) const
    {
      element2global.reset(new DefaultGeometry(ueg.subEntity().boundingBox().entity().geometry()));
      bbox2global.reset(new DefaultGeometry(ueg.subEntity().boundingBox()));
    }

    //! determine coefficients interpolating a given function
    template<typename F, typename C>
    void interpolate (const F& f, std::vector<C>& out) const
    {
      // the function to be interpolated expects values on elements of the fundamental mesh.
      // However, when calling interpolate of the wrapped local finite element, it is called with
      // local coordinates with respect ot the bounding box. We thus need to transform them (via
      // WrappedFunction) from bounding box --> global --> element local
      lfe.localInterpolation().interpolate(WrappedFunction<F>{f, element2global, bbox2global}, out);
    }
  };


  template<class LFE, class SE>
  class UnfittedFiniteElement
  {
  public:

    struct Traits {
      typedef UnfittedBasis<LFE,SE> Basis;
      typedef UnfittedInterpolation<LFE, SE> Interpolation;
      typedef UnfittedCoefficients<LFE> Coefficients;
    };

    typedef LFE Imp;
    typedef SE EntityPart;

    UnfittedFiniteElement(Dune::shared_ptr<const LFE> lfe, const int domain_index, const int layer, bool scaleToBBox = true)
        : lfe_(lfe->clone())
        , basis_(*lfe, domain_index, layer, scaleToBBox)
        , coefficients_(*lfe)
        , interpolation_(*lfe)
        , domain_index_(domain_index)
        , layer_(layer)
        , scaleToBBox_(scaleToBBox)
    {
    }

    UnfittedFiniteElement(const LFE& lfe, const int domain_index, const int layer, bool scaleToBBox = true)
        : lfe_(lfe.clone())
        , basis_(lfe, domain_index, layer, scaleToBBox)
        , coefficients_(*lfe_)
        , interpolation_(*lfe_)
        , domain_index_(domain_index)
        , layer_(layer)
        , scaleToBBox_(scaleToBBox)
    {
    }

    UnfittedFiniteElement(const UnfittedFiniteElement& other)
        : lfe_(other.lfe_->clone())
        , basis_(*lfe_, other.domain_index_, other.layer_, other.scaleToBBox_)
        , coefficients_(*lfe_)
        , interpolation_(*lfe_)
        , domain_index_(other.domain_index_)
        , layer_(other.layer_)
        , scaleToBBox_(other.scaleToBBox_)
    {
    }

    UnfittedFiniteElement& operator=(const UnfittedFiniteElement& other)
    {
      lfe_ = other.lfe_->clone();
      basis_ = UnfittedBasis<LFE, SE>(*lfe_);
      coefficients_ = UnfittedCoefficients<LFE>(*lfe_);
      interpolation_ = UnfittedInterpolation<LFE, SE>(*lfe_);
      // the remaining members are const. make sure that we are only assigning compatible elements
      assert((domain_index_ == other.domain_index_) && (layer_ == other.layer_)
             && (scaleToBBox_ == other.scaleToBBox_));
    }

    const typename Traits::Basis& basis () const
    {
      return basis_;
    }

    const typename Traits::Coefficients& coefficients () const
    {
      return coefficients_;
    }

    const typename Traits::Interpolation& interpolation () const
    {
      return interpolation_;
    }

    Dune::GeometryType type () const
    {
      return lfe_->type();
    }

    int domainIndex() const
    {
      return domain_index_;
    }

    const LFE& localFiniteElement() const
    {
      return *lfe_;
    }

  private:
    Dune::shared_ptr<const LFE> lfe_;
    UnfittedBasis<LFE,SE> basis_;
    UnfittedCoefficients<LFE> coefficients_;
    UnfittedInterpolation<LFE,SE> interpolation_;
    const int domain_index_;
    const int layer_;
    const bool scaleToBBox_;
  };

} // Unfitted

#endif
