// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_PDELAB_TWOPART_FEM_HH
#define DUNE_PDELAB_TWOPART_FEM_HH

#include <dune/geometry/type.hh>
#include <dune/pdelab/finiteelementmap/finiteelementmap.hh>

#include <dune/localfunctions/monomial.hh>

//#include "phantom_fem.hh"
#include "finiteelement.hh"
#include <dune/localfunctions/common/virtualwrappers.hh>
#include <dune/udg/common/shared_ptr_tools.hh>
#include <dune/udg/vtriangulation.hh>

namespace Dune {
  namespace PDELab {

    /**
       @defgroup DummyLocalFiniteElement Dummy Finite Element

       \brief This is the implementation of a local finite element which has
       zero degrees of freedom.

       It may be used to \e deactivate elements (e.g. for unfitted
       finite element approaches).

       @{
     */
    template<class T>
    class DummyLocalBasis
    {
    public:

      typedef T Traits;

      DummyLocalBasis(){}

      //! \brief number of shape functions
      unsigned int size () const
      {
        return 0;
      }

      //! \brief Evaluate all shape functions
      inline void evaluateFunction (const typename Traits::DomainType& in,
                                    std::vector<typename Traits::RangeType>& out) const
      {
        out.clear();
      }

      //! return given derivative of all components
      template<unsigned int k>
      inline void evaluate (const std::array<int,k>& directions,
                            const typename Traits::DomainType& in,
                            std::vector<typename Traits::RangeType>& out) const
      {
        out.clear();
      }

      //! \brief Evaluate Jacobian of all shape functions
      inline void
      evaluateJacobian (const typename Traits::DomainType& in,         // position
                        std::vector<typename Traits::JacobianType>& out) const      // return value
      {
        out.resize(0);
      }

      //! \brief Polynomial order of the shape functions
      unsigned int order () const
      {
        return 0;
      }

      //! \brief Evaluate partial derivative of all shape functions
      void partial(const std::array<unsigned int, Traits::dimDomain>& order,
                          const typename Traits::DomainType& in,
                          std::vector<typename Traits::RangeType>& out) const
      {
        out.clear();
      }

    };

    class DummyLocalCoefficients
    {
      LocalKey key;

    public:
      //! \brief Standard constructor
      DummyLocalCoefficients ()
        :  key(LocalKey(0,0,0))
      {}

      //! number of coefficients
      std::size_t size () const
      {
        return 0;
      }

      //! get i'th index
      const LocalKey& localKey (std::size_t i) const
      {
        return key;
      }

    };

    class DummyLocalInterpolation
    {
    public:

      DummyLocalInterpolation ()
      { }

      //! determine coefficients interpolating a given function
      template<typename F, typename C>
      void interpolate (const F& f, std::vector<C>& out) const
      {
        out.clear();
      }
    };

    template<class BT>
    class DummyLocalFiniteElement
    {

    public:

      typedef LocalFiniteElementTraits<
          DummyLocalBasis<BT>,
          DummyLocalCoefficients,
          DummyLocalInterpolation
          > Traits;

      DummyLocalFiniteElement (const Dune::GeometryType & _gtype) :
        gtype(_gtype)
      {}

      const typename Traits::LocalBasisType& localBasis () const
      {
        return basis;
      }

      const typename Traits::LocalCoefficientsType& localCoefficients () const
      {
        return coefficients;
      }

      const typename Traits::LocalInterpolationType& localInterpolation () const
      {
        return interpolation;
      }

      GeometryType type () const
      {
        return gtype;
      }

      std::size_t size() const
      {
        return basis.size();
      }

    private:
      const Dune::GeometryType gtype;

      DummyLocalBasis<BT> basis;
      DummyLocalCoefficients coefficients;
      DummyLocalInterpolation interpolation;
    };
    /**
        @}
     */

    /**
       @defgroup UnfittedFEGroup Unfitted Finite Element Map
       @{
     */

    template<class SE, class BT>
    struct UnfittedFiniteElementMapBaseTraits
    {
      typedef BT BasisTraits;
      typedef Dune::LocalFiniteElementVirtualInterface<BasisTraits> Interface;
      typedef Unfitted::UnfittedFiniteElement<Interface,SE> UnfittedFE;
      typedef LocalFiniteElementMapTraits< UnfittedFE > Traits;
    };



    template<class T, class ST>
    class UnfittedFiniteElementMapBase
    {
    public:
      typedef typename T::Traits Traits;
      typedef ST SubTriangulation;

    protected:
      typedef typename ST::GridView GridView;
      typedef typename GridView::template Codim<0>::Entity CellEntity;
      typedef typename T::Interface Interface;

      typedef DummyLocalFiniteElement<typename T::BasisTraits> DummyLFE;
      typedef Dune::LocalFiniteElementVirtualImp<DummyLFE> VirtualDummyLFE;

      typedef typename T::UnfittedFE UnfittedFE;

      typedef typename Dune::UDG::EntityPartInterface<GridView> EntityPart;
      typedef typename Dune::UDG::IntersectionPartInterface<GridView> IntersectionPart;

      static int getLayer(const ST & st, const int domainIndex){
        typedef MultiLayerVirtualSubTriangulation<GridView> MultiLayerST;
        const MultiLayerST * mst = dynamic_cast<const MultiLayerST*>(&st);
        if(!mst) return 0;
        return mst->layerOfDomain(domainIndex);
      }

      virtual const UnfittedFE& findImp(const CellEntity & e) const = 0;

    public:
      /** \brief Constructor

          \param lfe_ [in] The local finite element which should be
          used.

          \param def_ [in] An object which defines the extent of the
          domain (best interface is not yet clear)
       */
      UnfittedFiniteElementMapBase
        (const Dune::GeometryType & gtype_, const ST & def_, const int domain_index_ = 0)
        : triangulation(wrap_into_shared_ptr<const ST>(def_)),
          domain_index(domain_index_),
          layer(getLayer(def_,domain_index)),

          dummy_lfe(gtype_),
          v_dummy_lfe(dummy_lfe),
          dummy_ulfe(v_dummy_lfe, domain_index,layer)
      {}

      virtual ~UnfittedFiniteElementMapBase() {}

      //! The layer in the triangulation in which the fe are defined
      size_t triangulationLayer() const { return layer; }

      /** \brief Return finite element (by copy) for given entity */
      const UnfittedFE find (const CellEntity& e) const
      {
        // Check whether the entity is host cell of a entity part of
        // the requested domain
        if(!triangulation->isHostCell(e,domain_index))
          return dummy_ulfe;
        else
          return this->findImp(e);
      }

      /** \brief Return finite element (by copy) for given entity */
      const UnfittedFE find (const EntityPart& ep, const bool) const
      {
        ep.setLayer(layer);
        const int this_domain_index = ep.domainIndex();
        const CellEntity& host = ep.boundingBox().entity();
        ep.resetLayer();

        if( this_domain_index != domain_index )
          return dummy_ulfe;
        else
          return this->findImp(host);
      }

      /** \brief Return finite element (by copy) for given entity */
      const UnfittedFE find (const IntersectionPart& ip, const bool inside) const
      {
        ip.setLayer(layer);
        const int this_domain_index =
          inside ? ip.insideDomainIndex() : ip.outsideDomainIndex();
        const CellEntity& host =
          inside ?
          ip.boundingBoxInside().entity() :
          ip.boundingBoxOutside().entity();
        ip.resetLayer();

        if( this_domain_index != domain_index )
          return dummy_ulfe;
        else
          return this->findImp(host);
      }

    protected:
      //! The object defining the domain extent
      Dune::shared_ptr<const SubTriangulation> triangulation;

      //! The domain_index for which to map
      const int domain_index;

      //! The domain_index for which to map
      const int layer;

      //! \brief The dummy local finite element.
      const DummyLFE dummy_lfe;
      //! \brief The virtual dummy local finite element.
      const VirtualDummyLFE v_dummy_lfe;
      //! \brief The unfitted virtual dummy local finite element.
      const UnfittedFE dummy_ulfe;
    };



    /**
       \brief Traits class for the UnfittedFiniteElementMap

       \tparam LFE The local finite element which is to be used.

       \tparam SE The implementation of the sub entity on which to
       integrate locally.
     */
    template<class LFE, class SE>
    struct UnfittedFiniteElementMapTraits :
      UnfittedFiniteElementMapBaseTraits
      <SE,typename LFE::Traits::LocalBasisType::Traits>
    {
      typedef LFE EmbeddedLocalFiniteElementType;
    };


    /**
       \brief A FE map which may be used to employ unfitted finite
       element methods.

       This map allows to \e deactivate grid entities which are
       outside some arbitrarily defined domain such that the
       corresponding \a GridFunctionSpace will not generate degrees of
       freedom for the entity.

       \tparam T        Traits class
       \tparam ST       Sub triangulation
     */
    template<class T, class ST>
    class UnfittedFiniteElementMap :
      public UnfittedFiniteElementMapBase<T,ST>,
      public LocalFiniteElementMapInterface
      < typename T::Traits, UnfittedFiniteElementMap<T,ST> >
    {
    public:
      typedef UnfittedFiniteElementMapBase<T,ST> Base;
      using Base::find;
      typedef typename T::Traits Traits;

    protected:

      typedef typename T::EmbeddedLocalFiniteElementType LFE;
      typedef Dune::LocalFiniteElementVirtualImp<LFE> VirtualLFE;

      const typename Base::UnfittedFE& findImp
        (const typename Base::CellEntity& e) const
      {
        return ulfe;
      }

    public:
      /** \brief Constructor

          \param lfe_ [in] The local finite element which should be
          used.

          \param def_ [in] An object which defines the extent of the
          domain (best interface is not yet clear)
       */
      UnfittedFiniteElementMap
        (const LFE & lfe_, const ST & def_, const int domain_index_ = 0, bool scaleToBBox = true)
        : Base(lfe_.type(),def_,domain_index_),
          lfe(wrap_into_shared_ptr<const LFE>(lfe_)),
          v_lfe(*lfe),
          ulfe(v_lfe, domain_index_,Base::layer,scaleToBBox)
      {}

      bool fixedSize() const
      {
        return false;
      }

      bool hasDOFs(int codim) const
      {
        return codim == 0;
      }

      std::size_t size(GeometryType gt) const
      {
        DUNE_THROW(Dune::PDELab::VariableElementSize, "the unfitted fem has variable element size");
      }

      std::size_t maxLocalSize() const
      {
        std::cout << "fem: returning maxLocalSize: " << lfe->localBasis().size() << "\n";
        return lfe->localBasis().size();
      }

      const LFE& embeddedLocalFiniteElement() const
      {
        return *lfe;
      }

    private:
      //! \brief shared_ptr to the EmbeddedLocalFiniteElement
      //! this ptr ensures that references to the object are always valid
      Dune::shared_ptr<const LFE> lfe;

      //! \brief The local finite element wrapped by the \a
      //! Dune::LocalFiniteElementVirtualImp implementation.
      const VirtualLFE v_lfe;

      //! \brief The virtual local finite element v_lfe but wrapped by
      //! the UnfittedFiniteElement.
      const typename Base::UnfittedFE ulfe;
    };


    /** @} */
  }
}

#endif
