#ifndef DUNE_UDG_GHOST_PENALTY_HH
#define DUNE_UDG_GHOST_PENALTY_HH

#include <dune/pdelab/backend/istl.hh>
#include <dune/pdelab/ordering/chunkedblockordering.hh>

#include <dune/udg/pdelab/ghostpenaltylocaloperator.hh>
#include <dune/udg/pdelab/ghostpenaltylocaloperatorwrapper.hh>
#include <dune/udg/pdelab/gridfunctionspaceutilities.hh>
#include <dune/udg/pdelab/restrictedfiniteelementmap.hh>

#include <dune/udg/test/utility/vtk.hh>

namespace Dune
{
  namespace UDG
  {
    template <typename ST, typename UGFS, typename EdgeNormProvider, typename M>
    void add_ghost_penalty(const ST& subTriangulation, const UGFS& ugfs,
                           const EdgeNormProvider& edgeNormProvider, double penalty,
                           const std::vector<double>& conductivities, bool restrictToInterfaceZone,
                           unsigned int intorderadd, M& udgMatrix)
    {
      const int phases = UGFS::CHILDREN;
      // construct ghost penalty local operators
      using GPLOP = Dune::UDG::GhostPenaltyLocalOperator<EdgeNormProvider, ST::dim>;
      if (conductivities.size() != phases) {
        DUNE_THROW(Dune::Exception, "expected " << phases << " conductivity values but got "
                                                << conductivities.size());
      }
      std::vector<GPLOP> lops;
      for (unsigned int i = 0; i < phases; ++i) {
        lops.emplace_back(penalty, edgeNormProvider, conductivities[i], intorderadd);
      }
      using LOPW = Dune::UDG::GhostPenaltyLocalOperatorWrapper<ST, GPLOP>;
      LOPW lopw(Dune::stackobject_to_shared_ptr(subTriangulation), lops, restrictToInterfaceZone);

      // construct restricted finite element maps
      using GridView = typename UGFS::Traits::GridView;
      // assuming every domain has the same LFE
      using LFE =
          std::decay_t<decltype(ugfs.child(0).finiteElementMap().embeddedLocalFiniteElement())>;
      using WFEM = Dune::UDG::RestrictedFiniteElementMap<GridView, LFE>;
      std::array<std::shared_ptr<WFEM>, phases> fems;
      for (unsigned int i = 0; i < phases; ++i) {
        const auto& lfe = ugfs.child(i).finiteElementMap().embeddedLocalFiniteElement();
        fems[i] = std::make_shared<WFEM>(ugfs.gridView(), lfe,
                                         [&subTriangulation, i](const auto& element) {
                                           return !subTriangulation.isHostCell(element, i);
                                         });
      }

      // construct domain-local gridfunctionspaces
      using GPGFS = Dune::PDELab::
          GridFunctionSpace<Dune::PDELab::AllEntitySet<GridView>, WFEM, Dune::PDELab::NoConstraints,
                            Dune::PDELab::ISTL::VectorBackend<Dune::PDELab::ISTL::Blocking::fixed,
                                                              M::block_type::cols>>;
      std::array<std::shared_ptr<GPGFS>, phases> gfss;
      for (unsigned int i = 0; i < phases; ++i) {
        gfss[i] = std::make_shared<GPGFS>(ugfs.gridView(), *fems[i]);
      }

      // construct common glocal gridfunctionspace
      using PVBE = typename UGFS::Traits::Backend;
      using OrderingTag = typename UGFS::Traits::OrderingTag;
      using GPPGFS = Dune::PDELab::PowerGridFunctionSpace<GPGFS, phases, PVBE, OrderingTag>;
      auto gppgfs = make_power_gfs(gfss, ugfs.backend(), ugfs.orderingTag());

      // assemble ghost penalty jacobian into the udg matrix
      using MBE = Dune::PDELab::ISTL::BCRSMatrixBackend<>;
      Dune::PDELab::GridOperator<GPPGFS, GPPGFS, LOPW, MBE, double, double, double> gpgo(
          *gppgfs, *gppgfs, lopw, MBE(2 * GridView::dimension + 1));
      using GPV = Dune::PDELab::Backend::Vector<GPPGFS, double>;
      Dune::PDELab::Backend::Matrix<MBE, GPV, GPV, double> ghost_penalty_jacobian(gpgo, 0.0);
      GPV dummy(*gppgfs, 0.0);
      gpgo.jacobian(dummy, ghost_penalty_jacobian);
      using Dune::PDELab::Backend::native;
      native(ghost_penalty_jacobian) += native(udgMatrix);
      native(udgMatrix) = native(ghost_penalty_jacobian);
    }

    template <typename ST, typename UGFS, typename EdgeNormProvider, typename M>
    void add_ghost_penalty_cutfem(const ST& subTriangulation, const UGFS& ugfs,
                                  const EdgeNormProvider& edgeNormProvider, double penalty,
                                  const std::vector<double>& conductivities,
                                  unsigned int intorderadd, M& cutfemMatrix)
    {
      const int phases = UGFS::CHILDREN;
      // construct ghost penalty local operators
      using GPLOP = Dune::UDG::GhostPenaltyLocalOperator<EdgeNormProvider, ST::dim>;
      if (conductivities.size() != phases) {
        DUNE_THROW(Dune::Exception, "expected " << phases << " conductivity values but got "
                                                << conductivities.size());
      }
      std::vector<GPLOP> lops;
      for (unsigned int i = 0; i < phases; ++i) {
        lops.emplace_back(penalty, edgeNormProvider, conductivities[i], intorderadd);
      }
      using LOPW = Dune::UDG::GhostPenaltyLocalOperatorWrapper<ST, GPLOP>;
      LOPW lopw(Dune::stackobject_to_shared_ptr(subTriangulation), lops, true);

      // construct restricted finite element maps
      using GridView = typename UGFS::Traits::GridView;
      // assuming every domain has the same LFE
      using LFE =
          std::decay_t<decltype(ugfs.child(0).finiteElementMap().embeddedLocalFiniteElement())>;
      using WFEM = Dune::UDG::RestrictedFiniteElementMap<GridView, LFE>;
      std::array<std::shared_ptr<WFEM>, phases> fems;
      for (unsigned int i = 0; i < phases; ++i) {
        const auto& lfe = ugfs.child(i).finiteElementMap().embeddedLocalFiniteElement();
        fems[i] = std::make_shared<WFEM>(
            ugfs.gridView(), lfe, [&subTriangulation, i](const auto& element) {
                                           return !subTriangulation.isHostCell(element, i);
            });
      }

      // construct domain-local gridfunctionspaces
      using GPGFS = Dune::PDELab::GridFunctionSpace<Dune::PDELab::AllEntitySet<GridView>, WFEM,
                                                    Dune::PDELab::NoConstraints,
                                                    Dune::PDELab::ISTL::VectorBackend<>>;
      std::array<std::shared_ptr<GPGFS>, phases> gfss;
      for (unsigned int i = 0; i < phases; ++i) {
        gfss[i] = std::make_shared<GPGFS>(ugfs.gridView(), *fems[i]);
      }

      // construct common glocal gridfunctionspace
      using PVBE = typename UGFS::Traits::Backend;
      using OrderingTag = typename UGFS::Traits::OrderingTag;
      using GPPGFS = Dune::PDELab::PowerGridFunctionSpace<GPGFS, phases, PVBE, OrderingTag>;
      auto gppgfs = make_power_gfs(gfss, ugfs.backend(), ugfs.orderingTag());

      // assemble ghost penalty jacobian into the cutfem matrix
      using MBE = Dune::PDELab::ISTL::BCRSMatrixBackend<>;
      Dune::PDELab::GridOperator<GPPGFS, GPPGFS, LOPW, MBE, double, double, double> gpgo(
          *gppgfs, *gppgfs, lopw, MBE(2 * GridView::dimension + 1));
      using GPV = Dune::PDELab::Backend::Vector<GPPGFS, double>;
      Dune::PDELab::Backend::Matrix<MBE, GPV, GPV, double> ghost_penalty_jacobian(gpgo, 0.0);
      GPV dummy(*gppgfs, 0.0);
      gpgo.jacobian(dummy, ghost_penalty_jacobian);
      using Dune::PDELab::Backend::native;
      native(ghost_penalty_jacobian) += native(cutfemMatrix);
      native(cutfemMatrix) = native(ghost_penalty_jacobian);
    }
  }
}

#endif // DUNE_UDG_GHOST_PENALTY_HH
