// -*- tab-width: 4; indent-tabs-mode: nil -*-
#ifndef DUNE_UDG_GHOST_PENALTY_LOCAL_OPERATOR_HH
#define DUNE_UDG_GHOST_PENALTY_LOCAL_OPERATOR_HH

#include <dune/common/exceptions.hh>
#include <dune/common/fvector.hh>

#include <dune/geometry/quadraturerules.hh>
#include <dune/geometry/referenceelements.hh>

#include <dune/localfunctions/common/interfaceswitch.hh>

#include <dune/typetree/pairtraversal.hh>
#include <dune/typetree/transformation.hh>
#include <dune/typetree/visitor.hh>

#include <dune/pdelab/common/quadraturerules.hh>
#include <dune/pdelab/localoperator/flags.hh>
#include <dune/pdelab/localoperator/idefault.hh>
#include <dune/pdelab/localoperator/pattern.hh>

namespace Dune
{
  namespace UDG
  {
    template <class EdgeNormProvider, int dim>
    class GhostPenaltyLocalOperator
        : public Dune::PDELab::FullVolumePattern,
          public Dune::PDELab::FullSkeletonPattern,
          public Dune::PDELab::LocalOperatorDefaultFlags,
          public Dune::PDELab::InstationaryLocalOperatorDefaultMethods<double>
    {
    public:
      enum { doPatternSkeleton = true };
      enum { doPatternVolume = true };
      enum { doAlphaSkeleton = true };

      GhostPenaltyLocalOperator(double penalty, const EdgeNormProvider& edgeNormProvider,
                                double conductivity, unsigned int intorderadd = 0,
                                double quadrature_factor = 2.0)
          : penalty_(penalty)
          , edgeNormProvider_(edgeNormProvider)
          , conductivity_(conductivity)
          , intorderadd_(intorderadd)
          , quadrature_factor_(quadrature_factor)
      {
      }

      template <typename IG, typename LFSU, typename X, typename LFSV, typename R>
      void alpha_skeleton(const IG& ig, const LFSU& lfsu_s, const X& x_s, const LFSV& lfsv_s,
                          const LFSU& lfsu_n, const X& x_n, const LFSV& lfsv_n, R& r_s,
                          R& r_n) const
      {
        static_assert(std::is_same<LFSU, LFSV>::value, "assuming Galerkin method");
        // domain and range field type
        typedef Dune::FiniteElementInterfaceSwitch<typename LFSV::Traits::FiniteElementType>
            FESwitch;
        typedef Dune::BasisInterfaceSwitch<typename FESwitch::Basis> BasisSwitch;
        typedef typename BasisSwitch::DomainField DF;
        typedef typename BasisSwitch::RangeField RF;
        typedef typename BasisSwitch::Range RangeType;
        typedef typename LFSV::Traits::SizeType size_type;

        // dimensions
        //const int dim = IG::dimension;
        const int order_s = FESwitch::basis(lfsu_s.finiteElement()).order();
        const int order_n = FESwitch::basis(lfsu_n.finiteElement()).order();
        const int intorder = intorderadd_ + quadrature_factor_ * std::max(order_s, order_n);

        // select quadrature rule for face
        const auto gtface = ig.geometry().type();
        const auto& rule = Dune::QuadratureRules<DF, dim - 1>::rule(gtface, intorder);

        const auto& insideEntity = ig.inside();
        const auto& outsideEntity = ig.outside();
        const auto& localFaceCenter =
            Dune::ReferenceElements<DF, dim - 1>::general(gtface).position(0, 0);
        const auto n_F = ig.unitOuterNormal(localFaceCenter);

        std::vector<Dune::FieldMatrix<RF, 1, dim>> gradphi_s(lfsu_s.size());
        std::vector<Dune::FieldMatrix<RF, 1, dim>> gradphi_n(lfsu_n.size());

        RF h_F;
        edgeNormProvider_.edgeNorm(ig, h_F);

        for (auto&& qp : rule) {
          const auto iplocal_s = ig.geometryInInside().global(qp.position());
          const auto iplocal_n = ig.geometryInOutside().global(qp.position());

          BasisSwitch::gradient(FESwitch::basis(lfsu_s.finiteElement()), insideEntity.geometry(),
                                iplocal_s, gradphi_s);
          BasisSwitch::gradient(FESwitch::basis(lfsu_n.finiteElement()), outsideEntity.geometry(),
                                iplocal_n, gradphi_n);

          // compute gradient of u
          Dune::FieldVector<RF, dim> gradu_s(0.0);
          for (size_type i = 0; i < lfsu_s.size(); i++)
            gradu_s.axpy(x_s(lfsu_s, i), gradphi_s[i][0]);
          Dune::FieldVector<RF, dim> gradu_n(0.0);
          for (size_type i = 0; i < lfsu_n.size(); i++)
            gradu_n.axpy(x_n(lfsu_n, i), gradphi_n[i][0]);

          const RF harmonic_average = conductivity_;
          const RF factor = harmonic_average * penalty_ * h_F * qp.weight()
                            * ig.geometry().integrationElement(qp.position());

          const RF jump_gradu = factor * ((gradu_s * n_F) - (gradu_n * n_F));

          for (size_type i = 0; i < lfsu_s.size(); i++)
            r_s.accumulate(lfsu_s, i, jump_gradu * (n_F * gradphi_s[i][0]));
          for (size_type i = 0; i < lfsu_n.size(); i++)
            r_n.accumulate(lfsu_n, i, -jump_gradu * (n_F * gradphi_n[i][0]));
        }
      }

      template <typename IG, typename LFSU, typename X, typename LFSV, typename M>
      void jacobian_skeleton(const IG& ig, const LFSU& lfsu_s, const X& DUNE_UNUSED(x_s),
                             const LFSV& lfsv_s, const LFSU& lfsu_n, const X& DUNE_UNUSED(x_n),
                             const LFSV& lfsv_n, M& mat_ss, M& mat_sn, M& mat_ns, M& mat_nn) const
      {
        static_assert(std::is_same<LFSU, LFSV>::value, "assuming Galerkin method");
        //add_to_intersection(ig);
        // domain and range field type
        typedef Dune::FiniteElementInterfaceSwitch<typename LFSV::Traits::FiniteElementType>
            FESwitch;
        typedef Dune::BasisInterfaceSwitch<typename FESwitch::Basis> BasisSwitch;
        typedef typename BasisSwitch::DomainField DF;
        typedef typename BasisSwitch::RangeField RF;
        typedef typename BasisSwitch::Range RangeType;
        typedef typename LFSV::Traits::SizeType size_type;

        // dimensions
        //const int dim = IG::dimension;
        const int order_s = FESwitch::basis(lfsu_s.finiteElement()).order();
        const int order_n = FESwitch::basis(lfsu_n.finiteElement()).order();
        const int intorder = intorderadd_ + quadrature_factor_ * std::max(order_s, order_n);

        // select quadrature rule for face
        const auto gtface = ig.geometry().type();
        const auto& rule = Dune::QuadratureRules<DF, dim - 1>::rule(gtface, intorder);

        const auto& insideEntity = ig.inside();
        const auto& outsideEntity = ig.outside();
        const auto& localFaceCenter =
            Dune::ReferenceElements<DF, dim - 1>::general(gtface).position(0, 0);
        const auto& n_F = ig.unitOuterNormal(localFaceCenter);

        std::vector<Dune::FieldMatrix<RF, 1, dim>> gradphi_s(lfsu_s.size());
        std::vector<Dune::FieldMatrix<RF, 1, dim>> gradphi_n(lfsu_n.size());
        std::vector<RF> gradphi_n_F_s(lfsu_s.size());
        std::vector<RF> gradphi_n_F_n(lfsu_n.size());

        RF h_F;
        edgeNormProvider_.edgeNorm(ig, h_F);

        for (auto&& qp : rule) {
          const auto iplocal_s = ig.geometryInInside().global(qp.position());
          const auto iplocal_n = ig.geometryInOutside().global(qp.position());

          BasisSwitch::gradient(FESwitch::basis(lfsu_s.finiteElement()), insideEntity.geometry(),
                                iplocal_s, gradphi_s);
          BasisSwitch::gradient(FESwitch::basis(lfsu_n.finiteElement()), outsideEntity.geometry(),
                                iplocal_n, gradphi_n);
          for (unsigned int i = 0; i < lfsu_s.size(); ++i)
            gradphi_n_F_s[i] = gradphi_s[i][0] * n_F;
          for (unsigned int i = 0; i < lfsu_n.size(); ++i)
            gradphi_n_F_n[i] = gradphi_n[i][0] * n_F;

          const RF harmonic_average = conductivity_;
          const RF factor = harmonic_average * penalty_ * h_F * qp.weight()
                            * ig.geometry().integrationElement(qp.position());

          for (size_type j = 0; j < lfsu_s.size(); j++) {
            for (size_type i = 0; i < lfsu_s.size(); i++) {
              mat_ss.accumulate(lfsu_s, i, lfsv_s, j,
                                (gradphi_n_F_s[j] * gradphi_n_F_s[i]) * factor);
            }
          }
          for (size_type j = 0; j < lfsu_n.size(); j++) {
            for (size_type i = 0; i < lfsu_s.size(); i++) {
              mat_sn.accumulate(lfsu_s, i, lfsv_n, j,
                                -(gradphi_n_F_n[j] * gradphi_n_F_s[i]) * factor);
            }
          }
          for (size_type j = 0; j < lfsu_s.size(); j++) {
            for (size_type i = 0; i < lfsu_n.size(); i++) {
              mat_ns.accumulate(lfsu_n, i, lfsv_s, j,
                                -(gradphi_n_F_s[j] * gradphi_n_F_n[i]) * factor);
            }
          }
          for (size_type j = 0; j < lfsu_n.size(); j++) {
            for (size_type i = 0; i < lfsu_n.size(); i++) {
              mat_nn.accumulate(lfsu_n, i, lfsv_n, j,
                                (gradphi_n_F_n[j] * gradphi_n_F_n[i]) * factor);
            }
          }
        }
      }

      const std::vector<std::pair<std::vector<Dune::FieldVector<double, dim>>, double>>&
      intersections() const
      {
        return intersections_;
      }

    private:
      template <class IG>
      void add_to_intersection(const IG& ig) const
      {
        std::vector<Dune::FieldVector<double, dim>> corners;
        const auto& geo = ig.geometry();
        for (unsigned int i = 0; i < geo.corners(); ++i) {
          corners.push_back(geo.corner(i));
        }
        intersections_.push_back({corners, conductivity_});
      }
      double penalty_;
      EdgeNormProvider edgeNormProvider_;
      double conductivity_;
      unsigned int intorderadd_;
      double quadrature_factor_;
      mutable std::vector<std::pair<std::vector<Dune::FieldVector<double, dim>>, double>> intersections_;
    };
  }
}

#endif
