#ifndef DUNE_UDG_GHOST_PENALTY_LOCAL_OPERATOR_WRAPPER_HH
#define DUNE_UDG_GHOST_PENALTY_LOCAL_OPERATOR_WRAPPER_HH

#include <memory>

#include <dune/pdelab/localoperator/callswitch.hh>
#include <dune/pdelab/localoperator/flags.hh>
#include <dune/pdelab/localoperator/idefault.hh>
#include <dune/pdelab/localoperator/pattern.hh>

namespace Dune
{
  namespace UDG
  {
    template <class ST, class LOP>
    class GhostPenaltyLocalOperatorWrapper
        : public Dune::PDELab::LocalOperatorDefaultFlags,
          public Dune::PDELab::InstationaryLocalOperatorDefaultMethods<double>,
          public Dune::PDELab::FullVolumePattern,
          public Dune::PDELab::FullSkeletonPattern
    {
    public:
      enum { doPatternVolume = LOP::doPatternVolume };
      enum { doPatternSkeleton = LOP::doPatternSkeleton };

      enum { doAlphaSkeleton = LOP::doAlphaSkeleton };
      enum { doLambdaSkeleton = LOP::doLambdaSkeleton };

      enum { doSkeletonTwoSided = LOP::doSkeletonTwoSided };

      /**
       * \param lops ghost penalty local operators for each subdomain
       */
      explicit GhostPenaltyLocalOperatorWrapper(std::shared_ptr<const ST> subTriangulation,
                                                const std::vector<LOP>& lops,
                                                bool restrictToInterfaceZone = false)
          : subTriangulation_(subTriangulation)
          , lops_(lops)
          , restrictToInterfaceZone_(restrictToInterfaceZone)
      {
      }

      template <typename IG, typename LFSU, typename X, typename LFSV, typename R>
      void alpha_skeleton(const IG& ig, const LFSU& lfsu_s, const X& x_s, const LFSV& lfsv_s,
                          const LFSU& lfsu_n, const X& x_n, const LFSV& lfsv_n, R& r_s,
                          R& r_n) const
      {
        static_assert(LFSU::CHILDREN == LFSV::CHILDREN, "illegal LFS");
        assert(LFSU::CHILDREN == lops_.size());

        for (unsigned int i = 0; i < LFSU::CHILDREN; ++i) {
          if (lfsu_s.child(i).size() == 0 || lfsu_n.child(i).size() == 0
              || lfsv_s.child(i).size() == 0 || lfsv_n.child(i).size() == 0)
            continue;

          if (restrictToInterfaceZone_) {
            const auto& in = ig.inside();
            const auto& out = ig.outside();
            // dont add ghost penalty on completely internal edges
            if (subTriangulation_->cutCellInformation().filledBySingleCutCell(in)
                && subTriangulation_->cutCellInformation().filledBySingleCutCell(out)) {
              continue;
            }
          }
          Dune::PDELab::LocalAssemblerCallSwitch<LOP, LOP::doAlphaSkeleton CALLSWITCHLINEARFLAG>::alpha_skeleton(
              lops_[i], ig, lfsu_s.child(i), x_s, lfsv_s.child(i), lfsu_n.child(i), x_n,
              lfsv_n.child(i), r_s, r_n);
        }
      }

      template <typename IG, typename LFSU, typename X, typename LFSV, typename M>
      void jacobian_skeleton(const IG& ig, const LFSU& lfsu_s, const X& x_s, const LFSV& lfsv_s,
                             const LFSU& lfsu_n, const X& x_n, const LFSV& lfsv_n, M& mat_ss,
                             M& mat_sn, M& mat_ns, M& mat_nn) const
      {
        static_assert(LFSU::CHILDREN == LFSV::CHILDREN, "illegal LFS");
        assert(LFSU::CHILDREN == lops_.size());

        const auto& inside = ig.inside();
        const auto& outside = ig.outside();

        for (unsigned int i = 0; i < LFSU::CHILDREN; ++i) {
          if (lfsu_s.child(i).size() == 0 || lfsu_n.child(i).size() == 0
              || lfsv_s.child(i).size() == 0 || lfsv_n.child(i).size() == 0)
            continue;
          if (restrictToInterfaceZone_) {
            // dont add ghost penalty on completely internal edges
            if (subTriangulation_->cutCellInformation().filledBySingleCutCell(inside)
                && subTriangulation_->cutCellInformation().filledBySingleCutCell(outside)) {
              continue;
            }
          }
          Dune::PDELab::LocalAssemblerCallSwitch<LOP, LOP::doAlphaSkeleton CALLSWITCHLINEARFLAG>::jacobian_skeleton(
              lops_[i], ig, lfsu_s.child(i), x_s, lfsv_s.child(i), lfsu_n.child(i), x_n,
              lfsv_n.child(i), mat_ss, mat_sn, mat_ns, mat_nn);
        }
      }

      const LOP& lop(std::size_t i) const
      {
        return lops_[i];
      }

    private:
      std::shared_ptr<const ST> subTriangulation_;
      std::vector<LOP> lops_;
      bool restrictToInterfaceZone_;
    };
  }
}

#endif // DUNE_UDG_GHOST_PENALTY_LOCAL_OPERATOR_WRAPPER_HH
