// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef UNFITTED_DISCRETE_GF_HH
#define UNFITTED_DISCRETE_GF_HH

#include <dune/localfunctions/common/interfaceswitch.hh>

namespace Dune {
  namespace PDELab {

    /**
     * \brief Convert an \b unfitted grid function space and a
     *        coefficient vector into a grid function.
     *
     * \tparam GFS Type of grid function space
     * \tparam U   Type of coefficients vector
     * \tparam ST  Type of the sub triangulation
     */
    template <typename GFS, typename U, typename ST>
    class UnfittedDiscreteGridFunction
      : public GridFunctionInterface<
            GridFunctionTraits<
                typename GFS::Traits::GridViewType,
                typename BasisInterfaceSwitch<
                    typename FiniteElementInterfaceSwitch<
                        typename GFS::Traits::FiniteElementType
                        >::Basis
                    >::RangeField,
                BasisInterfaceSwitch<
                    typename FiniteElementInterfaceSwitch<
                        typename GFS::Traits::FiniteElementType
                        >::Basis
                    >::dimRange,
                typename BasisInterfaceSwitch<
                    typename FiniteElementInterfaceSwitch<
                        typename GFS::Traits::FiniteElementType
                        >::Basis
                    >::Range
                >,
            UnfittedDiscreteGridFunction<GFS,U,ST>
            >
    {
    protected:
      typedef BasisInterfaceSwitch<
          typename FiniteElementInterfaceSwitch<
              typename GFS::Traits::FiniteElementType
              >::Basis
          > BasisSwitch;
      typedef GridFunctionInterface<
          GridFunctionTraits<
              typename GFS::Traits::GridViewType,
              typename BasisSwitch::RangeField,
              BasisSwitch::dimRange,
              typename BasisSwitch::Range
              >,
          UnfittedDiscreteGridFunction<GFS,U,ST>
          > BaseT;

    public:
      typedef typename BaseT::Traits Traits;

      /**
       * \brief Construct an UnfittedDiscreteGridFunction.
       *
       * \param gfs          The grid function space
       * \param u            The coefficients vector
       * \param st           The sub triangulation object
       * \param domain_index The domain index of the function's domain
       */
      UnfittedDiscreteGridFunction (const GFS& gfs, const U& u,
                                    const ST& st, const int domain_index = 0, bool scaleToBBox = true)
        : pgfs_(stackobject_to_shared_ptr(gfs)), lfs_(gfs), lfsCache_(lfs_),
          u_view_(u), ul_(gfs.maxLocalSize()), yb_(gfs.maxLocalSize()),
          st_(st), domain_index_(domain_index), scaleToBBox_(scaleToBBox)
      {}

      //! evaluate grid function on element
      inline void evaluate (const typename Traits::ElementType& e,
                            const typename Traits::DomainType& x,
                            typename Traits::RangeType& y) const
      {
        const typename ST::BoundingBox& bbox = st_.BBox(domain_index_,e);

        lfs_.bind(*(bbox.entity()));
        lfsCache_.update();
        u_view_.bind(lfsCache_);
        u_view_.read(ul_);
        u_view_.unbind();

        typename Traits::DomainType gx;
        if (scaleToBBox_)
          gx = bbox.local(e.geometry().global(x));
        else
          gx = x;
        typedef FiniteElementInterfaceSwitch<
            typename LFS::Traits::FiniteElementType
            > FESwitch;
        FESwitch::basis(lfs_.finiteElement()).evaluateFunction(gx,yb_);

        y = 0.0;
        for (unsigned int i = 0; i < yb_.size(); i++)
          y.axpy(ul_[i],yb_[i]);
      }

      //! get a reference to the grid view
      inline const typename Traits::GridViewType& getGridView () const
      {
        return pgfs_->gridview();
      }

    protected:
      typedef LocalFunctionSpace<GFS> LFS;
      typedef LFSIndexCache<LFS> LFSCache;
      typedef typename U::template ConstLocalView<LFSCache> UView;

      shared_ptr<const GFS> pgfs_;
      mutable LFS lfs_;
      mutable LFSCache lfsCache_;
      mutable UView u_view_;
      mutable std::vector<typename Traits::RangeFieldType> ul_;
      mutable std::vector<typename Traits::RangeType> yb_;
      const ST& st_;
      const int domain_index_;
      bool scaleToBBox_;
    };


    /** \brief Unfitted DiscreteGridFunction for vector-valued functions
     *
     * convert a power function space of scalar function spaces into a
     * vector-valued grid function this is just an intermediate
     * solution to provide VTK output
     *
     * \tparam T Type of PowerGridFunctionSpace
     * \tparam X Type of coefficients vector
     * \tparam ST Type of sub triangulation
     */
    template<typename T, typename X, typename ST>
    class UnfittedVectorDiscreteGridFunction
      : public GridFunctionInterface<
            GridFunctionTraits<
                typename T::Traits::GridViewType,
                typename T::template Child<0>::Type::Traits::FiniteElementType
                ::Traits::Basis::Traits::RangeField,
                T::CHILDREN,
                Dune::FieldVector<
                    typename T::template Child<0>::Type::Traits::FiniteElementType
                    ::Traits::Basis::Traits::RangeField,
                    T::CHILDREN
                    >
                >,
            UnfittedVectorDiscreteGridFunction<T,X,ST>
            >
    {
      typedef T GFS;

      typedef BasisInterfaceSwitch<
          typename FiniteElementInterfaceSwitch<
              typename T::template Child<0>::Type::Traits::FiniteElementType
              >::Basis
          > BasisSwitch;

      typedef GridFunctionInterface<
          GridFunctionTraits<
              typename T::Traits::GridViewType,
              typename BasisSwitch::RangeField,
              T::CHILDREN,
              Dune::FieldVector<
                  typename BasisSwitch::RangeField,
                  T::CHILDREN
                  >
              >,
          UnfittedVectorDiscreteGridFunction<T,X,ST>
          > BaseT;

    public:
      typedef typename BaseT::Traits Traits;
      typedef typename T::template Child<0>::Type ChildType;
      typedef typename ChildType::Traits::FiniteElementType
      ::Traits::Basis::Traits::RangeField RF;
      typedef typename ChildType::Traits::FiniteElementType
      ::Traits::Basis::Traits::Range RT;

      UnfittedVectorDiscreteGridFunction (const GFS& gfs, const X& x_,
                                          const ST & st_, const int domain_index_ = 0, bool scaleToBBox = true)
        : pgfs(stackobject_to_shared_ptr(gfs)), xg(x_), lfs(gfs), xl(gfs.maxLocalSize()), yb(gfs.maxLocalSize()),
          st(st_), domain_index(domain_index_), scaleToBBox_(scaleToBBox)
      {}

      inline void evaluate (const typename Traits::ElementType& e,
                            const typename Traits::DomainType& x,
                            typename Traits::RangeType& y) const
      {
        typedef FiniteElementInterfaceSwitch<
            typename ChildType::Traits::FiniteElementType
            > FESwitch;

        lfs.bind(*( st.BBox(domain_index,e).entity() ));
        lfs.vread(xg,xl);
        for (std::size_t k=0; k<T::CHILDREN; k++)
        {
          typename Traits::DomainType gx;
          if (scaleToBBox_)
            gx = st.BBox(domain_index,e).local(e.geometry().global(x));
          else
            gx = x;
          FESwitch::basis(lfs.child(k).finiteElement()).evaluateFunction(gx,yb);
          y[k] = 0.0;
          for (unsigned int i=0; i<yb.size(); i++)
            y[k] += xl[lfs.child(k).localIndex(i)]*yb[i];
        }
      }

      //! get a reference to the GridView
      inline const typename Traits::GridViewType& getGridView () const
      {
        return pgfs->gridview();
      }

    private:
      shared_ptr<GFS const> pgfs;
      const X& xg;
      mutable LocalFunctionSpace<GFS> lfs;
      mutable std::vector<RF> xl;
      mutable std::vector<RT> yb;
      const ST & st;
      const int domain_index;
      bool scaleToBBox_;
    };
  }
}

#endif
