// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_UDG_PDELAB_OPERATOR_HH
#define DUNE_UDG_PDELAB_OPERATOR_HH

#include <dune/pdelab/gridfunctionspace/interpolate.hh>
#include <dune/pdelab/gridoperator/common/gridoperatorutilities.hh>
#include <dune/pdelab/gridoperator/common/borderdofexchanger.hh>
#include <dune/udg/pdelab/assembler/assembler.hh>
#include <dune/udg/pdelab/assembler/localassembler.hh>
#include <dune/istl/io.hh>

namespace Dune {
  namespace UDG {

    /**
       \brief A standard grid operator implementation for grids with a
       UDG subtriangulation

       \tparam GFSU GridFunctionSpace for ansatz functions
       \tparam GFSV GridFunctionSpace for test functions
       \tparam MB The matrix backend to be used for representation of the jacobian
       \tparam DF The domain field type of the operator
       \tparam RF The range field type of the operator
       \tparam ST The type of the sub triangulation
       \tparam nonoverlapping_mode Switch for nonoverlapping grids
       \tparam CU   Constraints maps for the individual dofs (trial space)
       \tparam CV   Constraints maps for the individual dofs (test space)

     */
    template<typename GFSU, typename GFSV, typename LOP,
        typename MB, typename DF, typename RF, typename JF,
        typename ST, bool nonoverlapping_mode = false,
        typename CU=Dune::PDELab::EmptyTransformation,
        typename CV=Dune::PDELab::EmptyTransformation>
    class UDGGridOperator
    {
    public:

      //! The global UDG assembler type
      typedef UDGAssembler<GFSU,GFSV,nonoverlapping_mode,ST> Assembler;

      //! The type of the domain (solution).
      typedef typename Dune::PDELab::Backend::impl::BackendVectorSelector<GFSU,DF>::Type Domain;
      //! The type of the range (residual).
      typedef typename Dune::PDELab::Backend::impl::BackendVectorSelector<GFSV,RF>::Type Range;
      //! The type of the jacobian.
      typedef typename Dune::PDELab::Backend::impl::BackendMatrixSelector<MB,Domain,Range,JF>::Type Jacobian;

      //! The sparsity pattern container for the jacobian matrix
      typedef typename MB::template Pattern<Jacobian,GFSV,GFSU> Pattern;

      //! The local UDG assembler type
      typedef UDGLocalAssembler<UDGGridOperator,LOP,ST,nonoverlapping_mode>
      LocalAssembler;

      typedef typename std::conditional<
        nonoverlapping_mode,
        PDELab::NonOverlappingBorderDOFExchanger<UDGGridOperator>,
        PDELab::OverlappingBorderDOFExchanger<UDGGridOperator>
        >::type BorderDOFExchanger;

      //! The grid operator traits
      typedef Dune::PDELab::GridOperatorTraits
      <GFSU,GFSV,MB,DF,RF,JF,CU,CV,Assembler,LocalAssembler> Traits;

      template <typename MFT>
      struct MatrixContainer {
        typedef typename Traits::Jacobian Type;
      };

      //! Constructor for non trivial constraints
      UDGGridOperator(const GFSU & gfsu_, const GFSV & gfsv_, const ST & st_, LOP & lop_,
                      const CU & cu_, const CV & cv_, const MB& mb_ = MB())
        : global_assembler(gfsu_,gfsv_,st_), local_assembler(lop_, cu_, cv_), backend(mb_)
      {}

      //! Constructor for empty constraints
      UDGGridOperator(const GFSU & gfsu_, const GFSV & gfsv_, const ST & st_, LOP & lop_, const MB& mb_ = MB())
        : global_assembler(gfsu_,gfsv_,st_), local_assembler(lop_), backend(mb_)
      {}

      //! Get the trial grid function space
      const GFSU& trialGridFunctionSpace() const
      {
        return global_assembler.trialGridFunctionSpace();
      }

      //! Get the test grid function space
      const GFSV& testGridFunctionSpace() const
      {
        return global_assembler.testGridFunctionSpace();
      }

      //! Get dimension of space u
      typename GFSU::Traits::SizeType globalSizeU () const
      {
        return trialGridFunctionSpace().globalSize();
      }

      //! Get dimension of space v
      typename GFSV::Traits::SizeType globalSizeV () const
      {
        return testGridFunctionSpace().globalSize();
      }

      const Assembler & assembler() const { return global_assembler; }

      Assembler & assembler() { return global_assembler; }

      LocalAssembler & localAssembler() const { return local_assembler; }

      //! Visitor which is called in the method setupGridOperators for
      //! each tuple element.
      template <typename GridOperatorTuple>
      struct SetupGridOperator {
        SetupGridOperator()
          : index(0), size(std::tuple_size<GridOperatorTuple>::value) {}

        template <class T>
        void operator()(T& elem) {
          elem.localAssembler().preProcessing(index == 0);
          elem.localAssembler().postProcessing(index == size-1);
          ++index;
        }

        int index;
        const int size;
      };

      //! Method to set up a number of grid operators which are used
      //! in a joint assembling. It is assumed that all operators are
      //! specializations of the same template type
      template<typename GridOperatorTuple>
      static void setupGridOperators(GridOperatorTuple tuple)
      {
        SetupGridOperator<GridOperatorTuple> setup_visitor;
        Dune::Hybrid::forEach(tuple, setup_visitor);
      }

      //! Interpolate the constrained dofs from given function
      template<typename F, typename X>
      void interpolate (const X& xold, F& f, X& x) const
      {
        // Interpolate f into grid function space and set corresponding coefficients
        Dune::PDELab::interpolate(f,global_assembler.trialGridFunctionSpace(),x);

        // Copy non-constrained dofs from old time step
        Dune::PDELab::copy_nonconstrained_dofs(local_assembler.trialConstraints(),xold,x);
      }

      //! Fill pattern of jacobian matrix
      void fill_pattern(Pattern & p) const {
        typedef typename LocalAssembler::LocalPatternAssemblerEngine PatternEngine;
        PatternEngine & pattern_engine = local_assembler.localPatternAssemblerEngine(p);
        global_assembler.assemble(pattern_engine);
      }

      //! Assemble residual
      void residual(const Domain & x, Range & r) const {
        typedef typename LocalAssembler::LocalResidualAssemblerEngine ResidualEngine;
        ResidualEngine & residual_engine = local_assembler.localResidualAssemblerEngine(r,x);
        global_assembler.assemble(residual_engine);
      }

      //! Assembler jacobian
      void jacobian(const Domain & x, Jacobian & a) const {
        typedef typename LocalAssembler::LocalJacobianAssemblerEngine JacobianEngine;
        JacobianEngine & jacobian_engine = local_assembler.localJacobianAssemblerEngine(a,x);
        global_assembler.assemble(jacobian_engine);
      }

      //! Get the matrix backend for this grid operator.
      const typename Traits::MatrixBackend& matrixBackend() const
      {
        return backend;
      }

    private:
      Assembler global_assembler;
      mutable LocalAssembler local_assembler;
      MB backend;
    };

  };
};
#endif
