#ifndef DUNE_UDG_RESTRICTED_FINITE_ELEMENT_MAP_HH
#define DUNE_UDG_RESTRICTED_FINITE_ELEMENT_MAP_HH

#include <functional>

#include <dune/localfunctions/common/interfaceswitch.hh>

#include <dune/pdelab/finiteelement/qkdglagrange.hh>
#include <dune/pdelab/finiteelementmap/finiteelementmap.hh>

namespace Dune
{
  namespace UDG
  {
    // Restricted local basis which can have a zero size on some elements
    template <typename Basis>
    class RestrictedBasis
    {
    public:
      using Traits = typename Basis::Traits;

      explicit RestrictedBasis(const Basis& basis_, bool zero = false)
          : basis(basis_), setZero(zero)
      {
      }

      unsigned int size() const
      {
        return setZero ? 0 : basis.size();
      }

      inline void evaluateFunction(const typename Basis::Traits::DomainType& in,
                                   std::vector<typename Basis::Traits::RangeType>& out) const
      {
        if (setZero)
          out.clear();
        else
          basis.evaluateFunction(in, out);
      }

      void evaluateJacobian(const typename Basis::Traits::DomainType& in,
                            std::vector<typename Basis::Traits::JacobianType>& out) const
      {
        if (setZero)
          out.clear();
        else
          basis.evaluateJacobian(in, out);
      }

      unsigned int order() const
      {
        return setZero ? 0 : basis.order();
      }

    private:
      const Basis& basis;
      bool setZero;
    };

    // wrapped QkDG local coefficients which can have a zero size on some elements
    template <typename Coefficients>
    class RestrictedCoefficients
    {
    public:
      explicit RestrictedCoefficients(const Coefficients& coefficients, bool zero = false)
          : coeffs(coefficients), setZero(zero)
      {
      }

      std::size_t size() const
      {
        return setZero ? 0 : coeffs.size();
      }

      const Dune::LocalKey& localKey(std::size_t i) const
      {
        if (setZero) {
          DUNE_THROW(Dune::Exception, "set zero");
        } else {
          return coeffs.localKey(i);
        }
      }

    private:
      Coefficients coeffs;
      bool setZero;
    };

    // wrapped Qk local interpolation which can have a zero size on some elements
    template <typename Interpolation>
    class RestrictedInterpolation
    {
    public:
      explicit RestrictedInterpolation(const Interpolation& interpolation_, bool zero = false)
          : interpolation(interpolation_), setZero(zero)
      {
      }

      template <typename F, typename C>
      void interpolate(const F& f, std::vector<C>& out) const
      {
        if (setZero)
          out.clear();
        else
          interpolation.interpolate(f, out);
      }

    private:
      Interpolation interpolation;
      bool setZero;
    };

    // wrapped Qk local finite element which can have a zero size on some elements
    template <typename FE>
    class RestrictedFiniteElement
    {
      using FESwitch = Dune::FiniteElementInterfaceSwitch<FE>;
      using Basis = RestrictedBasis<typename FESwitch::Basis>;
      using Coefficients = RestrictedCoefficients<typename FESwitch::Coefficients>;
      using Interpolation = RestrictedInterpolation<typename FESwitch::Interpolation>;

    public:
      typedef Dune::LocalFiniteElementTraits<Basis, Coefficients, Interpolation> Traits;

      explicit RestrictedFiniteElement(const FE& fe, bool zero = false)
          : fe_(fe)
          , basis(fe_.localBasis(), zero)
          , coefficients(fe_.localCoefficients(), zero)
          , interpolation(fe.localInterpolation(), zero)
      {
      }

      const typename Traits::LocalBasisType& localBasis() const
      {
        return basis;
      }

      const typename Traits::LocalCoefficientsType& localCoefficients() const
      {
        return coefficients;
      }

      const typename Traits::LocalInterpolationType& localInterpolation() const
      {
        return interpolation;
      }

      Dune::GeometryType type() const
      {
        return fe_.type();
      }

      RestrictedFiniteElement* clone() const
      {
        return new RestrictedFiniteElement(*this);
      }

    private:
      FE fe_;
      Basis basis;
      Coefficients coefficients;
      Interpolation interpolation;
    };

    // restrict the a given finite element to some elements based on a user provided predicate
    template <typename GV, typename FE>
    class RestrictedFiniteElementMap
        : public Dune::PDELab::
              LocalFiniteElementMapInterface<Dune::PDELab::
                                                 LocalFiniteElementMapTraits<RestrictedFiniteElement<FE>>,
                                             RestrictedFiniteElementMap<GV, FE>>
    {
    public:
      using Traits = Dune::PDELab::LocalFiniteElementMapTraits<RestrictedFiniteElement<FE>>;
      using Element = typename GV::template Codim<0>::Entity;

      template <class F>
      RestrictedFiniteElementMap(const GV& gv_, const FE& fe, F&& predicate)
          : fe_nonzero(fe, false), fe_zero(fe, true), gv(gv_), predicate_(predicate)
      {
      }

      bool fixedSize() const
      {
        return false;
      }

      bool hasDOFs(int codim) const
      {
        return codim == 0;
      }

      std::size_t size(Dune::GeometryType gt) const
      {
        DUNE_THROW(Dune::PDELab::VariableElementSize, "the fem has variable element size");
      }

      std::size_t maxLocalSize() const
      {
        // assuming constant size on nonzero space
        return fe_nonzero.localBasis().size();
      }

      template <class EntityType>
      const typename Traits::FiniteElementType& find(const EntityType& e) const
      {
        return predicate_(e) ? fe_zero : fe_nonzero;
      }

    private:
      mutable RestrictedFiniteElement<FE> fe_nonzero;
      mutable RestrictedFiniteElement<FE> fe_zero;
      GV gv;
      std::function<bool(Element)> predicate_;
    };
  }
}

#endif // DUNE_UDG_RESTRICTED_FINITE_ELEMENT_MAP_HH
