// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef UDG_UTILITIES_HH
#define UDG_UTILITIES_HH

#include <dune/typetree/typetree.hh>
#include <dune/udg/vtriangulation.hh>

namespace Dune {
  namespace PDELab {

    namespace UnfittedImp {

      /**
       * \defgroup UdgOperatorRebindVisitor Meta program for
       *                                    unfitted local funtion space
       *
       * This tree traversal calls rebind on every leaf node of the local
       * functions tree.
       *
       * @{
       */
      enum RebindTypes { BindEntityPart, BindInsideIntersectionPart, BindOutsideIntersectionPart,
                         BindIntersectionPart};

      template<RebindTypes type, typename UnfittedGeometry>
      struct RebindVisitor : public TypeTree::TreeVisitor, public TypeTree::DynamicTraversal
      {
        const UnfittedGeometry & ug;

        RebindVisitor(const UnfittedGeometry & ug_) : ug(ug_) {}

        template<typename LeafNode, typename TreePath>
        void leaf(const LeafNode& node, TreePath treePath) const
        {
          if(type == BindInsideIntersectionPart)
            node.finiteElement().basis().rebindToIntersectionInside(ug);
          else if(type == BindOutsideIntersectionPart)
            node.finiteElement().basis().rebindToIntersectionOutside(ug);
          else if(type == BindIntersectionPart)
            node.finiteElement().basis().rebindToIntersection(ug);
          else{DUNE_THROW(Dune::NotImplemented,"RebindType not implemented");}
        }
      };

      template<typename UnfittedGeometry>
      struct RebindVisitor<BindEntityPart,UnfittedGeometry>
        : public TypeTree::TreeVisitor, public TypeTree::DynamicTraversal
      {
        const UnfittedGeometry & ug;

        RebindVisitor(const UnfittedGeometry & ug_) : ug(ug_) {}

        template<typename LeafNode, typename TreePath>
        void leaf(const LeafNode& node, TreePath treePath) const
        {
          node.finiteElement().basis().rebindToEntityPart(ug);
          node.finiteElement().interpolation().rebind(ug);
        }
      };

    } // UnfittedImp


    /**
     * \brief Wrapper for SubTriangulationGeometry usage in PDELab.
     *        Objects of this class are used as ElementGeometries
     *        for entity parts.
     */
    template<typename SE>
    class UnfittedGeometryWrapper
    {
    public:
      //! Type of geometry maps
      typedef typename SE::Traits::Geometry Geometry;

      //! Wrapped entity part types
      typedef SE EntityPart;
      typedef Dune::shared_ptr<SE> EntityPartPointer;

      //! Type of bounding box of the wrapped entity part
      typedef typename EntityPart::Traits::BoundingBox BoundingBox;

      //! Entity types related to the wrapped entity part
      typedef typename BoundingBox::Entity Entity;

      //! \brief Construct from an entity part and its fundamental mesh
      //!        home entity, i.e. the element of the fundamental mesh
      //!        grid view for which the entity part was constructed.
      UnfittedGeometryWrapper (const Entity & e_, const EntityPartPointer se_)
        : e(e_), se(se_) {}

      //! \brief Return domain index of the wrapped entity part.
      int domainIndex() const
      { return se->domainIndex(); }

      //! \brief Return geometrical information about the wrapped
      //!        entity part in global coordinates.
      //!
      //! This method returns a Geometry object that provides a mapping
      //! from local coordinates of the wrapped entity part to global
      //! (world) coordinates.
      const Geometry& geometry () const
      {
        return se->geometry();
      }

      //! \brief Return geometrical information about the wrapped
      //!        entity part in local coordinates of its associated
      //!        bounding box.
      //!
      //! This method returns a Geometry object that provides a mapping
      //! from local coordinates of the wrapped entity part to local
      //! coordinates of its associated bounding box.
      const Geometry& localGeometry () const
      {
        return se->geometryInCell();
      }

      //! \brief Return geometrical information about the wrapped
      //!        entity part in local coordinates of its associated
      //!        bounding box.
      //!
      //! This method returns a Geometry object that provides a mapping
      //! from local coordinates of the wrapped entity part to local
      //! coordinates of its associated bounding box.
      const Dune::shared_ptr<Geometry> localGeometryPointer () const
      {
        return se->geometryInCellPointer();
      }

      //! \brief Return the fundamental mesh home entity, i.e. the element
      //!        of the fundamental mesh grid view for which the wrapped
      //!        entity part was constructed.
      const Entity& entity () const
      {
        return e;
      }

      //! \brief Return host entity of the wrapped entity part.
      //!
      //! Returns the entity to which the DOFs of this geometry (i.e. the
      //! DOFs associated with the wrapped entity part) have been attached.
      const Entity & hostEntity () const
      {
        return *(se->boundingBox().entity());
      }

      //! \brief Return the entity part object wrapped by an object of this
      //!        class.
      const EntityPart& subEntity () const
      {
        return *se;
      }

      //! \brief Return the bounding box of the wrapped entity part.
      const BoundingBox & boundingBox() const
      {
        return se->boundingBox();
      }

    private:
      const Entity & e;
      const EntityPartPointer se;
    };


    /**
     * \brief Wrapper for SubTriangulationGeometry usage in PDELab.
     *        Objects of this class are used as IntersectionGeometries
     *        for intersection parts.
     */
    template<typename SI>
    class UnfittedIntersectionWrapper
    {
    public:
      typedef typename SI::Traits::FaceGeometry Geometry;
      typedef typename SI::Traits::FaceGeometry LocalGeometry;
      typedef typename SI::Traits::BoundingBox BoundingBox;
      typedef typename SI::IntersectionPartPointer IntersectionPartPointer;
      typedef typename BoundingBox::Entity Entity;
#if DUNE_VERSION_NEWER(DUNE_GRID,2,4)
      typedef typename Entity::Geometry::ctype ctype;
#else
      typedef typename Entity::ctype ctype;
#endif

      //! Export grid dimension and dimension of world
      enum { dimension=Entity::dimension };
      enum { dimensionworld=Entity::Geometry::coorddimension };
      enum { mydimension=Entity::dimension };
      enum { coorddimension=Entity::Geometry::coorddimension };

      //! \brief Construct from an intersection part, the fundamental mesh home
      //!        entity on its inside, i.e. the element of the fundamental mesh
      //!        grid view for which the intersection part was constructed, and
      //!        the intersection part's index with respect to the home entity.
      UnfittedIntersectionWrapper
        (const IntersectionPartPointer i_,
        const Entity & e_,
        unsigned int index_)
        : iPart(i_), e(e_), index(index_)
      {}

      //! \brief Return the inside bounding box of wrapped intersection part.
      const BoundingBox & insideBBox() const { return iPart->boundingBoxInside(); }

      //! \brief Return the outside bounding box of wrapped intersection part.
      const BoundingBox & outsideBBox() const { return iPart->boundingBoxOutside(); }

      //! \brief Return domain index on the inside of wrapped intersection part.
      int insideDomainIndex() const { return iPart->insideDomainIndex(); }

      //! \brief Return domain index on the outside of wrapped intersection part.
      int outsideDomainIndex() const { return iPart->outsideDomainIndex(); }

      //! \brief Return true if wrapped intersection part is an intersection
      //!        with an interior or exterior boundary, i.e. if it is on a domain
      //!        boundary.
      //!
      //! \todo Isn't the implementation's first condition in general wrong
      //!       for an interior boundary on the fundamental mesh skeleton?
      //!       Why don't we just use iPart->boundary() here?
      bool boundary () const
      {
        return (iPart->boundingBoxInside().entity() == iPart->boundingBoxOutside().entity()) &&
               ( iPart->insideDomainIndex() != iPart->outsideDomainIndex() );
      }

      //! \brief Identifier for boundary segment from macro grid.
      //!
      //! One can attach a boundary Id to a boundary segment on the macro
      //! grid. This Id will also be used for all fragments of these
      //! boundary segments.
      //!
      //! The numbering is defined as:
      //!   - Id==0 for all intersections without boundary()==false
      //!   - Id>=0 for all intersections without boundary()==true
      //!
      //! The way the Identifiers are attached to the grid may differ
      //! between the different grid implementations.
      int boundaryId () const
      {
        return iPart->boundaryId();
      }

      //! \brief Return true if wrapped intersection part is shared with
      //!        a domain cell patch on the outside.
      //!
      //! \todo Why don't we just use iPart->neighbor() here?
      bool neighbor () const
      {
        return iPart->outsideDomainIndex() != -1;
      }

      //! \brief Return geometrical information about the wrapped
      //!        intersection part in local coordinates of the
      //!        inside bounding box.
      //!
      //! This method returns a Geometry object that provides a mapping
      //! from local coordinates of the wrapped intersection part to
      //! local coordinates of the inside bounding box.
      //!
      //! Note that the inside bounding box is NOT the inside() entity.
      const Geometry& geometryInInside () const
      {
        return iPart->geometryInInside();
      }

      //! \brief Return geometrical information about the wrapped
      //!        intersection part in local coordinates of the
      //!        outside bounding box.
      //!
      //! This method returns a Geometry object that provides a mapping
      //! from local coordinates of the wrapped intersection part to
      //! local coordinates of the outside bounding box.
      //!
      //! Note that the outside bounding box is NOT the outside() entity.
      const Geometry& geometryInOutside () const
      {
        return iPart->geometryInOutside();
      }

      //! \brief Return geometrical information about the wrapped
      //!        intersection part in global coordinates.
      //!
      //! This method returns a Geometry object that provides a mapping
      //! from local coordinates of the wrapped intersection part to
      //! global (world) coordinates.
      const Geometry& geometry () const
      {
        return iPart->geometry();
      }

      //! \brief Local number of the codim 1 subentity of the inside() entity
      //!        in which the wrapped intersection part is contained.
      int indexInInside () const
      {
        DUNE_THROW(Dune::NotImplemented,"The local index of the "
                   "intersection part is not known in UDG applications.");
        return 0;
      }

      //! \brief Local number of the codim 1 subentity of the outside() entity
      //!        in which the wrapped intersection part is contained.
      int indexInOutside () const
      {
        DUNE_THROW(Dune::NotImplemented,"The local index of the "
                   "intersection part is not known in UDG applications.");
        return 0;
      }

      //! \brief Return an outer normal (length not necessarily 1).
      //!
      //! The returned vector may depend on the local position within the
      //! wrapped intersection part.
      Dune::FieldVector<ctype, dimensionworld> outerNormal
        (const Dune::FieldVector<ctype, dimension-1>& local) const
      {
        return iPart->normal(local);
      }

      //! \brief Return outer normal scaled with the integration element.
      //!
      //! @copydoc outerNormal
      //! The normal is scaled with the integration element of the wrapped
      //! intersection part. This method is redundant but it may be more
      //! efficent to use this function rather than computing the
      //! integration element via intersectionGlobal().
      Dune::FieldVector<ctype, dimensionworld> integrationOuterNormal
        (const Dune::FieldVector<ctype, dimension-1>& local) const
      {
        return iPart->normal(local) * iPart->geometry().integrationElement(local);
      }

      //! \brief Return unit outer normal (length == 1).
      //!
      //! The returned vector may depend on the local position within the
      //! wrapped intersection part. It is scaled to have unit length.
      Dune::FieldVector<ctype, dimensionworld> unitOuterNormal
        (const Dune::FieldVector<ctype, dimension-1>& local) const
      {
        return iPart->unitOuterNormal(local);
      }

      //! \brief Return unit outer normal (length == 1).
      //!
      //! The returned vector is scaled to have unit length.
      //!
      Dune::FieldVector<ctype, dimensionworld> centerUnitOuterNormal () const
      {
        return iPart->unitOuterNormal(Dune::FieldVector<ctype, dimension-1>(0.5));
      }

      //! \brief Return EntityPointer to the host entity on the inside of
      //!        the wrapped intersection part.
      //!
      //! Returns a pointer to the entity to which the DOFs of this geometry
      //! have been attached on the inside of the wrapped intersection part.
      //!
      //! The returned entity is equal to the return type of
      //! insideBBox().entity(). Hence, the wrapped intersection part does
      //! not need to be embedded in any intersection of the returned entity
      //! in general. They need not even share a single vertex.
      //!
      //! \todo Actually this method shall return an EntityPointer to the
      //!       fundamental mesh home entity on the inside of the wrapped
      //!       intersection part in order to allow evaluation of data
      //!       functions on wrapped intersection parts the same way it
      //!       is done for wrapped intersections in DUNE-PDELab.
      //!       Additionally this would make the method consistent with
      //!       the entity() method of UnfittedGeometryWrapper.
      Entity inside () const
      {
        // We could just use iPart->inside() but the following implementation
        // reflects the current functionality of the method
        return iPart->boundingBoxInside().entity();
      }

      //! \brief Return EntityPointer to the fundamental mesh home entity
      //!        on the inside of the wrapped intersection part. This is
      //!        the element of the fundamental mesh grid view for which
      //!        the wrapped intersection part was constructed.
      //!
      //! The return value may in general differ from the value of
      //! insideBBox().entity().
      //!
      //! \note The following addition makes no sense to me (Sebastian W.):
      //!       "The returned entity is guaranteed to have an intersection
      //!       in which this intersection part may be embedded."
      //!       Shouldn't it be replaced by something like
      //!       "The closure of the returned entity is guaranteed to contain
      //!       the wrapped intersection part geometrically."?
      Entity insideHomeEntity () const
      {
        return e;
      }

      //! \brief Return EntityPointer to the host entity on the inside of
      //!        the wrapped intersection part.
      //!
      //! Equivalent to calling inside(). Compare to insideHomeEntity().
      Entity insideHostEntity () const
      {
        return inside();
      }

      //! \brief Return EntityPointer to the host entity on the outside of
      //!        the wrapped intersection part.
      //!
      //! Returns a pointer to the entity to which the DOFs of this geometry
      //! have been attached on the outside of the wrapped intersection part.
      //!
      //! The returned entity is equal to the return type of
      //! outsideBBox().entity(). Hence, the wrapped intersection part does
      //! not need to be embedded in any intersection of the returned entity
      //! in general. They need not even share a single vertex.
      //!
      //! \todo Actually this method shall return an EntityPointer to the
      //!       fundamental mesh home entity on the outside of the wrapped
      //!       intersection part in order to allow evaluation of data
      //!       functions on wrapped intersection parts the same way it
      //!       is done for wrapped intersections in DUNE-PDELab.
      //!       Additionally this would make the method consistent with
      //!       the entity() method of UnfittedGeometryWrapper.
      Entity outside () const
      {
        // We could just use iPart->outside() but the following implementation
        // reflects the current functionality of the method
        return iPart->boundingBoxOutside().entity();
      }

      //! \brief Return the intersection part object wrapped by an object of
      //!        this class.
      const SI& intersection () const
      {
        return *iPart;
      }

      //! \brief Return local index of the wrapped intersection part with
      //!        respect to insideHomeEntity().
      unsigned int intersectionIndex () const
      {
        return index;
      }

    private:
      const IntersectionPartPointer iPart;
      const Entity & e;
      const unsigned int index;
    };


    /**
     * \brief Wrapper for SubTriangulation usage in PDELab.
     */
    template<typename GV>
    class UnfittedSubTriangulation
    {
    public:
      static const bool hasSubTriangulation = true;

      typedef VirtualSubTriangulation<GV> SubTriangulationImp;

      typedef typename GV::Traits::template Codim<0>::Entity Entity;
      typedef typename GV::IntersectionIterator IntersectionIterator;
      typedef typename IntersectionIterator::Intersection Intersection;
      typedef typename SubTriangulationImp::BoundingBox BoundingBox;

      typedef typename SubTriangulationImp::EntityPart EntityPartImp;
      typedef typename Dune::UDG::EntityPartInterface<GV>::EntityPartPointer EntityPartImpPointer;
      typedef std::list<EntityPartImpPointer> EntityPartImpList;
      typedef typename SubTriangulationImp::IntersectionPart IntersectionPartImp;
      typedef typename Dune::UDG::IntersectionPartInterface<GV> IntersectionPartInterface;
      typedef typename IntersectionPartInterface::IntersectionPartPointer IntersectionPartPointer;
      typedef std::list<IntersectionPartPointer> IntersectionPartImpList;

      typedef UnfittedGeometryWrapper<EntityPartImp> EntityPart;
      typedef std::list<EntityPart> EntityPartList;
      typedef typename EntityPartList::iterator EntityPartIterator;
      typedef typename EntityPartList::iterator SubEntityIterator; //!< \deprecated

      typedef UnfittedIntersectionWrapper<IntersectionPartImp> IntersectionPart;
      typedef std::list<IntersectionPart> IntersectionPartList;
      typedef typename IntersectionPartList::iterator IntersectionPartIterator;
      typedef typename IntersectionPartList::iterator SubIntersectionIterator; //!< \deprecated
      typedef typename IntersectionPartList::const_iterator IntersectionPartConstIterator;

      typedef UnfittedImp::RebindVisitor
      <UnfittedImp::BindEntityPart,EntityPart>
      BindEntityPartVisitor;

      typedef UnfittedImp::RebindVisitor
      <UnfittedImp::BindInsideIntersectionPart,IntersectionPart>
      BindInsideIntersectionPartVisitor;

      typedef UnfittedImp::RebindVisitor
      <UnfittedImp::BindOutsideIntersectionPart,IntersectionPart>
      BindOutsideIntersectionPartVisitor;

      typedef UnfittedImp::RebindVisitor
      <UnfittedImp::BindIntersectionPart,IntersectionPart>
      BindIntersectionPartVisitor;

      //! \brief Construct from a fundamental mesh grid view and an associated
      //!        VirtualSubTriangulation.
      UnfittedSubTriangulation(const GV & gv_, const SubTriangulationImp & sti_)
        : gv(gv_), subTriangulation_imp(sti_)
      {}

      UnfittedSubTriangulation(const UnfittedSubTriangulation & c_)
        : gv(c_.gv), subTriangulation_imp(c_.subTriangulation_imp),
          entityParts(c_.entityParts), intersectionParts(c_.intersectionParts),
          entityParts_imp(c_.entityParts_imp),
          intersectionParts_imp(c_.intersectionParts_imp)
      {}

      //! \brief Create local sub triangulation of a fundamental mesh cell as a
      //!        set of entity parts and intersection parts that are wrapped as
      //!        ElementGeometries and IntersectionGeometries.
      void create(const Entity & e) const
      {
        entityParts_imp.clear();
        intersectionParts_imp.clear();

        subTriangulation_imp.create_entity_parts(e,entityParts_imp);
        subTriangulation_imp.create_intersection_parts(e,intersectionParts_imp);

        entityParts.clear();
        intersectionParts.clear();

        typename EntityPartImpList::iterator epit = entityParts_imp.begin();
        typename EntityPartImpList::iterator eepit = entityParts_imp.end();
        for(; epit != eepit; ++epit)
          entityParts.push_back(EntityPart(e,*epit));

        int intersectionPart_index(0);
        typename IntersectionPartImpList::iterator ipit = intersectionParts_imp.begin();
        typename IntersectionPartImpList::iterator eipit = intersectionParts_imp.end();
        for(; ipit != eipit; ++ipit) {
          intersectionParts.push_back(IntersectionPart(*ipit,e,intersectionPart_index));
          intersectionPart_index++;
        }
      }

      const SubTriangulationImp & imp() const
      {
        return subTriangulation_imp;
      }

      EntityPartIterator begin() const
      {
        return entityParts.begin();
      }

      EntityPartIterator end() const
      {
        return entityParts.end();
      }

      IntersectionPartIterator ibegin() const
      {
        return intersectionParts.begin();
      }

      IntersectionPartIterator iend() const
      {
        return intersectionParts.end();
      }

    private:
      const GV & gv;
      const SubTriangulationImp & subTriangulation_imp;
      mutable EntityPartList entityParts;
      mutable IntersectionPartList intersectionParts;
      mutable EntityPartImpList entityParts_imp;
      mutable IntersectionPartImpList intersectionParts_imp;
    };

  }
}

#endif
