// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef TWO_PHASE_OPERATOR_HH
#define TWO_PHASE_OPERATOR_HH

#include <dune/common/version.hh>
#include <dune/common/typetraits.hh>
#include <dune/common/exceptions.hh>

#include <dune/pdelab/localoperator/callswitch.hh>
#include <dune/pdelab/localoperator/defaultimp.hh>
#include <dune/pdelab/localoperator/pattern.hh>
#include <dune/pdelab/localoperator/flags.hh>
#include <dune/pdelab/localoperator/idefault.hh>
#include <dune/pdelab/gridoperator/common/assemblerutilities.hh>

#include "couplingutilities.hh"

/**
   \brief Helper class to provide alpha_coupling via the old interface
 */
template<typename LOP0, typename LOP1>
class SimpleSkeletonCouplingOperator
  : public Dune::PDELab::MultiDomain::CouplingOperatorDefaultFlags
    , public Dune::PDELab::MultiDomain::FullCouplingPattern
    , public Dune::PDELab::MultiDomain::NumericalJacobianApplyCoupling<SimpleSkeletonCouplingOperator<LOP0,LOP1> >
{
  LOP0 & lop0;

public:

  static const bool doSkeletonTwoSided = LOP0::doSkeletonTwoSided || LOP1::doSkeletonTwoSided;
  static const bool doAlphaCoupling = LOP0::doAlphaSkeleton || LOP1::doAlphaSkeleton;
  static const bool doLambdaCoupling = LOP0::doLambdaSkeleton || LOP1::doLambdaSkeleton;

  SimpleSkeletonCouplingOperator(LOP0 & lop_) : lop0(lop_) {}

  template<typename IG,
      typename LFSU0,typename LFSV0,
      typename LFSU1, typename LFSV1,
      typename X, typename R>
  void alpha_coupling(const IG& ig,
                      const LFSU0& lfsu0, const X& x0, const LFSV0& lfsv0,
                      const LFSU1& lfsu1, const X& x1, const LFSV1& lfsv1,
                      R& r0, R& r1) const
  {
#if DUNE_VERSION_NEWER(DUNE_COMMON,2,4)
    static_assert((std::is_same<LOP0, LOP1>::value), "Old-school operators work only if LOP0 == LOP1");
#else
    dune_static_assert((std::is_same<LOP0, LOP1>::value), "Old-school operators work only if LOP0 == LOP1");
#endif

#ifndef NDEBUG
    const int in_phase  = ig.intersection().insideDomainIndex();
    const int out_phase = ig.intersection().outsideDomainIndex();

    assert(in_phase < out_phase);
#endif

    Dune::PDELab::LocalAssemblerCallSwitch<LOP0,LOP0::doAlphaSkeleton CALLSWITCHLINEARFLAG>::
    alpha_skeleton(lop0, ig, lfsu0, x0, lfsv0, lfsu1, x1, lfsv1, r0, r1);
  }

  template<typename IG, typename LFSU, typename X, typename LFSV, typename M>
  void jacobian_coupling (const IG& ig,
                        const LFSU& lfsu_s, const X& x_s, const LFSV& lfsv_s,
                        const LFSU& lfsu_n, const X& x_n, const LFSV& lfsv_n,
                        M& mat_ss, M& mat_sn,
                        M& mat_ns, M& mat_nn) const
  {
    Dune::PDELab::LocalAssemblerCallSwitch<LOP0,LOP0::doAlphaSkeleton CALLSWITCHLINEARFLAG>::
    jacobian_skeleton(lop0, ig, lfsu_s,x_s,lfsv_s,lfsu_n,x_n,lfsv_n,mat_ss,mat_sn,mat_ns,mat_nn);

  }

  template<typename T>
  void setTime(T) {};
};

/**
   \brief Helper class to provide alpha_coupling which does nothing
 */
class NoCouplingOperator
  : public Dune::PDELab::MultiDomain::CouplingOperatorDefaultFlags
    , public Dune::PDELab::MultiDomain::FullCouplingPattern
    , public Dune::PDELab::MultiDomain::NumericalJacobianCoupling<NoCouplingOperator >
    , public Dune::PDELab::MultiDomain::NumericalJacobianApplyCoupling<NoCouplingOperator >
{
public:
  template<typename T>
  void setTime(T) {};
};

/**
   \brief A wrapping local operator which includes two nested local
   operators for a two-phase setup, where each local operator
   corresponds to one phase.

   Notice, that this wrapper can not be used directly as it does \b
   not derive from any of the pattern classes.

   \tparam The type of the nested local operators.
 */
template<typename LOP0, typename LOP1,
    typename COP,
    class TR = double>
class TwoPhaseLocalOperatorWrapper :
  public Dune::PDELab::LocalOperatorDefaultFlags,
  public Dune::PDELab::InstationaryLocalOperatorDefaultMethods<TR>
{
private:
  LOP0 & lop0;
  LOP1 & lop1;
  COP & cop;
public:
  // pattern assembly flags
  enum { doPatternVolume = (int)LOP0::doPatternVolume || (int)LOP1::doPatternVolume };
  enum { doPatternSkeleton = (int)LOP0::doPatternSkeleton || (int)LOP1::doPatternSkeleton };
  enum { doPatternBoundary = (int)LOP0::doPatternBoundary || (int)LOP1::doPatternBoundary };
  enum { doPatternVolumePostSkeleton = (int)LOP0::doPatternVolumePostSkeleton || (int)LOP1::doPatternVolumePostSkeleton };

  // residual assembly flags
  enum { doAlphaVolume = (int)LOP0::doAlphaVolume || (int)LOP1::doAlphaVolume };
  enum { doAlphaVolumePostSkeleton = (int)LOP0::doAlphaVolumePostSkeleton || (int)LOP1::doAlphaVolumePostSkeleton };
  enum { doLambdaVolume = (int)LOP0::doLambdaVolume || (int)LOP1::doLambdaVolume };
  enum { doLambdaVolumePostSkeleton = (int)LOP0::doLambdaVolumePostSkeleton || (int)LOP1::doLambdaVolumePostSkeleton };
  enum { doAlphaSkeleton = (int)LOP0::doAlphaSkeleton || (int)LOP1::doAlphaSkeleton };
  enum { doLambdaSkeleton = (int)LOP0::doLambdaSkeleton || (int)LOP1::doLambdaSkeleton };
  enum { doAlphaBoundary = (int)LOP0::doAlphaBoundary || (int)LOP1::doAlphaBoundary };
  enum { doLambdaBoundary = (int)LOP0::doLambdaBoundary || (int)LOP1::doLambdaBoundary };

  // select twoSided option for real skeleton terms
  enum { doSkeletonTwoSided = (int)LOP0::doSkeletonTwoSided || (int)LOP1::doSkeletonTwoSided };
  // NOTE:  coupling faces will be called two sided anyway
#if DUNE_VERSION_NEWER(DUNE_COMMON,2,4)
  static_assert((int)LOP0::doSkeletonTwoSided == (int)LOP1::doSkeletonTwoSided, "All LOPs must use TwoSided consistently");
#else
  dune_static_assert((int)LOP0::doSkeletonTwoSided == (int)LOP1::doSkeletonTwoSided, "All LOPs must use TwoSided consistently");
#endif

  TwoPhaseLocalOperatorWrapper(LOP0 & lop0_, LOP1 & lop1_, COP & cop_) :
    lop0(lop0_),
    lop1(lop1_),
    cop(cop_)
  {}

  void setTime(TR t_)
  {
    lop0.setTime(t_);
    lop1.setTime(t_);
    cop.setTime(t_);
  }

  // volume integral depending on test and ansatz functions
  template<typename EG, typename LFSU, typename X, typename LFSV, typename R>
  void alpha_volume (const EG& eg, const LFSU& lfsu, const X& x, const LFSV& lfsv, R& r) const
  {
    if(eg.subEntity().domainIndex() == 0) {
      typedef typename LFSU::template Child<0>::Type LFSU0;
      typedef typename LFSV::template Child<0>::Type LFSV0;
      const LFSU0& lfsu0 = lfsu.template child<0>();
      const LFSV0& lfsv0 = lfsv.template child<0>();
      Dune::PDELab::LocalAssemblerCallSwitch<LOP0,LOP0::doAlphaVolume CALLSWITCHLINEARFLAG>::
      alpha_volume(lop0,eg,lfsu0,x,lfsv0,r);
    }
    else if(eg.subEntity().domainIndex() == 1) {
      typedef typename LFSU::template Child<1>::Type LFSU1;
      typedef typename LFSV::template Child<1>::Type LFSV1;
      const LFSU1& lfsu1 = lfsu.template child<1>();
      const LFSV1& lfsv1 = lfsv.template child<1>();
      Dune::PDELab::LocalAssemblerCallSwitch<LOP1,LOP1::doAlphaVolume CALLSWITCHLINEARFLAG>::
      alpha_volume(lop1,eg,lfsu1,x,lfsv1,r);
    }
  }

  template<typename EG, typename LFSV, typename R>
  void lambda_volume (const EG& eg, const LFSV& lfsv, R& r) const
  {
    if(eg.subEntity().domainIndex() == 0) {
      typedef typename LFSV::template Child<0>::Type LFSV0;
      const LFSV0& lfsv0 = lfsv.template child<0>();
      Dune::PDELab::LocalAssemblerCallSwitch<LOP0,LOP0::doLambdaVolume CALLSWITCHLINEARFLAG>::
      lambda_volume(lop0,eg,lfsv0,r);
    }
    else if(eg.subEntity().domainIndex() == 1) {
      typedef typename LFSV::template Child<1>::Type LFSV1;
      const LFSV1& lfsv1 = lfsv.template child<1>();
      Dune::PDELab::LocalAssemblerCallSwitch<LOP1,LOP1::doLambdaVolume CALLSWITCHLINEARFLAG>::
      lambda_volume(lop1,eg,lfsv1,r);
    }
  }

  template<typename EG, typename LFSU, typename X, typename LFSV, typename M>
  void jacobian_volume (const EG& eg, const LFSU& lfsu, const X& x, const LFSV& lfsv,
                        M & mat) const
  {
    if(eg.subEntity().domainIndex() == 0) {
      typedef typename LFSU::template Child<0>::Type LFSU0;
      typedef typename LFSV::template Child<0>::Type LFSV0;
      const LFSU0& lfsu0 = lfsu.template child<0>();
      const LFSV0& lfsv0 = lfsv.template child<0>();
      Dune::PDELab::LocalAssemblerCallSwitch<LOP0,LOP0::doAlphaVolume CALLSWITCHLINEARFLAG>::
      jacobian_volume(lop0,eg,lfsu0,x,lfsv0,mat);
    }
    else if(eg.subEntity().domainIndex() == 1) {
      typedef typename LFSU::template Child<1>::Type LFSU1;
      typedef typename LFSV::template Child<1>::Type LFSV1;
      const LFSU1& lfsu1 = lfsu.template child<1>();
      const LFSV1& lfsv1 = lfsv.template child<1>();
      Dune::PDELab::LocalAssemblerCallSwitch<LOP1,LOP1::doAlphaVolume CALLSWITCHLINEARFLAG>::
      jacobian_volume(lop1,eg,lfsu1,x,lfsv1,mat);
    }
  }

  template<typename IG, typename LFSU, typename X, typename LFSV, typename R>
  void alpha_skeleton (const IG& ig,
                       const LFSU& lfsu_s, const X& x_s, const LFSV& lfsv_s,
                       const LFSU& lfsu_n, const X& x_n, const LFSV& lfsv_n,
                       R& r_s, R& r_n) const
  {
    const int in_phase  = ig.intersection().insideDomainIndex();
    const int out_phase = ig.intersection().outsideDomainIndex();

    typedef typename LFSU::template Child<0>::Type LFSU0;
    typedef typename LFSU::template Child<1>::Type LFSU1;
    typedef typename LFSV::template Child<0>::Type LFSV0;
    typedef typename LFSV::template Child<1>::Type LFSV1;

    if(in_phase == 0) {
      const LFSU0& lfsu_s0 = lfsu_s.template child<0>();
      const LFSV0& lfsv_s0 = lfsv_s.template child<0>();

      if(out_phase == 0) {
        const LFSU0& lfsu_n0 = lfsu_n.template child<0>();
        const LFSV0& lfsv_n0 = lfsv_n.template child<0>();
        Dune::PDELab::LocalAssemblerCallSwitch<LOP0,LOP0::doAlphaSkeleton CALLSWITCHLINEARFLAG>::
        alpha_skeleton(lop0, ig, lfsu_s0, x_s, lfsv_s0, lfsu_n0, x_n, lfsv_n0, r_s, r_n);
      }
      else if(out_phase == 1) {
        // we require COP::doSkeletonTwoSided == false
#if DUNE_VERSION_NEWER(DUNE_COMMON,2,4)
        static_assert(COP::doSkeletonTwoSided == false,
                           "Coupling Operator must use doSkeletonTwoSided == false");
#else
        dune_static_assert(COP::doSkeletonTwoSided == false,
                           "Coupling Operator must use doSkeletonTwoSided == false");
#endif

        Dune::PDELab::LocalAssemblerCallSwitch<LOP0,LOP0::doAlphaBoundary CALLSWITCHLINEARFLAG>::
        alpha_boundary(lop0, ig, lfsu_s0, x_s, lfsv_s0, r_s);
        //#warning how should I flip the intersection?
        // Dune::PDELab::LocalAssemblerCallSwitch<LOP1,LOP1::doAlphaBoundary CALLSWITCHLINEARFLAG>::
        //   alpha_boundary(lop1, ig, lfsu_s1, x_s, lfsv_s1, r_s);
        // always apply from Omega_0 to Omega_1
        if (COP::doAlphaCoupling)
        {
          const LFSU1& lfsu_n1 = lfsu_n.template child<1>();
          const LFSV1& lfsv_n1 = lfsv_n.template child<1>();
          cop.alpha_coupling(ig, lfsu_s0, x_s, lfsv_s0, lfsu_n1, x_n, lfsv_n1, r_s, r_n);
        }
      }
    }
    else if(in_phase == 1) {
      const LFSU1& lfsu_s1 = lfsu_s.template child<1>();
      const LFSV1& lfsv_s1 = lfsv_s.template child<1>();

      if(out_phase == 0) {
        // we require COP::doSkeletonTwoSided == false
#if DUNE_VERSION_NEWER(DUNE_COMMON,2,4)
        static_assert(COP::doSkeletonTwoSided == false,
                           "Coupling Operator must use doSkeletonTwoSided == false");
#else
        dune_static_assert(COP::doSkeletonTwoSided == false,
                           "Coupling Operator must use doSkeletonTwoSided == false");
#endif
        // always apply from Omega_0 to Omega_1

        // should never be called from the assembler
        DUNE_THROW(Dune::Exception, "During coupling, domain 0 must always be the interior domain");

        // const LFSU0& lfsu_n0 = lfsu_n.template child<0>();
        // const LFSV0& lfsv_n0 = lfsv_n.template child<0>();
        // Dune::PDELab::LocalAssemblerCallSwitch<COP10,COP10::doAlphaCoupling CALLSWITCHLINEARFLAG>::
        //   alpha_skeleton(cop10, ig, lfsu_s1, x_s, lfsv_s1, lfsu_n0, x_n, lfsv_n0, r_s, r_n);
      }
      else if(out_phase == 1) {
        const LFSU1& lfsu_n1 = lfsu_n.template child<1>();
        const LFSV1& lfsv_n1 = lfsv_n.template child<1>();
        Dune::PDELab::LocalAssemblerCallSwitch<LOP1,LOP1::doAlphaSkeleton CALLSWITCHLINEARFLAG>::
        alpha_skeleton(lop1, ig, lfsu_s1, x_s, lfsv_s1, lfsu_n1, x_n, lfsv_n1, r_s, r_n);
      }
    }
  }

  template<typename IG, typename LFSV, typename R>
  void lambda_skeleton(const IG& ig,const LFSV& lfsv_s, const LFSV& lfsv_n, R& r_s, R& r_n) const
  {
    const int in_phase  = ig.intersection().insideDomainIndex();
    const int out_phase = ig.intersection().outsideDomainIndex();

    typedef typename LFSV::template Child<0>::Type LFSV0;
    typedef typename LFSV::template Child<1>::Type LFSV1;

    if(in_phase == 0) {
      const LFSV0& lfsv_s0 = lfsv_s.template child<0>();

      if(out_phase == 0) {
        const LFSV0& lfsv_n0 = lfsv_n.template child<0>();
        Dune::PDELab::LocalAssemblerCallSwitch<LOP0,LOP0::doLambdaSkeleton CALLSWITCHLINEARFLAG>::
        lambda_skeleton(lop0, ig, lfsv_s0, lfsv_n0, r_s, r_n);
      }
      else if(out_phase == 1) {
        // we require COP::doSkeletonTwoSided == false
#if DUNE_VERSION_NEWER(DUNE_COMMON,2,4)
        static_assert(COP::doSkeletonTwoSided == false,
                           "Coupling Operator must use doSkeletonTwoSided == false");
#else
        dune_static_assert(COP::doSkeletonTwoSided == false,
                           "Coupling Operator must use doSkeletonTwoSided == false");
#endif
        Dune::PDELab::LocalAssemblerCallSwitch<LOP0,LOP0::doLambdaBoundary CALLSWITCHLINEARFLAG>::
        lambda_boundary(lop0, ig, lfsv_s0, r_s);
        // always apply from Omega_0 to Omega_1
        if (COP::doLambdaCoupling)
        {
          const LFSV1& lfsv_n1 = lfsv_n.template child<1>();
          cop.lambda_coupling(ig, lfsv_s0, lfsv_n1, r_s, r_n);
          // Dune::PDELab::LocalAssemblerCallSwitch<COP01,COP01::doLambdaCoupling CALLSWITCHLINEARFLAG>::
          //   lambda_coupling(cop01, ig, lfsv_s0, lfsv_n1, r_s, r_n);
        }
      }
    }
    else if(in_phase == 1) {
      const LFSV1& lfsv_s1 = lfsv_s.template child<1>();

      if(out_phase == 0) {
        // we require COP::doSkeletonTwoSided == false
#if DUNE_VERSION_NEWER(DUNE_COMMON,2,4)
        static_assert(COP::doSkeletonTwoSided == false,
                           "Coupling Operator must use doSkeletonTwoSided == false");
#else
        dune_static_assert(COP::doSkeletonTwoSided == false,
                           "Coupling Operator must use doSkeletonTwoSided == false");
#endif

        // should never be called from the assembler
        DUNE_THROW(Dune::Exception, "During coupling, domain 0 must always be the interior domain");

        // const LFSV0& lfsv_n0 = lfsv_n.template child<0>();
        // Dune::PDELab::LocalAssemblerCallSwitch<COP10,COP10::doLambdaCoupling CALLSWITCHLINEARFLAG>::
        //   lambda_coupling(lop, ig, lfsv_s1, lfsv_n0, r_s, r_n);
      }
      else if(out_phase == 1) {
        const LFSV1& lfsv_n1 = lfsv_n.template child<1>();
        Dune::PDELab::LocalAssemblerCallSwitch<LOP1,LOP1::doLambdaSkeleton CALLSWITCHLINEARFLAG>::
        lambda_skeleton(lop1, ig, lfsv_s1, lfsv_n1, r_s, r_n);
      }
    }
  }


  template<typename IG, typename LFSU, typename X, typename LFSV, typename M>
  void jacobian_skeleton
    ( const IG& ig,
    const LFSU& lfsu_s, const X& x_s, const LFSV& lfsv_s,
    const LFSU& lfsu_n, const X& x_n, const LFSV& lfsv_n,
    M & mat_ss, M & mat_sn,
    M & mat_ns, M & mat_nn) const
  {
    const int in_phase  = ig.intersection().insideDomainIndex();
    const int out_phase = ig.intersection().outsideDomainIndex();

    typedef typename LFSU::template Child<0>::Type LFSU0;
    typedef typename LFSU::template Child<1>::Type LFSU1;
    typedef typename LFSV::template Child<0>::Type LFSV0;
    typedef typename LFSV::template Child<1>::Type LFSV1;

    if(in_phase == 0) {
      const LFSU0& lfsu_s0 = lfsu_s.template child<0>();
      const LFSV0& lfsv_s0 = lfsv_s.template child<0>();

      if(out_phase == 0) {
        const LFSU0& lfsu_n0 = lfsu_n.template child<0>();
        const LFSV0& lfsv_n0 = lfsv_n.template child<0>();
        Dune::PDELab::LocalAssemblerCallSwitch<LOP0,LOP0::doAlphaSkeleton CALLSWITCHLINEARFLAG>::
        jacobian_skeleton(lop0, ig, lfsu_s0, x_s, lfsv_s0, lfsu_n0, x_n, lfsv_n0,
                          mat_ss, mat_sn, mat_ns, mat_nn);
      }
      else if(out_phase == 1) {
        // we require COP::doSkeletonTwoSided == false
#if DUNE_VERSION_NEWER(DUNE_COMMON,2,4)
        static_assert(COP::doSkeletonTwoSided == false,
                           "Coupling Operator must use doSkeletonTwoSided == false");
#else
        dune_static_assert(COP::doSkeletonTwoSided == false,
                           "Coupling Operator must use doSkeletonTwoSided == false");
#endif
        // always apply from Omega_0 to Omega_1
        if (COP::doAlphaCoupling)
        {
          const LFSU1& lfsu_n1 = lfsu_n.template child<1>();
          const LFSV1& lfsv_n1 = lfsv_n.template child<1>();
          cop.jacobian_coupling(ig, lfsu_s0, x_s, lfsv_s0, lfsu_n1, x_n, lfsv_n1,
                                mat_ss, mat_sn, mat_ns, mat_nn);
        }
      }
    }
    else if(in_phase == 1) {
      const LFSU1& lfsu_s1 = lfsu_s.template child<1>();
      const LFSV1& lfsv_s1 = lfsv_s.template child<1>();

      if(out_phase == 0) {
        // we require COP::doSkeletonTwoSided == false
#if DUNE_VERSION_NEWER(DUNE_COMMON,2,4)
        static_assert(COP::doSkeletonTwoSided == false,
                           "Coupling Operator must use doSkeletonTwoSided == false");
#else
        dune_static_assert(COP::doSkeletonTwoSided == false,
                           "Coupling Operator must use doSkeletonTwoSided == false");
#endif

        // should never be called from the assembler
        DUNE_THROW(Dune::Exception, "During coupling, domain 0 must always be the interior domain");

        // const LFSU0& lfsu_n0 = lfsu_n.template child<0>();
        // const LFSV0& lfsv_n0 = lfsv_n.template child<0>();
        // Dune::PDELab::LocalAssemblerCallSwitch<LOP,doAlphaSkeleton CALLSWITCHLINEARFLAG>::
        //   jacobian_skeleton(lop, ig, lfsu_s1, x_s, lfsv_s1, lfsu_n0, x_n, lfsv_n0,
        //                     mat_ss, mat_sn, mat_ns, mat_nn);
      }
      else if(out_phase == 1) {
        const LFSU1& lfsu_n1 = lfsu_n.template child<1>();
        const LFSV1& lfsv_n1 = lfsv_n.template child<1>();
        Dune::PDELab::LocalAssemblerCallSwitch<LOP1,LOP1::doAlphaSkeleton CALLSWITCHLINEARFLAG>::
        jacobian_skeleton(lop1, ig, lfsu_s1, x_s, lfsv_s1, lfsu_n1, x_n, lfsv_n1,
                          mat_ss, mat_sn, mat_ns, mat_nn);
      }
    }
  }

  template<typename IG, typename LFSU, typename X, typename LFSV, typename R>
  void alpha_boundary (const IG& ig,
                       const LFSU& lfsu_s, const X& x_s, const LFSV& lfsv_s,
                       R& r_s) const
  {
    const int in_phase  = ig.intersection().insideDomainIndex();
    const int out_phase = ig.intersection().outsideDomainIndex();

    typedef typename LFSU::template Child<0>::Type LFSU0;
    typedef typename LFSU::template Child<1>::Type LFSU1;
    typedef typename LFSV::template Child<0>::Type LFSV0;
    typedef typename LFSV::template Child<1>::Type LFSV1;

    if(out_phase == -1) {
      // boundary omega_0
      if(in_phase == 0) {
        const LFSU0& lfsu_s0 = lfsu_s.template child<0>();
        const LFSV0& lfsv_s0 = lfsv_s.template child<0>();
        Dune::PDELab::LocalAssemblerCallSwitch<LOP0,LOP0::doAlphaBoundary CALLSWITCHLINEARFLAG>::
        alpha_boundary(lop0, ig, lfsu_s0, x_s, lfsv_s0, r_s);
      }
      // boundary omega_1
      else if(in_phase == 1) {
        const LFSU1& lfsu_s1 = lfsu_s.template child<1>();
        const LFSV1& lfsv_s1 = lfsv_s.template child<1>();
        Dune::PDELab::LocalAssemblerCallSwitch<LOP1,LOP1::doAlphaBoundary CALLSWITCHLINEARFLAG>::
        alpha_boundary(lop1, ig, lfsu_s1, x_s, lfsv_s1, r_s);
      }
    }
    else{
      this->alpha_skeleton(ig, lfsu_s, x_s, lfsv_s, lfsu_s, x_s, lfsv_s, r_s, r_s);
    }
  }

  template<typename IG, typename LFSV, typename R>
  void lambda_boundary (const IG& ig, const LFSV& lfsv_s, R& r_s) const
  {
    const int in_phase  = ig.intersection().insideDomainIndex();
    const int out_phase = ig.intersection().outsideDomainIndex();

    typedef typename LFSV::template Child<0>::Type LFSV0;
    typedef typename LFSV::template Child<1>::Type LFSV1;

    if(out_phase == -1) {
      if(in_phase == 0) {
        const LFSV0& lfsv_s0 = lfsv_s.template child<0>();
        Dune::PDELab::LocalAssemblerCallSwitch<LOP0,LOP0::doLambdaBoundary CALLSWITCHLINEARFLAG>::
        lambda_boundary(lop0, ig, lfsv_s0, r_s);
      }
      else if(in_phase == 1) {
        const LFSV1& lfsv_s1 = lfsv_s.template child<1>();
        Dune::PDELab::LocalAssemblerCallSwitch<LOP1,LOP1::doLambdaBoundary CALLSWITCHLINEARFLAG>::
        lambda_boundary(lop1, ig, lfsv_s1, r_s);
      }
    }
    else{
      this->lambda_skeleton(ig, lfsv_s, lfsv_s, r_s, r_s);
    }
  }


  template<typename IG, typename LFSU, typename X, typename LFSV, typename M>
  void jacobian_boundary
    ( const IG& ig, const LFSU& lfsu_s, const X& x_s, const LFSV& lfsv_s,
    M & mat_ss) const
  {
    const int in_phase  = ig.intersection().insideDomainIndex();
    const int out_phase = ig.intersection().outsideDomainIndex();

    typedef typename LFSU::template Child<0>::Type LFSU0;
    typedef typename LFSU::template Child<1>::Type LFSU1;
    typedef typename LFSV::template Child<0>::Type LFSV0;
    typedef typename LFSV::template Child<1>::Type LFSV1;

    if(out_phase == -1) {
      if(in_phase == 0) {
        const LFSU0& lfsu_s0 = lfsu_s.template child<0>();
        const LFSV0& lfsv_s0 = lfsv_s.template child<0>();
        Dune::PDELab::LocalAssemblerCallSwitch<LOP0,LOP0::doAlphaBoundary CALLSWITCHLINEARFLAG>::
        jacobian_boundary(lop0, ig, lfsu_s0, x_s, lfsv_s0, mat_ss);
      }
      else if(in_phase == 1) {
        const LFSU1& lfsu_s1 = lfsu_s.template child<1>();
        const LFSV1& lfsv_s1 = lfsv_s.template child<1>();
        Dune::PDELab::LocalAssemblerCallSwitch<LOP1,LOP1::doAlphaBoundary CALLSWITCHLINEARFLAG>::
        jacobian_boundary(lop1, ig, lfsu_s1, x_s, lfsv_s1, mat_ss);
      }
    }
    else{
      this->jacobian_skeleton(ig, lfsu_s, x_s, lfsv_s, lfsu_s, x_s, lfsv_s,
                              mat_ss, mat_ss, mat_ss, mat_ss);
    }
  }
};

/**
   \brief Specialization of TwoPhaseLocalOperatorWrapper for DG type
   nested local operators.

   \tparam The type of the nested local operators.
 */
template<typename LOP0, typename LOP1=LOP0,
    typename COP=SimpleSkeletonCouplingOperator<LOP0,LOP1> >
class DGTwoPhaseLocalOperatorWrapper
  : public TwoPhaseLocalOperatorWrapper<LOP0, LOP1, COP>,
    public Dune::PDELab::FullVolumePattern,
    public Dune::PDELab::FullSkeletonPattern
{
  SimpleSkeletonCouplingOperator<LOP0,LOP0> &
  simpleSkeletonCouplingOperator(LOP0 & lop)
  {
    static SimpleSkeletonCouplingOperator<LOP0,LOP0> cop(lop);
    return cop;
  }
public:
  DGTwoPhaseLocalOperatorWrapper(LOP0 & lop)
    : TwoPhaseLocalOperatorWrapper<LOP0, LOP1, COP>(lop, lop, simpleSkeletonCouplingOperator(lop))
  {}

  DGTwoPhaseLocalOperatorWrapper(LOP0 & lop0, LOP1 & lop1, COP & cop)
    : TwoPhaseLocalOperatorWrapper<LOP0, LOP1, COP>(lop0, lop1, cop)
  {}
};


#endif
