#ifndef DUNE_UDG_SIMPLETPMCTRIANGULATION_HH
#define DUNE_UDG_SIMPLETPMCTRIANGULATION_HH

#include <dune/geometry/referenceelements.hh>
#include <dune/udg/vtriangulation.hh>
#include <dune/udg/common/entitypart.hh>
#include <dune/udg/common/intersectionpart.hh>
#include <dune/tpmc/tpmcrefinement.hh>
#include <dune/udg/simpletpmctriangulation/volumesubtriangulation.hh>
#include <dune/udg/simpletpmctriangulation/internalinterfacesubtriangulation.hh>
#include <dune/udg/simpletpmctriangulation/internalinterfaceareamap.hh>
#include <dune/udg/simpletpmctriangulation/cutcell.hh>
#include <dune/udg/simpletpmctriangulation/domainconfiguration.hh>
#include <dune/udg/simpletpmctriangulation/externalinterfacesubtriangulation.hh>
#include <dune/udg/simpletpmctriangulation/externalinterfaceareamap.hh>
#include <dune/udg/simpletpmctriangulation/cutcellinformation.hh>

namespace Dune
{
  namespace UDG
  {
    namespace SimpleTpmcTriangulationDetail
    {
      template <class GOut, class G>
      GOut createIdentityGeometry(const G& geometry)
      {
        using ctype = typename G::ctype;
        const auto& referenceElement
            = ReferenceElements<ctype, G::mydimension>::general(geometry.type());
        std::vector<FieldVector<ctype, G::mydimension> > corners;
        corners.reserve(referenceElement.size(G::mydimension));
        for (int i = 0; i < referenceElement.size(G::mydimension); ++i) {
          corners.push_back(referenceElement.position(i, G::mydimension));
        }
        return GOut(geometry.type(), corners);
      }
    }

    enum class SimpleTpmcIntersections { all, none, internal_only };

    inline SimpleTpmcIntersections simpleTPMCIntersectionsFromString(const std::string& x)
    {
      if (x == "all")
        return SimpleTpmcIntersections::all;
      if (x == "none")
        return SimpleTpmcIntersections::none;
      if (x == "internal_only")
        return SimpleTpmcIntersections::internal_only;
      DUNE_THROW(Dune::Exception, "Unknown intersections type \"" << x << "\"");
    }

    template <class GV, class LGV>
    class SimpleTpmcTriangulation : public VirtualSubTriangulation<GV>
    {
    public:
      using BaseT = VirtualSubTriangulation<GV>;
      enum { dim = GV::dimension };
      using ctype = typename BaseT::ctype;
      using Entity = typename BaseT::Entity;
      using Geometry = Dune::CachedMultiLinearGeometry<ctype, dim, dim>;
      using EntityPart = typename BaseT::EntityPart;
      using EntityPartGeometry = Dune::UDG::DefaultEntityPartGeometry<GV>;
      using EntityPartImpl = Dune::UDG::DefaultEntityPart<EntityPartGeometry>;
      using EntityPartInterface = Dune::UDG::EntityPartInterface<GV>;
      using EntityPartPointer = typename EntityPartInterface::EntityPartPointer;
      using EntityPartList = typename BaseT::EntityPartList;
      using IntersectionPartList = typename BaseT::IntersectionPartList;
      using IntersectionPart = typename BaseT::IntersectionPart;
      using IntersectionPartGeometry = AffineIntersectionPartGeometry<GV>;
      using IntersectionPartImpl = DefaultIntersectionPart<IntersectionPartGeometry>;
      using BoundingBox = typename BaseT::BoundingBox;

      explicit SimpleTpmcTriangulation(
          const GV& fundamentalGridView, const LGV& levelSetGridView,
          const DomainConfiguration<GV, LGV>& domainConfiguration, bool forceRefinement = false,
          SimpleTpmcIntersections intersectionTypes = SimpleTpmcIntersections::all,
          ctype valueTolerance = 1e-8)
          : fundamentalGridView_(fundamentalGridView)
          , fundamentalElementMapper_(fundamentalGridView_)
          , levelSetGridView_(levelSetGridView)
          , domainConfiguration_(domainConfiguration)
          , forceRefinement_(forceRefinement)
          , intersectionTypes_(intersectionTypes)
          , valueTolerance_(valueTolerance)
          , cutCellInformation_(fundamentalGridView_, levelSetGridView_, domainConfiguration_,
                                valueTolerance, 1)
      {
        cutCellInformation_.printStatistics();
      }

      virtual void create_entity_parts(const Entity& e, EntityPartList& entityparts) const override
      {
        if (!forceRefinement_ && cutCellInformation_.filledBySingleCutCell(e)) {
          auto domainIndex = cutCellInformation_.domainOfSingleCutCell(e);
          const auto& info = cutCellInformation_.information(e, domainIndex);
          VolumeSnippet<GV, LGV> snippet(
              0, SimpleTpmcTriangulationDetail::createIdentityGeometry<
                     typename VolumeSnippet<GV, LGV>::GeometryInFundamental>(e.geometry()),
              std::make_shared<Entity>(e));
          entityparts.push_back(createEntityPartFromSnippet(snippet, e,
              *domainConfiguration_.findDomain(domainIndex), info.boundingBox, info.volume));
        } else {
          VolumeSubTriangulation<GV, LGV> volumeTriangulation(
              levelSetGridView_, domainConfiguration_, e, forceRefinement_, valueTolerance_);

          // create entity parts from cc
          for (const Domain& domain : domainConfiguration_.domains()) {
            if (cutCellInformation_.cutCellsExist(e, domain.index())) {
              const auto& info = cutCellInformation_.information(e, domain.index());
              for (const CutCell<GV, LGV>& cutcell : volumeTriangulation.cutCells(domain.index())) {
                for (const VolumeSnippet<GV, LGV>& snippet : cutcell) {
                  entityparts.push_back(createEntityPartFromSnippet(
                      snippet, e, domain, info.boundingBox, info.volume));
                }
              }
            }
          }
        }
      }

      virtual void create_intersection_parts(
          const Entity& e, IntersectionPartList& intersectionparts) const override
      {
        if (!cutCellInformation_.filledBySingleCutCell(e)
            && intersectionTypes_ != SimpleTpmcIntersections::none) {
          InternalInterfaceSubTriangulation<GV, LGV> internalTriangulation(
              levelSetGridView_, domainConfiguration_, e, valueTolerance_);
          InternalInterfaceAreaMap<GV, LGV> internalInterfaceAreaMap(
              internalTriangulation, domainConfiguration_);

          for (const InternalCutCellInterface<GV, LGV>& ii :
              internalTriangulation.internalCutCellInterfaces()) {
            for (const InternalInterfaceSnippet<GV, LGV>& snippet : ii) {
              auto ip = createIntersectionPartFromSnippet(snippet, e, ii, internalInterfaceAreaMap);
              if (ip)
                intersectionparts.push_back(ip);
            }
          }
        }

        if (intersectionTypes_ == SimpleTpmcIntersections::all) {
          ExternalInterfaceSubTriangulation<GV, LGV> externalTriangulation(
              fundamentalGridView_, levelSetGridView_, domainConfiguration_, e, valueTolerance_);
          ExternalInterfaceAreaMap<GV, LGV> externalInterfaceAreaMap(
              fundamentalGridView_, externalTriangulation, domainConfiguration_);

          for (const Domain& domain : domainConfiguration_.domains()) {
            if (cutCellInformation_.cutCellsExist(e, domain.index())) {
              for (const ExternalCutCellInterface<GV, LGV>& ei :
                   externalTriangulation.externalCutCellInterfaces(domain.index())) {
                for (const ExternalInterfaceSnippet<GV, LGV>& snippet : ei) {
                  if (cutCellInformation_.cutCellsExist(snippet.outsideFundamentalElementIndex(),
                                                        domain.index())) {
                    intersectionparts.push_back(createIntersectionPartFromSnippet(
                        snippet, e, domain, externalInterfaceAreaMap));
                  }
                }
              }
            }
          }
        }
      }

      virtual const BoundingBox boundingBox(const EntityPart& ep) const override
      {
        return ep.boundingBox();
      }

      virtual const BoundingBox BBox(const unsigned int domainIndex, const Entity& e) const override
      {
        return cutCellInformation_.information(e, domainIndex).boundingBox;
      }

      virtual bool& forceRefinement() override
      {
        return forceRefinement_;
      }

      virtual bool isInDomain(const Entity& e, const int& domain_index) const override
      {
        assert(false);
        return false;
      }
      virtual bool isHostCell(const Entity& hostCandidate, const int& domainIndex) const override
      {
        return cutCellInformation_.cutCellsExist(hostCandidate, domainIndex);
      }
      virtual bool isHostCell(const Entity& hostCandidate) const
      {
        for (const auto& domain : domainConfiguration_.domains()) {
          if (isHostCell(hostCandidate, domain.index())) {
            return true;
          }
        }
        return false;
      }

      virtual ctype getEdgeNorm(const IntersectionPart& ipart) const override
      {
        // simply use the default implementation from mc33triangulation. This method should
        // probably not be part of the subtriangulation interface
        const ctype av = 1.0 / (1.0 / ipart.insideVolume() + 1.0 / ipart.outsideVolume());
        return std::pow(av, 1.0 / ctype(dim));
      }

      virtual const GV& gridView() const override
      {
        return fundamentalGridView_;
      }

      virtual const CutCellInformation<GV, LGV>& cutCellInformation() const
      {
        return cutCellInformation_;
      }

      virtual CutCellInformation<GV, LGV>& cutCellInformation()
      {
        return cutCellInformation_;
      }

      virtual ~SimpleTpmcTriangulation()
      {
      }

    private:
      std::shared_ptr<EntityPart> createEntityPartFromSnippet(const VolumeSnippet<GV, LGV>& snippet,
          const Entity& e, const Domain& domain, const BoundingBox& boundingBox, ctype volume) const
      {
        // transform coordinates in fundamental cell to global coordinates
        std::vector<FieldVector<ctype, dim> > globalCornerCoordinates;
        for (int i = 0; i < snippet.geometryInFundamental().corners(); ++i) {
          globalCornerCoordinates.push_back(
              e.geometry().global(snippet.geometryInFundamental().corner(i)));
        }

// create geometry of entity part
#if DUNE_VERSION_NEWER(DUNE_GEOMETRY, 2, 6)
        typename EntityPartGeometry::GeometryImp geometry(
            Dune::geometryTypeFromVertexCount(dim, globalCornerCoordinates.size()),
            globalCornerCoordinates);
#else
        GeometryType snippetType;
        snippetType.makeFromVertices(dim, globalCornerCoordinates.size());
        typename EntityPartGeometry::GeometryImp geometry(snippetType, globalCornerCoordinates);
#endif
        EntityPartGeometry partGeometry(geometry, *snippet.homeElement(), boundingBox);

        return std::make_shared<EntityPartImpl>(snippet.index(), partGeometry, domain.oldTag(),
                                                domain.tag(), domain.index(), volume);
      }

      std::shared_ptr<IntersectionPart> createIntersectionPartFromSnippet(
          const InternalInterfaceSnippet<GV, LGV>& snippet, const Entity& e,
          const InternalCutCellInterface<GV, LGV>& ii,
          const InternalInterfaceAreaMap<GV, LGV>& areaMap) const
      {
        // transform coordinates in fundamental cell to global coordinates
        std::vector<FieldVector<ctype, dim> > globalCornerCoordinates;
        for (int i = 0; i < snippet.globalGeometry().corners(); ++i) {
          globalCornerCoordinates.push_back(snippet.globalGeometry().corner(i));
        }

// create geometry of intersection part
#if DUNE_VERSION_NEWER(DUNE_GEOMETRY, 2, 6)
        typename IntersectionPartGeometry::FaceGeometryImp geometry(
            Dune::geometryTypeFromVertexCount(dim - 1, globalCornerCoordinates.size()),
            globalCornerCoordinates);
#else
        GeometryType snippetType;
        snippetType.makeFromVertices(dim - 1, globalCornerCoordinates.size());
        typename IntersectionPartGeometry::FaceGeometryImp geometry(snippetType,
                                                                    globalCornerCoordinates);
#endif

        // find interior and exterior domains along with their cut cell information

        auto interiorDomain
            = domainConfiguration_.findInteriorDomain(ii.interfaceRelativePositions());
        auto exteriorDomain
            = domainConfiguration_.findExteriorDomain(ii.interfaceRelativePositions());

        if (interiorDomain == domainConfiguration_.domainsEnd()) {
          if (exteriorDomain != domainConfiguration_.domainsEnd()
              && cutCellInformation_.cutCellsExist(e, exteriorDomain->index())) {
            auto area = areaMap.area(*exteriorDomain);
            // make sure domain with smaller index is exterior
            const auto& exteriorInformation
                = cutCellInformation_.information(e, exteriorDomain->index());

            // create intersection part geometry
            IntersectionPartGeometry igeom(geometry, exteriorInformation.boundingBox,
                exteriorInformation.boundingBox, snippet.globalUnitOuterNormal());

            return std::make_shared<IntersectionPartImpl>(
                igeom, 255, toDomainTag(ii.interfaceRelativePositions()), Dune::UDG::internal,
                exteriorDomain->index(), -1, area, exteriorInformation.volume, 0.0);
          }
        } else if (cutCellInformation_.cutCellsExist(e, interiorDomain->index())) {
          if (exteriorDomain == domainConfiguration_.domainsEnd()) {
            auto area = areaMap.area(*interiorDomain);
            // make sure domain with smaller index is interior
            const auto& interiorInformation
                = cutCellInformation_.information(e, interiorDomain->index());

            // create intersection part geometry
            IntersectionPartGeometry igeom(geometry, interiorInformation.boundingBox,
                interiorInformation.boundingBox, snippet.globalUnitOuterNormal());

            return std::make_shared<IntersectionPartImpl>(
                igeom, 255, toDomainTag(ii.interfaceRelativePositions()), Dune::UDG::internal,
                interiorDomain->index(), -1, area, interiorInformation.volume, 0.0);
          } else if (cutCellInformation_.cutCellsExist(e, exteriorDomain->index())) {
            auto area = areaMap.area(*interiorDomain, *exteriorDomain);
            // make sure domain with smaller index is interior
            if (interiorDomain->index() > exteriorDomain->index()) {
              std::swap(interiorDomain, exteriorDomain);
            }
            const auto& interiorInformation
                = cutCellInformation_.information(e, interiorDomain->index());
            const auto& exteriorInformation
                = cutCellInformation_.information(e, exteriorDomain->index());

            // create intersection part geometry
            IntersectionPartGeometry igeom(geometry, interiorInformation.boundingBox,
                exteriorInformation.boundingBox, snippet.globalUnitOuterNormal());

            return std::make_shared<IntersectionPartImpl>(
                igeom, 255, toDomainTag(ii.interfaceRelativePositions()), Dune::UDG::internal,
                interiorDomain->index(), exteriorDomain->index(), area, interiorInformation.volume,
                exteriorInformation.volume);
          }
        }
        return std::shared_ptr<IntersectionPart>();
        // DUNE_THROW(Dune::Exception, "cant create intersectionpart from snippet for which both "
        //"interior and exterior domain are not present");
      }

      std::shared_ptr<IntersectionPart> createIntersectionPartFromSnippet(
          const ExternalInterfaceSnippet<GV, LGV>& snippet, const Entity& e, const Domain& domain,
          const ExternalInterfaceAreaMap<GV, LGV>& areaMap) const
      {
        // transform coordinates in fundamental cell to global coordinates
        std::vector<FieldVector<ctype, dim> > globalCornerCoordinates;
        for (int i = 0; i < snippet.globalGeometry().corners(); ++i) {
          globalCornerCoordinates.push_back(snippet.globalGeometry().corner(i));
        }

        // create geometry of intersection part

#if DUNE_VERSION_NEWER(DUNE_GEOMETRY, 2, 6)
        typename IntersectionPartGeometry::FaceGeometryImp geometry(
            Dune::geometryTypeFromVertexCount(dim - 1, globalCornerCoordinates.size()),
            globalCornerCoordinates);
#else
        GeometryType snippetType;
        snippetType.makeFromVertices(dim - 1, globalCornerCoordinates.size());
        typename IntersectionPartGeometry::FaceGeometryImp geometry(snippetType,
                                                                    globalCornerCoordinates);
#endif

        const auto& interiorInformation = cutCellInformation_.information(e, domain.index());

        if (snippet.outsideFundamentalElementIndex() == fundamentalElementMapper_.index(e)) {

          IntersectionPartGeometry igeom(geometry, interiorInformation.boundingBox,
              interiorInformation.boundingBox, snippet.globalUnitOuterNormal());

          return std::make_shared<IntersectionPartImpl>(
              igeom, domain.oldTag(), domain.tag(), Dune::UDG::internal, domain.index(), -1,
              areaMap.area(snippet.codim1IndexInInside(), domain), interiorInformation.volume,
              interiorInformation.volume);
        } else {
          const auto& exteriorInformation = cutCellInformation_.information(
              snippet.outsideFundamentalElementIndex(), domain.index());

          IntersectionPartGeometry igeom(geometry, interiorInformation.boundingBox,
              exteriorInformation.boundingBox, snippet.globalUnitOuterNormal());

          return std::make_shared<IntersectionPartImpl>(
              igeom, domain.oldTag(), domain.tag(), Dune::UDG::internal, domain.index(),
              domain.index(), areaMap.area(snippet.codim1IndexInInside(), domain),
              interiorInformation.volume, exteriorInformation.volume);
        }
      }

      GV fundamentalGridView_;
      Dune::SingleCodimSingleGeomTypeMapper<GV, 0> fundamentalElementMapper_;
      LGV levelSetGridView_;
      DomainConfiguration<GV, LGV> domainConfiguration_;
      bool forceRefinement_;
      SimpleTpmcIntersections intersectionTypes_;
      ctype valueTolerance_;
      CutCellInformation<GV, LGV> cutCellInformation_;
    };
  }
}

#endif // DUNE_UDG_SIMPLETPMCTRIANGULATION_HH
