#ifndef DUNE_UDG_CUTCELL_HH
#define DUNE_UDG_CUTCELL_HH

#include <vector>
#include <dune/tpmc/tpmcrefinement.hh>
#include <dune/geometry/quadraturerules.hh>
#include <dune/udg/simpletpmctriangulation/interface.hh>
#include <dune/udg/simpletpmctriangulation/volumesnippet.hh>
#include <dune/udg/mc33triangulation/domaincell.hh> // for RawBoundingBox
#include <dune/udg/common/cuboid.hh>
#include <dune/udg/common/geometryinterface.hh>

namespace Dune
{
  namespace UDG
  {
    namespace CutCellDetail
    {
      template <int dim, class Snippet>
      void checkSnippet(const Snippet& snippet)
      {
        for (int i = 0; i < snippet.geometryInFundamental().corners(); ++i) {
          for (unsigned int d = 0; d < dim; ++d) {
            if (std::isnan(snippet.geometryInFundamental().corner(i)[d])) {
              DUNE_THROW(Dune::Exception, "corner is nan");
            }
          }
        }
      }

      template <class GOut, class G>
      GOut createIdentityGeometry(const G& geometry)
      {
        using ctype = typename G::ctype;
        const auto& referenceElement
            = ReferenceElements<ctype, G::mydimension>::general(geometry.type());
        std::vector<FieldVector<ctype, G::mydimension> > corners;
        corners.reserve(referenceElement.size(G::mydimension));
        for (int i = 0; i < referenceElement.size(G::mydimension); ++i) {
          corners.push_back(referenceElement.position(i, G::mydimension));
        }
        return GOut(geometry.type(), corners);
      }
    }
    template <class GV, class LGV>
    class CutCell
    {
    public:
      using FundamentalGridView = GV;
      using FundamentalElement = typename GV::template Codim<0>::Entity;
      using FundamentalGeometry = typename FundamentalElement::Geometry;
      using LevelSetGridView = LGV;
      using LevelSetElement = typename LGV::template Codim<0>::Entity;
      using ctype = typename GV::ctype;
      enum { dim = GV::dimension };
      using RawBoundingBox = MarchingCube33SubTriangulationImp::RawBoundingBox<ctype, dim>;
      using BoundingBox = Dune::GridCuboid<typename GV::Grid>;
      using const_iterator = typename std::vector<VolumeSnippet<GV, LGV> >::const_iterator;
      using Coordinate = Dune::FieldVector<ctype, dim>;

      CutCell(const FundamentalElement& fundamentalElement,
              const LevelSetGridView& levelSetGridView, ctype valueTolerance = 1e-8)
          : fundamentalGeometry_(fundamentalElement.geometry())
          , volumeInFundamental_(0.0)
          , fillsFundamentalCell_(true)
          , valueTolerance_(valueTolerance)
      {
        auto fundamentalGeometry = fundamentalElement.geometry();
        auto maxLevel = levelSetGridView.grid().maxLevel();

        if (fundamentalElement.isLeaf()
            && levelSetGridView.indexSet().contains(fundamentalElement)) {
          using GIF = typename VolumeSnippet<GV, LGV>::GeometryInFundamental;
          addSnippet(CutCellDetail::createIdentityGeometry<GIF>(fundamentalGeometry),
              std::make_shared<FundamentalElement>(fundamentalElement));
        } else {
          // loop through all child elements in the level set mesh and create snippets
          for (auto it = fundamentalElement.hbegin(maxLevel);
               it != fundamentalElement.hend(maxLevel); ++it) {
            if (!levelSetGridView.indexSet().contains(*it))
              continue;
            auto levelSetElement = std::make_shared<LevelSetElement>(*it);

            using GIF = typename VolumeSnippet<GV, LGV>::GeometryInFundamental;
            addSnippet(
                combineGeometries<GIF>(levelSetElement->geometry(), fundamentalGeometry, false),
                levelSetElement);
          }
        }
      }

      CutCell(const FundamentalGeometry& fundamentalGeometry, double valueTolerance)
          : fundamentalGeometry_(fundamentalGeometry)
          , volumeInFundamental_(0.0)
          , fillsFundamentalCell_(false)
          , valueTolerance_(valueTolerance)
      {
      }

      /**
       * @brief cut the given interface into this cut cell and return the resulting interior and
       * exterior cells
       *
       * @param interface the interface to be cut into this cell
       * @param out output iterator for the resulting cut cells
       */
      template <class OutIterator>
      void cut(const Interface<LGV>& interface, OutIterator out) const
      {
        using GIF = typename VolumeSnippet<GV, LGV>::GeometryInFundamental;

        CutCell interior(fundamentalGeometry_, valueTolerance_);
        CutCell exterior(fundamentalGeometry_, valueTolerance_);

        interior.interfaceRelativePositions_ = interfaceRelativePositions_;
        interior.interfaceRelativePositions_.push_back(InterfaceRelativePosition::interior);
        exterior.interfaceRelativePositions_ = interfaceRelativePositions_;
        exterior.interfaceRelativePositions_.push_back(InterfaceRelativePosition::exterior);

        std::vector<ctype> values;

        auto localInterfaceFunction = localFunction(interface.function());

        for (const VolumeSnippet<GV, LGV>& snippet : snippets_) {
          localInterfaceFunction.bind(*snippet.homeElement());
          values.clear();
          const auto& hgeo = snippet.homeElement()->geometry();

          // retrieve level set values at corners
          bool allInside = false;
          bool allOutside = false;
          for (int i = 0; i < snippet.geometryInFundamental().corners(); ++i) {
            values.push_back(localInterfaceFunction(hgeo.local(
                fundamentalGeometry_.global(snippet.geometryInFundamental().corner(i)))));
            if (values.back() >= 0.0 && values.back() < valueTolerance_) {
              values.back() = valueTolerance_;
            } else if (values.back() > -valueTolerance_ && values.back() < 0.0) {
              values.back() = -valueTolerance_;
            }
            allInside &= values.back() < 0.0;
            allOutside &= values.back() > 0.0;
          }

          if (allInside) {
            interior.addSnippet(snippet);
          } else if (allOutside) {
            exterior.addSnippet(snippet);
          } else {
            // retrieve interior and exterior elements using tpmc
            TpmcRefinement<ctype, dim> refinement(values.begin(), values.end());
            for (const auto& part : refinement.volume(tpmc::InteriorDomain)) {
              if (part.volume() == 0.0) {
                DUNE_THROW(Dune::Exception, "got an empty part");
              }
              interior.addSnippet(
                  combineGeometries<GIF>(part, snippet.geometryInFundamental(), true),
                  snippet.homeElement());
            }
            for (const auto& part : refinement.volume(tpmc::ExteriorDomain)) {
              if (part.volume() == 0.0) {
                DUNE_THROW(Dune::Exception, "got an empty part");
              }
              exterior.addSnippet(
                  combineGeometries<GIF>(part, snippet.geometryInFundamental(), true),
                  snippet.homeElement());
            }
          }
        }
        // check if interior or exterior fills the fundamental cell
        if (fillsFundamentalCell_ && (interior.empty() ^ exterior.empty())) {
          if (interior.empty()) {
            exterior.fillsFundamentalCell_ = true;
          } else {
            interior.fillsFundamentalCell_ = true;
          }
        }
        if (!interior.empty())
          *out++ = interior;
        if (!exterior.empty())
          *out++ = exterior;
      }

      /**
       * @brief add a new snippet with with given geometry, belonging to the given level set
       * element
       * @param geometry geometry describing the new snippet
       * @param element level set element the new snippet belongs to
       */
      void addSnippet(const typename VolumeSnippet<GV, LGV>::GeometryInFundamental& geometry,
          std::shared_ptr<LevelSetElement> element)
      {
        addSnippet(VolumeSnippet<GV, LGV>(snippets_.size(), geometry, element));
      }

      /**
       * @brief add a new snippet belonging to the given level set element
       */
      void addSnippet(const VolumeSnippet<GV, LGV>& snippet)
      {
#ifndef NDEBUG
        CutCellDetail::checkSnippet<dim>(snippet);
#endif
        snippets_.push_back(snippet);
        for (int i = 0; i < snippet.geometryInFundamental().corners(); ++i) {
          rawBoundingBox_.update(
              fundamentalGeometry_.global(snippet.geometryInFundamental().corner(i)));
        }

        centerOfMassInFundamental_ *= volumeInFundamental_;

        const auto& rule = Dune::QuadratureRules<ctype, dim>::rule(
            snippet.geometryInFundamental().type(), 2 * dim + 1);
        ctype snippetVolume = 0.;
        for (const auto& qp : rule) {
          snippetVolume +=
              qp.weight() * snippet.geometryInFundamental().integrationElement(qp.position());
        }
        volumeInFundamental_ += snippetVolume;

        centerOfMassInFundamental_.axpy(snippetVolume, snippet.geometryInFundamental().center());
        centerOfMassInFundamental_ /= volumeInFundamental_;
      }

      /**
       * @brief remove all snippets from the cut cell. leave the rest untouched
       */
      void clearSnippets()
      {
        snippets_.clear();
        volumeInFundamental_ = 0.0;
        centerOfMassInFundamental_ = 0.0;
      }

      //! \brief return the position relative to each cut interface
      const std::vector<InterfaceRelativePosition>& interfaceRelativePositions() const
      {
        return interfaceRelativePositions_;
      }

      const RawBoundingBox& rawBoundingBox() const
      {
        return rawBoundingBox_;
      }

      const_iterator begin() const
      {
        return snippets_.begin();
      }

      const_iterator end() const
      {
        return snippets_.end();
      }

      bool empty() const
      {
        return snippets_.empty();
      }

      std::size_t size() const
      {
        return snippets_.size();
      }

      ctype volumeInFundamental() const
      {
        return volumeInFundamental_;
      }

      bool fillsFundamentalCell() const
      {
        return fillsFundamentalCell_;
      }

      const Coordinate& centerOfMassInFundamental() const
      {
        return centerOfMassInFundamental_;
      }

    private:
      FundamentalGeometry fundamentalGeometry_;
      //! \brief the position of this cut cell relative to each cut interface
      std::vector<InterfaceRelativePosition> interfaceRelativePositions_;
      //! \brief snippets forming the geometrical part of this cut cell
      std::vector<VolumeSnippet<GV, LGV> > snippets_;
      //! \brief bounding box of the snippets of this cut cell
      RawBoundingBox rawBoundingBox_;
      //! \brief accumulated volumne of all snippets
      ctype volumeInFundamental_;
      //! \brief does the cut cell coincide with a complete fundamental mesh cell
      bool fillsFundamentalCell_;
      //! \brief minimal absolute value of the level-set function at corners
      ctype valueTolerance_;
      //! \brief center of mass in local coordinates of the fundamental mesh
      Coordinate centerOfMassInFundamental_;
    };
  }
}

#endif // DUNE_UDG_CUTCELL_HH
