#ifndef DUNE_UDG_EXTERNALCUTCELLINTERFACE_HH
#define DUNE_UDG_EXTERNALCUTCELLINTERFACE_HH

#include <dune/grid/common/scsgmapper.hh>
#include <dune/grid/common/rangegenerators.hh>
#include <dune/tpmc/tpmcrefinement.hh>
#include <dune/udg/common/geometryinterface.hh>
#include <dune/udg/simpletpmctriangulation/interface.hh>
#include <dune/udg/simpletpmctriangulation/interfacesnippet.hh>
#include <dune/udg/simpletpmctriangulation/cutcellinterfacebase.hh>

namespace Dune
{
  namespace UDG
  {
    namespace ExternalCutCellInterfaceDetail
    {
      template <class Guest, class Host>
      Host combineGeometries(const Guest& guest, const Host& host)
      {
        std::vector<typename Host::GlobalCoordinate> coords;
        coords.resize(guest.corners());
        for (unsigned int v = 0; v < coords.size(); ++v) {
          coords[v] = host.global(guest.corner(v));
        }
        return Host(guest.type(), coords);
      }

      template <class To, class From>
      To convertGeometry(const From& from)
      {
        std::vector<typename To::GlobalCoordinate> coords;
        coords.reserve(from.corners());
        for (int i = 0; i < from.corners(); ++i) {
          coords.push_back(from.corner(i));
        }
        return To(from.type(), coords);
      }
    }

    template <class GV, class LGV>
    class ExternalCutCellInterface
        : public CutCellInterfaceBase<GV, LGV, ExternalInterfaceSnippet<GV, LGV> >
    {
    public:
      using FundamentalElement = typename GV::template Codim<0>::Entity;
      using LevelSetElement = typename LGV::template Codim<0>::Entity;
      using FundamentalIntersection = typename GV::Intersection;
      using Intersection = typename GV::Intersection;
      using ctype = typename GV::ctype;
      enum { dim = GV::dimension };

      explicit ExternalCutCellInterface(const FundamentalElement& element,
                                        const GV& fundamentalGridView, const LGV& levelSetGridView,
                                        ctype valueTolerance = 1e-8)
          : fundamentalElement_(element), valueTolerance_(valueTolerance)
      {
        Dune::SingleCodimSingleGeomTypeMapper<GV, 0> elementMapper(fundamentalGridView);
        if (fundamentalElement_.isLeaf()) {
          for (const auto& intersection : intersections(fundamentalGridView, element)) {
            const auto& intersectionGeometry = intersection.geometry();
            using GG = typename ExternalInterfaceSnippet<GV, LGV>::GlobalGeometry;
            this->addSnippet(ExternalInterfaceSnippet<GV, LGV>(
                ExternalCutCellInterfaceDetail::convertGeometry<GG>(intersectionGeometry),
                intersection.unitOuterNormal(
                    ReferenceElements<ctype, dim - 1>::general(intersectionGeometry.type())
                        .position(0, 0)),
                elementMapper.index(intersection.neighbor() ? intersection.outside() : element),
                intersection.indexInInside(), std::make_shared<FundamentalElement>(element)));
          }
        } else {
          auto maxLevel = levelSetGridView.grid().maxLevel();
          // loop through all child elements in the level set mesh
          for (const auto& lelem : Dune::descendantElements(fundamentalElement_, maxLevel)) {
            if (!levelSetGridView.indexSet().contains(lelem))
              continue;

            // create copy of lelem to be passed to the snippets
            auto levelSetElement = std::make_shared<LevelSetElement>(lelem);

            for (const auto& intersection : intersections(levelSetGridView, lelem)) {
              FundamentalElement outsideFundamental = element;
              // check if the intersection is internal to this fundamental element
              if (intersection.neighbor()) {
                outsideFundamental = intersection.outside();
                while (!fundamentalGridView.indexSet().contains(outsideFundamental)) {
                  outsideFundamental = outsideFundamental.father();
                }
                if (outsideFundamental == element) {
                  continue;
                }
              }
              const auto& intersectionGeometry = intersection.geometry();
              FundamentalIntersection fundamentalIntersection;
              //bool foundIntersection = false;
              // find intersection in fundamental element
              if (outsideFundamental != element) {
                for (const auto& fintersection : intersections(fundamentalGridView, element)) {
                  if (fintersection.neighbor() && fintersection.outside() == outsideFundamental) {
                    fundamentalIntersection = fintersection;
                    //foundIntersection = true;
                    break;
                  }
                }
              } else { // meaning, no neighbor
                // find the intersection of the fundamental element which contains all corners of
                // the levelset intersection
                for (const auto& fintersection : intersections(fundamentalGridView, element)) {
                  const auto& fgeo = fintersection.geometry();
                  bool inside = true;
                  for (int i = 0; i < intersectionGeometry.corners(); ++i) {
                    auto diff = intersectionGeometry.corner(i);
                    diff -= fgeo.global(fgeo.local(intersectionGeometry.corner(i)));
                    if (diff.two_norm() > 1e-12) {
                      inside = false;
                      break;
                    }
                  }
                  if (inside) {
                    fundamentalIntersection = fintersection;
                    //foundIntersection = true;
                    break;
                  }
                }
              }
              //assert(foundIntersection);
              using GG = typename ExternalInterfaceSnippet<GV, LGV>::GlobalGeometry;
              this->addSnippet(ExternalInterfaceSnippet<GV, LGV>(
                  ExternalCutCellInterfaceDetail::convertGeometry<GG>(intersectionGeometry),
                  intersection.unitOuterNormal(
                      ReferenceElements<ctype, dim - 1>::general(intersectionGeometry.type())
                          .position(0, 0)),
                  elementMapper.index(outsideFundamental), fundamentalIntersection.indexInInside(),
                  levelSetElement));
            }
          }
        }
      }

      explicit ExternalCutCellInterface(const FundamentalElement& element,
                                        ctype valueTolerance = 1e-8)
          : fundamentalElement_(element), valueTolerance_(valueTolerance)
      {
      }

      template <class OutIterator>
      void cut(const Interface<LGV>& interface, OutIterator out) const
      {
        ExternalCutCellInterface interior(fundamentalElement_, valueTolerance_);
        ExternalCutCellInterface exterior(fundamentalElement_, valueTolerance_);

        interior.interfaceRelativePositions_ = interfaceRelativePositions_;
        interior.interfaceRelativePositions_.push_back(InterfaceRelativePosition::interior);
        exterior.interfaceRelativePositions_ = interfaceRelativePositions_;
        exterior.interfaceRelativePositions_.push_back(InterfaceRelativePosition::exterior);

        std::vector<ctype> values;

        auto localInterfaceFunction = localFunction(interface.function());

        for (const ExternalInterfaceSnippet<GV, LGV>& snippet : *this) {
          const auto& lgeo = snippet.levelSetElement()->geometry();
          localInterfaceFunction.bind(*snippet.levelSetElement());
          values.clear();
          values.reserve(snippet.globalGeometry().corners());
          // retrieve level set values at corners
          for (int i = 0; i < snippet.globalGeometry().corners(); ++i) {
            values.push_back(
                localInterfaceFunction(lgeo.local(snippet.globalGeometry().corner(i))));
            if (values.back() >= 0.0 && values.back() < valueTolerance_) {
              values.back() = valueTolerance_;
            } else if (values.back() > -valueTolerance_ && values.back() < 0.0) {
              values.back() = -valueTolerance_;
            }
          }

          // retrieve interior and exterior elements using tpmc
          TpmcRefinement<ctype, dim - 1> refinement(values.begin(), values.end());
          for (auto it = refinement.beginVolume(tpmc::InteriorDomain);
               it != refinement.endVolume(tpmc::InteriorDomain); ++it) {
            interior.addSnippet(ExternalInterfaceSnippet<GV, LGV>(
                ExternalCutCellInterfaceDetail::combineGeometries(*it, snippet.globalGeometry()),
                snippet.globalUnitOuterNormal(), snippet.outsideFundamentalElementIndex(),
                snippet.codim1IndexInInside(), snippet.levelSetElement()));
          }
          for (auto it = refinement.beginVolume(tpmc::ExteriorDomain);
               it != refinement.endVolume(tpmc::ExteriorDomain); ++it) {
            exterior.addSnippet(ExternalInterfaceSnippet<GV, LGV>(
                ExternalCutCellInterfaceDetail::combineGeometries(*it, snippet.globalGeometry()),
                snippet.globalUnitOuterNormal(), snippet.outsideFundamentalElementIndex(),
                snippet.codim1IndexInInside(), snippet.levelSetElement()));
          }
        }
        *out++ = interior;
        *out++ = exterior;
      }

      const std::vector<InterfaceRelativePosition>& interfaceRelativePositions() const
      {
        return interfaceRelativePositions_;
      }

    private:
      FundamentalElement fundamentalElement_;
      std::vector<InterfaceRelativePosition> interfaceRelativePositions_;
      ctype valueTolerance_;
    };
  }
}

#endif // DUNE_UDG_EXTERNALCUTCELLINTERFACE_HH
