#ifndef DUNE_UDG_INTERFACESNIPPET_HH
#define DUNE_UDG_INTERFACESNIPPET_HH

#include <dune/geometry/multilineargeometry.hh>

namespace Dune
{
  namespace UDG
  {
    template <class GV, class LGV>
    class InterfaceSnippetBase
    {
    public:
      using FundamentalGridView = GV;
      using LevelSetGridView = LGV;
      using LevelSetElement = typename LGV::template Codim<0>::Entity;
      using ctype = typename GV::ctype;
      enum { dim = GV::dimension };
      //! \brief geometry type of this snippet in the fundamental mesh cell
      using GlobalGeometry = Dune::CachedMultiLinearGeometry<ctype, dim - 1, dim,
          Dune::MultiLinearGeometryTraits<ctype> >;

      InterfaceSnippetBase(const GlobalGeometry& globalGeometry,
          const FieldVector<ctype, dim>& globalUnitOuterNormal,
          std::shared_ptr<LevelSetElement> levelSetElement)
          : globalGeometry_(globalGeometry)
          , globalUnitOuterNormal_(globalUnitOuterNormal)
          , levelSetElement_(levelSetElement)
      {
        assert(levelSetElement_);
      }

      InterfaceSnippetBase(Dune::GeometryType type,
          const std::vector<FieldVector<ctype, dim> >& corners,
          const FieldVector<ctype, dim>& globalUnitOuterNormal,
          std::shared_ptr<LevelSetElement> levelSetElement)
          : globalGeometry_(type, corners)
          , globalUnitOuterNormal_(globalUnitOuterNormal)
          , levelSetElement_(levelSetElement)
      {
        assert(levelSetElement_);
      }

      const GlobalGeometry& globalGeometry() const
      {
        return globalGeometry_;
      }

      const FieldVector<ctype, dim>& globalUnitOuterNormal() const
      {
        return globalUnitOuterNormal_;
      }

      std::shared_ptr<LevelSetElement> levelSetElement() const
      {
        return levelSetElement_;
      }

    private:
      GlobalGeometry globalGeometry_;
      FieldVector<ctype, dim> globalUnitOuterNormal_;
      std::shared_ptr<LevelSetElement> levelSetElement_;
    };

    template <class GV, class LGV>
    using InternalInterfaceSnippet = InterfaceSnippetBase<GV, LGV>;

    template <class GV, class LGV>
    class ExternalInterfaceSnippet : public InterfaceSnippetBase<GV, LGV>
    {
    public:
      using BaseT = InterfaceSnippetBase<GV, LGV>;
      using FundamentalElement = typename GV::template Codim<0>::Entity;
      using GlobalGeometry = typename BaseT::GlobalGeometry;
      using ctype = typename BaseT::ctype;
      enum { dim = BaseT::dim };
      using LevelSetElement = typename BaseT::LevelSetElement;

      ExternalInterfaceSnippet(const GlobalGeometry& globalGeometry,
          const FieldVector<ctype, dim>& globalUnitOuterNormal,
          std::size_t outsideFundamentalElementIndex, int codim1IndexInInside,
          std::shared_ptr<LevelSetElement> levelSetElement)
          : InterfaceSnippetBase<GV, LGV>(globalGeometry, globalUnitOuterNormal, levelSetElement)
          , outsideFundamentalElementIndex_(outsideFundamentalElementIndex)
          , codim1IndexInInside_(codim1IndexInInside)
      {
      }

      ExternalInterfaceSnippet(Dune::GeometryType type,
          const std::vector<FieldVector<ctype, dim> >& corners,
          const FieldVector<ctype, dim>& globalUnitOuterNormal,
          std::size_t outsideFundamentalElementIndex, int codim1IndexInInside,
          std::shared_ptr<LevelSetElement> levelSetElement)
          : InterfaceSnippetBase<GV, LGV>(type, corners, globalUnitOuterNormal, levelSetElement)
          , outsideFundamentalElementIndex_(outsideFundamentalElementIndex)
          , codim1IndexInInside_(codim1IndexInInside)
      {
      }

      std::size_t outsideFundamentalElementIndex() const
      {
        return outsideFundamentalElementIndex_;
      }

      int codim1IndexInInside() const
      {
        return codim1IndexInInside_;
      }

    private:
      std::size_t outsideFundamentalElementIndex_;
      int codim1IndexInInside_;
    };
  }
}

#endif // DUNE_UDG_INTERFACESNIPPET_HH
