#ifndef DUNE_UDG_INTERNALCUTCELLINTERFACE_HH
#define DUNE_UDG_INTERNALCUTCELLINTERFACE_HH

#include <dune/geometry/quadraturerules.hh>
#include <dune/udg/simpletpmctriangulation/interface.hh>
#include <dune/udg/simpletpmctriangulation/cutcellinterfacebase.hh>
#include <dune/udg/simpletpmctriangulation/interfacesnippet.hh>
#include <dune/tpmc/tpmcrefinement.hh>

namespace Dune
{
  namespace UDG
  {
    namespace InternalCutCellInterfaceDetail
    {
      template <class Guest, class Host>
      Host combineGeometries(const Guest& guest, const Host& host)
      {
        std::vector<typename Host::GlobalCoordinate> coords;
        coords.resize(guest.corners());
        for (unsigned int v = 0; v < coords.size(); ++v) {
          coords[v] = host.global(guest.corner(v));
        }
        return Host(guest.type(), coords);
      }

      template <class ctype>
      FieldVector<ctype, 3> crossProduct(
          const FieldVector<ctype, 3>& a, const FieldVector<ctype, 3>& b)
      {
        return { a[1] * b[2] - a[2] * b[1], a[2] * b[0] - a[0] * b[2], a[0] * b[1] - a[1] * b[0] };
      }

      template <class ctype>
      FieldVector<ctype, 3>
      computeUnitOuterNormal(const std::vector<FieldVector<ctype, 3>>& corners,
                             const std::vector<ctype>& values)
      {
        // move corner 0 to the origin
        auto a = corners[1];
        auto b = corners[2];
        a -= corners[0];
        b -= corners[0];
        // compute normal to corner 1 and 2
        auto cp = crossProduct(a, b);
        // normalize
        cp /= cp.two_norm();
        return cp;
      }

      template <class ctype>
      FieldVector<ctype, 2>
      computeUnitOuterNormal(const std::vector<FieldVector<ctype, 2>>& corners,
                             const std::vector<ctype>& values)
      {
        // move corner 0 to the origin
        auto diff = corners[0];
        diff -= corners[1];
        std::swap(diff[0], diff[1]);
        diff[0] *= -1;
        diff /= diff.two_norm();
        return diff;
      }

      template <class ctype>
      FieldVector<ctype, 1>
      computeUnitOuterNormal(const std::vector<FieldVector<ctype, 1>>& corners,
                             const std::vector<ctype>& values)
      {
        assert(values.size() == 2);
        if (values[0] < values[1])
          return {1};
        else
          return {-1};
      }
    }

    template <class GV, class LGV>
    class InternalCutCellInterface
        : public CutCellInterfaceBase<GV, LGV, InternalInterfaceSnippet<GV, LGV> >
    {
    public:
      using FundamentalElement = typename GV::template Codim<0>::Entity;
      using LevelSetElement = typename LGV::template Codim<0>::Entity;
      using ctype = typename GV::ctype;
      enum { dim = GV::dimension };

      InternalCutCellInterface(const FundamentalElement& fundamentalElement,
                               std::shared_ptr<LevelSetElement> levelSetElement,
                               const Interface<LGV>& interface, ctype valueTolerance = 1e-8)
          : fundamentalElement_(fundamentalElement)
          , initialInterface_(interface)
          , interfaceRelativePositions_(interface.index() + 1, InterfaceRelativePosition::any)
          , globalArea_(0.0)
          , valueTolerance_(valueTolerance)
      {
        std::vector<ctype> values;
        const auto& lgeo = levelSetElement->geometry();

        auto localInterfaceFunction = localFunction(interface.function());
        localInterfaceFunction.bind(*levelSetElement);
        const auto& lref = Dune::ReferenceElements<ctype, dim>::general(lgeo.type());
        for (int i = 0; i < lgeo.corners(); ++i) {
          values.push_back(localInterfaceFunction(lref.position(i, dim)));
          if (values.back() >= 0.0 && values.back() < valueTolerance_) {
            values.back() = valueTolerance_;
          } else if (values.back() > -valueTolerance_ && values.back() < 0.0) {
            values.back() = -valueTolerance_;
          }
        }

        TpmcRefinement<ctype, dim> refinement(values.begin(), values.end());
        for (const auto& i : refinement.interface()) {
          std::vector<FieldVector<ctype, dim> > corners;
          for (int c = 0; c < i.corners(); ++c) {
            corners.push_back(lgeo.global(i.corner(c)));
          }
          auto snippet = InternalInterfaceSnippet<GV, LGV>(
              i.type(), corners,
              InternalCutCellInterfaceDetail::computeUnitOuterNormal(corners, values),
              levelSetElement);
          const auto& rule
              = Dune::QuadratureRules<ctype, dim - 1>::rule(snippet.globalGeometry().type(), 1);
          for (const auto& qp : rule) {
            globalArea_ += qp.weight() * snippet.globalGeometry().integrationElement(qp.position());
          }
          this->addSnippet(snippet);
        }

        interfaceRelativePositions_[interface.index()] = InterfaceRelativePosition::interface;
      }

      explicit InternalCutCellInterface(const FundamentalElement& fundamentalElement,
                                        const Interface<LGV>& interface,
                                        ctype valueTolerance = 1e-8)
          : fundamentalElement_(fundamentalElement)
          , initialInterface_(interface)
          , globalArea_(0.0)
          , valueTolerance_(valueTolerance)
      {
      }

      template <class OutIterator>
      void cut(const Interface<LGV>& newInterface, OutIterator out) const
      {
        InternalCutCellInterface interior(fundamentalElement_, initialInterface_, valueTolerance_);
        InternalCutCellInterface exterior(fundamentalElement_, initialInterface_, valueTolerance_);

        // initialize interior and exteriors interface relative positions
        interior.interfaceRelativePositions_ = interfaceRelativePositions_;
        if (interior.interfaceRelativePositions_.size() < newInterface.index() + 1)
          interior.interfaceRelativePositions_.resize(newInterface.index() + 1,
                                                      InterfaceRelativePosition::any);
        interior.interfaceRelativePositions_[newInterface.index()] =
            InterfaceRelativePosition::interior;
        interior.globalArea_ = globalArea_;
        exterior.interfaceRelativePositions_ = interfaceRelativePositions_;
        if (exterior.interfaceRelativePositions_.size() < newInterface.index() + 1)
          exterior.interfaceRelativePositions_.resize(newInterface.index() + 1,
                                                      InterfaceRelativePosition::any);
        exterior.interfaceRelativePositions_[newInterface.index()] =
            InterfaceRelativePosition::exterior;
        exterior.globalArea_ = globalArea_;

        auto localInterfaceFunction = localFunction(newInterface.function());

        for (const InternalInterfaceSnippet<GV, LGV>& snippet : *this) {
          const auto& lgeo = snippet.levelSetElement()->geometry();
          localInterfaceFunction.bind(*snippet.levelSetElement());
          // retrieve level set values at corners
          std::vector<ctype> values;
          for (int i = 0; i < snippet.globalGeometry().corners(); ++i) {
            values.push_back(
                localInterfaceFunction(lgeo.local(snippet.globalGeometry().corner(i))));
            if (values.back() >= 0.0 && values.back() < valueTolerance_) {
              values.back() = valueTolerance_;
            } else if (values.back() > -valueTolerance_ && values.back() < 0.0) {
              values.back() = -valueTolerance_;
            }
          }

          // retrieve interior and exterior elements using tpmc
          TpmcRefinement<ctype, dim - 1> refinement(values.begin(), values.end());
          for (auto it = refinement.beginVolume(tpmc::InteriorDomain);
               it != refinement.endVolume(tpmc::InteriorDomain); ++it) {
            interior.addSnippet(InternalInterfaceSnippet<GV, LGV>(
                InternalCutCellInterfaceDetail::combineGeometries(*it, snippet.globalGeometry()),
                snippet.globalUnitOuterNormal(), snippet.levelSetElement()));
          }
          for (auto it = refinement.beginVolume(tpmc::ExteriorDomain);
               it != refinement.endVolume(tpmc::ExteriorDomain); ++it) {
            exterior.addSnippet(InternalInterfaceSnippet<GV, LGV>(
                InternalCutCellInterfaceDetail::combineGeometries(*it, snippet.globalGeometry()),
                snippet.globalUnitOuterNormal(), snippet.levelSetElement()));
          }
        }
        if (!interior.empty())
          *out++ = std::move(interior);
        if (!exterior.empty())
          *out++ = std::move(exterior);
        // cut interface into all snippets
      }

      const Interface<LGV>& interface() const
      {
        return initialInterface_;
      }

      const std::vector<InterfaceRelativePosition>& interfaceRelativePositions() const
      {
        return interfaceRelativePositions_;
      }

      ctype globalArea() const
      {
        return globalArea_;
      }

    private:
      //! \brief the element of the fundamental mesh this cut cell belongs to
      FundamentalElement const fundamentalElement_;

      //! \brief interface this class originally belonged to
      Interface<LGV> initialInterface_;

      std::vector<InterfaceRelativePosition> interfaceRelativePositions_;

      ctype globalArea_;

      ctype valueTolerance_;
    };
  }
}
#endif // DUNE_UDG_INTERNALCUTCELLINTERFACE_HH
