#ifndef DUNE_UDG_INTERNALINTERFACEAREAMAP_HH
#define DUNE_UDG_INTERNALINTERFACEAREAMAP_HH

#include <dune/geometry/referenceelements.hh>
#include <dune/udg/simpletpmctriangulation/domain.hh>
#include <dune/udg/simpletpmctriangulation/internalinterfacesubtriangulation.hh>

namespace Dune
{
  namespace UDG
  {
    //! \brief class that maps a pair of domains to the union of all areas separating those
    //  domains
    template <class GV, class LGV>
    class InternalInterfaceAreaMap
    {
    public:
      enum { dim = GV::dimension };
      using ctype = typename GV::ctype;

      InternalInterfaceAreaMap(const InternalInterfaceSubTriangulation<GV, LGV>& subTriangulation,
          const DomainConfiguration<GV, LGV>& domainConfiguration)
          : numberOfDomains_(domainConfiguration.size())
          , areas_(domainConfiguration.size() + domainConfiguration.size() * domainConfiguration.size(), 0.0)
      {
        for (const InternalCutCellInterface<GV, LGV>& ii :
            subTriangulation.internalCutCellInterfaces()) {
          auto interiorDomain
              = domainConfiguration.findInteriorDomain(ii.interfaceRelativePositions());
          auto exteriorDomain
              = domainConfiguration.findExteriorDomain(ii.interfaceRelativePositions());
          if (interiorDomain == domainConfiguration.domainsEnd()) {
            if (exteriorDomain != domainConfiguration.domainsEnd()) {
              area(*exteriorDomain, area(*exteriorDomain) + ii.globalArea());
            }
          } else {
            if (exteriorDomain == domainConfiguration.domainsEnd()) {
              area(*interiorDomain, area(*interiorDomain) + ii.globalArea());
            } else {
              area(*interiorDomain, *exteriorDomain,
                  area(*interiorDomain, *exteriorDomain) + ii.globalArea());
            }
          }
        }
      }

      ctype area(const Domain& interiorDomain, const Domain& exteriorDomain) const
      {
        auto index = linear(interiorDomain.index(), exteriorDomain.index());
        if (index > areas_.size() - 1) {
          std::cout << "index = " << index << " interiorDomain = " << interiorDomain.index() << " exteriorDomain = " << exteriorDomain.index() << std::endl;
          DUNE_THROW(Dune::Exception, "index " << index << " not in map");
        }
        return areas_[index];
      }

      ctype area(const Domain& interiorDomain) const
      {
        if (interiorDomain.index() > areas_.size() - 1) {
          DUNE_THROW(Dune::Exception, "domain " << interiorDomain.index() << " not in map");
        }
        return areas_[interiorDomain.index()];
      }

      void area(const Domain& interiorDomain, ctype v)
      {
        if (interiorDomain.index() > areas_.size() - 1) {
          DUNE_THROW(Dune::Exception, "domain " << interiorDomain.index() << " not in map");
        }
        areas_[interiorDomain.index()] = v;
      }

      void area(const Domain& interiorDomain, const Domain& exteriorDomain, ctype v)
      {
        auto index = linear(interiorDomain.index(), exteriorDomain.index());
        if (index > areas_.size() - 1) {
          DUNE_THROW(Dune::Exception, "index " << index << " not in map");
        }
        areas_[index] = v;
      }

    private:
      std::size_t numberOfDomains_;
      std::vector<ctype> areas_;

      std::size_t linear(std::size_t interior, std::size_t exterior) const
      {
        if (exterior < interior)
          std::swap(interior, exterior);
        return interior * numberOfDomains_ + exterior + numberOfDomains_;
      }
    };
  }
}

#endif // DUNE_UDG_INTERNALINTERFACEAREAMAP_HH
