#ifndef DUNE_UDG_VOLUMESUBTRIANGULATION
#define DUNE_UDG_VOLUMESUBTRIANGULATION

#include <dune/udg/mc33triangulation/domaincell.hh>
#include <dune/udg/simpletpmctriangulation/cutcell.hh>
#include <dune/udg/simpletpmctriangulation/domainconfiguration.hh>

namespace Dune
{
  namespace UDG
  {
    namespace VolumeSubTriangulationDetail
    {
      template <class GOut, class G>
      GOut createIdentityGeometry(const G& geometry)
      {
        using ctype = typename G::ctype;
        const auto& referenceElement =
            ReferenceElements<ctype, G::mydimension>::general(geometry.type());
        std::vector<FieldVector<ctype, G::mydimension>> corners;
        corners.reserve(referenceElement.size(G::mydimension));
        for (int i = 0; i < referenceElement.size(G::mydimension); ++i) {
          corners.push_back(referenceElement.position(i, G::mydimension));
        }
        return GOut(geometry.type(), corners);
      }
    }
    /**
     * @brief this class represents a sub triangulation of the volume of a given element
     *
     * @note we do not create a triangulation of the interface.
     *
     * @tparam GV type of the grid view of the fundamental mesh
     * @tparam LGV type of the grid view of the level set mesh
     */
    template <class GV, class LGV>
    class VolumeSubTriangulation
    {
    public:
      //! @brief field type of coordinates
      using ctype = typename GV::ctype;
      enum {
        //! @brief dimension of the given grid
        dim = GV::dimension
      };
      //! @brief raw bounding box type
      using RawBoundingBox = MarchingCube33SubTriangulationImp::RawBoundingBox<ctype, dim>;
      //! @brief type of a bounding box for a given domain
      using BoundingBox = Dune::GridCuboid<typename GV::Grid>;
      //! @brief type of elements of the fundamental mesh
      using FundamentalElement = typename GV::template Codim<0>::Entity;
      //! @brief type of coordinates
      using Coordinate = Dune::FieldVector<ctype, dim>;

      /**
       * @brief create a volume sub triangulation for the given element
       * @param levelSetGridView grid view of the level set mesh
       * @param domainConfiguration configuration describing the domains and interfaces
       * @param e element in the fundamental mesh for which the subtriangulation should be created
       */
      VolumeSubTriangulation(const LGV& levelSetGridView,
                             const DomainConfiguration<GV, LGV>& domainConfiguration,
                             FundamentalElement e, bool forceRefinement = false,
                             ctype valueTolerance = 1e-8)
          : fundamentalElement_(e)
          , domainToCutCells_(domainConfiguration.numberOfDomains())
          , domainToVolume_(domainConfiguration.numberOfDomains(), 0.0)
          , domainToBoundingBox_(domainConfiguration.numberOfDomains())
          , domainToCenterOfMass_(domainConfiguration.numberOfDomains())
      {
        auto fundamentalGeometry = e.geometry();
        auto volumeScaling = fundamentalGeometry.integrationElement(
            ReferenceElements<ctype, dim>::general(fundamentalGeometry.type()).position(0, 0));
        // create initial cut cell cc from all lgv subcells of e
        std::vector<CutCell<GV, LGV>> cutcells = {{e, levelSetGridView, valueTolerance}};
        // cut all interfaces into cut cells recursively
        for (const Interface<LGV>& interface : domainConfiguration.interfaces()) {
          std::vector<CutCell<GV, LGV>> newCutCells;
          for (const CutCell<GV, LGV>& cutcell : cutcells) {
            cutcell.cut(interface, std::back_inserter(newCutCells));
          }
          cutcells.swap(newCutCells);
        }
        // accumulate volumes of all cut cells belonging to the same domain
        for (CutCell<GV, LGV>& cutcell : cutcells) {
          auto domainIt = domainConfiguration.findDomain(cutcell.interfaceRelativePositions());
          if (!forceRefinement && cutcell.fillsFundamentalCell()) {
            cutcell.clearSnippets();
            cutcell.addSnippet(
                VolumeSubTriangulationDetail::createIdentityGeometry<
                    typename VolumeSnippet<GV, LGV>::GeometryInFundamental>(fundamentalGeometry),
                std::make_shared<FundamentalElement>(e));
          }
          if (domainIt != domainConfiguration.domainsEnd()) {
            // note: assuming the mapping from fundamental to global is linear, so that the
            // integrationElement is constant
            domainToVolume_[domainIt->index()] += cutcell.volumeInFundamental() * volumeScaling;
            domainToBoundingBox_[domainIt->index()].update(cutcell.rawBoundingBox());
            domainToCutCells_[domainIt->index()].push_back(std::move(cutcell));
            domainToCenterOfMass_[domainIt->index()].axpy(
                cutcell.volumeInFundamental() * volumeScaling,
                fundamentalGeometry.global(cutcell.centerOfMassInFundamental()));
          }
        }
        // rescale center of mass by complete domain volume
        for (std::size_t i = 0; i < domainConfiguration.numberOfDomains(); ++i) {
          if (!domainToCutCells_[i].empty()) {
            domainToCenterOfMass_[i] /= domainToVolume_[i];
          }
        }
      }

      //! @brief create the bounding box for the union of all cut cells with the given domain index
      BoundingBox createBoundingBox(std::size_t domainIndex) const
      {
        return BoundingBox(fundamentalElement_, domainToBoundingBox_[domainIndex].lower,
                           domainToBoundingBox_[domainIndex].higher, CuboidMode::corners);
      }

      //! @brief return the volume of the union of all cut cells with the given domain index
      ctype volume(std::size_t domainIndex) const
      {
        return domainToVolume_[domainIndex];
      }

      //! @brief return all cut cells belonging to the domain with the given index
      const std::vector<CutCell<GV, LGV>>& cutCells(std::size_t domainIndex) const
      {
        return domainToCutCells_[domainIndex];
      }

      const Coordinate& centerOfMass(std::size_t domainIndex) const
      {
        return domainToCenterOfMass_[domainIndex];
      }

      const FundamentalElement& fundamentalElement() const
      {
        return fundamentalElement_;
      }

      bool fillsFundamentalCell(std::size_t domainIndex) const
      {
        return domainToCutCells_[domainIndex].size() == 1
               && domainToCutCells_[domainIndex][0].fillsFundamentalCell();
      }

      std::size_t numberOfSnippets(std::size_t domainIndex) const
      {
        std::size_t result = 0;
        for (const auto& cc : cutCells(domainIndex)) {
          result += cc.size();
        }
        return result;
      }

    private:
      FundamentalElement fundamentalElement_;

      std::vector<std::vector<CutCell<GV, LGV>>> domainToCutCells_;
      std::vector<ctype> domainToVolume_;
      std::vector<RawBoundingBox> domainToBoundingBox_;
      std::vector<Coordinate> domainToCenterOfMass_;
    };
  }
}
#endif // DUNE_UDG_VOLUMESUBTRIANGULATION
