// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#include <cassert>

#include <dune/grid/common/mcmgmapper.hh>

#include "sparsitysetup.hh"

namespace Dune {

  template<int dim>
  struct CellLayout
  {
    bool contains (GeometryType gt)
    {
      return gt.dim()==dim;
    }
  };

  template<typename Matrix, typename Vector, typename GridView>
  void setupNeighborSparsity(Matrix &A, Vector &b, const GridView &gridview)
  {
    typedef MultipleCodimMultipleGeomTypeMapper< GridView, CellLayout >
    Mapper;

    Mapper mapper(gridview);

    setupNeighborSparsity(A, b, gridview, mapper);
  }

  template<typename Matrix, typename Vector, typename GridView, typename Mapper>
  void setupNeighborSparsity(Matrix &A, Vector &b, const GridView &gridview, const Mapper & mapper)
  {
    //typedefs
    typedef typename GridView::Grid Grid;
    typedef typename GridView::IndexSet IndexSet;
    typedef typename GridView::template Codim<0>::Iterator ElementIterator;
    typedef typename GridView::IntersectionIterator IntersectionIterator;

    int N = mapper.size();
    A.setSize(N,N);
    A.setBuildMode(Matrix::row_wise);
    typename Matrix::CreateIterator i=A.createbegin();
    // build up the matrix structure
    ElementIterator eit = gridview.template begin<0>();
    ElementIterator eitend = gridview.template end<0>();
    for (; eit != eitend; ++eit) {
      assert(i != A.createend());

      // insert a non zero entry for myself
      int eid = mapper.map(*eit);
      i.insert(eid);

      // insert a non zero entry for each neighbour
      IntersectionIterator endit = gridview.iend(*eit);
      IntersectionIterator it = gridview.ibegin(*eit);
      for(; it != endit; ++it) {
        if(it->neighbor())
        {
          int fid = mapper.map(*it->outside());
          assert (fid < N);
          i.insert(fid);
        }
      }

      // move CreateIterator to next row
      ++i;
    }
    // make sure we are at the end
    assert(i == A.createend());
    A = 0.0;

    b.resize(N, false);
    b = 0.0;
  }

  template<typename Matrix, typename Vector, typename GridView>
  void setupDiagonalSparsity(Matrix &A, Vector &b, const GridView &gridview)
  {
    typedef MultipleCodimMultipleGeomTypeMapper< GridView, CellLayout >
    Mapper;

    Mapper mapper(gridview);

    setupDiagonalSparsity(A, b, gridview, mapper);
  }

  template<typename Matrix, typename Vector, typename GridView, typename Mapper>
  void setupDiagonalSparsity(Matrix &A, Vector &b, const GridView &gridview, const Mapper &mapper)
  {
    //typedefs
    typedef typename GridView::Grid Grid;
    typedef typename GridView::IndexSet IndexSet;
    typedef typename GridView::template Codim<0>::Iterator ElementIterator;
    typedef typename GridView::IntersectionIterator IntersectionIterator;

    int N = mapper.size();

    A.setBuildMode(Matrix::row_wise);
    A.setSize(N, N, N);

    // build up the matrix structure
    typename Matrix::CreateIterator ci = A.createbegin();
    for(int i = 0; i < N; ++i) {
      assert(ci != A.createend());
      ci.insert(i);
      ++ci;
    }
    A = 0.0;

    b.resize(N, false);
    b = 0.0;
  }

} // end namespace Dune
