#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <dune/common/gmpfield.hh>
#include <dune/common/parallel/mpihelper.hh>

#include <dune/grid/common/rangegenerators.hh>
#include <dune/grid/yaspgrid.hh>

#include <dune/functions/gridfunctions/gridviewfunction.hh>

#include <dune/pdelab/boilerplate/pdelab.hh>
#include <dune/pdelab/finiteelement/l2orthonormal.hh>
#include <dune/pdelab/ordering/chunkedblockordering.hh>

#include <dune/udg/io/refinedvtkwriter.hh>
#include <dune/udg/pdelab/finiteelementmap.hh>
#include <dune/udg/pdelab/multiphaseoperator.hh>
#include <dune/udg/pdelab/subtriangulation.hh>
#include <dune/udg/simpletpmctriangulation.hh>
#include <dune/udg/test/common/levelsets.hh>
#include <dune/udg/vtriangulation.hh>

#include <iostream>

int main(int argc, char** argv)
{
  Dune::MPIHelper::instance(argc, argv);
  try {
    const int dim = 1;
    const unsigned int N = 3;
    using Grid = Dune::YaspGrid<dim>;
    using ctype = Grid::ctype;
    using GV = Grid::LevelGridView;
    using LGV = Grid::LeafGridView;
    using STT = Dune::UDG::SimpleTpmcTriangulation<GV, LGV>;

    Dune::Timer timer;
    // create fundamental and level set mesh
    Dune::PDELab::StructuredGrid<Grid> grid(Dune::GeometryType::cube, N);
    GV gv = grid->levelGridView(0);
    LGV lgv = grid->leafGridView();
    std::cout << "constructed structured grid: " << timer.elapsed() << "s\n";
    timer.reset();

    // create domains
    Dune::UDG::DomainConfiguration<GV, LGV> domainConfiguration;
    domainConfiguration.addDomain(Dune::UDG::Domain(0, {"i"}));
    domainConfiguration.addDomain(Dune::UDG::Domain(1, {"e"}));

    // create level sets
    domainConfiguration.addInterface(Dune::UDG::Interface<LGV>(
        0, Dune::Functions::makeGridViewFunction([](const auto& x) { return x - .5; }, lgv)));
    std::cout << "constructed domainConfiguration: " << timer.elapsed() << "s\n";
    timer.reset();

    // construct subtriangulation
    STT triangulation(gv, lgv, domainConfiguration);
    std::cout << "constructed subTriangulation: " << timer.elapsed() << "s\n";
    timer.reset();

    // output subtriangulation
    Dune::RefinedVtkWriter<GV, STT, double> vtkVolume(gv, triangulation);
    typedef Dune::UDG::DomainIndexUnfittedVTKGridFunction<GV> DVGF;
    vtkVolume.addVertexData(new DVGF(gv, true));
    vtkVolume.addVertexData(new DVGF(gv, false));
    vtkVolume.write("test_1d_volume", Dune::VTK::appendedraw, 255, Dune::writeVolume);
    std::cout << "wrote volume vtk output " << timer.elapsed() << "s\n";
    timer.reset();

    Dune::SingleCodimSingleGeomTypeMapper<GV, 0> elementMapper(gv);
    for (const auto& element : Dune::elements(gv)) {
      typename STT::EntityPartList entityParts;
      triangulation.create_entity_parts(element, entityParts);
      typename STT::IntersectionPartList intersectionParts;
      triangulation.create_intersection_parts(element, intersectionParts);
      std::cout << "element " << elementMapper.index(element) << " entity_parts "
                << entityParts.size() << " intersection_parts " << intersectionParts.size()
                << std::endl;
      for (const auto& it : entityParts) {
        std::cout << "entity part center: " << it->geometry().center() << std::endl;
      }
      for (const auto& it : intersectionParts) {
        std::cout << "intersection part center: " << it->geometry().center() << std::endl;
      }
    }
  } catch (Dune::Exception& ex) {
    std::cerr << ex.what() << "\n";
    return -1;
  }
}
