// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <array>
#include <iostream>
#include <sstream>

#include <dune/common/math.hh>

#include <dune/functions/gridfunctions/analyticgridviewfunction.hh>

#include <dune/grid/yaspgrid.hh>

#include <dune/udg/common/quadraturerules.hh>
#include <dune/udg/simpletpmctriangulation.hh>
#include <dune/udg/vtriangulation.hh>

#define DO_PI_TEST
#define DO_SINGLE_CUT_TEST

template<typename GV, typename Fnkt>
double runTest(GV& gv, Fnkt lset) {
  constexpr int intorder = 1;

  using DomainCFG = Dune::UDG::DomainConfiguration<GV,GV>;
  using VST = VirtualSubTriangulation<GV>;

  auto levelset = Dune::Functions::makeAnalyticGridViewFunction(lset, gv);

  //create cut domain
  DomainCFG dom;
  dom.addDomain(Dune::UDG::Domain(0, {"i"}));
  dom.addInterface(Dune::UDG::Interface<GV>(0, levelset));

  //create subtriangulation
  using ST = Dune::UDG::SimpleTpmcTriangulation<GV,GV>;
  ST subtriangulation(gv,gv,dom);

  //using Rule = Dune::VSTQuadratureRule<ST>;
  double result(0.0);
  // loop over cells
  for(const auto& cell : elements(gv)) {
    auto geo = cell.geometry();
    //auto rule = Rule(subtriangulation,cell,intorder);
    auto rule = Dune::UDG::quadratureRule(subtriangulation,cell,intorder);
    for(const auto& ip : rule) {
      result += geo.integrationElement(ip.position())*ip.weight(); //integrate constant function f(x)=1
    }
  }
  return result;
}

int main(int argc, char** argv) {
  constexpr int dim = 2;

  Dune::MPIHelper::instance(argc, argv);

#ifdef DO_PI_TEST
{
  std::cout << " ---- Try calculating pi ---- " << std::endl;

  // level set
  Dune::FieldVector<double,dim> center{1.5,1.5};
  double R = 1.0;
  auto lset = [&center,&R](auto x)
  {
    return (x - center).two_norm() - R;
  };

  // set up yasp-grid
  int N = 10;
  Dune::FieldVector<double,dim> Len(3.0);
  std::array<int,dim> s;
  std::fill(s.begin(), s.end(), N);
  Dune::YaspGrid<dim> grid(Len,s);

  double pi = Dune::MathematicalConstants<double>::pi();
  std::stringstream output;
  double result(0.0), error1(0.0), error0(0.0);
  const int maxIter = 6;
  int i=0;

  do {
    grid.globalRefine(1);
    auto lgv = grid.leafGridView();

    result = runTest(lgv,lset);
    error0 = std::abs(result-pi);
    output << "\n" << i+1 << ": error=" << error0;
    if(i>0)
      output << ", rate=" << std::abs(error0/error1);
    if(i>1)
      output << ", EOC=" << std::log(error1/error0)/std::log(2.0);
    output << "\n";
    error1 = error0;
    ++i;
  }
  while(i<maxIter && error0 > 1e-10);

  //output
  std::cout << output.str() << std::endl;
}
#endif
#undef DO_PI_TEST

#ifdef DO_SINGLE_CUT_TEST
{
  // set up yasp-grid
  int N = 2;
  Dune::FieldVector<double,dim> Len(1.0);
  std::array<int,dim> s;
  std::fill(s.begin(), s.end(), N);
  Dune::YaspGrid<dim> grid(Len,s);

  auto lgv = grid.leafGridView();

  // horizontal cut along {x=0.4} => area should be 0.4*1=0.4
  auto lset_horizontal = [](auto x)
  {
    return x[0] - 0.4;
  };
    // vertical cut along {x=0.25*y+0.25} => area should be (0.75+0.5)/2*1=5/8
  auto lset_vertical = [](auto x)
  {
    return 0.25*(x[1]+1) - x[0];
  };
  //cut the lower left corner => area should be 1-(0.5*0.4*0.4)=0.92
  auto lset_corner = [](auto x)
  {
    return 0.4 - x.one_norm() ;
  };
  std::array<double,3> result;
  result[0] = runTest(lgv,lset_horizontal);
  result[1] = runTest(lgv,lset_vertical);
  result[2] = runTest(lgv,lset_corner);

  std::array<double,3> error;
  error[0] = std::abs(result[0]-0.4);
  error[1] = std::abs(result[1]-0.625);
  error[2] = std::abs(result[2]-0.92);

  // output
  std::cout << "\n";
  std::cout << "error (horizontal cut): " << error[0] << std::endl;
  std::cout << "error (vertical cut): " << error[1] << std::endl;
  std::cout << "error (corner cut): " << error[2] << std::endl;
  for(auto e : error)
    if(e > 1e-8)
      std::cerr << "test failed: No exact integration!" << std::endl;
}
#endif
#undef DO_SINGLE_CUT_TEST
}
