#if HAVE_CONFIG_H
#include <config.h>
#endif

#include "gtest/gtest.h"
#include <dune/common/parallel/mpihelper.hh>
#include <dune/common/fvector.hh>
#include <dune/common/parametertreeparser.hh>
#include <dune/grid/common/scsgmapper.hh>
#include <dune/pdelab/common/function.hh>
#include <dune/pdelab/boilerplate/pdelab.hh>
#include <dune/udg/io/refinedvtkwriter.hh>
#include <dune/udg/io/vtkfunction.hh>

#include <dune/udg/simpletpmctriangulation.hh>

#include <dune/udg/test/common/levelsets.hh>
#include <dune/udg/test/common/domains.hh>
#include <dune/udg/test/utility/vtk.hh>

template <class ctype,int dim>
void runTest(const Dune::ParameterTree& param)
{
  // create grid
  const Dune::GeometryType::BasicType basicType = Dune::GeometryType::cube;
  typedef Dune::YaspGrid<dim> G;
  Dune::PDELab::StructuredGrid<G> grid(basicType,1);
  grid->globalRefine(param.get<unsigned int>("refinements"));
  typedef typename G::LeafGridView GV;
  GV gv = grid->leafGridView();

  // create cross domain
  using Range = ctype;
  using Domain = Dune::FieldVector<ctype, dim>;
  using LevelSet = Dune::Functions::GridViewFunction<Range(Domain), GV>;
  std::vector<LevelSet> levelSets;
  Dune::UDGTest::constructRotatedCrossLevelSets(
      gv, param.get<Domain>("height"), param.get<double>("phi"), std::back_inserter(levelSets));
  Dune::UDGTest::CrossDomainFactory<GV,LevelSet> factory;
  typedef Dune::UDG::DomainConfiguration<GV,GV> DC;
  DC dc = factory.construct(gv, levelSets);

  // create subtriangulation
  using ST = Dune::UDG::SimpleTpmcTriangulation<GV,GV>;
  ST subTriangulation(gv, gv, dc);

  std::vector<std::pair<std::vector<Dune::FieldVector<ctype,dim> >,int> > exteriorBoundaryParts;
  // loop over codim 0 elements
  for (typename GV::template Codim<0>::Iterator eit = gv.template begin<0>();
       eit != gv.template end<0>(); ++eit)
  {
    // create intersection parts
    typename ST::IntersectionPartList ips;
    subTriangulation.create_intersection_parts(*eit, ips);
    // loop over all intersection parts
    for (typename ST::IntersectionPartList::const_iterator it = ips.begin();
         it != ips.end(); ++it)
    {
      if ((*it)->outsideDomainIndex() == -1)
      {
        // since the domain 0 has neighboring domains on all sides it should
        // not occur at the inside of an intersecion with the exterior boundary
        EXPECT_NE((*it)->insideDomainIndex(),0) << "insideDomainIndex of an exterior boundary intersection should "
                                                << "be anything but 0 (entity " << gv.indexSet().index(*eit) << ")";
        // store corners along with their inside index for vtk output
        std::vector<Dune::FieldVector<ctype,dim> > corners;
        for (int i = 0; i<(*it)->geometry().corners(); ++i)
        {
          corners.push_back((*it)->geometry().corner(i));
        }
        int vtkIndex = (*it)->insideDomainIndex();
        exteriorBoundaryParts.push_back(std::make_pair(corners,vtkIndex));
      }
    }
  }
  if (param.get<bool>("enableVTK")) {
    Dune::UDGTest::writeVTK(param.get<std::string>("vtkname"),
                            "exterior boundary intersections",
                            "insideDomain",
                            exteriorBoundaryParts);
  }
}

/**
 * \brief In the case of a domain which is completely enclosed by other domains
 *        the outsideBoundaryIndex of an intersection of this domain should never
 *        be -1.
 */
TEST(DomainBoundaryIndexTest, NestedDomain)
{
  const int dim = 3;
  /*const bool enableVTK = true;
  double testHeights[][3] = {{0.37,0.52,0.52}, {0.3,0.4,0.5}, {0.4,0.5,0.6}};
  double testPhis[] =       {M_PI/6.0        , M_PI/6.0     , M_PI/4.0};
  unsigned int testRefs[] = {4               , 4            , 3};
  for (int i = 0; i<sizeof(testPhis)/sizeof(testPhis[0]); ++i)
  {
    Dune::FieldVector<double,dim> height;
    std::copy(testHeights[i],testHeights[i]+dim, height.begin());
    runTest(height, testPhis[i],testRefs[i], enableVTK);
  }*/
  Dune::ParameterTree param;
  Dune::ParameterTreeParser::readINITree("test_domainboundaryindices.ini",param);
  runTest<double,dim>(param);
}

int main(int argc, char** argv)
{
  Dune::MPIHelper::instance(argc,argv);
  testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}
