#if HAVE_CONFIG_H
#include <config.h>
#endif

#include "gtest/gtest.h"
#include <dune/common/fvector.hh>
#include <dune/grid/common/scsgmapper.hh>
#include <dune/pdelab/common/function.hh>
#include <dune/pdelab/boilerplate/pdelab.hh>
#include <dune/udg/mc33triangulation.hh>
#include <dune/udg/mc33triangulation/domaindefinition.hh>
#include <dune/udg/mc33triangulation/domainrefinement.hh>
#include <dune/udg/io/refinedvtkwriter.hh>
#include <dune/udg/io/vtkfunction.hh>

#include <dune/udg/test/common/levelsets.hh>
#include <dune/udg/test/common/domains.hh>

template <class ctype,int dim>
void runTest(const Dune::FieldVector<ctype,dim>& height, double phi,
             unsigned int refinements, double minEdgeEps)
{
  // create grid
  const Dune::GeometryType::BasicType basicType = Dune::GeometryType::cube;
  typedef Dune::YaspGrid<dim> G;
  Dune::PDELab::StructuredGrid<G> grid(basicType,1);
  grid->globalRefine(refinements);
  typedef typename G::LeafGridView GV;
  GV gv = grid->leafGridView();

  // create cross domain
  using Range = ctype;
  using Domain = Dune::FieldVector<ctype, dim>;
  using LevelSet = Dune::Functions::GridViewFunction<Range(Domain), GV>;
  std::vector<LevelSet> levelSets;
  Dune::UDGTest::constructRotatedCrossLevelSets(gv, height, phi,
                                                std::back_inserter(levelSets));
  Dune::UDGTest::CrossDomainFactory<GV,LevelSet> factory;
  typedef Dune::UDG::DomainDefinition<GV> DD;
  Dune::shared_ptr<DD> dd = factory.construct(gv, levelSets);

  // create domain refinement
  typedef Dune::UDG::DomainRefinement<DD> DR;
  const bool linearizeParts = false;
  DR dr(*dd, minEdgeEps, 0, linearizeParts);

  // loop over codim 0 elements
  for (typename GV::template Codim<0>::Iterator eit = gv.template begin<0>();
       eit != gv.template end<0>(); ++eit)
  {
    // create domain refinement parts
    typename DR::DomainPartList dpl;
    typename DR::DomainTag tag;
    dr.createDomainRefinement(*eit, dpl, tag);
    // count number of part indices
    std::map<typename DR::Index, unsigned int> counts;
    for (typename DR::DomainPartList::const_iterator dpit = dpl.begin();
         dpit != dpl.end(); ++dpit)
    {
      ++counts[dpit->index];
    }
    // check if every part index occurs only once
    for (typename std::map<typename DR::Index, unsigned int>::const_iterator mit = counts.begin();
         mit != counts.end(); ++mit)
    {
      EXPECT_EQ(mit->second, 1) << "Test [height = " << height << " phi = "
                                << phi << "]\n" << "partIndex " << mit->first
                                << " of entity " << gv.indexSet().index(*eit)
                                << " is not unique (" << mit->second
                                << " parts with the same index";
    }
  }
}

/**
 * \brief The domain refinements assures that the index of each refinement
 *        part is unique in each level set entity. So lets test it.
 */
TEST(RefinementTest, UniquePartIndex)
{
  const int dim = 3;
  const double minEdgeEps = 1e-2;
  double testHeights[][3] = {{0.37,0.52,0.51}, {0.3,0.4,0.5}, {0.4,0.5,0.51}};
  double testPhis[] =       {M_PI/6.0        , M_PI/6.0     , M_PI/6.0};
  unsigned int testRefs[] = {3               , 3            , 3};
  for (int i = 0; i<sizeof(testPhis)/sizeof(testPhis[0]); ++i)
  {
    Dune::FieldVector<double,dim> height;
    std::copy(testHeights[i],testHeights[i]+dim, height.begin());
    runTest(height, testPhis[i],testRefs[i], minEdgeEps);
  }
}

int main(int argc, char** argv)
{
  Dune::MPIHelper::instance(argc,argv);

  testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}
