#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <dune/common/parallel/mpihelper.hh>
#include <dune/common/gmpfield.hh>
#include <dune/grid/common/rangegenerators.hh>
#include <dune/grid/yaspgrid.hh>
#include <dune/pdelab/boilerplate/pdelab.hh>
#include <dune/pdelab/ordering/chunkedblockordering.hh>
#include <dune/udg/pdelab/finiteelementmap.hh>
#include <dune/udg/pdelab/subtriangulation.hh>
#include <dune/udg/pdelab/multiphaseoperator.hh>
#include <dune/udg/simpletpmctriangulation.hh>
#include <dune/udg/test/common/levelsets.hh>
#include <dune/udg/io/refinedvtkwriter.hh>
#include <dune/udg/vtriangulation.hh>

int main(int argc, char** argv)
{
  Dune::MPIHelper::instance(argc, argv);
  unsigned int N = 10;
  if (argc > 1) {
    N = std::stoi(argv[1]);
  }
  try {
    const int dim = 3;
    const int compartments = 4;
    const int degree = 1;
    using Grid = Dune::YaspGrid<dim>;
    using ctype = Grid::ctype;
    using GV = Grid::LevelGridView;
    using LGV = Grid::LeafGridView;
    using STT = Dune::UDG::SimpleTpmcTriangulation<GV, LGV>;

    Dune::Timer timer;
    // create fundamental and level set mesh
    Dune::PDELab::StructuredGrid<Grid> grid(Dune::GeometryType::cube, N);
    grid->globalRefine(1);
    GV gv = grid->levelGridView(0);
    LGV lgv = grid->leafGridView();
    std::cout << "constructed structured grid: " << timer.elapsed() << "s\n";
    timer.reset();

    // create domains
    Dune::UDG::DomainConfiguration<GV, LGV> domainConfiguration;
    domainConfiguration.addDomain(Dune::UDG::Domain(0, {"ii"}));
    domainConfiguration.addDomain(Dune::UDG::Domain(1, {"ei"}));
    domainConfiguration.addDomain(Dune::UDG::Domain(2, {"ee"}));
    domainConfiguration.addDomain(Dune::UDG::Domain(3, {"ie"}));

    // create level sets
    auto sphere1 = [](const Dune::FieldVector<ctype, dim>& x) -> ctype {
      auto t = x;
      t -= Dune::FieldVector<ctype, dim>(0.5);
      t[0] += 0.1;
      return t.two_norm() - 0.275;
    };
    auto sphere2 = [](const Dune::FieldVector<ctype, dim>& x) -> ctype {
      auto t = x;
      t -= Dune::FieldVector<ctype, dim>(0.5);
      t[0] -= 0.1;
      return t.two_norm() - 0.3;
    };
    auto lset1 = Dune::Functions::makeAnalyticGridViewFunction(sphere1, lgv);
    auto lset2 = Dune::Functions::makeAnalyticGridViewFunction(sphere2, lgv);
    domainConfiguration.addInterface(Dune::UDG::Interface<LGV>(0, lset1));
    domainConfiguration.addInterface(Dune::UDG::Interface<LGV>(1, lset2));
    std::cout << "constructed domainConfiguration: " << timer.elapsed() << "s\n";
    timer.reset();

    // construct subtriangulation
    STT triangulation(gv, lgv, domainConfiguration);
    std::cout << "constructed subTriangulation: " << timer.elapsed() << "s\n";
    timer.reset();

    // output subtriangulation
    Dune::RefinedVtkWriter<GV, STT, double> vtkVolume(gv, triangulation);
    typedef Dune::UDG::DomainIndexUnfittedVTKGridFunction<GV> DVGF;
    vtkVolume.addVertexData(new DVGF(gv, true));
    vtkVolume.addVertexData(new DVGF(gv, false));
    vtkVolume.write("test_simpletpmctriangulation_volume", Dune::VTK::appendedraw, 255,
                    Dune::writeVolume);
    std::cout << "wrote volume vtk output " << timer.elapsed() << "s\n";
    timer.reset();

    Dune::RefinedVtkWriter<GV, STT, double> vtkFaces(gv, triangulation);
    vtkFaces.addVertexData(new DVGF(gv, true));
    vtkFaces.addVertexData(new DVGF(gv, false));
    vtkFaces.write("test_simpletpmctriangulation_faces", Dune::VTK::appendedraw, 255,
                   Dune::writeFaces);
    std::cout << "wrote faces vtk output " << timer.elapsed() << "s\n";
    timer.reset();
  } catch (Dune::Exception& ex) {
    std::cerr << ex.what() << "\n";
    return -1;
  }
}
