// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#include "udgassembler.hh"
#include "udgparallel.hh"
#include "vtriangulation.hh"

#include <cmath>
#include <cassert>

namespace Dune {

  template < class G, class T, class L, class F, class M >
  inline
  const typename UDGAssembler<G,T,L,F,M>::ShapeFunctionSet &
  UDGAssembler<G,T,L,F,M>::getShapeFunctionSet(int component, const EntityPointer & ep) const
  {
    return discretization.getShapeFunctionSet(component, ep->geometry().type());
  }

  template < class G, class T, class L, class F, class M >
  void
  UDGAssembler<G,T,L,F,M>::assemble(int terms, bool cleanIsolatedElements)
  {
    /*
     * make sure that the sparsity pattern was set up by checking that A
     * has the correct size
     */
    assert(A.N() == static_cast<typename Matrix::size_type>(mapper.size()));

  #warning HACK
    std::vector< std::vector<int> > couplings(cleanIsolatedElements ? mapper.size() : 0);
    std::vector<int> todo;

    // loop over all elementes e
    ElementIterator it = gridview.template begin<0>();
    ElementIterator endit = gridview.template end<0>();
    for (; it != endit; ++it) {
      EntityPointer e = it;
      int eid = mapper.map(*e);
      if(Grid::dimension==3)
        if (!((eid+1) % (A.N() / 10 + 1)))
        {
          if (gridview.comm().rank() == 0)
            std::cout<<100*(eid+1)/A.N()<<" Percent Completed"<<std::endl;
        }
      // --- Volume terms ---
      // create geometries
      std::list<EntityPart> geometries;
      geometries.clear();
      triangulation.create_geometries(e, geometries);
      // sum over all quadrature points in all geometries
      typename std::list<EntityPart>::iterator geomend = geometries.end();
      typename std::list<EntityPart>::iterator geomit = geometries.begin();
      // use trivial problem if element is completely inside the boundary
      if (geometries.empty())
        continue;
      // ignore cell which are too small
      Dune::FieldVector<ctype, dim> zero(0.0);
      geomit = geometries.begin();
      if(terms&volumeTerm) {
        ElementParameters * ep = discretization.getBoundaryConditions().elementParameters(it);
#ifndef ASSEMBLE_FULL_VOLUME
        for (; geomit != geomend; ++geomit)
        {
          discretization.assembleVolumeTerm
            (*ep, e, eid, triangulation.BBox(e), geomit->rectLocal(),
            A[eid][eid], b[eid]);
        }
#else
        EntityPart se(triangulation.BBox(e));
        discretization.assembleVolumeTerm
          (*ep, e, eid, triangulation.BBox(e), se.rectLocal(),
          A[eid][eid], b[eid]);
#endif
        delete ep;
      }
      // === Edges / Boundaries ===

      // --- Edgeterms ---
      // create edges
      std::list<IntersectionPart> edges;
      triangulation.create_edges(e, edges);
      typename std::list<IntersectionPart>::const_iterator eend =
        edges.end();
      typename std::list<IntersectionPart>::const_iterator eit =
        edges.begin();
      // calculate edge norms
      std::map<int,ctype> edgenorm;
      for (; eit != eend; ++eit)
      {
        const EntityPointer & f = eit->outside();
        int fid = mapper.map(*f);
        Dune::QuadratureRule<ctype,dim-1> qr =
          Dune::QuadratureRules<ctype,dim-1>::rule(eit->intersectionGlobal().type(), dim-1);
        typename
        Dune::QuadratureRule<ctype,dim-1>::const_iterator
        qp = qr.begin();
        typename
        Dune::QuadratureRule<ctype,dim-1>::const_iterator
        qend = qr.end();
        for (; qp!=qend; ++qp)
          edgenorm[fid] += qp->weight() * eit->intersectionGlobal().integrationElement(qp->position());
        assert(edgenorm[fid] > 0.0);
      }
      // sum over all quadrature points on all edges
      for (eit = edges.begin(); eit != eend; ++eit)
      {
        const EntityPointer & f = eit->outside();
        int fid = mapper.map(*f);
        if(cleanIsolatedElements) couplings[eid].push_back(fid);
        if(terms&faceTerm) {
          BoundaryParameters * fp = discretization.getBoundaryConditions().boundaryParameters(*eit);
          discretization.assembleFaceTerm
            (*fp, eid, triangulation.BBox(e), fid, triangulation.BBox(f),
            *eit, edgenorm[fid],
            A[eid][eid], A[eid][fid], A[fid][eid], b[eid]);
          delete fp;
        }
      }
      // --- Boundary terms ---
      std::list<IntersectionPart> boundaries;
      triangulation.create_boundaries(e, boundaries);
      typename std::list<IntersectionPart>::const_iterator bend =
        boundaries.end();
      typename std::list<IntersectionPart>::const_iterator bit =
        boundaries.begin();
      // calculate edge norms
      // $\text{norm} := |\Gamma_{EF}|$
      std::map<int,ctype> boundarynorm;
      for (; bit != bend; ++bit)
      {
        int bid = bit->boundarySegmentIndex();
        Dune::QuadratureRule<ctype,dim-1> qr =
          Dune::QuadratureRules<ctype,dim-1>::rule(bit->intersectionGlobal().type(), dim-1);
        typename
        Dune::QuadratureRule<ctype,dim-1>::const_iterator
        qp = qr.begin();
        typename
        Dune::QuadratureRule<ctype,dim-1>::const_iterator
        qend = qr.end();
        for (; qp!=qend; ++qp)
          boundarynorm[bid] += qp->weight() * bit->intersectionGlobal().integrationElement(qp->position());
        assert(boundarynorm[bid] > 0.0);
      }
      for (bit = boundaries.begin(); bit != bend; ++bit)
      {
        // skip processor boundary
        if(bit->boundarySegmentIndex() == BoundaryId::processor)
        {
        #warning HACK
          // identify seeds
          if(cleanIsolatedElements)
          {
            todo.push_back(eid);
          }
          continue;
        }
        if(terms&boundaryTerm) {
          int bid = bit->boundarySegmentIndex();
          BoundaryParameters * bp = discretization.getBoundaryConditions().boundaryParameters(*bit);
          discretization.assembleBoundaryTerm
            (*bp, eid, triangulation.BBox(e), *bit, boundarynorm[bid],
            A[eid][eid], b[eid]);
        #warning HACK
          // identify seeds
          if(cleanIsolatedElements && bp->isDirichlet())
          {
            todo.push_back(eid);
          }
          delete bp;
        }
      }

    }

    gridview.comm().barrier();

    if(!(terms&faceTerm && terms&boundaryTerm))
      return;
    if(!cleanIsolatedElements)
      return;

    /*
     *
     * restructure matrix!
     *
     * insert trivial problems for isolated elements
     *
     */
  #warning HACK
    std::vector<bool> inside(mapper.size());
    if (verbose >= 1 && gridview.comm().rank() == 0)
      std::cout << "init inside " << inside.size() << "\n";
    for (size_t i=0; i<inside.size(); ++i) inside[i] = false;
    size_t count = 0;

    // recusively identify our domain
    while (! todo.empty())
    {
      // get entry id
      int i = todo.back();
      todo.pop_back();
      // did we already see this entry?
      if (inside[i] == true) continue;
      // register current entry
      inside[i] = true;
      count++;
      // search all couplings and put them on the stack
      typename std::vector<int>::const_iterator it=couplings[i].begin();
      for (; it!=couplings[i].end(); ++it)
      {
        int j=*it;
        if (i != j && inside[j] == false && couplings[j].size() != 0)
        {
          todo.push_back(j);
        }
      }
    }

    size_t countOuter = inside.size() - count;

    if (verbose >= 1)
    {
      std::cout << "identified domain of size " << count << "\n";
      std::cout << "remaining  domain of size " << countOuter << "\n";
    }

    size_t maxDOFS = 100000;
    size_t maxDomSize = maxDOFS / LocalStiffness::BlockSize;
    if (verbose >= 1)
    {
      std::cout << "maximum domain size " << maxDomSize << "\n";
    }

#if HAVE_SUPERLU
    size_t domsInner = (count / maxDomSize) + 1;
#ifdef SOLVE_OUTER
    size_t domsOuter = (countOuter / maxDomSize) + 1;
#else
    size_t domsOuter = 0;
#endif

    if (verbose >= 1)
    {
      std::cout << domsInner << " inner subdomains\n";
      std::cout << domsOuter << " outer subdomains\n";
    }

    domains.resize(domsInner + domsOuter);
    for (size_t i=0; i< domains.size(); i++) domains[i].clear();
    if (verbose >= 1)
    {
      std::cout << "init domains vector " << domains.size() << "\n";
    }

    // set up the final sparsity pattern
    size_t dI = 0;
    size_t cI = 0;
#ifdef SOLVE_OUTER
    size_t dO = 0;
    size_t cO = 0;
#endif
    for (size_t i=0; i<inside.size(); ++i)
    {
      if (inside[i])
      {
        assert(dI < domsInner);
        domains[dI].insert(i);
        typename std::vector<int>::const_iterator it=couplings[i].begin();
        for (; it!=couplings[i].end(); ++it)
        {
          size_t j=*it;
          domains[dI].insert(j);
        }
        ++cI;
        if (cI > count / domsInner)
        {
          cI = 0;
          dI++;
        }
        // save memory
        couplings[i].clear();
      }
#ifdef SOLVE_OUTER
      else
      {
        assert(dO < domsOuter);
        domains[dO + domsInner].insert(i);
        ++cO;
        if (cO > countOuter / domsOuter)
        {
          cO = 0;
          dO++;
        }
        couplings[i].clear();
      }
#endif
    }

    if (verbose >= 1)
    {
      for (size_t i=0; i< domains.size(); i++)
      {
        std::cout << ((i<domsInner) ? "inner " : "outer ")
                  << "subdomain " << i << ": size "
                  << domains[i].size() << "\n";
        if (verbose >= 3)
        {
          typename subdomain_type::const_iterator it=domains[i].begin();
          for (; it!=domains[i].end(); ++it)
          {
            std::cout << *it << " ";
          }
          std::cout << "\n\n";
        }
      }
    }
#endif

    // compute ghost cells
    {
      ElementIterator it = gridview.template begin<0>();
      ElementIterator endit = gridview.template end<0>();
      for (; it != endit; ++it)
      {
        bool clean = false;
        if (it->partitionType() == GhostEntity) {
          clean = true;
        }
        else {
          if (it->partitionType() == OverlapEntity) {
            typename GridView::IntersectionIterator iit = gridview.ibegin(*it);
            typename GridView::IntersectionIterator iend = gridview.iend(*it);
            for (; iit != iend; ++iit)
            {
              if (! iit->boundary() && ! iit->neighbor())
                clean = true;
            }
          }
        }
        if (clean) {
          int i = mapper.map(*it);
          inside[i] = false;
        }
      }
    }

    // eliminated entries, which are not part of the domain
    if (verbose >= 2)
    {
      std::cout << "remove: ";
    }
    for (size_t i=0; i<inside.size(); ++i)
    {
      if (! inside[i])
      {
        A[i] = 0.0;
        for (int k=0; k < LocalStiffness::BlockSize; ++k)
        {
          A[i][i][k][k] = 1.0;
        }
        if (verbose >= 2)
        {
          std::cout << i << ", ";
        }
        b[i] = 0.0;
      }
    }
    if (verbose >= 2)
    {
      std::cout << "\n\n";
    }

    UDG::VectorExchange<Mapper, Vector> dh(mapper, b);
    gridview.communicate(dh, Dune::InteriorBorder_All_Interface, Dune::ForwardCommunication);

    //  discretization.printCacheStatistics();
  }

} // end namespace Dune
