// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:

#ifndef DG_ASSEMBLER_HH
#define DG_ASSEMBLER_HH

#include <dune/common/fvector.hh>
#include <dune/common/fmatrix.hh>

#include <dune/grid/common/mcmgmapper.hh>
#include <dune/grid/common/quadraturerules.hh>

#include <dune/istl/bvector.hh>
#include <dune/istl/vbvector.hh>
#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/overlappingschwarz.hh>

#include <dune/udg/disc/functions.hh>

#include "sparsitysetup.hh"
#include <dune/udg/common/subentity.hh>
#include <dune/udg/common/subintersection.hh>

#include <dune/udg/localstiffness.hh>

namespace Dune {

  //! The terms to actually assemble
  /*
   * These can be |-ed together
   */
  enum {
    //! assemble the volume term
    volumeTerm   = 1,
    //! assemble the interior faces term
    faceTerm     = 2,
    //! assemble the exterior faces term
    boundaryTerm = 4
  };

  template < class G, class T, class L, class F,
      class M = MultipleCodimMultipleGeomTypeMapper< G, CellLayout > >
  class UDGAssembler
  {
  public:
    typedef G GridView;
    typedef typename G::Grid Grid;
    typedef T Triangulation;
    typedef typename LocalStiffnessTypedef<L>::type LocalStiffness;
    typedef F DGFunction;
    typedef M Mapper;
    enum
    { dimension = Grid::dimension /* !<dimension of the grid we are using */  };
    enum
    { dimensionworld = Grid::dimensionworld /* !<dimension of the world */  };
  private:
    // dimensions
    enum
    { dim = Grid::dimension /* !< dimension of the grid we are using */  };
    enum
    { dimw = Grid::dimensionworld /* !< dimension of the world */  };
    //
    typedef typename Grid::ctype ctype;
    // Iterators
    typedef typename GridView::template Codim < 0 >::Iterator ElementIterator;
    typedef typename GridView::template Codim < dim >::Iterator VertexIterator;
    typedef typename GridView::IntersectionIterator IntersectionIterator;
    typedef typename Grid::template Codim < 0 >::Entity Entity;
    typedef typename Grid::template Codim < dim >::Entity Vertex;

    // Geometries
    typedef Dune::Cuboid < dim, ctype > Geometry;
    typedef ::SubTriangulationGeometry < dim, dim, ctype > LocalGeometry;
    typedef ::SubTriangulationGeometry < dim - 1, dim, ctype > FaceGeometry;
    typedef ::SubTriangulationGeometry < dim - 1, dim, ctype > FaceLocalGeometry;
    typedef typename Triangulation::EntityPart EntityPart;
    typedef typename Triangulation::IntersectionPart IntersectionPart;
    // Typedefs needed to interface the Form class
    typedef typename LocalStiffness::matvec_t matvec_t;
    typedef typename LocalStiffness::Gradient Gradient;
    typedef typename LocalStiffness::FaceCoord FaceCoord;
    typedef typename LocalStiffness::LocalVectorBlock LocalVectorBlock;
    typedef typename LocalStiffness::LocalMatrixBlock LocalMatrixBlock;
    typedef typename LocalStiffness::BoundaryParameters BoundaryParameters;
    typedef typename LocalStiffness::ElementParameters ElementParameters;
    enum  { Components = LocalStiffness::Components };
    // Further typedefs
    typedef typename GridView::IndexSet IndexSet;
    typedef MultipleCodimMultipleGeomTypeMapper< GridView, CellLayout > MCMGMapper;
    typedef FunctionBase < ctype, matvec_t, dim, Components > Function;
  public:
    typedef BlockVector < LocalVectorBlock > Vector;
    typedef BCRSMatrix < LocalMatrixBlock > Matrix;
    // the contructor needs a grid and a triangulation

    // TODO: remove
    UDGAssembler (const GridView & g, const Triangulation & t,
                  const LocalStiffness & ls, Matrix & _A, Vector & _b,
                  DGFunction & _x, int v=1) :
      gridview (g), triangulation (t), discretization (ls),
      mcmgmapper(gridview),
      mapper(mcmgmapper),
      A(_A), b(_b),
      verbose(v) {};

    // TODO: remove
    UDGAssembler (const GridView & g, const Triangulation & t,
                  const LocalStiffness & ls, Matrix & _A, Vector & _b,
                  DGFunction & _x, const Mapper & _m, int v=1) :
      gridview (g), triangulation (t), discretization (ls),
      mcmgmapper(gridview),
      mapper(_m),
      A(_A), b(_b),
      verbose(v) {};

    UDGAssembler (const GridView & g, const Triangulation & t,
                  const LocalStiffness & ls, Matrix & _A, Vector & _b,
                  int v=1) :
      gridview (g), triangulation (t), discretization (ls),
      mcmgmapper(gridview),
      mapper(mcmgmapper),
      A(_A), b(_b),
      verbose(v) {};

    UDGAssembler (const GridView & g, const Triangulation & t,
                  const LocalStiffness & ls, Matrix & _A, Vector & _b,
                  const Mapper & _m, int v=1) :
      gridview (g), triangulation (t), discretization (ls),
      mcmgmapper(gridview),
      mapper(_m),
      A(_A), b(_b),
      verbose(v) {};

    // global assembly and solving
    void assemble (int terms = volumeTerm|faceTerm|boundaryTerm,
                   bool cleanIsolatedElements = true);

#if HAVE_SUPERLU
    typedef SeqOverlappingSchwarz<Matrix,Vector> Schwarz;
    typedef typename Schwarz::subdomain_type subdomain_type;
    typedef typename Schwarz::subdomain_vector subdomain_vector;
#else
    typedef int subdomain_vector;
#endif
    subdomain_vector domains;
  private:
    typedef typename LocalStiffness::ShapeFunctionSet ShapeFunctionSet;
    inline const ShapeFunctionSet & getShapeFunctionSet (int component, const Entity & ep) const;
    const GridView & gridview;
    const Triangulation & triangulation;
    const LocalStiffness & discretization;
    MCMGMapper mcmgmapper;
  public:
    const Mapper & mapper;
  private:
    Matrix & A;
    Vector & b;
    int verbose;
  };

} // end namespace Dune

#include "udgassembler.cc"

#endif // DG_ASSEMBLER_HH
