// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef UDGPARALLEL_HH
#define UDGPARALLEL_HH

#if HAVE_MPI
#include <mpi.h>
#endif

#include <dune/grid/common/datahandleif.hh>
#include <dune/grid/io/file/vtk/vtkwriter.hh>
#include <dune/istl/preconditioners.hh>
#include <dune/istl/scalarproducts.hh>
#include <dune/istl/operators.hh>

namespace Dune {
  namespace UDG {

    template<class GridView>
    class VTKProcessId :
      public Dune::VTKWriter<GridView>::VTKFunction
    {
      typedef typename GridView::Grid GridImp;
      enum { dim=GridImp::dimension };
      enum { dimw=GridImp::dimensionworld };
      typedef typename GridImp::ctype ctype;
      typedef typename Dune::VTKWriter<GridView>::VTKFunction VTKFunction;
      typedef typename GridImp::Traits::template Codim<0>::Entity Entity;
    public:
      VTKProcessId() { }
      virtual ~VTKProcessId() {}
      //! return number of components
      virtual int ncomps () const
      {
        return 1;
      }

      //! evaluate
      virtual double evaluate (int comp, const Entity& e, const Dune::FieldVector<ctype,dim>& xi) const
      {
#if HAVE_MPI
        int rank;
        MPI_Comm_rank(MPI_COMM_WORLD, &rank);
        return rank;
#else
        return 0;
#endif
      }

      //! get name
      virtual std::string name () const
      {
        return "ProcessId";
      }
    };

    // A DataHandle class to exchange entries of a vector
    template<class M, class V> // mapper type and vector type
    class VectorExchange
      : public Dune::CommDataHandleIF<VectorExchange<M,V>, typename V::value_type>
    {
    public:
      //! export type of data for message buffer
      typedef typename V::value_type DataType;

      //! returns true if data for this codim should be communicated
      bool contains (int dim, int codim) const
      {
        return (codim==0);
      }

      //! returns true if size per entity of given dim and codim is a constant
      bool fixedsize (int dim, int codim) const
      {
        return true;
      }

      /*! how many objects of type DataType have to be sent for a given entity

         Note: Only the sender side needs to know this size.
       */
      template<class EntityType>
      size_t size (EntityType& e) const
      {
        return 1;
      }

      //! pack data from user to message buffer
      template<class MessageBuffer, class EntityType>
      void gather (MessageBuffer& buff, const EntityType& e) const
      {
        buff.write(c[mapper.map(e)]);
      }

      /*! unpack data from message buffer to user

         n is the number of objects sent by the sender
       */
      template<class MessageBuffer, class EntityType>
      void scatter (MessageBuffer& buff, const EntityType& e, size_t n)
      {
        DataType x;
        buff.read(x);
        c[mapper.map(e)]=x;
      }

      //! constructor
      VectorExchange (const M& mapper_, V& c_)
        : mapper(mapper_), c(c_)
      {}

    private:
      const M& mapper;
      V& c;
    };

    // A DataHandle class to sum-up entries of a vector
    template<class M, class V> // mapper type and vector type
    class VectorAdd
      : public Dune::CommDataHandleIF<VectorAdd<M,V>, typename V::value_type>
    {
    public:
      //! export type of data for message buffer
      typedef typename V::value_type DataType;

      //! returns true if data for this codim should be communicated
      bool contains (int dim, int codim) const
      {
        return (codim==0);
      }

      //! returns true if size per entity of given dim and codim is a constant
      bool fixedsize (int dim, int codim) const
      {
        return true;
      }

      /*! how many objects of type DataType have to be sent for a given entity

         Note: Only the sender side needs to know this size.
       */
      template<class EntityType>
      size_t size (EntityType& e) const
      {
        return 1;
      }

      //! pack data from user to message buffer
      template<class MessageBuffer, class EntityType>
      void gather (MessageBuffer& buff, const EntityType& e) const
      {
        buff.write(c[mapper.map(e)]);
      }

      /*! unpack data from message buffer to user

         n is the number of objects sent by the sender
       */
      template<class MessageBuffer, class EntityType>
      void scatter (MessageBuffer& buff, const EntityType& e, size_t n)
      {
        DataType x;
        buff.read(x);
        c[mapper.map(e)]+=x;
      }

      //! constructor
      VectorAdd (const M& mapper_, V& c_)
        : mapper(mapper_), c(c_)
      {}

    private:
      const M& mapper;
      V& c;
    };

    // operator that only call seq operator
    template<class GV, class MP, class M, class X, class Y>
    class OverlappingOperator
      : public Dune::AssembledLinearOperator<M,X,Y>
    {
      std::vector<int> ghost;
    public:
      //! export types
      typedef M matrix_type;
      typedef X domain_type;
      typedef Y range_type;
      typedef typename X::field_type field_type;

      //redefine the category, that is the only difference
      enum {category=Dune::SolverCategory::overlapping};

      OverlappingOperator (const GV& gv, const MP& m, const M& A)
        : gridview(gv), mapper(m), _A_(A)
      {
        // compute ghost cells
        typename GV::template Codim<0>::Iterator it = gridview.template begin<0>();
        typename GV::template Codim<0>::Iterator endit = gridview.template end<0>();
        for (; it != endit; ++it)
        {
          bool clean = false;
          if (it->partitionType() == GhostEntity) {
            clean = true;
          }
          else {
            if (it->partitionType() == OverlapEntity) {
              typename GV::IntersectionIterator iit = gridview.ibegin(*it);
              typename GV::IntersectionIterator iend = gridview.iend(*it);
              for (; iit != iend; ++iit)
              {
                if (! iit->boundary() && ! iit->neighbor())
                  clean = true;
              }
            }
          }
          if (clean) {
            int i = mapper.map(*it);
            ghost.push_back(i);
          }
        }
        std::cout << "ghosts: " << ghost.size() << std::endl;
      }

      //! apply operator to x:  \f$ y = A(x) \f$
      virtual void apply (const X& x, Y& y) const
      {
        Y& y_(y);
        _A_.mv(x,y);
        for (size_t i=0; i<ghost.size(); ++i)
          y[ghost[i]] = y_[ghost[i]];
      }

      //! apply operator to x, scale and add:  \f$ y = y + \alpha A(x) \f$
      virtual void applyscaleadd (field_type alpha, const X& x, Y& y) const
      {
        Y& y_(y);
        _A_.usmv(alpha,x,y);
        for (size_t i=0; i<ghost.size(); ++i)
          y[ghost[i]] =  y_[ghost[i]];
      }

      //! get matrix via *
      virtual const M& getmat () const
      {
        return _A_;
      }

    private:
      const GV& gridview;
      const MP& mapper;
      const M& _A_;
    };

    // scalar product assuming unique partitioning of cells for parallelization
    template<class GV, class M, class X>
    class OverlappingScalarProduct
      : public Dune::ScalarProduct<X>
    {
    public:
      //! export types
      typedef X domain_type;
      typedef typename X::field_type field_type;

      //! define the category
      enum {category=Dune::SolverCategory::overlapping};

      /*! \brief Constructor needs to know the grid function space
       */
      OverlappingScalarProduct (const GV& gv, const M& m)
        : gridview(gv), mapper(m)
      {}


      /*! \brief Dot product of two vectors.
         It is assumed that the vectors are consistent on the interior+border
         partition.
       */
      virtual field_type dot (const X& x, const X& y)
      {
        // do local scalar product on unique partition
        field_type sum = 0;
        // loop over owner entities
        typename GV::template Codim<0>::template Partition<Interior_Partition>::Iterator
        it = gridview.template begin<0, Interior_Partition>();
        typename GV::template Codim<0>::template Partition<Interior_Partition>::Iterator
        endit = gridview.template end<0, Interior_Partition>();
        for (; it != endit; ++it)
        {
          int i = mapper.map(*it);
          // sum up defect
          sum += x[i]*y[i];
        }

        // compute global sum
        sum = gridview.comm().sum(sum);
        return sum;
      }

      /*! \brief Norm of a right-hand side vector.
         The vector must be consistent on the interior+border partition
       */
      virtual double norm (const X& x)
      {
        return sqrt(static_cast<double>(this->dot(x,x)));
      }

    private:
      const GV& gridview;
      const M& mapper;
    };

    // wrapped sequential preconditioner
    template<class GV, class M, class P>
    class OverlappingWrappedPreconditioner
      : public Dune::Preconditioner<typename P::domain_type,typename P::range_type>
    {
    public:
      //! \brief The domain type of the preconditioner.
      typedef typename P::domain_type domain_type;
      //! \brief The range type of the preconditioner.
      typedef typename P::range_type range_type;

      // define the category
      enum {
        //! \brief The category the preconditioner is part of.
        category=Dune::SolverCategory::overlapping
      };

      //! Constructor.
      OverlappingWrappedPreconditioner (const GV& gv, const M& m, P& prec_)
        : gridview(gv), mapper(m), prec(prec_), w(0.9/(1<<GV::dimension))
      {
        // compute ghost cells
        typename GV::template Codim<0>::Iterator it = gridview.template begin<0>();
        typename GV::template Codim<0>::Iterator endit = gridview.template end<0>();
        for (; it != endit; ++it)
        {
          bool clean = false;
          if (it->partitionType() == GhostEntity) {
            clean = true;
          }
          else {
            if (it->partitionType() == OverlapEntity) {
              typename GV::IntersectionIterator iit = gridview.ibegin(*it);
              typename GV::IntersectionIterator iend = gridview.iend(*it);
              for (; iit != iend; ++iit)
              {
                if (! iit->boundary() && ! iit->neighbor())
                  clean = true;
              }
            }
          }
          if (clean) {
            int i = mapper.map(*it);
            ghost.push_back(i);
          }
        }
        std::cout << "ghosts: " << ghost.size() << std::endl;
      }

      /*!
         \brief Prepare the preconditioner.

         \copydoc Preconditioner::pre(domain_type&,range_type&)
       */
      virtual void pre (domain_type& x, range_type& b)
      {
        prec.pre(x,b);
      }

      /*!
         \brief Apply the precondioner.

         \copydoc Preconditioner::apply(domain_type&,const range_type&)
       */
      virtual void apply (domain_type& v, const range_type& d)
      {
        range_type d2(d);
        // set defect for "ghost" cells to 0
        for (size_t i=0; i<ghost.size(); ++i)
          // assert(d2[ghost[i]] == typename domain_type::value_type(0.0));
          d2[ghost[i]] = 0;

        // sequentiell precoditioner
        prec.apply(v,d2);
        for (size_t i=0; i<ghost.size(); ++i)
          // assert(v[ghost[i]].two_norm() < 1e-14);
          assert(v[ghost[i]] == typename domain_type::value_type(0.0));

        // global sum of update
        VectorAdd<M, domain_type> add(mapper, v);
        gridview.communicate(add,Dune::All_All_Interface,Dune::ForwardCommunication);

        gridview.comm().barrier();
      }

      /*!
         \brief Clean up.

         \copydoc Preconditioner::post(domain_type&)
       */
      virtual void post (domain_type& x)
      {
        prec.post(x);
      }

    private:
      const GV& gridview;
      const M& mapper;
      P& prec;
      const double w;
      std::vector<int> ghost;
    };

  } // namespace UDG
} // namespace Dune

#endif // UDGPARALLEL_HH
