// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:

#ifndef DUNE_VRTRIANGULATION_HH
#define DUNE_VRTRIANGULATION_HH

#include <dune/grid/common/virtualrefinement.hh>

#include "vtriangulation.hh"
#include "notriangulation.hh"

template <class GridView, class BasicTriangulation>
class VirtualRefinementSubTriangulation {

public:
  typedef typename GridView::Grid Grid;
  typedef typename Grid::ctype ct;
  typedef typename Grid::template Codim<0>::Entity Entity;
  enum  { dim = Grid::dimension };
  enum  { dimw = Grid::dimensionworld };

  typedef Dune::UDG::EntityPart< Grid > EntityPart;
  typedef Dune::UDG::IntersectionPart< Grid > IntersectionPart;
  typedef Dune::GridCuboid< Grid > GridCuboid;

  typedef Dune::VirtualRefinement<dim, ct> Refinement;
  typedef typename Refinement::VertexIterator vIterator;
  typedef typename Refinement::ElementIterator eIterator;
  typedef typename Refinement::VertexIterator::CoordVector Coordinate;
  typedef typename Refinement::ElementIterator::IndexVector IndexVector;

  VirtualRefinementSubTriangulation(const BasicTriangulation & t, int d=2) :
    bt(t), depth(d) {}
  ~VirtualRefinementSubTriangulation() {}

  // enable / disable cube refinement
  virtual
  bool & cubeRefinement() { static bool b; return b; }

  // create Entities
  virtual
  void create_rectangles(const Entity & e,
                         std::list<GridCuboid> & rectangles) const
  {
    DUNE_THROW(Dune::NotImplemented, "VirtualRefinementSubTriangulation::" << __func__);
  }
  virtual
  void create_geometries(const Entity & ep,
                         std::list<EntityPart> & geometries) const
  {
    // dummy bbox
    static Dune::FieldVector<ct, dimw> zero(0.0);
    static Dune::FieldVector<ct, dimw> one(1.0);
    GridCuboid bbox(ep,zero,one);

    // create background geometries
    std::list<EntityPart> basicgeom;
    bt.create_geometries(ep, basicgeom);

    // refine geometries
    typename std::list<EntityPart>::iterator basicend = basicgeom.end();
    typename std::list<EntityPart>::iterator basicit = basicgeom.begin();
    for (; basicit != basicend; ++basicit)
    {
      EntityPart & e = *basicit;

      // create refinement
      static Dune::GeometryType coerceTo(Dune::GeometryType::simplex,dim);
      Refinement &refinement = Dune::buildRefinement<dim, ct>(e.global().type(),coerceTo);

      // calc vertices
      std::vector<Coordinate> vertices;
      vertices.reserve(refinement.nVertices(depth));
      vIterator vEnd = refinement.vEnd(depth);
      for(vIterator vIt = refinement.vBegin(depth); vIt != vEnd; ++vIt)
      {
        vertices.push_back(vIt.coords());
      }

      // calc sub elements
      eIterator eEnd = refinement.eEnd(depth);
      for(eIterator eIt = refinement.eBegin(depth); eIt != eEnd; ++eIt)
      {
        std::vector< Dune::FieldVector<double,dim> > coords;
        IndexVector vertexIds = eIt.vertexIndices();
        for (size_t c=0; c<vertexIds.size(); c++)
        {
          Dune::FieldVector<ct, dim> pos = vertices[vertexIds[c]];
          Dune::FieldVector<ct, dimw> global = e.global().global(pos);
          coords.push_back(global);
        }
        geometries.push_back(createEntityPart(bbox, coords));
      }
    }
  }
  virtual
  void create_edges(const Entity & e,
                    std::list<IntersectionPart> & edges) const
  {
    DUNE_THROW(Dune::NotImplemented, "VirtualRefinementSubTriangulation::" << __func__);
  }
  virtual
  void create_boundaries(const Entity & e,
                         std::list<IntersectionPart> & boundaries) const
  {
    DUNE_THROW(Dune::NotImplemented, "VirtualRefinementSubTriangulation::" << __func__);
  }
  // get the bounding box for this codim<0>::entity
  virtual
  const GridCuboid BBox(const Entity & e) const
  {
    DUNE_THROW(Dune::NotImplemented, "VirtualRefinementSubTriangulation::" << __func__);
  }
  // get the bounding box for this codim<0>::entity
  virtual
  const GridCuboid BBox(const Entity & e) const
  {
    DUNE_THROW(Dune::NotImplemented, "VirtualRefinementSubTriangulation::" << __func__);
  }

private:
  EntityPart createEntityPart(const GridCuboid & bbox, const std::vector< Dune::FieldVector<ct,dimw> > & c) const
  {
    switch (c.size()) {
    case 3 :
      return EntityPart(bbox, c[0], c[1], c[2]);
    case 4 :
      return EntityPart(bbox, c[0], c[1], c[2], c[3]);
    case 5 :
      return EntityPart(bbox, c[0], c[1], c[2], c[3], c[4]);
    case 6 :
      return EntityPart(bbox, c[0], c[1], c[2],
                        c[3], c[4], c[5]);
    case 8 :
      return EntityPart(bbox, c[0], c[1], c[2], c[3],
                        c[4], c[5], c[6], c[7]);
    default :
      DUNE_THROW(Dune::Exception, "Strange Element");
    }
  }
  const BasicTriangulation & bt;
  int depth;
};

#endif
