// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#ifndef DUNE_UDG_VTRIANGULATION_HH
#define DUNE_UDG_VTRIANGULATION_HH

// #include <list>  // unused header

#include <dune/udg/common/entitypart.hh>
#include <dune/udg/common/intersectionpart.hh>
#include <dune/udg/common/cuboid.hh>


/**
 * \brief Interface for sub triangulation classes.
 *
 * \tparam GridViewT Type of fundamental mesh grid view.
 *
 * \todo BBox shall return a reference.
 */
template <class GridViewT>
class VirtualSubTriangulation
{
public:
  typedef GridViewT GridView;

  typedef typename GridView::Grid Grid;
  typedef typename Grid::ctype ctype;
  typedef typename Grid::template Codim<0>::Entity Entity;

  typedef typename Dune::UDG::EntityPartInterface<GridView> EntityPart;
  typedef typename EntityPart::EntityPartList EntityPartList;
  typedef Dune::UDG::IntersectionPartInterface< GridView > IntersectionPart;
  typedef typename IntersectionPart::IntersectionPartList IntersectionPartList;

  typedef Dune::GridCuboid< Grid > BoundingBox;

  virtual
  ~VirtualSubTriangulation() {}

  /**
   * \brief Provide the entity parts for a given fundamental mesh cell.
   */
  virtual
  void create_entity_parts(const Entity & e,
                           EntityPartList & entityparts) const = 0;

  /**
   * \brief Provide the intersection parts for a given fundamental mesh cell.
   */
  virtual
  void create_intersection_parts(const Entity & e,
                                 IntersectionPartList & intersectionparts) const = 0;

  //! Get the bounding box for this codim<0>::entity.
  //! \todo DocMe right.
  virtual
  const BoundingBox BBox
    (const unsigned int domainIndex, const Entity & e) const = 0;

  //! Get the bounding box for this codim<0>::entity.
  //! \todo DocMe right.
  const BoundingBox boundingBox
    (const Entity & e, const unsigned int domainIndex) const
  {
    return BBox(domainIndex,e);
  }

  /**
   * \brief Provide bounding box for a given entity part.
   */
  virtual
  const BoundingBox boundingBox(const EntityPart & ep) const
  {
    Entity father(ep.entity());
    while( ! this->gridView().indexSet().contains(father) )
      father = father.father();
    return this->BBox(ep.domainIndex(),father);
  }

  /**
   * \brief Enable/disable an enforced refinement of fundamental mesh
   *        cells which are entirely filled by a single domain.
   *
   * \note Naturally, setting this flag is inconsequential if the grid
   *       views representing the fundamental mesh / the level set mesh
   *       are the same.
   */
  virtual
  bool & forceRefinement() = 0;

  //! Get additional information.
  //! \todo DocMe right.
  virtual
  bool isInDomain(const Entity & e, const int & domain_index) const = 0;

  /**
   * \brief Computes an edge norm for the domain cell intersection which
   *        contains a given intersection part.
   */
  virtual
  ctype getEdgeNorm(const IntersectionPart & intersection) const = 0;

  /**
   * \brief Constant access to fundamental mesh grid view.
   */
  virtual
  const GridView & gridView() const = 0;

  /**
   * \brief Check whether a given fundamental mesh cell is host cell of
   *        entity parts with a given domain index.
   *
   * \note This function may return false even if some entity parts with the
   *       given domain index are geometrically positioned in the given cell
   *       (which is therefore their fundamental mesh "home cell") but are
   *       merged to - and therefore associated with - another fundamental
   *       mesh cell (their "host cell").
   */
  virtual
  bool isHostCell(const Entity & e, const int & domain_index) const
  {
    return this->isInDomain(e,domain_index) &&
           this->BBox(domain_index,e).entity() == e;
  }
};


/**
 * \brief Interface for multi layer sub triangulation classes.
 *
 * \tparam GridViewT Type of fundamental mesh grid view.
 */
template <class GridViewT>
class MultiLayerVirtualSubTriangulation :
  public VirtualSubTriangulation<GridViewT>
{
public:
  //! Index types
  typedef typename GridViewT::IndexSet::IndexType Index;

  /**
   * \brief Get the layer in which the domain of the given index is defined.
   */
  virtual Index layerOfDomain(const Index domainIndex) const = 0;
};


#endif
