// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#include "config.h"
#include <dune/udg/common/geometry.hh>

/* === Implementations === */

/* --- private --- */

template<int mydim, int cdim, class ct>
void SubTriangulationGeometry<mydim,cdim,ct>::
setupJacobiMatrix(const LocalVector& local) const
{
  JinvT = 0;
  std::array<unsigned int,mydim> order;
  std::fill(order.begin(), order.end(), 0.0);
  for (int l=0; l<mydim; l++)
  {
    order[l] = 1;
    localBasis().partial(order, local, evalBuffer);
    for (int n=0; n<GeometryParamSize<mydim>::higherorder; n++)
    {
      // \frac{ \partial \phi_n }{ \partial \hat{e}_l }
      // std::cout << "d_" << shapeFnkt(n) << " / d_x_" << l << std::endl;
      for (int k=0; k<cdim; k++)
      {
        JinvT[l][k] += c[n][k] * evalBuffer[n];
      }
    }
    order[l] = 0;
  }
}

template<int mydim, int cdim, class ct>
inline
void SubTriangulationGeometry<mydim,cdim,ct>::
setupInverseTransposed() const
{
  if (mydim == cdim)
  {
    JinvT.invert();
  }
}

template<int mydim, int cdim, class ct>
inline
ct SubTriangulationGeometry<mydim,cdim,ct>::
setupDetJ() const
{
  LocalMatrix AtA(0.0);
  for (int m=0; m<cdim; m++)
  {
    for (int k=0; k<mydim; k++)
    {
      for (int l=0; l<mydim; l++)
      {
        AtA[k][l] += JinvT[k][m]*JinvT[l][m];
      }
    }
  }
  return sqrt(std::abs(AtA.determinant()));
}

/* --- public --- */

template<int mydim, int cdim, class ct>
SubTriangulationGeometry<mydim,cdim,ct>::
SubTriangulationGeometry(const Dune::GeometryType & t,
                         const LinearParamMatrix & _c) :
  c(0), _type(t), islinear(true)
{
  evalBuffer.reserve(GeometryParamSize<mydim>::higherorder);
  for (int i=0; i<GeometryParamSize<mydim>::linear; i++)
    for (int j=0; j<cdim; j++)
    {
      c[i][j] = _c[i][j];
      assert(std::isfinite(c[i][j]));
    }
  for (int i=GeometryParamSize<mydim>::linear;
       i<GeometryParamSize<mydim>::higherorder; i++)
    for (int j=0; j<cdim; j++)
      c[i][j] = 0;
  setupJacobiMatrix(LocalVector(0));
  detj = setupDetJ();
  setupInverseTransposed();
}

template<int mydim, int cdim, class ct>
SubTriangulationGeometry<mydim,cdim,ct>::
SubTriangulationGeometry(const Dune::GeometryType & t,
                         const ParamMatrix & _c) :
  c(_c), _type(t), islinear(true), detj(0)
{
  evalBuffer.reserve(GeometryParamSize<mydim>::higherorder);
  for (int i=0; i<GeometryParamSize<mydim>::higherorder; i++)
    for (int j=0; j<cdim; j++)
    {
      assert(std::isfinite(c[i][j]));
    }
  for (int i=GeometryParamSize<mydim>::linear;
       i<GeometryParamSize<mydim>::higherorder; i++)
    for (int j=0; j<cdim; j++)
      if (c[i][j] != 0)
        islinear = false;
  if (islinear == true)
  {
    setupJacobiMatrix(LocalVector(0));
    detj = setupDetJ();
    setupInverseTransposed();
  }
}

template<int mydim, int cdim, class ct>
inline
ct SubTriangulationGeometry<mydim,cdim,ct>::
integrationElement(const LocalVector &local) const
{
  if (isLinear())
  {
    return detj;
  }
  else
  {
    setupJacobiMatrix(local);
    return setupDetJ();
  }
}

template<int mydim, int cdim, class ct>
inline
const typename SubTriangulationGeometry<mydim,cdim,ct>::GlobalVector
SubTriangulationGeometry<mydim,cdim,ct>::
global (const LocalVector& local) const
{
  Dune::FieldVector<ct, cdim> v(0);
  localBasis().evaluateFunction(local, evalBuffer);
  for (int i=0; i<GeometryParamSize<mydim>::higherorder; i++)
  {
    for (int j=0; j<cdim; j++)
    {
      v[j] += c[i][j] * evalBuffer[i];
    }
  }
  return v;
}

template<int mydim, int cdim, class ct>
inline
const typename SubTriangulationGeometry<mydim,cdim,ct>::JacobianMatrix &
SubTriangulationGeometry<mydim,cdim,ct>::
jacobianInverseTransposed (const LocalVector & l) const
{
  // compile time assertion
  static_assert(cdim == mydim,
                     "geometry dimension must be equal to coordinate dimension for jacobianInverseTransposed");

  if (!isLinear())
  {
    setupJacobiMatrix(l);
    setupInverseTransposed();
  }
  return JinvT;
}

/* === instantiating 2D objects === */

template
SubTriangulationGeometry<1,2,double>::
SubTriangulationGeometry(const Dune::GeometryType & t,
                         const ParamMatrix & _c);
template
SubTriangulationGeometry<1,2,double>::
SubTriangulationGeometry(const Dune::GeometryType & t,
                         const LinearParamMatrix & _c);
template
double SubTriangulationGeometry<1,2,double>::
integrationElement(const LocalVector &local) const;
template
const SubTriangulationGeometry<1,2,double>::GlobalVector
SubTriangulationGeometry<1,2,double>::
global (const LocalVector &local) const;

template
SubTriangulationGeometry<2,2,double>::
SubTriangulationGeometry(const Dune::GeometryType & t,
                         const ParamMatrix & _c);
template
SubTriangulationGeometry<2,2,double>::
SubTriangulationGeometry(const Dune::GeometryType & t,
                         const LinearParamMatrix & _c);
template
double SubTriangulationGeometry<2,2,double>::
integrationElement(const LocalVector &local) const;
template
const SubTriangulationGeometry<2,2,double>::JacobianMatrix &
SubTriangulationGeometry<2,2,double>::
jacobianInverseTransposed (const LocalVector & local) const;
template
const SubTriangulationGeometry<2,2,double>::GlobalVector
SubTriangulationGeometry<2,2,double>::
global (const LocalVector &local) const;

/* === instantiating 3D objects === */

template
SubTriangulationGeometry<2,3,double>::
SubTriangulationGeometry(const Dune::GeometryType & t,
                         const ParamMatrix & _c);
template
SubTriangulationGeometry<2,3,double>::
SubTriangulationGeometry(const Dune::GeometryType & t,
                         const LinearParamMatrix & _c);
template
double SubTriangulationGeometry<2,3,double>::
integrationElement(const LocalVector &local) const;
template
const SubTriangulationGeometry<2,3,double>::GlobalVector
SubTriangulationGeometry<2,3,double>::
global (const LocalVector &local) const;

template
SubTriangulationGeometry<3,3,double>::
SubTriangulationGeometry(const Dune::GeometryType & t,
                         const ParamMatrix & _c);
template
SubTriangulationGeometry<3,3,double>::
SubTriangulationGeometry(const Dune::GeometryType & t,
                         const LinearParamMatrix & _c);
template
double SubTriangulationGeometry<3,3,double>::
integrationElement(const LocalVector &local) const;
template
const SubTriangulationGeometry<3,3,double>::JacobianMatrix &
SubTriangulationGeometry<3,3,double>::
jacobianInverseTransposed (const LocalVector & local) const;
template
const SubTriangulationGeometry<3,3,double>::GlobalVector
SubTriangulationGeometry<3,3,double>::
global (const LocalVector &local) const;
