// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
#include <config.h>

#include <dune/udg/2dtriangulation/sphere.hh>
#include <dune/common/ios_state.hh>

#include <math.h>

#include <iostream>

template<int dim, class ct>
const Dune::FieldVector< ct, dim >
Dune::UDG::Sphere<dim,ct>::normal (const Dune::FieldVector< ct, dim > & p) const {
  Dune::FieldVector< ct, dim > n;
  n[0] = (c[0] - p[0]);
  n[1] = (c[1] - p[1]);
  ct l = sqrt( sqr(n[0]) + sqr(n[1]) );
  n *= 1.0/l;
  return n;
}

template<int dim, class ct>
void Dune::UDG::Sphere<dim,ct>::polynom(const Dune::FieldVector<ct,dim> & A,
                                        const Dune::FieldVector<ct,dim> & B, int & side,
                                        Dune::FieldVector<ct,3> & a,
                                        Dune::FieldVector<ct,3> & b) const {
  double epsilon = 1e-8 * std::abs(A[1] - B[1]) + std::numeric_limits<ct>::epsilon();

  if (!isInsideInc(A)) {
    std::cout << "=> A = " << A[0] << " " << A[1] << std::endl;
    std::cout << "B = " << B[0] << " " << B[1] << std::endl;
    std::cout << "c = " << c[0] << " " << c[1] << std::endl;
    std::cout << "r = " << r << std::endl;
    double eps = 2*r*std::numeric_limits<ct>::epsilon();
    std::cout << "eps = " << eps << std::endl;
    std::cout << "d = " << sqr(A[0]-c[0]) + sqr(A[1]-c[1]) - sqr(r) << std::endl;
  }
  if (!isInsideInc(B)) {
    std::cout << "A = " << A[0] << " " << A[1] << std::endl;
    std::cout << "=> B = " << B[0] << " " << B[1] << std::endl;
    std::cout << "c = " << c[0] << " " << c[1] << std::endl;
    std::cout << "r = " << r << std::endl;
    double eps = 2*r*std::numeric_limits<ct>::epsilon();
    std::cout << "eps = " << eps << std::endl;
    std::cout << "d = " << sqr(B[0]-c[0]) + sqr(B[1]-c[1]) - sqr(r) << std::endl;
  }
  assert(isInsideInc(A));
  assert(isInsideInc(B));

  ct len = sqrt(sqr(A[0]+B[0]-2*c[0])+sqr(A[1]+B[1]-2*c[1]));
  Dune::FieldVector<ct,dim> D;
  D[0] = c[0] + r*(A[0]+B[0]-2*c[0])/len;
  // if (A[1] >= c[1] && B[1] >= c[1])
  if ( ( std::abs(A[1] - c[1]) < epsilon || A[1] > c[1] )
       &&
       ( std::abs(B[1] - c[1]) < epsilon || B[1] > c[1] ) )
    side = 1;
  else
    side = -1;

  D[1]=c[1]+side*sqrt(sqr(r)-sqr(D[0]-c[0]));

  a[0]=A[0];
  a[1]=-3*A[0]-B[0]+4*D[0];
  a[2]=2*A[0]+2*B[0]-4*D[0];

  b[0]=A[1];
  b[1]=-3*A[1]-B[1]+4*D[1];
  b[2]=2*A[1]+2*B[1]-4*D[1];

  //  debug output
  ct y = b[0] + b[1] * 0.5 + b[2] * sqr(0.7);
  if ((A[1] > B[1] && ( y > A[1] || y < B[1] ))
      ||
      (A[1] < B[1] && ( y < A[1] || y > B[1] )))
  {
    std::cout << "A = " << A[0] << " " << A[1] << std::endl;
    std::cout << "B = " << B[0] << " " << B[1] << std::endl;
    std::cout << "c = " << c[0] << " " << c[1] << std::endl;
    std::cout << "r = " << r << std::endl;

    std::cout << "polynom... "
              << a[0] << " "
              << a[1] << " "
              << a[2] << " "
              << " - "
              << b[0] << " "
              << b[1] << " "
              << b[2] << " "
              << std::endl;
  }

  return;
  // linear
  a[0]=A[0];
  a[1]=B[0]-A[0];
  a[2]=0;

  b[0]=A[1];
  b[1]=B[1]-A[1];
  b[2]=0;
}

template<int dim, class ct>
void Dune::UDG::Sphere<dim,ct>::getIntersections(const Line<dim,ct>& l, std::vector<Dune::FieldVector<ct,dim> >& points) const
{
  Dune::ios_base_all_saver formatsaver(std::cout);
#ifdef DEBUG_2DST
  std::cout.precision(12);
  std::cout << "INTERSECTIONS " << l.c << " -- " << l.c+l.d << " with SP " << c << "/" << r << std::endl;
#endif
  Dune::FieldVector<ct,dim> D = l.d;
  Dune::FieldVector<ct,dim> O = c-l.c;
  ct DO = D*O;
  ct D2 = D*D;
  ct O2 = O*O;
  ct r2 = r*r;
  ct S = DO*DO - D2*O2 + D2*r2;
  if (S < 0.0)
  {
    return;
  }
  if (S == 0.0)
  {
    ct d = DO / D2;
    if (d >= 0.0 && d <= 1.0)
    {
      Dune::FieldVector<ct,dim> t = l.d; t*=d;
      points.push_back(l.c + t);
#ifdef DEBUG_2DST
      std::cout << "found d "<< (l.c+t) << std::endl;
#endif
    }
    return;
  }
  ct d1 = (DO - sqrt(S)) / D2;
  ct d2 = (DO + sqrt(S)) / D2;

  if (d1 >= 0.0 && d1 <= 1.0)
  {
    Dune::FieldVector<ct,dim> t = l.d; t*=d1;
    points.push_back(l.c + t);
#ifdef DEBUG_2DST
    std::cout << "found d1 "<< (l.c+t) << std::endl;
#endif
  }
  if (d2 >= 0.0 && d2 <= 1.0)
  {
    Dune::FieldVector<ct,dim> t = l.d; t*=d2;
    points.push_back(l.c + t);
#ifdef DEBUG_2DST
    std::cout << "found d2 "<< (l.c+t) << std::endl;
#endif
  }
}

template
const Dune::FieldVector<double,2>
Dune::UDG::Sphere<2,double>::normal (const Dune::FieldVector<double,2> & p) const;
template
void Dune::UDG::Sphere<2,double>::polynom(const Dune::FieldVector<double,2> & A,
                                          const Dune::FieldVector<double,2> & B,
                                          int & side,
                                          Dune::FieldVector<double,3> & a,
                                          Dune::FieldVector<double,3> & b) const;
template
void Dune::UDG::Sphere<2,double>::getIntersections(const Line<2,double>& l,
                                                   std::vector<Dune::FieldVector<double,2> >& points) const;
